#include "resource.h"
#include "BCDB_Functions.h"
#include "secKeySN.h"
#include "PKParse.h"

#include "..\Includes\stddlgs.h"

#include "..\CertificateLib\PGPPkts\Packets.h"
#include "..\CertificateLib\PGPPkts\PGPPkts.h"
#include "..\CertificateLib\x509Pkts\base64funcs.h"
#include "..\CertificateLib\hashAlgs\hashAlgs.h"

#include "..\Includes\OpenFiles.h"
#include "..\Includes\FileView.h"
#include "..\Includes\UnicodeFuncs.h"
#include "..\Includes\shell32dll.h"

#include "..\Setups\SetupIncludesT\GetBCPath.h"

#include <process.h>
#include <stdio.h>
#include <io.h>

bool BCDB_GenerateLocalKeyID( cPTB *ptb, HPKA hka, bool secret )
{
	int size = 0;

	int res = secret ?
				ska_generateLocalKeyID( hka, NULL, &size ):
				pka_generateLocalKeyID( hka, NULL, &size );

	if ( res != pkae_bufferSizeTooSmall )
		return false;
	
	if ( ptb->Allocate( size ) == NULL )
		return false;

	res = secret ? 
			ska_generateLocalKeyID( hka, ptb->buffer, &size ):
			pka_generateLocalKeyID( hka, ptb->buffer, &size );

	return ( res == pkae_ok );
}

bool BCDB_GetCrtSerialNumber( int *pSerialNumber, BYTE *issuerLocalKeyID, int size, bool autoIncrement )
{
	cSecKeySN sksn;

	if ( sksn.m_err )
		return false;

	*pSerialNumber = sksn.findKeyID( issuerLocalKeyID, size, autoIncrement );

	if ( *pSerialNumber || autoIncrement )
		sksn.writeToFile();

	return *pSerialNumber != 0;
}

//
//	Version 1.02.3 - 18.02.2008 choose My Documments as default folder for BCDB
//
static bool GetDefaultDataBasePathT( wchar_t *path, int size )
{
	ODSFunc;
	//cShell32Dll dll;

	//if ( !dll.m_SHGetSpecialFolderPathW( 0, path, CSIDL_PERSONAL, true ) || 
	if ( !cShell32Dll::getStatic().m_SHGetSpecialFolderPathW( 0, path, CSIDL_PERSONAL, true ) || 
		!FileExistsT( path ) )
	{
		int len = GetModuleFileNameT( hMainInst, path, size );

		if ( len >= size )
		{ ODS_fls( L"assert( false );" ); assert( false ); return false; }

		len = splitPath( path, path );
	}

	wcscat_s( path, size, L"\\BCDB" );

	CreateDirectoryT( path, NULL );

	bool res = FileExistsT( path ) == TRUE; // _waccess( path, 0 ) == 0;

	wcscat_s( path, size, L"\\" );

	ODSFuncRes( path );

	return res;
}

//static bool GetDefaultDataBasePath( char *path, int size )
//{
//	cWCStr wcs( size );
//	if ( wcs.m_len != 0 )
//		return false;
//
//	if ( !GetDefaultDataBasePathT( wcs.m_szBody, wcs.m_len ) )
//		return false;
//
//	getMBStr( wcs.m_szBody, path, size, CP_ACP );
//
//	return true;
//}

HKEY hKeyRoot = HKEY_CURRENT_USER;
wchar_t *szBCKeyManKey = L"Software\\Jetico\\BCKeyMan";
static char *szBCDBPath = "BCDBPath";
static wchar_t *szBCDBPathW = L"BCDBPath";

static bool getDataBasePathT( wchar_t *path, int size )
{
	ODSFunc;

	HKEY hKey;
	DWORD res = RegCreateKeyExT( hKeyRoot, szBCKeyManKey, 0, NULL, REG_OPTION_NON_VOLATILE, KEY_QUERY_VALUE, NULL, &hKey, NULL );
	if ( res != ERROR_SUCCESS )
		return GetDefaultDataBasePathT( path, size );

	DWORD cbValue = size;
	DWORD dwType;
	res = RegQueryValueExT( hKey, szBCDBPathW, 0, &dwType, (BYTE *)path, &cbValue );

	RegCloseKey( hKey );

	if ( ( res != ERROR_SUCCESS ) || ( size < (int)cbValue ) )
		return GetDefaultDataBasePathT( path, size );

	int len = wcslen( path );
	if ( path[ len - 1 ] != L'\\' )
		wcsncpy( path + len, L"\\", size - len );

	ODSFuncRes( path );
	
	return true;
}

bool BCDB_GetDataBasePath( wchar_t *path, int size, bool silent )
{
	return getDataBasePathT( path, size );

	//bool isNT = GetVersion() < 0x80000000;
	//if ( isNT )
	//	return getDataBasePathT( path, size );

	//char pathA[ MAX_PATH ];
	//bool res = BCDB_GetDataBasePath( pathA, sizeof( pathA ), silent );

	//if ( res )
	//	mbstowcs( path, pathA, size );

	//return res;
}

static bool setDataBasePathT( wchar_t *path )
{
	HKEY hKey;
	DWORD res = RegCreateKeyExT( hKeyRoot, szBCKeyManKey, 0, NULL, REG_OPTION_NON_VOLATILE, KEY_ALL_ACCESS, NULL, &hKey, NULL );
	if ( res != ERROR_SUCCESS )
		return false;

	DWORD cbValue = ( wcslen( path ) + 1 ) * 2;
	DWORD dwType = REG_SZ;
	res = RegSetValueExT( hKey, szBCDBPathW, 0, dwType, (BYTE *)path, cbValue );

	RegCloseKey( hKey );

	if ( res != ERROR_SUCCESS )
		return false;
	
	return true;
}

bool BCDB_SetDataBasePath( wchar_t *path )
{
	return setDataBasePathT( path );
}


bool BCDB_GetLocalKeyRingPath( wchar_t *path, int size, bool secret )
{
	if ( !BCDB_GetDataBasePath( path, size ) )
	{
		ODS_fls( L"Assertion!" );
		return false;
	}

	int pos = wcslen( path );

	if ( pos + 10 > size )
	{
		ODS_fls( L"Assertion!" );
		return false;
	}

	wcsncpy( path + pos, ( secret ? L"local.skr" : L"local.pkr" ), size - pos );

	if ( FileExistsT( path ) )
		return true;

	return false;
}

//
//	AddPublicKeyToStorage
//		Adds PGP public key to local key ring
//		Writes x509 certificate file to BCDB
//
bool AddPublicKeyToStorage( HWND hWnd, BYTE *pkt, int pktSize, type_kff kff )
{
	return AddKeyToStorage( hWnd, pkt, pktSize, kff, false );
}

bool AddKeyToStorage( HWND hWnd, BYTE *pkt, int pktSize, type_kff kff, bool secret )
{
	return AddKeyToStorageEx( hWnd, pkt, pktSize, kff, secret, NULL );
}

bool AddKeyToStorageEx( HWND hWnd, BYTE *pkt, int pktSize, type_kff kff, bool secret, wchar_t *path )
{
	cWCStr wcs;

	if ( kff == kffPGP )
	{
		int ids = BCDB_AddPGPKeyToLocalKeyRing( hWnd, pkt, pktSize, secret );
		if ( ids != NULL )
			wcs.Add( _LoadIDS( ids ) );
		else
			return true;
	}
	else
	if ( kff == kffX509 )
	{
		if ( secret )
		{	
			if ( BCDB_AddP12SecretKeyToStorage( pkt, pktSize, &wcs, 0 ) )
				return true;
		}
		else
		{
			if ( BCDB_AddCertificateToStorage( pkt, pktSize, &wcs, 0, path ) )
				return true;
		}
	}
	else
		wcs.Add( _LoadIDS( IDS_UndefinedKeyFileFormat ) );

	MessageBoxT( hWnd, wcs.m_szBody,  _LoadIDS( IDS_AppName ), MB_ICONSTOP );

	return false;
}

static int CheckSecretPGPKeyBlock(
		BYTE *pkt,	int pktSize, bool withHeader
	)
{
	sSecretKey *sk;
	int skSize;

	if ( withHeader )
	{
		BYTE tag;
		int dataSize;
		int hdrSize;
		bool partial = 0;
		int res;

		int parsed = 0;

		//
		//	Parses key packet.
		//
		
		res = pgp_PacketHeader_Parse( &tag, &hdrSize, &dataSize, &partial, pkt, pktSize );
		if ( res != ppe_ok ) return res;

		if ( tag != pkt_SecretKey )
			return ppe_invalidBlockType;

		sk = (sSecretKey *)(pkt + parsed + hdrSize);
		skSize = dataSize;
	}
	else // without header
	{
		sk = (sSecretKey *)(pkt);
		skSize = pktSize;
	}

	int publicMpiSize;
	BYTE symAlgID;		// identifier of the symmetric algorithm is used for decryption of private part
	S2K  s2k;			// mode for string to key conversion for symmetrin algorithm
	BYTE *IV;			// initial vector
	BYTE *encMpi;		// encrypted buffer with secret mpis
	int encMpiSize;		// size of buffer to be decrypted

	return PKT_SecretKey_Parse(
			sk, skSize,
			&publicMpiSize,
			&symAlgID,
			&s2k,
			&IV,
			&encMpi,
			&encMpiSize
		);
}


static int CheckPGPKeyBlock( BYTE *keyBuff, int keyBuffSize, bool secret, bool withHeader )
{
	if ( secret )
		return CheckSecretPGPKeyBlock( keyBuff, keyBuffSize, withHeader );
	else
	if( withHeader )
		return ExtractDataAndVerifyPGPPublicKeyBlock( keyBuff, keyBuffSize, NULL, 0, NULL, NULL, NULL, NULL, NULL, true, NULL );
	else
		return PKT_PublicKey_Parse( (sPublicKey*)keyBuff, keyBuffSize, NULL, NULL );
}

static int getKeyIDTbl( cPTB *ptbTbl, BYTE *keyPkt, int keyPktSize, bool secret = false )
{
	int size = 0;

	int res = pgp_EnumKeyIDs( NULL, &size, keyPkt, keyPktSize, secret );

	if ( res == ppe_bufferTooSmall )
	{
		if ( ptbTbl->Allocate( size ) == NULL )
			return ppe_memoryAllocation;

		res = pgp_EnumKeyIDs( ptbTbl->buffer, &size, keyPkt, keyPktSize, secret );
	}

	return res;
}

int BCDB_GetPGPKeyUserIDT(
		cWCStr *wcs, //cPTB *ptb, // [OUT] userID
		BYTE *krBuff, int krBuffSize,
		BYTE *keyID,
		bool secret )
{
	cpgp_keyInfo keyInfo, *pki = NULL;

	int res = pgp_FindKeyByIDEx(
				&keyInfo,
				keyID,
				krBuff, krBuffSize,
				secret,
				&pki );

	if ( res != ppe_ok )
		return res;

	if ( keyInfo.userID == NULL )
	{
		//char text[ 100 ];// = "Empty";
		//LoadStringTA( hMainInst, IDS_Empty, text, sizeof( text ) );
		//ptb->WriteString( text, false );
		wcs->Add( _LoadIDS( IDS_Empty ) );
		return ppe_ok;
	}
	else
	if ( pki != NULL ) // Sub key was found
	{
		//char text[ 100 ];// = "Subkey of ";
		//LoadStringTA( hMainInst, IDS_SubkeyOf, text, sizeof( text ) );
		//ptb->WriteString( text, false );
		//ptb->WriteString( " ", false );
		wcs->Add( _LoadIDS( IDS_SubkeyOf ) );
		wcs->Add( L" " );
	}

	//ptb->WriteBytes( keyInfo.userID->body, keyInfo.userID->size );
	//ptb->PutByte( 0 );
	wcs->Add( (char*)keyInfo.userID->body, keyInfo.userID->size, CP_UTF8 );
	wcs->Add( L"" );

	return ppe_ok;
}

int BCDB_AddPGPKeyToLocalKeyRing( HWND hWnd, BYTE *keyBuff, int keyBuffSize, bool secret )
{
	ODSFunc;

	cPTB ptb, ptbEntirePkt;

	if ( CheckPGPKeyBlock( keyBuff, keyBuffSize, secret, true ) == ppe_ok )
	{
		// It has header - does nothing
	}
	else
	if ( CheckPGPKeyBlock( keyBuff, keyBuffSize, secret, false ) == ppe_ok )
	{
		// Adds header to pure packet

		int  pktSize = 0;
		BYTE tag = secret ? pkt_SecretKey : pkt_PublicKey;

		int res = pgp_PacketHeader_Format(
			NULL,			// [Out]The packet with header.
			&pktSize,		// [In] Size of allcated buffer for the packet.
							// [Out] Space written or required if the buffer is too small.
							//	If last it returns error ppe_bufferTooSmall.
			tag,			// Content tag, 0 if it is partial tail
			keyBuff,		// Packet body ( data )
			keyBuffSize,	// The packet body size
			false			// Pack as partial data
			);

		if ( res != ppe_bufferTooSmall )
			return res;
		
		if ( ptbEntirePkt.Allocate( pktSize ) == NULL )
			return pkae_memoryAllocation;

		res = pgp_PacketHeader_Format(
			ptbEntirePkt.buffer,
			&pktSize,
			tag,
			keyBuff,
			keyBuffSize,
			false
			);

		if ( res != ppe_ok )
			return res;

		keyBuff = ptbEntirePkt.buffer;
		keyBuffSize = ptbEntirePkt.bufferSize;
	}
	else
		return IDS_szInvalidKeyBlock;

	wchar_t path[ MAX_PATH ];
	if ( BCDB_GetLocalKeyRingPath( path, maxLen( path ), secret ) ) // File exists.
	{
		cFileView cfv;
		if ( !cfv.Open( path ) )
			return IDS_szErrorInReadingLocalKeyRingFile;

		//
		//	Checks if same key already exists
		//
		if ( cfv.m_size.QuadPart != 0 )
		{
			if ( ptb.WriteBytes( cfv.m_ptr, (int)cfv.m_size.QuadPart, true ) == 0 )
				return IDS_szMemoryAllocationError;

			cPTB KR_IDtbl;
			int res = getKeyIDTbl( &KR_IDtbl, cfv.m_ptr, (int)cfv.m_size.QuadPart, secret );
			if ( res != ppe_ok )
				return IDS_szErrorInReadingLocalKeyRingFile;

			cPTB key_IDtbl;
			res = getKeyIDTbl( &key_IDtbl, keyBuff, keyBuffSize, secret );
			if ( res != ppe_ok )
				return IDS_szErrorInReadingLocalKeyRingFile;

			for( int i = 0; i * pgp_keyIDSize < key_IDtbl.bufferSize; i++ )
			{
				if ( key_IDtbl.bufferSize - i * pgp_keyIDSize < pgp_keyIDSize )
					return IDS_szInvalidKeyBlock;

				for ( int j = 0; j * pgp_keyIDSize < KR_IDtbl.bufferSize; j++ )
				{
					if ( memcmp( KR_IDtbl.buffer + j * pgp_keyIDSize, key_IDtbl.buffer + i * pgp_keyIDSize, pgp_keyIDSize )
						== 0 ) // same key ID found
					{
						//cPTB ptbUserID;
						cWCStr wcsUserID;

						int res = ppe_ok; 
						
						if ( !secret  )
						{
							res = BCDB_GetPGPKeyUserIDT( &wcsUserID, //&ptbUserID,
									ptb.buffer, ptb.bufferSize, key_IDtbl.buffer + i * pgp_keyIDSize, secret );
							if ( res != ppe_ok )
								return IDS_KeyRingReadError;
						}
						else
						{
							//
							//	Gets user id from local public key ring
							//
							wchar_t pkrPath[ MAX_PATH ];
							if ( BCDB_GetLocalKeyRingPath( pkrPath, maxLen( pkrPath ), false ) ) // File exists.
							{
								cFileView fvPKR( pkrPath );
								if ( fvPKR.m_err == cFileView::cfv_ok )
									res = BCDB_GetPGPKeyUserIDT( &wcsUserID, // &ptbUserID,
										fvPKR.m_ptr, (int)fvPKR.m_size.QuadPart, key_IDtbl.buffer + i * pgp_keyIDSize, false );

								if ( res != ppe_ok )
								{
									//ptbUserID.SeekToBegin();
									wcsUserID.Free();
	//								return IDS_KeyRingReadError;
								}
							}
						}
					
						//ptbUserID.WriteString( "", true );
						wcsUserID.Add( L"" );

						wchar_t keyIDStr[ 1024 ];
						BCDB_bytes2string(
								keyIDStr, maxLen( keyIDStr ),
								key_IDtbl.buffer + i * pgp_keyIDSize, pgp_keyIDSize
							);

						wchar_t message[ 1024 ];
						//wchar_t bf[256];
						FormatT( message, maxLen( message ),
							_LoadIDS( IDS_KeyAlreadyExists_SSS ),
							_LoadIDS( secret ? IDS_szSecret : IDS_szPublic ),
							//getWCStr( (char *)ptbUserID.buffer, bf, maxLen( bf ) ),
							wcsUserID.m_szBody,							
							keyIDStr
						);

						int idRes = MessageBoxT( hWnd, message, _LoadIDS( IDS_AppName ), MB_ICONQUESTION | MB_YESNO );

						if ( idRes == IDYES )
						{
							res = pgp_RemoveKeyByID( ptb.buffer, &ptb.bufferSize, KR_IDtbl.buffer + j * pgp_keyIDSize, secret );
						}
						else
							return IDS_szKeyWasNotOverwritten;
					}
				}
			}
		}
	}
	else
	{
		cWCStr wcsDir( MAX_PATH );
		splitPath( path, wcsDir.m_szBody );

		RecCreateDirectory( wcsDir.m_szBody ); 
	}

	if ( keyBuffSize != 0 )
	{
		int res = pgp_getValidSize( keyBuff, &keyBuffSize );
		if ( res != ppe_ok )
			return IDS_KeyRingReadError;
	}

	if ( ( keyBuffSize != 0 ) && ( ptb.WriteBytes( keyBuff, keyBuffSize ) == 0 ) )
		return IDS_szMemoryAllocationError;

	DWORD err = 0;

	if ( WriteToFile( path, ptb.buffer, ptb.bufferSize, false, &err ) == 0 )
	{
		ODS_fls( cWCStr().FormatT( L"Error %d in opening file '%s'", err, path ) );
		return IDS_szErrorInWrittingLocalKeyRingFile;
	}

	return 0;
}

bool BCDB_AddP12SecretKeyToStorage( BYTE *pkt, int pktSize, cWCStr *wcsInfo, HPKA *pSKA )
{
	wchar_t errStr[ 256 ];
	LoadStringT( hMainInst, IDS_NoError, errStr, maxLen( errStr ) );

	cPTB ptbLKID;

	if ( !BCDB_GetP12SecretKeyInfo( hMainInst, 0, pkt, pktSize, pSKA, NULL, &ptbLKID, errStr, maxLen( errStr ) ) )
	{
		if ( wcsInfo )	wcsInfo->Add( errStr );

		return false;
	}

	wchar_t path[ MAX_PATH ];
	if ( !BCDB_SecKeyFileName( path, maxLen( path ), ptbLKID.buffer, ptbLKID.bufferSize, true ) )
	{
		assert( false );
		return false;
	}

	if ( WriteToFile( path, pkt, pktSize ) == 0 )
	{
		if ( wcsInfo ) wcsInfo->Add( _LoadIDS( IDS_ErrorInWrittingSKFile ) );
		return false;
	}

	return true;
}

static bool addCertificate( BYTE *crtPkt, int crtPktSize, cWCStr *wcsInfo, HPKA *pPKA, wchar_t *pPath )
{
	HWND hParent = 0;
	HINSTANCE hInst = hMainInst;

	wchar_t path[ MAX_PATH * 2 ];
	if ( !BCDB_GetDataBasePath( path, maxLen( path ), false ) )
	{
		MessageBoxT( hParent, _LoadIDS( IDS_ErrorIngettingStorageFolder ), _LoadIDS( IDS_AppName ), MB_ICONSTOP );
		if ( wcsInfo != NULL )	wcsInfo->Add( _LoadIDS( IDS_ErrorIngettingStorageFolder ) );
		return false;
	}

	int len = wcslen( path );

	x509_certificate crt;
	int res = crt.parse( crtPkt, crtPktSize );
	if ( res != x509e_ok )
	{
		if ( wcsInfo != NULL )
		{
			wcsInfo->Add( _LoadIDS( IDS_InvalidCertificatePacket ) );
			MessageBoxT( hParent, _LoadIDS( IDS_InvalidCertificatePacket ), _LoadIDS( IDS_AppName ), MB_ICONSTOP );
		}

		return false;
	}

	const char *szCNValue = crt.tbsCertificate.subjectName.getAttributeValue( sNameAttribute::getOid( sNameAttribute::id_at_commonName ) );
	const char *szEmailValue = crt.tbsCertificate.subjectName.getAttributeValue( sNameAttribute::getOid( sNameAttribute::pkcs9_emailAddress ) );

	HPKA pka = NULL;
	HPKA generatedPKA = NULL;

	if ( pPKA == NULL )
	{
		res = pka_createAlgorithm_X509( &generatedPKA, &crt.tbsCertificate.subjectPKInfo );
		if ( res != pkae_ok )
		{
			MessageBoxT( hParent, _LoadIDS( IDS_ErrorInCreatingPKAlg ), _LoadIDS( IDS_AppName ), MB_ICONSTOP );
			if ( wcsInfo != NULL ) wcsInfo->Add( _LoadIDS( IDS_ErrorInCreatingPKAlg ) );
			return false;
		}

		pka = generatedPKA;
	}
	else
		pka = *pPKA;

	cPTB ptbID;
	bool err = !BCDB_GenerateLocalKeyID( &ptbID, pka, false );

	if ( generatedPKA != NULL )
		pka_free( generatedPKA );

	if ( err )
	{
		MessageBoxT( hParent, _LoadIDS( IDS_ErrorInGeneratingPKCrtPkt ), _LoadIDS( IDS_AppName ), MB_ICONSTOP );
		if ( wcsInfo != NULL ) wcsInfo->Add( _LoadIDS( IDS_ErrorInGeneratingPKCrtPkt ) );

		return false;
	}

	bool writeIt = true;

	wchar_t fileNameW[ MAX_PATH ];
	//cPTB existenUserID;
	cWCStr existenUserID;
	wchar_t lkIdStr[ MAX_PATH ];

	BCDB_bytes2string( lkIdStr, maxLen( lkIdStr ), ptbID.buffer, ptbID.bufferSize );

	//
	// Checks if certificate with same local key id already exists in BCDB.
	// 
	while( BCDB_GetX509KeyByID( NULL, ptbID.buffer, ptbID.bufferSize, false, &existenUserID, fileNameW ) )
	{
		//existenUserID.PutByte( 0 );
		existenUserID.Add( L"" );

		wchar_t message[ MAX_PATH + 256 ];
		FormatT( message, maxLen( message ),
			_LoadIDS( IDS_CrtFileAlreadyExists_SSQ ),			
			//getWCStr( (char*)existenUserID.buffer ),
			existenUserID.m_szBody,
			fileNameW,
			lkIdStr
			);

		//existenUserID.ClearBuffer(); // for next file found.
		existenUserID.Free();

		int idRes = MessageBoxT( hParent, message, _LoadIDS( IDS_AppName ), MB_ICONQUESTION | MB_YESNO );

		if ( idRes == IDYES )
			DeleteFileT( fileNameW );
		else
		{
			writeIt = false;
			FormatT( message, maxLen( message ),
				_LoadIDS( IDS_OverwrittingOfCrtIsSkipped_S ),			
				fileNameW );

			if ( wcsInfo != NULL ) wcsInfo->Add( message );

			break;
		}
	}
	
	if ( writeIt )
	{
		if ( szCNValue != NULL )
		{
			wcsncpy( path + len, getWCStr( szCNValue ), maxLen( path ) - len );
			wcsncat( path + len, L".cer", maxLen( path ) - len );
			if( ( FileExistsT( path ) ) && ( szEmailValue != NULL ) )
			{
				wcsncpy( path + len, getWCStr( szCNValue ), maxLen( path ) - len );
				wcsncat( path + len, L" - ", maxLen( path ) - len );
				wcsncat( path + len, getWCStr( szEmailValue ), maxLen( path ) - len );
				wcsncat( path + len, L".cer", maxLen( path ) - len );
			}
		}
		else
		if ( szEmailValue != NULL )
		{
			wcsncpy( path + len, getWCStr( szEmailValue ), maxLen( path ) - len );
			wcsncat( path + len, L".cer", maxLen( path ) - len );
		}

		//
		//	We may bravly use lkIdStr as file name because it has been checked that 
		//	there is no certificate with same lkid in BCDB
		//	See loop above.
		//
		if( ( wcslen( path ) > MAX_PATH ) || ( FileExistsT( path ) ) )
		{
			wcsncpy( path + len, lkIdStr, maxLen( path ) - len );
			wcsncat( path + len, L".cer", maxLen( path ) - len );
		}

		if ( WriteToFile( path, crtPkt, crtPktSize ) == 0 )
		{
			cWCStr wcs;
			wcs.FormatT( _LoadIDS( IDS_ErrorInWrittingCrtFile_S ), path );

			if ( wcsInfo != NULL )
				wcsInfo->Add( wcs.m_szBody );

			MessageBoxT( hParent, wcs.m_szBody, _LoadIDS( IDS_AppName ), MB_ICONSTOP );

			return false;
		}

		if ( pPath != NULL )
			wcsncpy( pPath, path, MAX_PATH );

		if ( wcsInfo != NULL )
			wcsInfo->Add( _FormatT( _LoadIDS( IDS_CrtFileWrittingSuccess_S ), path ) );
	}

	return true;
}

bool BCDB_AddCertificateToStorage( BYTE *crtPkt, int crtPktSize, cWCStr *wcsInfo, HPKA *pPKA, wchar_t *path )
{
	p12_safeBag pfx;
	char *oidT[] = { "pkcs", "7", "signedData", NULL };

	int res = pfx.parse( crtPkt, crtPktSize, 0,0, (const char **)oidT );
	if ( res == x509e_ok )
	{
		p7_signedData sd;
		res = sd.parse( pfx.value.B, pfx.value.L );

		if ( res == x509e_ok )
		{
			crtPkt = sd.certificates.B;
			crtPktSize = sd.certificates.L;
		}
	}

	return addCertificate( crtPkt, crtPktSize, wcsInfo, pPKA, path );
}

int BCDB_bytes2string( wchar_t *str, int strLen, BYTE *bf, int bfSize )
{
	int len = 0;

	for( int pos = 0; pos < bfSize; pos++, len += 2 )
	{
		if ( len + 2 > strLen )
			break;

		FormatT( str + len, strLen - len, L"%02X", (DWORD)bf[pos] );
	}

	return len;
}

int BCDB_bytes2string( char *str, int strLen, BYTE *bf, int bfSize )
{
	int len = 0;

	for( int pos = 0; pos < bfSize; pos++, len += 2 )
	{
		if ( len + 2 > strLen )
			break;

		sprintf( str + len, "%02X", (DWORD)bf[pos] );
	}

	return len;
}

//
//	BCDB_SecKeyFileName
//		Generates name of secret key from local key ID, it simplifies find process
//		in BCStorage folder.
//
bool BCDB_SecKeyFileName( wchar_t *nameBuff, int buffLen, BYTE *lkid, int lkidSize, bool addBCStoragePath )
{
	char str[ 1024 ];

	if ( BCDB_bytes2string( str, sizeof( str ), lkid, lkidSize ) == 0 )
		return false;

	strcat_s( str, ".p12" );

	int pathLen = 0;

	if ( addBCStoragePath )
	{
		if ( !BCDB_GetDataBasePath( nameBuff, buffLen, false ) )
			return false;

		pathLen = wcslen( nameBuff );
	}

	cWCStr wcs( str );

	if ( pathLen + (int)wcslen( wcs.m_szBody ) >= buffLen )
		return false;

	wcsncpy( nameBuff + pathLen, wcs.m_szBody, buffLen - pathLen );

	return true;
}

bool BCDB_SecKeyFileName( wchar_t *nameBuff, int buffLen, HPKA ska, bool addBCStoragePath )
{
	cPTB ptb;

	if ( !BCDB_GenerateLocalKeyID( &ptb, ska ) )
		return false;

	return BCDB_SecKeyFileName( nameBuff, buffLen, ptb.buffer, ptb.bufferSize, addBCStoragePath );
}

//
//	BCDB_GetCerPublicKeyInfo - high level function works in user interface mode
//	1. Parses certificate public key.
//	2. Create public key handler ( use pka_free functions to free the handle ).
//
bool BCDB_GetCerPublicKeyInfo(
		HINSTANCE hInst, HWND hWnd,
		BYTE *buff, int buffSize,
		HPKA *pPKA,
		cUniversalString *pUSFriendlyName,
		cPTB *ptbLocalKeyID,
		cPTB *ptbThumbprint
	)
{
	bool silent = pUSFriendlyName != NULL;

	x509_certificate crt;

	HPKA pka = NULL;

	int res = crt.parse( buff, buffSize );
	if ( res == x509e_ok )
		res = pka_createAlgorithm_X509( &pka, &crt.tbsCertificate.subjectPKInfo );

	if ( res != x509e_ok )
	{
		if ( !silent )
		{
			DlgMessageBoxT(
				hWnd,
				cWCStr().FormatT( _LoadIDS( IDS_ErrorInKeyReading_D ), res ),
				MB_ICONSTOP );
		}

		return false;
	}

	if ( ptbLocalKeyID != NULL )
		BCDB_GenerateLocalKeyID( ptbLocalKeyID, pka, false );
	
	if ( ptbThumbprint != NULL )
		crt.generateThumbprint( ptbThumbprint->Allocate( 20 ) );

	if ( pPKA == NULL )
		pka_free( pka );
	else
		*pPKA = pka;

	if ( pUSFriendlyName != NULL )
	{
		cPTB ptbUserID;
		cWCStr wcUserID;

		int size = 0;
//		if ( crt.tbsCertificate.subjectName.GenerateName( NULL, &size ) != x509e_bufferTooSmall )
		if ( crt.tbsCertificate.subjectName.getAttributesString( NULL, &size ) != x509e_bufferTooSmall )
		{
			assert( false );
			//ptbUserID.WriteString( LoadStringOnce( hMainInst, IDS_Undefined ) );
			wcUserID.Add( _LoadIDS( IDS_Undefined ) );
		}
		else
		if ( ptbUserID.Allocate( size ) == NULL )
			return false;
		else
//		if ( crt.tbsCertificate.subjectName.GenerateName( (char *)ptbUserID.buffer, &size ) != x509e_ok )
		if ( crt.tbsCertificate.subjectName.getAttributesString( (char *)ptbUserID.buffer, &size ) != x509e_ok )
		{
			assert( false );
			//ptbUserID.WriteString( LoadStringOnce( hMainInst, IDS_Undefined ) );
			wcUserID.Add( _LoadIDS( IDS_Undefined ) );
		}
		else
		{
			wcUserID.Add( (const char*)ptbUserID.buffer, size );
		}
		
		pUSFriendlyName->Set( wcUserID.m_szBody );
		// Set( (char *)ptbUserID.buffer ); //, size );
		// pUSFriendlyName->GenerateUnicodeString();
		// pUSFriendlyName->MakeFromUnicodeString();
	}

	return true;
}

class cPasswordDlgLoop_P12_PFX_PDU_parse : public cPasswordDlgLoop
{
	BYTE *buff;
	int buffSize;
	cPTB *ptbKey;
	wchar_t *errStr;
	int errStrSize;

public:

	cPasswordDlgLoop_P12_PFX_PDU_parse(
				const wchar_t *dlgText, int IDText, int IDPwd, const wchar_t *caption,
				int idsInvalidPassword, int idsErrorInProcessing,
				BYTE *_buff, int _buffSize, cPTB *_ptbKey, wchar_t *_errStr, int _errStrSize
				)
		: cPasswordDlgLoop( dlgText, IDText, IDPwd, caption, idsInvalidPassword, idsErrorInProcessing )
	{
		buff		= _buff;
		buffSize	= _buffSize;
		ptbKey		= _ptbKey;
		errStr		= _errStr;
		errStrSize	= _errStrSize;
	}

	virtual int pwdActor() 
	{
		int res = P12_PFX_PDU_parseEx( buff, buffSize, NULL, &m_pwd, ptbKey, NULL );
		switch ( res )
		{
			case ppe_ok : return errNo;
			case ppe_invalidPassword: return errInvalidPassword;
			default: return errInProcessing;
		}
	}

	virtual void showResult( UINT ids )
	{
		if ( errStr )	LoadStringT( m_hInst, ids, errStr, errStrSize );
		else			DlgMessageBoxT( ids, MB_ICONINFORMATION  );
	}

	cPassword *getPwd()
	{
		return &m_pwd;
	}
};

bool BCDB_GetP12SecretKeyInfo( 
		HINSTANCE hInst,
		HWND hWnd,
		BYTE *buff, int buffSize,
		HPKA *pSka,
		cUniversalString *pUSFriendlyName,
		cPTB *ptbLocalKeyId,
		wchar_t *errStr, int errStrSize )
{
	cPTB ptbAttrSet;
	cPTB ptbKey;

	int res = P12_PFX_PDU_parseEx( buff, buffSize, &ptbAttrSet, NULL, &ptbKey, NULL );

	if ( res != ppe_invalidPassword ) // x509e_emptyParameters )
	{
		if ( errStr )
			LoadStringT( hInst, IDS_SecretKeyPacketIsInvalid, errStr, errStrSize );
		else
			DlgMessageBoxT( hInst, hWnd, IDS_SecretKeyPacketIsInvalid, MB_ICONSTOP );
		return false;
	}

	cUniversalString us;

	if ( ptbAttrSet.bufferSize == 0 )
	{
		us.Set( _LoadIDS( IDS_Empty ) );
		if ( pUSFriendlyName != NULL )
			pUSFriendlyName->Set( _LoadIDS( IDS_Empty ) );
	}
	else
	{
		P12_attrSet attrSet;
		res = attrSet.parse( ptbAttrSet.buffer, ptbAttrSet.bufferSize, NULL, NULL, asn1tag_CONSTRUCTED_SET );
		if ( res != x509e_ok )
		{
			if ( errStr )
				LoadStringT( hInst, IDS_SecretKeyPacketIsInvalid, errStr, errStrSize );
			else
				DlgMessageBoxT( hInst, hWnd, IDS_SecretKeyPacketIsInvalid, MB_ICONSTOP );
			return false;
		}

		if ( ptbLocalKeyId != NULL )
		{
			x509_struct sLocalKeyID;
			res = attrSet.getLocalKeyId( &sLocalKeyID );
			if ( res == x509e_ok )
				ptbLocalKeyId->WriteBytes( sLocalKeyID.B, sLocalKeyID.L, true );
			else
				ptbLocalKeyId->SeekToBegin();
		}

		x509_struct sUnicodeString;
		res = attrSet.getFriendlyName( &sUnicodeString );

		if ( ( res == x509e_ok ) && ( sUnicodeString.L > 0 ) )
		{
			// Really FriendlyName attribute is defenitelly big-endian, so we may do not check.
			us.SetUnicode_CheckBigEndianInput( sUnicodeString.B, sUnicodeString.L );
			us.MakeFromUnicodeString();
			
			if ( pUSFriendlyName != NULL )
			{
				pUSFriendlyName->SetUnicode_CheckBigEndianInput( sUnicodeString.B, sUnicodeString.L );
				pUSFriendlyName->MakeFromUnicodeString();
			}
		}
		else
		{
			us.Set( _LoadIDS( IDS_szIsEmpty ) );

			if ( pUSFriendlyName != NULL )
			{
				pUSFriendlyName->Set( _LoadIDS( IDS_szIsEmpty ) );
			}
		}
	}

	if ( pSka == NULL )
		return true;

	cPasswordDlgLoop_P12_PFX_PDU_parse pwDlgLoop(
			_FormatT( _LoadIDS( IDS_UserName_S ), us.wcBuffer ), IDC_Text, IDC_Password, _LoadIDS( IDS_pwdDlgCaption ),
			IDS_InvalidPassword, IDS_ErrorInReadingKeyPacket,
			buff, buffSize, &ptbKey, errStr, errStrSize	);
	
	if ( pwDlgLoop.DoModalLoop( hInst, IDD_DlgPassword, hWnd ) != IDOK )
		return false;

	x509_privateKeyInfo pki;
	
	if ( pki.parse( ptbKey.buffer, ptbKey.bufferSize ) != x509e_ok  )
	{
		if ( errStr )
			LoadStringT( hInst, IDS_ErrorInReadingKeyPacket, errStr, errStrSize );
		else
			DlgMessageBoxT( hInst, hWnd, IDS_ErrorInReadingKeyPacket, MB_ICONSTOP  );
		return false;
	}

	if ( ska_createAlgorithm_X509( pSka, &pki ) != x509e_ok )
	{
		if ( errStr )
			LoadStringT( hInst, IDS_ErrorInReadingKeyPacket, errStr, errStrSize );
		else
			DlgMessageBoxT( hInst, hWnd, IDS_ErrorInReadingKeyPacket, MB_ICONSTOP  );
		return false;
	}

	return res == ppe_ok;
}

bool BCDB_GetCrtInfo( x509_certificate *crt, cWCStr *wcsUserID, cPTB *ptbKeyID, cPTB *ptbThumbprint )
{
	int size = 0;

	if ( wcsUserID != NULL )
	{
		cPTB ptb;

		if ( crt->tbsCertificate.subjectName.GenerateName( NULL, &size ) != x509e_bufferTooSmall )
		{
			assert( false );
			wcsUserID->Add( _LoadIDS( IDS_Undefined ) );
		}
		else
		if ( ptb.Allocate( size ) == NULL )
			return false;
		else
		if ( crt->tbsCertificate.subjectName.GenerateName( (char *)ptb.buffer, &size ) != x509e_ok )
		{
			assert( false );
			wcsUserID->Add( _LoadIDS( IDS_Undefined ) );
		}

		wcsUserID->Add( (const char*)ptb.buffer );
	}

	if ( ptbKeyID != NULL )
	{
		BYTE *keyID = NULL;
		int keyIDSize = 0;

		int res = crt->tbsCertificate.getSubjectKeyID( &keyID, &keyIDSize );

		if ( res == x509e_ok )
		{
			ptbKeyID->WriteBytes( keyID, keyIDSize, true );
		}
		else
		if ( res == x509e_extensionIsNotPresent )
		{
			keyIDSize = 0;
			res = crt->tbsCertificate.generateSubjectKeyID( NULL, &keyIDSize );
			if ( res != pkae_bufferSizeTooSmall )
				return false;

			if ( ptbKeyID->Allocate( keyIDSize ) == NULL )
				return false;

			res = crt->tbsCertificate.generateSubjectKeyID( ptbKeyID->buffer, &keyIDSize );
			if ( res != x509e_ok )
				return false;
		}
	}
	
	if ( ptbThumbprint != NULL )
		crt->generateThumbprint( ptbThumbprint->Allocate( 20 ) );

	return true;
}

//bool BCDB_EnumKeysT( cWCStr *wcsText, bool secret )
//{
//	return 
//		BCDB_EnumX509KeysT( wcsText, secret ) ||
//		BCDB_EnumPGPKeysT( wcsText, secret );
//}

int BCDB_EnumKeysBf( wchar_t *wc, size_t *pLen, bool secret )
{
	size_t len = *pLen;
	size_t lenPGP = 0;
	wchar_t *wcsPGP = NULL;

	int res = BCDB_EnumX509KeysBf( wc, &len, secret );

	if ( res == ppe_ok )
	{
		lenPGP = *pLen - len;
		wcsPGP = wc + len;
		if ( ppe_bufferTooSmall == BCDB_EnumPGPKeysBf( wcsPGP, &lenPGP, secret ) )
			res = ppe_bufferTooSmall;
	}
	else
	if ( res == ppe_bufferTooSmall )
	{
		lenPGP = 0;
		wcsPGP = NULL;
		BCDB_EnumPGPKeysBf( wcsPGP, &lenPGP, secret );
	}
	else
		res = BCDB_EnumPGPKeysBf( wc, pLen, secret );

	*pLen = len + lenPGP;

	return res;
}

int BCDB_EnumPGPKeysBf( wchar_t *wc, size_t *pLen, bool secret )
{
	cWCStr wcsText;

	int cnt = 0;

	wchar_t path[ MAX_PATH ];
	if ( !BCDB_GetLocalKeyRingPath( path, maxLen( path ), secret ) )
		return ppe_ok;

	cFileView cfv( path );

	if ( cfv.m_err )
		return ppe_keyNotFound;

	BYTE *kb = cfv.m_ptr;
	int kbSize = (int)cfv.m_size.QuadPart;

	cPTB key_IDtbl;
	int res = getKeyIDTbl( &key_IDtbl, kb, kbSize, secret );
	if ( res != ppe_ok )
		return res;

	wcsText.Add( cWCStr().FormatT( L"%s PGP keys\r\n", ( secret ? L"Secret" : L"Public" ) ) );

	for( int i = 0; i * pgp_keyIDSize < key_IDtbl.bufferSize; i++ )
	{
		if ( key_IDtbl.bufferSize - i * pgp_keyIDSize < pgp_keyIDSize )
			break;

		//cPTB ptbUserID;
		cWCStr wcsUserID;

		int res = ppe_ok; 
			
		if ( !secret  )
		{
			res = BCDB_GetPGPKeyUserIDT( &wcsUserID, kb, kbSize, key_IDtbl.buffer + i * pgp_keyIDSize, false );
			if ( res != ppe_ok )
				break;
		}
		else
		{
			//
			//	Gets user id from local public key ring
			//
			wchar_t pkrPath[ MAX_PATH ];
			if ( BCDB_GetLocalKeyRingPath( pkrPath, maxLen( pkrPath ), false ) ) // File exists.
			{
				cFileView fvPKR( pkrPath );
				if ( fvPKR.m_err == cFileView::cfv_ok )
					res = BCDB_GetPGPKeyUserIDT( &wcsUserID, fvPKR.m_ptr, (int)fvPKR.m_size.QuadPart, key_IDtbl.buffer + i * pgp_keyIDSize, false );

				if ( res != ppe_ok )
				{
					wcsUserID.Free(); // .SeekToBegin();
				}
			}
		}
		
		wcsUserID.Add( L"" );

		wchar_t keyIDStr[ 1024 ];
		BCDB_bytes2string(
				keyIDStr, maxLen( keyIDStr ),
				key_IDtbl.buffer + i * pgp_keyIDSize, pgp_keyIDSize
			);

		wcsText.Add( cWCStr().FormatT( L"%d) KeyID=", ++cnt ) );
		wcsText.Add( keyIDStr );
		wcsText.Add( L", ", false );
		wcsText.Add( wcsUserID.m_szBody, false );
		wcsText.Add( L"\r\n", false );
	}

	if ( *pLen <= (size_t)wcsText.m_len )
	{
		*pLen = wcsText.m_len + 1;
		return ppe_bufferTooSmall;
	}
	
	*pLen = wcsText.m_len + 1;
	copyString( wc, wcsText.m_szBody, *pLen, true );

	return ppe_ok;
}

int BCDB_EnumX509KeysBf( wchar_t *wc, size_t *pLen, bool secret )
{
	cWCStr wcsText;

	int cnt = 0;

	wchar_t path[ MAX_PATH ];
	if ( !BCDB_GetDataBasePath( path, maxLen( path ) ) ) 
		return ppe_keyNotFound;

	int len = wcslen( path );
	wcsncat( path + len, ( secret ? L"*.p12" : L"*.cer" ), maxLen( path ) - len );

	wcsText.Add( cWCStr().FormatT( L"%s\r\n", ( secret ? L"Secret P12 keys" : L"Public X509 certificates" ) ) );

	WIN32_FIND_DATAW findData;

	HANDLE hFind = FindFirstFileT( path, &findData );
	if ( hFind == INVALID_HANDLE_VALUE )
		return ppe_keyNotFound;

	do
	{
		if ( ( findData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY ) != 0 )
			continue;

		wcsncpy( path + len, findData.cFileName, maxLen( path ) - len );
		cFileView cfv( path );

		if ( cfv.m_err )
			continue;

		BYTE *keyID = NULL;
		int keyIDSize = 0;
		cPTB ptb;
		cPTB ptbAttrSet;
		P12_attrSet attrSet;
		x509_certificate crt;

		//cPTB ptbName;
		cWCStr wcsName;

		if ( secret )
		{
			int res = P12_PFX_PDU_parseEx( cfv.m_ptr, (int)cfv.m_size.QuadPart, &ptbAttrSet, NULL, &ptb, NULL );
			if ( ( res != x509e_ok ) && ( res != x509e_emptyParameters ) && ( res != ppe_invalidPassword ) )
				continue;

			res = attrSet.parse( ptbAttrSet.buffer, ptbAttrSet.bufferSize, NULL, NULL, asn1tag_CONSTRUCTED_SET );
			if ( res != x509e_ok )
				continue;

			x509_struct sLocalKeyId;
			res = attrSet.getLocalKeyId( &sLocalKeyId );
			if ( res != x509e_ok )
				continue;

			keyIDSize = sLocalKeyId.L;
			keyID = sLocalKeyId.B;

			x509_struct sFriendlyName;
			attrSet.getFriendlyName( &sFriendlyName );

			if ( res == x509e_ok )
			{
				cUniversalString us;
				// Really FriendlyName attribute is defenitelly big-endian, so we may do not check.
				us.SetUnicode_CheckBigEndianInput( sFriendlyName.B, sFriendlyName.L );
				us.MakeFromUnicodeString();
				wcsName.Add( us.wcBuffer );
			}
			else
			{
				//cPTB ptbUserID;
				cWCStr wcsUserID;
				BCDB_GetX509KeyByID( NULL, keyID, keyIDSize, false, &wcsUserID ); // ptbUserID );
				//cWCStr wcs( (char*)ptb.buffer, ptb.bufferSize );
				//ptbName.WriteString( wcs.m_szBody, true );
				//ptbName.WriteString( wcsUserID.m_szBody, true );
				wcsName.Add( wcsUserID.m_szBody );
			}
		}
		else
		{
			if ( crt.parse( cfv.m_ptr, (int)cfv.m_size.QuadPart ) != x509e_ok )
				continue;

			int res = crt.tbsCertificate.getSubjectKeyID( &keyID, &keyIDSize );

			if ( res == x509e_extensionIsNotPresent )
			{
				keyIDSize = 0;
				res = crt.tbsCertificate.generateSubjectKeyID( NULL, &keyIDSize );
				if ( res != pkae_bufferSizeTooSmall )
					continue;

				if ( ptb.Allocate( keyIDSize ) == NULL )
					continue;

				res = crt.tbsCertificate.generateSubjectKeyID( ptb.buffer, &keyIDSize );
				if ( res != x509e_ok )
					continue;

				keyID = ptb.buffer;
			}

			cWCStr wcsUserID;
			if ( BCDB_GetCrtInfo( &crt, &wcsUserID ) )
				wcsName.Add( wcsUserID.m_szBody );
			else
				wcsName.Add( L"<Invalid>" );
		}

		if ( keyID == NULL )
		{
			assert( false );
			continue;
		}

		wchar_t keyIDStr[ 1024 ];
		BCDB_bytes2string(
				keyIDStr, maxLen( keyIDStr ),
				keyID, keyIDSize
			);

		wcsText.Add( cWCStr().FormatT( L"%d) KeyID=", ++cnt ));
		wcsText.Add( keyIDStr );
		wcsText.Add( L", " );
		wcsText.Add( wcsName.m_szBody );
		wcsText.Add( L"\r\n" );
	}
	while( FindNextFileT( hFind, &findData ) );

	FindClose( hFind );

	if ( *pLen <= (size_t)wcsText.m_len )
	{
		*pLen = wcsText.m_len + 1;
		return ppe_bufferTooSmall;
	}
	
	*pLen = wcsText.m_len + 1;
	copyString( wc, wcsText.m_szBody, *pLen, true );

	return ppe_ok;
}

bool BCDB_GetX509KeyByID( cPTB *ptbKey, BYTE *localKeyIDBuffer, int bufferSize, bool secret, cWCStr *wcsUserID, wchar_t *fileName  )
{
	if ( ( ptbKey == NULL ) && ( fileName == NULL ) )
		return false;

	wchar_t path[ MAX_PATH ];
	if ( !BCDB_GetDataBasePath( path, maxLen( path ) ) ) 
		return false;

	int len = wcslen( path );
	wcsncat( path + len, ( secret ? L"*.p12" : L"*.cer" ), maxLen( path ) - len );

	WIN32_FIND_DATAW findData;

	HANDLE hFind = FindFirstFileT( path, &findData );
	if ( hFind == INVALID_HANDLE_VALUE )
		return false;

	bool found = false;

	do
	{
		if ( ( findData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY ) != 0 )
			continue;

		wcsncpy( path + len, findData.cFileName, maxLen( path ) - len );
		cFileView cfv( path );

		if ( cfv.m_err )
			continue;

		BYTE *keyID = NULL;
		int keyIDSize = 0;
		cPTB ptb;
		cPTB ptbAttrSet;
		P12_attrSet attrSet;
		x509_certificate crt;
		BYTE crtThumbprint[ 20 ]; // SHA1 digest of entire certificate.

		if ( secret )
		{
			int res = P12_PFX_PDU_parseEx( cfv.m_ptr, (int)cfv.m_size.QuadPart, &ptbAttrSet, NULL, &ptb, NULL );
			if ( ( res != x509e_ok ) && ( res != x509e_emptyParameters ) && ( res != ppe_invalidPassword ) )
				continue;

			res = attrSet.parse( ptbAttrSet.buffer, ptbAttrSet.bufferSize, NULL, NULL, asn1tag_CONSTRUCTED_SET );
			if ( res != x509e_ok )
				continue;

			x509_struct sLocalKeyId;
			res = attrSet.getLocalKeyId( &sLocalKeyId );
			if ( res != x509e_ok )
				continue;

			keyIDSize = sLocalKeyId.L;
			keyID = sLocalKeyId.B;
		}
		else
		{
			if ( crt.parse( cfv.m_ptr, (int)cfv.m_size.QuadPart ) != x509e_ok )
				continue;

			//
			//	Version 1.05.1 - 18.11.2015
			//		We have to check both thumbprint and local key ID, to save compatibility with old version.
			//		New PFX PDU uses thumprint, old - local key ID. See comments at Version 1.05.1
			//
			int res = crt.generateThumbprint( crtThumbprint );

			if ( ( res == ppe_ok ) && ( memcmp( localKeyIDBuffer, crtThumbprint, sizeof( crtThumbprint ) ) == 0 ) )
			{
				keyID = crtThumbprint;
				keyIDSize = sizeof( crtThumbprint );
			}
			else
			{
				res = crt.tbsCertificate.getSubjectKeyID( &keyID, &keyIDSize );

				if ( res == x509e_extensionIsNotPresent )
				{
					keyIDSize = 0;
					res = crt.tbsCertificate.generateSubjectKeyID( NULL, &keyIDSize );
					if ( res != pkae_bufferSizeTooSmall )
						continue;

					if ( ptb.Allocate( keyIDSize ) == NULL )
						continue;

					res = crt.tbsCertificate.generateSubjectKeyID( ptb.buffer, &keyIDSize );
					if ( res != x509e_ok )
						continue;

					keyID = ptb.buffer;
				}
			}
		}

		if ( bufferSize != keyIDSize )
			continue;

		if ( memcmp( localKeyIDBuffer, keyID, keyIDSize ) != 0 )
			continue;

		found = true;

		if ( ptbKey != NULL )
		{
			if ( ptbKey->WriteBytes( cfv.m_ptr, (int)cfv.m_size.QuadPart, true ) == 0 )
			{
				found = false;
				break;
			}
		}

		if ( fileName != NULL )
			wcsncpy( fileName, path, MAX_PATH );

		if ( wcsUserID )
		{
			if( secret )
			{
				x509_struct sFriendlyName;
				if ( ( attrSet.getFriendlyName( &sFriendlyName ) == ppe_ok ) )
				{
					cUniversalString us;
					// Really FriendlyName attribute is defenitelly big-endian, so we may do not check.
					us.SetUnicode_CheckBigEndianInput( sFriendlyName.B, sFriendlyName.L );
					us.MakeFromUnicodeString();
					wcsUserID->Add( us.wcBuffer );
				}
				else
				{
					assert( false );
					wcsUserID->Add( _LoadIDS( IDS_Undefined ) );
				}
			}
			else
			{
				if ( !BCDB_GetCrtInfo( &crt, wcsUserID ) )
					return false;
			}
		}

		break;
	}
	while( FindNextFileT( hFind, &findData ) );

	FindClose( hFind );

	return found;
}

type_kff getKeyFileFormat( const wchar_t *fileName )
{
	if ( fileName == NULL )
		return kffNone;

	size_t len = wcslen( fileName );

	if ( len <= 4 )
		return kffNone;

	const wchar_t *extPos = fileName + len - 4;

	if ( _wcsicmp( extPos, L".p12" ) == 0 ) return kffX509;
	if ( _wcsicmp( extPos, L".pfx" ) == 0 ) return kffX509;
	if ( _wcsicmp( extPos, L".cer" ) == 0 ) return kffX509;
	if ( _wcsicmp( extPos, L".skr" ) == 0 ) return kffPGP;
	if ( _wcsicmp( extPos, L".pkr" ) == 0 ) return kffPGP;
	if ( _wcsicmp( extPos, L".gpg" ) == 0 ) return kffPGP;
	if ( _wcsicmp( extPos, L".asc" ) == 0 ) return kffPGPAscii;
	if ( _wcsicmp( extPos, L".pem" ) == 0 ) return kffPEM;

	return kffNone;
}

bool BCDB_GetPGPPublicKeyInfo( HINSTANCE hInst, HWND hWnd, BYTE *buffer, int bufferSize, HPKA *pHPKA )
{
	int res = PKT_CreatePublicKeyAlgFromPacket( pHPKA, ( sPublicKey * )buffer, bufferSize );
	if ( res != ppe_ok )
	{
		DlgMessageBoxT(
			hWnd,
			cWCStr().FormatT( _LoadIDS( IDS_ErrorInKeyReading_D ), res ),
			MB_ICONSTOP );

		return false;
	}

	return true;
}

bool BCDB_GetPGPSecretKeyInfo( HINSTANCE hInst, HWND hWnd, BYTE *buffer, int bufferSize, HPKA *pHPKA, const wchar_t *friendlyName )
{
	//
	//	Checks if valid PGP secret key packet.
	//	The procedure parses all but key and cannot decrypt secret key because pwd was passed NULL
	//
	int res = PKT_CreateSecretKeyAlgFromPacket( pHPKA, ( sSecretKey * )buffer, bufferSize, NULL );

	if ( res == ppe_invalidPassword )
	{
		cPassword pwd;
		cPasswordDlg pwDlg( cWCStr().FormatT( _LoadIDS( IDS_UserID_S ), friendlyName ), IDC_Text, IDC_Password, _LoadIDS( IDS_pwdDlgCaption ), &pwd );

		while( true )
		{
			if ( pwDlg.DoModal( hInst, IDD_DlgPassword, hWnd ) != IDOK )
				return false;

			res = PKT_CreateSecretKeyAlgFromPacket( pHPKA, ( sSecretKey * )buffer, bufferSize, pwd.buffer );

			if ( res == ppe_ok )
				return true;

			if ( res == ppe_invalidPassword )
			{
				DlgMessageBoxT( hInst, hWnd, IDS_InvalidPassword, MB_ICONINFORMATION  );
				continue;
			}
			else
			{
				break;
			}
		}
	}

	DlgMessageBoxT( hWnd, cWCStr().FormatT( _LoadIDS( IDS_ErrorInKeyReading_D), res ), MB_ICONSTOP );
	return false;
}


//
//	GetBCArcPath
//		1. Tries to get name of parent module,
//			if the module file name is BCArchive returns the path of parent module.
//		2. Read Uninstall info to get home folder of BCArchive firstly, and BestCrypt secondly
//			checks existance of BCArchive.exe i the home dir. 
//		
//		Returns true if BCArchive path is found.
//
bool GetBCArcPath( wchar_t *exePath )
{
	wchar_t buff[ MAX_PATH ];

	// 1.

	DWORD res = GetModuleFileNameT( NULL, buff, maxLen( buff )  );

	int pos = splitPath( buff );
	wchar_t *fname = buff + pos;

	if ( _wcsnicmp( fname, szBCArchive, wcslen( szBCArchive ) ) == 0 )
	{
		wcsncpy( exePath, buff, MAX_PATH );
		return true;
	}

	// 2.

	if( !GetBCHomeDir( buff, szBCArchive ) &&
		!GetBCHomeDir( buff, szBestCrypt ) )
		return false;

	int len = wcslen( buff );

	if ( buff[ len - 1 ] != L'\\' )
	{
		buff[ len++ ] = L'\\';
	}
 
	wcsncpy( buff + len, szBCArchive, maxLen( buff ) - len );
	wcsncat( buff + len, L".exe", maxLen( buff ) - len );

	if ( !FileExistsT( buff ) )
	{
		wcsncpy( buff + len, szBCArchive, maxLen( buff ) - len );
		wcscat_s( buff, L"\\");
		wcscat_s( buff, szBCArchive );
		wcscat_s( buff, L".exe" );

		if ( !FileExistsT( buff ) )
			return false;
	}

	wcsncpy( exePath, buff, MAX_PATH );
	return true;
}

//static bool RunProcess(
//		wchar_t *cmdLine,
//		int *pWaitResult = NULL, // Null means - do not wait. Otherwise here is process result code.
//		bool show = true )
//{
//	STARTUPINFO stInfo;
//	memset(&stInfo,0,sizeof(stInfo));
//	stInfo.cb = sizeof(stInfo);
//	stInfo.dwFlags = STARTF_USESHOWWINDOW;
//	stInfo.wShowWindow = show ? SW_SHOW : SW_HIDE;
//
//	PROCESS_INFORMATION prInfo;
//
//	memset( &prInfo, 0, sizeof( prInfo ) );
//
//	DWORD dwCreationFlag = 0; // wait ?  : DETACHED_PROCESS; //NORMAL_PRIORITY_CLASS;
//
//	BOOL ok = CreateProcess(
//		NULL,
//		cmdLine,
//		NULL, // Security process attribute...
//		NULL, // Security thread attribute...
//		FALSE, // Inherit handlers
//		dwCreationFlag,
//		NULL, // Enviroment
//		NULL,
//		&stInfo,
//		&prInfo
//		);
//
//	if ( ok )
//	{
//		if (  pWaitResult )
//		{
//			_cwait( pWaitResult, (int)prInfo.hProcess, NULL );
//		}
//	}
//	else
//	{
//		assert( FALSE );
//	}
//
////	if ( prInfo.hProcess != NULL )
////		CloseHandle( prInfo.hProcess );
//	if ( prInfo.hThread != NULL )
//		CloseHandle( prInfo.hThread );
//
//	return ok != FALSE;
//}

//
//	Export or import BCDB to/form file
//
bool BCDB_Port( HWND hWnd, bool bExport )
{
	static wchar_t exePath[ MAX_PATH ] = L"";

	if ( !GetBCArcPath( exePath ) || 
		( !FileExistsT( exePath ) )
		)
	{
		MessageBoxT( hWnd, exePath, _LoadIDS( IDS_FileNotExists ), MB_ICONSTOP );
		return false;
	}

	wchar_t path[ MAX_PATH ] = L"\0";

	wchar_t szFilters[ 1024 ];
	LoadStringT( hMainInst, IDS_szFilterJDB, szFilters, maxLen( szFilters ) );

	bool res = BrowseFiles(
			hWnd,
			path, maxLen( path ),
			szFilters,
			_LoadIDS( bExport ? IDS_ChooseDatabaseExportFile : IDS_ChooseDatabaseImportFile ),
			false,
			( bExport ? OFN_OVERWRITEPROMPT : 0 )
		)
			== TRUE;

	if ( !res )
		return false;

	wchar_t bcdbPath[ MAX_PATH ];

	if ( !BCDB_GetDataBasePath( bcdbPath, maxLen( bcdbPath ), false ) )
		return false;

	wchar_t cmdLine[ MAX_PATH * 2 ];

	if ( bExport )
	{
		wcscat_s( bcdbPath, L"*" );
		FormatT( cmdLine, maxLen( cmdLine ), L"\"%s\" add \"%s\" \"%s\"", exePath, path, bcdbPath );
	}
	else // import
	{
		FormatT( cmdLine, maxLen( cmdLine ), L"\"%s\" extract \"%s\" \"%s\"", exePath, path, bcdbPath );
	}

	int retCode = -1;

	HANDLE hProcess = RunProcessWithWaitingT( cmdLine, true, false, NULL, &retCode );  // ( cmdLine, &resCode, true );

	if ( hProcess <= 0 )
	{
		MessageBoxT( hWnd, cmdLine, _LoadIDS( IDS_ErrorInStartingProcess ), MB_ICONSTOP );
		return false;
	}

	return true;
}