#ifndef __DlgViewLocalDatabase__
#define __DlgViewLocalDatabase__

#include "BCKeyMan.h"
#include "resource.h"
#include "DlgFindParams.h"

#include "..\Includes\FileView.h"
#include "..\Includes\cDlg.h"
#include "..\Includes\UnicodeFuncs.h"
#include "..\Includes\cSysDll.h"
#include "..\Includes\menuItem.h"

#include "..\Includes\IconCache.h"
#include "..\Includes\cSplitter.h"

class cDlgViewLocalDatabase : public cColoredDlg
{
	cPlainTextBlock *m_ptbKey;
	//cPlainTextBlock	*m_ptbUserID;
	cWCStr *m_wcsUserID;
		// [OUT] Pgp subkey does not have an user ID ( master key has only )
		// but we have to ask password for secret key with a name.
		// The BrowseKeyRingForUser generates here the subkey name as 
		// as "Subkey of ( master key user ID )" or  key user ID if the key has it

	type_kff *m_pKFF;

	HWND m_hList;
	HWND m_hTree;
	HWND m_hToolbar16;
	HWND m_hToolbar24;
	HWND m_hToolbar; // Current tool bar
	int m_columnQ;

	//
	//	PGP specific values
	//		All the values are assigned in FillListView_PGP
	//
	HIMAGELIST m_himlPGP, m_himlPGPLarge;
	cFileView m_fvKeyRing;
	cPlainTextBlock m_krImage;// file view has to be closed to be enable for modification
	cPlainTextBlock m_keyIDtbl;
	int m_keyPktSize;
	BYTE *m_keyPkt;
	bool m_secret;
	void expandItem( int item );	//	Expands PGP subkeys


	cIconCache m_iconCache;

	cSplitter *m_splitter;

	sImageInfo *m_siiT;
	int m_siiQ;

public:

	enum eViewType { vt_crtPublicKey = 0, vt_p12SecretKey, vt_pgpPublicKey, vt_pgpSecretKey };
	eViewType m_viewType;
	bool m_useTree;
	bool m_bLargeToolbar;

	enum { IDD = IDD_BCDB_Manager };

	cDlgViewLocalDatabase( DWORD viewFlags, cPlainTextBlock *ptbKey, type_kff *pKFF, cWCStr *wcsUserID )
			//cPlainTextBlock *ptbUserID )
	{
		m_ptbKeyOut = ptbKey;
		m_pKFF = pKFF;
		m_outFormat = viewFlags;

		m_viewType = vt_crtPublicKey;
		m_bLargeToolbar = true;
		//m_ptbUserID = ptbUserID;
		m_wcsUserID = wcsUserID;
		m_splitter = NULL;

		m_siiT = NULL;
		m_siiQ = 0;
	}

	~cDlgViewLocalDatabase()
	{
	}

	enum { IDM_VIEW = 200, IDM_Find, IDM_GenerateNew, IDM_PROPERTIES };
 private:


	LRESULT RealDlgProc( UINT uMsg, WPARAM wParam, LPARAM lParam );

	wchar_t m_caption[128];

	BOOL InitDialog();
	void DeleteDialog();

	//
	//	GetSelected
	//		Fills selected key block
	//
	bool GetSelected( cPlainTextBlock *ptb, cWCStr *wcsUserID );
	bool GetSelected( cPlainTextBlock *ptb, cWCStr *wcsUserID, bool secret );
	bool GetSelected_PGP( cPlainTextBlock *ptb, cWCStr *wcsUserID, bool secret );
	//bool GetSelected( cPlainTextBlock *ptb, cPlainTextBlock *ptbUserID );
	//bool GetSelected( cPlainTextBlock *ptb, cPlainTextBlock *ptbUserID, bool secret );
	//bool GetSelected_PGP( cPlainTextBlock *ptb, cPlainTextBlock *ptbUserID, bool secret );

	int RefreshView( eViewType viewType );
	int FillListView();
	int FillListView_internal();
	BOOL FillTreeView();
	int SelectTreeView( eViewType viewType );
	void ShowTreeView();

	int AddLVItem( const wchar_t *path, WIN32_FIND_DATAW *findData, int itemNum );

	//
	//	Reads public or secret local key ring and fills list view
	//
	int FillListView_PGP( bool secret );
	int SetPGPItemValues( BYTE *krPkt, int krSize, int item, BYTE *keyID );

	//
	//	Walks all public certificates or secret pkcs#12 files in BCDB folder and fills list view.
	//
	int FillListView( bool secret );

	int InitToolBar();
	void ToolBarUpdateSize();
	void ToolBarUpdateButtons();
	int ToolBarNotify( NMHDR *pnmh );
	int ListViewNotify( NMHDR *pnmh );
	int TreeViewNotify( NMHDR *pnmh );
	BOOL checkHitPlus( POINT pt );
	BOOL OnInitMenuPopup( HMENU hmInit, int nIndex, BOOL fSystemMenu );
	BOOL OnCommand( WPARAM wParam, LPARAM lParam );

	BOOL OnKeyProperties();
	BOOL OnKeyProperties( bool secret );
	BOOL OnKeyProperties_PGP( bool secret );

	enum waRes { wa_Cancel = 0, wa_Yes, wa_YesToAll, wa_No, wa_NoToAll };
	typedef waRes typeWalkActor(
			int iItem, // = -1,
			waRes prevRes // = wa_Yes
		);

	BOOL DeleteSelected();
	waRes OnDeleteItem(
			int i,// = -1,  // -1 means delete selected
			waRes prevRes// = wa_Yes
		);
	//
	//	Uses BCArchive to compress content of BCDB folder to export.bca file
	//
	bool OnExportDatabase();

	//
	//	Uses BCArchive to extract export.bca file content to current BCDB folder
	//
	bool OnImportDatabase();

	bool OnRefresh();
	bool OnChooseStoreFolder();
	bool OnGenerateKeyPair();
	bool OnAddPK_SearchInNet();
	bool OnAddPK_BrowseLocaly();
	int FindItem( BYTE *keyID, int keyIDSize );

	//
	//	Find function and params.
	//
	cFindParams m_findParams;
	bool IsFindParamsEmpty();
	bool GetFindParams();
	bool FindAndSelect();

	//
	//	Send armored publc key by eMail
	//	1. Sends selected certificate if x.509 crt is viewed.
	//	2. Sends corresponded certifiacte for selected secret key p12.
	//	3. Sends PGP key selected from local key ring.
	//
	bool SendPKToEMailRecipient();
	bool GetSelectedPublicKey( cPlainTextBlock *ptbBase64Encoded, cWCStr *wcsUserID );
	bool CopyPKToClipboard();

	//
	//	ReadPKFromClipboard
	//		1. Finds ascii armored PGP or Certificate blocks
	//		2. Adds the public key to local database.
	//
	bool ReadPKFromClipboard();
	bool isClipboardEmpty();
	bool ReadASCIIArmoredBlock( const char *asciiBlock );

	static DWORD GetIDSbyCmd( int cmd );
	static int LV_SetStyle( HWND hList, DWORD idCmd );

	void OnHelp();

	//
	//	Note :: GetKeyOnly -> Output values
	//		If the params are specified, the dialogue is used to get the specified key from local database.
	//		Double click on a key fills the parameters and close the Database manager.
	//		1. If vld_flag_viewSecret specified only so the secret key view is available only. All other views 
	//			are disabled. ( see cDlgViewLocalDatabase::OnInitMenuPopup )
	//		2. The vld_flag_viewPublic allows to view secret key but returns the coresponded public key from
	//			the database. ( see cDlgViewLocalDatabase::OnGetKey )
	//		3. Right-click pop-up menu has Get the Key entry and it is default instead usual view has 
	//			the Properties entry as default. ( see cDlgViewLocalDatabase::OnInitMenuPopup -> ID_RBCLICK_GETTHEKEY )
	//		4. It does not save view style and positions if key is getting only.
	//			see cDlgViewLocalDatabase::DeleteDialog
	//
	//	The ViewLocalDatabase calls the dialogus with corresponded parameters
	//
	cPlainTextBlock *m_ptbKeyOut;
	DWORD m_outFormat;

	bool GetKeyOnly();
	int OnGetKey();

	wchar_t *GetHelpFileName( wchar_t *path, int wcSize );

	void ReloadMenu();
};

int lvGetSelectedItem( HWND hLV );
BYTE *LV_GetKeyID(
		HWND hList,
		int item = -1 // -1 means selected item
	);

int deleteSubKeys( HWND hList, int item );
extern const int fileNamePos;

#endif // __DlgViewLocalDatabase__