/////////////////////////////////////////////////////////////////
//
//	Copyright (c) 2005-2013 Jetico, Inc., Finland
//	All rights reserved.
//
//	File: KM_Funcs.h
//	Revision: 
//	Created:
//	Description:
//		Interface module for access functions 
//		from BCKeyMan.dll
//
/////////////////////////////////////////////////////////////////
#pragma once

#include "..\Includes\cTrustDll.h"
#include "..\Includes\UnicodeFuncs.h"
#include "..\CertificateLib\SMemory\SBlocks.h"

#include <assert.h>

enum type_kff {
	kffNone = 0,
	kffX509 = 1,
	kffPGP = 2,
	kffPGPAscii = 3
};
typedef class cPKA * HPKA; // Handle to public/private key algorithm

bool KM_BrowseForPKBlock(
		HWND hWnd,
		const wchar_t *initialPath,// if initialPath == NULL it runs in BCDB
		bool secret,
		cPTB *ptbKey,// [OUT] key in x509 or PGP format
		type_kff *pKeyPktFormat,// [OUT] Format of key packet
		cWCStr			// [OUT] Pgp subkey does not have an user ID ( master key has only )
			*wcsUserID = NULL	// but we have to ask password for secret key with a name.
								// The BrowseKeyRingForUser generates here the subkey name as 
								// as "Subkey of ( master key user ID )" or  key user ID if the key has it
	);

bool KM_GetPKBlockByID(
		HWND hWnd,
		BYTE *keyID,
		int keyIDSize,
		bool secret,
		cPTB *ptbKey,// [OUT] key in x509 or PGP format
		type_kff *pKeyPktFormat, // [OUT] Format of key packet
		cWCStr *wcsUserID = NULL
	);

int KM_SearchPKUser(
		HWND hWnd,
		cPTB *ptbKey = NULL, //if null the function is used to search key and save it to BCDB
		BYTE *keyID = NULL, // If key id specified the key ID filed is filled and then locked.
		cWCStr *wcsUserID = NULL
	);

//
//	Function:
//		GenerateKeyPair
//	1. Generates key pair
//	2. Saves private key to file pfx_pdu format with .p12 extension
//		Adds the file to BCStorage.
//	3. Saves public key to certificate file with .cer extension.
//		The PK may be signed by its private key
//			Note, private key has to be added to BCStorage because the function takes
//			secret key attributes by ska local key ID
//		Or any other specified secret from BCStorage ( or may be from system storage WinNT )
//		Adds the file to BCStorage.
//	4. Fills localKeyID of the key been generated.
//
int KM_GenerateKeyPair(
		HINSTANCE hInst,
		HWND hWnd,
		BYTE *localKeyID = NULL,
		int *pLocalKeyIDBuffSize = NULL
	);

//
//	AddPublicKeyToStorage
//		Adds PGP public key to local key ring
//		Writes x509 certificate file to BCDB
//
bool KM_AddPublicKeyToStorage( HWND hWnd, BYTE *pkt, int pktSize, type_kff kff );
bool KM_AddKeyToStorage( HWND hWnd, BYTE *pkt, int pktSize, type_kff kff, bool secret );

//
//	BrowseForSecretKey
//	Uses OpenFile to choose the secret key file specified by keyID.
//	It may be as '.p12' secret key file as well as '.pgp, .skr, .asc' secret key ring file.
//
bool KM_BrowseForSecretKey(
		BYTE *keyID,
		int keyIDSize,
		cPTB *ptbKey,// [OUT] key in x509 or PGP format
		type_kff *pKeyPktFormat, // [OUT] Format of key packet
		cWCStr *wcsUserID // = NULL
	);

//
//	BrowseKeyFiles
//		Browses key file in BCDB ( default if pathBuff is empty ) or an other directory
//		Uses detailed view of p12 or cer files.
//		Returns true if key file is selected and fills pathBuff by file path name.
//
#define KM_BrowseFiles_Flag_UseAddKeyButton 1
#define KM_BrowseFiles_Flag_UseNewKeyButton 2
#define KM_BrowseFiles_Flag_DisableBrowseButtons 4

bool KM_BrowseKeyFiles(
		HWND hWnd,
		wchar_t *pathBuff, int buffSize,
		const wchar_t *filters,
		const wchar_t *boldCaptionText,
		const wchar_t *notes = L"",
		const wchar_t *OKBtnText = L"&OK",
		DWORD dwFlags = 0
	);

//
//	BrowseKeyRingForUser
//		Visual interfase for choosing a public or secret key from PGP key ring ( *.pkr, *.skr, *.gpg )
//		It allows to dynamically change key file path, so krOut contains the new path
//		ptbUserID is 
//			UserID for PGP key packet,
//			Friendly name for P12 secret key packet,
//			Subject RDN for certificate
//
//		If ptbKey is NULL the function is used for viewing and deleting a key
//		OK button in corresponded dialogue is changed as Delete and does not close the dialog.
//
int KM_BrowseKeyRingForUser(
		HWND hWnd,
		cPTB *ptbKey, // [OUT] 
		const char *keyRingPath, // [IN ]
		char *krOut = NULL,		 // [OUT] if krOut != NULL it is to be MAX_PATH size buffer.
		bool secret = false,
		cWCStr *wcsUserID = NULL
				// [OUT] Pgp subkey does not have an user ID ( master key has only )
				// but we have to ask password for secret key with a name.
				// The BrowseKeyRingForUser generates here the subkey name 
				// as "Subkey of ( master key user ID )" or  key user ID if the key has it
	);

//
//	KM_GetDataBasePath
//		Returns path to local database.
//
enum { KM_GetDataBasePath_Flag_Folder = 0, KM_GetDataBasePath_Flag_PGPPublicKeyRing = 1, KM_GetDataBasePath_Flag_PGPSecretKeyRing =2 };
bool KM_GetDataBasePath( wchar_t *buff, int buffLen, DWORD dwFlag = KM_GetDataBasePath_Flag_Folder );

//
//	UpdatePublicRecipientList
//		Add all public key encounterred to new allocated cPublicRecipientList
//			1. Walks all certificate files.
//			2. Gets all public key from local key ring
//		Returns the list;
//
//	Note, all allocation and deallocations are performed in the dll.
//
class cPublicRecipientList;
cPublicRecipientList *KM_UpdatePublicRecipientList( cPublicRecipientList *prl = NULL );
void KM_FreePublicRecipientList( cPublicRecipientList *prl );


//
//	ViewLocalDatabase
//		1. Shows and manages keys in local data base.
//		2. If paramters ptbKey and pKFF are specified it allows to choose
//			key from local database.
//

#define vld_flag_viewSecret 1
#define vld_flag_viewPublic 2

bool KM_ViewLocalDatabase(
		HWND hWnd,
		DWORD viewFlags = vld_flag_viewSecret | vld_flag_viewPublic,
		cPTB *ptbKey = NULL,
		type_kff *pKFF = NULL,
		cWCStr			// [OUT] Pgp subkey does not have an user ID ( master key has only )
			*wcsUserID	// but we have to ask password for secret key with a name.
				 = NULL	// The BrowseKeyRingForUser generates here the subkey name as 
						// as "Subkey of ( master key user ID )" or  key user ID if the key has it
	);

//
//	KM_EnumKeys uses BCDB_EnumPGPKeysBf and/or BCDB_EnumX509KeysBf functions according kff type.
//
//	BCDB_EnumKeys
//		Walks BCDB and fills wcsText
//			the text contains list of key info
//			keyID, userName
//	Sample:
//  123456789, <qqq@qqqq.qqq>
//
//	Note, the key id may be used for extracting key from BCDB by GetPKBlockByID
//
bool KM_EnumKeys( cWCStr *wcsText, bool secret = false, type_kff kff = kffNone );

class CKeyManDll : public cTrustDll
{
public:

	static CKeyManDll &getStatic( bool bShowMessage = false );

	CKeyManDll( BOOL usePath, const wchar_t *libraryName, bool bShowMessage = false )
	{
#ifndef _ExtractorOnly_
		//
		//	 WinVerifyTrust requires full path only.
		//
		cWCStr wcs;

		if ( usePath )
		{
			wcs.Set( libraryName );
		}
		else
		{
			wcs.SetLength( MAX_PATH );

			if ( !GetModuleFileNameW( NULL, wcs.m_szBody, wcs.m_len ) )
				return;

			cWCStr dir;
			splitPath( wcs.m_szBody, &dir );

			wcs.FormatT( L"%s\\%s", dir.m_szBody, libraryName );

			// If BCArchive is a part of BestCrypt it uses BCKeyMan.dll from BestCrypt root folder.
			if ( !FileExistsT( wcs.m_szBody ) )
			{
				wcs.SetLength( splitPath( dir.m_szBody ) );
				wcs.Add( libraryName );
			}

			if ( !FileExistsT( wcs.m_szBody ) )
				return;
		}

		VerifyAndLoad( wcs.m_szBody, NULL, bShowMessage );
#endif
	}

	BOOL GetProcAddress( const char *procName, BOOL showMessage = TRUE )
	{
		if ( hInst == NULL )
			return false; // Trusted dll has to be loaded at constructor with full path.
		else
			return CDll::GetProcAddress( procName, showMessage );
	}
};
