#include "resource.h"
#include "BCKeyMan.h"
#include "cWizardDlgs.h"

#include "..\CERTIFICATELIB\PUBLICKEYALGS\POWER\Windefs.h"
#include "..\CertificateLib\hashAlgs\hashAlgs.h"
#include "..\CertificateLib\PGPPkts\pktdefs.h"

#include "..\Includes\cDlg.h"

#include <windows.h>
#include <windowsx.h>
#include <commctrl.h>
#include <stdio.h>
#include <time.h>

//
//	Function:
//		GenerateKeyPair
//	1. Generates key pair
//	2. Saves private key to file pfx_pdu format with .p12 extension
//		Adds the file to BCStorage.
//	3. Saves public key to certificate file with .cer extension.
//		The PK may be signed by its private key
//			Note, private key has to be added to BCStorage because the function takes
//			secret key attributes by ska local key ID
//		Or any other specified secret from BCStorage ( or may be from system storage WinNT )
//		Adds the file to BCStorage.
//	4. Fills localKeyID of the key been generated.
//
int GenerateKeyPair(
		HINSTANCE hInst,
		HWND hWnd,
		BYTE *localKeyID,
		int *pLocalKeyIDBuffSize
	)
{
	if ( hInst == NULL )
		hInst = hMainInst;

	if ( pLocalKeyIDBuffSize != NULL )
	{
		int hashID = pgphaSHA1;

		int digetSize = 0;
		int res = ha_getDigestSize( &digetSize, hashID );
		if ( *pLocalKeyIDBuffSize < digetSize )
		{
			*pLocalKeyIDBuffSize = digetSize;
			return x509e_bufferTooSmall;
		}
	}

	cWizardBlock wBlock( hInst, hWnd );

	// Wizard contains 

	// 1. Wellcome dialogue with generation options.
	// 2. Create secret key packet dialogue with IDD_CreateCertificate
	// 3. Create certificate ( public key )packet dialogue with IDD_CreateCertificate

	const int dlgQ = 3;
	int i;
	cWizardDlg *dlgT[dlgQ];

	for( i = dlgQ; i--; )
		dlgT[ i ] = NULL;

	dlgT[ 0 ] = new cWizardDlg_Welcome( &wBlock, hInst, hWnd );
	dlgT[ 1 ] = new cWizardDlg_SecKeyPkt( &wBlock, hInst, hWnd );
	dlgT[ 2 ] = new cWizardDlg_PubKeyCrt( &wBlock, hInst, hWnd );
	
	for( i = 0; ; )
	{
		if ( ( i >= dlgQ ) || ( i < 0 ) )
		{	assert( false );
			break; 
		}

		int res = dlgT[i]->RunIt();
	
		if ( res == IDCANCEL ) break;
		else
		if ( res == IDOK )
		{
			//cPlainTextBlock ptbLog;
			cWCStr wcsLog;

			if ( wBlock.Generate( &wcsLog ) )
			{
				//ptbLog.WriteString( L"", true );
				
				cWizardDlg_FinalInfo dlg( &wBlock, hInst, hWnd, &wcsLog );
				res = dlg.RunIt();

				if ( res == IDCANCEL )
					break;
				else
					continue;
			}
			else
			{
				//ptbLog.WriteString( L"", true );
				DlgMessageBoxT( hWnd, _LoadIDS( IDS_ErrorInGeneratingProcess ), MB_ICONSTOP );
				continue;
			}
		}
		else
			if ( res == IDC_btPrev ) { i--; if ( ( i == 1 ) && !wBlock.generateSecKey ) i--; }
		else
			if ( res == IDC_btNext ) { i++; if ( ( i == 1 ) && !wBlock.generateSecKey ) i++; }
		else
			assert( false );
	}

	for( i = dlgQ; i--; )
		if ( dlgT[ i ] != NULL )
			delete dlgT[i];

	if ( wBlock.wasGenerated )
	{
		if ( ( pLocalKeyIDBuffSize != NULL ) && ( localKeyID != NULL ) )
		{
			int res = ska_generateLocalKeyID( wBlock.ska, localKeyID, pLocalKeyIDBuffSize );
			return res;			
		}

		return x509e_ok;
	}
	else
		return ppe_internalError;
}