#include "resource.h"
#include "BCKeyMan.h"
#include "PublicKeyProps.h"
#include "PKParse.h"
#include "BCDB_Functions.h"

#include "..\JpegRead\jpegread.h"

#include "..\CertificateLib\PGPPkts\PGPPkts.h"
#include "..\CertificateLib\PGPPkts\PKTSignature.h"
#include "..\CertificateLib\PGPPkts\Packets.h"
#include "..\CertificateLib\SMemory\SBlocks.h"
#include "..\CertificateLib\PGPPkts\PKTSignature.h"

#include "..\Includes\cDlg.h"
#include "..\Includes\FileView.h"
#include "..\Includes\OpenFiles.h"
#include "..\Includes\menuItem.h"

#include <stdio.h>
#include <time.h>
#include <assert.h>
#include <windowsx.h>
#include <commctrl.h>

class cDlgKeyProperties : public cColoredDlg
{

public:
	//
	//	User key inforamtion...
	//
	int keySize, userIDSize, //signatureSize,
		imgSize;
	BYTE *keyPtr, // *sigPtr,
		*userIDPtr, *imgPtr;

	cpgp_pkt *sigList, *curSig;
	cpgp_keyInfo keyInfo;

	BYTE *m_keyID;
	BYTE *m_krPkt;
	int m_krSize;
	bool m_secret;

	//
	//	Signature issuer information...
	//
	bool m_issuerUndefined; // if true button Show issuer works as Find issuer.
	cFileView m_issuerKRFileView;
	cPlainTextBlock m_ptbIssuerKey;
	BYTE *m_signatureIssuerKR;
	int m_signatureIssuerKRSize;
	BYTE *m_signatureIssuerID;
	BYTE *m_issuerUserIDPtr;
	int m_issuerUserIDSize;
	BYTE *m_signatureIssuerKey;
	int m_signatureIssuerKeySize;


	//
	//	User image objects...
	//
	HWND hImg;
	cPlainTextBlock ptbBmp;
	BITMAPINFOHEADER bmiHeader;


 public:

	enum { ID = IDD_DlgPublicKeyProperties, IDSmall = IDD_DlgPublicKeyPropertiesSmall };
	cDlgKeyProperties( BYTE *keyID, BYTE *krPkt, int krSize, bool secret );
	~cDlgKeyProperties();

 private:

	LRESULT RealDlgProc( UINT uMsg, WPARAM wParam, LPARAM lParam );
	int VerifyKey();
	BOOL InitDialog();

	BOOL FillSignatureInfo( int num );
	int  FillIssuerInfo( BYTE *krPkt, int krSize );
	BOOL FindSignatureIssuerKey();
	void updateSpinRange();

	void PrepareBitmap();

	void OnPaint();
};

cDlgKeyProperties::cDlgKeyProperties( BYTE *keyID, BYTE *krPkt, int krSize, bool secret )
{
	m_keyID = keyID;
	m_krPkt = krPkt;
	m_krSize = krSize;
	m_secret = secret;

	curSig = NULL;


	//
	//	Checks if user image exists
	//
	cpgp_keyInfo *pki = NULL;

	int res = pgp_FindKeyByIDEx(
				&keyInfo,
				m_keyID,
				m_krPkt, m_krSize,
				m_secret,
				&pki );

	if ( res != ppe_ok )
		return;

	if ( pki == NULL )
		pki = &keyInfo;

	if ( pki->img )
	{
		imgPtr  = pki->img->body;
		imgSize = pki->img->size;
	}
	else
	{
		imgPtr  = NULL;
		imgSize = NULL;
	}
}

cDlgKeyProperties::~cDlgKeyProperties()
{
//	if ( m_issuerKRFileView != NULL )
//		delete m_issuerKRFileView;

	m_issuerKRFileView.Close();
}

static BYTE endZero[1] = { 0 };

static void MakeKeyIDStr( cPlainTextBlock *ptb, BYTE *keyID )
{
	ptb->Allocate( pgp_keyIDSize * 3 );

	int c = 0;

	char str[4];

	sprintf( str, "%2X", 0xFF & (int)keyID[c] );
	ptb->WriteBytes( (BYTE *)str, 2, true );

	for ( ; ++c < pgp_keyIDSize; )
	{
		sprintf( str, " %02X", 0xFF & (int)keyID[c] );
		ptb->WriteBytes( (BYTE *)str, 3 );
	}

	ptb->WriteBytes( endZero, 1 );
}

static void MakeTimeString( char *str, long dwTime )
{
	time_t time = ( time_t )dwTime;
	sprintf( str, ctime( &time ) );	
}

static BOOL MakeAlgString( cPlainTextBlock *ptb, BYTE *key, int keySize )
{
	ptb->WriteBytes( endZero, 1, true );

	BYTE version = key[0];
	const char *algName;
	DWORD size;

	if ( version > 3 )
	{
		sPublicKey *pk = (sPublicKey *)key;

		if ( PKT_PublicKey_Parse( pk, keySize ) != ppe_ok )
			return FALSE;

		algName = PKT_GetPublicKeyAlgName( pk->algID );
		size = ((DWORD)pk->mpi[0] << 8 ) | pk->mpi[1];
	}
	else
	{
		sPublicKeyV3 *pk = (sPublicKeyV3 *)key;

		if ( PKT_PublicKeyV3_Parse( pk, keySize ) != ppe_ok )
			return FALSE;

		algName = PKT_GetPublicKeyAlgName( pk->algID );
		size = ((DWORD)pk->mpi[0] << 8 ) | pk->mpi[1];
	}

	ptb->Allocate( strlen( algName ) + 10 );
	sprintf( (char *)ptb->buffer, "%s - %d", algName, size );

	return TRUE;
}

int cDlgKeyProperties::FillIssuerInfo( BYTE *krPkt, int krSize )
{
	m_issuerUserIDPtr = NULL;
	m_issuerUserIDSize = 0;

	cpgp_keyInfo keyInfo, *pki;

	int res = pgp_getValidSize( krPkt, &krSize );

	if ( res == ppe_ok )
		res = pgp_FindKeyByIDEx(
				&keyInfo,
				m_signatureIssuerID,
				krPkt, krSize,
				m_secret,
				&pki
			);

	if ( res != ppe_ok )
	{
		SetDlgItemTextT( IDC_stxVerification, IDS_szIssuerNotFound );

		m_issuerUndefined = true;

		ShowWindow( GetDlgItem( m_hDlg, IDC_rbFindLocal ), SW_SHOW );
		ShowWindow( GetDlgItem( m_hDlg, IDC_rbBrowseNet ), SW_SHOW );

		SetDlgItemTextT( IDC_btnShowIssuer, IDS_szBtnFindIssuer );
		EnableWindow( GetDlgItem( m_hDlg, IDC_btnShowIssuer ), TRUE );
	}

	if ( res == ppe_keyNotFound )
		return res;
	else
	if ( res != ppe_ok )
	{
		DlgMessageBoxT( IDS_KeyRingReadError );
		return res;
	}

//	if ( pki == NULL )
		pki = &keyInfo;

	while( memcmp( pki->keyID, m_signatureIssuerID, pgp_keyIDSize ) != 0 )
	{
		if ( pki->next == NULL )
			return ppe_internalError;

		pki = pki->next;
	}
	
	m_signatureIssuerKey  = pki->keyPkt;
	m_signatureIssuerKeySize = pki->keyPktSize;

	if ( pki->userID )
	{
		m_issuerUserIDPtr  = pki->userID->body;
		m_issuerUserIDSize = pki->userID->size;
	}
	else
	{
		m_issuerUserIDPtr  = NULL;
		m_issuerUserIDSize = NULL;
	}

	//cPlainTextBlock ptb;

	if ( res != ppe_ok )
		return res;
	
	//if ( ptb.Assign( m_issuerUserIDPtr, m_issuerUserIDSize ) == NULL )
	//	return ppe_memoryAllocation;

	//ptb.WriteBytes( endZero, 1 );

	//SetDlgItemTextA( m_hDlg, IDC_stxIssuerID, ( char * )ptb.buffer );
	cWCStr wcs;
	wcs.Set( (char*)m_issuerUserIDPtr, m_issuerUserIDSize, CP_UTF8 );
	::SetDlgItemTextW( m_hDlg, IDC_stxIssuerID, wcs.m_szBody );

	m_signatureIssuerKR = krPkt;
	m_signatureIssuerKRSize = krSize;

	m_issuerUndefined = false;

	res = VerifyKey();

	if ( res == ppe_ok ) // The issuer is valid we can show his/her info
	{
		ShowWindow( GetDlgItem( m_hDlg, IDC_rbFindLocal ), SW_HIDE );
		ShowWindow( GetDlgItem( m_hDlg, IDC_rbBrowseNet ), SW_HIDE );

		SetDlgItemTextT( IDC_btnShowIssuer, IDS_szBtnShowIssuer );
		EnableWindow( GetDlgItem( m_hDlg, IDC_btnShowIssuer ), TRUE );
	}

	return res;
}

BOOL cDlgKeyProperties::FillSignatureInfo( int num )
{
	cpgp_pkt *prev = curSig;

	for( curSig = sigList; curSig; curSig = curSig->next )
	{
		if ( --num <= 0 )
			break;
	}

	if ( curSig == 0 )
		return FALSE;

	if ( curSig == prev ) // Already filled.
		return true;

	BYTE version = curSig->body[0];

	BYTE sigType;

	if ( version <= 3 )
	{
		sSignatureV3 *sig = ( sSignatureV3 *) curSig->body;

		sigType = sig->type;

		char str[80];

		MakeTimeString( str, sig->time.getTime() );
		SetDlgItemTextA( m_hDlg, IDC_stxSigTime, str );

		SetDlgItemTextA( m_hDlg, IDC_stxHashAlg, PKT_GetHashAlgName( sig->hashAlgID ) );

		m_signatureIssuerID = sig->keyID;
	}
	else
	if ( version == 4 )
	{
		sSignature *sig = ( sSignature *) curSig->body;

		sigType = sig->type;

		BYTE *spHashed;
		WORD spHashedSize;
		BYTE *spUnhashed;
		WORD spUnhashedSize;
		BYTE *hashAlders;
		BYTE *MPIs;

		int res = PKT_Signature_Parse( sig, curSig->size, &spHashed, &spHashedSize, &spUnhashed, &spUnhashedSize, &hashAlders, &MPIs );
		if ( res != ppe_ok )
			return res;

		char str[80];

		sTime *time = PKT_findSignatureTime( spHashed, spHashedSize );

		if ( time == NULL )
			return FALSE;

		MakeTimeString( str, time->getTime() );
		SetDlgItemTextA( m_hDlg, IDC_stxSigTime, str );

		SetDlgItemTextA( m_hDlg, IDC_stxHashAlg, PKT_GetHashAlgName( sig->hashAlgID ) );

		m_signatureIssuerID = PKT_findIssuerKeyID ( spUnhashed, spUnhashedSize );

		if ( m_signatureIssuerID == NULL )
			return FALSE;

	}
	else
		return FALSE;
	
	SetDlgItemTextA( m_hDlg, IDC_SigType, PKT_Signature_GetTypeName( sigType ) );

	//
	//	Checks issuer information...
	//
	if ( memcmp( m_signatureIssuerID, m_keyID, pgp_keyIDSize ) == 0 )
	{
		// 1. Self signed key

		m_signatureIssuerKey = keyPtr;
		m_signatureIssuerKeySize = keySize;

		SetDlgItemTextT( IDC_stxIssuerID, IDS_szSelfSigned );

		ShowWindow( GetDlgItem( m_hDlg, IDC_rbFindLocal ), SW_HIDE );
		ShowWindow( GetDlgItem( m_hDlg, IDC_rbBrowseNet ), SW_HIDE );

		SetDlgItemTextT( IDC_btnShowIssuer, IDS_szBtnShowIssuer );

		EnableWindow( GetDlgItem( m_hDlg, IDC_btnShowIssuer ), false );

		m_issuerUndefined = false;

		return VerifyKey() == ppe_ok;
	}

	//	2. Checks if issuer key is placed in same key ring.

	if ( FillIssuerInfo( m_krPkt, m_krSize ) == ppe_ok )
		return TRUE;

	//	2.1 Tryes to find issuer in BCDB local public key ring
	wchar_t path[ MAX_PATH ];
	if ( BCDB_GetLocalKeyRingPath( path, maxLen( path ), false ) )
	{
		if ( !m_issuerKRFileView.Open( path ) )
		{
			DlgMessageBoxT( IDS_KeyRingReadError, MB_ICONSTOP );
			return false;
		}

		int res = FillIssuerInfo( m_issuerKRFileView.m_ptr, (int)m_issuerKRFileView.m_size.QuadPart );
		if ( res == ppe_ok )
			return true;
	}

	//	3. We can try to find issuer key in other key ring.

	cPlainTextBlock ptb;
	MakeKeyIDStr( &ptb, m_signatureIssuerID );
	SetDlgItemTextA( m_hDlg, IDC_stxIssuerID, ( char * )ptb.buffer );
	
	//	3.1. Show issuer button is changed to Find issuer

	m_issuerUndefined = true;

	SetDlgItemTextT( IDC_btnShowIssuer, IDS_szBtnFindIssuer );

	ShowWindow( GetDlgItem( m_hDlg, IDC_rbFindLocal ), SW_SHOW );
	ShowWindow( GetDlgItem( m_hDlg, IDC_rbBrowseNet ), SW_SHOW );

	return TRUE;
}


void cDlgKeyProperties::OnPaint()
{
	HWND hwnd = GetDlgItem( m_hDlg, IDC_imgUser );

	if ( hwnd == NULL )
		return;

	PAINTSTRUCT ps; 
	HDC dc;
	dc = BeginPaint(hwnd, &ps); 

//	HDC memDC = CreateCompatibleDC(dc);

//	SetMapMode( memDC , GetMapMode(dc) );
       
    // Drawing background

    RECT rect;
    GetWindowRect( hwnd , &rect );
    rect.right = rect.right - rect.left - 4;
    rect.bottom = rect.bottom - rect.top - 4;
    rect.left = rect.top = 0;
   
    
	// Brush for filled rectangle

	if ( ptbBmp.bufferSize != 0 )
	{
//		static raster = SRCCOPY;
//		BOOL res = BitBlt( dc, 0, 0, rect.right, rect.bottom, memDC, 0, 0, raster );

		int lines = StretchDIBits(
						dc,
						0, rect.bottom-1,
						rect.right, -rect.bottom,
						0,0,
						bmiHeader.biWidth, bmiHeader.biHeight,
						ptbBmp.buffer,
						(LPBITMAPINFO)&bmiHeader,
						DIB_RGB_COLORS,
						SRCCOPY );
/*
		int lines = SetDIBitsToDevice(
						dc,
						0, rect.bottom-1,
						rect.right, -rect.bottom,
						0,0,
						bmiHeader.biWidth, bmiHeader.biHeight,
						ptbBmp.buffer,
						(LPBITMAPINFO)&bmiHeader,
						DIB_RGB_COLORS );
*/
	}
	else
	{
		cWCStr wcs;
		wcs.LoadStringT( IH.m_hInst, IDS_szNoImage );

		int length = wcslen( wcs.m_szBody );

		HFONT font;

		font = CreateFontA( 
			35		,0		,0		,0			,FW_BOLD	,TRUE,		FALSE,	0,			
			//nHeight,nWidth,nEscapement,nOrientation,nWeight,bItalic,bUnderline,cStrikeOut,
			DEFAULT_CHARSET, OUT_DEFAULT_PRECIS,CLIP_DEFAULT_PRECIS,PROOF_QUALITY,
			//nCharSet,  nOutPrecision, nClipPrecision,			 BYTE nQuality,
			FF_DONTCARE,"Times New Roman");
			//BYTE nPitchAndFamily, LPCTSTR lpszFacename );

		SelectObject(dc,font);

		int x = 35,y = 85;

		SetBkMode( dc, TRANSPARENT );
		SetTextColor( dc, RGB(0,0,0) );
		TextOutT( dc,x + 4,y + 4, wcs.m_szBody, length );

		SetTextColor( dc, RGB(0xff,0xff,0xff) );
		TextOutT( dc, x, y, wcs.m_szBody, length );
	}

	EndPaint(hwnd, &ps); 
}

void cDlgKeyProperties::PrepareBitmap()
{
	if ( imgPtr == NULL )
		return;

	cJPegRead jr( imgPtr + 0x16, imgSize - 0x16 );

	if ( jr.err )
		return;

	//
	//	The bitmap width has to be DWORD alligned. 
	//	
	int dwAllignedWidth =  (jr.width * jr.bytesPerPixel + 3 ) & ~3;

	if ( ptbBmp.Allocate( jr.height * dwAllignedWidth ) == NULL )
		return;

	if ( jr.ScanLines( ptbBmp.buffer, jr.height, dwAllignedWidth ) != jr.height )
		return;

	bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
	bmiHeader.biWidth = jr.width;
	bmiHeader.biHeight = jr.height;
	bmiHeader.biPlanes = 1;
	bmiHeader.biBitCount = 24;
	bmiHeader.biCompression = BI_RGB;
	bmiHeader.biSizeImage = 0;
	bmiHeader.biXPelsPerMeter = 0;
	bmiHeader.biYPelsPerMeter = 0;
	bmiHeader.biClrUsed = 0;
	bmiHeader.biClrImportant = 0;

	jr.SwapRGB( ptbBmp.buffer, dwAllignedWidth, jr.height );
}

#define SpinSetRange( hWnd, nLower, nUpper) SendMessage( hWnd, UDM_SETRANGE, 0, MAKELPARAM(nUpper, nLower))
#define SpinSetPos( hWnd, nPos ) SendMessage( hWnd, UDM_SETPOS, 0, MAKELPARAM( 0, nPos ))
#define SpinGetPos( hWnd ) SendMessage( hWnd, UDM_GETPOS, 0, 0 )

void cDlgKeyProperties::updateSpinRange()
{
	cpgp_pkt *tmp = sigList;
	int q = 0;

	for( ; tmp; tmp = tmp->next )	q++;

	HWND hWnd = GetDlgItem( m_hDlg, IDC_SigSpin );

	SpinSetRange( hWnd, 1, q );
	SpinSetPos( hWnd, 0 );
}

BOOL cDlgKeyProperties::InitDialog()
{
	if ( m_secret )
		SetWindowTextT( m_hDlg, _LoadIDS( IDS_SKProperties ) );

//	CenterPosition();
	SendMessage( m_hDlg, WM_SETICON, 0, (LPARAM)LoadIcon( m_hInst, MAKEINTRESOURCE( IDI_Main ) ) );
	SetRCValues( IDC_Line );


	//
	//	1. Gets key info
	//

	cpgp_keyInfo *pki = NULL;

	int res = pgp_getValidSize(	m_krPkt, &m_krSize );

	if ( res == ppe_ok )
		res = pgp_FindKeyByIDEx(
				&keyInfo,
				m_keyID,
				m_krPkt, m_krSize,
				m_secret,
				&pki );

	if ( res != ppe_ok )
	{
		DlgMessageBoxT( IDS_KeyRingReadError );
		return FALSE;
	}

	bool isSubKey = true;

	if ( pki == NULL )
		pki = &keyInfo;
	else
		isSubKey = true;

	keyPtr  = pki->keyPkt;
	keySize = pki->keyPktSize;

	sigList = pki->sig;
	

	if ( pki->userID )
	{
		userIDPtr  = pki->userID->body;
		userIDSize = pki->userID->size;
	}
	else
	{
		userIDPtr  = NULL;
		userIDSize = NULL;
	}

	if ( pki->img )
	{
		imgPtr  = pki->img->body;
		imgSize = pki->img->size;
	}
	else
	{
		imgPtr  = NULL;
		imgSize = NULL;
	}


	//
	//	2. Fills user and key information
	//

	BYTE version = keyPtr[0];

	char szVer[2] = "4";
	szVer[0] = '0' + version;

	SetDlgItemTextA( m_hDlg, IDC_keyVersion, szVer );

	sTime time = ((sPublicKey *)keyPtr)->time;
	char szTime[100];
	MakeTimeString( szTime, time.getTime() );
	SetDlgItemTextA( m_hDlg, IDC_keyTime, szTime );


	if ( userIDPtr != NULL )
	{
		cWCStr wcs( (char*)userIDPtr, userIDSize, false, CP_UTF8 );
		SetDlgItemTextT( IDC_stxUserID, wcs.m_szBody );
	}
	else
	{
		if ( isSubKey && keyInfo.userID )
		{
			cWCStr wcs( _LoadIDS( IDS_SubkeyOf ) );
			wcs.Add( L"\n" );
			wcs.Add( (char*)keyInfo.userID->body, keyInfo.userID->size, CP_UTF8 );

			SetDlgItemTextT( IDC_stxUserID, wcs.m_szBody );
		}
		else
			SetDlgItemTextT( IDC_stxUserID, IDS_Undefined );
	}

	cPlainTextBlock ptb;

	MakeKeyIDStr( &ptb, m_keyID );
	SetDlgItemTextA( m_hDlg, IDC_stxKeyID, (char *)ptb.buffer );

	MakeAlgString( &ptb, keyPtr, keySize );
	SetDlgItemTextA( m_hDlg, IDC_stxAlgorithm, (char *)ptb.buffer );
	

	//
	//	3. Fills signature information
	//

	updateSpinRange();

	Button_SetCheck( GetDlgItem( m_hDlg, IDC_rbFindLocal ), BST_CHECKED );

	if ( !FillSignatureInfo( 0 ) )
		return FALSE;

	//
	//	4. Creates image bitmap.
	//
	if ( imgPtr != NULL )
	{
		PrepareBitmap();
	}

	return TRUE;
}

BOOL cDlgKeyProperties::FindSignatureIssuerKey()
{
	int res;

	BOOL bFindLocaly = Button_GetCheck( GetDlgItem( m_hDlg, IDC_rbFindLocal ) ) == BST_CHECKED;

	if ( bFindLocaly )
	{
		wchar_t path[ MAX_PATH ];
			path[0] = L'\0';

		wchar_t szFiltersPublic[ 1024 ];
		LoadStringT( m_hInst, IDS_szFiltersPublicPGP, szFiltersPublic, maxLen( szFiltersPublic ) );		
		//char *szFilters = 
		//	"Public key ring ( *.pkr,*gpg )\0*.pkr;*gpg\0"
		//	"All files ( *.* )\0*.*\0";

		if ( !BrowseFiles( m_hDlg, path, maxLen( path ), szFiltersPublic, _LoadIDS( IDS_OpenIssuerKR ), false ) )
			return false;

		if ( !m_issuerKRFileView.Open( path ) )
		{
			DlgMessageBoxT( IDS_KeyRingReadError, MB_ICONSTOP );
			return false;
		}

		res = FillIssuerInfo( m_issuerKRFileView.m_ptr, (int)m_issuerKRFileView.m_size.QuadPart );
	}
	else // Browse in net
	{
//		cPlainTextBlock ptbMessage;

//		if ( SearchPKUser( m_hInst, m_hDlg, &ptbMessage, m_signatureIssuerID ) != ppe_ok )
//			return false;

//		char *headers;
//		res = DecodePGPPublicKeyBlock( ptbMessage.buffer, ptbMessage.bufferSize, &m_ptbIssuerKey, NULL, &headers );
//		if ( res == ppe_ok )

		if ( SearchPKUserT( m_hInst, m_hDlg, &m_ptbIssuerKey, m_signatureIssuerID ) != ppe_ok )
			return false;

		res = FillIssuerInfo( m_ptbIssuerKey.buffer, m_ptbIssuerKey.bufferSize );
	}

	if ( res == ppe_invalidSignature )
		return DlgMessageBoxT( IDS_InvalidSignature, MB_ICONSTOP );

	if ( res == ppe_keyNotFound )
		return DlgMessageBoxT( IDS_IssuerKeyNotFound, MB_ICONSTOP );

	if ( res != ppe_ok )
		return DlgMessageBoxT( IDS_KeyRingReadError, MB_ICONSTOP );
	else
		return true;
}

int cDlgKeyProperties::VerifyKey()
{
	if ( ( curSig == NULL ) || ( m_signatureIssuerKey == NULL ) )
	{
		SetDlgItemTextT( IDC_stxVerification, IDS_szInternalError );
		assert( false );
		return ppe_internalError;
	}

	SetDlgItemTextT( IDC_stxVerification, IDS_szPleaseWaitForVerification );

	EnableWindow( GetDlgItem( m_hDlg, IDC_SigSpin ), FALSE );

	BYTE version = curSig->body[0];

	cHashedList hashList;

	BYTE sigType = pgp_GetSigType( curSig->body, curSig->size );

	int res;

	switch ( sigType )
	{
		case sig_GenericCertificationOfAUserIDAndPublicKeyPacket:
		case sig_PersonaCertificationOfAUserIDAndPublicKeyPacket:
		case sig_CasualCertificationOfAUserIDAndPublicKeyPacket:
		case sig_PositiveCertificationOfAUserIDAndPublicKeyPacket:
			res = pgp_MakeKeySignature_0x10_HashList(
					&hashList,
					keyPtr, keySize,
					userIDPtr, userIDSize,
					version );

			break;
		case sig_SubkeyBindingSignature:
			//
			//	For the signature type issuer is main key and the key is a subkey
			//	See RFC 2440 5.2.4. Computing Signatures ( type 0x18 )
			//
			res = pgp_AddKeyPktToHashList(
					&hashList,
					m_signatureIssuerKey,
					m_signatureIssuerKeySize );

			if ( res == ppe_ok )
				res = pgp_AddKeyPktToHashList(
						&hashList,
						keyPtr, keySize );
/*
			if ( res == ppe_ok )
				res = PKT_VerifySignatureEx(
					sigPtr, signatureSize,
					&hashList,
					m_signatureIssuerKey, m_signatureIssuerKeySize );
*/
			break;
		case sig_SignatureDirectlyOnAKey:
			res = pgp_AddKeyPktToHashList(
					&hashList,
					keyPtr, keySize );
			break;

		default:
			res = ppe_invalidSignatureType;
	}


	if ( res == ppe_ok )
		res = pgp_VerifySignaturePacket(
				curSig->body, curSig->size,
				&hashList,
				m_signatureIssuerKey, m_signatureIssuerKeySize );

	EnableWindow( GetDlgItem( m_hDlg, IDC_SigSpin ), TRUE );

	if ( res != ppe_ok )
		SetDlgItemTextT( IDC_stxVerification, IDS_szInvalid );
	else
		SetDlgItemTextT( IDC_stxVerification, IDS_szVerified );

	return res;
}

LRESULT cDlgKeyProperties::RealDlgProc( UINT uMsg, WPARAM wParam, LPARAM lParam )
{
	LRESULT result;
	
	if ( cMenuItem::DlgProc( &result, m_hDlg, uMsg, wParam, lParam ) )
		return result;

	switch(uMsg)
	{
		case WM_INITDIALOG:
			if ( !InitDialog() )
			{
				EndDialog( m_hDlg, IDCANCEL );
			}
			break;

		case WM_COMMAND:
		{
			//
			//	Check if menu command...
			//
			if ( HIWORD( wParam ) == 0 )
			{
				switch ( LOWORD( wParam ) )
				{
					case IDC_btnShowIssuer:
						if ( m_issuerUndefined )
							return FindSignatureIssuerKey();
						else
							return ShowKeyPropertiesEx(
										m_hInst,m_hDlg,
										m_signatureIssuerID,
										m_signatureIssuerKR, m_signatureIssuerKRSize );

					case IDOK:
					case IDCANCEL:
						EndDialog( m_hDlg, LOWORD( wParam ) );
						return true;
					default:
						return false;
				}
			}
			else
			if ( ( HIWORD( wParam ) == EN_CHANGE ) &&  ( LOWORD( wParam ) == IDC_edtSigNumber ) )
			{
				HWND hWnd = GetDlgItem( m_hDlg, IDC_SigSpin );
				int pos = SpinGetPos( hWnd );

				if ( HIWORD( pos ) != 0 ) // An error
					return false;

				FillSignatureInfo( pos );
			}
		}
		case WM_PAINT:
		{
			OnPaint();
			return 0;
		}
		default:
			return false;
	}

	return true;
}

int ShowKeyProperties( HINSTANCE hInst, HWND hWnd, BYTE *keyID, const char *keyRingPath, bool secret )
{
	if ( hInst == NULL )
		hInst = hMainInst; // GetModuleHandle( "BCKeyMan.dll" );

	cFileView kr( keyRingPath );

	if ( kr.m_err )
		return FALSE;

	return ShowKeyPropertiesEx( hInst, hWnd, keyID, kr.m_ptr, (int)kr.m_size.QuadPart, secret );
}

int ShowKeyPropertiesEx( HINSTANCE hInst, HWND hWnd, BYTE *keyID, BYTE *krPkt, int krSize, bool secret )
{
	if ( hInst == NULL )
		hInst = hMainInst; // GetModuleHandle( "BCKeyMan.dll" );

	cDlgKeyProperties dlg( keyID, krPkt, krSize, secret );

	dlg.DoModal( hInst, dlg.imgPtr == NULL ? dlg.IDSmall : dlg.ID, hWnd );

	return TRUE;
}