#include "cDlgBrowseKeyFile.h"
#include "BCDB_Functions.h"
#include "cWizardDlgs.h"
#include "resource.h"

#include "..\CertificateLib\SMemory\SBlocks.h"
#include "..\CertificateLib\PGPPkts\Packets.h"

#include "..\Includes\EquStrings.h"
#include "..\Includes\ptbRegs.h"
#include "..\Includes\UnicodeFuncs.h"
#include "..\Includes\FileView.h"
#include "..\Includes\menuItem.h"
#include "..\Includes\TranslateDll.h"

#include <stdio.h>
#include <windowsx.h>

#include "..\Includes\ODS.h"

enum { IDM_Browse = 200, IDM_Up, IDM_Delete, IDM_Find, IDM_GenerateNew };

static DWORD GetIDSbyCmd( int cmd )
{
	switch ( cmd )
	{
		case IDM_Browse : return IDS_szBrowse;
		case IDM_Up		: return IDS_szParentFolder;
		case IDM_Delete : return IDS_szDelete;
		case IDM_Find	: return IDS_szFind;
		case IDM_GenerateNew : return IDS_szGenerateNew;

		default : return NULL;
	}
}

int cDlgBrowseKeyFile::InitToolBar()
{
	INITCOMMONCONTROLSEX commCtrl;

	commCtrl.dwSize = sizeof( INITCOMMONCONTROLSEX );
	commCtrl.dwICC = 
		ICC_BAR_CLASSES |
		ICC_LISTVIEW_CLASSES |
		ICC_PROGRESS_CLASS |
		ICC_TAB_CLASSES |
		ICC_TREEVIEW_CLASSES;

	InitCommonControlsEx( &commCtrl );

	HWND hPath = GetDlgItem( m_hDlg, IDC_ebPath );

	RECT r;
	GetWindowRect( hPath, &r );
//	POINT ptTBPos = { r.right + 2, r.top };
//	ScreenToClient( m_hDlg, &ptTBPos );
	r.right += 2;
	ScreenToClientRect( &r );
	POINT ptTBPos = { r.right + 2, r.top };

	m_hToolbar = CreateWindowExT( 
			WS_EX_TRANSPARENT |
			0 ,
			TOOLBARCLASSNAMEW, L"",
			CCS_NOPARENTALIGN |
			CCS_ADJUSTABLE |
			CCS_NOMOVEY |
			CCS_NODIVIDER |
			CCS_TOP |
			CCS_NORESIZE |
			TBSTYLE_FLAT |
			TBSTYLE_TOOLTIPS |
			TBSTYLE_TRANSPARENT |
			TBSTYLE_AUTOSIZE |
			WS_CHILD | WS_VISIBLE ,
			ptTBPos.x, ptTBPos.y, 150, r.bottom - r.top + 1,
			m_hDlg, NULL, m_hInst, 0 );


	if ( m_hToolbar == NULL )
		return 0;

	SendMessage( m_hToolbar, TB_SETEXTENDEDSTYLE, 0 , TBSTYLE_EX_DRAWDDARROWS | WS_EX_TRANSPARENT );
	SendMessage( m_hToolbar, TB_BUTTONSTRUCTSIZE, (WPARAM) sizeof(TBBUTTON), 0);

	// Add the bitmap containing button images to the toolbar. 

	TBADDBITMAP tbab;
	memset( &tbab, 0, sizeof( tbab ) );

	tbab.hInst = HINST_COMMCTRL ; 
	tbab.nID   = IDB_STD_SMALL_COLOR; 
	int idbStd = SendMessage( m_hToolbar, TB_ADDBITMAP, (WPARAM) 0, (WPARAM) &tbab); 

	tbab.hInst = HINST_COMMCTRL ; 
	tbab.nID   = IDB_VIEW_SMALL_COLOR; 
	int idbView = SendMessage( m_hToolbar, TB_ADDBITMAP, (WPARAM) 0, (WPARAM) &tbab); 

	tbab.hInst = HINST_COMMCTRL ; 
	tbab.nID   = IDB_HIST_SMALL_COLOR; 
	int idbHist = SendMessage( m_hToolbar, TB_ADDBITMAP, (WPARAM) 0, (WPARAM) &tbab); 

	tbab.hInst = m_hInst; 
	tbab.nID   = IDB_ToolbarXP; 
	int idbInternal = SendMessage( m_hToolbar, TB_ADDBITMAP, (WPARAM) 7, (WPARAM) &tbab); 

	if ( ( m_dwFlags & flag_DisableBrowseButtons ) == 0 )
	{
		// v. 1.04.6 Icon browse '...' is added as number 6 to toolbar.
		ToolBar_AddButton( m_hToolbar, idbInternal + 6, IDM_Browse );
		ToolBar_AddButton( m_hToolbar, idbView + VIEW_PARENTFOLDER     , IDM_Up );
		ToolBar_AddButton( m_hToolbar, 0, 0, TBSTYLE_SEP );
	}

	if ( m_dwFlags & flag_UseNewKeyButton )
		ToolBar_AddButton( m_hToolbar, idbInternal + 2, IDM_GenerateNew );
	if ( m_dwFlags & flag_UseAddKeyButton )
		ToolBar_AddButton( m_hToolbar, idbInternal + 1, IDM_Find, TBSTYLE_DROPDOWN );

	if ( m_dwFlags != 0 )
		ToolBar_AddButton( m_hToolbar, 0, 0, TBSTYLE_SEP );

	ToolBar_AddButton( m_hToolbar,idbStd + STD_DELETE      , IDM_Delete );

//	SendMessage(m_hToolbar, TB_AUTOSIZE, 0, 0);
//	SetWindowPos( m_hToolbar, HWND_BOTTOM, ptTBPos.x, ptTBPos.y, 100, 15, SWP_SHOWWINDOW );

	return 1;
}

static UINT ColumnLabelIDS[] = 
{
	IDS_szFileName, IDS_szUserName, IDS_szAlgorithm, IDS_LocalKeyID
};

const int NUM_COLUMNS = sizeof( ColumnLabelIDS ) / sizeof( UINT );

static int _gnColumnFmt[ NUM_COLUMNS ] =
{
	LVCFMT_LEFT, LVCFMT_LEFT, LVCFMT_LEFT, LVCFMT_LEFT
};

static int _gnColumnWidth[NUM_COLUMNS]=
{
	75, 168, 85, 80
};

bool cDlgBrowseKeyFile::InitListViewColumns()
{
	int i;
	LV_COLUMNW lvc;

	memset( &lvc, 0, sizeof( lvc ) );

	lvc.mask = LVCF_WIDTH;

	//
	//	Check if already initialised...
	//
	if ( ListView_GetColumnT( m_hList, 0, &lvc ) )
		return TRUE;

	lvc.mask = LVCF_FMT | LVCF_WIDTH | LVCF_TEXT | LVCF_SUBITEM;

	wchar_t text[ 80 ];

	int res;

	for(i=0; i<NUM_COLUMNS; i++)
	{
		LoadStringT( m_hInst, ColumnLabelIDS[i], text, maxLen( text ) );

		lvc.iSubItem=i;
		lvc.pszText = text;
		lvc.cx = _gnColumnWidth[i];
		lvc.fmt=_gnColumnFmt[i];

		res = ListView_InsertColumnT( m_hList, i, &lvc );

		if ( res == -1 )
		{
			MessageBoxT( m_hDlg, text, _LoadIDS( IDS_ErrorInAddLVColumn ), MB_OK );
			return FALSE;
		}
	}

	DWORD dwStyle = LVS_EX_FULLROWSELECT; // | LVS_EX_GRIDLINES;
	ListView_SetExtendedListViewStyleEx( m_hList, dwStyle, dwStyle );

	return TRUE;
}

static void SetItemValues( HWND hList, int item, wchar_t **subItemT, int subItemQ )
{
	for( int subItem = subItemQ; subItem; subItem -- )
		ListView_SetItemTextT( hList, item, subItem, subItemT[ subItem - 1 ] ); 
}

int cDlgBrowseKeyFile::AddLVItem( const wchar_t *path, WIN32_FIND_DATAW *findData, int itemNum )
{
	LVITEMW lvi;
	memset( &lvi, 0, sizeof( lvi ) );
	lvi.mask = LVIF_IMAGE | LVIF_TEXT | LVIF_PARAM;
	lvi.iItem = itemNum;
	lvi.lParam = findData->dwFileAttributes;

	const int subItemQ = 3;
	wchar_t *subItemT[ subItemQ ];
	wchar_t buff[ MAX_PATH ] = L"";
	wchar_t buff1[ MAX_PATH ] = L"";
	wchar_t algDesc[ MAX_PATH ] = L"";
	subItemT[0] = buff;
	subItemT[1] = algDesc;
	subItemT[2] = buff1;

	wchar_t text[ 1024 ];
	wcscpy_s( text, findData->cFileName );

	lvi.iImage = m_iconCache.GetIconNum( path, NULL );

	if ( ( findData->dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY ) == 0 )
	{
		cFileView fv( path );
		cUniversalString friendlyName;
		cPlainTextBlock ptbLocalKeyID;
		wchar_t errStr[ 1024 ];
		LoadStringT( m_hInst, IDS_NoError, errStr, maxLen( errStr ) );

		HPKA pka = NULL;

		bool res = BCDB_GetP12SecretKeyInfo(
			m_hInst, m_hDlg,
			fv.m_ptr, (int)fv.m_size.QuadPart,
			NULL,
			&friendlyName, &ptbLocalKeyID,
			errStr, maxLen( errStr )
			);


		if ( res )
		{
			// The secret key file was succefully read, so we try to find corresponded
			//	certificate file to get algorithm descryption ( secret key is encrypted and has no
			//	opened descryption )

			cPlainTextBlock ptbCrt;

			bool lRes = BCDB_GetX509KeyByID( &ptbCrt, ptbLocalKeyID.buffer, ptbLocalKeyID.bufferSize, false, NULL );

			if ( lRes )
			{
				if ( !BCDB_GetCerPublicKeyInfo(
						m_hInst, m_hDlg,
						ptbCrt.buffer, ptbCrt.bufferSize,
						&pka )
					)
						wcscpy_s( algDesc, _LoadIDS( IDS_Undefined ) );
			}
		}
		else // Tries to read the file as certificate
		{
			res = BCDB_GetCerPublicKeyInfo(
				m_hInst, m_hDlg,
				fv.m_ptr, (int)fv.m_size.QuadPart,
				&pka, // NULL,
				&friendlyName, &ptbLocalKeyID );
		}

		//
		//	Gets algorithm description from public key algorithm handle.
		//
		if ( pka != NULL )
		{
			int algID = pka_getAlgID( pka );
			int bitSize = pka_getModuleSize( pka ) * 8;
			pka_free( pka );

			FormatT( algDesc, maxLen( algDesc ), _LoadIDS( IDS_NameAndBits_SD ), cWCStr( PKT_GetPublicKeyAlgName( (BYTE)algID ) ).m_szBody, bitSize );
		}

		if ( !res )
		{
			SHFILEINFOW sfi;
			if ( SHGetFileInfoT( path, 0, &sfi, sizeof( sfi ), SHGFI_DISPLAYNAME | SHGFI_TYPENAME ) != 0 )
			{
				wcsncpy( subItemT[0], sfi.szTypeName, maxLen( sfi.szTypeName ) );
			}
			else
				wcsncpy( subItemT[0], errStr, maxLen( buff )  );
		}
		else
			wcsncpy( subItemT[0], friendlyName.wcBuffer, maxLen( buff ) );

		char buffA[ 1024 ];
		int size = sizeof( buffA );
		bytes2string( buffA, &size, ptbLocalKeyID.buffer, ptbLocalKeyID.bufferSize );
		mbstowcs( buff1, buffA, maxLen( buff1 ) );
	}
	
	lvi.pszText = text;

	int i = ListView_InsertItemT( m_hList, &lvi );

	SetItemValues( m_hList, i, subItemT, subItemQ );

	return i;
}

static bool isCompatiblePattern( const wchar_t *fileName, const wchar_t *p )
{

	// Pattern sample:
	//	p = "*.gpg;*.skr"

	wchar_t curPattern[ 128 ];

	for( int i = 0;; p++, i++ )
	{
		curPattern[i] = *p;

		if ( curPattern[i] == L';' )
			curPattern[i] = L'\0';

		if ( curPattern[i] == L'\0' )
		{
			int ok = EquStrings( fileName, curPattern );

			if ( ok )
			{
				return true;
			}

			i = 0;

			if ( *p == L'\0' )
				break;
		}
	}

	return false;
}

bool cDlgBrowseKeyFile::UpdateListView( const wchar_t *folder, const wchar_t *pattern )
{
	ListView_DeleteAllItems( m_hList );

	wcsncpy( m_path, folder, MAX_PATH );
	wchar_t path[ MAX_PATH ];
	wcscpy_s( path, m_path );

	SetWindowTextT( GetDlgItem( m_hDlg, IDC_ebPath ), m_path );

	int len = wcslen( path );
	if ( path[ len - 1 ] != L'\\')
	{
		path[ len++ ] = L'\\';
		path[ len ] = L'\0';
	}

	// Fills directory first
	wcsncpy( path + len, L"*", maxLen( path ) - len );

	WIN32_FIND_DATAW  findData;
	memset( &findData, 0, sizeof( findData ) );

	HANDLE hFind = FindFirstFileT( path, &findData );
	if ( hFind == INVALID_HANDLE_VALUE )
		return false;

	int lastDirItem = 0;

	do
	{
		wcsncpy( path + len, findData.cFileName, maxLen( path ) - len );

		if ( ( findData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY ) != 0 )
		{
			if ( ( wcscmp( findData.cFileName, L"." ) == 0 ) ||
				( wcscmp( findData.cFileName, L".." ) == 0 ) )
				continue;

			AddLVItem( path, &findData, lastDirItem++ );
		}
		else
		if ( isCompatiblePattern( findData.cFileName, pattern ) )
		{
			AddLVItem( path, &findData, 0x7FFFffff ); // Add to the list end
		}
	}
	while( FindNextFileT( hFind, &findData ) != 0 );

	FindClose( hFind );

	return true;
}

#ifndef INVALID_FILE_ATTRIBUTES
#define INVALID_FILE_ATTRIBUTES (-1)
#endif

bool cDlgBrowseKeyFile::SetCurrentFolder( const wchar_t *folder, const wchar_t *pattern )
{
	DWORD attr = GetFileAttributesT( folder );
	if ( attr == INVALID_FILE_ATTRIBUTES )
	{
		assert( false );
		return false;
	} 

	if ( ( attr & FILE_ATTRIBUTE_DIRECTORY ) == 0 )
	{
		assert( false );
		return false;
	}

	return UpdateListView( folder, pattern );	
}

class cBrowseKeyFileValues : public cRegDlgValues
{
public:
	wchar_t m_path[ MAX_PATH ];
	int m_filterID;

	cBrowseKeyFileValues( HKEY hRoot, const wchar_t *keyName, const wchar_t *valueName, int columns )
		: cRegDlgValues( hRoot, keyName, valueName, sizeof( *this ), columns )
	{
		memset( m_path, 0, sizeof( m_path ) );
	}

	bool Save( HWND dlg, HWND listView )
	{
		GetWindowTextT( GetDlgItem( dlg, IDC_ebPath ), m_path, maxLen( m_path ) );
		HWND hCmb = GetDlgItem( dlg, IDC_cmbFileType );
		m_filterID = ComboBox_GetCurSel( hCmb );
		return cRegDlgValues::Save( dlg, listView );
	}
};

cBrowseKeyFileValues dlgVals( hKeyRoot, szBCKeyManKey, L"BrwsKF", NUM_COLUMNS ); 

bool cDlgBrowseKeyFile::InitDialog()
{
	SetRCValues( IDC_Line, IDC_STATIC_BOLD, IDC_Text, IDI_Main );

	if ( m_caption )
		SetWindowTextT( GetDlgItem( m_hDlg, IDC_STATIC_BOLD ), m_caption );

	if ( m_notes )
		SetWindowTextT( GetDlgItem( m_hDlg, IDC_Notes ), m_notes );

	if ( m_OKBtnText )
		SetWindowTextT( GetDlgItem( m_hDlg, IDOK ), m_OKBtnText );

//	SendMessage( m_hDlg, WM_SETICON, 0, (LPARAM)LoadIcon( m_hInst, MAKEINTRESOURCE( IDI_Main ) ) );

	InitToolBar();
	m_hList = GetDlgItem( m_hDlg, IDC_List );
	InitListViewColumns();
	m_iconCache.Init( m_hInst, m_hList, 0, 0 );
	InitFilters();

//	m_sizedCtrls.Init( m_hDlg );
	m_sizedCtrls.AddCtrl( IDC_Line, m_sizedCtrls.ptSR );
	m_sizedCtrls.AddCtrl( IDC_STATIC_BOLD,	cSizedCtrls::ptSR );
	m_sizedCtrls.AddCtrl( IDC_Notes,		cSizedCtrls::ptSR );
	m_sizedCtrls.AddCtrl( IDC_bmpHeader, m_sizedCtrls.ptPR );
	m_sizedCtrls.AddCtrl( IDC_List, m_sizedCtrls.ptSDR );
	m_sizedCtrls.AddCtrl( IDOK, m_sizedCtrls.ptPDR );
	m_sizedCtrls.AddCtrl( IDCANCEL, m_sizedCtrls.ptPDR );
	m_sizedCtrls.AddCtrl( IDC_ebPath, m_sizedCtrls.ptSR );
	m_sizedCtrls.AddCtrl( m_hToolbar, m_sizedCtrls.ptPR );
	m_sizedCtrls.AddCtrl( IDC_stFileName, m_sizedCtrls.ptPD );
	m_sizedCtrls.AddCtrl( IDC_stFileType, m_sizedCtrls.ptPD );
	m_sizedCtrls.AddCtrl( IDC_ebFileName, m_sizedCtrls.ptPDSR );
	m_sizedCtrls.AddCtrl( IDC_cmbFileType, m_sizedCtrls.ptPDSR );

	wchar_t dir[ MAX_PATH ];
	splitPath( m_path, dir, NULL );

	dlgVals.Restore( m_hDlg, m_hList );
	if ( ( wcslen( dlgVals.m_path ) != 0 ) &&
		 ( _waccess( dlgVals.m_path, 0 ) == 0 ) )
			wcscpy_s( dir, dlgVals.m_path );

	ComboBox_SetCurSel( GetDlgItem( m_hDlg, IDC_cmbFileType ), dlgVals.m_filterID );

	const wchar_t *filter = GetCurrentFilter();
	SetCurrentFolder( dir, filter );
	UpdateButtons();

	return true;
}


bool cDlgBrowseKeyFile::OnCommand( WPARAM wParam, LPARAM lParam )
{
	if ( (HWND)lParam == GetDlgItem( m_hDlg, IDC_cmbFileType ) )
	{
		if ( HIWORD(wParam) == CBN_CLOSEUP ) //CBN_SELENDOK ) // CBN_SELCHANGE )
		{
			const wchar_t *filter = GetCurrentFilter();
			SetCurrentFolder( m_path, filter );
			UpdateButtons();
			return true;
		}

		return false;
	}
	else
	if ( (HWND)lParam == m_hToolbar )
	{
		wchar_t dir[ MAX_PATH ];
		if ( wParam == IDM_Up )
		{
			splitPath( m_path, dir, NULL );
		}
		else
		if ( wParam == IDM_Browse )
		{
			wcscpy_s( dir, m_path );
			if ( !BrowseForFolderT( m_hInst, IDS_szBrowseForFolder, m_hDlg, dir ) )
				return false;
		}
		else
		if ( wParam == IDM_Delete )
		{
			if ( DeleteSelected() != 0 )
				UpdateButtons();

			return true;
		}
		else
		if ( wParam == IDM_GenerateNew )
		{
//			DlgMessageBox( IDS_szGenerateNew );
			GenerateKeyPair( m_hInst, m_hDlg );
			const wchar_t *filter = GetCurrentFilter();
			SetCurrentFolder( m_path, filter );
			return true;
		}
		else
		if ( wParam == IDM_Find )
		{
			int menuID = -1;

			if ( wParam == IDM_Find )
				menuID = 0;
			else
				return false;

			RECT rc;
			SendMessage( m_hToolbar, TB_GETRECT, (WPARAM)IDM_Find, (LPARAM)&rc );

			MapWindowPoints( m_hToolbar, HWND_DESKTOP, (LPPOINT)&rc, 2 );

			HMENU hPopupMenu = cMenuItem::LoadPopupMenu( m_hInst, IDR_MENU, menuID );

			cMenuItem::TrackPopupMenu(
				hPopupMenu,
				TPM_LEFTALIGN|TPM_LEFTBUTTON|TPM_VERTICAL,
				rc.left,
				rc.bottom,
				m_hDlg,
				NULL);

			return true;
		}
		else
			return false;

		const wchar_t *filter = GetCurrentFilter();
		SetCurrentFolder( dir, filter );
		UpdateButtons();

		return true;
	}

	switch ( LOWORD( wParam ) )
	{
		case IDM_Browse:
		{
			wchar_t dir[ MAX_PATH ];
			wcscpy_s( dir, m_path );

			if ( !BrowseForFolderT( m_hInst, IDS_szBrowseForFolder, m_hDlg, dir ) )
				return false;

			const wchar_t *filter = GetCurrentFilter();
			SetCurrentFolder( dir, filter );
			UpdateButtons();
			return true;
		}
		case IDCANCEL:
			EndDialog( m_hDlg, IDCANCEL );
			return true;
		case IDOK:
			return OnOK();
		case ID_ADDKEY_SEARCHINTHEINTERNET:
			return SearchPKUserT( m_hInst, m_hDlg, NULL ) == ppe_ok; // ptbKey is null so it will be saved to local database only

		case ID_ADDKEY_BROWSEFORFILES:
		{	
			type_kff kff;
			cPlainTextBlock ptbKey;

			bool res = BrowseForPKBlockT( m_hInst, m_hDlg, L"C:\\", false, &ptbKey, &kff, NULL );

			if ( res && ( ptbKey.bufferSize != 0 ) )
				res = AddPublicKeyToStorage( m_hDlg, ptbKey.buffer, ptbKey.bufferSize, kff );

			return res;
		}
		default:
			break;
	}

	return false;
}

bool cDlgBrowseKeyFile::OnLVNotify( NMHDR *pNMHdr )
{
	if ( pNMHdr->code == LVN_ITEMCHANGED )// LVN_ITEMACTIVATE )
	{
		UpdateButtons();
	}
	else
	if ( pNMHdr->code == NM_DBLCLK )
	{
		NMITEMACTIVATE *lpnmitem = (LPNMITEMACTIVATE)pNMHdr;
		OnOK( lpnmitem->iItem );
	}

	return true;
}

int cDlgBrowseKeyFile::OnGetMinMaxInfo( MINMAXINFO *pMMI )
{
	pMMI->ptMinTrackSize.x = 400;
	pMMI->ptMinTrackSize.y = 200;
	return 0;
}

// Fills filter combobox
void cDlgBrowseKeyFile::InitFilters()
{
	HWND hCmb = GetDlgItem( m_hDlg, IDC_cmbFileType );

	const wchar_t *ptr = m_filters;

	while( *ptr != L'\0' )
	{
		//char buff[ 100 ];
		//wcstombs( buff, ptr, sizeof( buff ) ); 
		ComboBox_AddStringT( hCmb, ptr );
		int len = wcslen( ptr );
		ptr += len + 1;
		len = wcslen( ptr );
		ptr += len + 1;
	}

	ComboBox_SetCurSel( hCmb, 0 );
}

// Returns pointer to current selected filter string f.e. "*.p12\0...." or "*.skr;*.gpg;*.pgp\0"
const wchar_t *cDlgBrowseKeyFile::GetCurrentFilter()
{
	HWND hCmb = GetDlgItem( m_hDlg, IDC_cmbFileType );
	int cur = ComboBox_GetCurSel( hCmb );
	if ( cur == CB_ERR )
	{
		cur = 0;
		ComboBox_SetCurSel( hCmb, cur );
	}

	const wchar_t *ptr = m_filters;

	for( int i = 0; *ptr != L'\0'; i++ )
	{
		ptr += wcslen( ptr ) + 1;

		if ( i == cur )
			return ptr;

		ptr += wcslen( ptr ) + 1;
	}

	assert( false );

	static cWCStr wcs;
	wcs.LoadStringT( IH.m_hInst, IDS_Undefined );
	return wcs.m_szBody;
}

int cDlgBrowseKeyFile::ToolBarNotify( NMHDR *pnmh )
{
	LPNMHDR lpnm  = (LPNMHDR)pnmh;
	LPNMTOOLBAR lpnmTB = (LPNMTOOLBAR)pnmh;
	LPNMTTDISPINFOA lpnmtdi = (LPNMTTDISPINFOA)pnmh;

	switch(lpnm->code)
	{
		//case TTN_GETDISPINFOW:
		//{
		//	LPNMTTDISPINFOW lpnmtdiW = (LPNMTTDISPINFOW)pnmh;

		//	// stores the info
		//	lpnmtdiW->uFlags = TTF_DI_SETITEM; //0;

		//	if ( pTranslateDll != NULL )
		//	{
		//		if ( pTranslateDll->m_flipDialog )
		//			lpnmtdiW->uFlags |= TTF_RTLREADING;

		//		lpnmtdiW->hinst = NULL;
		//		LoadStringT( m_hInst, GetIDSbyCmd( lpnmtdiW->hdr.idFrom ), lpnmtdiW->szText, 80 );
		//	}
		//	else
		//	{
		//		lpnmtdiW->hinst = m_hInst;
		//		lpnmtdiW->lpszText = MAKEINTRESOURCEW( GetIDSbyCmd( lpnmtdi->hdr.idFrom ) );
		//	}

		//	return TRUE;
		//}
		//case TTN_GETDISPINFOA:
		//{
		//	// stores the info
		//	lpnmtdi->uFlags = TTF_DI_SETITEM; // 0;

		//	if ( pTranslateDll != NULL )
		//	{
		//		if ( pTranslateDll->m_flipDialog )
		//			lpnmtdi->uFlags |= TTF_RTLREADING;

		//		lpnmtdi->hinst = NULL;
		//		LoadStringTA( m_hInst, GetIDSbyCmd( lpnmtdi->hdr.idFrom ), lpnmtdi->szText, 80 );
		//	}
		//	else
		//	{
		//		lpnmtdi->hinst = m_hInst;
		//		lpnmtdi->lpszText = MAKEINTRESOURCEA( GetIDSbyCmd( lpnmtdi->hdr.idFrom ) );
		//	}

		//	return TRUE;
		//}
		case TTN_GETDISPINFOW:
			return GetDispInfo( m_hInst, (LPNMTTDISPINFOW)pnmh, GetIDSbyCmd( pnmh->idFrom ) );

		case TTN_GETDISPINFOA:
			return GetDispInfo( m_hInst, (LPNMTTDISPINFOA)pnmh, GetIDSbyCmd( pnmh->idFrom ) );

		case TBN_DROPDOWN:
		{
			int menuID = -1;

			if ( lpnmTB->iItem == IDM_Find )
				menuID = 0;
			else
				return false;

			RECT rc;
			SendMessage( lpnmTB->hdr.hwndFrom, TB_GETRECT, (WPARAM)lpnmTB->iItem, (LPARAM)&rc );

			MapWindowPoints( lpnmTB->hdr.hwndFrom, HWND_DESKTOP, (LPPOINT)&rc, 2 );

			HMENU hPopupMenu = cMenuItem::LoadPopupMenu( m_hInst, IDR_MENU, menuID );

			cMenuItem::TrackPopupMenu(
				hPopupMenu,
				TPM_LEFTALIGN|TPM_LEFTBUTTON|TPM_VERTICAL,
				rc.left,
				rc.bottom,
				m_hDlg,
				NULL);

			return (FALSE);
		}
		default :
			return FALSE;
	}

	return FALSE;
}

LRESULT cDlgBrowseKeyFile::RealDlgProc( UINT uMsg, WPARAM wParam, LPARAM lParam )
{
	LRESULT result;
	
	if ( cMenuItem::DlgProc( &result, m_hDlg, uMsg, wParam, lParam ) )
		return result;

	switch(uMsg)
	{
		case WM_INITDIALOG:
			InitDialog();
			break;

		case WM_SIZE:
			m_sizedCtrls.OnSize( lParam );
			break;

		case WM_COMMAND:
			return OnCommand( wParam, lParam );

		case WM_NOTIFY:
		{
			NMHDR *hdr = (NMHDR *)lParam;

			if ( ( hdr->code == TTN_GETDISPINFOA ) || ( hdr->code == TTN_GETDISPINFOW ) )
				ToolBarNotify( hdr );
			else
			if ( wParam == IDC_List )
				return OnLVNotify( (LPNMHDR) lParam );
			else
			if ( ( hdr->hwndFrom == m_hToolbar ) ||
				( GetDlgCtrlID( m_hToolbar ) == wParam ) )
					return ToolBarNotify( hdr );

			return false;
		}

		case WM_KEYDOWN: // 0x100
		{
			if ( wParam == VK_DELETE )
			{
				if ( DeleteSelected() != 0 )
				UpdateButtons();
			}

			break;
		}

		case WM_GETMINMAXINFO:
			return OnGetMinMaxInfo( (MINMAXINFO *)lParam );

		case WM_DESTROY:
			dlgVals.Save( m_hDlg, m_hList );
			break;

		default:
			return false;
	}

	return true;
}

int cDlgBrowseKeyFile::UpdateButtons()
{
//	return 0;

	int selected = ListView_GetNextItem( m_hList, -1, LVNI_SELECTED );

	wchar_t wBuff[ MAX_PATH ] = L"";
	LVITEMW lvi;
	memset( &lvi, 0, sizeof( lvi ) );
	lvi.iItem = selected;
	lvi.mask = LVIF_IMAGE | LVIF_TEXT | LVIF_PARAM;
	lvi.pszText = wBuff;
	lvi.cchTextMax = maxLen( wBuff );
	
	ListView_GetItemT( m_hList, &lvi );

	DWORD attributes = (DWORD)lvi.lParam;
	bool bIsFile = ( attributes & FILE_ATTRIBUTE_DIRECTORY ) == 0;

	bool enable = 
		( ListView_GetSelectedCount( m_hList ) == 1 ) &&
		( bIsFile ); // ( lvi.iImage != cIconCache::FolderIcon  );

	ToolBar_EnableButton( m_hToolbar, IDM_Delete, enable );

	enable = (m_dwFlags & flag_DisableBrowseButtons) == 0;
	ToolBar_EnableButton( m_hToolbar, IDM_Up, enable  );
	ToolBar_EnableButton( m_hToolbar, IDM_Browse, enable  );

	enable = bIsFile; // lvi.iImage != cIconCache::FolderIcon;
	EnableWindow( GetDlgItem( m_hDlg, IDOK ), enable );

	if ( enable )
		SetDlgItemTextT( IDC_ebFileName, wBuff );
	else
		SetDlgItemTextT( IDC_ebFileName, L"" );

	return 0;
}

// Asks confirmation and delete all selected items.
int cDlgBrowseKeyFile::DeleteSelected(void)
{
	int q = ListView_GetSelectedCount( m_hList );

	if ( q == 0 )
	{
		assert( false );
		return 0;
	}

	int res = DlgMessageBoxT( IDS_DeleteSelected_Q, MB_ICONQUESTION | MB_YESNO | MB_DEFBUTTON2 );

	if ( res != IDYES )
		return 0;

	int count = 0;

	int item;
	while( ( item = ListView_GetNextItem( m_hList, -1, LVNI_SELECTED ) ) != -1 )
	{
		wchar_t fileName[ MAX_PATH ];
		ListView_GetItemTextT( m_hList, item, 0, fileName, maxLen( fileName ) );

		wchar_t filePath[ MAX_PATH ];
		wcscpy_s( filePath, m_path );
		wcscat_s( filePath, L"\\" );
		wcscat_s( filePath, fileName );

		if ( DeleteFileT( filePath ) )
		{
			ListView_DeleteItem( m_hList, item );
			count ++;
		}
		else
		{
			res = MessageBoxT( m_hDlg, filePath, _LoadIDS( IDS_ErrorInDeletingFile ), MB_ICONEXCLAMATION | MB_RETRYCANCEL );

			if ( res == IDCANCEL )
				break;
		}
	}

	return count;
}

bool cDlgBrowseKeyFile::OnOK( int iItem )
{
	if ( iItem == -1 )
	{
		if ( ListView_GetSelectedCount( m_hList ) != 1 )
			return false;
	}

	int selected = iItem;

	if ( selected == -1 )
	{
		selected = ListView_GetNextItem( m_hList, -1, LVNI_SELECTED );

		if ( selected == -1 )
			return false;
	}

	wchar_t wBuff[ MAX_PATH ];
	LVITEMW lvi;
	memset( &lvi, 0, sizeof( lvi ) );
	lvi.iItem = selected;
	lvi.mask = LVIF_IMAGE | LVIF_TEXT | LVIF_PARAM;
	lvi.pszText = wBuff;
	lvi.cchTextMax = maxLen( wBuff );
	
	ListView_GetItemT( m_hList, &lvi );

	DWORD attributes = (DWORD)lvi.lParam;
	bool bIsDirectory = ( attributes & FILE_ATTRIBUTE_DIRECTORY ) != 0; // lvi.iImage == cIconCache::FolderIcon )

	int len = wcslen( m_path );
	if ( ( len != 0 ) && ( m_path[ len - 1 ] != L'\\' ) )
		wcsncat( m_path, L"\\", MAX_PATH );

	if ( bIsDirectory )
	{
		wchar_t path[ MAX_PATH ];
		wcscpy_s( path, m_path );
		wcscat_s( path, lvi.pszText );

		const wchar_t *filter = GetCurrentFilter();
		SetCurrentFolder( path, filter );
		UpdateButtons();
	}
	else
	{
		wcsncat( m_path, lvi.pszText, MAX_PATH );
		EndDialog( m_hDlg, IDOK );
	}

	return true;
}