#include "cDlgCreateCertificate.h"
#include "..\..\CertificateLib\SMemory\SBlocks.h"
#include "..\..\CertificateLib\x509pkts\oidFuncs.h"
#include "..\..\CertificateLib\pgpPkts\Packets.h"
#include "..\..\CertificateLib\publicKeyAlgs\PublicKeyAlgs.h"
#include "..\..\CertificateLib\x509pkts\x509_structs.h"
#include "resource.h"
#include <commctrl.h>
#include <stdio.h>


BOOL cDlgCreateCertificate::RealDlgProc( UINT uMsg, WPARAM wParam, LPARAM lParam )
{
	switch(uMsg)
	{
		case WM_INITDIALOG:
			if ( !InitDialog() )
				EndDialog( m_hDlg, IDCANCEL );
			break;

		case WM_COMMAND:
			if ( HIWORD( wParam ) == 0 )
			{
				switch ( LOWORD( wParam ) )
				{
					case IDOK:
					case IDCANCEL:
						EndDialog( m_hDlg, LOWORD( wParam ) );
						return true;
				}
			}
			break;
		default:
			return false;
	}

	return true;
}



static UINT ColumnLabelIDS[] = 
{
	IDS_szField, IDS_szValue
};

const int NUM_COLUMNS = sizeof( ColumnLabelIDS ) / sizeof( UINT );

static int _gnColumnFmt[ NUM_COLUMNS ] =
{
	LVCFMT_LEFT, LVCFMT_LEFT
};

static int _gnColumnWidth[NUM_COLUMNS]=
{
	100, 156
};

static BOOL InitListViewColunms( HINSTANCE hInst, HWND hLView )
{
	int i;
	LV_COLUMN lvc;

	memset( &lvc, 0, sizeof( lvc ) );

	lvc.mask = LVCF_WIDTH;

	//
	//	Check if already initialised...
	//
	if ( ListView_GetColumn( hLView, 0, &lvc ) )
	{
//		MessageBox( 0, "Already inited", "InitListViewColumns", MB_OK );

		return TRUE;
	}

	lvc.mask = LVCF_FMT | LVCF_WIDTH | LVCF_TEXT | LVCF_SUBITEM;

	char text[ 80 ];

	int res;

	for(i=0; i<NUM_COLUMNS; i++)
	{
		LoadString( hInst, ColumnLabelIDS[i], text, sizeof( text ) );

		lvc.iSubItem=i;
		lvc.pszText = text; // _gszColumnLabel[i];
		lvc.cx = _gnColumnWidth[i];
		lvc.fmt=_gnColumnFmt[i];

		res = ListView_InsertColumn( hLView, i, &lvc );

		if ( res == -1 )
		{
			MessageBox( 0, text, "Error in adding List View Column", MB_OK );
			return FALSE;
		}
	}

	return TRUE;
}

static int bytes2string( char *str, int strSize, BYTE *bf, int bfSize )
{
	int len = 0;
	int pos = 0;

	for(; pos < bfSize; pos++ )
	{
		if ( len + 3 > strSize )
			break;

		sprintf( str + len, "%02X ", (DWORD)bf[pos] );
		len += 3;
	}

	return pos;
}

bool cDlgCreateCertificate::UpdateFields( bool editableOnly )
{
	char buff[ 2 * 1024 ];

	// lvId_version

	sprintf( buff, "V%d", m_version + 1 ); // 2 means 3
	ListView_SetItemText( m_hLView, lvId_version, 1, buff );

	// lvId_serialNumber,

	sprintf( buff, "%d", m_serialNumber );
	ListView_SetItemText( m_hLView, lvId_version, 1, buff );

	// lvId_signatureAlgorithm,

	BYTE pkID = ska_getAlgID( *m_pSka );

	const char *oidStr = getSignatureAlgOidStr( m_hashAlgID, pkID );
	if ( oidStr == NULL )
	{
		assert( false );

		char message[ 256 ];
		sprintf( message, "hash alg %d, public key alg %d", m_hashAlgID, pkID );
		MessageBox( m_hDlg, message, "Invalid signature algorithm", MB_ICONSTOP );
		return false;
	}

	tUnit *t = NULL;
	if ( ( generateOIDName( NULL, oidStr, &t ) == 0 ) || ( t== NULL ) ||
		( ( t->name == NULL ) && ( ( t->desc == NULL ) ) )
		)
	{
		assert( false );
		MessageBox( m_hDlg, oidStr, "Signature algorithm OID is invalid", MB_ICONSTOP );
		return false;
	}

	ListView_SetItemText( m_hLView, lvId_signatureAlgorithm, 1, t->name ? t->name : t->desc );

	// lvId_issuerRdn,

	int size = 0;
	int res = m_rdnIssier.GenerateName( NULL, &size );
	if ( res != 0 )
	{
		assert( false );
	}
	cPlainTextBlock ptb;
	if ( ptb.Allocate( size ) == NULL )
		return false;

	res =  m_rdnIssier.GenerateName( ( char *)ptb.buffer, &size );
	ListView_SetItemText( m_hLView, lvId_issuerRdn, 1, ( char * )ptb.buffer );

	// lvId_validFrom,

	SYSTEMTIME tm;
	memset( &tm, 0, sizeof( tm ) );
	tm.wYear = m_nbYe;
	tm.wMonth = m_nbMo;
	tm.wDay = m_nbDa;

	char dateStr[ 100 ];

	GetDateFormat(
		LOCALE_USER_DEFAULT, LOCALE_NOUSEROVERRIDE, //LOCALE_USE_CP_ACP
		&tm, NULL, dateStr, sizeof( dateStr ) );

	ListView_SetItemText( m_hLView, lvId_validFrom, 1, dateStr );

	// lvId_validTo,
	memset( &tm, 0, sizeof( tm ) );
	tm.wYear = m_naYe;
	tm.wMonth = m_naMo;
	tm.wDay = m_naDa;

	GetDateFormat(
		LOCALE_USER_DEFAULT, LOCALE_NOUSEROVERRIDE, //LOCALE_USE_CP_ACP
		&tm, NULL, dateStr, sizeof( dateStr ) );

	ListView_SetItemText( m_hLView, lvId_validTo, 1, dateStr );

	// lvId_subjectRdn,
	size = 0;
	res = m_rdnIssier.GenerateName( NULL, &size );
	if ( res != 0 )
	{
		assert( false );
	}

	if ( ptb.Allocate( size ) == NULL )
		return false;

	res =  m_rdnSubject.GenerateName( (char *)ptb.buffer, &size );
	ListView_SetItemText( m_hLView, lvId_subjectRdn, 1, ( char * )ptb.buffer );

	if ( editableOnly )
		return true;

	// lvId_publicKey,

	char text[ 256 ];
	sprintf( text, "%s( %d Bits )", PKT_GetPublicKeyAlgName( pkID ), ska_getModuleSize( *m_pSka ) * 8 );
	ListView_SetItemText( m_hLView, lvId_publicKey, 1, text );

	// lvId_subjectKeyID,
	size = 0;
	res = pka_generateLocalKeyID( *m_pPka, NULL, &size );
	cPlainTextBlock ptbSubjKeyID;
	if ( res != x509e_bufferTooSmall ) { assert( false ); }
	else
	{
		if ( ptbSubjKeyID.Allocate( size ) == NULL ) return false;
		res = pka_generateLocalKeyID( *m_pPka, ptbSubjKeyID.buffer, &size );
		if ( res != x509e_ok ) return false;
		bytes2string( text, sizeof( text ), ptbSubjKeyID.buffer, size );
		ListView_SetItemText( m_hLView, lvId_subjectKeyID, 1, text );
	}

	// lvId_authrityKeyID,
	size = 0;
	cPlainTextBlock ptbAuthKeyID;
	res = ska_generateLocalKeyID( *m_pSka, NULL, &size );
	if ( res != x509e_bufferTooSmall ) { assert( false ); }
	else
	{
		if ( ptbAuthKeyID.Allocate( size ) == NULL ) return false;
		res = pka_generateLocalKeyID( *m_pPka, ptbAuthKeyID.buffer, &size );
		if ( res != x509e_ok ) return false;
		bytes2string( text, sizeof( text ), ptbAuthKeyID.buffer, size );
		ListView_SetItemText( m_hLView, lvId_authrityKeyID, 1, text );
	}

	// lvId_basicConstraints,

	m_bCA = ( ptbAuthKeyID.bufferSize == ptbSubjKeyID.bufferSize ) &&
		( memcmp( ptbAuthKeyID.buffer, ptbSubjKeyID.buffer, ptbSubjKeyID.bufferSize ) == 0 );

	if ( m_bCA )
		sprintf( text, "Subject Type=CA, Path Length Constraint=None" );

	return true;
}

int ids_tab[] = 
{
	IDS_szVersion,
	IDS_szSerialNumber,
	IDS_szSignatureAlgorithm,
	IDS_szIssuerRdn,
	IDS_szValidFrom,
	IDS_szValidTo,
	IDS_szSubjectRdn,
	IDS_szPublicKey,
	IDS_szSubjectKeyID,
	IDS_szAuthrityKeyID,
	IDS_szBasicConstraints
};

BOOL cDlgCreateCertificate::InitDialog()
{
	m_version = 2; // means 3
	m_hashAlgID = pgphaSHA1;

	if ( !GetRDN( m_hInst, m_hDlg, &m_rdnSubject, "Get certificate subject relative destinguished name" ) )
		return false;

	m_hLView = GetDlgItem( m_hDlg, IDC_lvFields );

	InitListViewColunms( m_hInst, m_hLView );
	// Inserts list view items
	LVITEM lvi;

	memset( &lvi, 0, sizeof( LVITEM ) );

	lvi.mask = 0; // LVIF_IMAGE;

	for ( lvi.iItem = lvId_first; lvi.iItem <= lvId_last; (lvi.iItem)++ )
	{
		if ( ListView_InsertItem( m_hLView, &lvi ) != lvi.iItem )
		{
			assert( false );
			return false;
		}

		ListView_SetItemText( m_hLView, lvi.iItem, 0, LoadStringOnce( ids_tab[lvi.iItem] ) );
	}

	return UpdateFields();
}

//////////////////////////////////////////////
//
//	RDN dialogue
//

static struct 
{
	sNameAttribute::at_type atNum;
	int dlgID;
} idsTbl[] = 
{
	{ sNameAttribute::id_at_countryName,			IDC_edtCountry },
	{ sNameAttribute::id_at_stateOrProvinceName,	IDC_edtState },
	{ sNameAttribute::id_at_localityName,			IDC_edtLocality },
	{ sNameAttribute::id_at_commonName,				IDC_edtCommonName },
	{ sNameAttribute::pkcs9_emailAddress,			IDC_edtEmailAddress },
	{ sNameAttribute::id_at_organizationName,		IDC_edtOrganization },
	{ sNameAttribute::id_at_organizationalUnitName,	IDC_edtOrganizationUnit }
//	IDC_edtOthers
};

int findDlgID( int atNum )
{
	int i = sizeof ( idsTbl ) / sizeof( *idsTbl );

	while ( --i )
	{
		if ( idsTbl[i].atNum == atNum )
			return idsTbl[i].dlgID;
	}

	return -1;
}

int findAtNum( int dlgID )
{
	int i = sizeof ( idsTbl ) / sizeof( *idsTbl );

	while ( --i )
	{
		if ( idsTbl[i].dlgID == dlgID )
			return idsTbl[i].atNum;
	}

	return -1;
}

class cDlgRDN : public cDlg
{
public:
	enum { IDD = IDD_DlgRDN };

	cDlgRDN( sNameAttributeList *rdn, const char *caption = NULL )
	{
		m_caption = caption;
		m_rdn = rdn;
	}

private:
	BOOL RealDlgProc( UINT uMsg, WPARAM wParam, LPARAM lParam )
	{
		switch(uMsg)
		{
			case WM_INITDIALOG:
				InitDialog();
				break;

			case WM_COMMAND:
				if ( HIWORD( wParam ) == 0 )
				{
					switch ( LOWORD( wParam ) )
					{
						case IDOK:
							if ( updateRDN() != 0 )
							{
								MessageBox( m_hDlg, "Invalid name attributes", "RDN", MB_ICONSTOP );
								return false;
							}
						case IDCANCEL:
							EndDialog( m_hDlg, LOWORD( wParam ) );
							return true;
						default:
							break;
					}
				}
				break;
			default:
				return false;
		}

		return true;
	}

	sNameAttributeList *m_rdn;
	const char *m_caption;

	void InitDialog()
	{

		if ( m_caption )
			SetWindowText( m_hDlg, m_caption );

		char othersStr[ 1024 ] = "";

		for( sNameAttribute *na = m_rdn->head; na; na = na->next )
		{
			int dlgID = findDlgID( na->getIdByName( na->szOID ) );
			if ( dlgID != -1 )
				SetDlgItemText( m_hDlg, dlgID, na->szValue );
			else
			{
				char str[ 256 ];
				sprintf( str, "%s = %s", na->getName(), na->szValue );
				if ( strlen( othersStr ) != 0 )
					strcat( othersStr, "; " );
				strcat( othersStr, str );
			}
		}

		SetDlgItemText( m_hDlg, IDC_edtOthers, othersStr );
	}

	int updateRDN()
	{
		m_rdn->freeList();

		int i = sizeof ( idsTbl ) / sizeof( *idsTbl );

		while ( --i )
		{
			char str[ 1024 ];
			if ( GetDlgItemText( m_hDlg, idsTbl[i].dlgID, str, sizeof( str ) ) != 0 )
				m_rdn->Add( idsTbl[i].atNum, str );
		}

		char othersStr[ 1024 ];
		if ( GetDlgItemText( m_hDlg, IDC_edtOthers, othersStr, sizeof( othersStr ) ) != 0 )
			return m_rdn->ParseString( othersStr, ';' );

		return 0;
	}
};

bool GetRDN( HINSTANCE hInst, HWND hWnd, sNameAttributeList *rdn, const char *caption )
{
	cDlgRDN dlg( rdn, caption );
	return dlg.DoModal( hInst, dlg.IDD, hWnd ) == IDOK;
}
