#include "cDlgSKGen.h"
#include "..\BCArchive_V2\Frame\SeedGenDlg.h"
#include "..\CertificateLib\crtLibErrs.h"
#include "..\CertificateLib\publicKeyAlgs\publicKeyAlgs.h"
#include <stdio.h>
#include <windowsx.h>
#include "..\CertificateLib\x509pkts\pkcs12_funcs.h"
#include "..\CertificateLib\RNG\cRNG.h"
#include "..\CertificateLib\PGPPkts\pktdefs.h"
#include "..\CertificateLib\progressActor.h"
#include <process.h>
#include "cWizardDlgs.h"
#include "..\Includes\menuItem.h"
#include "..\Includes\UnicodeFuncs.h"

int GenerateSeed( HINSTANCE hInst, HWND hParent, BYTE *seedBuff, int seedSize )
{
	cGenerateSeedDlg dlg( seedBuff, seedSize );

	return dlg.DoModal( hInst, IDD_DlgGenerateSeed, hParent );
}

static bool SmallString( HWND hLB, const char *str )
{
	return getLastVisible( hLB, getWCStr( str ) ) >= (int)strlen( str );
 
//	return strlen( str ) < 80;

}

const int MSG_GenerationFinished = WM_USER + 1;

//
//	cProgressObj
//	The actor block is sent do prime genration process for feedback
//
class cProgressObj : public cProgressActor
{
	HWND m_hLB;
	cDlgPrimesGen *m_dlg;
	HINSTANCE m_hInst;
	char m_buff[ 1024 * 2 ];
	int m_textLen;
	int paramCount;

	int m_num;

public:
	//
	//	Progress functions return true if not canceled
	//
	virtual bool enableCancelButton( bool bEnable )
	{
		return !m_bCancelled;
	}

	virtual bool setMessageText( const char *text )
	{
		func( text, 0, 0 );
		return !m_bCancelled;
	}

	virtual bool setCaptionText( const char *text )
	{
		SetWindowTextA( m_dlg->m_hDlg, text );
		return !m_bCancelled;
	}

	cProgressObj( cDlgPrimesGen *dlg )
	{
		m_dlg = dlg;
		m_buff[0] = 0;
		m_textLen = 0;
		paramCount = 0;

		m_hLB = GetDlgItem( m_dlg->m_hDlg, IDC_List );
		m_dlg->InitTimer();
		m_num = 0;
	}
	~cProgressObj()
	{
//		SetWindowText( m_hText, "" );
	}

	virtual int func( const char *name, int res, int param  )
	{

		if ( m_dlg->m_cancelled )
		{
			return pkae_cancelled;
		}

		if ( name != NULL )
		{
			strcpy_s( m_buff, name );
			int len = strlen( m_buff );
			m_buff[ --len ] = '\0'; // cut new line.

			if ( m_num == 0 )
				sprintf( m_buff + len, " [ %d ]", param );
			else
			if ( m_num > 1 )
				m_num++;

			int res = ListBox_InsertStringT( m_hLB, m_num, _getWcs( m_buff ) );			

			m_buff[0] = '\0';
			m_num++;

			return 0;
		}
		else
		if ( m_num <= 1 )
		{
//			sprintf( m_buff + m_textLen, " [ %d ]", param );
		}
		else
		if ( param < 0 )
			strcat_s( m_buff, "." );
		else
			strcat_s( m_buff, "*" );

		//if ( SmallString( m_hLB, m_buff ) )
		//{
		//	ListBox_InsertString( m_hLB, m_num, m_buff );
		//	ListBox_DeleteString( m_hLB, m_num + 1 );
		//	ListBox_SetTopIndex( m_hLB, m_num );
		//}
		//else
		//{
		//	m_buff[0] = '\0';
		//	m_num++;
		//	ListBox_InsertString( m_hLB, m_num, m_buff );
		//	ListBox_SetTopIndex( m_hLB, m_num );
		//}

		if ( !SmallString( m_hLB, m_buff ) )
			m_buff[0] = '\0';

		ListBox_DeleteString( m_hLB, m_num );
		ListBox_InsertStringT( m_hLB, m_num, _getWcs( m_buff ) );

		return 0;
	}

};

//
//	It generates additional random buffer for any purpouses
//
void *cDlgPrimesGen::m_ptbAdditionalRnd = NULL;
//cPlainTextBlock *cDlgPrimesGen::m_ptbAdditionalRnd = NULL;
int cDlgPrimesGen::m_additionalRndSizeRequired;

DWORD WINAPI GenerateKeys( void *handle )
{
	cDlgPrimesGen *dlg = (cDlgPrimesGen *)handle;
	ShowWindow( dlg->m_hDlg, SW_HIDE );
	//bool useSafePrime = true;

//	GetDlgItemText( dlg->m_hDlg, IDC_edtLength, text, sizeof( text ) );
	int size = dlg->m_keyBitSize; //atoi( text );

	int res = ppe_ok;
	bool cancelled = false;
	cPlainTextBlock ptbSeed;
	cPlainTextBlock ptbRnd;

	int saltSize = 8;

	size = dlg->m_keyBitSize / 8;

	if ( dlg->m_hSka )
	{
		ska_free( dlg->m_hSka );
		dlg->m_hSka = NULL;
	}

	int additionalRndSize = 0;

	cPlainTextBlock *ptbPtr = ( cPlainTextBlock * )dlg->m_ptbAdditionalRnd;
	
	if ( ptbPtr != NULL )
	{
		if ( ptbPtr->Allocate( dlg->m_additionalRndSizeRequired ) != NULL )
			additionalRndSize = dlg->m_additionalRndSizeRequired;
	}

	if ( ( ptbSeed.Allocate( saltSize + 3 * ( size + additionalRndSize ) ) == NULL ) || 
		 ( ptbRnd.Allocate( size + additionalRndSize ) == NULL )
		)
	{
		dlg->DlgMessageBoxT( IDS_ErrorInAllocatingMemory, MB_ICONSTOP );
		goto exit;
	}

	res = GenerateSeed( dlg->m_hInst, dlg->m_hDlg, ptbSeed.buffer, ptbSeed.bufferSize );
	if ( res == IDCANCEL )
	{
		res = ppe_ok;
		cancelled = true;
		goto exit;
	}

	//
	//	It is useful to 'little' shake the generated seed before using.
	//
	res = P12_producePseudoRandomBytes( ptbRnd.buffer, ptbRnd.bufferSize, ptbSeed.buffer, saltSize, 1, ptbSeed.buffer + saltSize, ptbSeed.bufferSize - saltSize );
	if ( res != x509e_ok )	goto exit;

	if ( additionalRndSize != 0 )
	{
		if ( ptbPtr == NULL )
		{
			assert( false );
		}

		memcpy( ptbPtr->buffer, ptbRnd.buffer + size, additionalRndSize );
	}

	cRNG::setPool( ptbRnd.buffer, ptbRnd.bufferSize );

	ska_RSA_generateSafePrime( dlg->m_useSafePrime );

	{
		// Initialisation of cProgressObj

		cProgressObj progressObj( dlg );
		ShowWindow( dlg->m_hDlg, SW_SHOW );

		res = ska_generateKeyAndCreateAlgorithm( &dlg->m_hSka, pgppkaRSA, size, &progressObj );

		progressObj.func( " ", 0,0 ); // Clears text.
	}
	if ( res != pkae_ok )	goto exit;

exit:

	dlg->m_cancelled |= cancelled;

	if ( dlg->m_cancelled )
		res = pkae_cancelled;
	else
	if ( res != ppe_ok )
	{
		wchar_t text[ 128 ];
		FormatT( text, maxLen( text ), _LoadIDS( IDS_ErrorInGeneratingKey_D ), res );
		dlg->DlgMessageBoxT( text, MB_ICONSTOP );
	}

	if ( dlg->m_thread != NULL )
		CloseHandle( dlg->m_thread );

	dlg->m_thread = NULL;

	::SendMessage( dlg->m_hDlg, MSG_GenerationFinished, res, 0 );

	return 0;
}
void cDlgPrimesGen::InitTimer()
{
	m_timeStart = GetTickCount();
	m_timerID = SetTimer( m_hDlg, NULL, 1000, (TIMERPROC)SetTimeValue );
}

LRESULT cDlgPrimesGen::RealDlgProc( UINT uMsg, WPARAM wParam, LPARAM lParam )
{
	LRESULT result;
	
	if ( cMenuItem::DlgProc( &result, m_hDlg, uMsg, wParam, lParam ) )
		return result;

	switch(uMsg)
	{
		case MSG_GenerationFinished :
		{
			EndDialog( m_hDlg, m_cancelled ? IDCANCEL : IDOK );
			break;
		}

		case WM_SIZE:
			sizedCtrls.OnSize( lParam );
			break;

		case WM_DESTROY:
			if ( m_timerID != NULL )
				KillTimer( m_hDlg, m_timerID );
			break;

		case WM_INITDIALOG:
			m_timerID = 0;
			if ( !InitDialog() )
				EndDialog( m_hDlg, IDCANCEL );
			break;

		case WM_COMMAND:
			if ( HIWORD( wParam ) == 0 )
			{
				switch ( LOWORD( wParam ) )
				{
					case IDOK:
					{
/*
						if ( !CheckValues() )
							return true;

						m_waitDlg.DoModeless( m_hInst, m_waitDlg.IDD, m_hDlg );

						DWORD threadId;
						m_thread = CreateThread( NULL, NULL, GenerateKeys, this, 0, &threadId );
						EnableWindow( GetDlgItem( m_hDlg, IDOK ), FALSE );
						m_cancelled = FALSE;
*/
						return true;
					}
					case IDCANCEL:
						if ( m_thread )
						{
							if ( m_cancelled )
							{
								TerminateThread( m_thread, 0 );
								m_thread = NULL;
								m_cancelled = FALSE;
								return true;
							}

							if ( MessageBoxT( 
									m_hDlg,
									_LoadIDS( IDS_TerminateProcess_Q ),
									_LoadIDS( IDS_GenerateKey ),
									MB_ICONQUESTION | MB_YESNO )
								== IDNO )
								return true;

							m_cancelled = true;
//							EnableWindow( GetDlgItem( m_hDlg, IDOK ), TRUE );
						}
						else
							EndDialog( m_hDlg, LOWORD( wParam ) );

						return true;
				}
			}
			break;
//		case WM_NOTIFY:
//			return OnNotify( (NMHDR *)lParam );
		default:
			return false;
	}

	return true;
}
/*
BOOL cDlgPrimesGen::OnNotify( NMHDR *hdr )
{
	switch( hdr->code )
	{
		case UDN_DELTAPOS:
			if ( hdr->idFrom == IDC_spinSize )
				return UDNDeltapos( (NMUPDOWN *)hdr, IDC_edtLength, 512, 8 * 1024, 512 );
			else
			if ( hdr->idFrom == IDC_spinIterations )
				return UDNDeltapos( (NMUPDOWN *)hdr, IDC_edtIterations, 1, 0xFFFF, 1 );
		default : break;
	}

	return FALSE;
}
*/

cDlgPrimesGen::~cDlgPrimesGen()
{
	if ( m_hSka != NULL )
		ska_free( m_hSka );
}

void CALLBACK cDlgPrimesGen::SetTimeValue(
  HWND hDlg,         // handle to window
  UINT uMsg,         // WM_TIMER message
  UINT_PTR idEvent,  // timer identifier
  DWORD dwTime       // current system time
)
{
	cDlgPrimesGen *pThis = (cDlgPrimesGen *)GetWindowLong(hDlg, DWL_USER);
	assert( pThis != NULL );

	assert( GetDlgItem( hDlg, IDC_TimeBox ) != NULL );

	DWORD totalSec = ( dwTime - pThis->m_timeStart ) / 1000;

	char str[ 20 ];
	int hour = totalSec / 3600;
	int min  = ( totalSec % 3600 ) / 60;
	int sec  = totalSec % 60;

	sprintf( str, "%02d:%02d:%02d", hour, min, sec );

	SetDlgItemTextA( hDlg, IDC_TimeBox, str );
}

BOOL cDlgPrimesGen::InitDialog()
{
	SendMessage( m_hDlg, WM_SETICON, ICON_SMALL, (LPARAM)LoadIcon( m_hInst, MAKEINTRESOURCE( IDI_Main ) ) );

	//
	//	Hides the window before seed generation process
	//
	ShowWindow( m_hDlg, SW_HIDE );

	sizedCtrls.Init( m_hDlg );
	sizedCtrls.AddCtrl( IDC_TimeBox, cSizedCtrls::ptPR );
	sizedCtrls.AddCtrl( IDC_List,	cSizedCtrls::ptSDR );
	sizedCtrls.AddCtrl( IDCANCEL,	cSizedCtrls::ptPDR );

	CenterPosition();

	unsigned int threadId;
	
//	m_thread = CreateThread( NULL, NULL, GenerateKeys, this, 0, &threadId );
	m_thread = (HANDLE)_beginthreadex( NULL, NULL, (unsigned (__stdcall *) (void *))GenerateKeys, this, 0, &threadId );
//	EnableWindow( GetDlgItem( m_hDlg, IDOK ), FALSE );
	m_cancelled = FALSE;

	return TRUE;
	
/*
	HWND hWnd = GetDlgItem( m_hDlg, IDC_cmbAlgorithm );
	ComboBox_InsertString( hWnd, algRSA, "RSA" );
	ComboBox_InsertString( hWnd, algRSAWithSafePrime, "RSA ( generate safe prime )" );
	ComboBox_SetCurSel( hWnd, 0 );

	SetDlgItemText( m_hDlg, IDC_edtLength, "256" );
	SetDlgItemText( m_hDlg, IDC_edtIterations, "1" );

	hWnd = GetDlgItem( m_hDlg, IDC_cmbPBEAlg );
	ComboBox_InsertString( hWnd, 0, "SHA1 with 3Key TripleDES CBC" ); // { pkcs-12 pbe pbeWithSHAAnd3KeyTripleDES_CBC }
	ComboBox_SetCurSel( hWnd, 0 );

	BYTE salt[ 8 ] = { 0x5F, 0x03, 0x59, 0xF4, 0x42, 0x72, 0xCD, 0x14 };
	int res;

	//
	//	Generates random salt from current time and clock;
	//
	time_t tm = time(0) ^ clock();
	res = generateRandomKey( salt, sizeof( salt ), (BYTE *)&tm, sizeof( tm ) );

	WriteHexString( GetDlgItem( m_hDlg, IDC_edtSalt) , salt, sizeof( salt ) );

	CheckDlgButton( m_hDlg, IDC_chkGenerateCertificate, BST_CHECKED );
	return TRUE;
*/
}
