#include "cWizardBlock.h"
#include "cDlgSKGen.h"
#include "BCDB_Functions.h"

#include "..\CertificateLib\PGPPkts\pktdefs.h"
#include "..\CertificateLib\x509pkts\oidFuncs.h"
#include "..\CertificateLib\x509pkts\pkcs12_funcs.h"
#include "..\CertificateLib\RNG\cRNG.h"

#include "..\Includes\FileView.h"
#include "..\Includes\OpenFiles.h"
#include "..\Includes\UnicodeFuncs.h"

#include <stdio.h>
#include <io.h>

cWizardBlock::cWizardBlock( HINSTANCE phInst, HWND phParent )
{
	hInst = phInst;
	hParent = phParent;
	srand( (int)time( 0 ) );
	for( int i = sizeof( salt ); i--; )
		salt[ i ] = rand();

	iterCount = 2048;

	ska = NULL;
	skaIssuer = NULL;
	pka = NULL;
	// Version 1.04.2 - 11.09.2014 New public key by defalt has size 4098
	keyBitSize = 4096; // 2048;
	algPos = 0;
	generateNew = true;
	generateCertificate = true;
	memset( path, 0, sizeof( path ) );
	skShowNum = 1;
	pkShowNum = 1;
	memset( friendlyName, 0, sizeof( friendlyName ) );
	saveToFiles = false;
	saveToBCStorage = true;
	wasGenerated = false;

	hFixedFont = GetAFont();

	m_titleFont = NULL;
	m_boldFont  = NULL;

	version = 2; // means 3
	serialNumber = 1;
	serialNumberDWL = 0; // OpenSSL certificate has 8 bytes serial number.

	// hashAlgID = pgphaSHA1;
	hashAlgID = pgphaSHA512;
	pkAlgID = pgppkaRSA;

	m_view = false;
	bCA = 0;
	bcPathLength = -1;

	SYSTEMTIME tm;
	GetSystemTime( &tm );
	nb = tm;
	tm.wYear++;
	na = tm;
}

HFONT cWizardBlock::GetAFont()
{ 
	static LOGFONT lf;  // structure for font information  

	// Get a handle to the ANSI fixed-pitch font, and copy 
	// information about the font to a LOGFONT structure. 

	GetObject(GetStockObject( ANSI_FIXED_FONT ), sizeof(LOGFONT), &lf); 

	// Set the font attributes, as appropriate.  

	lf.lfWeight = FW_NORMAL; 

	lf.lfItalic = false; // (fnFont == ITALIC); 
	lf.lfItalic = false; // (fnFont == ULINE); 

	// Create the font, and then return its handle.  

	return CreateFont(lf.lfHeight, lf.lfWidth, 
		lf.lfEscapement, lf.lfOrientation, lf.lfWeight, 
		lf.lfItalic, lf.lfUnderline, lf.lfStrikeOut, lf.lfCharSet, 
		lf.lfOutPrecision, lf.lfClipPrecision, lf.lfQuality, 
		lf.lfPitchAndFamily, lf.lfFaceName); 
} 

const wchar_t *cWizardBlock::getAlgName( int pAlgPos )
{
	if ( pAlgPos == -1 )
		pAlgPos = algPos;

	if ( pAlgPos == algRSA )
		m_algName.LoadStringT( hInst, IDS_szRSA );
	else
	if ( pAlgPos == algRSAWithSafePrime )
		m_algName.LoadStringT( hInst, IDS_szRSASafePrime );
	else
	{
		assert( false );
		m_algName.LoadStringT( hInst, IDS_szUndefined );
	}

	return m_algName.m_szBody;
}
bool cWizardBlock::CheckPasswords( HWND hDlg, int bitMask, bool silent ) // 0 bit - pwd, 1 bit pwdConfirm; silent = false
{
	bool pwdIsEmpty = pwd.wcBuffer[0] == 0;
	bool pwdConfirmIsEmpty = pwdConfirm.wcBuffer[0] == 0;

	int errIDS = 0;

	if ( bitMask == 1 ) // Check pwd only
	{
		if ( pwdIsEmpty )
			errIDS = IDS_PWDIsEmpty;
		else
		if ( !pwd.wcCheckValidSize() )
			errIDS = IDS_PWDTooSmall;
		else
		if ( !pwdConfirmIsEmpty && ( pwd.wcCompare( &pwdConfirm ) != 0 ) )
			errIDS = IDS_IncompatiblePWD;
	}
	else
	if ( bitMask == 2 ) // Check pwd confirm only
	{
		if ( pwdConfirmIsEmpty )
			errIDS = IDS_PWDIsEmpty;
		else
		if ( !pwdConfirm.wcCheckValidSize() )
			errIDS = IDS_PWDTooSmall;
		else
		if ( !pwdIsEmpty && ( pwd.wcCompare( &pwdConfirm ) != 0 ) )
			errIDS = IDS_IncompatiblePWD;
	}
	else
	if ( bitMask == 3 ) // Check both pwds
	{
		if ( pwdIsEmpty || pwdConfirmIsEmpty )
			errIDS = IDS_PWDIsEmpty;
		else
		if ( pwd.wcCompare( &pwdConfirm ) != 0 )
			errIDS = IDS_IncompatiblePWD;
		else
		if ( !pwd.wcCheckValidSize() )
			errIDS = IDS_PWDTooSmall;
	}

	if ( errIDS == 0 )
		return true;

	if( !silent )
		MessageBoxT( hDlg, _LoadIDS( errIDS ), _LoadIDS( IDS_AppName ), MB_ICONSTOP );

	return false;
}

bool cWizardBlock::isSKReady()
{
	return CheckPasswords( 0, 3, true ) && ( friendlyName[0] != 0 );
}

bool cWizardBlock::isPKReady()
{
	bool selfSigned = ( skaIssuer == NULL );

//	if ( ( ska == NULL ) && ( skaIssuer == NULL ) )
//		return generateNew;

	if ( rdnSubject.isEmpty() )
		return false;

	if ( getSignatureAlgOidStr() == NULL )
		return false;

	return true;
}

const char *cWizardBlock::getSignatureAlgOidStr( BYTE *ppkID )
{
	if ( !m_view )
	{
		pkAlgID = ( skaIssuer != NULL ? 
						ska_getAlgID( skaIssuer ) :
						(( ska != NULL ) ? ska_getAlgID( ska ) : pgppkaRSA )
					);
	}

	if ( ppkID != NULL )
		*ppkID = pkAlgID;

	const char *oidStr = ::getSignatureAlgOidStr( hashAlgID, pkAlgID );

	return oidStr;
}

const int saltSize = 8;
const int rndSizeReq = 160 / 8; // DSA_SMALL_PRIME_BYTE_SIZE;

bool cWizardBlock::generateCrtPkt( cPTB *ptb )
{
	const char *szSigAlgOID = getSignatureAlgOidStr();
	assert( szSigAlgOID != NULL );

	HPKA signatureSka = skaIssuer != NULL ? skaIssuer : ska;

	int size = 0;

	x509_certificate crt;
	int res = crt.createPacket(
		NULL, &size,
		pka, 
		signatureSka, 
		rdnSubject.head,
		rdnIssuer.head,
		nb.wYear, nb.wMonth, nb.wDay,
		na.wYear, na.wMonth, na.wDay, 
		szSigAlgOID,
		serialNumber
		);


	if ( res != x509e_bufferTooSmall )
		return false;

	cPlainTextBlock ptbCrt;
	if ( ptbCrt.Allocate( size ) == NULL )
		return false;

	res = crt.createPacket(
		ptbCrt.buffer, &size,
		pka, 
		signatureSka, // skaIssuer != NULL ? skaIssuer : ska, 
		rdnSubject.head,
		rdnIssuer.head,
		nb.wYear, nb.wMonth, nb.wDay,
		na.wYear, na.wMonth, na.wDay, 
		szSigAlgOID,
		serialNumber
		);

	// DSA algorithm requires random buffer
	if ( res == pkae_randomBufferRequired )
	{
		//
		// If it was not generated together with prime,
		//	see cDlgPrimesGen::m_ptbAdditionalRnd initialisation 
		//	at the beginning of cWizardBlock::Generate function
		//
		if ( m_ptbRnd.bufferSize == 0 )
		{
			cPlainTextBlock ptbSeed;
			if ( ( ptbSeed.Allocate( saltSize + 3 * ( rndSizeReq ) ) == NULL ) || 
				( m_ptbRnd.Allocate( rndSizeReq ) == NULL )
				)
			{
				return false;
			}

			if ( GenerateSeed( hInst, hParent, ptbSeed.buffer, ptbSeed.bufferSize ) == IDCANCEL )
			{
				return false;
			}

			res = P12_producePseudoRandomBytes( m_ptbRnd.buffer, m_ptbRnd.bufferSize, ptbSeed.buffer, saltSize, 1, ptbSeed.buffer + saltSize, ptbSeed.bufferSize - saltSize );
			if ( res != x509e_ok )
				return false;
		}

		cRNG::setPool( m_ptbRnd.buffer, m_ptbRnd.bufferSize );

		res = crt.createPacket(
			ptbCrt.buffer, &size,
			pka, 
			signatureSka, // skaIssuer != NULL ? skaIssuer : ska, 
			rdnSubject.head,
			rdnIssuer.head,
			nb.wYear, nb.wMonth, nb.wDay,
			na.wYear, na.wMonth, na.wDay, 
			szSigAlgOID,
			serialNumber
			);

		cRNG::clearPool();
	}

	if ( res != x509e_ok )
		return false;

	//
	//	Set signature secret key serial number 
	//
	cPlainTextBlock ptbID;
	if ( BCDB_GenerateLocalKeyID( &ptbID, signatureSka ) )
	{
		int serNum = 0;
		BCDB_GetCrtSerialNumber( &serNum, ptbID.buffer, ptbID.bufferSize, true );
	}

	return ptb->WriteBytes( ptbCrt.buffer, ptbCrt.bufferSize ) != 0;
}

bool cWizardBlock::Generate( cWCStr *wcsLog )
{
	//cPlainTextBlock *ptbLog

	int recordCount = 1;

	//
	// Random generation required for DSA signature
	//	We may generate it together with random prime
	//
	if ( pkAlgID == pgppkaDSA )
	{
		cDlgPrimesGen::m_ptbAdditionalRnd = &m_ptbRnd;
		cDlgPrimesGen::m_additionalRndSizeRequired = rndSizeReq;
	}

	//
	//	Generates primes and create algorithm.
	//
	if ( ( ska == NULL ) && ( generateSecKey ) )
	{
		assert( generateSecKey );

		bool useSafePrime = ( algPos == algRSAWithSafePrime );
		cDlgPrimesGen dlg( keyBitSize, useSafePrime );
		int res = dlg.DoModal( hInst, dlg.IDD, hParent );

		if ( res != IDOK )
		{
			if ( dlg.m_cancelled || ( res == IDCANCEL ) )
				wcsLog->Add( _LoadIDS( IDS_CancelledByUser ) );

			return false;
		}

		ska = dlg.m_hSka;
		dlg.m_hSka = NULL;

		wchar_t message[ 1024 ];
		FormatT( message, maxLen( message ), _LoadIDS( IDS_RSAPrimesWereGenerates_DD ), recordCount++, keyBitSize );
		wcsLog->Add(  message );

		cPlainTextBlock ptbID;
		if ( BCDB_GenerateLocalKeyID( &ptbID, ska, true ) &&
			 BCDB_bytes2string( message, maxLen( message ), ptbID.buffer, ptbID.bufferSize ) 
		   )
		{
			wcsLog->Add( _LoadIDS( IDS_szLocalKeyIDn ) );
			wcsLog->Add( message );
			wcsLog->Add( L"\n" );
		}
	}

	if ( pka == NULL )
	{
		if ( ska_derivePublicKeyAlgorithm( &pka, ska ) != x509e_ok )
		{
			MessageBoxT( hParent, _LoadIDS( IDS_ErrorInDerivingPK ), _LoadIDS( IDS_AppName ), MB_ICONSTOP );
			return false;
		}
	}

	cPlainTextBlock ptbCrt;
	cWCStr crtName;

	if ( generateCertificate )
	{
		HPKA signatureSka = skaIssuer != NULL ? skaIssuer : ska;
		if ( signatureSka == NULL )
		{
			MessageBoxT( hParent, _LoadIDS( IDS_ErrorIssuerKeyIsNotLoaded ), _LoadIDS( IDS_AppName ), MB_ICONSTOP );
			return false;
		}

		if ( !generateCrtPkt( &ptbCrt ) )
		{
			MessageBoxT( hParent, _LoadIDS( IDS_ErrorInGeneratingCrtPkt ), _LoadIDS( IDS_AppName ), MB_ICONSTOP );
			return false;
		}

		wchar_t message[ 1024 ];
		FormatT( message, maxLen( message ), _LoadIDS( IDS_STRING116 ), recordCount++, version + 1 );
		wcsLog->Add( message );

		const char *szCNValue = rdnSubject.getAttributeValue( sNameAttribute::getOid( sNameAttribute::id_at_commonName ) );

		if ( saveToBCStorage )
		{
			wchar_t path[ MAX_PATH ];
			if ( !BCDB_GetDataBasePath( path, maxLen( path ), false ) )
			{
				MessageBoxT( hParent, _LoadIDS( IDS_ErrorInGettingStorageFolder ), _LoadIDS( IDS_AppName ), MB_ICONSTOP );
				return false;
			}

			int len = wcslen( path );

			if ( szCNValue == NULL )
			{
				cPlainTextBlock ptbID;
				if ( !BCDB_GenerateLocalKeyID( &ptbID, pka, false ) ||
					 !BCDB_bytes2string( path + len, maxLen( path ) - len, ptbID.buffer, ptbID.bufferSize )
					 )
				{
					MessageBoxT( hParent, _LoadIDS( IDS_ErrorInGeneratingPKCrtPkt ), _LoadIDS( IDS_AppName ), MB_ICONSTOP );
					return false;
				}
			}
			else
			{
				wcscat_s( path, getWCStr( szCNValue ) );
	
				const char *szCNValueIssuer = rdnIssuer.getAttributeValue( sNameAttribute::getOid( sNameAttribute::id_at_commonName ) );

				if ( ( szCNValueIssuer != NULL ) && ( strcmp( szCNValueIssuer, szCNValue ) != 0 ) )
				{
					wcscat_s( path, L" - " );
					wcscat_s( path, getWCStr( szCNValueIssuer ) );
				}
			}

			crtName.Set( path );
			wcscat_s( path, L".cer" );

			bool writeIt = true;

			if ( FileExistsT( path ) ) // Exists
			{
				cFileView cfv( path );
				x509_certificate crt;
				cPlainTextBlock ptbThumbprint, ptbUserID;
				cWCStr wcsUserID;

				//
				//	Version 1.05.1 - 19.11.2015
				//	Generates name of secret key as thumbprint instead of local key ID
				//
				if ( crt.parse( cfv.m_ptr, (int)cfv.m_size.QuadPart ) == x509e_ok )
					BCDB_GetCrtInfo( &crt, &wcsUserID, &ptbUserID, &ptbThumbprint );

				//ptbUserID.PutByte( 0 );
				wcsUserID.Add( L"" );

				wchar_t lkIdStr[ MAX_PATH ];
				BCDB_bytes2string( lkIdStr, maxLen( lkIdStr ), ptbUserID.buffer, ptbUserID.bufferSize );
				
				cWCStr msg;
				msg.FormatT( _LoadIDS( IDS_CrtFileAlreadyExists_SSQ ), wcsUserID.m_szBody, path, lkIdStr );

				wchar_t thumbprintStr[ MAX_PATH ];
				BCDB_bytes2string( thumbprintStr, maxLen( thumbprintStr ), ptbUserID.buffer, ptbUserID.bufferSize );

				msg.AddF( L"\nThumbprint: %s", thumbprintStr );

				int idRes =
					MessageBoxT(
							hParent,
							msg.m_szBody,
							_LoadIDS( IDS_AppName ),
							MB_ICONQUESTION | MB_YESNO );

				if ( idRes == IDYES )
					DeleteFileT( path );
				else
					writeIt = false;
			}

			if ( writeIt )
			{
				wchar_t message[ 1024 ];
				FormatT( message, maxLen( message ), L"%d. ", recordCount++ );

				if ( WriteToFile( path, ptbCrt.buffer, ptbCrt.bufferSize ) == 0 )
				{
					wcscat_s( message, _FormatT( _LoadIDS( IDS_ErrorInWrittingCrtFile_S ), path ) );
					wcsLog->Add( message );

					MessageBoxT( hParent, message, _LoadIDS( IDS_AppName ), MB_ICONSTOP );

					return false;
				}

				wcscat_s( message, _FormatT( _LoadIDS( IDS_CrtFileWrittingSuccess_S ), path ) );
				wcsLog->Add( message );
			}
		}

		if ( saveToFiles )
		{
			wchar_t path[ MAX_PATH ];
			path[0] = '\0';

			wchar_t szFilters[ 1024 ];
			LoadStringT( hInst, IDS_szFiltersPublicP12, szFilters, maxLen( szFilters ) );

			bool res = BrowseFiles( hParent, path, maxLen( path ), szFilters, _LoadIDS( IDS_SavePKCrtFile ), false ) == TRUE;
			if ( !res )
			{
				MessageBoxT( hParent, _LoadIDS( IDS_CrtWasNotSaved ), _LoadIDS( IDS_AppName ), MB_ICONSTOP );
			}
			else
			if ( WriteToFile( path, ptbCrt.buffer, ptbCrt.bufferSize ) == 0 )
			{
				wchar_t message[ 1024 ];
				FormatT( message, maxLen( message ), L"%d.", recordCount++ );
				wcscat_s( message, _FormatT( _LoadIDS( IDS_ErrorInWrittingCrtFile_S ), path ) );
				wcsLog->Add( message );
				MessageBoxT( hParent, message, _LoadIDS( IDS_AppName ), MB_ICONSTOP );

				return false;
			}
			else
			{
				wchar_t message[ 1024 ];
				FormatT( message, maxLen( message ), L"%d.", recordCount++ );
				wcscat_s( message, _FormatT( _LoadIDS( IDS_CrtFileWrittingSuccess_S ), path ) );
				wcsLog->Add( message );
			}
		}
	}

	//
	//	Creates secret key packet.
	//
	if ( generateSecKey )
	{
		cPlainTextBlock pfxPduPtb;

		cUniversalString usFriendlyName;
		usFriendlyName.wstr2unicodeBuff( friendlyName );

		int res = P12_MakeSecretKeyPkt(
			&pfxPduPtb, 
			ska,
			(wchar_t *)usFriendlyName.unicodeBuffer, usFriendlyName.unicodeBufferSize - sizeof( wchar_t ), // whitout one zero wchar_t
			&pwd,
			salt, iterCount,
			ptbCrt.buffer, ptbCrt.bufferSize );

		if ( res != x509e_ok )
		{
			MessageBoxT(
					hParent,
					_LoadIDS( IDS_ErrorInGeneratingSKPacket ),
					_LoadIDS( IDS_AppName ),
					MB_ICONSTOP
				);

			return false;
		}

		wchar_t message[ 1024 ];
		FormatT(
				message, maxLen( message ), 
				_LoadIDS( IDS_STRING134 ),
				(wchar_t *)recordCount++
			);

		wcsLog->Add( message );

		if ( saveToBCStorage )
		{
			BYTE digest[20];
			if ( x509e_ok != x509_certificate::generateThumbprint( ptbCrt.buffer, ptbCrt.bufferSize, digest ) )
			{	assert( false ); return false; 	}

			cWCStr path( MAX_PATH );

			if ( crtName.isEmpty() )
			{
				if ( !BCDB_SecKeyFileName( path.m_szBody, path.m_allocatedLen, digest, sizeof( digest ), true ) )
				{	assert( false ); return false; 	}
			}
			else
				path.FormatT( L"%s.p12", crtName.m_szBody );

			if ( WriteToFile( path.m_szBody, pfxPduPtb.buffer, pfxPduPtb.bufferSize ) == 0 )
			{
				FormatT(
						message, maxLen( message ),
						_LoadIDS( IDS_STRING167 ),
						recordCount++,
						path.m_szBody
					);

				wcsLog->Add( message ); // WriteBytes( (BYTE*)message, strlen( message ) );

				MessageBoxT( hParent, message, _LoadIDS( IDS_AppName ), MB_ICONSTOP );

				return false;
			}

			FormatT( message, maxLen( message ), _LoadIDS( IDS_STRING168 ), recordCount++, path.m_szBody );
			wcsLog->Add( message );
		}

		if ( saveToFiles )
		{
			wchar_t path[ MAX_PATH ];
			path[0] = '\0';

			wchar_t szFilters[ 1024 ];
			LoadStringT( hInst, IDS_szFiltersSecretP12, szFilters, maxLen( szFilters ) );

			bool res = BrowseFiles( hParent, path, maxLen( path ), szFilters, _LoadIDS( IDS_szSaveSecretKeyFile ), false ) == TRUE;
			if ( !res )
			{
				MessageBoxT( hParent, _LoadIDS( IDS_szSecretKeyWasSaved ), _LoadIDS( IDS_AppName ), MB_ICONSTOP );
			}
			else
			if ( WriteToFile( path, pfxPduPtb.buffer, pfxPduPtb.bufferSize ) == 0 )
			{
				wchar_t message[ MAX_PATH + 256 ];
				FormatT( message, maxLen( message ),
					_LoadIDS( IDS_ErrorInWrittingSecretKey_S ),
					recordCount++,
					path );

				wcsLog->Add( message );

				MessageBoxT( hParent, message, _LoadIDS( IDS_AppName ), MB_ICONSTOP );

				return false;
			}
			else
			{
				wchar_t message[ 1024 ];
				FormatT( message, maxLen( message ),
					_LoadIDS( IDS_SuccessInSKSaving_S ),
					recordCount++, path );

				wcsLog->Add( message );
			}
		}
	}

	wasGenerated = true;

	return true;
}

///////////////////////////////////////////////
//
// List view NM_CUSTOMDRAW message handler  
//

LRESULT cListViewCustomDrawHandler::OnCustomDraw( LPNMHDR pnmh ) //, BOOL& bHandled )  
{  
//	T* pT = static_cast<T*>(this);  
//	pT->SetMsgHandled(TRUE);  

	LPNMCUSTOMDRAW lpNMCustomDraw = (LPNMCUSTOMDRAW)pnmh;  
	DWORD dwRet = 0;  
	switch(lpNMCustomDraw->dwDrawStage) {  
	case CDDS_PREPAINT:  
		dwRet = OnPrePaint( lpNMCustomDraw );  
		break;  
	case CDDS_ITEMPREPAINT:  
		dwRet = OnItemPrePaint( lpNMCustomDraw );  
		break;  
//	#if (_WIN32_IE >= 0x0400)  
	case (CDDS_ITEMPREPAINT | CDDS_SUBITEM):  
		dwRet = OnSubItemPrePaint( lpNMCustomDraw );  
		break;  
//	#endif //(_WIN32_IE >= 0x0400)  
/*
	case CDDS_POSTPAINT:  
		dwRet = OnPostPaint( lpNMCustomDraw);  
		break;  
	case CDDS_PREERASE:  
		dwRet = OnPreErase( lpNMCustomDraw);  
		break;  
	case CDDS_POSTERASE:  
		dwRet = OnPostErase( lpNMCustomDraw);  
		break;  
	case CDDS_ITEMPOSTPAINT:  
		dwRet = OnItemPostPaint( lpNMCustomDraw);  
		break;  
	case CDDS_ITEMPREERASE:  
		dwRet = OnItemPreErase( lpNMCustomDraw);  
		break;  
	case CDDS_ITEMPOSTERASE:                 break;  
		dwRet = OnItemPreErase( lpNMCustomDraw);  
		break;  
	case CDDS_ITEMPOSTERASE:  
		dwRet = OnItemPostErase( lpNMCustomDraw);  
		break;  
*/
	default:  
//		pT->SetMsgHandled(FALSE);  
		break;  
	}  
//	bHandled = pT->IsMsgHandled();  
	return dwRet;  
}

DWORD cListViewCustomDrawHandler::OnPrePaint( LPNMCUSTOMDRAW lpNMCustomDraw )  
{  
    return CDRF_NOTIFYITEMDRAW;  
}  
  
DWORD cListViewCustomDrawHandler::OnItemPrePaint( LPNMCUSTOMDRAW lpNMCustomDraw )  
{  
    return CDRF_NOTIFYSUBITEMDRAW;  
}  
  
DWORD cListViewCustomDrawHandler::OnSubItemPrePaint( LPNMCUSTOMDRAW lpNMCustomDraw)  
{  
	LPNMLVCUSTOMDRAW pLVCD = (LPNMLVCUSTOMDRAW)lpNMCustomDraw;  
	if (NULL == pLVCD)  
		return CDRF_DODEFAULT;  

	if ( ( 0 == pLVCD->iSubItem ) || (lpNMCustomDraw->lItemlParam == 0 ) )
	{  
		pLVCD->clrText = ::GetSysColor(COLOR_WINDOWTEXT);  
		pLVCD->clrTextBk = ::GetSysColor(COLOR_WINDOW);  
		return CDRF_NEWFONT;  
	}  

	if ( lpNMCustomDraw->lItemlParam != 0 )
	{
		pLVCD->clrText = RGB(255,0,0);  
		pLVCD->clrTextBk = ::GetSysColor(COLOR_WINDOW);  
	}

	return CDRF_NEWFONT;  
}

int cWizardBlock::LoadSecretKey( wchar_t *fileName )
{
	m_view = true;

	cFileView fv( fileName );
	if ( fv.m_err )
		return x509e_internalError;

	cUniversalString usFriendlyName;

	bool res = BCDB_GetP12SecretKeyInfo( hInst, hParent, fv.m_ptr, (int)fv.m_size.QuadPart, NULL, &usFriendlyName, &ptbLKID );

	if ( !res )
		return x509e_internalError;

	wcsncpy( friendlyName, usFriendlyName.wcBuffer, maxLen( friendlyName ) );
	friendlyName[ maxLen( friendlyName ) - 1 ] = L'\0';

	// Find corresponded public key

	wchar_t filePath[ MAX_PATH ];
	res = BCDB_GetX509KeyByID( NULL, ptbLKID.buffer, ptbLKID.bufferSize, false, NULL, filePath );

	if ( res )
	{
		LoadPublicKey( filePath );
	}

	return x509e_ok;
}

int cAuthorityInfo::readExtension( x509_extensions *extensions )
{
	BYTE *keyIDPtr = NULL;
	int keyIDSize = 0;
	BYTE *generalName = NULL;
	int generalNameSize = 0;

	int res = extensions->getAuthorityKeyID(
			&keyIDPtr, &keyIDSize,
			&choice,
			&generalName,
			&generalNameSize,
			&serialNumber
		);

	if ( res == x509e_ok )
	{
		bInited = true;

		if ( keyIDSize != 0 )
			ptbKeyID.WriteBytes( keyIDPtr, keyIDSize );

		if ( choice == 4 )
			res = x509_generalName::readDirectoryName( &name, generalName, generalNameSize );
	}

	return res;
}

void cAuthorityInfo::writeToPtb( cWCStr *wcs, bool multyStrings )
{
	wcs->Add( _LoadIDS( IDS_KeyID ) );
	bytes2string( wcs, ptbKeyID.buffer, ptbKeyID.bufferSize );
	wcs->Add( multyStrings ? L"\n" : L", " );

	if ( name.head != NULL )
	{
		wcs->Add( _LoadIDS( IDS_Names ) );
		wcs->Add( multyStrings ? L"\n" : L", " );

		wcs->Add( cWCStr().FormatT( L"%s:", cWCStr( x509_generalName::getDescription( choice ) ).m_szBody ) );
		wcs->Add( multyStrings ? L"\n" : L", " );

		for( sNameAttribute *att = name.head; att; att = att->next )
		{
			wcs->Add( multyStrings ? L"  " : L"" );
			wcs->Add( getWCStr( att->getName() ) );
			wcs->Add( L"=" );
			wcs->Add( getWCStr( att->szValue ) );
			wcs->Add( multyStrings ? L"\n" : L", " );
		}
	}

	wcs->Add( cWCStr().FormatT( _LoadIDS( IDS_SerialNumber_D ), serialNumber ) );
}

int cWizardBlock::VerifyCertificate()
{
	if ( crt.L == 0 )
		return x509e_emptyParameters;

	// Find corresponded issuer public key
	
	BYTE *issuerKeyID;
	int issuerKeyIDSize;

	int res = crt.tbsCertificate.getIssuerKeyID( &issuerKeyID, &issuerKeyIDSize );

	if ( res != x509e_ok ) return res;
	
	cPlainTextBlock crtPtb;
	if ( !BCDB_GetX509KeyByID( &crtPtb, issuerKeyID, issuerKeyIDSize, false ) )
		return ppe_keyNotFound;

	x509_certificate issuerCrt;

	res = issuerCrt.parse( crtPtb.buffer, crtPtb.bufferSize );
	if ( res != x509e_ok ) return res;

	res = crt.verify( &(issuerCrt.tbsCertificate.subjectPKInfo) );
	return res;
}

int cWizardBlock::LoadPublicKey( wchar_t *fileName )
{
	m_view = true;

	pkShowNum = 0;

	// v. 1.03.1 saves the file content to m_crtPtb and closes file. Prev version does not allow to update the opened file.
	cFileView crtFile( fileName );
	if ( crtFile.m_err )
//	if ( !fv.Open( fileName ) )
		return x509e_internalError;

	m_crtPtb.WriteBytes( crtFile.m_ptr, (int)crtFile.m_size.QuadPart );
//	x509_certificate crt; 

//	int res = crt.parse( fv.m_ptr, fv.m_size );
	int res = crt.parse( m_crtPtb.buffer, m_crtPtb.bufferSize );
	if ( res != x509e_ok ) return res;

	res = pka_createAlgorithm_X509( &pka, &crt.tbsCertificate.subjectPKInfo );
	if ( res != x509e_ok ) return res;

	version = crt.tbsCertificate.version - 1; // 2 means 3
	//res = crt.tbsCertificate.serialNumber.getUIValue( (unsigned int *)&serialNumber );

	// Version 1.05.2.6 - 30.01.2018
	//	Uses 64 bit serial number and safe function _getUIValue.
	serialNumber = -1;
	res = _getUIValue( crt.tbsCertificate.serialNumber.L, crt.tbsCertificate.serialNumber.B, &serialNumberDWL );
	if ( res != x509e_ok ) return res;

	serialNumber = (DWORD)serialNumberDWL; // cWizardDlg_PubKeyCrt::UpdateSerialNumber uses serialNumberDWL instead if it is not zero.
	if ( serialNumber == serialNumberDWL )
		serialNumberDWL = 0;

	hashAlgID = crt.signatureAlg.tu->hAlg;
	pkAlgID = crt.signatureAlg.tu->pAlg;

	//
	//	Generates authority info
	//

	res = m_authorityInfo.readExtension( &crt.tbsCertificate.extensions );

	rdnSubject.AddCopyList( crt.tbsCertificate.subjectName.head ),
	rdnIssuer.AddCopyList( crt.tbsCertificate.issuerName.head );

	// Validity
	na = crt.tbsCertificate.validity.notAfter.ConvertToSystemTime();	// Not after
	nb = crt.tbsCertificate.validity.notBefore.ConvertToSystemTime();	// Not before

	BYTE bt;
	res = crt.tbsCertificate.extensions.checkBasicConstraint( &bt, &bcPathLength ); // BasicConstraints CertificateAuthority
	if ( res != x509e_ok ) return res;

	bCA = ( bt != 0 );

	return x509e_ok;
}
