#ifndef __cWizardBlock__
#define __cWizardBlock__

#include "resource.h"

#include "..\Includes\cDlg.h"
#include "..\Includes\FileView.h"

#include "..\CertificateLib\x509pkts\x509_structs.h"
#include "..\CertificateLib\publicKeyAlgs\publicKeyAlgs.h"
#include "..\CertificateLib\SMemory\cPassword.h"

#include <time.h>

enum { algRSA, algRSAWithSafePrime };

class cAuthorityInfo
{
public:
	cPlainTextBlock ptbKeyID;
	int choice;
	x509_name name;
	int serialNumber;

	bool bInited;

	cAuthorityInfo()
	{
		serialNumber = 1;
		bInited = false;
	}


	int readExtension( x509_extensions *extensions );
	void writeToPtb( cWCStr *wcs, bool multyStrings = true );
};

class cWizardBlock
{
	//
	//	Additional random buff.
	//	see cDlgPrimesGen::m_ptbAdditionalRnd initialisation 
	//	at the beginning of cWizardBlock::Generate function
	//
	cPlainTextBlock m_ptbRnd;

public:

	// Public key certificate parsed info, see LoadPublicKey and VerifyCertificate
	//cFileView fv;

	cPlainTextBlock m_crtPtb;
	x509_certificate crt; 
	int VerifyCertificate();

	cDlgRect dlgRect;

	// Save created packages options.
	bool saveToFiles;		// Ask file names and location when generation finishes.
	bool saveToBCStorage;	// Save the packages to storage with corresponded local key IDs

	// Values for Generate secret key dialogue
	int keyBitSize;
	int algPos; // Position of algorithm in list 
	bool generateNew; // Defined by Generate new or Use an existing key radio buttons
	wchar_t path[ MAX_PATH ]; // Path of existent key file

	// Values for Secret key packet dialogue
	bool generateSecKey;
	int skShowNum;

	const wchar_t *getAlgName( int pAlgPos = -1 );

	wchar_t friendlyName[ 1024 ];
	cPlainTextBlock ptbLKID;

	cPassword pwd, pwdConfirm ;
	BYTE salt[ 8 ];
	int iterCount;

	//
	// List view fields
	// Has to be filled 
	// 1. Algorithm - may be changed
	// 2. Key size - may be changed
	// 3. friendly name
	// 4. password
	// 5. password ( confirmation )
	// if showAll 
	//	6. PBE algorithm f.e. { pkcs-12-pbe-pbeWithSHAAnd3-KeyTripleDES-CBC }
	//	7. Salt
	//	8. Iterations
	//
	// If key is already generated so Info view shows
	// 1. Local key ID
	// 2. ASN1 buffer with key info
	//
	

	// Values for Public key certificate packet dialogue
	bool generateCertificate;

	int version;		// 2 ( means v3 )
	int serialNumber;	// current value get from storage for assigned issuer
	DWORDLONG serialNumberDWL;

	// Signature hash and pk algorithms values
	BYTE hashAlgID;
	BYTE pkAlgID;

	sNameAttributeList	
		rdnSubject,
		rdnIssuer;

	//
	//	Authority Info is filled in LoadPublicKey function 
	//		by the info is get with crt.tbsCertificate.extensions.getAuthorityKeyID
	//
	cAuthorityInfo m_authorityInfo;

	//
	//	HANDLE to certificate issuer, if it is NULL, so self signed certificate has to be generated only.
	//	rdnIssuer is updated if skaIssuer is changed or assigned ( see cWizardDlg_PubKeyCrt::OnBrowseIssuer() ).
	//
	HPKA skaIssuer;
	cPTB issuerLocalKeyID; // SHA1 digest of issuer tbcertificate.

	// Validity
	SYSTEMTIME na;	// Not after
	SYSTEMTIME nb;	// Not before
//	int nbYe, nbMo, nbDa; // Not before Year, Month, Day
//	int naYe, naMo, naDa; // Not after Year, Month, Day

	bool bCA; // BasicConstraints CertificateAuthority
	unsigned int bcPathLength; // BasicConstraints CertificateAuthority path length
	int pkShowNum;

	HPKA ska;
	HPKA pka;

	HINSTANCE hInst;
	HWND hParent;
	HFONT hFixedFont;
	HFONT m_titleFont;
	HFONT m_boldFont;

	bool wasGenerated;

	cWizardBlock( HINSTANCE phInst, HWND phParent );

	~cWizardBlock()
	{
		if ( ska != NULL ) ska_free( ska );
		if ( pka != NULL ) pka_free( pka );
		if ( hFixedFont != NULL ) DeleteObject( hFixedFont );
		if ( m_titleFont != NULL ) DeleteObject( m_titleFont );
		if ( m_boldFont != NULL ) DeleteObject( m_boldFont );
	}

	bool Generate( cWCStr *wcsLog );
	bool CheckPasswords( HWND hDlg, int bitMask, bool silent = false ); // 0 bit - pwd, 1 bit pwdConfirm
	bool isSKReady();
	bool isPKReady();
	const char *getSignatureAlgOidStr( BYTE *ppkID = NULL );
	bool generateCrtPkt( cPTB *ptb );

	static HFONT GetAFont();

	//
	//	Load wizard block functions
	//		are used for read-only view dialog
	//	Both function return x509e_ok if successfull
	//
	bool m_view;
	int LoadSecretKey( wchar_t *fileName );
	int LoadPublicKey( wchar_t *fileName );

	cWCStr m_algName;
};

class cListViewCustomDrawHandler
{
public:
	static LRESULT OnCustomDraw( LPNMHDR pnmh );
	static DWORD OnPrePaint( LPNMCUSTOMDRAW lpNMCustomDraw );
	static DWORD OnItemPrePaint( LPNMCUSTOMDRAW lpNMCustomDraw );
	static DWORD OnSubItemPrePaint( LPNMCUSTOMDRAW lpNMCustomDraw);
};

//
//	Return number of filled bytes
//	if  buffer is too small returns -1 and sets *pStrSize as required size.
//
int bytes2string( char *str, int *pStrSize, BYTE *bf, int bfSize, int bytesInLine = 0 );
int bytes2string( wchar_t *str, int *pStrLen, BYTE *bf, int bfSize, int bytesInLine = 0 );
int bytes2string( cWCStr *wcs, BYTE *bf, int bfSize, int bytesInLine = 0 );

#endif // __cWizardBlock__