#include "cWizardDlgs.h"
#include <windowsx.h>
//#include "..\BCZip\Help\ctxHelpIDs.h"
#include "..\Includes\menuItem.h"

int dlgColor = COLOR_WINDOW;

bool cWizardDlg::OnStdButton( WPARAM wParam )
{
	GetWindowRect( m_hDlg, &m_pBlock->dlgRect );

	if ( (LOWORD(wParam) == IDOK) ||
		 (LOWORD(wParam) == IDC_btPrev ) ||
		 (LOWORD(wParam) == IDC_btNext )
		 )
	{
		EndDialog( m_hDlg, LOWORD(wParam) );
		return true;
	}
	else
	if ( LOWORD(wParam) == IDCANCEL )
	{
		if ( ( m_pBlock->m_view ) ||
			 ( m_pBlock->wasGenerated ) ||
			 ( IDYES == MessageBoxT(
						m_hDlg,
						_LoadIDS( IDS_ConfirmTermination ),
						_LoadIDS( IDS_AppName ),
						MB_ICONQUESTION | MB_YESNO | MB_DEFBUTTON2 )
			 )
		   )
				EndDialog( m_hDlg, LOWORD(wParam) );

		return true;
	}

	return false;
}

LRESULT cWizardDlg::RealDlgProc( UINT uMsg, WPARAM wParam, LPARAM lParam )
{
	LRESULT result;
	
	if ( cMenuItem::DlgProc( &result, m_hDlg, uMsg, wParam, lParam ) )
		return result;

	switch ( uMsg )
	{
		case WM_HELP:
			OnCtxHelp( (HELPINFO*)lParam );
			return true;

		case WM_CTLCOLORSTATIC:
		{
			// Skips the items lay below line.

			HWND hLine = GetDlgItem( m_hDlg, IDC_Line );
			if ( hLine == NULL )
				break;

			RECT itemRect;
			GetWindowRect( (HWND)lParam, &itemRect );

			RECT lineRect;
			GetWindowRect( hLine, &lineRect );

			if ( itemRect.top > lineRect.top )
				break;

			// Change text and bg colors.

			HWND hBoldText = GetDlgItem( m_hDlg, IDC_STATIC_BOLD );
			if ( hBoldText != NULL )
			{
				if ((LPARAM)hBoldText == lParam )
				{
					SelectFont( (HDC)wParam, m_pBlock->m_boldFont );
					SetBkColor( (HDC)wParam, GetSysColor( dlgColor ) );
					return (BOOL)GetSysColorBrush( dlgColor );
				}
			}
			else
			{
				hBoldText = GetDlgItem( m_hDlg, IDC_Text );

				if ((LPARAM)hBoldText == lParam )
					SelectFont( (HDC)wParam, m_pBlock->m_titleFont );
			}
		
			SetBkColor( (HDC)wParam, GetSysColor( dlgColor ) );
			return (BOOL)GetSysColorBrush( dlgColor );
		}
		case WM_DESTROY:
			GetWindowRect( m_hDlg, &( m_pBlock->dlgRect ) );
			m_wasInited = false;
			break;

		case WM_INITDIALOG:
		{
			SendMessage( m_hDlg, WM_SETICON, ICON_SMALL, (LPARAM)LoadIcon( m_hInst, MAKEINTRESOURCE( IDI_Main ) ) );

			if ( m_pBlock->m_titleFont == NULL )
			{
				HDC hDC = GetDC( m_hDlg );
				int nHeight = -MulDiv(12, GetDeviceCaps( hDC, LOGPIXELSY ), 72);
				m_pBlock->m_titleFont = CreateFontA( nHeight, 0, 0, 0, 
					FW_BOLD, FALSE, FALSE, FALSE, DEFAULT_CHARSET, 
					OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS, DEFAULT_QUALITY, VARIABLE_PITCH | FF_SWISS, "ARIAL");

				nHeight = -MulDiv( 10, GetDeviceCaps( hDC, LOGPIXELSY ), 72);
				m_pBlock->m_boldFont = CreateFontA( nHeight, 0, 0, 0, 
					FW_BOLD, FALSE, FALSE, FALSE, DEFAULT_CHARSET, 
					OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS, DEFAULT_QUALITY, VARIABLE_PITCH | FF_SWISS, "MS Shell Dlg" );
			}

			SetWindowTextT( m_hDlg, _LoadIDS( IDS_AppName ) );

			if ( m_pBlock->dlgRect.left != 0 )
				SetWindowPos( m_hDlg, NULL,
					m_pBlock->dlgRect.left,
					m_pBlock->dlgRect.top,
					0, 0, SWP_NOSIZE | SWP_NOZORDER );

			//
			//	Strange behaivior.
			//	The SetWindowPos sends WM_COMMAND with 0x417 :(((
			//	m_wasInited exclude all command for Welcome dialog.
			//
			m_wasInited = true;
			return OnInitDialog();
		}
		case WM_ERASEBKGND:
		{
			HWND hLine = GetDlgItem( m_hDlg, IDC_Line );
			if ( hLine == NULL )
				break;

			RECT r;
			GetWindowRect( hLine, &r );
			POINT lt = { r.left, r.top };
			ScreenToClient( m_hDlg, &lt );

			GetClientRect( m_hDlg, &r );
			FillRect( (HDC)wParam, &r, GetSysColorBrush( COLOR_BTNFACE ) );

			r.bottom = lt.y;
			FillRect( (HDC)wParam, &r, GetSysColorBrush( dlgColor ) );

			return 1;
		}
		case WM_COMMAND :
			return OnCommand( wParam, lParam );
		case WM_NOTIFY :
			return OnNotify( (NMHDR *)lParam );
		default:
			break;
	}

	return FALSE;
}

int getLastVisible( HWND lBox, const wchar_t *string )
{
	RECT r;
	GetWindowRect( lBox, &r );

	int scrollBarSize = 0;

//	DWORD style = GetWindowStyle( lBox );

//	if ( ( style & WS_VSCROLL ) != 0 )
		scrollBarSize = 16;

	int windowSize = r.right - r.left - scrollBarSize;

//	assert( windowSize );

	HDC hDC = GetDC( lBox );

//	LOGFONT lf;
//	memset( &lf, 0, sizeof(lf) );
	HFONT current = GetWindowFont( lBox );
	HFONT save = ( HFONT )SelectObject( hDC, current );

	SIZE size;

	int len = wcslen( string );

	for ( ; len; len-- )
	{
		GetTextExtentPoint32T( hDC, string, len, &size );
		if ( size.cx < windowSize )
			break;
	}

	SelectObject( hDC, save );

	ReleaseDC( lBox, hDC );

	return len;
}

void AddString( HWND lBox, char *string )
{
	cWCStr wcs( string );
	AddString( lBox, wcs.m_szBody );
}

void AddString( HWND lBox, wchar_t *string )
{
	int len = wcslen( string );

	int last = getLastVisible( lBox, string );

	if ( last >= len )
		ListBox_AddStringT( lBox, string );
	else
	{
		int lastProbel = last;
		for( ;; lastProbel-- )
		{	if ( lastProbel == 0 )
			{
				lastProbel = last;
				break;
			}
			else
			if ( string[ lastProbel ] == L' ')
				break;
		}

		wchar_t save = string[ lastProbel + 1];
		string[ lastProbel + 1] = 0;

		ListBox_AddStringT( lBox, string );

		string[ lastProbel + 1 ] = save;

		AddString( lBox, string + lastProbel + 1);
	}
}

int AddStrings( HWND hLB, char *str )
{
	cWCStr wcs( str );
	return AddStrings( hLB, wcs.m_szBody );
}

int AddStrings( HWND hLB, wchar_t *str )
{
	int i = 1;
	
	for( wchar_t *p = str;; p++ )
	{
		if ( *p == L'\0' )
		{
			AddString( hLB, str );
			break;
		}
		else
		if ( *p == L'\n' )
		{
			*p = L'\0';
			AddString( hLB, str );
			str = p + 1;
			i++;
		}
	}

	return i;
}
