#include "cWizardDlgs.h"
#include "resource.h"
#include "BCDB_Functions.h"

#include "..\CertificateLib\SMemory\SBlocks.h"
#include "..\CertificateLib\x509pkts\oidFuncs.h"
#include "..\CertificateLib\pgpPkts\Packets.h"
#include "..\CertificateLib\pgpPkts\PKTSignature.h"
#include "..\CertificateLib\publicKeyAlgs\PublicKeyAlgs.h"
#include "..\CertificateLib\x509pkts\x509_structs.h"

#include "..\Includes\OpenFiles.h"
#include "..\Includes\FileView.h"
#include "..\Includes\UnicodeFuncs.h"
#include "..\Includes\menuItem.h"

#include <commctrl.h>
#include <stdio.h>
#include <windowsx.h>

static UINT ColumnLabelIDS[] = 
{
	IDS_szField, IDS_szValue
};

const int NUM_COLUMNS = sizeof( ColumnLabelIDS ) / sizeof( UINT );

static int _gnColumnFmt[ NUM_COLUMNS ] =
{
	LVCFMT_LEFT, LVCFMT_LEFT
};

static int _gnColumnWidth[NUM_COLUMNS]=
{
	140, 288
};

static BOOL InitListViewColunms( HINSTANCE hInst, HWND hLView )
{
	int i;
	LV_COLUMNW lvc;

	memset( &lvc, 0, sizeof( lvc ) );

	lvc.mask = LVCF_WIDTH;

	//
	//	Check if already initialised...
	//
	if ( ListView_GetColumnT( hLView, 0, &lvc ) )
		return TRUE;

	lvc.mask = LVCF_FMT | LVCF_WIDTH | LVCF_TEXT | LVCF_SUBITEM;

	wchar_t text[ 80 ];

	int res;

	for(i=0; i<NUM_COLUMNS; i++)
	{
		LoadStringT( hInst, ColumnLabelIDS[i], text, maxLen( text ) );

		lvc.iSubItem=i;
		lvc.pszText = text; // _gszColumnLabel[i];
		lvc.cx = _gnColumnWidth[i];
		lvc.fmt=_gnColumnFmt[i];

		res = ListView_InsertColumnT( hLView, i, &lvc );

		if ( res == -1 )
		{
			MessageBoxT( 0, text, _LoadIDS( IDS_ErrorInAddLVColumn ), MB_OK );
			return FALSE;
		}
	}

	return TRUE;
}

bool cWizardDlg_PubKeyCrt::VerifyCertificate()
{
	// See ->  cWizardBlock::LoadPublicKey( wchar_t *fileName )

	int res = m_pBlock->VerifyCertificate();

	if ( res == x509e_ok )
		return true;

	wchar_t szErr[ 128 ] = L"";
	switch ( res  )
	{
		case ppe_keyNotFound:
			LoadStringT( m_hInst, IDS_IssuerPublicKeyIsNotFoundInBCDB, szErr, maxLen( szErr ) );
			break;
		case x509e_emptyParameters:
		case x509e_extensionIsNotPresent:
			::FormatT( szErr, maxLen( szErr ), L"Paramaters are invalid!", res );
			break;
		default:
			::FormatT( szErr, maxLen( szErr ), _LoadIDS( IDS_ErrorNumber_X ), res );
	}


	return DlgMessageBoxT(
		cWCStr().FormatT( L"%s\n\n%s", szErr, _LoadIDS( IDS_ContinueQ ) ),
		MB_YESNO,
		_LoadIDS( IDS_ErrorInCertificateVerification ) )
			== IDYES;
}

bool cWizardDlg_PubKeyCrt::UpdateDateItem( int id, SYSTEMTIME *tm )
{
	wchar_t dateStr[ 100 ];

	GetDateFormatW(
		LOCALE_USER_DEFAULT, LOCALE_NOUSEROVERRIDE, //LOCALE_USE_CP_ACP
		tm, NULL, dateStr, sizeof( dateStr ) );

	ListView_SetItemTextT( m_hLV, id, 1, dateStr );

	return true;
}

bool cWizardDlg_PubKeyCrt::UpdateSubjectRDN( cWCStr *wcs )
{
	int size = 0;
	int res = m_pBlock->rdnSubject.GenerateName( NULL, &size );
	if ( res != x509e_bufferTooSmall )
	{
		assert( false );
		return false;
	}

	if ( size <= 1 )
	{
		ListView_SetItemTextT( m_hLV, lvId_subjectRdn, 1, _LoadIDS( IDS_szIsEmpty ) );
		ListView_SetItemParam( m_hLV, lvId_subjectRdn, 1 );
	}
	else
	{
		cPlainTextBlock ptb;
		if ( ptb.Allocate( size ) == NULL )
			return false;

		res = m_pBlock->rdnSubject.GenerateName( (char *)ptb.buffer, &size );
		if ( res != x509e_ok )
		{
			assert( false );
		}

		ListView_SetItemTextT( m_hLV, lvId_subjectRdn, 1, getWCStr( ( char * )ptb.buffer ) );
		ListView_SetItemParam( m_hLV, lvId_subjectRdn, 0 );

		if ( wcs != NULL )
			wcs->Set( ( char * )ptb.buffer );
	}

	return true;
}

bool cWizardDlg_PubKeyCrt::UpdateAuthorityKeyID()
{
	cWCStr wcs;
	int lParam = 0;

	if ( m_pBlock->m_authorityInfo.bInited )
	{
		m_pBlock->m_authorityInfo.writeToPtb( &wcs, false );

		BCDB_GetCrtSerialNumber( &m_pBlock->m_authorityInfo.serialNumber, m_pBlock->m_authorityInfo.ptbKeyID.buffer, m_pBlock->m_authorityInfo.ptbKeyID.bufferSize, false );
	}
	else
	if ( m_pBlock->skaIssuer != NULL )
		getLocalKeyID( m_pBlock->skaIssuer, &wcs, true, false );
	else
	if ( m_editState != esEditionInited )
		wcs.Add( _LoadIDS( IDS_szSelfSigned ) );
	else
	{
		wcs.Add( _LoadIDS( IDS_Undefined ) );
		lParam = 1;
	}
	
	ListView_SetItemTextT( m_hLV, lvId_authrityKeyID, 1, wcs.m_szBody );
	ListView_SetItemParam( m_hLV, lvId_authrityKeyID, lParam );

	return true;
}

bool cWizardDlg_PubKeyCrt::UpdateSignatureAlgorithm()
{
//	BYTE pkID = m_pBlock->pkAlgID;

	const char *oidStr = m_pBlock->getSignatureAlgOidStr(); //  &pkID );
	if ( oidStr == NULL )
	{
//		assert( false );

		wchar_t bf[ 64 ];
		cWCStr wcs;
		wcs.FormatT( _LoadIDS( IDS_STRING135 ),
					getWCStr( PKT_GetHashAlgName( m_pBlock->hashAlgID ) ),
					m_pBlock->hashAlgID,
					getWCStr( PKT_GetPublicKeyAlgName( (BYTE)m_pBlock->pkAlgID ), bf, maxLen( bf ) ),
					m_pBlock->pkAlgID );

		MessageBoxT( m_hDlg, wcs.m_szBody, _LoadIDS( IDS_STRING136 ), MB_ICONSTOP );
//		return false;

		ListView_SetItemTextT( m_hLV, lvId_signatureAlgorithm, 1, _LoadIDS( IDS_Undefined ) );
	}
	else
	{
		tUnit *t = NULL;
		if ( ( generateOIDName( NULL, 0, oidStr, &t ) == 0 ) || ( t== NULL ) ||
			( ( t->name == NULL ) && ( ( t->desc == NULL ) ) )
			)
		{
			assert( false );
			MessageBoxT( m_hDlg, getWCStr( oidStr ), _LoadIDS( IDS_STRING137 ), MB_ICONSTOP );
			return false;
		}

		cWCStr wcs( ( t->name != NULL ? t->name : t->desc ) );

		ListView_SetItemTextT( m_hLV, lvId_signatureAlgorithm, 1, wcs.m_szBody );
	}

	return true;
}

bool cWizardDlg_PubKeyCrt::UpdateFields( bool editableOnly )
{
	int iFocused = ListView_GetNextItem( m_hLV, -1, LVNI_FOCUSED );
	if ( iFocused == -1 )
		iFocused = 0;

	// lvId_version
	ListView_SetItemTextT( m_hLV, lvId_version, 1, _FormatT( L"V%d", m_pBlock->version + 1 ) );  // 2 means 3

	// lvId_serialNumber,

	UpdateSerialNumber();

	// lvId_signatureAlgorithm,
	UpdateSignatureAlgorithm();

	// lvId_issuerRdn,
	UpdateIssuerRDN();


	// lvId_validFrom,
	UpdateDateItem( lvId_validFrom, &m_pBlock->nb );

	// lvId_validTo,
	UpdateDateItem( lvId_validTo, &m_pBlock->na );

	// lvId_subjectRdn,
	UpdateSubjectRDN();

	if ( editableOnly )
	{
		if ( iFocused > lvId_subjectRdn )
			iFocused = lvId_subjectRdn;

		ListView_SetItemState( m_hLV, iFocused, LVIS_FOCUSED, LVIS_FOCUSED );

		return true;
	}

	// lvId_publicKey,

	wchar_t text[ 256 ];
//	int pkID = pgppkaRSA;
	if ( m_pBlock->ska != NULL )
	{
		m_pBlock->keyBitSize = ska_getModuleSize( m_pBlock->ska ) * 8;
		m_pBlock->pkAlgID = ska_getAlgID( m_pBlock->ska );
	}
	else
	if ( m_pBlock->pka != NULL )
	{
		m_pBlock->keyBitSize = pka_getModuleSize( m_pBlock->pka ) * 8;
		m_pBlock->pkAlgID = pka_getAlgID( m_pBlock->pka );
	}
	
	FormatT( text, maxLen( text ), _LoadIDS( IDS_NameAndBits_SD ), cWCStr( PKT_GetPublicKeyAlgName( m_pBlock->pkAlgID ) ).m_szBody, m_pBlock->keyBitSize );
	ListView_SetItemTextT( m_hLV, lvId_publicKey, 1, text );

	// lvId_subjectKeyID,
	//cPlainTextBlock ptbSubjKeyID;
	cWCStr wcs;
	if ( ( m_pBlock->ska != NULL ) && ( m_pBlock->pka == NULL ) )
	{
		ska_derivePublicKeyAlgorithm( &m_pBlock->pka, m_pBlock->ska );
	}

	getLocalKeyID( m_pBlock->pka, &wcs, false, false );
	ListView_SetItemTextT( m_hLV, lvId_subjectKeyID, 1, wcs.m_szBody );

	// lvId_authrityKeyID,
	UpdateAuthorityKeyID();

	// lvId_basicConstraints,

	if ( !m_pBlock->m_view )
		m_pBlock->bCA = m_pBlock->skaIssuer == NULL;

	if ( m_pBlock->bCA )
	{
		wchar_t szPathLength[ 32 ];

		if ( m_pBlock->bcPathLength == -1 )
			LoadStringT( m_hInst, IDS_szNone, szPathLength, maxLen( szPathLength ) );
		else
			::FormatT( szPathLength, maxLen( szPathLength ), L"%d", m_pBlock->bcPathLength );

		FormatT( text, maxLen( text ), _LoadIDS( IDS_STRING139 ), szPathLength );
	}
	else
		wcscpy_s( text, L"" );
	
	ListView_SetItemTextT( m_hLV, lvId_basicConstraints, 1, text );

	ListView_SetItemState( m_hLV, iFocused, LVIS_FOCUSED, LVIS_FOCUSED );

	return true;
}

int ids_tab[] = 
{
	IDS_szVersion,
	IDS_szSerialNumber,
	IDS_szSignatureAlgorithm,
	IDS_szIssuerRdn,
	IDS_szValidFrom,
	IDS_szValidTo,
	IDS_szSubjectRdn,
	IDS_szPublicKey,
	IDS_szSubjectKeyID,
	IDS_szAuthrityKeyID,
	IDS_szBasicConstraints
};

LONG cWizardDlg_PubKeyCrt::lpNativeLVProc = NULL;

static DWORD keySizeMin = 512;
static DWORD keySizeMax = 8 * 1024;
static DWORD keySizeStep = keySizeMin;

BOOL CALLBACK cWizardDlg_PubKeyCrt::LVWndProc( HWND hLV, UINT uMsg, WPARAM wParam, LPARAM lParam )
{
	if ( lpNativeLVProc == NULL )
	{
		assert( false );
		return FALSE;
	}

	if ( uMsg == WM_NOTIFY )
	{
/*
		NMHDR *hdr = (NMHDR *)lParam;
		if ( hdr->code == UDN_DELTAPOS )
		{
			HWND hDlg = GetParent( hLV );
			cWizardDlg_PubKeyCrt *pThis = (cWizardDlg_PubKeyCrt *)GetWindowLong( hDlg, DWL_USER);

			if ( hdr->hwndFrom == pThis->m_popup1 )
				return cDlg::UDNDeltapos( (NMUPDOWN *)hdr, pThis->m_popup, keySizeMin, keySizeMax, keySizeStep );
		}
*/
	}
	else
	if ( uMsg == WM_COMMAND )
	{
		HWND hDlg = GetParent( hLV );
		cWizardDlg_PubKeyCrt *pThis = (cWizardDlg_PubKeyCrt *)GetWindowLong( hDlg, DWL_USER);

		if ( (HWND)lParam == pThis->m_popup )
		{
			if ( pThis->m_editNum == lvId_issuerRdn )
			{
				pThis->OnBrowseIssuer();
//				pThis->UpdateIssuerRDN();
				pThis->StopEdit();
//				pThis->OnEditItem( pThis->m_editNum, true );
			}
			else
			if ( pThis->m_editNum == lvId_subjectRdn )
			{
				pThis->OnGetSubjectRDN();
				pThis->StopEdit();
//				pThis->OnEditItem( pThis->m_editNum, true );
				
				cWCStr wcs;
				pThis->UpdateSubjectRDN( &wcs );
				pThis->FillInfo( &wcs, false );
			}
		}
	}

	return CallWindowProc( (WNDPROC)lpNativeLVProc, hLV, uMsg, wParam, lParam );
}

//
//	friendlyNameToRDN
//		Converts friendly name to attribute list 
//		friendly name : 'John Smith <JohnSmit@aol.com>'
//		attribute list string: CN = John Smith, E = JohnSmith@aol.com
//
bool friendlyNameToRDN( cUniversalString *friendlyName, sNameAttributeList *attList )
{
	cUniversalString us;
	us.SetUnicode( (BYTE *)friendlyName->unicodeBuffer, friendlyName->unicodeBufferSize  );
	us.MakeFromUnicodeString();

	if ( wcscmp( us.wcBuffer, _LoadIDS( IDS_szIsEmpty ) ) == 0 )
		us.ClearBuffers();

	int len = strlen( us.buffer );

	int i;

	for ( i = len; i--; )
	{
		if ( us.buffer[i] == '>' )
			us.buffer[i] = 0;
		if ( us.buffer[i] == '<' )
		{
			us.buffer[i] = '\0';

			attList->Add( sNameAttribute::pkcs9_emailAddress, us.buffer + i + 1, asn1tag_IA5String );

			break;
		}
	}

	if ( i > 0 ) // means e-mail was found.
	{
		for ( ; i--; )
		{
			if ( ( us.buffer[i] == ' ' ) || ( us.buffer[i] == '-' ) )
				us.buffer[i] = '\0';
			else
				break;
		}
	}

	len = strlen( us.buffer );
	if ( len != 0 )
		attList->Add( sNameAttribute::id_at_commonName, us.buffer );

	return true;
}

bool cWizardDlg_PubKeyCrt::OnGetSubjectRDN()
{
	sNameAttributeList rdn;
	rdn.AddCopyList( m_pBlock->rdnSubject.head );

	if ( !GetRDN( m_hInst, m_hDlg, &rdn, _LoadIDS( IDS_GetCrtSubjRDN ), &m_pBlock->dlgRect ) ) // "Get certificate subject relativedistinguished name" ) )
		return false;

	m_pBlock->rdnSubject.freeList();
	m_pBlock->rdnSubject.AddCopyList( rdn.head );
	UpdateSubjectRDN();
	UpdateSTDButtons();
	return true;
}

//
//	OnBrowseIssuer
//	1. Browses secret key in BCStorage and opens it's content ( password required ) 
//	2. Parses the key block and creates m_pBlock->skaIssuer secret key
//	The issuer RDN is ordinary friendly name for P12 secret key packet, so
//	3. Tries to find corresponded ( by local key ID ) public key certificate to get it RDN to m_pBlock->rdnIssuer.
//		If the certificate is not found, so the friendly name may be parsed as 'Common Name' and <e-Mail address>
//		both the values are parts of RDN. 
//
bool cWizardDlg_PubKeyCrt::OnBrowseIssuer()
{
	// 1.
	//wchar_t *szFiltersW = 
	//	L"PKCS 12 ( *.p12 )\0*.p12\0"
	//	L"All files ( *.* )\0*.*\0";

	wchar_t szFilters[ 1024 ];
	LoadStringT( hMainInst, IDS_szFiltersSecretP12, szFilters, maxLen( szFilters ) );

	wchar_t pathW[ MAX_PATH ] = L"\0";

	wchar_t text[ 1024 ];
	LoadStringT( hMainInst, IDS_szChooseSecretKeyAndPressGet, text, maxLen( text ) );
	wcscat_s( text, _LoadIDS( IDS_TheSKWillBeUsedForNewPairGeneration ) );

	bool res = BrowseKeyFiles( 
			m_hDlg,
			pathW, maxLen( pathW ),
			szFilters,
			_LoadIDS( IDS_ChooseSecretKey ),
			text,
			_LoadIDS( IDS_szGetItBtnText )
		);

	if ( !res )
	{
		if ( m_pBlock->skaIssuer != NULL )
		{
			if ( IDYES == 
					MessageBoxT(
						m_hDlg,
						_LoadIDS( IDS_szDoYouWantToCreateSelfSignedCRT_Q ),
						_LoadIDS( IDS_AppName ),
						MB_YESNO | MB_ICONQUESTION | MB_DEFBUTTON2)
				)
			{
				ska_free( m_pBlock->skaIssuer );
				m_pBlock->skaIssuer = NULL;
				m_pBlock->rdnIssuer.freeList();
				m_newIssuer = true;
			}
		}
		return false;
	}

	// 2.
	cUniversalString usFriendlyName;

	cFileView cfv( pathW );

	if ( cfv.m_err )
		return false;

	struct autoDestruct
	{
		HPKA ska;
		sNameAttributeList rdn;

		autoDestruct()
		{
			ska = NULL;
		}
		~autoDestruct()
		{
			if ( ska != NULL )
				ska_free( ska );
		}

	} issuer;


	if ( !BCDB_GetP12SecretKeyInfo( m_hInst, m_hDlg, cfv.m_ptr, (int)cfv.m_size.QuadPart, &issuer.ska, &usFriendlyName ) )
		return false;
	
	// 3.
	cPlainTextBlock ptbKeyID;
	if ( !BCDB_GenerateLocalKeyID( &ptbKeyID, issuer.ska ) )
	{ return false; }

	cPlainTextBlock crtPtb;
	x509_certificate crt;
	int errIDS = 0;

	if ( !BCDB_GetX509KeyByID( &crtPtb, ptbKeyID.buffer, ptbKeyID.bufferSize, false ) )
	{
		errIDS = IDS_crtWasNotFound;
	}
	else
	if ( crt.parse( crtPtb.buffer, crtPtb.bufferSize ) != x509e_ok )
	{
		errIDS = IDS_crtIsInvalid;
	}
	else
	{
		if ( issuer.rdn.AddCopyList( crt.tbsCertificate.subjectName.head ) <= 0 )
			errIDS = IDS_crtSubjectIsInvalid;
		else
			errIDS = 0;
	}

	if ( errIDS )
	{
		// Certificate with issuer local key ID was not found or invalid.
		// We'll split issuer secret key friendly name as RDN 'John Smith <JohnSmith@aol.com>'
		// CN = John Smith, E = JohnSmith@aol.com

		friendlyNameToRDN( &usFriendlyName, &issuer.rdn );

		//cPlainTextBlock ptbRDN;
		cWCStr wcs;
		if ( !rdn2ptb( &wcs, &issuer.rdn ) )
			return false;

		int res = MessageBoxT( m_hDlg, cWCStr().FormatT( _LoadIDS( IDS_STRING109 ) , _LoadIDS( errIDS ), wcs.m_szBody ), _LoadIDS( IDS_AppName ), MB_ICONQUESTION | MB_OKCANCEL );

		if ( res == IDCANCEL )
			return false;
	}
	else
	{
		// Certificate with issuer local key ID was found and m_pBlock->rdnIssuer successfully filled.
	}

	if ( m_pBlock->skaIssuer )
		ska_free( m_pBlock->skaIssuer );

	m_pBlock->skaIssuer = issuer.ska;
	issuer.ska = NULL; // To skip destruction.
	m_pBlock->rdnIssuer.freeList();
	m_pBlock->rdnIssuer.AddCopyList( issuer.rdn.head );
	m_newIssuer = true;

	UpdateIssuerRDN();

	UpdateSTDButtons();

	return true;
}
bool cWizardDlg_PubKeyCrt::InsertLVItems( bool editableOnly )
{
	ListView_DeleteAllItems( m_hLV );

	// Inserts list view items
	LVITEMW lvi;

	memset( &lvi, 0, sizeof( lvi ) );

	lvi.mask = 0; // LVIF_IMAGE;

	int last = editableOnly ? lvId_subjectRdn : lvId_last;

	for ( lvi.iItem = lvId_first; lvi.iItem <= last; (lvi.iItem)++ )
	{
		if ( ListView_InsertItemT( m_hLV, &lvi ) != lvi.iItem )
		{
			assert( false );
			return false;
		}

		ListView_SetItemTextT( m_hLV, lvi.iItem, 0, _LoadIDS( ids_tab[lvi.iItem] ) );
	}

	return true;
}

static DWORD dlgHelpIDs[] = 
{
    //IDC_lbFieldDetails,	IDH_lbFieldDetails,
    //IDC_lvFields,		IDH_lvFields,
    //IDC_cmdShowDetails,	IDH_cmdShowDetails,
    //IDC_Notes,			IDH_Notes,
	0,0
};

static DWORD lvHelpIDs[] = 
{
	//lvId_version,			IDH_lvId_version,
	//lvId_serialNumber,		IDH_lvId_serialNumber,
	//lvId_signatureAlgorithm,IDH_lvId_signatureAlgorithm,
	//lvId_issuerRdn,			IDH_lvId_issuerRdn,
	//lvId_validFrom,			IDH_lvId_validFrom,
	//lvId_validTo,			IDH_lvId_validTo,
	//lvId_subjectRdn,		IDH_lvId_subjectRdn,
	//lvId_publicKey,			IDH_lvId_publicKey,
	//lvId_subjectKeyID,		IDH_lvId_subjectKeyID,
	//lvId_authrityKeyID,		IDH_lvId_authrityKeyID,
	//lvId_basicConstraints,	IDH_lvId_basicConstraints,
	0,0
};

static void LV_HelpPopup( HWND hLV, HELPINFO *hInfo, DWORD *helpIDs )
{
	LVHITTESTINFO hi;
	memset( &hi, 0, sizeof( hi ) );
	hi.flags = LVHT_ONITEM;

	//GetCursorPos( &hi.pt );
	hi.pt = hInfo->MousePos;
	ScreenToClient( hLV, &hi.pt );
	hi.pt.x = 10;

	int res = ListView_HitTest( hLV, &hi );
	if ( res == -1 )
		return;
	
	hInfo->iCtrlId = res;

	cDlg::OnHelpPopup( hInfo, helpIDs );
}

void cWizardDlg_PubKeyCrt::OnCtxHelp( HELPINFO *hInfo )
{
	int ids = 0;

	if ( hInfo->iCtrlId == IDC_lvFields )
	{
		LV_HelpPopup( m_hLV, hInfo, lvHelpIDs );
	}
	else
		cDlg::OnHelpInfo( hInfo, dlgHelpIDs );
}

const int aNumber = 25; 

bool cWizardDlg_PubKeyCrt::OnInitDialog()
{
	//
	// (WM_USER + 1) allows to catch WM_COMMAND with the number if VK_RETURN pressed
	//
	SendMessage( m_hDlg, WM_USER + 1, aNumber, 0 );

	INITCOMMONCONTROLSEX commCtrl;
	commCtrl.dwSize = sizeof( INITCOMMONCONTROLSEX );
	commCtrl.dwICC = ICC_DATE_CLASSES;
	InitCommonControlsEx( &commCtrl );

	m_popup = NULL;

	//
	//	Set dialogue caption 
	//
	if ( m_pBlock->m_view )
	{
		SetDlgItemTextT( IDC_STATIC_BOLD, _LoadIDS( IDS_PublicKeyCertificate ) );
		SetDlgItemTextT( IDC_Notes, L"" );
	}

	if ( !m_pBlock->m_view )
	{
		SYSTEMTIME tm;
		GetSystemTime( &tm );
		m_pBlock->nb = tm;
		tm.wYear++;
		m_pBlock->na = tm;
	}

	m_hLV = GetDlgItem( m_hDlg, IDC_lvFields );
	ListView_SetExtendedListViewStyleEx( m_hLV, LVS_EX_GRIDLINES | LVS_EX_FULLROWSELECT, LVS_EX_GRIDLINES | LVS_EX_FULLROWSELECT );

	InitListViewColunms( m_hInst, m_hLV );

	//
	// Get address of List View WinProc and change it to internal LVWndProc
	//	It is required for interception of UPDOWN control nitification messages.
	//
	lpNativeLVProc = GetWindowLong( m_hLV, GWL_WNDPROC );
	SetWindowLong( m_hLV, GWL_WNDPROC, (LONG) LVWndProc );

	ComboBox_AddStringT( GetDlgItem( m_hDlg, IDC_cmdShowDetails ), _LoadIDS( IDS_ShowAll ) );

	if ( !m_pBlock->m_view )
	{
		ComboBox_AddStringT( GetDlgItem( m_hDlg, IDC_cmdShowDetails ), _LoadIDS( IDS_ShowHastToBeFilled ) );
		ComboBox_SetCurSel( GetDlgItem( m_hDlg, IDC_cmdShowDetails ), m_pBlock->pkShowNum );
	}
	else
		ComboBox_SetCurSel( GetDlgItem( m_hDlg, IDC_cmdShowDetails ), 0 );

	HWND hLB = GetDlgItem( m_hDlg, IDC_lbFieldDetails );
	m_fontOriginal = GetWindowFont( hLB );


	//
	//	Fills rnd subject from secret key friendly name value.
	//
	if ( m_pBlock->rdnSubject.head == NULL )
	{
		cUniversalString us;
		char friendlyNameA[ 1024 ];
		wcstombs( friendlyNameA, m_pBlock->friendlyName, sizeof( friendlyNameA ) );
		us.Set( friendlyNameA );
		us.GenerateUnicodeString();
		sNameAttributeList rdn;
		friendlyNameToRDN( &us, &rdn );
		m_pBlock->rdnSubject.AddCopyList( rdn.head );
	}

	SetItemsValues();

	UpdateSTDButtons();

	return true;
}

bool cWizardDlg_PubKeyCrt::SetItemsValues()
{
	bool editableOnly = m_pBlock->pkShowNum != 0;

	if ( !InsertLVItems( editableOnly ) )
		return false;

	return UpdateFields( editableOnly );
}

bool cWizardDlg_PubKeyCrt::OnCommand( WPARAM wParam, LPARAM lParam )
{
	if ( m_popup )
	{
		bool cancelled = ( LOWORD(wParam) == IDCANCEL );
		StopEdit( false, cancelled );
		return true;
	}
	else
	if ( wParam == aNumber )
	{
		int iFocused = ListView_GetNextItem( m_hLV, -1, LVNI_FOCUSED );
		return OnEditItem( iFocused, false, true );
	}

	if ( LOWORD(wParam) == IDOK )
	{
		if ( m_editState == esAllowEdition )
		{
			if ( !VerifyCertificate() )
				return true;

			m_editState = esEditionInited;
			if ( m_pBlock->skaIssuer == NULL )
			{
				m_pBlock->rdnIssuer.freeList();
				UpdateIssuerRDN();
				m_pBlock->m_authorityInfo.bInited = false;
				UpdateAuthorityKeyID();
			}
			UpdateSerialNumber( false );
			UpdateSTDButtons();
			return true;
		}

		if ( m_editState == esEditionInited )
		{
			cWCStr wcsLog;
			m_pBlock->generateSecKey = false;
			m_pBlock->generateCertificate = true;
			m_pBlock->saveToBCStorage = true;
			bool ok = m_pBlock->Generate( &wcsLog );
			if ( ok )
				UpdateSerialNumber( true );

			DlgMessageBoxT( wcsLog.m_szBody, MB_OK, _LoadIDS( IDS_UpdateCertificate ) );
			//m_editState = esAllowEdition;
			//UpdateSTDButtons();
			EndDialog( m_hDlg, LOWORD(wParam) );
			return true;
		}
	}

	if ( OnStdButton( wParam ) )
	{
		return true;
	}
	else
	{
		switch ( LOWORD(wParam) )
		{
			case IDC_cmdShowDetails:
				if ( HIWORD( wParam ) == CBN_SELCHANGE )
				{
					m_pBlock->pkShowNum = ComboBox_GetCurSel( GetDlgItem( m_hDlg, IDC_cmdShowDetails ) );
					SetItemsValues();
				}
				break;
			default:
				return false;
		}
		return true;
	}
}

bool cWizardDlg_PubKeyCrt::OnNotify( NMHDR *hdr )
{
	bool dblClk = false;

	switch( hdr->code )
	{
		case UDN_DELTAPOS:
			break;
		case NM_KILLFOCUS :
//			StopEdit( true );
			break;
		case NM_DBLCLK:
			dblClk = true;
		case NM_CLICK :
		{
			LVHITTESTINFO hi;
			memset( &hi, 0, sizeof( hi ) );
			hi.flags = LVHT_ONITEM;

			GetCursorPos( &hi.pt );
			ScreenToClient( m_hLV, &hi.pt );
			hi.pt.x = 10;

			int res = ListView_HitTest( m_hLV, &hi );

			//
			//	1. Do not stop edit if current item was clicked
			//	2. Stop current edit item before edit new.
			//
			if ( ( m_popup != NULL ) && ( res != -1 ) )
			{
				if ( hi.iItem != m_editNum )
				{
					StopEdit();
					OnEditItem( hi.iItem, true, dblClk ); // type info only == true
				}

				break;
			}

			StopEdit();

			if ( hdr->idFrom == IDC_lvFields )
			{
				if ( res != -1 )
				{
					OnEditItem( hi.iItem, false, dblClk );
				}
			}
		
			break;
		}
		case NM_CUSTOMDRAW:
		{
			int res = cListViewCustomDrawHandler::OnCustomDraw( hdr );
			SetWindowLong( m_hDlg, DWL_MSGRESULT, res ); 
			return TRUE;
		}

		default : break;
	}

	return FALSE;
}

//
//	UpdateSerialNumber
//	if issuer is undefined serial number is 1 ( means self signed certificate with new key )
//	Otherwise, issuer has to be choosen from BestCrypt DataBase, BCDB stores certificate serial number for
//	every secret key.
//
bool cWizardDlg_PubKeyCrt::UpdateSerialNumber( bool increment )
{
	if ( ( !m_pBlock->m_view ) || ( m_editState == esEditionInited ) )
	{
		int sn = -1;

		HPKA skaIssuer = m_pBlock->skaIssuer;

		if ( skaIssuer == NULL )
		{
			if ( m_pBlock->m_authorityInfo.bInited )
				sn = m_pBlock->m_authorityInfo.serialNumber;
			else
			{
				//
				//	Self signed
				//
				skaIssuer = m_pBlock->ska;
			}
		}

		if ( skaIssuer == NULL )
		{
			if ( sn == -1 )
			{
				//assert( false );
				ODS_fls( L"Assertion!" );
				sn = 1;
			}
		}
		else
		{
			// 1. Generates local key ID
			cPlainTextBlock ptb;
			if ( !BCDB_GenerateLocalKeyID( &ptb, skaIssuer ) )
				return false;


			// 2. Gets number from BestCrypt Database.
			if ( !BCDB_GetCrtSerialNumber( &sn, ptb.buffer, ptb.bufferSize, increment ) && increment )
			{
				return false;
			}
		}

		m_pBlock->serialNumber = sn;
	}

	wchar_t text[ 256 ];

	if ( m_pBlock->serialNumberDWL != 0 )
		::FormatT( text, maxLen( text ), L"0x%I64X", m_pBlock->serialNumberDWL );
	else
	if ( m_pBlock->serialNumber < 0 )
		LoadStringT( m_hInst, IDS_Undefined, text, maxLen( text ) );
	else
		::FormatT( text, maxLen( text ), L"%d", m_pBlock->serialNumber );

	ListView_SetItemTextT( m_hLV, lvId_serialNumber, 1, text );

	return true;
}

//
//	UpdateIssuerRDN
//	
//
bool cWizardDlg_PubKeyCrt::UpdateIssuerRDN()
{
	cWCStr wcs;

	if ( m_pBlock->rdnIssuer.isEmpty() ) // It assumes self signed
	{
		if ( m_editState != esEditionInited )
			wcs.Add( _LoadIDS( IDS_szSelfSigned ) );
	}
	else
	{
		char buff[1024];
		int size = maxLen( buff );
		m_pBlock->rdnIssuer.GenerateName( buff, &size );
		wcs.Add( buff );
	}

	int lParam = 0;

	if ( wcs.isEmpty() )
	{
		wcs.Add( _LoadIDS( IDS_szIsEmpty ) );
		lParam = 1;
	}

	ListView_SetItemTextT( m_hLV, lvId_issuerRdn, 1, wcs.m_szBody );
	ListView_SetItemParam( m_hLV, lvId_issuerRdn, lParam );

	if ( m_newIssuer )
	{
		UpdateAuthorityKeyID();
		UpdateSignatureAlgorithm();
	}

	UpdateSerialNumber();
	m_newIssuer = false;

	return true;
}


bool cWizardDlg_PubKeyCrt::FillInfo( cWCStr *wcs, bool useFixedFont )
{
	HWND hLB = GetDlgItem( m_hDlg, IDC_lbFieldDetails );
	ListBox_ResetContent( hLB );

	HFONT font = useFixedFont ? m_pBlock->hFixedFont : m_fontOriginal;
	SetWindowFont( hLB, font, 1 );

	//ptb->PutByte( 0 ); // EOS
	AddStrings( hLB, wcs->m_szBody );//(char *)ptb->buffer );
	return true;
}

int cWizardDlg_PubKeyCrt::OnEditDate( RECT *r )
{
	int delta = 1;
	m_popup = CreateWindowExT( 
		0, 
		DATETIMEPICK_CLASSW,   // predefined class 
		L"DateTime",       // window text 
		WS_VISIBLE | WS_CHILD,// | DTS_SHOWNONE, // styles 
//		20,50,220,20,

		r->left - delta + 1,	// starting x position 
		r->top - delta - 2,	// starting y position 
		r->right - r->left + 2 * delta,				// width 
		r->bottom - r->top + 2 * delta + 2,			// height 

		m_hLV,							// parent window 
		NULL,							// No menu 
		m_hInst,						// (HINSTANCE) GetWindowLong(hwnd, GWL_HINSTANCE), 
		NULL);

	HFONT hFont = GetWindowFont( m_hLV );
	if ( hFont )
		SetWindowFont( m_popup, hFont, false );

	SYSTEMTIME tm;
	if ( m_editNum == lvId_validFrom )
		tm = m_pBlock->nb;
	else
		tm = m_pBlock->na;

	DateTime_SetSystemtime( m_popup, GDT_VALID, &tm );

	SetFocus( m_popup );

//	ShowWindow( m_popup, SW_SHOW );

	return 0;
}

int cWizardDlg_PubKeyCrt::CreateBrowseButton( RECT *r )
{
	int size = r->bottom - r->top;

	m_popup = CreateWindowExT( 
		0, 
		L"BUTTON",   // predefined class 
		L"",       // window text 
		WS_VISIBLE | WS_CHILD, // styles 
		r->right - size - 2,	// starting x position 
		r->top - 2,	// starting y position 
		size + 3,				// width 
		size + 3,			// height 
		m_hLV,							// parent window 
		NULL,							// No menu 
		m_hInst,						// (HINSTANCE) GetWindowLong(hwnd, GWL_HINSTANCE), 
		NULL);

	HFONT hFont = GetWindowFont( m_hLV );
	if ( hFont )
		SetWindowFont( m_popup, hFont, false );

	SetFocus( m_popup );

	SetWindowTextA( m_popup, "..." );

	return 0;
}

int cWizardDlg_PubKeyCrt::OnEditIssuer( RECT *r, cWCStr *wcs, bool typeInfoOnly )
{
	rdn2ptb( wcs, &m_pBlock->rdnIssuer, "\n" );

	if ( typeInfoOnly && ( m_editState != esEditionInited  ) )
		return 0;

	return CreateBrowseButton( r );
}

int cWizardDlg_PubKeyCrt::OnEditSubjectRDN( RECT *r, cWCStr *wcs, bool typeInfoOnly )
{
	rdn2ptb( wcs, &m_pBlock->rdnSubject, "\n" );

	if ( typeInfoOnly && ( m_editState != esEditionInited  ) )
		return 0;

	return CreateBrowseButton( r );
}

//
//	OnEditItem
//	1. Creates edit window above subitem with same size
//		window type may be
//			a. edit box
//			b. spin
//			c. fixed combo-box
//			d. browse button in left side of item rectangle
//	2. Set dialogue to edit item mode
//		- save edit item number and edit window type and handle
//		- and tries to catch mouse and key board activity
//	3. Close edit window and save updated value.
//
bool cWizardDlg_PubKeyCrt::OnEditItem( int itemNum, bool typeInfoOnly, bool dblClk )
{
	bool useFixedFontForInfo = false;
	m_editNum = itemNum;

	if ( dblClk && !m_pBlock->m_view )
	{
		if ( m_editNum == lvId_issuerRdn )
			return OnBrowseIssuer();
		else
		if ( m_editNum == lvId_subjectRdn )
			return OnGetSubjectRDN();
	}

	if ( m_pBlock->m_view )
		typeInfoOnly = true;

	RECT r;

	//cPlainTextBlock ptb;
	cWCStr wcs;

	if ( ListView_GetSubItemRect( m_hLV, m_editNum, 1, LVIR_BOUNDS, &r ) == 0 )
	{
		assert( false );
		return false;
	}

	if ( m_editNum == lvId_version )
	{
		wcs.Add( _LoadIDS( IDS_STRING140 ) );
	}
	else
	if ( m_editNum == lvId_signatureAlgorithm )
	{
		wcs.Add( _LoadIDS( IDS_STRING141 ) );

		const char *oidStr = m_pBlock->getSignatureAlgOidStr();

		if ( oidStr != NULL )
		{
			wcs.Add( getWCStr( oidStr ) );

			char name[ 1024 ];
			if ( generateOIDName( name, maxLen( name ), oidStr ) != 0 )
			{
				wcs.Add( L"\n  " );
				wcs.Add( getWCStr( name ) );
			}
		}
		else
		{
			wcs.Add( _LoadIDS( IDS_Undefined ) );
		}
	}
	else
	if ( m_editNum == lvId_issuerRdn )
	{
		OnEditIssuer( &r, &wcs, typeInfoOnly );
	}
	else
	if ( m_editNum == lvId_validFrom )
	{
		if ( !typeInfoOnly || ( m_editState == esEditionInited ) )
			OnEditDate( &r );
	}
	else
	if ( m_editNum == lvId_validTo )
	{
		if ( !typeInfoOnly || ( m_editState == esEditionInited ) )
			OnEditDate( &r );
	}
	else
	if ( m_editNum == lvId_subjectRdn )
	{
		OnEditSubjectRDN( &r, &wcs, typeInfoOnly );
	}
	else
	if ( m_editNum == lvId_publicKey )
	{
		if ( ( !m_pBlock->m_view ) && ( m_pBlock->ska == NULL ) )
			wcs.Add( _LoadIDS( IDS_STRING142 ) );
		else
		{
			if ( m_pBlock->pka == NULL )
			{
				int res = ska_derivePublicKeyAlgorithm( &m_pBlock->pka, m_pBlock->ska );
				if ( res != pkae_ok ) return false;
			}
			else
			{
				getLocalKeyID( m_pBlock->pka, &wcs, false );
				getKeyASN1Buffer( m_pBlock->pka, &wcs, false );
				useFixedFontForInfo = true;
			}
		}
	}
	else
	if ( m_editNum == lvId_subjectKeyID )
	{
		if ( m_pBlock->pka != NULL )
		{
			getLocalKeyID( m_pBlock->pka, &wcs, false );
			getKeyASN1Buffer( m_pBlock->pka, &wcs, false );
			useFixedFontForInfo = true;
		}
	}
	else
	if ( m_editNum == lvId_authrityKeyID )
	{
		if ( m_pBlock->m_authorityInfo.bInited )
		{
			m_pBlock->m_authorityInfo.writeToPtb( &wcs, true );
		}
		else
		if ( m_pBlock->skaIssuer != NULL )
		{
			getLocalKeyID( m_pBlock->skaIssuer, &wcs, true );
			getKeyASN1Buffer( m_pBlock->skaIssuer, &wcs, true );
			useFixedFontForInfo = true;
		}
		else
		if ( m_editState != esEditionInited )
			wcs.Add( _LoadIDS( IDS_szSelfSigned ) );
		else
			wcs.Add( _LoadIDS( IDS_CertificateIssuerIsUndefined ) );
	}

	FillInfo( &wcs, useFixedFontForInfo );

	return false;
}

bool cWizardDlg_PubKeyCrt::StopEdit( bool checkPos, bool cancelled )
{
	if ( m_popup == NULL )
		return false;

	if ( cancelled )
	{
		if ( m_popup1 != NULL  )
		{
			DestroyWindow( m_popup1 );
			m_popup1 = NULL;
		}

		if ( m_popup != NULL )
		{
			DestroyWindow( m_popup ); 
			m_popup = NULL;
		}

		return true;
	}

	if ( checkPos )
	{
		HWND hFocus = GetFocus();

		if ( hFocus == m_popup )
			return false;

		if ( hFocus == m_popup1 )
			return false;
	}

	if ( m_editNum == lvId_issuerRdn )
	{
		DestroyWindow( m_popup ); m_popup = NULL;
	}
	else
	if ( m_editNum == lvId_subjectRdn )
	{
		DestroyWindow( m_popup ); m_popup = NULL;
	}
	else
	if( m_editNum == lvId_validFrom )
	{
		DateTime_GetSystemtime( m_popup, &m_pBlock->nb );
		UpdateDateItem( lvId_validFrom, &m_pBlock->nb );
		DestroyWindow( m_popup ); m_popup = NULL;
	}
	else
	if( m_editNum == lvId_validTo )
	{
		DateTime_GetSystemtime( m_popup, &m_pBlock->na );
		UpdateDateItem( lvId_validTo, &m_pBlock->na );
		DestroyWindow( m_popup ); m_popup = NULL;
	}
	else
		return false;

	m_pBlock->wasGenerated = false;
	UpdateSTDButtons();
	return true;
}

bool cWizardDlg_PubKeyCrt::UpdateSTDButtons()
{
	if ( m_pBlock->m_view )
	{
		HWND hWnd = GetDlgItem( m_hDlg, IDOK );

		ShowWindow( hWnd, true );
		EnableWindow( hWnd, true );

		if (  m_editState == esAllowEdition ) 
		{
			SetWindowTextT( hWnd, _LoadIDS( IDS_szEdit ) );
		}
		else
		if ( m_editState == esEditionInited ) 
		{
			SetWindowTextT( hWnd, _LoadIDS( IDS_szUpdate ) );
			if ( m_pBlock->skaIssuer == NULL )
				EnableWindow( hWnd, false );
		}
		else
		{
			ShowWindow( GetDlgItem( m_hDlg, IDOK ), false );
		}

		ShowWindow( GetDlgItem( m_hDlg, IDC_btNext ), false );
		ShowWindow( GetDlgItem( m_hDlg, IDC_btPrev ), false );
		EnableWindow( GetDlgItem( m_hDlg, IDCANCEL ), true );

		return true;
	}

	bool enableSKGeneration = m_pBlock->isSKReady();
	bool skLoaded = m_pBlock->ska != NULL;
	bool enablePKGeneration = m_pBlock->isPKReady();

	EnableWindow( GetDlgItem( m_hDlg, IDC_btNext ), false );
	if ( m_pBlock->wasGenerated )
		EnableWindow( GetDlgItem( m_hDlg, IDOK ), false );
	else
		EnableWindow( GetDlgItem( m_hDlg, IDOK ), enablePKGeneration && ( enableSKGeneration || skLoaded ) );
	return true;
}

//////////////////////////////////////////////
//
//	RDN dialogue
//

static struct 
{
	sNameAttribute::at_type atNum;
	int dlgID;
} idsTbl[] = 
{
	{ sNameAttribute::id_at_countryName,			IDC_edtCountry },
	{ sNameAttribute::id_at_stateOrProvinceName,	IDC_edtState },
	{ sNameAttribute::id_at_localityName,			IDC_edtLocality },
	{ sNameAttribute::id_at_commonName,				IDC_edtCommonName },
	{ sNameAttribute::pkcs9_emailAddress,			IDC_edtEmailAddress },
	{ sNameAttribute::id_at_organizationName,		IDC_edtOrganization },
	{ sNameAttribute::id_at_organizationalUnitName,	IDC_edtOrganizationUnit }
//	IDC_edtOthers
};

int findDlgID( int atNum )
{
	int i = sizeof ( idsTbl ) / sizeof( *idsTbl );

	while ( i-- )
	{
		if ( idsTbl[i].atNum == atNum )
			return idsTbl[i].dlgID;
	}

	return -1;
}

int findAtNum( int dlgID )
{
	int i = sizeof ( idsTbl ) / sizeof( *idsTbl );

	while ( i-- )
	{
		if ( idsTbl[i].dlgID == dlgID )
			return idsTbl[i].atNum;
	}

	return -1;
}

class cDlgRDN : public cColoredDlg
{
public:
	enum { IDD = IDD_DlgRDN };

	cDlgRDN( sNameAttributeList *rdn, const wchar_t *caption = NULL, cDlgRect *rect = NULL )
	{
		m_dlgRect = rect;
		m_caption = caption;
		m_rdn = rdn;
		SetRCValues( IDC_Line, IDC_STATIC_BOLD, 0, IDI_Main );
	}

private:
	LRESULT RealDlgProc( UINT uMsg, WPARAM wParam, LPARAM lParam )
	{
		LRESULT result;
		
		if ( cMenuItem::DlgProc( &result, m_hDlg, uMsg, wParam, lParam ) )
			return result;

		switch(uMsg)
		{
			case WM_INITDIALOG:
				InitDialog();
				break;

			case WM_COMMAND:
				if ( HIWORD( wParam ) == EN_CHANGE )
				{
					UpdateResultWindow();
				}
				else
				if ( HIWORD( wParam ) == 0 )
				{
					switch ( LOWORD( wParam ) )
					{
						case IDOK:
							if ( updateRDN() != 0 )
							{
								MessageBoxT( m_hDlg, _LoadIDS( IDS_NameAttrIsInvalied ), L"RDN", MB_ICONSTOP );
								return false;
							}
						case IDCANCEL:
							EndDialog( m_hDlg, LOWORD( wParam ) );
							return true;
						default:
							break;
					}
				}
				break;
			default:
				return false;
		}

		return true;
	}

	sNameAttributeList *m_rdn;
	const wchar_t *m_caption;

	void InitDialog()
	{
		if ( m_caption )
			SetDlgItemTextT( IDC_STATIC_BOLD, m_caption );

		char othersStr[ 1024 ] = "";

		for( sNameAttribute *na = m_rdn->head; na; na = na->next )
		{
//			int dlgID = findDlgID( na->getIdByName( na->szValue ) );
			int dlgID = findDlgID( na->getIdByOID( na->szOID ) );
			if ( dlgID != -1 )
			{
				SetDlgItemTextA( m_hDlg, dlgID, na->szValue );
			}
			else
			{
				char str[ 256 ];
				sprintf_s( str, "%s = %s", na->getName(), na->szValue );
				if ( strlen( othersStr ) != 0 )
					strcat_s( othersStr, "; " );
				strcat_s( othersStr, str );
			}
		}

		SetDlgItemTextA( m_hDlg, IDC_edtOthers, othersStr );

		UpdateResultWindow();
	}

	int updateRDN()
	{
		m_rdn->freeList();

		int i = sizeof ( idsTbl ) / sizeof( *idsTbl );

		while ( i-- )
		{
			char str[ 1024 ];
			if ( GetDlgItemTextA( m_hDlg, idsTbl[i].dlgID, str, sizeof( str ) ) != 0 )
			{
				m_rdn->Add( idsTbl[i].atNum, str );
			}
		}

		char othersStr[ 1024 ];
		if ( GetDlgItemTextA( m_hDlg, IDC_edtOthers, othersStr, sizeof( othersStr ) ) != 0 )
			return m_rdn->ParseString( othersStr, ';' );

		return 0;
	}

	void UpdateResultWindow()
	{
		int i = sizeof ( idsTbl ) / sizeof( *idsTbl );

		sNameAttributeList rdn;

		while ( i-- )
		{
			char str[ 1024 ];
			if ( GetDlgItemTextA( m_hDlg, idsTbl[i].dlgID, str, sizeof( str ) ) != 0 )
			{
				rdn.Add( idsTbl[i].atNum, str );
			}
		}

		char othersStr[ 1024 ];
		if ( GetDlgItemTextA( m_hDlg, IDC_edtOthers, othersStr, sizeof( othersStr ) ) != 0 )
			rdn.ParseString( othersStr, ';' );

		//cPlainTextBlock ptbRDN;
		cWCStr wcs;
		if ( !rdn2ptb( &wcs, &rdn ) )
			return;

		SetDlgItemTextT( IDC_edtResult, wcs.m_szBody ); //( char *)ptbRDN.buffer );

		EnableWindow( GetDlgItem( m_hDlg, IDOK ), !wcs.isEmpty() ); //ptbRDN.bufferSize != 0 );
	}
};

bool GetRDN( HINSTANCE hInst, HWND hWnd, sNameAttributeList *rdn, const wchar_t *caption, cDlgRect *rect )
{
	cDlgRDN dlg( rdn, caption, rect );
	return dlg.DoModal( hInst, dlg.IDD, hWnd ) == IDOK;
}

bool rdn2ptb( cWCStr *wcs, sNameAttributeList *rdn, char *separator )
{
	int size = 0;

	if ( rdn->getAttributesString( NULL, &size, separator ) != x509e_bufferTooSmall )
		return false;

	cPlainTextBlock ptb;
	if ( ptb.Allocate( size ) == NULL )
		return false;

	if ( rdn->getAttributesString( ( char *)ptb.buffer, &size, separator ) != x509e_ok )
		return false;

	wcs->Add( (char*)ptb.buffer );

	return true;
}
