#include "cWizardDlgs.h"
#include "BCKeyMan.h"

#include "..\Includes\stddlgs.h"
#include "..\CertificateLib\PGPPkts\Packets.h"
#include "..\CertificateLib\x509pkts\pkcs12_funcs.h"

#include "..\Includes\FileView.h"
#include "..\Includes\UnicodeFuncs.h"
#include "..\Includes\OpenFiles.h"
#include "..\Includes\AutoDeallocate.h"

#include <windowsx.h>
#include <windows.h>
#include <io.h>
#include <stdio.h>

static UINT ColumnLabelIDS[] = 
{
	IDS_szField, IDS_szValue
};

const int NUM_COLUMNS = sizeof( ColumnLabelIDS ) / sizeof( UINT );

static int _gnColumnFmt[ NUM_COLUMNS ] =
{
	LVCFMT_LEFT, LVCFMT_LEFT
};

static int _gnColumnWidth[NUM_COLUMNS]=
{
	120, 288
};

//
//	Return number of filled bytes
//	if  buffer is too small returns -1 and sets *pStrSize as required size.
//
int bytes2string( cWCStr *wcs, BYTE *bf, int bfSize, int bytesInLine )
{
	int size = 0;
	bytes2string( (wchar_t*)NULL, &size, bf, bfSize, bytesInLine );

	if ( size == 0 )
	{
		assert( false );
		return -1;
	}

	int pos = 0;
	if ( !wcs->isEmpty() )
		pos = wcslen( wcs->m_szBody );

	// Exclude last zero, it will automatically added for cWCStr.
	if ( wcs->Add( size - 1 ) == 0 )
	{
		assert( false );
		return -1;
	}

	bytes2string( wcs->m_szBody + pos, &size, bf, bfSize, bytesInLine );

	return size;
}

int bytes2string( wchar_t *str, int *pStrLen, BYTE *bf, int bfSize, int bytesInLine )
{
	if ( str == NULL )
		return bytes2string( (char*)NULL, pStrLen, bf, bfSize, bytesInLine );

	cAutoDeallocate < char > mbs;

	if ( mbs.Allocate( *pStrLen ) == NULL )
		return -1;

	int res = bytes2string( mbs.m_ptr, pStrLen, bf, bfSize, bytesInLine );

	if ( res == -1 )
		return -1;

	*pStrLen = _mbstowcs( str, *pStrLen, mbs.m_ptr );

	return res;
}

int bytes2string( char *str, int *pStrSize, BYTE *bf, int bfSize, int bytesInLine )
{
	int strSize = *pStrSize;

	int len = 0;
	int pos = 0;

	for( ; pos < bfSize; pos++ )
	{
		if ( len + 3 >= strSize )
			break;

		sprintf_s( str + len, strSize, "%02X ", (DWORD)bf[pos] );
		len += 3;

		if ( pos && bytesInLine && !( pos % bytesInLine ) )
		{
			const char *newLine = "\n";
			int newLineLen = strlen( newLine );

			if ( len + newLineLen >= strSize )
				break;
			strcat_s( str + len, strSize, newLine );
			len += newLineLen;
		}
	}

	if ( len < strSize )
		str[ len ] = '\0';

	int required = bfSize * 3 + 1;

	if ( bytesInLine != 0 )
	{
		required += pos / bytesInLine;
	}

	if ( *pStrSize < required )
	{
		*pStrSize = required;
		return -1;
	}

	return pos;
}

void cWizardDlg_SecKeyPkt::UpdateFriendlyName()
{
	int itemNum = 2;

	if ( m_pBlock->friendlyName[0] == 0 )
	{
		ListView_SetItemTextT( m_hLV, itemNum, 1, _LoadIDS( IDS_szIsEmpty ) );
		ListView_SetItemParam( m_hLV, itemNum, 1 );
	}
	else
	{
		ListView_SetItemTextT( m_hLV, itemNum, 1, m_pBlock->friendlyName );
		ListView_SetItemParam( m_hLV, itemNum, 0 );
	}
}

bool cWizardDlg_SecKeyPkt::SetItemsValues() // HINSTANCE hInst, HWND hLView, cWizardBlock *pBlock  )
{
	int iFocused = ListView_GetNextItem( m_hLV, -1, LVNI_FOCUSED );
	if ( iFocused == -1 )
		iFocused = 0;

	ListView_DeleteAllItems( m_hLV );

	int itemNum = 0;
	int subItem = 0;
	int showNum = m_pBlock->skShowNum;

	wchar_t buff[ MAX_PATH ];

	LVITEMW lvi;
	memset( &lvi, 0, sizeof( lvi ) );
	lvi.mask = LVIF_TEXT;

	// 1. Algorithm
	itemNum = 0;
	lvi.iItem = itemNum;

	cWCStr wcs;
	wcs.LoadStringT( IH.m_hInst, IDS_szAlgorithm );
	lvi.pszText = wcs.m_szBody;

	if ( ListView_InsertItemT( m_hLV, &lvi ) != itemNum ) { assert( false ); return false; }

	if ( m_pBlock->m_view )
	{
		if ( m_pBlock->pka )
		{	int algID = pka_getAlgID( m_pBlock->pka );
			ListView_SetItemTextT( m_hLV, itemNum, 1, getWCStr( (char *)PKT_GetPublicKeyAlgName( (BYTE)algID ) ) );
		}
		else
			ListView_SetItemTextT( m_hLV, itemNum, 1, _LoadIDS( IDS_Undefined ) );
	}
	else
	if ( m_pBlock->ska != NULL )
	{
		int algID = ska_getAlgID( m_pBlock->ska );
		ListView_SetItemTextT( m_hLV, itemNum, 1, getWCStr( (char *)PKT_GetPublicKeyAlgName( (BYTE)algID ) ) );
	}
	else
		ListView_SetItemTextT( m_hLV, itemNum, 1, m_pBlock->getAlgName( m_pBlock->algPos ) );

	// 2. Key size - may be changed
	itemNum = 1;
	lvi.iItem = itemNum;
	wcs.LoadStringT( IH.m_hInst, IDS_szKeySize );
	lvi.pszText = wcs.m_szBody;
	if ( ListView_InsertItemT( m_hLV, &lvi ) != itemNum ) { assert( false ); return false; }

	if ( m_pBlock->m_view )
	{
		if ( m_pBlock->pka )
		{
			m_pBlock->keyBitSize = pka_getModuleSize( m_pBlock->pka ) * 8;
			FormatT( buff, maxLen( buff ), _LoadIDS( IDS_szKeySizePattern ), m_pBlock->keyBitSize );
			ListView_SetItemTextT( m_hLV, itemNum, 1, buff );
		}
		else
			ListView_SetItemTextT( m_hLV, itemNum, 1, _LoadIDS( IDS_Undefined ) );
	}
	else
	{
		FormatT( buff, maxLen( buff ), _LoadIDS( IDS_szKeySizePattern ), m_pBlock->keyBitSize );
		ListView_SetItemTextT( m_hLV, itemNum, 1, buff );
	}

	// 3. friendly name
	itemNum = 2;
	lvi.iItem = itemNum;
	wcs.LoadStringT( IH.m_hInst, IDS_szFriendlyName );
	lvi.pszText = wcs.m_szBody;
	if ( ListView_InsertItemT( m_hLV, &lvi ) != itemNum ) { assert( false ); return false; }

	UpdateFriendlyName();

	if ( m_pBlock->m_view )
		return true;

	// 4. password
	itemNum = 3;
	lvi.iItem = itemNum;

	wcs.LoadStringT( IH.m_hInst, IDS_szPassword );
	lvi.pszText = wcs.m_szBody;

	if ( ListView_InsertItemT( m_hLV, &lvi ) != itemNum ) { assert( false ); return false; }

	FillLVPassword( &(m_pBlock->pwd), itemNum );

	// 5. password ( confirmation )
	itemNum = 4;
	lvi.iItem = itemNum;

	wcs.LoadStringT( IH.m_hInst, IDS_szConfirmPassword );
	lvi.pszText = wcs.m_szBody;

	if ( ListView_InsertItemT( m_hLV, &lvi ) != itemNum ) { assert( false ); return false; }

	FillLVPassword( &(m_pBlock->pwdConfirm), itemNum );

	if ( showNum == 0 ) // show all
	{
		//	6. PBE algorithm f.e. { pkcs-12-pbe-pbeWithSHAAnd3-KeyTripleDES-CBC }
		itemNum = 5;
		lvi.iItem = itemNum;
		wcs.LoadStringT( IH.m_hInst, IDS_szPBEAlg );
		lvi.pszText = wcs.m_szBody;
		if ( ListView_InsertItemT( m_hLV, &lvi ) != itemNum ) { assert( false ); return false; }
		ListView_SetItemTextT( m_hLV, itemNum, 1, _LoadIDS( IDS_PBEwSHA1a3KeyTripleDESCBC ) );
		//	7. Salt
		itemNum = 6;
		lvi.iItem = itemNum;
		wcs.LoadStringT( IH.m_hInst, IDS_szSalt );
		lvi.pszText = wcs.m_szBody;
		if ( ListView_InsertItemT( m_hLV, &lvi ) != itemNum ) { assert( false ); return false; }
		int size = maxLen( buff );
		bytes2string( buff, &size, m_pBlock->salt, sizeof( m_pBlock->salt ) );
		ListView_SetItemTextT( m_hLV, itemNum, 1, buff );

		//	8. Iterations
		itemNum = 7;
		lvi.iItem = itemNum;
		wcs.LoadStringT( IH.m_hInst, IDS_szIterations );
		lvi.pszText = wcs.m_szBody;
		if ( ListView_InsertItemT( m_hLV, &lvi ) != itemNum ) { assert( false ); return false; }
		::FormatT( buff, maxLen( buff ), L"%d", m_pBlock->iterCount );
		ListView_SetItemTextT( m_hLV, itemNum, 1, buff );
	}

	if ( iFocused > itemNum )
		iFocused = itemNum;

	ListView_SetItemState( m_hLV, iFocused, LVIS_FOCUSED, LVIS_FOCUSED );

	return true;
}
     	  
static BOOL InitListViewColumns( HINSTANCE hInst, HWND hLView )
{
	int i;
	LV_COLUMNW lvc;

	memset( &lvc, 0, sizeof( lvc ) );

	lvc.mask = LVCF_WIDTH;

	//
	//	Check if already initialised...
	//
	if ( ListView_GetColumnT( hLView, 0, &lvc ) )
		return TRUE;

	lvc.mask = LVCF_FMT | LVCF_WIDTH | LVCF_TEXT | LVCF_SUBITEM;

	wchar_t text[ 80 ];

	int res;

	for(i=0; i<NUM_COLUMNS; i++)
	{
		LoadStringT( hInst, ColumnLabelIDS[i], text, maxLen( text ) );

		lvc.iSubItem=i;
		lvc.pszText = text; // _gszColumnLabel[i];
		lvc.cx = _gnColumnWidth[i];
		lvc.fmt=_gnColumnFmt[i];

		res = ListView_InsertColumnT( hLView, i, &lvc );

		if ( res == -1 )
		{
			MessageBoxT( 0, text, _LoadIDS( IDS_ErrorInAddLVColumn ), MB_OK );
			return FALSE;
		}
	}

	return TRUE;
}

static bool useCmbEx = false; // true;

LONG cWizardDlg_SecKeyPkt::lpNativeLVProc = NULL;

DWORD keySizeMin = 1024; // 512;
DWORD keySizeMax = 16 * 1024;
DWORD keySizeStep = keySizeMin;

BOOL CALLBACK cWizardDlg_SecKeyPkt::LVWndProc( HWND hLV, UINT uMsg, WPARAM wParam, LPARAM lParam )
{
	if ( lpNativeLVProc == NULL )
	{
		assert( false );
		return FALSE;
	}

	if ( uMsg == WM_KEYDOWN )
	{
	}
	else
	if ( uMsg == WM_NOTIFY )
	{
		NMHDR *hdr = (NMHDR *)lParam;
		if ( hdr->code == UDN_DELTAPOS )
		{
			HWND hDlg = GetParent( hLV );
			cWizardDlg_SecKeyPkt *pThis = (cWizardDlg_SecKeyPkt *)GetWindowLong( hDlg, DWL_USER);

			if ( hdr->hwndFrom == pThis->m_popup1 )
				return cDlg::UDNDeltapos( (NMUPDOWN *)hdr, pThis->m_popup, keySizeMin, keySizeMax, keySizeStep );
		}
	}
	else
	if ( uMsg == WM_COMMAND )
	{
/*
		if ( HIWORD( wParam ) == EN_CHANGE )
		{
			char msg[10];
			sprintf( msg, "qqq" );
		}
*/
	}

	return CallWindowProc( (WNDPROC)lpNativeLVProc, hLV, uMsg, wParam, lParam );
}

const int aNumber = 25; 

bool cWizardDlg_SecKeyPkt::OnInitDialog()
{
	//
	// (WM_USER + 1) allows to catch WM_COMMAND with the number if VK_RETURN pressed
	//
	SendMessage( m_hDlg, WM_USER + 1, aNumber, 0 );

	if ( useCmbEx )
	{
		INITCOMMONCONTROLSEX InitCtrls;
		InitCtrls.dwSize = sizeof( INITCOMMONCONTROLSEX );
		InitCtrls.dwICC = ICC_USEREX_CLASSES;

		BOOL res = InitCommonControlsEx( &InitCtrls );
	}

	//
	//	Set dialogue caption 
	//
	if ( m_pBlock->m_view )
	{
		SetDlgItemTextT( IDC_STATIC_BOLD, _LoadIDS( IDS_P12SecretKey ) );
		SetDlgItemTextT( IDC_Notes, L"" );
	}
	else
		SetDlgItemTextT( IDC_STATIC_BOLD, _LoadIDS( IDS_CreateSecKeyPkt ) );

	ComboBox_AddStringT( GetDlgItem( m_hDlg, IDC_cmdShowDetails ), _LoadIDS( IDS_ShowAll ) );
	if ( m_pBlock->m_view )
	{
		ComboBox_SetCurSel( GetDlgItem( m_hDlg, IDC_cmdShowDetails ), 0 );
	}
	else
	{
		ComboBox_AddStringT( GetDlgItem( m_hDlg, IDC_cmdShowDetails ), _LoadIDS( IDS_ShowHastToBeFilled ) );
		ComboBox_SetCurSel( GetDlgItem( m_hDlg, IDC_cmdShowDetails ), m_pBlock->skShowNum );
	}

	HWND hLB = GetDlgItem( m_hDlg, IDC_lbFieldDetails );
	m_fontOriginal = GetWindowFont( hLB );

	m_hLV = GetDlgItem( m_hDlg, IDC_lvFields );

	ListView_SetExtendedListViewStyleEx( m_hLV, LVS_EX_GRIDLINES | LVS_EX_FULLROWSELECT, LVS_EX_GRIDLINES | LVS_EX_FULLROWSELECT );
	m_popup = NULL;

	InitListViewColumns( m_hInst, m_hLV );

	if ( m_pBlock->ska != NULL )
	{
		m_pBlock->keyBitSize = ska_getModuleSize( m_pBlock->ska ) * 8;
	}

	if ( wcscmp( m_pBlock->friendlyName, _LoadIDS( IDS_szIsEmpty ) ) == 0 )
		m_pBlock->friendlyName[0] = 0;

	SetItemsValues();// m_hInst, m_hLV, m_pBlock );

	//
	// Get address of List View WinProc and change it to internal LVWndProc
	//	It is required for interception of UPDOWN control notification messages.
	//
	lpNativeLVProc = GetWindowLong( m_hLV, GWL_WNDPROC );
	SetWindowLong( m_hLV, GWL_WNDPROC, (LONG) LVWndProc );

	int err = 0;

	if ( !UpdateSTDButtons() )
	{
		if ( SetFocus( m_hLV ) == NULL )
		{
			err = GetLastError();
		}
	}

	return true;
}

bool cWizardDlg_SecKeyPkt::FillLVPassword( cPassword *pwd, int itemNum ) // = -1 )
{
	wchar_t buff[ 256 ];

	if ( itemNum == -1 )
		itemNum = m_editNum;

	DWORD lParam = 0;

	if ( ( pwd->wcBufferSize == 0 ) || ( pwd->wcBuffer[0] == 0 ) )
	{
		LoadStringT( IH.m_hInst, IDS_szIsEmpty, buff, maxLen( buff ) );
		lParam = 1;
	}
	else
	{
		int len = wcslen( pwd->wcBuffer );
		buff[len] = L'\0';
		while( len-- )
			buff[ len ] = L'*'; // '\7';
	}

	ListView_SetItemTextT( m_hLV, itemNum, 1, buff );
	ListView_SetItemParam( m_hLV, itemNum, lParam );

	return true;
}
bool cWizardDlg_SecKeyPkt::CheckRequiredFields()
{
	bool silent = true;

	if ( m_pBlock->friendlyName[0] == 0 )
	{
		if ( silent )
			MessageBoxT( m_hDlg, _LoadIDS( IDS_FriendlyNameIsEmpty ), _LoadIDS( IDS_AppName ), MB_ICONSTOP );
		return false;
	}
	else
	if ( !m_pBlock->CheckPasswords( m_hDlg, 3, silent ) ) // Check both password
		return false;

	return true;
}

bool cWizardDlg_SecKeyPkt::OnCommand( WPARAM wParam, LPARAM lParam )
{

	if ( m_popup )
	{
		bool cancelled = ( LOWORD(wParam) == IDCANCEL );
		StopEdit( false, cancelled );
		if ( wParam == aNumber )
		{
			// Select next item
			int iFocused = ListView_GetNextItem( m_hLV, -1, LVNI_FOCUSED );
			ListView_SetItemState( m_hLV, iFocused, 0, LVIS_FOCUSED | LVIS_SELECTED );
			ListView_SetItemState( m_hLV, iFocused + 1, LVIS_FOCUSED | LVIS_SELECTED, LVIS_FOCUSED | LVIS_SELECTED );
		}
		return true;
	}
	else
	if ( wParam == aNumber )
	{
		int iFocused = ListView_GetNextItem( m_hLV, -1, LVNI_FOCUSED );
		bool res = OnEditItem( iFocused );
		return res;
	}


	if ( OnStdButton( wParam ) )
	{
		return true;
	}
	else
	{
		switch ( LOWORD(wParam) )
		{
			case IDC_cmdShowDetails:
				if ( HIWORD( wParam ) == CBN_SELCHANGE )
				{
					m_pBlock->skShowNum = ComboBox_GetCurSel( GetDlgItem( m_hDlg, IDC_cmdShowDetails ) );
					SetItemsValues(); // m_hInst, GetDlgItem( m_hDlg, IDC_lvFields ), m_pBlock );
				}
				break;
			default:
				return false;
		}
		return true;
	}
}

bool cWizardDlg_SecKeyPkt::StopEdit( bool checkPos, bool cancelled )
{
	if ( m_popup == NULL )
		return false;

	if ( cancelled )
	{
		if ( m_popup1 != NULL  )
		{
			DestroyWindow( m_popup1 );
			m_popup1 = NULL;
		}

		if ( m_popup != NULL )
		{
			DestroyWindow( m_popup ); 
			m_popup = NULL;
		}

		return true;
	}

	if ( checkPos )
	{
		HWND hFocus = GetFocus();

		if ( hFocus == m_popup )
			return false;

		if ( hFocus == m_popup1 )
			return false;
	}

	if ( m_editNum == 0 )
	{
		m_pBlock->algPos = ComboBox_GetCurSel( m_popup );
		ListView_SetItemTextT( m_hLV, m_editNum, 1, m_pBlock->getAlgName() );
		DestroyWindow( m_popup ); m_popup = NULL;
	}
	else
	if ( m_editNum == 1 )
	{
		wchar_t buff[ 256 ];
		GetWindowTextT( m_popup, buff, maxLen( buff ) );

		DWORD size = _wtoi( buff );

		if ( size == 0 )
		{
			assert( false );
		}
		else
		{
			if ( size > keySizeMax )
				size = keySizeMax;
			else
			if ( size < keySizeMin )
				size = keySizeMin;

			m_pBlock->keyBitSize = size;
			FormatT( buff, maxLen( buff ), _LoadIDS( IDS_szKeySizePattern ), m_pBlock->keyBitSize );
			ListView_SetItemTextT( m_hLV, m_editNum, 1, buff );
		}

		DestroyWindow( m_popup1 ); m_popup1 = NULL;
		DestroyWindow( m_popup ); m_popup = NULL;
	}
	else
	if ( m_editNum == 2 )
	{
		GetWindowTextT( m_popup, m_pBlock->friendlyName, maxLen( m_pBlock->friendlyName ) );
		UpdateFriendlyName();
		DestroyWindow( m_popup ); m_popup = NULL;
	}
	else
	if ( m_editNum == 3 )
	{
		GetWindowTextT( m_popup, m_pBlock->pwd.wcBuffer, m_pBlock->pwd.wcBufferSize / sizeof( wchar_t ) );
		DestroyWindow( m_popup ); m_popup = NULL;

		FillLVPassword( &(m_pBlock->pwd) );
		m_pBlock->CheckPasswords( m_hDlg, 1 );
	}
	else
	if ( m_editNum == 4 )
	{
		GetWindowTextT( m_popup, m_pBlock->pwdConfirm.wcBuffer, m_pBlock->pwdConfirm.wcBufferSize / sizeof( wchar_t ) );
		DestroyWindow( m_popup ); m_popup = NULL;

		FillLVPassword( &(m_pBlock->pwdConfirm) );
		m_pBlock->CheckPasswords( m_hDlg, 2 );
	}
	else
		return false;

	m_pBlock->wasGenerated = false;
	UpdateSTDButtons();
	return true;
}

bool cWizardDlg_SecKeyPkt::FillInfo( cWCStr *wcs, bool useFixedFont )
{
	HWND hLB = GetDlgItem( m_hDlg, IDC_lbFieldDetails );
	ListBox_ResetContent( hLB );

	HFONT font = useFixedFont ? m_pBlock->hFixedFont : m_fontOriginal;
	SetWindowFont( hLB, font, 1 );

//	ptb->PutByte( 0 ); // EOS
	AddStrings( hLB, wcs->m_szBody );//(char *)ptb->buffer );
	return true;
}

//
//	OnEditItem
//	1. Creates edit window above subitem with same size
//		window type may be
//			a. edit box
//			b. spin
//			c. fixed combo-box
//	2. Set dialogue to edit item mode
//		- save edit item number and edit window type and handle
//		- and tries to catch mouse and key board activity
//	3. Close edit window and save updated value.
//
bool cWizardDlg_SecKeyPkt::OnEditItem( int itemNum, bool typeInfoOnly )
{
	m_editNum = itemNum;
	bool useFixedFontForInfo = false;

	if ( m_pBlock->m_view )
		typeInfoOnly = true;

	RECT r;

	//cPlainTextBlock ptb;
	cWCStr wcs;

	if ( ListView_GetSubItemRect( m_hLV, m_editNum, 1, LVIR_BOUNDS, &r ) == 0 )
	{
		assert( false );
		return false;
	}

	// 1. Algorithm - may be changed
	if ( m_editNum == 0 )
	{
		OnEditAlgorithm( &r, &wcs, typeInfoOnly );
		useFixedFontForInfo = true;
	}
	else
	// 2. Key size - may be changed
	if ( m_editNum == 1 )
	{
		OnEditKeySize( &r, &wcs, typeInfoOnly );
		useFixedFontForInfo = true;
	}
	else
	// 3. friendly name
	if ( m_editNum == 2 )
	{
		OnEditFriendlyName( &r, &wcs, typeInfoOnly );
	}
	else
	// 4. password
	if ( m_editNum == 3 )
	{
		if ( !typeInfoOnly )
			OnEditPassword( &r, &m_pBlock->pwd );
	}
	else
	// 5. password ( confirmation )
	if ( m_editNum == 4 )
	{
		if ( !typeInfoOnly )
			OnEditPassword( &r, &m_pBlock->pwdConfirm );
	}
	// if showAll 
	//	6. PBE algorithm f.e. { pkcs-12-pbe-pbeWithSHAAnd3-KeyTripleDES-CBC }
	//	7. Salt
	//	8. Iterations
	// predefined
	// 9. Local key ID
	// 10. ASN1 buffer with key info

	FillInfo( &wcs, useFixedFontForInfo );

	return false;
}

bool cWizardDlg_SecKeyPkt::OnNotify( NMHDR *hdr )
{
	switch( hdr->code )
	{
		case UDN_DELTAPOS:
			break;
		case NM_KILLFOCUS :
			StopEdit( true );
			break;
		case NM_CLICK :
		case NM_DBLCLK:
		{
			LVHITTESTINFO hi;
			memset( &hi, 0, sizeof( hi ) );
			hi.flags = LVHT_ONITEM;

			GetCursorPos( &hi.pt );
			ScreenToClient( m_hLV, &hi.pt );
			hi.pt.x = 10;

			int res = ListView_HitTest( m_hLV, &hi );

			//
			//	1. Do not stop edit if current item was clicked
			//	2. Stop current edit item before edit new.
			//
			if ( ( m_popup != NULL ) && ( res != -1 ) )
			{
				if ( hi.iItem != m_editNum )
				{
					StopEdit();
					OnEditItem( hi.iItem, true ); // type info only == true
				}

				break;
			}

			StopEdit();

			if ( hdr->idFrom == IDC_lvFields )
			{
				if ( res != -1 )
				{
					OnEditItem( hi.iItem );
				}
			}
		
			break;
		}
		case NM_CUSTOMDRAW:
		{
			int res = cListViewCustomDrawHandler::OnCustomDraw( hdr );
			SetWindowLong( m_hDlg, DWL_MSGRESULT, res ); 
			return TRUE;
		}

		default : break;
	}



	return FALSE;
}

//
//	Add string to combobox with prefixed spaces
//
static int myComboBox_AddString( HWND hCmb, const wchar_t *str, int spaceCount )
{
	wchar_t buff[ 1024 ];
	wchar_t *ptr = buff;

	int len = wcslen( str );

	if ( len > maxLen( buff ) )
		ptr = (wchar_t *)str;
	else
	{
		if ( len + spaceCount < maxLen( buff ) )
		{
			for( int i = spaceCount; i--; )
			{
				*(ptr++) = L' ';
				*ptr = L'\0';
			}
		}

		wcscat_s( buff, str );
		ptr = buff;
	}

	return ComboBox_AddStringT( hCmb, ptr );
}

//
//	getLocalKeyID
//		Generates unique value corresponded to the public key ( itself or derived from secret key )
//		It is used as certificate 'Subject Key Identifier'.
//		See RFC 3280 Internet X.509 Public Key Infrastructure p. 4.2.1.2  Subject Key Identifier
//
bool getLocalKeyID( HPKA ska, cWCStr *wcsLocalKeyID, bool secret, bool withText )
{
	int size = 0;

	if ( ska == NULL )
	{
		wcsLocalKeyID->Add( _LoadIDS( IDS_szNotReady  ) );
		return true;
	}

	if ( withText )
	{
		wcsLocalKeyID->Add( _LoadIDS( IDS_szLocalKeyIDn  ) );
	}

	int res = secret ?
			ska_generateLocalKeyID( ska, NULL, &size ):
			pka_generateLocalKeyID( ska, NULL, &size );

	if ( res == x509e_bufferTooSmall )
	{
		cPlainTextBlock ptb;
		if ( ptb.Allocate( size ) == NULL )
		{
			assert( false );
			return false;
		}

		res = secret ?
				ska_generateLocalKeyID( ska, ptb.buffer, &size ):
				pka_generateLocalKeyID( ska, ptb.buffer, &size );

		if (( res != x509e_ok ) || ( size != ptb.bufferSize ) )
		{
			assert( false );
			return false;
		}

		bytes2string( wcsLocalKeyID, ptb.buffer, ptb.bufferSize );
	}

	if ( withText )
	{
		wcsLocalKeyID->Add( L"\n" );
	}

	return true;
}

bool getKeyASN1Buffer( HPKA ska, cWCStr *wcsKey, bool secret )
{
	wcsKey->Add(  _LoadIDS( secret ? IDS_ASN1SKBuffer : IDS_ASN1PKBuffer ) );

	int size = 0;
	int res = secret ? 
			ska_writeKeyToASN1Buffer( ska, NULL, &size ):
			pka_writeKeyToASN1Buffer( ska, NULL, &size );

	if ( res != pkae_bufferSizeTooSmall )
		return false;

	cPlainTextBlock ptb;
	if ( ptb.Allocate( size ) == NULL )
	{
		assert( false );
		return false;
	}

	res = secret ? 
			ska_writeKeyToASN1Buffer( ska, ptb.buffer, &size ):
			pka_writeKeyToASN1Buffer( ska, ptb.buffer, &size );

	if ( res != pkae_ok )
	{
		assert( false );
		return false;
	}
	
	bytes2string( wcsKey, ptb.buffer, ptb.bufferSize );

	return true;
}

int cWizardDlg_SecKeyPkt::OnEditAlgorithm( RECT *r, cWCStr *wcs, bool typeInfoOnly )
{
	//
	//	Disable edition of existing key algorithm parameters
	//
	if ( m_pBlock->ska != NULL )
	{
		// Generate local key ID and key ASN.1 byte buffers
		getLocalKeyID( m_pBlock->ska, wcs );
		getKeyASN1Buffer( m_pBlock->ska, wcs );
		return 0;
	}
	else
	if ( m_pBlock->pka != NULL )
	{
		getLocalKeyID( m_pBlock->pka, wcs, false );
	}

	if ( typeInfoOnly )
		return 0;

	int height = r->bottom - r->top - ( useCmbEx ? 2 : 1 );

	POINT p;
	p.x = r->left;
	p.y = r->top;

//
//	:( Ident value works only for item, but not for subitems
//	But text is really moved at least one space
//
//	LVITEM lvItem;
//	memset( &lvItem, 0, sizeof( lvItem ) );
//	lvItem.mask = LVIF_INDENT;
//	lvItem.iItem = 0;
//	lvItem.iSubItem = 1;
//
//	ListView_GetItem( m_hLV, &lvItem );
//
//	int spaceCount = lvItem.iIndent;
//
	int spaceCount = 1;

	m_popup = CreateWindowExT( 
		0, //useCmbEx ? CBES_EX_NOEDITIMAGEINDENT | WS_EX_STATICEDGE : 0,
		useCmbEx ? WC_COMBOBOXEXW : WC_COMBOBOXW, //L"COMBOBOX",   // predefined class 
		L"",       // window text 
		WS_VISIBLE | WS_CHILD |
		CBS_AUTOHSCROLL | CBS_DROPDOWNLIST,// | CBS_NOINTEGRALHEIGHT | CBS_OEMCONVERT,  // styles 
		p.x + (useCmbEx ? 0 : -1 ),	// starting x position 
		p.y + (useCmbEx ? -3 : -4),	// starting y position 
		r->right - r->left  + (useCmbEx ? 1: 3),				// width 
		r->bottom - r->top + 100,			// height 
		m_hLV,							// parent window 
		NULL,							// No menu 
		m_hInst,						// (HINSTANCE) GetWindowLong(hwnd, GWL_HINSTANCE), 
		NULL);

	HFONT hFont = GetWindowFont( m_hLV );
	if ( hFont )
		SetWindowFont( m_popup, hFont, false );

	SetFocus( m_popup );

	if ( useCmbEx )
	{
		//COMBOBOXEXITEMW cmbe;
		//memset( &cmbe, 0, sizeof( cmbe ) );

		//cmbe.mask = CBEIF_TEXT; // | CBEIF_INDENT;
		//cmbe.iIndent = 1;
		//cmbe.pszText = (wchar_t *)m_pBlock->getAlgName( 0 );
		//cmbe.iItem = 0;
		//SendMessageT( m_popup, CBEM_INSERTITEMW, CBEM_INSERTITEMA, 0, (LPARAM)&cmbe );
		//cmbe.pszText = (wchar_t *)m_pBlock->getAlgName( 1 );
		//cmbe.iItem = 1;
		//SendMessageT( m_popup, CBEM_INSERTITEMW, CBEM_INSERTITEMA, 0, (LPARAM)&cmbe );

		//cmbe.mask = CBEIF_INDENT;
		//cmbe.iIndent = 1;
		//cmbe.iItem = -1;
		//cmbe.pszText = 0;
		//int res = SendMessageT( m_popup, CBEM_SETITEMW, CBEM_SETITEMA, 0, (LPARAM)&cmbe );

		ComboBoxEx_InsertStringT( m_popup, 0, (wchar_t *)m_pBlock->getAlgName( 0 ), 1 );
		ComboBoxEx_InsertStringT( m_popup, 0, (wchar_t *)m_pBlock->getAlgName( 1 ), 1 );

		SendMessage( m_popup, CB_SETITEMHEIGHT, -1, height );
		ComboBox_SetCurSel( m_popup, m_pBlock->algPos );
	}
	else
	{
		SendMessage( m_popup, CB_SETITEMHEIGHT, -1, height );

		myComboBox_AddString( m_popup, m_pBlock->getAlgName( 0 ), spaceCount );
		myComboBox_AddString( m_popup, m_pBlock->getAlgName( 1 ), spaceCount );
		ComboBox_SetCurSel( m_popup, m_pBlock->algPos );
	}
	return 0;
}

const int ebShift = 5;

int cWizardDlg_SecKeyPkt::OnEditKeySize( RECT *r, cWCStr *wcs, bool typeInfoOnly )
{
	//
	//	Disable edition of existing key algorithm parameters
	//
	if ( m_pBlock->ska != NULL )
	{
		// Generate local key ID and key ASN.1 byte buffers
		getLocalKeyID( m_pBlock->ska, wcs );
		getKeyASN1Buffer( m_pBlock->ska, wcs );
		return 0;
	}
	else
	if ( m_pBlock->pka != NULL )
	{
		getLocalKeyID( m_pBlock->pka, wcs, false );
	}

	if ( typeInfoOnly )
		return 0;

	int height = r->bottom - r->top - 2;
	int sizeX = height;

	m_popup = CreateWindowExT( 
		0, 
		L"Edit",   // predefined class 
		L"",       // window text 
		WS_VISIBLE | WS_CHILD | 
		ES_LEFT | ES_NUMBER | ES_AUTOHSCROLL,
		r->left + 1,// + ebShift,	// starting x position 
		r->top,	// starting y position 
		r->right - r->left - 5 - sizeX, // - ebShift,				// width 
		height + 1,			// height 
		m_hLV,			// parent window 
		NULL,			// No menu 
		m_hInst,		// (HINSTANCE) GetWindowLong(hwnd, GWL_HINSTANCE), 
		NULL);

	if ( m_popup == NULL )
	{
		assert( false );
		return 0;
	}

	SetFocus( m_popup );

	HFONT hFont = GetWindowFont( m_hLV );
	if ( hFont )
		SetWindowFont( m_popup, hFont, false );

	SendMessage( m_popup,EM_SETMARGINS, EC_LEFTMARGIN, ebShift );

	wchar_t buff[ 256 ];
	::FormatT( buff, maxLen( buff ), L"%d", m_pBlock->keyBitSize );
	SetWindowTextT( m_popup, buff );
	SendMessage( m_popup, EM_SETSEL, 0, wcslen( buff ) );

	m_popup1 = CreateWindowExT(
		0, 
		UPDOWN_CLASSW,
		L"",
		WS_VISIBLE | WS_CHILD | 
		UDS_SETBUDDYINT,
		r->right - 4 - sizeX,	// starting x position 
		r->top - 1,					// starting y position 
		sizeX,					// width 
		height + 3,					// height 
		m_hLV,
		NULL,
		m_hInst,
		NULL);

	if ( m_popup1 == NULL )
	{
		assert( false );
		return 0;
	}

	SendMessage( m_popup1, UDM_SETBUDDY, (WPARAM)m_popup, 0 );
//	SendMessage( m_popup1, UDM_SETRANGE, 0, (LPARAM) MAKELONG((short) 4 * 1024, (short) 128 ));

//	cDlg::UDNDeltapos SetBuddy

	return 1;
}

int cWizardDlg_SecKeyPkt::OnEditFriendlyName( RECT *r, cWCStr *wcs, bool typeInfoOnly )
{
	wcs->Add( m_pBlock->friendlyName[0] == 0 ? _LoadIDS( IDS_UsageSample ) : L"" );

	if ( typeInfoOnly )
		return 0;

	int height = r->bottom - r->top - 1;

	m_popup = CreateWindowExT( 
		0, 
		L"Edit",   // predefined class 
		L"",       // window text 
		WS_VISIBLE | WS_CHILD | 
		ES_LEFT | ES_AUTOHSCROLL,
		r->left + 1, // + ebShift,	// starting x position 
		r->top,	// starting y position 
		r->right - r->left - 1, //- ebShift,				// width 
		height,			// height 
		m_hLV,							// parent window 
		NULL,							// No menu 
		m_hInst,						// (HINSTANCE) GetWindowLong(hwnd, GWL_HINSTANCE), 
		NULL);

	if ( m_popup == NULL )
	{
		assert( false );
		return 0;
	}

	HFONT hFont = GetWindowFont( m_hLV );
	if ( hFont )
		SetWindowFont( m_popup, hFont, false );

	SendMessage( m_popup,EM_SETMARGINS, EC_LEFTMARGIN, ebShift );

	SetWindowTextT( m_popup, m_pBlock->friendlyName );
	SendMessage( m_popup, EM_SETSEL, 0, wcslen( m_pBlock->friendlyName ) );
	SetFocus( m_popup );

	return 1;
}

int cWizardDlg_SecKeyPkt::OnEditPassword( RECT *r, cPassword *pwd )
{
	int height = r->bottom - r->top - 1;

	m_popup = CreateWindowExT( 
		0, 
		L"Edit",   // predefined class 
		L"",       // window text 
		WS_VISIBLE | WS_CHILD | 
		ES_LEFT | ES_AUTOHSCROLL | ES_PASSWORD,
		r->left + 1,// + ebShift,	// starting x position 
		r->top,	// starting y position 
		r->right - r->left - 1, // - ebShift ),				// width 
		height,			// height 
		m_hLV,							// parent window 
		NULL,							// No menu 
		m_hInst,						// (HINSTANCE) GetWindowLong(hwnd, GWL_HINSTANCE), 
		NULL);

	if ( m_popup == NULL )
	{
		assert( false );
		return 0;
	}

	HFONT hFont = GetWindowFont( m_hLV );
	if ( hFont )
		SetWindowFont( m_popup, hFont, false );

	SendMessage( m_popup,EM_SETMARGINS, EC_LEFTMARGIN, ebShift );

	// SetPasswordChar
	SendMessage( m_popup, EM_SETPASSWORDCHAR, (WPARAM)'*', 0 );

	SetWindowTextT( m_popup, pwd->wcBuffer );
	SendMessage( m_popup, EM_SETSEL, 0, wcslen( pwd->wcBuffer ) );
	SetFocus( m_popup );

//	lpNativePwdEditProc = GetWindowLongPtr( m_popup, GWLP_WNDPROC );
//	SetWindowLong( m_popup, GWL_WNDPROC, (LONG) PwdEditWndProc );

	return 1;
}

bool cWizardDlg_SecKeyPkt::UpdateSTDButtons()
{
	if ( m_pBlock->m_view )
	{
		ShowWindow( GetDlgItem( m_hDlg, IDC_btNext ), false );
		ShowWindow( GetDlgItem( m_hDlg, IDC_btPrev ), false );
		ShowWindow( GetDlgItem( m_hDlg, IDOK ), false );
		EnableWindow( GetDlgItem( m_hDlg, IDCANCEL ), true );
		return true;
	}

	bool enableSKGeneration = m_pBlock->isSKReady();
	bool enablePKGeneration = m_pBlock->isPKReady();

	if ( !enableSKGeneration )
	{
		EnableWindow( GetDlgItem( m_hDlg, IDC_btNext ), false );

		if ( m_pBlock->wasGenerated )
			EnableWindow( GetDlgItem( m_hDlg, IDOK ), false );
		else
			EnableWindow( GetDlgItem( m_hDlg, IDOK ), false );

		return false;
	}
	else
	if ( m_pBlock->generateCertificate )
	{
		EnableWindow( GetDlgItem( m_hDlg, IDC_btNext ), true );
		if ( m_pBlock->wasGenerated )
			EnableWindow( GetDlgItem( m_hDlg, IDOK ), false );
		else
			EnableWindow( GetDlgItem( m_hDlg, IDOK ), enablePKGeneration && enableSKGeneration );
		return true;
	}
	else
	{
		EnableWindow( GetDlgItem( m_hDlg, IDC_btNext ), false );

		if ( m_pBlock->wasGenerated )
			EnableWindow( GetDlgItem( m_hDlg, IDOK ), false );
		else
			EnableWindow( GetDlgItem( m_hDlg, IDOK ), enableSKGeneration );

		return true;
	}
}
//
//	End of cWizardDlg_SecKeyPkt
//
///////////////////////////////////////////////////////////////////////////////////////