#ifndef __WizardDlgs__
#define __WizardDlgs__

#include "cWizardBlock.h"

//
// Certificate Edit state
// 1. If m_editState == esAllowEdition ( set the value before InitDialog )
//	The Edit button is available ( it is the IDOK button with 'Edit' text )
//
// 2. If 'Edit' is inited so,
//	- m_editState = esEditionInited
//  - The 'Issuer' and 'time stamp' values are editable
//  - And 'Edit' button text is changed to 'Save'
//
// 3. If the sertificate is saved it's name is 'subject + issuer'.cer
//
enum { esNone = 0, esAllowEdition, esEditionInited };

class cWizardDlg : public cDlg
{
	HWND m_hParent;
	int m_ID;

public:
	bool m_wasInited;

	//
	// See Certificate Edit state above.
	//
	int m_editState;

	cWizardBlock *m_pBlock;

	cWizardDlg ( cWizardBlock *pBlock, HINSTANCE hInst, HWND hWnd, int ID )
	{
		m_pBlock = pBlock;
		m_hInst = hInst;
		m_hParent = hWnd;
		m_ID = ID;
		m_editState = esNone;

		m_wasInited = false;
	}

	int RunIt() { return DoModal( m_hInst, m_ID, m_hParent ); }

//	static const int CustomMessage;
//	static const DWORD CustomMessageSig;
//	virtual int OnCustomMessage( WPARAM wParam, LPARAM lParam ) { return false; }

	virtual LRESULT RealDlgProc( UINT uMsg, WPARAM wParam, LPARAM lParam );

	bool OnStdButton( WPARAM wParam );
	virtual bool OnNotify( NMHDR *hdr ) { return true; }
	virtual bool OnInitDialog() { return true; }
	virtual bool OnCommand( WPARAM wParam, LPARAM lParam ) { return OnStdButton( wParam ); }
	virtual void OnCtxHelp( HELPINFO *hInfo ) { return; }
};

//
//	cWizardDlg_Welcome
//	Allows to assign algorithm and parameters for secret key to be generated.
//	Or allows to choose an existent key in PKCS12 or PGP format
//	The parameters a stored in cWizardBlock structure passed at constructor
//
class cWizardDlg_Welcome : public cWizardDlg
{
public:
	cWizardDlg_Welcome( cWizardBlock *pBlock, HINSTANCE hInst, HWND hWnd ) 
		: cWizardDlg ( pBlock, hInst, hWnd, IDD_GenerateKeyWelcome )//IDD_GenerateSecretKey )
	{
	}
	
	bool OnInitDialog();
	bool OnNotify( NMHDR *hdr );
	bool OnCommand( WPARAM wParam, LPARAM lParam );
	bool OnBrowseKeyFile();
	bool UpdateControls();
	bool ReadSecretKeyFile( wchar_t *path );
	bool SetExistentKeyInfo();
	void OnCtxHelp( HELPINFO *hInfo );
};

//
//	cWizardDlg_SecKeyPkt
//
class cWizardDlg_SecKeyPkt : public cWizardDlg
{
	//
	//	Creates and stops editable item window
	//
	HWND m_popup,m_popup1;	// Popup windows' handles
	int m_editNum;			// Number of currrntly edited item

	//
	//	OnEditItem
	//	1. Creates popup windows for editable item
	//	2. Fills info to be displayed in down view.
	//
	bool OnEditItem( int itemNum, bool typeInfoOnly = false );
	int OnEditAlgorithm( RECT *r, cWCStr *wcs, bool typeInfoOnly );
	int OnEditKeySize( RECT *r, cWCStr *wcs, bool typeInfoOnly );
	int OnEditFriendlyName( RECT *r, cWCStr *wcs, bool typeInfoOnly );
	int OnEditPassword( RECT *r, cPassword *pwd );

	//
	//	StopEdit
	//	1. Assigns information from active popup window to wizardBlock values.
	//	2. Updates corresponded list view item.
	//	3. Closes popup windows.
	//
	bool StopEdit( bool checkPos = false, bool cancelled = false );

	//
	//	UpdateSTDButtons
	//	Changes standard buttons states
	//
	bool UpdateSTDButtons();

	//
	//	List view params
	//
	HWND m_hLV;

	//
	//	LVWndProc
	//	List view wnd proc is reassigned to catch Up-Down control notfication messages
	//
	static LONG lpNativeLVProc;
	static BOOL CALLBACK LVWndProc( HWND hLV, UINT uMsg, WPARAM wParam, LPARAM lParam );
//	virtual int OnCustomMessage( WPARAM wParam, LPARAM lParam );
//	static LONG_PTR lpNativePwdEditProc;
//	static BOOL CALLBACK PwdEditWndProc( HWND hEDT, UINT uMsg, WPARAM wParam, LPARAM lParam );

	//
	//	SetItemsValues
	//	Set values for list view.
	//
	bool SetItemsValues();

	//
	//	FillLVPassword
	//	Sets text for list view password items
	//
	bool FillLVPassword( cPassword *pwd, int itemNum = -1 );

	//
	//	FillInfo
	//	Fills info to be displayed in down view.
	//
	bool FillInfo( cWCStr *wcs, bool useFixedFont );

	//
	//	CheckRequiredFields
	//	If a required item is not properly filled sends report and returns false
	//	otherwise return true
	//
	bool CheckRequiredFields();

	void UpdateFriendlyName();

	HFONT m_fontOriginal;

public:
	cWizardDlg_SecKeyPkt( cWizardBlock *pBlock, HINSTANCE hInst, HWND hWnd )
		: cWizardDlg ( pBlock, hInst, hWnd, IDD_CreateCertificate )
	{
	}
	
	bool OnInitDialog();
	bool OnCommand( WPARAM wParam, LPARAM lParam );
	bool OnNotify( NMHDR *hdr );
};


//
//	cWizardDlg_PubKeyCrt
//

class cWizardDlg_PubKeyCrt : public cWizardDlg
{
//	x509_certificate m_cer;
	bool VerifyCertificate();

	bool m_newIssuer;

	//
	//	Creates and stops editable item window
	//
	HWND m_popup,m_popup1;	// Popup windows' handles
	int m_editNum;			// Number of currrntly edited item

	HFONT m_fontOriginal;

	//
	//	OnEditItem
	//	1. Creates popup windows for editable item
	//	2. Fills info to be displayed in down view.
	//
	bool OnEditItem( int itemNum, bool typeInfoOnly = false, bool dblClk = false );
	int OnEditIssuer( RECT *r, cWCStr *wcs, bool typeInfoOnly );
	int OnEditDate( RECT *r );
	int OnEditSubjectRDN( RECT *r, cWCStr *wcs, bool typeInfoOnly );
	bool OnGetSubjectRDN();

	//
	//	StopEdit
	//	1. Assigns information from active popup window to wizardBlock values.
	//	2. Updates corresponded list view item.
	//	3. Closes popup windows.
	//
	bool StopEdit( bool checkPos = false, bool cancelled = false );

	//
	//	UpdateSTDButtons
	//	Changes standard buttons states
	//
	bool UpdateSTDButtons();

	//
	//	FillInfo
	//	Fills info to be displayed in down view.
	//
	bool FillInfo( cWCStr *wcs, bool useFixedFont );

	//
	//	UpdateSerialNumber
	//	if issuer is undefined serial number is 1 ( means self signed certificate with new key )
	//	Otherwise, issuer has to be choosen from BestCrypt DataBase, BCDB stores certificate serial number for
	//	every secret key ( It uses BCDB_GetCrtSerialNumber( localKeyID ) function, the last returns the stored SN 
	//	and increments it ).
	//
	bool UpdateSerialNumber( bool increment = false );
	bool UpdateIssuerRDN();
	bool UpdateDateItem( int id, SYSTEMTIME *tm );
	bool UpdateSubjectRDN( cWCStr *wcs = NULL );
	bool UpdateAuthorityKeyID();
	bool UpdateSignatureAlgorithm();

	static LONG lpNativeLVProc;
	static BOOL CALLBACK LVWndProc( HWND hLV, UINT uMsg, WPARAM wParam, LPARAM lParam );
public:

	bool OnBrowseIssuer();

	HWND m_hLV;

	cWizardDlg_PubKeyCrt( cWizardBlock *pBlock, HINSTANCE hInst, HWND hWnd )
		: cWizardDlg ( pBlock, hInst, hWnd, IDD_CreateCertificate )
	{
		m_newIssuer = false;
	}
	
	bool OnInitDialog();
	bool OnCommand( WPARAM wParam, LPARAM lParam );
	bool OnNotify( NMHDR *hdr );
	void OnCtxHelp( HELPINFO *hInfo );

 private:

	enum {
		lvId_first = 0,
		lvId_version = lvId_first,
		lvId_serialNumber,
		lvId_signatureAlgorithm,
		lvId_issuerRdn,
		lvId_validFrom,
		lvId_validTo,
		lvId_subjectRdn,
		lvId_publicKey,
		lvId_subjectKeyID,
		lvId_authrityKeyID,
		lvId_basicConstraints,
		lvId_last = lvId_basicConstraints
	};

	bool UpdateFields( bool editableOnly );
	bool InsertLVItems( bool editableOnly );
	bool SetItemsValues();
	int CreateBrowseButton( RECT *r );
};

bool GetRDN( HINSTANCE hInst, HWND hWnd, sNameAttributeList *rdn, const wchar_t *caption, cDlgRect *rect );
bool getLocalKeyID( HPKA ska, cWCStr *wcsLocalKeyID, bool secret = true, bool withText = true );
bool getKeyASN1Buffer( HPKA ska, cWCStr *wcsKey, bool secret = true );

//
//	friendlyNameToRDN
//		Converts friendly name to attribute list 
//		friendly name : 'John Smith <JohnSmith@aol.com>'
//		attribute list string: CN = John Smith, E = JohnSmith@aol.com
//
bool friendlyNameToRDN( cUniversalString *friendlyName, sNameAttributeList *attList );
bool rdn2ptb( cWCStr *wcs, sNameAttributeList *rdn, char *separator = ", " );
/*
#define ListView_SetItemTextW(hwndLV, i, iSubItem_, pszText_) \
{ LV_ITEMW _ms_lvi;\
  _ms_lvi.iSubItem = iSubItem_;\
  _ms_lvi.pszText = pszText_;\
  SNDMSG((hwndLV), LVM_SETITEMTEXTW, (WPARAM)(i), (LPARAM)(LV_ITEM *)&_ms_lvi);\
}
#define ListView_GetItemTextW(hwndLV, i, iSubItem_, pszText_, cchTextMax_) \
{ LV_ITEMW _ms_lvi;\
  _ms_lvi.iSubItem = iSubItem_;\
  _ms_lvi.cchTextMax = cchTextMax_;\
  _ms_lvi.pszText = pszText_;\
  SNDMSG((hwndLV), LVM_GETITEMTEXTW, (WPARAM)(i), (LPARAM)(LV_ITEM *)&_ms_lvi);\
}
*/

class cWizardDlg_FinalInfo : public cWizardDlg
{
	cWCStr *m_finalInfo;
public:
	cWizardDlg_FinalInfo( cWizardBlock *pBlock, HINSTANCE hInst, HWND hWnd, cWCStr *finalInfo ) :
	  cWizardDlg ( pBlock, hInst, hWnd, IDD_FinalInfo )
	{
		m_finalInfo = finalInfo;
	}
	bool OnInitDialog();
};

////////////
//
//	FillLBText functions.
//

//HFONT save,current;
//static windowSize = 0;

//
//	Returns last visible of non scrolled list box.
//
int getLastVisible( HWND lBox, const wchar_t *string );

//
//	Adds wrapped string.
//	It calkulates size of list box and wraps the string by the size
//
void AddString( HWND lBox, char *string );
void AddString( HWND lBox, wchar_t *string );
int AddStrings( HWND hLB, char *str );
int AddStrings( HWND hLB, wchar_t *str );

#endif // __WizardDlgs__