#include "networkclient.h"

//
//	Function:
//		ncInit()
//	Purpouse: 
//		Allocates and initializes internal structures.
//		Use ncShutdown to free allocated data.
//	If successful returns nce_ok.
//
int ncInit(
		HNCC *hc // pointer to connection handler
	);

//
//	Function:
//		ncConnect()
//	Purpouse: 
//		Establishes connection with remote server. 
//	Returns nce_invalidConnection or nce_invalidServerAddress if no connection established.
//	If successful returns nce_ok.
//
int ncConnect(
		HNCC hc, // connection handler
		sNC_URL *dstSrv, // Destination server URL
		sNC_URL *proxySrv = 0 // Proxy server URL
	);

//
//	Function:
//		ncGetCertificate()
//	Purpouse: 
//		Gets server certificate for SSL connection ( HTTPS or LDAPS protoco types ).
//		Validity of certificate has to be checked by caller.
//	If successful returns nce_ok.
//
int ncGetCertificate(
		HNCC hc, // connection handler
		unsigned char *crtBuff, // Allocated buffer for certificate.
		int *pCrtBuffSize	// [IN] Pointer to size of allocated buffer.
							//	If allocated size is less then required 
							//	returns nce_bufferTooSmall
							// [OUT] Really written size,
	);


//
//	Handle to NC write context
//
typedef void *HNCWCTX;

//
//	Definitions of NC write function
//
typedef int ncWriteFuncType( HNCWCTX hctx, unsigned char *buff, int buffSize );



//
//	Function:
//		ncGetData()
//	Purpouse: 
//		Sends data request defined by path.
//	If successful returns nce_ok.
//
int ncGetData(
		HNCC hc,					// connection handler
		char *path,					// String with path or script on destination server.
		ncWriteFuncType *ncWrite,	// Writing function, see typedefs above,
		HNCWCTX hwctx				// The write context, it is used with ncWrite( hwctx, buff, size )
	);

//
//	Function:
//		ncShutdown()
//	Purpouse: 
//		Clears and frees all allocated data. 
//	Returns nce_ok.
//
int ncShutDown( HNCC hc );


