#ifndef _HTTP_GET_H
#define _HTTP_GET_H

///////// Socks definitions ////////////

#define	S5_VER			5		//Socks version
#define	S4_VER			4		//Socks version

// AUTHENTICATION METODS
#define	S5_AM_NO_AUTH	0		//NO AUTHENTICATION
#define	S5_AM_GSSAPI	1		//GSSAPI
#define S5_AM_CLEAR		2		//CLEAR TEXT USER/PASSWORD
#define S5_AM_NO_ACCEPT	0xff	//NO ACCEPTABLE METHODS

//COMMANDS
#define S5_CMD_CONNECT	1		//CONNECT
#define S5_CMD_BIND		2		//BIND
#define S4_CMD_CONNECT	1		//CONNECT

//ADDRESS TYPES
#define S5_AT_IPV4		1		//IP VERSION 4	
#define S5_AT_IPV6		4		//IP VERSION 6
#define S5_AT_DOMAIN	3		//FULLY-QUALIFIED DOMAIN NAME

//REPLIES
#define S5_REP_SUCCESS	0		//SUCCEEDED
#define S5_REP_FAILURE	1		//GENERAL SOCKS FAILURE
#define S5_REP_RULE		2		//CONNECTION NOT ALLOWED BY RULESET
#define S5_REP_NU		3		//NETWORK UNREACHABLE
#define S5_REP_HU		4		//HOST UNREACHABLE
#define S5_REP_REFUSED	5		//CONNECTION REFUSED
#define S5_REP_TTL		6		//TTL EXPIRED
#define S5_REP_ILL_CMD	7		//COMMAND NOT SUPPORTED
#define S5_REP_ILL_AT	8		//ADDRESS TYPE NOT SUPPORTED

#define S4_REP_SUCCESS	90		//SUCCEEDED

///////// End of Socks definitions ////////////

#define HTTP_OK                  0
#define HTTP_FAILURE            -1
#define HTTP_NO_URL             -2
#define HTTP_BAD_URL            -3
#define HTTP_BAD_PROXY          -4
#define HTTP_BAD_PROTOCOL       -5
#define HTTP_INVALID_HOST       -6      //see system error for more information
#define HTTP_INVALID_SOCKET     -7      //see system error for more information
#define HTTP_CONNECT_ERROR      -8      //see system error for more information
#define HTTP_CANCELED           -9
#define HTTP_SEND_ERROR         -10     //see system error for more information
#define HTTP_RECV_ERROR         -11     //see system error for more information
#define HTTP_INVALID_HEADER     -12
#define HTTP_FILE_OPEN_ERROR    -13     //see system error for more information
#define HTTP_FILE_WRITE_ERROR   -14     //see system error for more information
#define HTTP_SOCKS_ERROR        -15 
#define HTTP_SCONNECT_ERROR     -16     //Socks connect error


#define HTTP_RPL_OK				200 //   ; OK
#define HTTP_RPL_CREATED		201 //   ; Created
#define HTTP_RPL_ACCEPTED		202 //   ; Accepted
#define HTTP_RPL_NO_CONTENT		204 //   ; No Content
#define HTTP_RPL_MOVED_PERM		301 //   ; Moved Permanently
#define HTTP_RPL_MOVED_TEMP		302 //   ; Moved Temporarily
#define HTTP_RPL_NOT_MODIFIED	304 //   ; Not Modified
#define HTTP_RPL_BAD_REQUEST	400 //   ; Bad Request
#define HTTP_RPL_UNAUTHORIZED	401 //   ; Unauthorized
#define HTTP_RPL_FORBIDDEN		403 //   ; Forbidden
#define HTTP_RPL_NOT_FOUND		404 //   ; Not Found
#define HTTP_RPL_INTERNAL_ERR	500 //   ; Internal Server Error
#define HTTP_RPL_NOT_IMPLEMENTED 501 //   ; Not Implemented
#define HTTP_RPL_BAD_GATEWAY	502 //   ; Bad Gateway
#define HTTP_RPL_UNAVAILABLE	503 //   ; Service Unavailable

#define HTTP_ST_CONNECTING      100
#define HTTP_ST_CONNECTED       101
#define HTTP_ST_SENDING_REQUEST 102
#define HTTP_ST_WAITING_REPLY   103
#define HTTP_ST_RECV_INPROGRESS 104     // see value for number of received bytes
#define HTTP_ST_DISCONNECTING   105
#define HTTP_ST_RECEIVED_HEADER 106     // in value HTTP response (200 == OK)
#define HTTP_ST_CONTENT_LENGHT  107     // in value content lenght
#define HTTP_ST_DONE            108


/* information_callback returns HTTP_OK if OK, any other return values are signal to terminate a whole process */                  
typedef int (*information_callback)(int stage, int value, void *ctx );

//
//	cWSABlock 
//		initiates use of WS2_32.DLL by a process
//
class cWSABlock
{
public:
	cWSABlock();
	~cWSABlock();
	int getError();
};

int http_get_file(
		char *URL, 
		const char *store_file, 
		information_callback cbf = NULL, 
		void *cb_data = NULL, 
		char *proxy_URL = NULL 
	);

int http_get_file(
		char *URL, 
		const wchar_t *store_file, 
		information_callback cbf = NULL, 
		void *cb_data = NULL, 
		char *proxy_URL = NULL 
	);

#endif // _HTTP_GET_H

