#include "socks.h"

#ifndef WIN32
	#include <netinet/in.h>
	#include <sys/socket.h>
	#include <netdb.h>
#else
	#include <winsock.h>
#endif  //WIN32


#ifdef _MSC_VER
    #pragma pack(push, 1)
#endif


int s4connect (int s, const struct sockaddr *sock_addr, int namelen, st_url *socks)
{
	
	struct s4_struct {
		u_char ver;				//protocol version
		union {
			u_char cmd;			//command
			u_char rep;			//reply
		};
		u_short port;			//port number
		struct in_addr addr;	//address
	}command,reply;

	int res;
	struct hostent* he;
	struct sockaddr_in sin;
	char  *buff;	


    if ( !(he = gethostbyname((const char*)socks_host)) )
    {
        return SOCKET_ERROR;
    }
	
    memset(&sin, 0, sizeof(sin));
	sin.sin_family=AF_INET;
	sin.sin_port=htons(socks_port);
	sin.sin_addr.s_addr=*((u_long *)he->h_addr);
	
	

	res = connect(s,(const SOCKADDR*)&sin, sizeof(sin));
	if (res){
		return res;
	}


	command.ver			= S4_VER;	//socks version 4
	command.cmd			= S4_CMD_CONNECT;	//command connect
	command.port		= ((struct sockaddr_in*)lpSockAddr)->sin_port;	
	command.addr.s_addr	= ((struct sockaddr_in*)lpSockAddr)->sin_addr.s_addr;
	
	buff = new char[userlen+sizeof(command)];
	memcpy(buff,&command,sizeof(command));
	memcpy(buff+sizeof(command),(const char*)user,userlen);
	
	
	res = send(s,buff,sizeof(command)+userlen,0);
	if ((sizeof(command)+userlen) != (u_int)res)
	{
		return SOCKET_ERROR;
	}

	res = recv(s,(char *)&reply,sizeof(reply),0);
	if (sizeof(reply)!=res)
	{
		return SOCKET_ERROR;
	}

	if (reply.ver != 0)       //socks4 should return 0
	{
		return SOCKET_ERROR;
	}

	
	if(reply.rep != S4_REP_SUCCESS)		//reply = 90 is success
	{
		return SOCKET_ERROR;
	}
	
	
	return 0;
}




BOOL DoAuth(int s, st_url *socks)
{
	int res;
	char *buff;
	u_char pl,ul;

	if (user.IsEmpty() || passwd.IsEmpty())
		return FALSE;

	ul=user.GetLength();
	pl=passwd.GetLength();


	buff = new char[ul+pl+3];
	buff[0]=1;
	buff[1]=ul;
	memcpy(&buff[2],(const char*)user,ul);
	buff[2+ul]=pl;
	memcpy(&buff[2+ul+1],(const char*)passwd,pl);

	res = send(s,buff,ul+pl+3,0);

	if (ul+pl+3!=res)
	{
		return FALSE;
	}

	res = recv(s,buff,2,0);
	if (2!=res)
	{
		return FALSE;
	}

	if (buff[0]!=1 && buff[1]!=S5_REP_SUCCESS)
	{
		return FALSE;
	}

	delete []buff;	
	return TRUE;
}


int s5connect (int s, const struct sockaddr sock_addr, int namelen, st_url *socks)
{
	BOOL bUseAuth = (NULL == socks->user || NULL == socks->pass) ? FALSE : TRUE;
	
	struct s5_struct {
		u_char ver;			//protocol version
		union {
			u_char cmd;			//command
			u_char rep;			//reply
		};
		u_char rsv;			//reserved (must be 0)
		u_char atyp;		//address type (must be S5_AT_IPV4)
		struct in_addr addr;//address
		u_short port;		//port number
	}command,reply;

	int res;
	struct hostent* he;
	struct sockaddr_in sin;
	u_char	buff[4];

	if (socks_host.IsEmpty()) {
								//Please specify socks host
		return SOCKET_ERROR;
	}
	if (!(he = gethostbyname((const char*)socks_host))) {
								//Socks host unknown
		return SOCKET_ERROR;
	}
	memset(&sin, 0, sizeof(sin));
	sin.sin_family=AF_INET;
	sin.sin_port=htons(socks_port);
	sin.sin_addr.s_addr=*((u_long *)he->h_addr);
	
	res = connect(s,(const SOCKADDR*)&sin, sizeof(sin));
	if (res){
								//Can not connect to Socks server
		return res;
	}


	buff[0]=S5_VER;
	buff[1] = 1;				//number of AUTH methods

	if (bUseAuth)
		buff[2] = S5_AM_CLEAR;
	else
		buff[2] = S5_AM_NO_AUTH;

	res = send(s,(char *)buff,3,0);
	if (3!=res){
		return SOCKET_ERROR;
	}

	res = recv(s,(char *)buff,2,0);
	if (2!=res){
		return SOCKET_ERROR;
	}

	if (buff[0]!=S5_VER || ( buff[1]!=S5_AM_NO_AUTH && buff[1]!=S5_AM_CLEAR))
	{
		return SOCKET_ERROR;
	}

	if (buff[1]==S5_AM_CLEAR)
	{
		if(!bUseAuth || !DoAuth(s))
		{
			return SOCKET_ERROR;
		}
	}


	command.ver			= S5_VER;
	command.atyp		= S5_AT_IPV4;
	command.cmd			= S5_CMD_CONNECT;
	command.rsv			= 0;
	command.addr.s_addr	= ((struct sockaddr_in*)lpSockAddr)->sin_addr.s_addr;
	command.port		= ((struct sockaddr_in*)lpSockAddr)->sin_port;	

	
	res = send(s,(char *)&command,sizeof(command),0);
	if (sizeof(command)!=res)
	{
		return SOCKET_ERROR;
	}

	res = recv(s,(char *)&reply,sizeof(reply),0);
	if (sizeof(reply)!=res)
	{
		return SOCKET_ERROR;
	}

	if (reply.ver!=S5_VER)
	{
		return SOCKET_ERROR;
	}

	switch(reply.rep)
	{
	case S5_REP_SUCCESS:
		break;
	case S5_REP_FAILURE:
		return SOCKET_ERROR;
		break;
	case S5_REP_RULE:
		return SOCKET_ERROR;
		break;
	case S5_REP_NU:
		return SOCKET_ERROR;
		break;
	case S5_REP_HU:
		return SOCKET_ERROR;
		break;
	case S5_REP_TTL:
		return SOCKET_ERROR;
		break;
	case S5_REP_ILL_CMD:
		return SOCKET_ERROR;
		break;
	case S5_REP_ILL_AT:
		return SOCKET_ERROR;
		break;
	default:	
		return SOCKET_ERROR;
	}
	
	return 0;
}

#ifdef _MSC_VER
#pragma pack(pop)
#endif
