#include <windows.h>
#include "nwclient\networkClient.h"
#include "proxySrv.h"

//
//	Internet Explorer proxy
//
//	HKEY_CURRENT_USER\SOFTWARE\Microsoft\Windows\CurrentVersion\Internet Settings
//
//	ProxyEnable =
//	 0 
//	or
//	 1
//
//	ProxyServer = 
//	 193.124.169.58:8080
//	or 
//	 ftp=193.124.169.58:8080;gopher=193.124.169.58:8080;http=193.124.169.58:8080;https=193.124.169.58:8080
//
//	ProxyOverride = 
//	 diadema.ru; nsk.su; nsc.ru;<local>
//


cProxySrv curProxy;

static bool regGetValue(
        HKEY RootKey,
        const char *szSection,
        const char *szSubSection,
        DWORD dwType,   // flag for value type
        BYTE *lpData,   // address of value data
        DWORD cbData
        )
{
	HKEY hKey;

	if (RegOpenKeyExA(RootKey,szSection,0,KEY_READ,&hKey) != ERROR_SUCCESS)
			return FALSE;

	if (RegQueryValueExA(hKey,szSubSection,0,&dwType,lpData,&cbData) != ERROR_SUCCESS)
			return FALSE;

	return (RegCloseKey(hKey) == ERROR_SUCCESS);
}

static int findchar( char *str, char ch )
{
	for( int i = 0; str[i]; i++ )
	{
		if ( str[i] == ch )
			return i;
	}

	return 0;
}

static int setURLData(
		sURL *url,
		int type,
		char *str // 193.124.169.58:8080;
		)
{
	int end = findchar( str, ';' );

	if ( end == 0 )
		end = (int)strlen( str );

	int portPos = findchar( str, ':' );

	if ( ( portPos > end ) || ( portPos == 0 ) )
	{
		url->port = 0;
	}
	else
	{
		url->port = atoi( str + portPos + 1);
		end = portPos;
	}

	if ( url->addr != NULL )
		delete url->addr;

	url->addr = new char[ end + 1 ];

	if ( url->addr == NULL )
		return 2;

	strncpy( url->addr, str, end );
	url->addr[ end ] = 0;

	url->pType = type;

	return 0;
}

char *findUrlType( 
	char *type,	// "http"
	char *srvName // ftp=193.124.169.58:8080;gopher=193.124.169.58:8080;http=193.124.169.58:8080;https=193.124.169.58:8080
	)
{

	if ( findchar( srvName, ';' ) == 0 ) // Same proxy address
		return srvName;

	int typeLen = (int)strlen( type );
	for ( int pos = 0, cur = 0;  cur = findchar( srvName + pos , '=' ); pos += cur )
	{
		if ( ( cur == typeLen ) && ( strncmp( srvName + pos, type, typeLen ) ) )
		{
			//	Sample:
			//		..;https=193.124.169.58:8080...
			//		pos^ cur^

			return srvName + pos + cur;
		}
	}

	return "udf"; // specified type is not found;
}

bool cProxySrv::readMSInternetExplorerData()
{
	HKEY hKey;

	if ( RegOpenKeyExA( HKEY_CURRENT_USER, "SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\Internet Settings",0,KEY_READ,&hKey) != ERROR_SUCCESS )
			return false;

	DWORD dwType = REG_DWORD;
	DWORD cbData = sizeof ( enable );

	if ( ( RegQueryValueExA( hKey, "ProxyEnable",0,&dwType, (BYTE *)&enable, &cbData) != ERROR_SUCCESS ) ||
		 ( dwType != REG_DWORD ) ||
		 ( cbData != sizeof ( enable ) ||
		 !enable )
		)
	{
		RegCloseKey(hKey);
		return false;
	}

	//
	//	Gets override names
	//
	dwType = REG_SZ;
	cbData = 0;

	if ( ( RegQueryValueExA( hKey, "ProxyOverride",0,&dwType, NULL, &cbData) == ERROR_SUCCESS ) && 
		 ( cbData != 0 )
		)
	{
		if ( szOverride != NULL ) delete szOverride;
		szOverride = new char[ cbData ];
		if ( szOverride == 0 ) return false;

		if ( RegQueryValueExA( hKey, "ProxyOverride",0,&dwType, (BYTE *)szOverride, &cbData) != ERROR_SUCCESS )
		{
			RegCloseKey(hKey);
			return false;
		}
	}

	//
	//	Gets server name
	//
	dwType = REG_SZ;
	cbData = 0;

	RegQueryValueExA( hKey, "ProxyServer",0,&dwType, NULL, &cbData);
	if ( cbData == 0 )
	{
		RegCloseKey(hKey);
		return false;
	}

	char *srvName = new char[ cbData ];
	if ( srvName == NULL )
		return false;

	bool ok = ( RegQueryValueExA( hKey, "ProxyServer",0,&dwType, (BYTE *)srvName, &cbData) == ERROR_SUCCESS );

	RegCloseKey(hKey);

	//
	//	Parses retrieved name...
	//

	urlHTTP = new sURL;
	urlHTTPS = new sURL;

	if (( urlHTTP == NULL ) || ( urlHTTPS == NULL ) )
		return false;

	memset( urlHTTP, 0, sizeof( sURL ) );
	memset( urlHTTPS, 0, sizeof( sURL ) );

	if ( ok &&
		(
		 ( setURLData( urlHTTP, ncpt_HTTP,  findUrlType( "http", srvName ) ) != 0 ) || 
		 ( setURLData( urlHTTPS, ncpt_HTTPS,  findUrlType( "https", srvName ) ) != 0 ) 
		)
	   )
	{
		ok = false;
	}

	delete srvName;

	return ok;
}

void cProxySrv::readNetscapeData()
{
}

void cProxySrv::readMozillaData()
{
}

void cProxySrv::readFireFoxData()
{
}

cProxySrv::cProxySrv()
{
	memset( this, 0, sizeof( *this ) );
}

static void	freeUrlData( sURL *url )
{
	if ( url->addr ) delete url->addr;
	if ( url->pwd ) delete url->pwd;
	if ( url->usr ) delete url->usr;
}


cProxySrv::~cProxySrv()
{
	if ( urlHTTP ) { freeUrlData( urlHTTP ); delete urlHTTP; }
	if ( urlHTTPS ) { freeUrlData( urlHTTPS ); delete urlHTTPS; }

	if ( szOverride != NULL )
		delete szOverride;
}

static const char *szNetscapeName = "netscape";
static const char *szMozillaName = "mozilla";
static const char *szFireFoxName = "firefox";

void cProxySrv::getCurrent()
{
	char *wbName = getDefWebBrowser();

	if ( wbName[0] == 0 )
	{
		readMSInternetExplorerData();
		return;
	}

	char drive[_MAX_DRIVE];
	char dir[_MAX_DIR];
	char fname[_MAX_FNAME];
	char ext[_MAX_EXT];

	_splitpath( wbName, drive, dir, fname, ext );

	if ( _stricmp( fname, szNetscapeName ) == 0 )
		readNetscapeData();
	if ( _stricmp( fname, szMozillaName ) == 0 )
		readMozillaData();
	if ( _stricmp( fname, szFireFoxName ) == 0 )
		readFireFoxData();
	else
		readMSInternetExplorerData();
}

sURL *cProxySrv::getProxyURL( int type )
{
	if ( !curProxy.installed )
		curProxy.getCurrent();

	if ( !curProxy.enable )
		return NULL;

	switch ( type )
	{
		case ncpt_HTTP :	return curProxy.urlHTTP;
		case ncpt_HTTPS:	return curProxy.urlHTTPS;
		default :			return NULL;
	}
}

int cProxySrv::setProxyURL( int type, char *urlStr )
{
	int res;

	sURL **ppURL;

	switch ( type )
	{
		case ncpt_HTTP :	ppURL = &curProxy.urlHTTP; break;
		case ncpt_HTTPS:	ppURL = &curProxy.urlHTTPS; break;
		default :			return 1;
	}

	if ( *ppURL != NULL ) freeUrlData( *ppURL );
	else
	{
		*ppURL = new sURL;
		if ( *ppURL == NULL )
			return 2;
	}
	
	res = setURLData( *ppURL, type, urlStr );

	if ( res == 0 )
	{
		curProxy.enable = true;
		curProxy.installed = true;
	}

	return res;
}

static bool extractPath( char *path, int pathSize, const char *buff, int buffSize )
{
	if ( pathSize == 0 )
		return false;

	char sep;
	int shift;
	int i;

	if ( buff[0] == '\"' )
	{
		sep = '\"';
		i = 1;
		shift = 1;
	}
	else
	{
		sep = ' ';
		i = 0;
		shift = 0;
	}

	for( ; ; i++ )
	{
		char ch = buff[i];

		if ( ch == sep )
			break;

		if (( i >= buffSize ) ||
			(( i - shift ) == pathSize )
			)
		{
			path[ pathSize - 1 ] = '\0';
			return false;
		}

		path[ i - shift ] = ch;
	}

	path[ i - shift ] = '\0';

	return true;
}

static char *regGetDefBrowser()
{
	static char exeName[ MAX_PATH ] = "";

	static bool found = false;

	if ( found )
		return exeName;

//	The key contains
//	HKEY_CLASSES_ROOT\http\shell\open\command 
//	For explorer: 
//	Default Value = "C:\Program Files\Internet Explorer\iexplore.exe" -nohome
//	For Mozilla: 
//	Default Value = C:\Progra~1\Mozilla\Mozilla.exe URL %s bla-bla

	found = true;

	HKEY hKey;

	if ( RegOpenKeyA( HKEY_CLASSES_ROOT, "http\\shell\\open\\command", &hKey ) != ERROR_SUCCESS )
		return exeName;

	char buff[ 1024 ];

	LONG size = sizeof( buff );

	LONG res = RegQueryValueA( hKey, "", buff, &size );
	
	if ( res == ERROR_SUCCESS )
		extractPath( exeName, sizeof( exeName ), buff, size );

	RegCloseKey( hKey );

	return exeName;
}

char *cProxySrv::getDefWebBrowser()
{
	return regGetDefBrowser();
}

