#include "checkLicense.h"
#include <time.h>
#include <stdio.h>
#include <stdlib.h>
#include "../publicKeyAlgs/publicKeyAlgs.h"
#include "../x509pkts/x509_structs.h"
#include "../hashAlgs/SHA1/Sha1.h"
#include "../SMemory/SBlocks.h"
#include "../x509pkts/base64funcs.h"
#include "../x509pkts/oidFuncs.h"
#include "../PGPPkts/Packets.h"

//#include "../../Includes/cStringBuffer.h"

template < class T > T cl_string2int( const char *str, int radix = 10 )
{
	while( *str == ' ' ) str++;
	if ( ( str[0] == '0' ) && ( str[1] == 'x' ) ) // skip '0x' prefix
	{
		str += 2;
		radix = 16;
	}

	T dw = 0;

	for( ; *str; str++ )
	{
		BYTE bt = 0;
		if ( ( *str >= '0' ) && ( *str <= '9' ) )
			bt = *str - '0';
		else
		if ( radix != 0x10 )
			break;
		else
		if ( ( *str >= 'a' ) && ( *str <= 'f' ) )
			bt = *str - 'a' + 0xA;
		else
		if ( ( *str >= 'A' ) && ( *str <= 'F' ) )
			bt = *str - 'A' + 0xA;
		else
			break;

		T DW = dw * radix;
		
		if ( dw > DW ) // overflow
			return T(-1);

		dw = DW + bt;
	}

	return dw;
}

#ifdef WIN32
static wchar_t cl_hex2wc( BYTE hex )
{
	hex &= 0x0F;

	if ( hex >= 10 )
		return L'A' + ( hex - 10 );

	return L'0' + hex;
}

static wchar_t *cl_byte2hex( wchar_t *hexStr, const BYTE byte )
{
	hexStr[0] = cl_hex2wc( ( byte >> 4 ) );
	hexStr[1] = cl_hex2wc( ( byte ) );
	hexStr[2] = L' ';
	hexStr[3] = L'\0';

	return hexStr;
}

static int cl_bytes2string( wchar_t *str, int strSize, BYTE *bf, int bfSize )
{
	int len = 0;
	int pos = 0;

	for(; pos < bfSize; pos++ )
	{
		if ( len + 3 > strSize )
			break;

		wchar_t wcs[4]; // 4 wide char 'FF '
		cl_byte2hex( wcs, bf[pos] );
		wcscpy_s( str + len, strSize - len, wcs );

		len += 3;
	}

	return pos;
}
#endif

#ifndef WIN32
//#include "../../Includes/unix/misc.h"
#define MessageBoxA( _$_hWnd, _$_text, _$_caption, _$_mbId ) fprintf( stderr, "%s: %s", _$_caption, _$_text )
#endif

const char *szBeginMSG = "-----BEGIN SIGNED MESSAGE-----";
const char *szSignature = "Signature:";
const char *szEndMSG = "-----END SIGNED MESSAGE-----";

const char *szBase64BlockStart = "-----BEGIN ARMORED MESSAGE-----";
const char *szBase64BlockEnd = "-----END ARMORED MESSAGE-----";

const char *szPKStart = "-----BEGIN PUBLIC KEY-----";
const char *szPKEnd = "-----END PUBLIC KEY-----";

#ifdef __UseEmailsubject__

const char *cl_szLicenseSubjectBestCryptWindows	= "Subject: Order of BestCrypt software,";
const char *cl_szLicenseSubjectBCWipeWindows	= "Subject: Order of BCWipe software,";
const char *cl_szLicenseSubjectBestCryptLinux	= "Subject: Order of BestCrypt software for Linux,";
const char *cl_szLicenseSubjectBCWipeLinux		= "Subject: Order of BCWipe software for Linux,";
const char *cl_szLicenseSubjectCryptoSwap		= "Subject: Order of CryptoSwap software,";
const char *cl_szLicenseSubjectFireWall			= "Subject: Order of Jetico Personal Firewall software,";
const char *cl_szLicenseSubjectBCArchive		= "Subject: Order of BCArchive software,";
const char *cl_szLicenseSubjectBCWipeTWO		= "Subject: Order of BCWipePD software,";
const char *cl_szLicenseSubjectVE				= "Subject: Order of Volume Enctrytion software,";
const char *cl_szLicenseSubjectBestCryptMAC		= "Subject: Order of BestCrypt software for Mac,";
const char *cl_szLicenseSubjectBCWipeMAC		= "Subject: Order of BCWipe software for Mac,";
const char *cl_szLicenseSubjectBCVEMAC			= "Subject: Order of BCVE software for Mac,";

const char *cl_szLicenseSubjectForBestCryptWindows	= "Subject: Order for BestCrypt,";
const char *cl_szLicenseSubjectForBCWipeWindows		= "Subject: Order for BCWipe,";
const char *cl_szLicenseSubjectForBestCryptLinux	= "Subject: Order for BestCrypt for Linux,";
const char *cl_szLicenseSubjectForBCWipeLinux		= "Subject: Order for BCWipe for Linux,";
const char *cl_szLicenseSubjectForCryptoSwap		= "Subject: Order for CryptoSwap,";
const char *cl_szLicenseSubjectForFireWall			= "Subject: Order for Jetico Personal Firewall,";
const char *cl_szLicenseSubjectForBCArchive			= "Subject: Order for BCArchive,";
const char *cl_szLicenseSubjectForBCWipeTWO			= "Subject: Order for BCWipePD,";
const char *cl_szLicenseSubjectForVE				= "Subject: Order for Volume Encryption,";
const char *cl_szLicenseSubjectForBestCryptMAC		= "Subject: Order for BestCrypt for Mac,";
const char *cl_szLicenseSubjectForBCWipeMAC			= "Subject: Order for BCWipe for Mac,";
const char *cl_szLicenseSubjectForVEMac				= "Subject: Order for Volume Encryption for Mac,";

const char *cl_szLicenseSubject_Ver2_BestCryptWindows	= "Subject: \r\n          Order for BestCrypt,";
const char *cl_szLicenseSubject_Ver2_BCWipeWindows		= "Subject: \r\n          Order for BCWipe,";
const char *cl_szLicenseSubject_Ver2_BestCryptLinux		= "Subject: \r\n          Order for BestCrypt for Linux,";
const char *cl_szLicenseSubject_Ver2_BCWipeLinux		= "Subject: \r\n          Order for BCWipe for Linux,";
const char *cl_szLicenseSubject_Ver2_CryptoSwap			= "Subject: \r\n          Order for CryptoSwap,";
const char *cl_szLicenseSubject_Ver2_FireWall			= "Subject: \r\n          Order for Jetico Personal Firewall,";
const char *cl_szLicenseSubject_Ver2_BCArchive			= "Subject: \r\n          Order for BCArchive,";
const char *cl_szLicenseSubject_Ver2_BCWipeTWO			= "Subject: \r\n          Order for BCWipePD,";
const char *cl_szLicenseSubject_Ver2_VE					= "Subject: \r\n          Order for Volume Encryption,";
const char *cl_szLicenseSubject_Ver2_BestCryptMAC		= "Subject: \r\n          Order for BestCrypt for Mac,";
const char *cl_szLicenseSubject_Ver2_BCWipeMAC			= "Subject: \r\n          Order for BCWipe for Mac,";
const char *cl_szLicenseSubject_Ver2_VEMac				= "Subject: \r\n          Order for VE for Mac,";

const char *cl_szLicenseSubject_Ver2_BestCryptCorporate	= "Subject: \r\n          Order for BestCrypt Corporate,";

#endif

const char *cl_szNumberOfYears= "NUMBER_OF_YEARS:";

const char *cl_szLicenseBestCryptWindows= "SITE_LICENSE_BESTCRYPT_WINDOWS:";
const char *cl_szLicenseBCWipeWindows	= "SITE_LICENSE_BCWIPE_WINDOWS:";
const char *cl_szLicenseBestCryptLinux	= "SITE_LICENSE_BESTCRYPT_LINUX:";
const char *cl_szLicenseBCWipeLinux		= "SITE_LICENSE_BCWIPE_LINUX:";
const char *cl_szLicenseCryptoSwap		= "CRYPTOSWAP_SITE_LICENSE:";
const char *cl_szLicenseFireWall		= "FIREWALL_SITE_LICENSE:";
const char *cl_szLicenseBCArchive		= "SITE_LICENSE_BCARCHIVE:";
const char *cl_szLicenseBCArchiveW		= "SITE_LICENSE_BCARCHIVE_WINDOWS:";
const char *cl_szLicenseBCWipeTWO		= "SITE_LICENSE_BCWIPEPD:";
const char *cl_szLicenseVE				= "SITE_LICENSE_VOLUME_ENCRYPTION:";
const char *cl_szLicenseBestCryptMAC	= "SITE_LICENSE_BESTCRYPT_MAC:";
const char *cl_szLicenseBCWipeMAC		= "SITE_LICENSE_BCWIPE_MAC:";
const char *cl_szLicenseVEMac			= "SITE_LICENSE_VE_MAC:";

const char *cl_szUnitBestCryptWindows	= "BESTCRYPT_WINDOWS_UNITS:";
const char *cl_szUnitBCWipeWindows		= "BESTCRYPT_WINDOWS_UNITS:";
const char *cl_szUnitBestCryptLinux		= "BESTCRYPT_LINUX_UNITS:";
const char *cl_szUnitBCWipeLinux		= "BCWIPE_LINUX_UNITS:";
const char *cl_szUnitCryptoSwap			= "CRYPTOSWAP_UNITS:";
const char *cl_szUnitFireWall			= "FIREWALL_UNITS:";
const char *cl_szUnitBCArchive			= "BCARCHIVE_UNITS:";
const char *cl_szUnitBCWipeTWO			= "BCWIPE_TOTAL_WIPEOUT_UNITS:";
const char *cl_szUnitVE					= "VOLUME_ENCRYPTION_UNITS:";
const char *cl_szUnitBestCryptMAC		= "BESTCRYPT_MAC_UNITS:";
const char *cl_szUnitBCWipeMAC			= "BESTCRYPT_MAC_UNITS:";
const char *cl_szUnitVEMac				= "VE_MAC_UNITS:";

const char *cl_szLicenseBestCryptWindowsUpgrade	= "UPGRADE_BESTCRYPT_WINDOWS:";
const char *cl_szLicenseBCWipeWindowsUpgrade	= "UPGRADE_BCWIPE_WINDOWS:";
const char *cl_szLicenseBestCryptLinuxUpgrade	= "UPGRADE_BESTCRYPT_LINUX:";
const char *cl_szLicenseBCWipeLinuxUpgrade		= "UPGRADE_BCWIPE_LINUX:";
const char *cl_szLicenseCryptoSwapUpgrade		= "UPGRADE_CRYPTOSWAP_SITE_LICENSE:";
const char *cl_szLicenseFireWallUpgrade			= "UPGRADE_FIREWALL_SITE_LICENSE:";
const char *cl_szLicenseBCArchiveUpgrade		= "UPGRADE_BCARCHIVE_SITE_LICENSE:";
const char *cl_szLicenseBCWipeTWOUpgrade		= "UPGRADE_BCWIPEPD_SITE_LICENSE:";
const char *cl_szLicenseVEUpgrade				= "UPGRADE_VOLUME_ENCRYPTION_SITE_LICENSE:";
const char *cl_szLicenseBestCryptMACUpgrade		= "UPGRADE_BESTCRYPT_MAC:";
const char *cl_szLicenseBCWipeMACUpgrade		= "UPGRADE_BCWIPE_MAC:";
const char *cl_szLicenseVEUpgradeMac			= "UPGRADE_VE_MAC:";

const char *cl_szIncludesUpgradedBestCryptWindows	= "INCLUDES_UPGRADED_BESTCRYPT_WINDOWS:";
const char *cl_szIncludesUpgradedBCWipeWindows		= "INCLUDES_UPGRADED_BCWIPE_WINDOWS:";
const char *cl_szIncludesUpgradedBestCryptLinux		= "INCLUDES_UPGRADED_BESTCRYPT_LINUX:";
const char *cl_szIncludesUpgradedBCWipeLinux		= "INCLUDES_UPGRADED_BCWIPE_LINUX:";
const char *cl_szIncludesUpgradedCryptoSwap			= "INCLUDES_UPGRADED_CRYPTOSWAP:";
const char *cl_szIncludesUpgradedFireWall			= "INCLUDES_UPGRADED_FIREWALL:";
const char *cl_szIncludesUpgradedBCArchive			= "INCLUDES_UPGRADED_BCARCHIVE:";
const char *cl_szIncludesUpgradedBCWipeTWO			= "INCLUDES_UPGRADED_BCWIPEPD:";
const char *cl_szIncludesUpgradedVE					= "INCLUDES_UPGRADED_VOLUME_ENCRYPTION:";
const char *cl_szIncludesUpgradedBestCryptMAC		= "INCLUDES_UPGRADED_BESTCRYPT_MAC:";
const char *cl_szIncludesUpgradedBCWipeMAC			= "INCLUDES_UPGRADED_BCWIPE_MAC:";
const char *cl_szIncludesUpgradedVEMac				= "INCLUDES_UPGRADED_VE_MAC:";

const char *cl_szBestCryptWindowsRenewal	= "BESTCRYPT_WINDOWS_RENEWAL:";
const char *cl_szBCWipeWindowsRenewal		= "BCWIPE_WINDOWS_RENEWAL:";
const char *cl_szBestCryptLinuxRenewal		= "BESTCRYPT_LINUX_RENEWAL:";
const char *cl_szBCWipeLinuxRenewal			= "BCWIPE_LINUX_RENEWAL:";
const char *cl_szCryptoSwapRenewal			= "CRYPTOSWAP_RENEWAL:";
const char *cl_szFireWallRenewal			= "FIREWALL_RENEWAL:";
const char *cl_szBCArchiveRenewal			= "BCARCHIVE_RENEWAL:";
const char *cl_szBCWipeTWORenewal			= "BCWIPEPD_RENEWAL:";
const char *cl_szVERenewal					= "VOLUME_ENCRYPTION_RENEWAL:";
const char *cl_szBestCryptMACRenewal		= "BESTCRYPT_MAC_RENEWAL:";
const char *cl_szBCWipeMACRenewal			= "BCWIPE_MAC_RENEWAL:";
const char *cl_szVEMacRenewal				= "VE_MAC_RENEWAL:";
		
const char *cl_szVersion = "VERSION:";

const char *cl_szBestCryptCorporateRenewal	= "BESTCRYPT_CORPORATE_RENEWAL:";
const char *cl_szBestCryptCorporateEdition = "BESTCRYPT_CORPORATE_EDITION:";
const char *cl_szBestCryptBaseEdition = "BESTCRYPT_BASE_EDITION:";
const char *cl_szShowExpirationWarning = "SHOW_EXPIRATION_WARNING:";

const char *cl_szExpirationDate = "EXPIRATION_DATE:";
const char *cl_szLicenseType = "LICENSE_TYPE:";
const char *cl_szDate = "Date:";
const char *cl_szCreationDate = "Created:";
const char *cl_szCodePage = "CP:";

const char *cl_szName		= "NAME:";
const char *cl_szNameFirst	= "FIRST NAME:";
const char *cl_szNameLast	= "LAST NAME:";
const char *cl_szCardHolderName    = "CARDHOLDER:";
const char *cl_szCardNumber = "CARDNUM:";
const char *cl_szCardBillingAddress = "Card billing address:";
const char *cl_szCompany = "COMPANY:";
const char *cl_szAddress = "ADDRESS:";

const char *cl_szCity = "CITY:";
const char *cl_szState = "STATE:";

const char *cl_szStreetAddress = "STREET ADDRESS:";
const char *cl_szZipCode = "ZIP:";
const char *cl_szPhone   = "PHONE:";
const char *cl_szEmail	 = "EMAIL:";
const char *cl_szResellerEmail	 = "RESELLER_EMAIL:";
const char *cl_szCountry = "COUNTRY:";
const char *cl_szAdditionalInfo = "ADDITIONAL_INFO:";

/*
bool probelsAdded = false; 

static bool cutProbel( char **firstPos, int *pRestSize )
{
	if ( ( **firstPos != '\r' )			// Checks last probel in the string.
		&& ( *(*firstPos - 1 ) != ' ' )
		)
		return false;
	
	--(*pRestSize);
	memcpy( *firstPos - 1, *firstPos, *pRestSize );

	(*firstPos)--;

	assert( **firstPos == '\r' );

	return true;
}
*/

//
// Hiper text format adds a probels to the end of every strings
// We have to cut them to make valid license verification.
//
static void cutLastProbels( char *p, int dataSize )
{
	for(; *p && dataSize; p++, dataSize-- )
	{
		if ( ( *p == ' ' ) && ( *(p+1) == '\r' ) )
			memmove( p, p+1, dataSize-- );
	}
}

static bool FindFirstString(
	const char **firstPos, int *pRestSize,
	const char *szData, int dataSize )
{
	if ( ( szData == NULL ) || ( dataSize == 0 ) )
		return false;

	const char *p = szData;
	int beginLen = (int)strlen( szBeginMSG );

	for(;;)
	{
		while ( *p != '-' ) 
		{
			if ( ( dataSize-- == 0 ) || ( *(p++) == 0 ) )
				return false; // No more strings
		}

		if ( dataSize <= beginLen )
			return false; // No more strings

		if ( strncmp( p, szBeginMSG, beginLen ) == 0 )
		{
			if ( p[ beginLen ] == ' ' ) // Hiper text format adds the probels.
			{
				cutLastProbels( (char *)p, dataSize );

				return FindFirstString( firstPos, pRestSize, p, dataSize );
			}
			//else
			//{
			//	beginLen += 2; // "\r\n";
			//}

			//*firstPos = p + beginLen;
			//*pRestSize = dataSize - beginLen;

			p += beginLen;
			dataSize -= beginLen;

			while(  (*p == '\n' ) || ( *p == '\r' ) )
			{
				p++;
				if ( dataSize-- == 0 )
					return false;
			}

			*firstPos = p;
			*pRestSize = dataSize;

			return true;
		}
		else
		if ( ( dataSize-- == 0 ) || ( *(p++) == 0 ) )
				return false; // No more strings
	}
}

static bool FindString(
	const char **firstPos, int *pRestSize,
	const char *szData, int dataSize,
	const char *str,
	bool begin = true ) // otherwise end
{
	if ( ( szData == NULL ) || ( dataSize == 0 ) )
		return false;

	const char *p = szData;
	int len = (int)strlen( str );

	for(;;)
	{
		while ( *p != '-' ) 
		{
			if ( ( dataSize-- == 0 ) || ( *(p++) == 0 ) )
					return false; // No more strings
		}

		if ( strncmp( p, str, len ) == 0 )
		{
			if ( begin )
			{
				*firstPos = p;
				*pRestSize = dataSize;
			}
			else
			{
				*firstPos = p + len;
				*pRestSize = dataSize - len;
			}

			return true;
		}
		else
		if ( ( dataSize-- == 0 ) || ( *(p++) == 0 ) )
				return false; // No more strings
	}
}

//
//	extractText
//		Extracts text form message beatween strings beginMSG and endMSG
//
static int extractText(
		const char **pos, int *size,
		const char *buff, int buffSize,
		const char *beginMSG,
		const char *endMSG )

{
	if ( !FindString( (const char **)pos, size, buff, buffSize, beginMSG, true ) )
		return pkae_invalidInputData;

	const char *p;
	int s;

	if ( !FindString( &p, &s, (char *)*pos, *size, endMSG, false ) )
		return pkae_invalidInputData;

	*size = int(p - *pos);

	return pkae_ok;
}

static bool FindNextStringWithNamedValue(
	const char **nextPos, int *pRestSize,
	const char *szData, int dataSize )
{
	if ( ( szData == NULL ) || ( dataSize == 0 ) )
		return false;

	const char *p = szData;
	int endMsgLen = (int)strlen( szEndMSG );

	if ( dataSize == 0 )
		return false; // No more strings by reason 1.b) 

	if ( ( *p == '-' ) && !strncmp( p, szEndMSG, endMsgLen ) )
		return false; // No more strings by reason 1.a)

	// 0. Finds next string \r\n
	for( ; ( *p != '\n' ) && ( *p != '\r' ) ; p++, dataSize-- ) 
	{
		if ( dataSize == 0 )
		{	*nextPos = p;
			*pRestSize = dataSize;
			return true; // a) Last string found;
		}

		if ( *p == 0 )
			return false;

		if ( ( *p == '-' ) && !strncmp( p, szEndMSG, endMsgLen ) )
		{
			*nextPos = p;
			*pRestSize = dataSize;
			return true; // b) Last string found;
		}
	}

	// 1. Finds ':'
	for( ; *p != ':'; p++, dataSize-- ) 
	{
		if ( dataSize == 0 )
		{	*nextPos = p;
			*pRestSize = dataSize;
			return true; // a) Last string found;
		}

		if ( *p == 0 )
			return false;

		if ( ( *p == '-' ) && !strncmp( p, szEndMSG, endMsgLen ) )
		{
			*nextPos = p;
			*pRestSize = dataSize;
			return true; // b) Last string found;
		}
	}

	// 2. Goes back to line start.
	for(; p != szData; p--, dataSize++ )
	{
		if ( ( *p == '\r' ) || ( *p == '\n' ) )
		{
			p += 1;
			dataSize -= 1;
			break;
		}
	}

	*nextPos = p;
	*pRestSize = dataSize;
	return true;
}

static bool FindStringValue( const char *szData, int dataSize, const char *fieldName, const char **pValue, int *pValueSize )
{
	int len = (int)strlen( fieldName );

	while( ( szData ) && ( dataSize ) )
	{
		const char *next;
		int  nextRestSize;
		bool nextFound = FindNextStringWithNamedValue( &next, &nextRestSize, szData, dataSize );

		if ( strncmp( szData, fieldName, len ) == 0 )
		{
			szData += len;
			dataSize -= len;
			
			*pValue = szData;
			*pValueSize = (int)(next - szData); // - 2; // \n\r
			while( 
				( szData[ *pValueSize - 1 ] == '\n' ) || 
				( szData[ *pValueSize - 1 ] == '\r' ) 
				)
			{
				(*pValueSize)--;
			}

			return true;
		}

		if ( !nextFound )
			break;

		szData = next;
		dataSize = nextRestSize;
	}

	return false;
}

#include "PublicKey.tbl"

//
//	Function getString
//		Extracts single string from buffer without '\n' or '\r' symbols
//		Returns position to the next string, if exists, 
//		otherwise returns NULL		
//
static const char *getString( char *str, const char *ptr, int *pRestSize )
{
	if ( *pRestSize == 0 )
		return NULL;

	const char *p = ptr;

	while( ( *p != '\r' ) && ( *p != '\n' ) )
	{
		(*pRestSize)--;
		*(str++) = *(p++);

		//assert( *pRestSize != 0 );
	}

	*str = '\0';

	while( ( *p == '\r' ) || ( *p == '\n' ) )
	{
		//assert( *pRestSize != 0 );

		(*pRestSize)--;
		p++;
	}

	return p;
}

//
// Function : cl_makeInteligentDigest
//	Extrates every strings without any '\n' and '\r' and adds both "\r\n"
//
int cl_makeInteligentDigest( const BYTE *buff, int size, BYTE *digest, bool skipSpaceSymbols )
{
	int res;
	void *ctx;

	res = SHA1Allocate( &ctx );
	if ( res !=	SHA1_ERROR_NO )
		return res;

	res = SHA1Init( ctx );

	if ( res == SHA1_ERROR_NO )
	{
		if ( skipSpaceSymbols )
		{
			char ch;

			for( int i = 0; i < size; i++ )
			{
				ch = buff[i];

				// isspace depends on locale, so we'll skip 
				//	directly 09 0a 0d 20 -> ( tab cr nl ' ' )

				if ( ( ch == 0x09 ) ||
					 ( ch == 0x0a ) ||
					 ( ch == 0x0d ) ||
					 ( ch == 0x20 ) )
						continue;

				res = SHA1Process( ctx, (BYTE *)&ch, 1 );
				if ( res != SHA1_ERROR_NO )
					break;
			}
		}
		else
		{
			char str[ 512 ];
			//const char *p = (char *)buff;
			// while( p = getString( str, p, &size ) )

			for( const char *p = getString( str, (char*)buff, &size ); p; p = getString( str, p, &size ) )
			{
				strcat_s( str, sizeof(str), "\r\n" );

				res = SHA1Process( ctx, (BYTE *)str, strlen( str )  );
				if ( res != SHA1_ERROR_NO )
					break;
			}
		}

		if ( res == SHA1_ERROR_NO )
			res = SHA1Final( ctx, digest );
	}

	SHA1Free( &ctx );

	return res;
}

//
//	cl_GetDateString
//		Finds position of Date: string in user license.
//		Returns true if the string was found.
//
bool cl_GetDateString(
		const char *liPos,  // license info position
		int liSize,			// license info size
		const char **pDataStringPos,
		int	*pDataStringSize
		)
{
	return FindStringValue( liPos, liSize, cl_szDate, pDataStringPos, pDataStringSize );
}

//
//	cl_GetYearsQ
//		Returns value specified by cl_szNumberOfYears if the string is found
//		otherwise returns 1.
//
int cl_GetYearsQ(
		const char *liPos,  // license info position
		int liSize			// license info size
	)
{
	const char *pos;
	int size;

	if ( !FindStringValue( liPos, liSize, cl_szNumberOfYears, &pos, &size ) )
		return 1;

	return atoi( (char *)pos );
}

int cl_GetCodePage(
		const char *liPos,  // license info position
		int liSize			// license info size
	)
{

	const char *pos;
	int size;

	if ( !FindStringValue( liPos, liSize, "CP:", &pos, &size ) )
		return CP_ACP;

	return atoi( (char *)pos );
}

static bool getTimeFromDateString( const char *dateStr, time_t *pt )
{
	// Fri Apr 29 12:25:12 1994 (12345678)
	//if ( sscanf( (char *)pos, "%s %s %d %d:%d:%d %d (%x)", tmpBuf, tmpBuf, &tmp, &tmp, &tmp, &tmp, &tmp, &t ) != 8 )

	int pos = 0;

	for (;;)
	{
		if ( pos > 80 )
			return false;

		if ( dateStr[ pos++ ] == '(' )
			break;
	}

	unsigned int t = cl_string2int < unsigned int >( dateStr + pos, 0x10 );

	*pt = time_t( t );

	return true;
}

#ifdef _WIN32

//
//	13.12.2012
//		1. CheckLicense does not use SystemTimeToFileTime, for Linux compatibility
//		2. CheckLicense it does not use sscanf for time extracting for Veracode recuirement, the getTimeFromDateString used instead
//
//
// cl_getDaysLeft uses SystemTimeToFileTime it has to be implemented for Unix by another way ( if required )
//

//static int getMonthNumber( const char *szMonth )
//{
//	if ( strncmp( szMonth, "Jan", 3 ) == 0 ) return 1;
//	if ( strncmp( szMonth, "Feb", 3 ) == 0 ) return 2;
//	if ( strncmp( szMonth, "Mar", 3 ) == 0 ) return 3;
//	if ( strncmp( szMonth, "Apr", 3 ) == 0 ) return 4;
//	if ( strncmp( szMonth, "May", 3 ) == 0 ) return 5;
//	if ( strncmp( szMonth, "Jun", 3 ) == 0 ) return 6;
//	if ( strncmp( szMonth, "Jul", 3 ) == 0 ) return 7;
//	if ( strncmp( szMonth, "Aug", 3 ) == 0 ) return 8;
//	if ( strncmp( szMonth, "Sep", 3 ) == 0 ) return 9;
//	if ( strncmp( szMonth, "Oct", 3 ) == 0 ) return 10;
//	if ( strncmp( szMonth, "Nov", 3 ) == 0 ) return 11;
//	if ( strncmp( szMonth, "Dec", 3 ) == 0 ) return 12;
//
//	return 0;
//}

//
//	Returns number of passed days
//
int cl_getDaysPast( const char *datePos, int dateSize )
{
	return cl_getDaysLeft( datePos, dateSize, 0 );
}

//
//	cl_getDaysLeft
//		1. Extracts license generation date and adds yearsQ to get license finish date.
//		2. Substract current date from the finish date to get days left.
//	11.03.2013
//	!!!Returns the days left if it is positive, otherwise returns zero
//	Upgraded licenses mat be generated  for future time ( if old period does not expired ). But the old approach reject this licenses.
//	The version allows negative days to be added.
//
int cl_getDaysLeft( const char *datePos, int dateSize, int yearsQ )
{
	if ( dateSize == 0 )
		return 0;

	time_t t;
	if ( !getTimeFromDateString( datePos, &t ) )
		return 0;

	time_t cur = time(NULL);

//	if ( t > cur )
//		return 0;

	__int64 iDiff = cur - t;

//	if ( iDiff < 0 )
//		return 0;

	int secondsInDay = 24 * 60 * 60;

	__int64 dayCount = 365 * yearsQ - iDiff / secondsInDay;

	return (int)dayCount;

//	char szDay[100];
//	char szMonth[100];
//
//	SYSTEMTIME licenseTime;
//	memset( &licenseTime, 0, sizeof( licenseTime ) );
//
//	cPlainTextBlock ptb;
//	
//	const char *dataStr = ( char * )ptb.Assign( (const BYTE *)datePos, dateSize + 1 );
//
//	//					" Mon Sep 27 11:57:30 2004",
//	sscanf( dataStr, " %s %s %d %d:%d:%d %d",
//			szDay,
//			szMonth,
//			&licenseTime.wDay,
//			&licenseTime.wHour,
//			&licenseTime.wMinute,
//			&licenseTime.wSecond,
//			&licenseTime.wYear );
//
//	licenseTime.wMonth = (WORD)getMonthNumber( szMonth );
//
//	// Allows old one year license license to be valid up to 1 December 2004
//	// So last valid data is
//	// That is made to do no shock the customers which buy the product several years ago
//	int lastValidYear	= 2003;
//	int lastValidMonth	= 12;
//	int lastValidDay	= 1;
//
//	if ( ( licenseTime.wYear < lastValidYear ) ||
//		 ( ( licenseTime.wYear == lastValidYear ) && 
//		   ( licenseTime.wMonth < lastValidMonth ) &&
//		   ( licenseTime.wDay   < lastValidDay ) 
//		  )
//		)
//	{
//		memset( &licenseTime, 0, sizeof( licenseTime ) );
//
//		licenseTime.wYear	= (WORD)lastValidYear;
//		licenseTime.wMonth	= (WORD)lastValidMonth;
//		licenseTime.wDay	= (WORD)lastValidDay;
//	}
//
//	FILETIME fLicenseTime;
//	SystemTimeToFileTime( &licenseTime, &fLicenseTime );
//
//	SYSTEMTIME curTime;
//	GetSystemTime( &curTime );
//
//	FILETIME fCurTime;
//	SystemTimeToFileTime( &curTime, &fCurTime );
//
//	LARGE_INTEGER liDiff;
//
//	liDiff.HighPart = fCurTime.dwHighDateTime;
//	liDiff.LowPart  = fCurTime.dwLowDateTime;
//
//	liDiff.QuadPart -= fLicenseTime.dwLowDateTime;
//	liDiff.HighPart -= fLicenseTime.dwHighDateTime;
//
//	__int64 _100NanoSecInDay = ( __int64(24) * 60 * 60 ) * ( 1000000000L / 100 );
//
//	int dayCount = 365 * yearsQ - (int)(liDiff.QuadPart / _100NanoSecInDay);
//
////	if ( dayCount < 0 )
////		dayCount = 0;
//
//	return dayCount;
}

#endif // #ifndef _WIN32_

//
//	Subkey strategy notes.
//		Every license message generated by reseller has to contain the reseller public key
//			inside the message text.
//		Verification module finds the subkey, verifies it by main public key and uses the 
//			verified reseller public key for license verification.
//

//
//	getSubkey
//
//		1. Parses certificate buffer
//		2. Verify the public key by main pka
//		3. Creates public key handle.
//		4. Checks validity time. If validityTime is 0, so current time is used for validity.
//
//		Returns the verified public key or NULL.
//
static HPKA getSubkey( const BYTE *crtBuff, int crtSize, HPKA pka, time_t validityTime )
{
	if ( crtSize == 0 )
		return NULL;

	// 3.
	x509_certificate crt;
	
	int res = crt.parse( (BYTE*)crtBuff, crtSize );
	if ( res != pkae_ok ) return NULL;

	// 4.
	tUnit *tu;

	if ( generateOIDName( NULL, 0, crt.signatureAlg.oid.oidStr, &tu ) == 0 )
		return NULL;

	BYTE haID = tu->hAlg;
	BYTE pkID = tu->pAlg;

	//
	//	05.06.2013
	//	We check validity of certificate by licecense creation date. Web shop and standard license generators add 'Created:' value to license.
	//
	if ( !crt.tbsCertificate.validity.isValid( validityTime ) )
	{
		MessageBoxA( 0, "Certificate validity period is finished", "Check license error", MB_ICONSTOP );
		return NULL;
	}
	
	res = crt.verify( haID, pkID, pka );
	if ( res != pkae_ok ) return NULL;

	// 5.
	HPKA subKey = NULL;
	res = pka_createAlgorithm_X509( &subKey, &crt.tbsCertificate.subjectPKInfo );
	if ( res != pkae_ok  ) return NULL;

	return subKey;
}

//
//	getSubkey
//		1. Finds public key certificate at buffer
//		2. Decodes it from base 64 armored text.
//
static HPKA getSubkeyEx( const char *buff, int buffSize, HPKA pka, time_t validityTime )
{
	cPlainTextBlock ptbCrt;

	if ( !cl_readArmoredBlock( &ptbCrt, buff, buffSize, szPKStart, szPKEnd ) )
		return NULL;

	return getSubkey( ptbCrt.buffer, ptbCrt.bufferSize, pka, validityTime );
}

//
//	cl_verifyCertificate
//		Verifies certificate by assigned public key, 
//		if pka is NULL it uses Jetico Public Key as default.
//
int cl_verifyCertificate( const BYTE *crtBuff, int crtSize, HPKA pka, time_t validityTime ) // pka = NULL, validityTime = 0
{
	x509_publicKeyInfo pki;
	HPKA lPka = NULL;

	int res;

	if ( pka == NULL )
	{
		res = pki.parse( (BYTE*)publicKey, sizeof( publicKey ) );
		if ( res != x509e_ok ) 	return res;

		res = pka_createAlgorithm_X509( &lPka, &pki );
		if ( res != pkae_ok ) return res;

		pka = lPka;
	}

	HPKA subKey = getSubkey( crtBuff, crtSize, (HPKA)pka, validityTime );

	if ( lPka != NULL )
		pka_free( lPka );

	if ( subKey == NULL )
		return pkae_invalidSignature;

	pka_free( subKey );

	return pkae_ok;
}

//
//	cl_VerifySignature
//		1. parses signed message
//		2. extracts license info
//		3. returns values
//			pkae_invalidSignature - if signature invalid,
//			pkae_invalidInputData - if message buffer does not contain label strings
//			pkae_ok	- if signature is valid.
//
int cl_VerifySignature(
		const char *messageBuff, int messageBuffSize,
		const BYTE *pk, int pkSize,
		const char **pLIPos, int *pLISize, // OUT position and size of license info 
		UINT *pSN, UINT *pTQ,
		HPKA hPka_param // = NULL
	)
{
	char *pos;
	int size;

	if ( !FindFirstString( (const char **)&pos, &size, (char *)messageBuff, messageBuffSize ) )
		return pkae_invalidInputData;

	BYTE *b64EncSigPos;
	int b64EncSigSize;

	if ( !FindStringValue( (const char *)pos, size, szSignature, (const char **)&b64EncSigPos, &b64EncSigSize ) )
		return pkae_invalidInputData;

	//
	//	27.03.2012 
	//		Without the check base64Decode returns ppe_ok for empty signature.
	//		An as result license is treated as valid.
	//
	if ( b64EncSigSize == 0 )
		return pkae_invalidInputData;

	size = (int)( (char *)b64EncSigPos - pos ) - (int)strlen( szSignature );

	if ( ( pLIPos != NULL ) && ( pLISize != NULL ) )
	{
		*pLIPos = (const char *)pos;
		*pLISize = size;
	}

	// Make digest with space symbols
	BYTE digest[ SHA1_DIGEST_SIZE ];
	int res = cl_makeInteligentDigest( (BYTE *)pos, size, digest, false );
	if ( res != ppe_ok )
		return res;

	// Make digest without space symbols
	BYTE digestEx[ SHA1_DIGEST_SIZE ];
	res = cl_makeInteligentDigest( (BYTE *)pos, size, digestEx, true );
	if ( res != ppe_ok )
		return res;

	int sigSize = 0;

	res = base64Decode( NULL, &sigSize, b64EncSigPos, b64EncSigSize );

	if ( res != pkae_bufferSizeTooSmall )
		return res;

	cPlainTextBlock ptbSig;

	if ( ptbSig.Allocate( sigSize ) == NULL )
		return x509e_memoryAllocation;

	res = base64Decode( ptbSig.buffer, &sigSize, b64EncSigPos, b64EncSigSize );
	if ( res != ppe_ok ) return res;


	// Decrypts

	HPKA pka;

	x509_publicKeyInfo pki;

	if ( hPka_param != NULL )
	{
		pka = (HPKA)hPka_param;
	}
	else
	{
		res = pki.parse( (BYTE*)publicKey, sizeof( publicKey ) );
		if ( res != x509e_ok ) 	return res;

		res = pka_createAlgorithm_X509( &pka, &pki );
		if ( res != pkae_ok ) return res;
	}

	//
	//	06.08.2012 - Gets license time.
	//	The time is required for subkey validation. 
	//	It has to be valid for license generation time.
	//	Previous version checked subkey validity by current time.
	//
	//	21.05.2013 - Gets creation time first.
	// 
	time_t licenseTime = 0;
	res = cl_GetLicenseTime( &licenseTime, (char *)messageBuff, messageBuffSize, cl_szCreationDate );

	if ( res != pkae_ok )
		res = cl_GetLicenseTime( &licenseTime, (char *)messageBuff, messageBuffSize, cl_szDate );
	
	if ( res != pkae_ok )
		licenseTime = 0;

	//
	//	Tries to find verification key at specified public key buffer ( is decoded ) first and
	//		same message buffer ( is base64 encoded text ).
	//	See subkey strategy notes.
	//
	HPKA subKey = getSubkey( pk, pkSize, pka, licenseTime );

	if ( subKey == NULL )
		subKey = getSubkeyEx( messageBuff, messageBuffSize, pka, licenseTime );

	HPKA verifyKey = subKey != NULL ? subKey : pka;

	int decSize = 0;
	res = pka_decrypt_X509( verifyKey, NULL, &decSize, ptbSig.buffer, ptbSig.bufferSize );

	if ( res != pkae_bufferSizeTooSmall )
		return res;

	cPlainTextBlock ptbDec;

	if ( ptbDec.Allocate( decSize ) == NULL )
	{
		if ( subKey != NULL )
			pka_free( subKey );

		if ( hPka_param == NULL )
			pka_free( pka );

		return x509e_memoryAllocation;
	}

	res = pka_decrypt_X509( verifyKey, ptbDec.buffer, &decSize, ptbSig.buffer, ptbSig.bufferSize );

	if ( subKey != NULL )
		pka_free( subKey );

	if ( hPka_param == NULL )
		pka_free( pka );

	if ( res != pkae_ok )
		return res;

	// Parses and checks public key formatted block...

	int valueSize = 0;

	//	v. 7.06.3 disables using of block type 02 in signature
	if ( ptbDec.buffer[1] == 0x02 )
		return pkae_internalError;

	//
	//	Error in block formatting has been fixed in Certificate lib v. 1.01
	//	But we have to use the rudimentary function for compatibility with old licenses.
	//
	res = PKT_PublicKeyEncryptionBlock_Parse_Rudiment( NULL, &valueSize, ptbDec.buffer, decSize );

	if ( res != pkae_bufferSizeTooSmall )
		return res;

	//	No need to extract value, we know positions of digest and number of copies bytes SN and TQ
	//	So we ommit the resonable using of PKT_PublicKeyEncryptionBlock_Parse.
	//	It is useful to check only size of parsed value ( really it checkes also one padded byte ).

	if ( valueSize != sizeof( digest ) +  2 * sizeof( UINT ) )
		return ppe_invalidBlock;

	BYTE *ptr = ptbDec.buffer + decSize - sizeof( digest );

	if ( pSN && pTQ )
	{
		UINT *dw = (UINT *)(ptr - 2 * sizeof( UINT ));

		*pSN = dw[0];
		*pTQ = dw[1];

//		BYTE *type = ( BYTE *)pSN;
//		type[0] = ptbDec.buffer[1];
	}

	// Compares digest values.

	if ( ( memcmp( ptr, digest,   sizeof( digest   ) ) != 0 ) &&
		 ( memcmp( ptr, digestEx, sizeof( digestEx ) ) != 0 ) )
			return pkae_invalidSignature;

	return pkae_ok;
}


//
//	cl_GetLicenseValue
//		Finds position and size of named value
//
int cl_GetLicenseValue(
		const char **pValuePos, int *pValueSize,
		const char *messageBuff, int messageBuffSize,
		const char *valueName
	)
{
	if ( !FindFirstString( (const char **)&messageBuff, &messageBuffSize, (char *)messageBuff, messageBuffSize ) )
		return pkae_invalidInputData;

	if ( !FindStringValue( (char *)messageBuff, messageBuffSize, valueName, (const char **)pValuePos, pValueSize ) )
		return pkae_invalidInputData;

	return pkae_ok;	
}

//
//	cl_GetLicenseValue
//		Extract integer data of named value
//
int cl_GetLicenseValueInt(
		int *pValue,
		const char *messageBuff, int messageBuffSize,
		const char *valueName
	)
{
	const char *pos;
	int size;

	if ( cl_GetLicenseValue( &pos, &size, messageBuff, messageBuffSize, valueName ) != pkae_ok )
		return pkae_invalidInputData;

	*pValue = atoi( (char *)pos );

	return pkae_ok;
}

//
//	cl_GetLicenseTime
//		Extracts date of szTimeField.
//			The value may be one of them cl_szExpirationDate or cl_szDate. 
//			Both the values are filled as Fri Apr 29 12:25:12 1994 (12345678)
//			Really it extracts last (12345678) as (time_t).
//		returns values
//			pkae_invalidInputData - if message buffer does not contain 
//				the szTimeField value.
//			ppe_invalidBlock - if value was found but was not properly scanned.
//			pkae_ok	- if value was found and properly parsed.
//
int cl_GetLicenseTime(
		time_t *pTime,
		const char *messageBuff, int messageBuffSize, // whole message with Begin and End strings
		const char *szTimeField // = cl_szExpirationDate
	)
{
	const char *pos;
	int size;

	if ( cl_GetLicenseValue( &pos, &size, messageBuff, messageBuffSize, szTimeField ) != pkae_ok )
		return pkae_invalidInputData;

	// Obtain coordinated universal time:
	// Fri Apr 29 12:25:12 1994 (12345678)

	// unsigned int t;
	//int tmp;
	//char tmpBuf[ 100 ];
	//if ( sscanf( (char *)pos, "%s %s %d %d:%d:%d %d (%x)", tmpBuf, tmpBuf, &tmp, &tmp, &tmp, &tmp, &tmp, &t ) != 8 )
	//*pTime = t;

	if ( !getTimeFromDateString( (char *)pos, pTime ) )
		return ppe_invalidBlock;

	return pkae_ok;
}

//
//	cl_CheckExpiredDemo
//		Extracts date of cl_szExpirationDate
//		returns values
//			pkae_invalidInputData - if message buffer does not contain 
//				the cl_szExpirationDate value, so the license is not demo version, 
//				and cannot be expired.
//			pkae_ok	- if value was found, so the licenss is demo version
//				*pValue will contains number of evaluation days
//
int cl_CheckExpiredDemo(
		__int64 *pValue,
		const char *messageBuff, int messageBuffSize // whole message with Begin and End strings
	)
{
	//const char *pos;
	//int size;

	//if ( cl_GetLicenseValue( &pos, &size, messageBuff, messageBuffSize, cl_szExpirationDate ) != pkae_ok )
	//	return pkae_invalidInputData;

	time_t licenseTime = 0, curTime = 0;

	time( &curTime );

	const __int64 secInDay = ( 24 * 60 * 60 ); // seconds in day

	//int tmp;
	//char tmpBuf[ 100 ];
	//if ( sscanf( (char *)pos, "%s %s %d %d:%d:%d %d (%x)", tmpBuf, tmpBuf, &tmp, &tmp, &tmp, &tmp, &tmp, &licenseTime ) != 8 )
	int res = cl_GetLicenseTime( &licenseTime, messageBuff, messageBuffSize, cl_szExpirationDate );
	
	if ( res == pkae_invalidInputData )
		return pkae_invalidInputData;

	if ( res != pkae_ok )
		*pValue = 0;
	else
		*pValue = (__int64)((licenseTime - curTime + secInDay - 1 ) / secInDay );

	return pkae_ok;
}

//
//	cl_ExtractDemoBody
//		Extracts content between the szBeginMSG and cl_szExpirationDate.
//		if properly extracted returns  pkae_ok.
//		Note, it does not check license validity, so we can use any unsigned patern
//
int cl_ExtractDemoBody(
		const char **pBody, int *pBodySize,
		const char *templateBuff, int templateBuffSize )
{
	char *body;
	int bodySize;

	if ( !FindFirstString( (const char **)&body, &bodySize, (char *)templateBuff, templateBuffSize ) )
		return pkae_invalidInputData;

	char *expDate = NULL;
	int expDateSize = 0;

	if ( !FindStringValue( (char *)body, bodySize, cl_szExpirationDate, (const char **)&expDate, &expDateSize ) )
		return pkae_invalidInputData;

	expDate -= (int)strlen( cl_szExpirationDate );

	*pBody = body;
	*pBodySize = int( expDate - body );

	return pkae_ok;	
}

//
//	cl_ExtractSignedLicense
//		Finds position of szBeginMSG string in buffer
//		Gets size upto end of szEndMSG
//		if successful returns pkae_ok;
//
int cl_ExtractSignedLicense(
		const char **pos, int *size,
		const char *buff, int buffSize )
{
	return extractText( pos, size, buff, buffSize, szBeginMSG, szEndMSG );

	//if ( !FindString( (const char **)pos, size, buff, buffSize, szBeginMSG, true ) )
	//	return pkae_invalidInputData;

	//const char *p;
	//int s;

	//if ( !FindString( &p, &s, (char *)*pos, *size, szEndMSG, false ) )
	//	return pkae_invalidInputData;

	//*size = int(p - *pos);

	//return pkae_ok;
}

bool cl_createArmoredBlock( cPlainTextBlock *ptb, const BYTE *buff, int buffSize, const char *szStart, const char *szEnd )
{	return createArmoredBlock( ptb, buff, buffSize, szStart, szEnd ); }

bool cl_readArmoredBlock( cPlainTextBlock *ptb, const char *text, int textSize, const char *szStart, const char *szEnd )
{	return readArmoredBlock( ptb, text, textSize, szStart, szEnd ); }

int cl_readArmoredBlock( BYTE *bf, int *bfSize, const char *text, int textSize, const char *szStart, const char *szEnd )
{	return readArmoredBlock( bf, bfSize, text, textSize, szStart, szEnd ); }

bool cl_checkBase64Encoding(
		const BYTE *encBlock, 	int encBlockSize,
		const BYTE *decBlock,	int decBlockSize,
		const char *szBlockStart,
		const char *szBlockEnd
	)
{	return checkBase64Encoding(	encBlock, encBlockSize,	decBlock, decBlockSize,	szBlockStart, szBlockEnd ); }

#ifdef WIN32
int cl_getFingerprint(
	wchar_t *bfOut,		// wide character buffer with prepared fingerprint
	int *pBfOutSize,	// the count of wide characters
	const BYTE *bfIn,	// byte buffer with text
	int bfInSize
   )
{
	BYTE digest[ SHA1_DIGEST_SIZE ];

	int res = cl_makeInteligentDigest( bfIn, bfInSize, digest, true );

	if ( res != hae_ok )
		return res;

	if ( *pBfOutSize < sizeof( digest ) * 3 + 1 )
		res = ppe_bufferTooSmall;

	*pBfOutSize = sizeof( digest ) * 3 + 1;

	if ( res != hae_ok )
		return res;
		
	res = cl_bytes2string( bfOut, *pBfOutSize, digest, sizeof( digest ) );

	if ( res != sizeof( digest ) )
		return hae_internalError;

	return hae_ok;
}
#endif
//
//	Human readable values...
//
static const char *hrT[] = 
{
	cl_szName,		"User name:\t",
	cl_szCompany,	"Company:\t",
	cl_szAddress,	"Address:\t",
	cl_szZipCode,	"Zip:\t",
	cl_szCountry,	"Country:\t",
	cl_szPhone,		"Phone:\t",
	cl_szEmail,		"eMail:\t",
	cl_szBestCryptCorporateEdition, "Enterprise version",
	cl_szLicenseBestCryptWindows, "BestCrypt for Windows:",
	cl_szLicenseBCWipeWindows, "BCWipe for Windows:",
	cl_szLicenseBestCryptLinux, "BestCrypt for Linux:",
	cl_szLicenseBCWipeLinux, "BCWipe for Linux:",
	cl_szLicenseFireWall, "Jetico Personal Firewall:",
	cl_szLicenseBCArchive, "BCArchive:",
	cl_szLicenseBCWipeTWO, "BCWipe Total Wipe Out:",
	cl_szLicenseVE, "BestCrypt Volume Encryption:",
	cl_szLicenseBestCryptMAC, "BestCrypt for Mac:",
	cl_szLicenseBCWipeMAC, "BCWipe for Mac:",
	cl_szNumberOfYears, "Number of years:",
	cl_szDate, "Date:",
	cl_szExpirationDate, "License for evaluation purposes.",
	cl_szLicenseVEMac, "BC Volume Encryption for Mac:",
};

static const char *getHumanReadbleValue( const char *szValue, int len, bool *pUseValaue )
{
	for ( unsigned int q = 0; q < sizeof( hrT ) / sizeof( *hrT ); q += 2 )
	{
		if ( strncmp( hrT[q], szValue, len ) == 0 )
		{
			if ( hrT[q] == cl_szExpirationDate )
				*pUseValaue = false;

			return hrT[q + 1 ];
		}
	}

	return NULL;
}

int cl_getDescription( cPlainTextBlock *ptbOut, BYTE *bfLicence, int bfSize, int *pCP, const char *szPrefix ) // = NULL )
{
	//int res = cl_VerifySignature( bfLicence, bfSize, hPka_param );
	//
	//if ( res != pkae_ok )
	//	return res;

	cPlainTextBlock ptb;

	if ( !cl_readArmoredBlock( &ptb, (const char *)bfLicence, bfSize, szBase64BlockStart, szBase64BlockEnd ) )
		return pkae_invalidInputData;

	const char *szData = (const char *)ptb.buffer;
	int dataSize = ptb.bufferSize;

	const char *szValue;
	int valueSize;

	*pCP = cl_GetCodePage( szData, dataSize );

	while( ( szData ) && ( dataSize ) )
	{
		const char *next;
		int  nextRestSize;
		bool nextFound = FindNextStringWithNamedValue( &next, &nextRestSize, szData, dataSize );

		if ( szData[0] != '-' )
		{
			int len = 0;

			while( szData[len++] != ':' )
			{
				if ( len > dataSize )
					return pkae_invalidInputData;
			}

			szValue = szData + len;

			valueSize = (int)(next - szValue ); // - 2; // \n\r

			while( 
				( szValue[ valueSize - 1 ] == '\n' ) || 
				( szValue[ valueSize - 1 ] == '\r' ) 
				)
			{
				valueSize--;
			}

			bool bUseValue = true;

			const char *szDescription = getHumanReadbleValue( szData, len, &bUseValue );

			if ( szDescription != NULL )
			{
				if ( szPrefix != NULL )
					ptbOut->WriteString( szPrefix, false );

				ptbOut->WriteString( szDescription, false );
				
				if ( bUseValue )
				{
					ptbOut->WriteString( "\t", false );
					ptbOut->WriteBytes( (const BYTE*)szValue, valueSize );
				}

				ptbOut->WriteString( "\r\n", false );
			}
		}

		if ( !nextFound )
			break;

		szData = next;
		dataSize = nextRestSize;
	}

	ptbOut->WriteString( "\r\n", false );

	return pkae_ok;
}
