/////////////////////////////////////////////////////////////////
//
//      Copyright (c) 2017 Jetico, Inc., Finland
//      All rights reserved.
//
//      File: camellia.h
//      Revision:
//      Created:	22.07.2017
//      Description:
//			RFC 3713             Camellia Encryption Algorithm            April 2004
//
/////////////////////////////////////////////////////////////////

#ifndef __cAlgCamellia__
#define __cAlgCamellia__

#include "ALG_DEF.H"
#include "cSymAlg.h"

class cAlgCamellia: public cSymAlg
{
 private:
	void *m_ctx;
	int  m_keySize;

	BOOL getAlgConstants( int *blockSize, int *keySize );

protected:
	static cSymAlg *_createAlg( cAlgCamellia *alg, BYTE *key, int keySize );

 public:

     cAlgCamellia();
	 ~cAlgCamellia();

	static int getBlockSize();
	static const char *getDisplayName();

	// returns minimum key size in bytes, 
	// if algorithm supports other key length, pMax contains maximum key size
	// if algorithm supports fixed key length, pMax is equal to returned value
	static int getKeySize( int *pMax = NULL );

	static cSymAlg *createAlg( BYTE *key, int keySize );  // keySize in bytes

	// Encryption and decription in Electronic Code Book mode

	virtual int encryptECB(
		const BYTE *inBf,		// Input plaintext buffer. 
		BYTE *outBf,	// Output encrypted buffer
		int  bfSize		// In/out buffers' size.
		);

	virtual int decryptECB(
		const BYTE *inBf,		// Input encrypted buffer.
		BYTE *outBf,	// Output plaintext buffer.
		int  bfSize		// In/out buffers' size.
		);
};

class cAlgCamellia128: public cAlgCamellia
{
public:
	static int getKeySize( int *pMax = NULL )
	{
		if (pMax)
			*pMax = 128 / 8;

		return 128 / 8;
	}

	static cSymAlg *createAlg( BYTE *key, int keySize );  // keySize in bytes
};

class cAlgCamellia192: public cAlgCamellia
{
public:
	static int getKeySize( int *pMax = NULL )
	{
		if (pMax)
			*pMax = 196 / 8;

		return 196 / 8;
	}

	static cSymAlg *createAlg( BYTE *key, int keySize );  // keySize in bytes
};

#endif // __cAlgCamellia__

