/*---------------------------------------------------------------------------*\

      Copyright (c) 2018 Jetico, Inc., Finland
      All rights reserved.

      File: textEncode.h
      Revision:
      Created:
      Description:

		Password based encryption.
		Creates base 64 encoded packet with SymmetricKeyEncryptedSessionKey
		and Sym Encrypted and Integrity Protected Data packet ( tag 18 )

		5.3. Symmetric-Key Encrypted Session-Key Packets (Tag 3), see RFC 2440
			The body of this packet consists of:
			 - A one-octet version number. The only currently defined version
			   is 4.
			 - A one-octet number describing the symmetric algorithm used.
			 - A string-to-key (S2K) specifier, length as defined above.
			 - Optionally, the encrypted session key itself, which is decrypted
			   with the string-to-key object.


\*---------------------------------------------------------------------------*/

#ifndef _textEncode_
#define _textEncode_

#include "../SMemory/SBlocks.h"
#include "secretKey.h"

int encodeTextByPassword(
		cPTB &ptbBase64Text,
		const char *plainTextBuffer, int plainTextBufferSize,
		int saAlgID,
		const char *utf8pwd,
		const unsigned char *seedBuffer, int seedBufferSize,
		const char *szVersionInfo );

typedef class cPKA * HPKA; // Handle to public/private key algorithm

int encodeTextByPublicKey(
		char *encodedBuff, int &encodedBufferSize,
		const char *plainTextBuffer, int plainTextBufferSize,
		int saAlgID,
		HPKA pka, unsigned char *keyID,
		const unsigned char *seedBuffer, int seedBufferSize
	);

int decodeText(
		cPTB &ptbPlainText,
		const char *base64encodedBuffer, int base64encodedBufferSize,
		const char *utf8pwd,
		cSecretKey &secretKey
	);

int encodeTextTest();

#endif // _textEncode_
