#ifndef _AutoDeallocate_h_
#define _AutoDeallocate_h_

#include <assert.h>
#include <memory.h>

template < class T > bool swapValues( T *v0, T *v1 )
{
	char *pT = new char[ sizeof( T ) ];

	if ( pT == NULL )
	{
		assert( false );
		return false;
	}

	memcpy( pT, v0, sizeof( T ) );
	memcpy( v0, v1, sizeof( T ) );
	memcpy( v1, pT, sizeof( T ) );

	delete [] pT;

	return true;
}

//
//	class cADBlock
//	 Auto Deallocated Block
//
template < class T > class cADBlock
{
public:

	T *m_T;
	int m_Q;

	bool m_setZero;

	bool Add( T in )
	{
		return Add( &in, 1 );
	}

	bool Add( T *inT, int size = 1 )
	{
		if ( size <= 0 )
			return false;

		T *newT = allocate( m_Q + size, m_setZero );

		if ( newT == NULL )
			return false;

		int q = 0;
		for( ; q < m_Q; q++ )
			newT[q] = m_T[q];

		Free();

		m_T = newT;
		m_Q = q;

		for( q = 0; q < size; q++,m_Q++ )
		{
			if ( inT != NULL )
				m_T[m_Q] = inT[q];
			else
			if ( m_setZero )
				memset( m_T + m_Q, 0, sizeof( T ) );
		}

		return true;
	}

	T *allocate( int size, bool setZero )
	{
		if ( size == 0 )
			return NULL;

		T *pT = new T [ size ];

		if ( ( pT != NULL ) && setZero )
			memset( pT, 0, sizeof( T ) * size );

		return pT;
	}

	void Free()
	{
		if ( m_T != NULL )
		{
			if ( m_setZero )
				memset( m_T, 0, sizeof( T ) * m_Q );

			delete [] m_T;
		}

		m_T = NULL;
		m_Q = 0;
	}

	cADBlock( int Q = 1, bool setZero = true )
	{
		memset( this, 0, sizeof( *this ) );

		m_setZero = setZero;

		m_T = allocate( Q, setZero );

		if ( m_T != NULL )
			m_Q = Q;
	}

	~cADBlock()
	{
		Free();
	}

	int Find( T item )
	{
		int i = m_Q;

		while( i-- )
		{
			if ( m_T[i] == item )
				break;
		}

		return i;
	}
	
	int Remove( int i )
	{
		if ( i >= m_Q )
			return -1;

		for ( ; i < m_Q - 1; i++ )
		{
			m_T[i] = m_T[i + 1];
		}

		m_T[--m_Q] = NULL;

		return i;
	}
};

//
//	Resizeable table of T elements.
//	The class T element has to contain Free function.
//
template < class T > class cADTable : public cADBlock < T >
{
public:
	cADTable( int size = 0, bool setZero = false ) 
	    : cADBlock < T > ( size, setZero )
	{
	}

	T * Add( T *in, int size = 1 )
	{
		if ( size == 0 )
			return NULL;

		if ( !Resize( this->m_Q + size ) )
			return NULL;

		if ( in != NULL )
			memcpy( this->m_T + this->m_Q - size, in, sizeof ( T ) * size );

		return this->m_T + this->m_Q - size;
	}

	T * AddCopy( T *in, int size = 1 )
	{
		T *added = Add( NULL, size );

		if ( added == NULL )
			return NULL;

		if ( in != NULL )
		{
			for( int i = 0; i < size; i++ )
			{
				added[i].CopyFrom( in + i );
			}
		}

		return added;
	}

	bool Resize( int size )
	{
		while( size < this->m_Q )
		{
			this->m_T[ --(this->m_Q) ].Free(); // Free content...

			if ( this->m_setZero )
				memset( this->m_T + this->m_Q, 0, sizeof( T ) );
		}

		if ( this->m_Q == 0 ) 
			this->Free();

		if ( size == this->m_Q )
			return true;

		T *pT = allocate( size, this->m_setZero );

		if ( pT == NULL )
			return false;

		if ( this->m_T != NULL )
		{
			for( int i = 0; i < this->m_Q; i++ )
				swapValues( this->m_T + i, pT + i );

			this->Free();
		}
		
		this->m_T = pT;
		this->m_Q = size;

		return true;
	}
	
	~cADTable()
	{
		Free();
	}
};


//
//	cAdArray did not call destructor for every element insted of cADTable does.
//

template < class T > class cADArray : public cADBlock < T >
{
public:
	cADArray( int size ) : cADBlock < T > ( size, true ) {}
};


template < class T > class cAutoDeallocate
{
public:
	int m_len;
	T *m_ptr;

	void Free()
	{
		if ( m_ptr != NULL )
		{
			Clear();
			delete m_ptr;
			m_ptr = NULL;
			m_len = 0;
		}
	}

	void Clear()
	{
		if ( m_ptr == NULL )
		{
			assert( m_len == 0  );
			return;
		}

		memset( m_ptr, 0, m_len * sizeof( T ) );
	}

	T* Allocate( int len )
	{
		if ( len <= m_len )
		{
			Clear();
			return m_ptr;
		}

		Free();

		m_ptr = new T[ len ];

		if ( m_ptr != NULL )
			m_len = len;
		else
		{
			assert( false );
		}

		Clear();

		return m_ptr;
	}

	cAutoDeallocate( int len = 0 )
	{
		m_ptr = NULL;
		m_len = 0;

		if ( len != 0 )
			Allocate( len );
	}

	~cAutoDeallocate()
	{
		Free();
	}

	T *AllocateAndCopy( const T* ptr, int len = -1 )
	{
		if ( len == -1 )
			while( ptr[++len] != 0 ) {}

		Allocate( len + 1 );

		for( int i = len; i--; )
		{
			m_ptr[i] = ptr[i];
		}

		return m_ptr;
	}

	T *Add( const T* ptr, int len = -1 )
	{
		if ( len == -1 )
			while( ptr[++len] != 0 ) {}

		cAutoDeallocate < T > newT( m_len + len );

		int i = 0;

		for( ; i < m_len; i++ )
			newT.m_ptr[i] = m_ptr[i];

		for( ; i < newT.m_len; i++ )
		{
			if ( ptr != NULL )
				newT.m_ptr[i] = ptr[ i - m_len ];
		}

		// swap newT with this.

		swapValues( this, newT );

		return m_ptr;
	}
};

#endif // #ifndef _AutoDeallocate_h_
