#pragma once

#include "UnicodeFuncs.h"
#include "cDll.h"
#include "cWCS.h"

#include <winnt.h>
#include <Accctrl.h>
#include <Aclapi.h>
#include <WinSvc.h>

class CAdvAPIDll : public CDll
{
	bool m_loaded;

	bool checkFunction( void *ptr, bool showError = true );

	bool checkFunction_FL( const char *function, int line, void *ptr, bool showError = true );

public:

	int m_intErr;
	DWORD m_dwLastErr;

	// Do not use the in global constructor, and avoid global contruction at all.
	static __declspec(noinline) CAdvAPIDll &getStatic()
	{
		static cSysCriticalSection cs;
		SYS_CRITICAL_SECTION_BLOCK( &cs );

		static CAdvAPIDll dll;
		return dll;
	}

	SC_HANDLE m_hScManager;

	CAdvAPIDll( bool loadNow = false, bool showError = false );
	~CAdvAPIDll();

	bool Load( bool showError );

	static bool CheckAdministratorPrivileges( bool bForced = false );
	static bool StopService( bool bWait, const wchar_t *serviceName, DWORD maxWaitInMilliSec = INFINITE );

	LONG RegDeleteKeyEx( HKEY hKey, LPCSTR lpSubKey, REGSAM samDesired, DWORD Reserved = 0 );
	LONG RegDeleteKeyExW( HKEY hKey, LPCWSTR lpSubKey, REGSAM samDesired, DWORD Reserved = 0 );

	BOOL LookupPrivilegeValue( IN LPCSTR lpSystemName, IN LPCSTR lpName, OUT PLUID lpLuid );
	BOOL LookupPrivilegeValue( IN LPCWSTR lpSystemName, IN LPCWSTR lpName, OUT PLUID lpLuid );

	BOOL LogonUser( 
		IN LPCSTR lpszUsername,
		IN LPCSTR lpszDomain,
		IN LPCSTR lpszPassword,
		IN DWORD dwLogonType,
		IN DWORD dwLogonProvider,
		OUT PHANDLE phToken
	);

	BOOL LogonUser( 
		IN LPCWSTR lpszUsername,
		IN LPCWSTR lpszDomain,
		IN LPCWSTR lpszPassword,
		IN DWORD dwLogonType,
		IN DWORD dwLogonProvider,
		OUT PHANDLE phToken
	);


	BOOL StopDriver( IN LPCSTR DriverName );
	BOOL RemoveDriver( IN LPCSTR DriverName );

	BOOL RemoveDriver( IN LPCWSTR DriverName );
	BOOL StartDriver( IN LPCWSTR serviceName );
	BOOL StopDriver( IN LPCWSTR DriverName );

	BOOL PauseDriver	( IN LPCWSTR DriverName )	{	return ControlService( DriverName, SERVICE_CONTROL_PAUSE ); }
	BOOL ContinueDriver	( IN LPCWSTR DriverName )	{	return ControlService( DriverName, SERVICE_CONTROL_CONTINUE ); }

	BOOL ControlService( IN LPCWSTR DriverName, DWORD dwControl, DWORD *dwErr = NULL );

	//
	//	The realisation uses setupAPI. ( 10.09.08 )
	//
	bool UninstallDriver( const wchar_t *serviceName );
	bool UninstallDriverInternal( const wchar_t *serviceName );

	SC_HANDLE OpenSCManager( const wchar_t *lpMachineName, const wchar_t* lpDatabaseName, DWORD   dwDesiredAccess );
	SC_HANDLE OpenSCManager( const char *lpMachineName, const char* lpDatabaseName, DWORD   dwDesiredAccess );

	BOOL CloseServiceHandle( SC_HANDLE   hSCObject );

	SC_HANDLE CreateService(
			SC_HANDLE    hSCManager,
			LPCWSTR     lpServiceName,
			LPCWSTR     lpDisplayName,
			DWORD        dwDesiredAccess,
			DWORD        dwServiceType,
			DWORD        dwStartType,
			DWORD        dwErrorControl,
			LPCWSTR     lpBinaryPathName,
			LPCWSTR     lpLoadOrderGroup,
			LPDWORD      lpdwTagId,
			LPCWSTR     lpDependencies,
			LPCWSTR     lpServiceStartName,
			LPCWSTR     lpPassword
			);

	SC_HANDLE CreateService(
			SC_HANDLE    hSCManager,
			LPCSTR     lpServiceName,
			LPCSTR     lpDisplayName,
			DWORD        dwDesiredAccess,
			DWORD        dwServiceType,
			DWORD        dwStartType,
			DWORD        dwErrorControl,
			LPCSTR     lpBinaryPathName,
			LPCSTR     lpLoadOrderGroup,
			LPDWORD      lpdwTagId,
			LPCSTR     lpDependencies,
			LPCSTR     lpServiceStartName,
			LPCSTR     lpPassword
			);


	BOOL StartService( SC_HANDLE hService, DWORD dwNumServiceArgs = 0, LPCSTR *lpServiceArgVectors = NULL );
	BOOL StartService( SC_HANDLE hService, DWORD dwNumServiceArgs, LPCWSTR *lpServiceArgVectors );

	SC_HANDLE OpenService( SC_HANDLE hSCManager, LPCSTR lpServiceName, DWORD dwDesiredAccess );
	SC_HANDLE OpenService( SC_HANDLE hSCManager, LPCWSTR lpServiceName, DWORD dwDesiredAccess );

	BOOL DeleteService( SC_HANDLE hService );

	BOOL ControlService( SC_HANDLE hService, DWORD dwControl, LPSERVICE_STATUS lpServiceStatus );

	BOOL AllocateAndInitializeSid(
			IN PSID_IDENTIFIER_AUTHORITY pIdentifierAuthority,
			IN BYTE nSubAuthorityCount,
			IN DWORD nSubAuthority0,
			IN DWORD nSubAuthority1,
			IN DWORD nSubAuthority2,
			IN DWORD nSubAuthority3,
			IN DWORD nSubAuthority4,
			IN DWORD nSubAuthority5,
			IN DWORD nSubAuthority6,
			IN DWORD nSubAuthority7,
			OUT PSID *pSid
	);

	DWORD SetEntriesInAcl( IN ULONG cCountOfExplicitEntries, IN PEXPLICIT_ACCESS_A pListOfExplicitEntries, IN PACL OldAcl, OUT PACL *NewAcl	);
	DWORD SetEntriesInAcl( IN ULONG cCountOfExplicitEntries, IN PEXPLICIT_ACCESS_W pListOfExplicitEntries, IN PACL OldAcl, OUT PACL *NewAcl	);

	BOOL InitializeSecurityDescriptor( OUT PSECURITY_DESCRIPTOR pSecurityDescriptor, IN DWORD dwRevision );

	BOOL SetSecurityDescriptorDacl( IN OUT PSECURITY_DESCRIPTOR pSecurityDescriptor, IN BOOL bDaclPresent, IN PACL pDacl, IN BOOL bDaclDefaulted );

	BOOL RevertToSelf();
	BOOL CreateProcessWithLogonW(
      LPCWSTR lpUsername,
      LPCWSTR lpDomain,
      LPCWSTR lpPassword,
      DWORD   dwLogonFlags,
      LPCWSTR lpApplicationName,
      LPWSTR lpCommandLine,
      DWORD dwCreationFlags,
      LPVOID lpEnvironment,
      LPCWSTR lpCurrentDirectory,
      LPSTARTUPINFOW lpStartupInfo,
      LPPROCESS_INFORMATION lpProcessInformation
      );

	BOOL ImpersonateLoggedOnUser( IN HANDLE  hToken );
	BOOL ImpersonateNamedPipeClient( IN HANDLE hNamedPipe );

	BOOL ChangeServiceConfig( SC_HANDLE hService, DWORD dwServiceType, DWORD dwStartType, DWORD dwErrorControl, LPCWSTR lpBinaryPathName, LPCWSTR lpLoadOrderGroup, LPDWORD lpdwTagId, LPCWSTR lpDependencies, LPCWSTR lpServiceStartName, LPCWSTR lpPassword, LPCWSTR lpDisplayName );
	BOOL ChangeServiceConfig( SC_HANDLE hService, LPQUERY_SERVICE_CONFIGW lpServiceConfig );
	BOOL QueryServiceConfig(  SC_HANDLE hService, LPQUERY_SERVICE_CONFIGW lpServiceConfig, DWORD cbBufSize, LPDWORD pcbBytesNeeded );

	BOOL LogonUserEx( LPCWSTR lpszUsername, LPCWSTR lpszDomain, LPCWSTR lpszPassword, DWORD dwLogonType, DWORD dwLogonProvider, PHANDLE phToken, PSID  *ppLogonSid, PVOID *ppProfileBuffer, LPDWORD pdwProfileLength, PQUOTA_LIMITS pQuotaLimits );
	BOOL OpenProcessToken( IN HANDLE ProcessHandle, IN DWORD DesiredAccess, OUT PHANDLE TokenHandle );
	BOOL GetTokenInformation( IN HANDLE TokenHandle, IN TOKEN_INFORMATION_CLASS TokenInformationClass, OUT LPVOID TokenInformation, IN DWORD TokenInformationLength, OUT PDWORD ReturnLength );

	BOOL QueryServiceStatus( SC_HANDLE hService, LPSERVICE_STATUS lpServiceStatus );

	BOOL ConvertSidToStringSid( IN  PSID Sid, OUT LPWSTR  *StringSid );
	BOOL ConvertStringSidToSid( LPCWSTR StringSid, PSID* Sid );
	PVOID FreeSid( PSID pSid );

	//
	//	GetSidByProcessId
	//		StringSid has to be deallocated b LocalFree.
	//
	bool GetSidByProcessId( LPWSTR  *StringSid, DWORD pid );

	//
	//	GetSidByToken
	//		If success use LocalFree to free StringSid.
	//
	bool GetSidByToken( LPWSTR *StringSid, HANDLE hToken );
	const wchar_t *GetSidByToken( cWCS &wcsSid, HANDLE hToken )
	{
		wchar_t *szSid = NULL;
		if ( !GetSidByToken( &szSid, hToken ) )
			return NULL;

		wcsSid.Set( szSid );
		LocalFree( szSid );
		return wcsSid.m_szBody;	
	}


	TOKEN_USER *GetTokenUser( HANDLE hToken );

	DWORD ServiceGetStatus( const wchar_t *sService, const wchar_t *sMachine = NULL );

	DWORD SetNamedSecurityInfo( LPCWSTR pObjectName, SE_OBJECT_TYPE ObjectType, SECURITY_INFORMATION SecurityInfo, PSID psidOwner, PSID psidGroup, PACL pDacl, PACL pSacl );

	static bool SetPrivilege( const wchar_t *lpszPrivilege, bool bEnablePrivilege );
	static bool SetPrivilege( const char *lpszPrivilege, bool bEnablePrivilege )
	{ return SetPrivilege( cWCStr( lpszPrivilege ).m_szBody, bEnablePrivilege );}

	bool SetPrivilege( HANDLE hToken, const wchar_t *lpszPrivilege, bool bEnablePrivilege );
	bool SetPrivilege( HANDLE hToken, const char *lpszPrivilege, bool bEnablePrivilege )
	{ return SetPrivilege( hToken, cWCStr( lpszPrivilege ).m_szBody, bEnablePrivilege );}

	bool AdjustTokenPrivileges( HANDLE TokenHandle, BOOL DisableAllPrivileges, PTOKEN_PRIVILEGES NewState, DWORD BufferLength, PTOKEN_PRIVILEGES PreviousState,  PDWORD ReturnLength  );

private:
	typedef WINADVAPI BOOL WINAPI TypeQueryServiceStatus( SC_HANDLE hService, LPSERVICE_STATUS lpServiceStatus );
	typedef WINADVAPI SC_HANDLE WINAPI TypeOpenSCManagerA( LPCSTR lpMachineName, LPCSTR lpDatabaseName, DWORD   dwDesiredAccess );
	typedef WINADVAPI SC_HANDLE WINAPI TypeCreateServiceA(
		SC_HANDLE    hSCManager,
		LPCSTR     lpServiceName,
		LPCSTR     lpDisplayName,
		DWORD        dwDesiredAccess,
		DWORD        dwServiceType,
		DWORD        dwStartType,
		DWORD        dwErrorControl,
		LPCSTR     lpBinaryPathName,
		LPCSTR     lpLoadOrderGroup,
		LPDWORD      lpdwTagId,
		LPCSTR     lpDependencies,
		LPCSTR     lpServiceStartName,
		LPCSTR     lpPassword
		);
	typedef WINADVAPI BOOL WINAPI TypeStartServiceA( SC_HANDLE hService, DWORD dwNumServiceArgs, LPCSTR *lpServiceArgVectors );
	typedef WINADVAPI SC_HANDLE WINAPI TypeOpenServiceA( SC_HANDLE hSCManager, LPCSTR lpServiceName, DWORD dwDesiredAccess );
	typedef WINADVAPI DWORD WINAPI TypeSetEntriesInAclA(
		IN  ULONG               cCountOfExplicitEntries,
		IN  PEXPLICIT_ACCESS_A  pListOfExplicitEntries,
		IN  PACL                OldAcl,
		OUT PACL              * NewAcl
		);


	typedef WINADVAPI BOOL WINAPI TypeControlService( SC_HANDLE hService, DWORD dwControl, LPSERVICE_STATUS lpServiceStatus );
	typedef WINADVAPI BOOL WINAPI TypeDeleteService( SC_HANDLE hService );
	typedef WINADVAPI BOOL WINAPI TypeCloseServiceHandle( SC_HANDLE   hSCObject );
	typedef WINADVAPI BOOL WINAPI TypeAllocateAndInitializeSid (
		IN PSID_IDENTIFIER_AUTHORITY pIdentifierAuthority,
		IN BYTE nSubAuthorityCount,
		IN DWORD nSubAuthority0,
		IN DWORD nSubAuthority1,
		IN DWORD nSubAuthority2,
		IN DWORD nSubAuthority3,
		IN DWORD nSubAuthority4,
		IN DWORD nSubAuthority5,
		IN DWORD nSubAuthority6,
		IN DWORD nSubAuthority7,
		OUT PSID *pSid
		);

	typedef WINADVAPI BOOL WINAPI TypeInitializeSecurityDescriptor (   OUT PSECURITY_DESCRIPTOR pSecurityDescriptor, IN DWORD dwRevision );
	typedef WINADVAPI BOOL WINAPI TypeSetSecurityDescriptorDacl (
		IN OUT PSECURITY_DESCRIPTOR pSecurityDescriptor,
		IN BOOL bDaclPresent,
		IN PACL pDacl,
		IN BOOL bDaclDefaulted
		);

	typedef WINADVAPI LONG WINAPI TypeRegDeleteKeyEx(
	HKEY hKey,
	LPCSTR lpSubKey,
	REGSAM samDesired,
	DWORD Reserved
	);

	typedef WINADVAPI LONG WINAPI TypeRegDeleteKeyExW(
	HKEY hKey,
	LPCWSTR lpSubKey,
	REGSAM samDesired,
	DWORD Reserved
	);

	typedef WINADVAPI SC_HANDLE WINAPI TypeOpenSCManagerW( LPCWSTR lpMachineName, LPCWSTR lpDatabaseName, DWORD   dwDesiredAccess );
	typedef WINADVAPI SC_HANDLE WINAPI TypeCreateServiceW(
		SC_HANDLE    hSCManager,
		LPCWSTR     lpServiceName,
		LPCWSTR     lpDisplayName,
		DWORD        dwDesiredAccess,
		DWORD        dwServiceType,
		DWORD        dwStartType,
		DWORD        dwErrorControl,
		LPCWSTR     lpBinaryPathName,
		LPCWSTR     lpLoadOrderGroup,
		LPDWORD      lpdwTagId,
		LPCWSTR     lpDependencies,
		LPCWSTR     lpServiceStartName,
		LPCWSTR     lpPassword
		);
	typedef WINADVAPI BOOL WINAPI TypeStartServiceW( SC_HANDLE hService, DWORD dwNumServiceArgs, LPCWSTR *lpServiceArgVectors );
	typedef WINADVAPI SC_HANDLE WINAPI TypeOpenServiceW( SC_HANDLE hSCManager, LPCWSTR lpServiceName, DWORD dwDesiredAccess );
	typedef WINADVAPI DWORD WINAPI TypeSetEntriesInAclW(
		IN  ULONG               cCountOfExplicitEntries,
		IN  PEXPLICIT_ACCESS_W  pListOfExplicitEntries,
		IN  PACL                OldAcl,
		OUT PACL              * NewAcl
		);

	typedef WINADVAPI BOOL WINAPI TypeCloseServiceHandle( SC_HANDLE   hSCObject );
	typedef WINADVAPI BOOL WINAPI TypeAllocateAndInitializeSid (
		IN PSID_IDENTIFIER_AUTHORITY pIdentifierAuthority,
		IN BYTE nSubAuthorityCount,
		IN DWORD nSubAuthority0,
		IN DWORD nSubAuthority1,
		IN DWORD nSubAuthority2,
		IN DWORD nSubAuthority3,
		IN DWORD nSubAuthority4,
		IN DWORD nSubAuthority5,
		IN DWORD nSubAuthority6,
		IN DWORD nSubAuthority7,
		OUT PSID *pSid
		);

	typedef WINADVAPI BOOL WINAPI TypeOpenProcessToken( IN HANDLE ProcessHandle, IN DWORD DesiredAccess, OUT PHANDLE TokenHandle );
	typedef WINADVAPI BOOL WINAPI TypeGetTokenInformation( IN HANDLE TokenHandle, IN TOKEN_INFORMATION_CLASS TokenInformationClass, OUT LPVOID TokenInformation, IN DWORD TokenInformationLength, OUT PDWORD ReturnLength );

	typedef WINADVAPI BOOL WINAPI TypeLogonUserEx( LPCWSTR lpszUsername, LPCWSTR lpszDomain, LPCWSTR lpszPassword, DWORD dwLogonType, DWORD dwLogonProvider, PHANDLE phToken, PSID  *ppLogonSid, PVOID *ppProfileBuffer, LPDWORD pdwProfileLength, PQUOTA_LIMITS pQuotaLimits );
	TypeLogonUserEx *procLogonUserEx;

	TypeOpenProcessToken *procOpenProcessToken;
	TypeGetTokenInformation *procGetTokenInformation;

	TypeQueryServiceStatus *procQueryServiceStatus;
	TypeOpenSCManagerA *procOpenSCManager;
	TypeCreateServiceA *procCreateService;
	TypeStartServiceA *procStartService;
	TypeOpenServiceA *procOpenService;
	TypeSetEntriesInAclA *procSetEntriesInAcl;

	TypeOpenSCManagerW *procOpenSCManagerW;
	TypeCreateServiceW *procCreateServiceW;
	TypeStartServiceW *procStartServiceW;
	TypeOpenServiceW *procOpenServiceW;
	TypeSetEntriesInAclW *procSetEntriesInAclW;

	TypeCloseServiceHandle *procCloseServiceHandle;
	TypeDeleteService *procDeleteService;
	TypeControlService *procControlService;

	TypeAllocateAndInitializeSid *procAllocateAndInitializeSid;
	TypeInitializeSecurityDescriptor *procInitializeSecurityDescriptor;
	TypeSetSecurityDescriptorDacl *procSetSecurityDescriptorDacl;
	
	TypeRegDeleteKeyEx *procRegDeleteKeyEx;
	TypeRegDeleteKeyExW *procRegDeleteKeyExW;

	typedef WINADVAPI BOOL WINAPI TypeLookupPrivilegeValueW( IN LPCWSTR lpSystemName, IN LPCWSTR lpName, OUT PLUID lpLuid );
	TypeLookupPrivilegeValueW *procLookupPrivilegeValueW;

	typedef WINADVAPI BOOL WINAPI TypeLookupPrivilegeValueA( IN LPCSTR lpSystemName, IN LPCSTR lpName, OUT PLUID lpLuid );
	TypeLookupPrivilegeValueA *procLookupPrivilegeValue;

	typedef WINADVAPI BOOL WINAPI TypeRevertToSelf( VOID );
	TypeRevertToSelf *procRevertToSelf;

	typedef WINADVAPI BOOL WINAPI TypeLogonUserA( IN LPCSTR lpszUsername, IN LPCSTR lpszDomain, IN LPCSTR lpszPassword, IN DWORD dwLogonType, IN DWORD dwLogonProvider, OUT PHANDLE phToken );
	typedef WINADVAPI BOOL WINAPI TypeLogonUserW( IN LPCWSTR lpszUsername, IN LPCWSTR lpszDomain, IN LPCWSTR lpszPassword, IN DWORD dwLogonType, IN DWORD dwLogonProvider, OUT PHANDLE phToken );
	TypeLogonUserA *procLogonUserA;
	TypeLogonUserW *procLogonUserW;

	typedef WINADVAPI BOOL WINAPI TypeCreateProcessWithLogonW(
      LPCWSTR lpUsername,
      LPCWSTR lpDomain,
      LPCWSTR lpPassword,
      DWORD   dwLogonFlags,
      LPCWSTR lpApplicationName,
      LPWSTR lpCommandLine,
      DWORD dwCreationFlags,
      LPVOID lpEnvironment,
      LPCWSTR lpCurrentDirectory,
      LPSTARTUPINFOW lpStartupInfo,
      LPPROCESS_INFORMATION lpProcessInformation
      );
	TypeCreateProcessWithLogonW *procCreateProcessWithLogonW;

	typedef WINADVAPI BOOL WINAPI TypeImpersonateLoggedOnUser( IN HANDLE  hToken );
	TypeImpersonateLoggedOnUser *procImpersonateLoggedOnUser;

	typedef WINADVAPI BOOL WINAPI TypeImpersonateNamedPipeClient( IN HANDLE hNamedPipe );
	TypeImpersonateNamedPipeClient *procImpersonateNamedPipeClient;

	typedef WINADVAPI BOOL WINAPI TypeQueryServiceConfig( SC_HANDLE hService, LPQUERY_SERVICE_CONFIGW lpServiceConfig, DWORD cbBufSize, LPDWORD pcbBytesNeeded );
	TypeQueryServiceConfig *procQueryServiceConfig;

	typedef WINADVAPI BOOL WINAPI TypeChangeServiceConfig( SC_HANDLE hService, DWORD dwServiceType, DWORD dwStartType, DWORD dwErrorControl, LPCWSTR lpBinaryPathName, LPCWSTR lpLoadOrderGroup, LPDWORD lpdwTagId, LPCWSTR lpDependencies, LPCWSTR lpServiceStartName, LPCWSTR lpPassword, LPCWSTR lpDisplayName );
	TypeChangeServiceConfig *procChangeServiceConfig;

	bool GetProcAddress( const char *procName, void **ppProc );
};

//
//	CreateWritablePipe
//		creates named pipe writable by 'Everyone' group
//		so any loggen user can communicate with service
//
extern HANDLE CreateWritablePipe(
		const wchar_t *szPipeName,
		DWORD openMode = PIPE_ACCESS_DUPLEX | FILE_FLAG_OVERLAPPED,
		DWORD pipeMode = PIPE_TYPE_BYTE,
		DWORD instances = 1,
		DWORD outBuffSize = sizeof( DWORD ),
		DWORD inBuffSize = sizeof(DWORD),
		DWORD defTimeout = 1000
	);


bool GetSidByProcessId( cWCStr *wcsSid, DWORD pid );

//
//	Return sid of impersonated token or sid of current process.
//
bool GetCurrentThreadSid( cWCStr *wcsSid );

//
//	CreateSecureSemaphore
//		Is same as CreateSemaphore but it creates semaphore acceessable for Everyone group.
//
extern HANDLE CreateSecureSemaphore(
    LONG lInitialCount,
    LONG lMaximumCount,
    LPCWSTR lpName,
	int *pErr = NULL );

//
//	CreateSecureMutex
//		Is same as CreateMutexW but it creates semaphore acceessable for Everyone group.
//
extern HANDLE CreateSecureMutex(
	BOOL bInitialOwner,
	LPCWSTR lpName,
	int *pErr = NULL );

//
//	SetFileSizeAndValidData
//		Set required file size.
//		Uses SetFileValidData function to sets the logical end of a file,
//		so the file will contain all garbage from disk instead of zero ( if we extend the file ).
//		Note, the validation required administrator privileges to assign SE_MANAGE_VOLUME_NAME.
//
extern bool SetFileSizeAndValidData( const wchar_t *filePath, DWORDLONG fileSize, bool bFlushBuffers = false );
extern bool SetFileValidDataT( HANDLE hFile, DWORDLONG fileSize );

inline bool IsServiceStarted( const wchar_t *szService )
{	return CAdvAPIDll::getStatic().ServiceGetStatus( szService ) == SERVICE_RUNNING; }

inline TOKEN_USER *GetTokenUser( HANDLE hToken ) 
{	return CAdvAPIDll::getStatic().GetTokenUser( hToken ); }

inline const wchar_t *ServiceGetStatusStr( const wchar_t *sService, const wchar_t *sMachine = NULL )
{
	DWORD dwStat = CAdvAPIDll::getStatic().ServiceGetStatus( sService, sMachine );
	switch ( dwStat )
	{
		case SERVICE_STOPPED			: return L"STOPPED";
		case SERVICE_START_PENDING		: return L"START_PENDING";
		case SERVICE_STOP_PENDING		: return L"STOP_PENDING";
		case SERVICE_RUNNING			: return L"RUNNING";
		case SERVICE_CONTINUE_PENDING	: return L"CONTINUE_PENDING";
		case SERVICE_PAUSE_PENDING		: return L"PAUSE_PENDING";
		case SERVICE_PAUSED				: return L"PAUSED";
		default							: return L"Undefined";
	}
}

class cImpersonateUserToken
{
	bool m_bImpersonated;

public:

	cImpersonateUserToken( HANDLE hToken )
	{
		m_bImpersonated = false;

		if ( hToken == NULL )
			return;

		if ( !CAdvAPIDll::getStatic().ImpersonateLoggedOnUser( hToken ) )
			ODS_fls( _FormatT( L"ImpersonateLoggedOnUser error %s", _GetSystemError( -1 ) ) );
		else
			m_bImpersonated = true;
	}

	~cImpersonateUserToken()
	{
		if ( m_bImpersonated )
			CAdvAPIDll::getStatic().RevertToSelf();
	}
};

//
//	Disables 'Program Compatibility Assistant Service'
//	It allows to skip annoing message 'Program might not have installed correctly'
//	Call it just before exit.
//
void RestartPCA();
