#ifndef _AutoDeallocate_h_
#define _AutoDeallocate_h_

#include <assert.h>
#include <memory.h>
#include "ODS.h"
#include <exception>

template < class T > bool swapValues( T *v0, T *v1 )
{
	char *pT = new char[ sizeof( T ) ];

	if ( pT == NULL )
	{
		ODS_fls( L"Assertion!" );
		return false;
	}

	memcpy( pT, v0, sizeof( T ) );
	memcpy( v0, v1, sizeof( T ) );
	memcpy( v1, pT, sizeof( T ) );

	delete [] pT;

	return true;
}

//
//	class cADBlock
//	 Auto Deallocated Block
//
template < class T > class cADBlock
{

public:

	T& getAt( int pos )
	{
		if ( pos >= m_Q )
		{
			ODS_fls( L"Assertion! pos >= m_Q" );
			static T errObj;
			return errObj;
		}

		return  m_T[pos];
	}

	T& getNext( int &cur )
	{
		if ( cur >= m_Q )
		{
			ODS_fls( L"Assertion! pos >= m_Q" );
			static T errObj;
			return errObj;
		}

		T &t = getAt( cur );
		if ( t != NULL )
			cur++;

		return  t;
	}

	T& getFirst( int &cur )
	{
		cur = 0;
		return getNext( cur );
	}

	bool Remove( T item )
	{
		for( int i = m_Q; i--; )
		{
			if ( m_T[i] == item )
			{
				RemoveByNumber( i );
				return true;
			}
		}

		return false;
	}

	T *m_T;
	int m_Q;

	bool m_setZero;

	bool isEmpty()
	{
		return m_Q == 0;
	}

	bool Add( T in )
	{
		return Add( &in, 1 );
	}

	bool Add( T *inT, int size = 1 )
	{
		if ( size <= 0 )
			return false;

		T *newT = allocate( m_Q + size, m_setZero );

		if ( newT == NULL )
			return false;

		int q = 0;
		for( ; q < m_Q; q++ )
			newT[q] = m_T[q];

		Free();

		m_T = newT;
		m_Q = q;

		for( q = 0; q < size; q++,m_Q++ )
		{
			if ( inT != NULL )
				m_T[m_Q] = inT[q];
			else
			if ( m_setZero )
				memset( m_T + m_Q, 0, sizeof( T ) );
		}

		return true;
	}

	T *reallocate( int size, bool setZero = true )
	{
		if ( size < m_Q )
		{
			memset( m_T + size, 0, sizeof(T) * ( m_Q - size ) );
			m_Q = size;
			return m_T;
		}

		T *pT = allocate( size, setZero );

		if ( pT == NULL )
		{
			Free();
			return NULL;
		}

		memcpy( pT, m_T, sizeof(T) * m_Q );
		m_T = pT;
		m_Q = size;

		return m_T;
	}

	static T *allocate( int size, bool setZero )
	{
		if ( size == 0 )
			return NULL;

		T *pT;

		try
		{
			pT = new T [ size ];
		}
// DDK build report: error C2039: 'bad_alloc' : is not a member of 'std'
		//catch( std::bad_alloc& ba )
		//{
		//	ba.what();
		//	ODSF( L"%s:%d Assertion! BAD_ALLOC sizeof(T) = %d, len = %d", _CRT_WIDE( __FUNCTION__ ), __LINE__, sizeof(T), size );
		//	pT = NULL;
		//}
		catch( ... )
		{
			ODSF( L"%s:%d Assertion! UNHANDLED_EXCEPTION sizeof(T) = %d, len = %d" _CRT_WIDE( __FUNCTION__ ), __LINE__, sizeof(T), size );
			pT = NULL;
		}

		if ( ( pT != NULL ) && setZero )
			memset( pT, 0, sizeof( T ) * size );

		return pT;
	}

	void Free()
	{
		while( m_Q )
			RemoveByNumber( m_Q - 1 );

		if ( m_T != NULL )
		{
			//if ( m_setZero )
			//	memset( m_T, 0, sizeof( T ) * m_Q );

			delete [] m_T;
		}

		m_T = NULL;
		m_Q = 0;
	}

	cADBlock( int Q = 1, bool setZero = true )
	{
		memset( this, 0, sizeof( *this ) );

		m_setZero = setZero;

		m_T = allocate( Q, setZero );

		if ( m_T != NULL )
			m_Q = Q;
	}

	~cADBlock()
	{
		Free();
	}

	int Find( T item )
	{
		int i = m_Q;

		while( i-- )
		{
			if ( m_T[i] == item )
				break;
		}

		return i;
	}
	
	int RemoveByNumber( int i )
	{
		if ( i >= m_Q )
			return -1;

		for ( ; i < m_Q - 1; i++ )
		{
			swapValues( m_T + i, m_T + i + 1 );
		}

		if ( m_setZero )
			memset( m_T + i, 0, sizeof( T ) );

		m_Q--;

		// assert( m_Q == i );

		return i;
	}

	void DelegateTable( cADBlock < T > &adb )
	{
		adb.Free();
		adb.m_T = m_T;
		adb.m_Q = m_Q;
		m_Q = 0;
		m_T = NULL;
	}

	T *delegetePtr()
	{
		T *ptr = m_T;
		m_Q = 0;
		m_T = NULL;

		return ptr;
	}

	int Compare( cADBlock < T > &adb )
	{
		if ( adb.m_Q != m_Q )
			return adb.m_Q - m_Q;

		for( int i = 0; i < m_Q; i++ )
		{
			int diff = adb.m_T[i] - m_T[i];
			if ( diff != 0 )
				return diff;
		}

		return 0;
	}
};

//
//	Resizeable table of T elements.
//	The class T element has to contain Free function.
//
template < class T > class cADTable : public cADBlock < T >
{
public:
	cADTable( int size = 0, bool setZero = false ) 
	    : cADBlock < T > ( size, setZero )
	{
	}

	T * Add( T *in, int size = 1 )
	{
		if ( size == 0 )
			return NULL;

		if ( !Resize( this->m_Q + size ) )
			return NULL;

		if ( in != NULL )
			memcpy( this->m_T + this->m_Q - size, in, sizeof ( T ) * size );

		return this->m_T + this->m_Q - size;
	}

	T * AddCopy( T *in, int size = 1 )
	{
		T *added = Add( NULL, size );

		if ( added == NULL )
			return NULL;

		if ( in != NULL )
		{
			for( int i = 0; i < size; i++ )
			{
				added[i].CopyFrom( in + i );
			}
		}

		return added;
	}

	bool Resize( int size )
	{
		while( size < this->m_Q )
		{
			this->m_T[ --(this->m_Q) ].Free(); // Free content...

			if ( this->m_setZero )
				memset( this->m_T + this->m_Q, 0, sizeof( T ) );
		}

		if ( this->m_Q == 0 ) 
			this->Free();

		if ( size == this->m_Q )
			return true;

		T *pT = allocate( size, this->m_setZero );

		if ( pT == NULL )
			return false;

		if ( this->m_T != NULL )
		{
			for( int i = 0; i < this->m_Q; i++ )
				swapValues( this->m_T + i, pT + i );

			this->Free();
		}
		
		this->m_T = pT;
		this->m_Q = size;

		return true;
	}

	int Remove( int i )
	{
		if ( i >= m_Q )
			return -1;

		if ( !m_setZero )
			m_T[i].Free();

		return cADBlock::RemoveByNumber( i );
	}
};


//
//	cADArray did not call Free for every element instead of cADTable does.
//

template < class T > class cADArray : public cADBlock < T >
{
public:
	cADArray( int size ) : cADBlock < T > ( size, true ) {}
};

template < class T > class cADPtr : public cADBlock < T * >
{
public:
	cADPtr() : cADBlock < T * > ( 0, true ) {}

	int Remove( int i )
	{
		return RemoveByNumber( i );
	}

	int RemoveByNumber( int i )
	{
		if ( i >= m_Q )
			return -1;

		T *ptr = m_T[i];

		int res = cADBlock::RemoveByNumber( i );

		if ( ptr != NULL )
			delete ptr;

		return res;
	}

	void Free()
	{
		while( m_Q )
			RemoveByNumber( m_Q - 1 );

		if ( m_T != NULL )
			delete [] m_T;

		m_T = NULL;
		m_Q = 0;
	}
	
	~cADPtr()
	{
		Free();
	}
};


template < class T > class cAutoDeallocate
{
public:
	int m_len;
	T *m_ptr;
	bool m_bUseVirtualAlloc;

	void Free()
	{
		if ( m_ptr != NULL )
		{
			Clear();
#ifdef _WIN32
			if ( m_bUseVirtualAlloc )
				VirtualFree( m_ptr, 0, MEM_RELEASE );
			else
#endif
				delete [] m_ptr;

			m_ptr = NULL;
			m_len = 0;
		}
	}

	void Clear()
	{
		if ( m_ptr == NULL )
		{
			assert( m_len == 0  );
			return;
		}

		memset( m_ptr, 0, m_len * sizeof( T ) );
	}

	T* Allocate( int len )
	{
		if ( len <= m_len )
		{
			Clear();
			return m_ptr;
		}

		Free();

		try
		{
#ifdef _WIN32
			if ( m_bUseVirtualAlloc )
				m_ptr = (T*)VirtualAlloc( NULL, sizeof( T ) * len, MEM_COMMIT , PAGE_READWRITE );
			else
#endif
				m_ptr = new T[ len ];
		}
// DDK build report: error C2039: 'bad_alloc' : is not a member of 'std'
		//catch( std::bad_alloc& ba )
		//{
		//	ba.what();

		//	ODSF( L"%s:%d Assertion! BAD_ALLOC sizeof(T) = %d, len = %d", _CRT_WIDE( __FUNCTION__ ), __LINE__, sizeof(T), len );
		//	m_ptr = NULL;
		//}
		catch( ... )
		{
			ODSF( L"%s:%d Assertion! UNHANDLED_EXCEPTION sizeof(T) = %d, len = %d", _CRT_WIDE( __FUNCTION__ ), __LINE__, sizeof(T), len );
			m_ptr = NULL;
		}

		if ( m_ptr != NULL )
			m_len = len;
		else
		{
			ODS_fls( L"Assertion!" );
		}

		if ( !m_bUseVirtualAlloc ) // VirtuallAlloc automatically clear memory by MEM_COMMIT flag
			Clear();

		return m_ptr;
	}

	cAutoDeallocate( int len = 0, bool bUseVirtualAlloc = false )
	{
		m_ptr = NULL;
		m_len = 0;
		m_bUseVirtualAlloc = false;

#ifdef _WIN32
		m_bUseVirtualAlloc = bUseVirtualAlloc;
#else
		assert( bUseVirtualAlloc == false );
#endif

		if ( len != 0 )
			Allocate( len );
	}

	~cAutoDeallocate()
	{
		Free();
	}

	T *AllocateAndCopy( const T* ptr, int len = -1 )
	{
		if ( len == -1 )
			while( ptr[++len] != 0 ) {}

		Allocate( len + 1 );

		for( int i = len; i--; )
		{
			m_ptr[i] = ptr[i];
		}

		return m_ptr;
	}

	T *Add( const T* ptr, int len = -1 )
	{
		if ( len == -1 )
			while( ptr[++len] != 0 ) {}

		cAutoDeallocate < T > newT( m_len + len );

		int i = 0;

		for( ; i < m_len; i++ )
			newT.m_ptr[i] = m_ptr[i];

		for( ; i < newT.m_len; i++ )
		{
			if ( ptr != NULL )
				newT.m_ptr[i] = ptr[ i - m_len ];
		}

		// swap newT with this.

		swapValues( this, &newT );

		//int swp_len = m_len;
		//T *swp_ptr = m_ptr;

		//m_len = newT.m_len;
		//m_ptr = newT.m_ptr;

		//newT.m_len = swp_len;
		//newT.m_ptr = swp_ptr;

		return m_ptr;
	}
};

#endif // #ifndef _AutoDeallocate_h_
