#include "EquStrings.h"
#include <ctype.h>

//
// Functions for string comparation
//           with *,%  symbols .
// Master function EqualStrings(s1,s2) returns true in following case:
//      s1 - word    Example : "qwerty",
//      s2 - pattern Example : "q%e*y*".
//

#define EOS '\0'
#define WEOS L'\0'

static bool All_check( const char *s1, const char *s2)
{
	for(;;)
	{
		if ( EquStrings(s1,s2) ) return true;
		if ( *(s1++) == EOS ) return false;
	}
}

static bool All_check( const wchar_t *s1, const wchar_t *s2)
{
	for(;;)
	{
		if ( EquStrings(s1,s2) ) return true;
		if ( *(s1++) == EOS ) return false;
	}
}

//
//  EquStrings
//
bool equ( char c1, char c2, bool matchCase )
{
	if ( matchCase )
		return c1 == c2;
	else
		return toupper( c1 ) == toupper( c2 );
}

bool EquStrings( const char *s1, const char *s2, bool matchCase )
{ 
	for(;;s1++)
	{
		if ( *s2 == '*' ) return All_check(s1,++s2);
		if ( !equ( *s2, *s1, matchCase ) && (( *s2 != '%' ) || ( *s1 == EOS ))) return false;
//		if ((*s2 != *s1) && (( *s2 != '%' ) || ( *s1 == EOS ))) return false;
		if ( *(s2++) == EOS ) return true;
	}
}

bool EquStrings( const char *s1, const char *s2 )
{ 
	for(;;s1++)
	{
		if ( *s2 == '*' ) return All_check(s1,++s2);
		if ((*s2 != *s1) && (( *s2 != '%' ) || ( *s1 == EOS ))) return false;
		if ( *(s2++) == EOS ) return true;
	}
}

bool EquStrings( const wchar_t *s1, const wchar_t *s2 )
{ 
	for(;;s1++)
	{
		if ( *s2 == L'*' ) return All_check(s1,++s2);
		if ((*s2 != *s1) && (( *s2 != L'%' ) || ( *s1 == WEOS ))) return false;
		if ( *(s2++) == WEOS ) return true;
	}
}

static bool All_check( const wchar_t *s1, const wchar_t *s2, bool matchCase )
{
	for(;;)
	{
		if ( EquStrings(s1,s2,matchCase) ) return true;
		if ( *(s1++) == EOS ) return false;
	}
}

bool equ( wchar_t c1, wchar_t c2, bool matchCase )
{
	if ( matchCase )
		return c1 == c2;
	else
		return towupper( c1 ) == towupper( c2 );
}

bool EquStrings( const wchar_t *s1, const wchar_t *s2, bool matchCase )
{ 
	for(;;s1++)
	{
		if ( *s2 == L'*' ) return All_check(s1,++s2,matchCase);
		if ( !equ( *s2, *s1, matchCase ) && (( *s2 != L'%' ) || ( *s1 == WEOS ))) return false;
		if ( *(s2++) == WEOS ) return true;
	}
}

bool HasStarSymbols( const wchar_t *str )
{
	for( int p = 0; str[ p ] != L'\0'; p++ )
	{
		if ( str[ p ] == L'*' ) return true;
		if ( str[ p ] == L'%' ) return true;
	}

	return false;
}

bool RemoveStarSymbols( wchar_t *str, const wchar_t *startSymbols ) // = NULL
{
	if ( startSymbols == NULL )
		startSymbols = L"*%";

	bool bFound = false;

	int len = wcslen( str );

	for( int p = 0; p < len ; p++ )
	{
		if ( wcschr( startSymbols, str[ p ] ) != NULL )
		{
			bFound = true;

			memmove( str + p, str + p + 1, ( len - p ) * sizeof( wchar_t ) );

			p--;
			len--;
		}
	}

	return bFound;
}

bool EquList( const wchar_t *str, const wchar_t *patternList, bool matchCase )
{
		if ( patternList == NULL )
			return false;

		const wchar_t *s;

		for( s = patternList; ; s = s + wcslen( s ) + 1 )
		{
			if ( *s == L'\0' )
				return false;

			if ( EquStrings( str, s, matchCase ) )
				return true;
		}

		return false;
}

bool IsIncluded( const wchar_t *str, const wchar_t *patternInclude, const wchar_t *patternExclude, bool matchCase )
{
	bool included = true;

	if ( patternInclude != NULL )
		included = EquList( str, patternInclude, matchCase );

	if ( !included )
		return false;
	
	if ( patternExclude != NULL )
	{
		bool excluded = EquList( str, patternExclude, matchCase );

		if ( excluded )
			return false;
	}

	return true;
}
