#include "FileView.h"
#include <io.h>
#include <memory.h>
#include "UnicodeFuncs.h"
#include <assert.h>
#include "ODS.h"
#include "AutoDeallocate.h"

#ifndef _UseFileViewA_
BOOL cFileView::OpenMapping( bool readOnly )
{
	m_err = cfv_errorInOpeningFile;

	if ( m_hFile == INVALID_HANDLE_VALUE )
	{
		ODS_fl;
		return FALSE;
	}

	m_size.QuadPart = GetFileSizeT( m_hFile );

	if ( m_size.QuadPart == (DWORDLONG)-1 )
	{
		ODS_fl;
		return FALSE;
	}

	if ( m_size.QuadPart == 0 )
	{
		m_err = cfv_ok;
		ODS_fl;
		return TRUE;
	}

	if ( ( m_stepSize == 0 ) && ( (SIZE_T)m_size.QuadPart != m_size.QuadPart ) )
	{
		ODS_fls( _FormatT( L"Assertion in MapViewOfFile, file size %s", _GetSizeText( m_size.QuadPart ) ) );
		return FALSE;
	}
	
	m_hMap = CreateFileMapping( m_hFile, NULL, ( readOnly ? PAGE_READONLY : PAGE_READWRITE ), m_size.HighPart, m_size.LowPart, NULL ); // fileName );

	if ( m_hMap == NULL )
	{
		int err = GetLastError();
		ODS_fls( _FormatT( L"Error %s in CreateFileMapping", _GetSystemError( err ) ) );
		return FALSE;
	}
	
	m_err = cfv_ok;

	m_mapAccess = ( readOnly ? FILE_MAP_READ : FILE_MAP_ALL_ACCESS );

	//
	//	Skip for optimization, to avoid mapping again at cStringFile::Init
	//
	if ( !m_bSkipInitAtOpening )
	{
		if ( !Init() )
			return false;
	}

	//m_ptr = (BYTE *)MapViewOfFile( m_hMap, m_mapAccess, m_curPosition.HighPart, m_curPosition.LowPart, m_windowSize );

	//if ( m_ptr == NULL )
	//{
	//	int err = GetLastError();
	//	ODS_fls( _FormatT( L"Error %d in MapViewOfFile", err ) );
	//	return FALSE;
	//}

	return TRUE;
}

bool cFileView::Init()
{
	m_curPosition.QuadPart = 0;
	m_nextPosition = 0;

	return MappingNextWindow();
}

bool cFileView::MappingNextWindow()
{
	if ( m_stepSize != 0 )
	{
		SYSTEM_INFO systemInfo;
		GetSystemInfo( &systemInfo );

		if ( ( m_stepSize % systemInfo.dwAllocationGranularity ) != 0 )
		{
			ODS_fls( L"Assertion!!! Invalid mapping step granularity." );

			m_stepSize = ( ( m_stepSize / systemInfo.dwAllocationGranularity ) + 1 ) * systemInfo.dwAllocationGranularity;

			//m_err = cfv_errorInMapingFile;
			//return false;
		}
	}

	if ( m_ptr )
	{
		//
		//	It is not required unmap and map again same part of file.
		//
		if ( m_stepSize == 0 )
			return true;

		UnmapViewOfFile( m_ptr );
		m_ptr = NULL;
	}
	
	//
	//	Checks if no more windows are available for next ( non first initial step ).
	//
	if ( m_nextPosition != 0 )
		if ( m_curPosition.QuadPart + m_windowSize >= m_size.QuadPart )
			return false;

	m_curPosition.QuadPart = m_nextPosition;

	if ( m_curPosition.QuadPart >= m_size.QuadPart )
	{
		if ( m_size.QuadPart == 0 )
		{
			ODS_fls( L"size is zero" );
			return true;
		}

		ODS_fls( L"Assertion!!!" );
		return false;
	}

	m_windowSize = 2 * m_stepSize;

	ULONGLONG tailSize = m_size.QuadPart - m_curPosition.QuadPart;

	if ( m_windowSize == 0 )
		m_windowSize = (SIZE_T)tailSize;
	else
	if ( m_windowSize > tailSize )
		m_windowSize = (SIZE_T)tailSize;

	m_ptr = (BYTE *)MapViewOfFile( m_hMap, m_mapAccess, m_curPosition.HighPart, m_curPosition.LowPart, m_windowSize );

	if ( m_ptr == NULL )
	{
		m_sysErr = GetLastError();
 		ODS_fls( _FormatT( L"Error %d in MapViewOfFile", m_sysErr ) );
		return false;
	}

	m_nextPosition = m_curPosition.QuadPart + m_stepSize;

	return true;
}

void cFileView::CloseMapping()
{
	if ( m_ptr )	{ UnmapViewOfFile( m_ptr );	m_ptr  = NULL; }
	if ( m_hMap )	{ CloseHandle( m_hMap );	m_hMap = NULL; }
}

void cFileView::Close( )
{
	if( m_hFile == NULL )
		return;

	CloseMapping();

	if ( m_hFile )
		CloseHandle( m_hFile );

	commomConstructor();
}
#endif // _UseFileViewA_

BOOL cFileView::Open( const wchar_t *fileName, bool readOnly )
{
	ODSFuncS( fileName );

	Close();
	m_err = cfv_fileDoesNotExist;

	if ( !FileExistsT( fileName ) && readOnly )
	{
		//ODS_f( _FormatT( L"File '%s' does not exist", fileName ) );
		m_sysErr = ERROR_FILE_NOT_FOUND;
		return false;
	}

	m_err = cfv_errorInOpeningFile;

	m_hFile =
		CreateFileT(
			fileName,
			( readOnly ? GENERIC_READ : GENERIC_READ | GENERIC_WRITE ),
			FILE_SHARE_READ, // ( readOnly ? FILE_SHARE_READ | FILE_SHARE_WRITE : FILE_SHARE_READ ),
			NULL,
			( readOnly ? OPEN_EXISTING : OPEN_ALWAYS ),
			FILE_ATTRIBUTE_NORMAL,
			NULL);

	if ( m_hFile == INVALID_HANDLE_VALUE )
	{
		m_sysErr = GetLastError();
		ODS_fls( _FormatT( L"%s::Error %X in opening file '%s'", _CRT_WIDE( __FUNCTION__ ), m_sysErr, fileName ) );
		return false;
	}

	return OpenMapping( readOnly );
}

int WriteToFile( HANDLE hFile, const BYTE *ptr, DWORD size, bool append, DWORD *pError )
{
	DWORD err = 0;

	if ( pError == NULL )
		pError = &err;

	if ( hFile == INVALID_HANDLE_VALUE )
	{
		*pError = GetLastError();
		return 0;
	}

	DWORD pos = 0;

	if ( append )
	{
		pos = SetFilePointer( hFile, 0, NULL, FILE_END );

		if ( pos == INVALID_SET_FILE_POINTER )
		{
			*pError = GetLastError();

		}
	}

	if ( !WriteFile( hFile, ptr, size, &size, NULL ) )
	{
		ODS_fls( _FormatT( L"Error %s in writting file", _GetSystemError( -1 ) ) );
		size = 0;
	}

	return size;
}

int WriteToFile( const char *fileName, const BYTE *ptr, DWORD size, bool append, DWORD *pError, DWORD shareMode )
{
	HANDLE hFile = CreateFileA(
				fileName,
				GENERIC_WRITE | ( append ? GENERIC_READ : 0 ),
				shareMode,
				NULL,
				append ? OPEN_ALWAYS : CREATE_ALWAYS,
				FILE_ATTRIBUTE_NORMAL,
				NULL
			);


	int res = WriteToFile( hFile, ptr, size, append, pError );
	CloseHandle( hFile );
	return res;
}

int WriteToFile( const wchar_t *fileName, const BYTE *ptr, DWORD size, bool append, DWORD *pError, DWORD shareMode ) // = FILE_SHARE_READ )
{
	HANDLE hFile = CreateFileT(
				fileName,
				GENERIC_WRITE | ( append ? GENERIC_READ : 0 ),
				shareMode,
				NULL,
				append ? OPEN_ALWAYS : CREATE_ALWAYS,
				FILE_ATTRIBUTE_NORMAL,
				NULL
			);

	DWORD err = 0;

	if ( pError == NULL )
		pError = &err;

	if ( hFile == INVALID_HANDLE_VALUE )
	{
		*pError = GetLastError();
		ODS_fls( _FormatT( L"Error %s in opening file", _GetSystemError( *pError ) ) );
		return 0;
	}

	int res = WriteToFile( hFile, ptr, size, append, pError );
	CloseHandle( hFile );
	return res;
}

int WriteToFile( const wchar_t *fileName, const wchar_t *str, int len, bool append, DWORD *pError )
{
	if ( len == -1 )
		len = (int)wcslen( str );

	return WriteToFile( fileName, (BYTE*)str, len * sizeof( wchar_t ), append, pError );
}

#ifndef INVALID_SET_FILE_POINTER
#define INVALID_SET_FILE_POINTER (-1) 
#endif

int WriteToFile( HANDLE hFile, const BYTE *ptr, DWORD size, ULONGLONG pos, DWORD *pError )
{
	DWORD err = 0;

	if ( pError == NULL )
		pError = &err;

	if ( hFile == INVALID_HANDLE_VALUE )
	{
		*pError = GetLastError();
		return 0;
	}

	LARGE_INTEGER liPos;
	
	liPos.QuadPart = pos;


	if ( SetFilePointer( hFile, liPos.LowPart, &liPos.HighPart, FILE_BEGIN ) == INVALID_SET_FILE_POINTER )
	{
		*pError = GetLastError();
		
		if ( *pError != NO_ERROR )
			return 0;
	}

	return WriteToFile( hFile, ptr, size, false, pError );
}

int WriteToFile( const wchar_t *fileName, const BYTE *ptr, DWORD size, ULONGLONG pos, DWORD *pError )
{
	HANDLE hFile = CreateFileT(
				fileName,
				GENERIC_WRITE | GENERIC_READ,
				FILE_SHARE_READ,
				NULL,
				OPEN_ALWAYS,
				FILE_ATTRIBUTE_NORMAL,
				NULL
			);

	if ( hFile == INVALID_HANDLE_VALUE )
	{
		if ( pError != NULL )
			*pError = GetLastError();

		return -1;
	}

	int res = WriteToFile( hFile, ptr, size, pos, pError );
	CloseHandle( hFile );
	return res;
}

int ReadFromFile( HANDLE hFile, BYTE *ptr, DWORD size, DWORD *pError )
{
	DWORD err = 0;

	if ( pError == NULL )
		pError = &err;

	if ( hFile == INVALID_HANDLE_VALUE )
	{
		*pError = GetLastError();
		return 0;
	}

	DWORD dwRead = 0;

	if ( !ReadFile( hFile, ptr, size, &dwRead, NULL ) )
		ODS_fls( _FormatT( L"Error %s in writting file", _GetSystemError( -1 ) ) );

	return dwRead;
}

int ReadFromFile( HANDLE hFile, BYTE *ptr, DWORD size, ULONGLONG pos, DWORD *pError )
{
	DWORD err = 0;

	if ( pError == NULL )
		pError = &err;

	if ( hFile == INVALID_HANDLE_VALUE )
	{
		*pError = GetLastError();
		return 0;
	}

	if ( SetFilePointerT( hFile, pos, FILE_BEGIN ) == (DWORDLONG)-1 )
	{
		*pError = GetLastError();
		
		if ( *pError != NO_ERROR )
			return 0;
	}

	return ReadFromFile( hFile, ptr, size, pError );
}

int ReadFromFile( const wchar_t *fileName, BYTE *ptr, DWORD size, ULONGLONG pos, DWORD *pError )
{
	HANDLE hFile = CreateFileT(
				fileName,
				GENERIC_READ,
				FILE_SHARE_READ,
				NULL,
				OPEN_ALWAYS,
				FILE_ATTRIBUTE_NORMAL,
				NULL
			);

	if ( hFile == INVALID_HANDLE_VALUE )
	{
		if ( pError != NULL )
			*pError = GetLastError();

		return -1;
	}

	int res = ReadFromFile( hFile, ptr, size, pos, pError );
	CloseHandle( hFile );
	return res;
}
//
//	Copies opened file ( hFileSrc ) to new file ( fileNameDst )
//		1. Saves current hFile position
//		2. Read portions from source and write to destination
//		3. Returns current position of source file back
//	Returns number of bytes copied.
//
DWORDLONG CopyToFile( HANDLE hFileSrc, const wchar_t *fileNameDst )
{
	ULARGE_INTEGER save;
	save.QuadPart = 0;
	save.LowPart = SetFilePointer( hFileSrc, 0, (PLONG)&save.HighPart, FILE_CURRENT );

	DWORD pos = SetFilePointer( hFileSrc, 0, NULL, FILE_BEGIN );
	if ( pos == INVALID_SET_FILE_POINTER )
		return 0;

	HANDLE hFileDst = CreateFileT( fileNameDst, GENERIC_WRITE, 0, NULL, CREATE_NEW, 0, NULL );
	if ( hFileDst == INVALID_HANDLE_VALUE )
		return 0;

	const int buffSize = 64 * 1024;

	BYTE *buff = new BYTE[ buffSize ];
	if ( buff == NULL )
		return 0;

	DWORDLONG dwlCopied = 0;

	DWORD dwRead;

	while( ReadFile( hFileSrc, buff, buffSize, &dwRead, NULL ) )
	{
		if ( dwRead == 0 )
			break;

		DWORD  dwWritten = 0;
		if ( !WriteFile( hFileDst, buff, dwRead, &dwWritten, NULL ) || ( dwRead != dwWritten ) )
		{
			dwlCopied = 0;
			break;
		}

		dwlCopied += dwWritten;
	}

	SetFilePointer( hFileSrc, save.LowPart, (PLONG)&save.HighPart, FILE_BEGIN );

	CloseHandle( hFileDst );

	delete( buff );

	return dwlCopied;
}

//
//	cStringFile
//

//cStringFile::cStringFile( const char *fileName )
//{
//	m_isUnicode = false;
//	m_pos = 0;
//
//	m_bf = NULL;
//	m_bfSize = 0;
//
//	if ( fileName != NULL )
//		Open( fileName );
//}

bool cStringFile::commonConstructor( const wchar_t *fileName )
{
	m_isUnicode = false;
	m_pos = 0;

	m_bf = NULL;
	m_bfSize = 0;

	if ( fileName == NULL )
		return true;
	
	return Open( fileName );
}

//bool cStringFile::Open( const char *fileName, bool readOnly )
//{
//	bool res = cFileView::Open( fileName, readOnly ) == TRUE;
//
//	if ( res )
//		Init();
//
//	return res;
//}
//
bool cStringFile::Open( const wchar_t *fileName, bool readOnly )
{
	m_bSkipInitAtOpening = true;

	bool res = cFileView::Open( fileName, readOnly ) == TRUE;

	if ( res )
		res = Init();

	return res;
}


template < class T > int getLen( T *s )
{
	for( int i = 0; ; i++ )
	{
		if ( s[i] == 0 )
			return i;
	}
}

template < class T > int cutOffServiceSymbolsAtTail( T *s )
{
	int len = getLen( s );

	while ( ( s[ len - 1 ] == L'\n' ) ||
			( s[ len - 1 ] == L'\r' ) ||
			( s[ len - 1 ] == L' ' ) )
	{
		s[ --len ] = 0;
		if ( len == 0 ) break;
	}

	return len;
}

bool cStringFile::getNextFile( cMBStr *nextFileName )
{
	char bf[ MAX_PATH ];
	
	nextFileName->Free();

	while( true )
	{
		if ( !ReadString( bf, maxLen( bf ) ) )
			return false;
		
		nextFileName->Add( bf );
		
		if ( m_bCompleted )
			break;
	}

	if ( nextFileName->m_szBody == NULL )
		return false;

	if ( nextFileName->m_szBody[0] == '#' )
		return getNextFile( nextFileName );

	//
	//	Cuts new line, carrige return and probel symbols at the end of string.
	//
	//int len = (int)strlen( nextFileName->m_szBody );

	//while ( ( nextFileName->m_szBody[ len - 1 ] == '\n' ) ||
	//		( nextFileName->m_szBody[ len - 1 ] == '\r' ) ||
	//		( nextFileName->m_szBody[ len - 1 ] == ' ' ) )
	//{
	//	nextFileName->m_szBody[ --len ] = 0;
	//	if ( len == 0 ) break;
	//}

	int len = cutOffServiceSymbolsAtTail( nextFileName->m_szBody );

	if ( len == 0 )
		return getNextFile( nextFileName );

	return true;
}

bool cStringFile::Init()
{
	//
	//	cStringFile::Open assigns the m_bSkipInitAtOpening
	//
	if ( !cFileView::Init() )
	{
		ODS_fls( L"Assertion!" );
		m_err = cfv_errorInOpeningFile;
	}

	if ( m_err )
		cStringBuffer::Init( NULL, 0 );
	else
		cStringBuffer::Init( m_ptr, m_windowSize );

	return m_err == cfv_ok;
}

//virtual 
bool cStringFile::FindString( wchar_t *buff, int buffLen, const wchar_t *str, bool getTailOnly )
{
	if ( !Init() )
		return false;

	return cStringBuffer::FindString( buff, buffLen, str, getTailOnly );
}

bool cStringFile::getNextFileT( cWCStr *nextFileName )
{
	if ( m_isUnicode )
		return getNextFileW( nextFileName );

	//char pathA[ MAX_PATH ];
	//if ( !getNextFile( pathA ) )
	//	return false;

	//_mbstowcs( nextFileName, MAX_PATH, pathA );

	cMBStr mbs;
	if ( !getNextFile( &mbs ) )
		return false;

	nextFileName->Set( mbs.m_szBody );

	return true;
}

bool cStringFile::UpdateMapping()
{
	if ( m_stepSize == 0 )
		return true;

	if ( m_pos < m_stepSize )
		return true;

	if ( m_curPosition.QuadPart + m_windowSize >= m_size.QuadPart )
		return true;

	if ( !MappingNextWindow() )
		return false;

	m_startPos = m_pos - m_stepSize;
	m_bf = m_ptr;
	m_bfSize = m_windowSize;

	cStringBuffer::Restart();

	return true;
}

bool cStringFile::getNextFileW( cWCStr *nextFileName )
{
	cWCStr bf( 1024 * 4 ); // 4K

	if ( bf.m_szBody == NULL )
		return false;
	
	nextFileName->Free();

	while( true )
	{
		if ( !ReadStringW( bf.m_szBody, bf.m_len ) )
			return false;
		
		nextFileName->Add( bf.m_szBody );
		
		if ( m_bCompleted )
			break;
	}

	if ( nextFileName->m_szBody == NULL )
		return false;

	if ( nextFileName->m_szBody[0] == L'#' )
		return getNextFileW( nextFileName );

	//
	//	Cuts new line, carrige return and probel symbols at the end of string.
	//
	int len = cutOffServiceSymbolsAtTail( nextFileName->m_szBody );

	if ( len == 0 )
		return getNextFileW( nextFileName );

	return true;
}

bool AddStrToFile( HANDLE hFile, const wchar_t *str, int len  )
{
	cWCStr wcsBf;
	wcsBf.FormatT( L"%s\r\n", str );

	if ( len == -1 )
		len = (int)wcslen( wcsBf.m_szBody );

	DWORD written = 0;

	if ( !WriteFile( hFile, (BYTE *)wcsBf.m_szBody, len * sizeof( wchar_t ), &written, NULL ) || ( (int)written != len * sizeof( wchar_t ) ) )
	{
		ODS_fls( _FormatT( L"Assertion! Error '%s' in writing file 0x%X", _GetSystemError( -1 ), hFile ) );
		return false;
	}

	return true;
}

//
//	Unicode text file has the prefix.
//
const BYTE unicodeLEPrefix[ 2 ] = { 0xFF, 0xFE }; 
const BYTE unicodeBEPrefix[ 2 ] = { 0xFE, 0xFF }; 
const BYTE utf8prefix[ 3 ] = { 0xEF, 0xBB, 0xBF };

bool InitUnicodeFile( HANDLE hFile, int codePage )
{
	SetFilePointer( hFile, 0, NULL, FILE_BEGIN );
	SetEndOfFile( hFile );

	const BYTE *prefix = NULL;
	int prefixSize = 0;

	switch ( codePage )
	{
		case CP_ACP:	return true;
		case CP_UTF8:	prefix = utf8prefix; prefixSize = sizeof( utf8prefix ); break;
		default:
			prefix = unicodeLEPrefix; prefixSize = sizeof( unicodeLEPrefix ); break;
	}

	DWORD written = 0;

	return 
		WriteFile( hFile, prefix, prefixSize, &written, NULL )
			&& ( (int)written == prefixSize );
}

bool InitUnicodeFile( const wchar_t *fileName, HANDLE *pHFile, int codePage )
{
	HANDLE hFile;

	hFile = CreateFileT( fileName, GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ, NULL, OPEN_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL );

	if ( hFile == INVALID_HANDLE_VALUE )
		return false;
	
	bool ok = InitUnicodeFile( hFile, codePage );

	if ( pHFile != NULL )
		*pHFile = hFile;
	else
		CloseHandle( hFile );

	return ok;
}

//
//	MoveCurrentFilePointer
//	Usual variants do not work to move file pointer back:
//		v.1
//			SetFilePointer( hFile, -size, &(uli.HighPart), FILE_CURRENT );
//			It does not work becouse it uses common H+L signed value.
//		v.2
//			LARGE_INTEGER uli = { 0, 0 };
//			uli.QuardPart -= size;
//			SetFilePointer( hFile, uli.LowPart, &(uli.HighPart), FILE_CURRENT );
//			1. it reports error move before file beginning.
//			2. it do never work for Windows 95 ( as MSDN report uli.HighPart has to be zero ).
//
bool MoveCurrentFilePointer( HANDLE hFile, LONGLONG *ll, int *pLastError )
{
	int err = 0;
	if ( pLastError == NULL )
		pLastError = &err;

	LARGE_INTEGER uli = { 0, 0 };
	uli.LowPart = SetFilePointer( hFile, 0, &(uli.HighPart), FILE_CURRENT );

	if ( uli.LowPart == INVALID_SET_FILE_POINTER )
	{
		*pLastError = GetLastError();
		if ( *pLastError != NO_ERROR )
		{
			ODS_fls( _FormatT( L"%s::File pointer is invalid %X", _CRT_WIDE( __FUNCTION__ ), 0 ) );
			return false;
		}
	}

	uli.QuadPart += *ll;

	uli.LowPart = SetFilePointer( hFile, uli.LowPart, &(uli.HighPart), FILE_BEGIN );
	if ( uli.LowPart == INVALID_SET_FILE_POINTER )
	{
		*pLastError = GetLastError();
		if ( *pLastError != NO_ERROR )
		{
			ODS_fls( _FormatT( L"%s::File pointer is invalid %X", _CRT_WIDE( __FUNCTION__ ), 1 ) );
			return false;
		}
	}

	*ll = uli.QuadPart;

	return true;
}

static bool SafeReadLastData( HANDLE hFile, BYTE *bf, int bfSize, int *pLastError )
{
	int err = 0;
	if ( pLastError == NULL )
		pLastError = &err;

	FlushFileBuffers( hFile );

	LONGLONG ll = -bfSize;
	if ( !MoveCurrentFilePointer( hFile, &ll, pLastError ) )
	{
		ODS_fls( L"Assertion!" );
		return false;
	}

	DWORD dwRead; 
	BOOL ok = ReadFile( hFile, bf, bfSize, &dwRead, NULL );

	*pLastError = GetLastError();

	if ( !ok )
	{
		ODS_fls( _FormatT( L"%s::Read file error (%d)", _CRT_WIDE( __FUNCTION__ ), *pLastError ) );
		ODS_fls( L"Assertion!" );
		return false;
	}

	if ( dwRead != bfSize )
	{
		ODS_fls( _FormatT( L"%s::Read file error read size %X, was read %X", _CRT_WIDE( __FUNCTION__ ), bfSize, dwRead ) );
		ODS_fls( L"Assertion!" );
		return false;
	}

	return true;
}

//
//	19.05.2011
//	Strange behaivior on NTFS W7 is detected!!!
//	It reports that 0 bytes available for caller, but WriteFile allows to write 4K several times.
//	So the 'modification 4K' is implemented.
//		It tryes to write on zero free space drive 4K buffers until WriteFile returns OK.
//	Note, caller fuction have to send large enough buffer ( It concern Wipe free space with cWipeDrive::Process() ).
//
static bool writeWithTriesBy4K( HANDLE hFile, const wchar_t *fileName, BYTE *bf, DWORD size, const wchar_t *szRootDir, DWORD *pWritten, int *pLastError  )
{
	bool ok = false;
	int maxQ = 16;
	int modification4KCount = 0;

	for( int tryCount = 0; ; tryCount++ )
	{
		if ( tryCount > maxQ )
		{
			ODS_fls( _FormatT( L"Try count exceeded!, %s", getWCStr(ok) ) );
			*pLastError = ERROR_DISK_FULL;
			ok = true;
			break;
		}

		ULARGE_INTEGER freeSpace;
		if ( !GetDiskFreeSpaceExT( szRootDir, &freeSpace ) )
		{
			ODS_fls( _FormatT( L"Action: %s\tError: %s\tPath: %s", L"Get Disk Free Space", getWCStr( GetLastError() ), szRootDir ) );
			break;
		}
		else
		{
			ODS_fls( _FormatT( L"Action: %s\tError: %s\tPath: %s\tFree space: %I64X\tBuffer size: %X\tWritten size: %X\tTry: %d",
				L"Write file", L"There is not enough space on the disk.", 
										fileName,
													freeSpace.QuadPart,		size,			*pWritten,		tryCount ) );

			DWORD newSize = (DWORD)( freeSpace.QuadPart );

			if ( newSize == 0 )
			{
				modification4KCount ++;
				newSize = 4096;
			}

			if ( newSize >= ( size - *pWritten ) )
			{
				//
				// Nonsence!
				// Error when free space was enough for writting.
				//
				ok = false;
				ODS_fls( _FormatT( L"Assertion! New size %s >= ( buffer size %s - written size %s )", _GetSizeText( newSize ), _GetSizeText( size ), _GetSizeText( *pWritten ) ) );
				break;
			}
			else
			{
				// Tries to write left part on the disk...
				BYTE *newBf = bf + *pWritten;
				DWORD newWritten = 0;

				ok = WriteFile( hFile, newBf, newSize, &newWritten, NULL ) == TRUE;
				
				if ( ok )
				{
					*pWritten += newWritten;
					ODS_fls( _FormatT( L"Written new %s, total %s", _GetSizeText( newWritten ), _GetSizeText( *pWritten ) ) );
				}
				else
				{
					*pLastError = GetLastError();
					GetDiskFreeSpaceExT( szRootDir, &freeSpace );

					ODS_fls( _FormatT( L"Action: %s\tError: %s\tFree space: %I64X\tNew buffer size: %X\tNew written size: %X",
							L"Write file", cWCStr().GetSystemError( *pLastError ),
											freeSpace.QuadPart,		newSize,			newWritten ) );

					if ( *pLastError != ERROR_DISK_FULL )
						break;

					//
					//	See comment at 'modification 4K' above.
					//
					if ( freeSpace.QuadPart == 0 )
					{
						ok = true;
						ODS_fls( _FormatT( L"Modification 4K count = %d", modification4KCount ) );
						break;
					}
				}
			}
		}
	}

	return ok;
}

//
//	WriteFileVerifyInternal
//		1. Writes required data to disk
//		2. If verification is on, it flushes file buffers
//		3. Moves file poiter back and reads the data is written before
//		4. Compares the data was read from file and original.
//		19.05.2011
//		5. If disk is full the WriteFile function returns false and returns error ERROR_DISK_FULL.
//			It does not write even left part. So following improvement is added:
//			It gets free space and tries to write at least left part of required buffer.
//			
//
//	Note, ewfv_errorSize is not treat as error. It means that *pWritten size is properly written and verified.
//
static int WriteFileVerifyInternal( HANDLE hFile, BYTE *bf, DWORD size, DWORD *pWritten = NULL, bool verify = false, int *pLastError = NULL )
{
	int err = 0;
	if ( pLastError == NULL )
		pLastError = &err;

	DWORD written = 0;
	if ( pWritten == NULL )
		pWritten = &written;

	// 1.
	
	BOOL ok;
	{
		//ODSFunc;
		ok = WriteFile( hFile, bf, size, pWritten, NULL );
		//ODSFuncRes( ok );
	}

	*pLastError = GetLastError();

	bool bDiskFull = false;

	if ( !ok )
	{
		//
		// 5. Full disk improvenment...
		//
		if ( ( *pLastError == ERROR_DISK_FULL ) ||
			 ( *pLastError == ERROR_ALREADY_EXISTS ) // It strange but sometimes it returns the error when written buffer is more then available free space.
			 )
		{
			bDiskFull = true;

			cWCStr fileName( MAX_PATH + 1 );

			//
			// GetFileNameFromHandle does not work for zero size file.
			// So get the file size and write at least 4K.
			//
			DWORD dwFileSizeHi = 0;
			DWORD dwFileSizeLo = GetFileSize( hFile, &dwFileSizeHi );

			if( dwFileSizeLo == 0 && dwFileSizeHi == 0 )
				ok = WriteFile( hFile, bf, 4096, pWritten, NULL );
			else
				ok = true;

			if ( ok )
			{
				ok = GetFileNameFromHandle( hFile, fileName.m_szBody );

				if ( ok )
				{
					cWCStr rootDir( MAX_PATH );

					if ( !GetDriveRootDirT( rootDir.m_szBody, fileName.m_szBody ) )
						ODS_fls( _FormatT( L"Action: %s\tError: %s\tPath: %s", L"Get Disk Root Dir", _GetSystemError(-1), rootDir.m_szBody ) );
					else
						ok = writeWithTriesBy4K( hFile, fileName.m_szBody, bf, size, rootDir.m_szBody, pWritten, pLastError );
				}
			}
		}
	}

	if ( !ok )
	{
		ODS_fls( _FormatT( L"Error %d in writting file, size %d, written %d", *pLastError, size, *pWritten ) );
		return ewfv_errorInWrittingFile;
	}

	// It verifies all written data
	if ( *pWritten == 0 )
	{
		ODS_fls( L"Zero size written" );
		*pLastError = 0;
		return ewfv_errorNo;
//		return ewfv_errorSize;
	}

	if ( !verify )
		return ewfv_errorNo;

	// 2. & 3.
	cAutoDeallocate < BYTE > ad;
	if ( ad.Allocate( *pWritten ) == NULL )
	{
		ODS_fls( _FormatT( L"Asertion! Error in memory allocation, size %s", _GetSizeText( *pWritten ) ) );
		return ewfv_errorInAllocatingMemory;
	}

	if ( !SafeReadLastData( hFile, ad.m_ptr, ad.m_len, pLastError ) )
	{
		ODS_fls( L"Assertion!" );
		return ewfv_errorInReadingFile;
	}

	// 4.
	if ( memcmp( ad.m_ptr, bf, *pWritten ) != NULL )
	{
		if ( bDiskFull )
		{
			ODS_fls( _FormatT( L"Error in verification after disk full detection! Written size %s.", _GetSizeText( *pWritten ) ) );
			return ewfv_errorNo;
		}
		else
		{
			ODS_fls( _FormatT( L"Error in verification %s", _GetSizeText( *pWritten ) ) );
			return ewfv_errorInVerifing;
		}
	}

	if ( *pWritten != size )
	{
		ODS_fls( _FormatT( L"Size 0x%X != written size 0x%X", size, *pWritten ) );
		//return ewfv_errorSize;
		return ewfv_errorNo;
	}

	return ewfv_errorNo;
}

const char *ewfv_str( int err )
{
	switch ( err )
	{
		case ewfv_errorNo					: return "No error";
		case ewfv_errorSize					: return "Size is invalid";
		case ewfv_errorInAllocatingMemory	: return "Allocating memory";
		case ewfv_errorInWrittingFile		: return "In Writing";
		case ewfv_errorInReadingFile		: return "In Reading";
		case ewfv_errorInVerifing			: return "In Verifing";
		default								: return "Undefined";
	}
}

//
//	WriteFileVerify
//		Writes data first if verification is on, it reads data from file for verification.
//		If the data is not compatible it tries to write and verify again according with try count.
//
int WriteFileVerify( int tryCount, HANDLE hFile, BYTE *bfWrite, DWORD size, DWORD *pWritten, bool bVerify, int *pLastError )
{
	//ODSFuncS( getWCStr( size ) );

	DWORD written = 0;
	if ( pWritten == NULL )
		pWritten = &written;

	int iRes = WriteFileVerifyInternal( hFile, bfWrite, size, pWritten, bVerify, pLastError );

	for( ; iRes == ewfv_errorInVerifing; tryCount-- )
	{
		if ( tryCount == 0 )
			break;

		LONGLONG ll = 0;
		ll -= size;
		MoveCurrentFilePointer( hFile, &ll );

		cWCStr wcs;
		wcs.FormatT( L"Error in verification ( file position %I64X, buffer size %X ). Try once more!", ll, size );

		ODS_fls( wcs.m_szBody );

		iRes = WriteFileVerifyInternal( hFile, bfWrite, size, pWritten, true, pLastError );
	}

	//ODSFuncRes( *pWritten );

	return iRes;
}

bool DeletePrivateProfileString(
		LPCSTR lpAppName,
		LPCSTR lpKeyName,
		LPCSTR lpString,
		LPCSTR lpFileName
	)
{
	ODSFuncS( _FormatT( L"%s, %s, %s, %s", _getWcs( lpAppName ), _getWcs( lpKeyName ), _getWcs( lpString ), _getWcs( lpFileName ) ) );

	cStringFile sf( lpFileName );
	cMBStr newBody;

	cMBStr mbsSection;
	mbsSection.Add( '[' ); mbsSection.Add( lpAppName ); mbsSection.Add( ']' );

	char bf[1024];
	bool bIsSection = false;
	int iSkipped = 0;

	size_t len = strlen( lpKeyName );

	int q = 0;

	while( sf.ReadString( bf, maxLen( bf ) ) )
	{
		q++;

		if ( bf[0] == '[' )
			bIsSection = mbsSection.equ( bf );
		else
		if ( bIsSection )
		{
			bool bSkipIt =
				( ( strncmp( bf, lpKeyName, len ) == 0 ) && 
				( bf[len] == '=' ) &&
				( strcmp( bf + len + 1, lpString ) == 0  ) );

			if ( bf[0] == '\0' )
				bSkipIt = true;

			if ( bSkipIt )
			{
				iSkipped++;
				continue;
			}
		}

		newBody.Add( bf );
		newBody.Add( "\r\n" );
	}

	sf.Close();

	if ( q == 0 )
	{
		ODS_fls( _FormatT( L"File is empty!" ) );
		return true;
	}
	
	if ( iSkipped == 0 )
	{
		ODS_fls( _FormatT( L"No on string is found!" ) );
		return true;
	}

	return WriteToFile( lpFileName, (BYTE*)newBody.m_szBody, newBody.m_len ) != 0;
}

