/*******************************************************************\
 *
 *	Copyright (c) 2000-2015 Jetico, Inc., Finland
 *	All rights reserved.
 *
 *	File:		IconCache.h
 *	Revision:	04.03.2015
 *	Modified:	01.01.2000
 *  Author:		A. Pichuev
 *	Description:
 *		Gets and caches shell icons.
 *		04.03.2015 Uses real icons for directries.
 *
\*******************************************************************/

#pragma once

#include <windows.h>
#include <commctrl.h>
#include <assert.h>

class cIconCache
{
	HINSTANCE m_hInst;

public:
	//
	//	Large and small image lists...
	//
	HIMAGELIST m_hImlLg;
	HIMAGELIST m_hImlSm;
	int m_defNum;

	//
	//	If error occuares all functions return m_undefNum number.
	//	Otherwise they return number of icon in large and small image lists.
	//	the undefined and folder icons may be added to image list at common constructor...
	//
	int m_undefNum;
	int m_firstNum;

	cIconCache( HINSTANCE hInst, HWND hListView, int idiFolder, int idiUnknown, HICON hDefIcon = NULL );
	cIconCache( HINSTANCE hInst, HWND hListView, HICON hFolder = NULL, HICON hUnknown = NULL, HICON hDefIcon = NULL );
	void CommonConstructor( HINSTANCE hInst, HWND hListView, HICON hFolder, HICON hUnknown, HICON hDefIcon );

	void Init( HINSTANCE hInst, HWND hListView, int idiFolder, int idiUnknown );

	cIconCache()
	{
		int i = 0;
		m_firstNum = -1;
	}

	~cIconCache();

	void DestroyImageLists();

	//
	//	AddIcon.
	//	Adds both large and small icons 
	//	 and returns it's position into image list.
	//
	int AddIcon( HICON hIcon );

	//
	//	GetIconNum.
	//	Returns specified icon number into image lists.
	//	If icon was not inserted to image list it will be added by AddShellIcon
	//
	int GetIconNum( const wchar_t *fileName, wchar_t **pTypeName );

	//
	//	AddIconWithPath
	//		Adds directly the small icon to both small and large lists.
	//		This icon was get by GetAppIconT function.
	//		Returns specified icon number into image lists.
	//
	int AddIconWithPath( const wchar_t *filePathName, HICON hSmallIcon, HICON hLargeIcon = NULL );

protected:

	//
	//	AddShellIcon.
	//	Uses SHGetFileInfo to extract file icon by extention.
	//
	int AddShellIcon( const wchar_t *ext, const wchar_t *filePathName, bool bDirectory );

	//
	//	GetCachedNum.
	//	Checks cache table of extensions. 
	//
	int GetCachedNum( const wchar_t *ext, const wchar_t *filePathName, bool bdirectory );

	//
	//	extTbl is the table with cached extensions.
	//
	wchar_t (*extTbl)[5];
	wchar_t **typeNameTbl;
	int extTblSize;

	//
	//	Adds new one item to cache table.
	//	Returns number of added item.
	//
	int ReallocExtTbl();

public:
	//
	//      Shared image list my be assigned and has to be destroied
	//
	void SetImageList( HWND hListView );
	void DestroySharedImageLists();
};