/*
 *	Copyright 1995-2015 Jetico, Inc. All rights reserved.
 *	OpenFiles.h
 *	Description:
 *		1. GetSaveFileName and GetOpenFileName are used.
 *		2. Filters.
 *			a. this functions above are used file filters in following format
 *					"Text (*.txt)\0*.TXT\0All (*.*)\0*.*\0"
 *				Filters are zero separated and the string is doble zero terminated.
 *			b. It is not possible to save string in this format in resources.
 *				This function allows another format:
 *					"Text (*.txt)%*.TXT%All (*.*)%*.*%"
 *				It automatically converted to format a.
 *		3. This function automatically save last filter index and set it as default for following calls.
 *		4. Output buffer is double zero terminated string.
 *			It may be single string ( whole file path ) or zero separated multistring buffer where
 *			first string is folder path next strings are names of selected files ( just name without folder path ).
 *			cOpenFilesList parses this format in both cases.
 *		5. It saves initial path after first call and uses this path for next call. The initial path may be assigned at start.
 *
 *	Modified:	04.03.2015
 *	Author:		Alexander Pichuev.
 */

#pragma once

#include "UnicodeFuncs.h"
#include "cOpenFilesList.h"

//
//	BrowseFiles
//		Use helper class cOpenFilesList ( see cOpenFilesList.h ) for parsing multy or single string result.
//
bool BrowseFiles(
		HWND hWnd,
		wchar_t *buff,	int buffSize,	// See description 4. Output buffer.
		const wchar_t *filters = NULL,	// See description 2. Filters
		const wchar_t *title = NULL,
		bool allowMultiselect = false,
		DWORD additionalFlags = 0 // the OFN_OVERWRITEPROMPT flag uses GetSaveFileName, otherwise GetOpenFileName
	);

//
//	Last filter index. See description 2.b
//
void BF_SetLastFilterIndex( int index );
int BF_GetLastFilterIndex( const wchar_t *filterPairs = NULL );

//
//	Initial path. See description 5.
//
bool BF_SetInitialPath( wchar_t *buffer, int nFileOffset = 0 );
wchar_t *BF_GetInitialPathW( HWND hWnd = NULL );

//
//	BF_isSinglePathBuffer
//		Just check szzBuffer ( double zero terminated string  ) according output buffer format ( see description 4. )
//
bool BF_isSinglePathBuffer( const wchar_t *szzBuff );

//
//	BF_getNextString
//		Finds first zero and return pointer next symbol.
//		If the symbol is zero it is end of double zero terminated buffer.
//
const wchar_t *BF_getNextString( const wchar_t *szzStr );

//void convertMultiStringBuffer( wchar_t *buffW, const char *buffA, int buffSize );
//void convertMultiStringBuffer( char *buffA, const wchar_t *buffW, int buffSize );

