#include "stddlgs.h"
#include "..\CertificateLib\SMemory\cPassword.h"
#include <stdio.h>
#include <assert.h>
#include <windowsx.h>


LRESULT cTextDlg::RealDlgProc( UINT uMsg, WPARAM wParam, LPARAM lParam )
{
	if ( uMsg == WM_INITDIALOG )
	{
		CenterPosition();

		if ( m_dlgTextW )	UpdateDlgItemTextT( m_hDlg, m_IDText, m_dlgTextW );
		else				UpdateDlgItemText( m_hDlg, m_IDText, m_dlgText );

		if ( m_captionW )	SetWindowTextT( m_hDlg, m_captionW );
		else
		if ( m_caption )	SetWindowTextA( m_hDlg, m_caption );
	}
	else
	if ( uMsg == WM_COMMAND )
	{
		EndDialog( m_hDlg, LOWORD(wParam) );
	}

	return FALSE;
}

LRESULT cPasswordDlg::RealDlgProc( UINT uMsg, WPARAM wParam, LPARAM lParam )
{
	if ( uMsg == WM_INITDIALOG )
	{
		if ( m_dlgText )
		{
			cWCStr pattern;
			pattern.GetWindowTextT( GetItem( m_IDText ) );

			cWCStr text;

			if ( wcschr( pattern.m_szBody, L'%' ) == NULL )
				text.Set( m_dlgText );
			else
				text.FormatT( pattern.m_szBody, m_dlgText );

			SetDlgItemTextT( m_IDText, text.m_szBody );
		}


		if ( m_caption )
			SetWindowTextT( m_hDlg, m_caption );

		HWND hEdt = GetDlgItem( m_hDlg, m_IDPwdA );

#ifdef _useKBFilter_
		m_kbd.GenerateEncodingTable( m_useSecureTyping );
		m_edtA.Attach( hEdt );
		SetFocus( hEdt );

		hEdt = GetDlgItem( m_hDlg, m_IDPwdB );
		if ( hEdt != NULL )
			m_edtB.Attach( hEdt );
#else
		SetFocus( hEdt );
#endif // #ifdef _useKBFilter_

	}
	else
	if ( uMsg == WM_DESTROY )
	{
#ifdef _useKBFilter_
		m_kbd.FreeTables( m_useSecureTyping );
#endif // #ifdef _useKBFilter_
	}
	else
	if ( uMsg == WM_COMMAND )
	{
		HWND hWnd;

		if (LOWORD(wParam) == IDOK)
		{
			bool reportError = true;

			if ( m_pwd != NULL )
			{
				int pwdErr = ecpeOK;

#ifdef _useKBFilter_
				pwdErr = CheckAndGetPassword( m_pwd, &m_edtA, &m_edtB );
#else
				pwdErr = CheckAndGetPassword( m_pwd, GetDlgItem( m_hDlg, m_IDPwdA ), GetDlgItem( m_hDlg, m_IDPwdB ) );
#endif // #ifdef _useKBFilter_

				if ( pwdErr == ecpeSecondPwdIsEmpty )
				{
					// The CheckAndGetPassword set focus to corresponded edit box.
					return false;
				}

				if ( pwdErr != ecpeOK )
				{
					errorActor( pwdErr );
					return false;
				}
			}
		}
		else
		if (LOWORD(wParam) == IDCANCEL)
		{
		}
		else
		{
		//	assert( FALSE );
			return FALSE;
		}

		//
		//	Clear the edit box text memory.
		//
		char tmpText[ m_pwd->MaxLenght ];
		memset( tmpText, '0', maxLen( tmpText ) - 1 );
		tmpText[ maxLen( tmpText ) - 1 ] = 0;

		hWnd = GetDlgItem( m_hDlg, m_IDPwdA );
		SetWindowTextA( hWnd, tmpText );

		hWnd = GetDlgItem( m_hDlg, m_IDPwdB );
		if ( hWnd != NULL )
			SetWindowTextA( hWnd, tmpText );


		EndDialog( m_hDlg, LOWORD(wParam) );
	}

	return FALSE;
}

#ifdef _useKBFilter_

int cPasswordDlg::CheckAndGetPassword( cPassword *pwd, cKbdEdit *edt )
{
	if ( edt->m_hEdit == 0 )
	{
		assert( false );
		return ecpeInternal;
	}

	pwd->ClearBuffers();

	int pwdLength = Edit_LineLength( edt->m_hEdit, 0 );

	if ( pwdLength == 0 ) 
	{
		SetFocus(  edt->m_hEdit );
		return ecpeEmpty;
    } 

	if ( edt->dlg->m_showPwdChars // m_keyGenFlags & FLAG_ADVANCED_SHOW_PWD_CHARS) != 0) 
		  || !edt->dlg->m_useSecureTyping  )
		pwd->ReadFromEditBox( edt->m_hEdit );
	else
		pwd->ReadFromBuffer( edt->m_buffer );

	if ( pwdLength < cPassword::MinimumLenght ) 
	{
		SetFocus( edt->m_hEdit );
		return ecpeTooSmall;
    } 

	return ecpeOK;
}

//
//	CheckAndGetPassword
//		1. Checks password length, 
//		2. Verifies if edtB is not NULL,
//
int cPasswordDlg::CheckAndGetPassword( cPassword *pwd, cKbdEdit *edtA, cKbdEdit *edtB, cPassword *pPwdConfirm )
{
	if ( edtA->m_hEdit == NULL )
	{
		assert( false );
		return ecpeInternal;
	}

	int res = CheckAndGetPassword( pwd, edtA );

	if ( res != ecpeOK )
		return res;

	if ( ( edtB == NULL ) || ( edtB->m_hEdit == NULL ) )
		return ecpeOK;

	int pwdConfirmLength = Edit_LineLength( edtB->m_hEdit, 0 );

	if ( pwdConfirmLength == 0 )
	{
		SetFocus( edtB->m_hEdit );
		return ecpeSecondPwdIsEmpty; // let user to type the second copy of password
	}
	
	cPassword pwdConfirm;

	if ( pPwdConfirm == NULL )
		pPwdConfirm = &pwdConfirm;
	
	res = CheckAndGetPassword( pPwdConfirm, edtB );

	if ( ( res != ecpeOK ) ||
		 ( wcscmp( pwd->wcBuffer, pPwdConfirm->wcBuffer ) != 0 ) )
	{
		SetFocus( edtB->m_hEdit );
		return ecpeMismatch;
	}

	return ecpeOK;
}
#else

int cPasswordDlg::CheckAndGetPassword( cPassword *pwd, HWND hEdtA, HWND hEdtB, cPassword *pPwdConfirm )
{
	pwd->ReadFromEditBox( hEdtA );

	if ( hEdtB == NULL ) // Dialog with confirmation does report errors
		return ecpeOK;

	SetFocus( hEdtA );

	if ( pwd->IsEmpty() )
		return ecpeEmpty;

	if ( !pwd->CheckValidSize() )
	{
		SetFocus( hEdtA );
		return ecpeTooSmall;
	}

	cPassword pwdConfirm;

	if ( pPwdConfirm == NULL )
		pPwdConfirm = &pwdConfirm;

	pPwdConfirm->ReadFromEditBox( hEdtB );

	SetFocus( hEdtB );

	if ( pPwdConfirm->IsEmpty() )
		return ecpeSecondPwdIsEmpty;

	if ( pPwdConfirm->Compare( pwd ) != 0 )
		return ecpeMismatch;

	return ecpeOK;
}

#endif // #ifdef _useKBFilter_
