#pragma once 

#include "cDlg.h"
#include "..\CertificateLib\SMemory\cPassword.h"

#ifdef _useKBFilter_
#include "cKBFilter.h"
#endif // #ifdef _useKBFilter_

//
//	Dialog with text item...
//
class cTextDlg : public cDlg
{
	const char *m_caption;
	const char *m_dlgText;
	const wchar_t *m_captionW;
	const wchar_t *m_dlgTextW;
	int m_IDText;

 public:
	cTextDlg( const char *dlgText, int IDText, const char *caption = NULL )
	{
		m_dlgText = dlgText;
		m_IDText = IDText;
		m_caption = caption;
		m_captionW = NULL;
		m_dlgTextW = NULL;
	}

	cTextDlg( const wchar_t *dlgText, int IDText, const wchar_t *caption = NULL )
	{
		m_dlgTextW = dlgText;
		m_IDText = IDText;
		m_captionW = caption;
		m_caption = NULL;
		m_dlgText = NULL;
	}
 private:
	LRESULT RealDlgProc( UINT uMsg, WPARAM wParam, LPARAM lParam );
};

//
//	Dialog with text password item...
//

class cPasswordDlg :
#ifdef _useKBFilter_
	public cDlgKbd
{
	cKbdEdit m_edtA;
	cKbdEdit m_edtB;
#else
	public cDlg
{
#endif // #ifdef _useKBFilter_
protected:
	wchar_t *m_caption;
	wchar_t *m_dlgText;
	int m_IDText;
	cPassword *m_pwd;
	int m_IDPwdA;
	int m_IDPwdB;

	cWCStr bf[2];

	//
	//	Table with error IDS, according the error number below.
	//
	int *m_errIDS_T;

public:

	//
	//	Any checks passwords function returns ecpeOK if no error, otherwise returns corresponded error
	//	and set focus to corresponded window where error is detected.
	//
	enum eCheckPwdErr {
		ecpeOK = 0,
		ecpeSecondPwdIsEmpty,
		ecpeInternal,
		ecpeEmpty,
		ecpeTooSmall,
		ecpeMismatch,
		ecpeQ
	};

	//
	//	The confirmation dialog checks passwords by CheckAndGetPassword.
	//	The errors above are sent to the virtual errorActor to be reported as MessageBox with corresponded resources.
	//	Or the error notification may be overridden by another virtual actor.
	//
	virtual int errorActor( int errNum )
	{
		if ( m_errIDS_T == NULL )
		{
			assert( false );
			return -1;
		}

		DlgMessageBoxT( m_errIDS_T[ errNum ], MB_ICONEXCLAMATION );
		return 0;
	}

	void Init( const wchar_t *dlgText, int IDText, int IDPwdA, int IDPwdB, int *errIDS_T, const wchar_t *caption, cPassword *pwd )
	{
		updateString( &m_dlgText, dlgText );
		updateString( &m_caption, caption );
		m_pwd = pwd;

		m_IDText = IDText;
		m_IDPwdA = IDPwdA;
		m_IDPwdB = IDPwdB;
		m_errIDS_T = errIDS_T;
	}

	//
	//	Dialog with confirmation has to be initialised by following constructor
	//
	cPasswordDlg( const wchar_t *dlgText, int IDText, int IDPwdA, int IDPwdB, int errIDS_T[ecpeQ], const wchar_t *caption, cPassword *pwd )
	{
		m_dlgText = NULL;
		m_caption = NULL;
		Init( dlgText, IDText, IDPwdA, IDPwdB, errIDS_T, caption, pwd );
	}

	//
	// Dialog without confirmation has to be initialised by following constructor
	//
	cPasswordDlg( const wchar_t *dlgText, int IDText, int IDPwd, const wchar_t *caption, cPassword *pwd )
	{
		m_dlgText = NULL;
		m_caption = NULL;
		Init( dlgText, IDText, IDPwd, 0, NULL, caption, pwd );
	}
	
	~cPasswordDlg()
	{
		deleteString( &m_dlgText );
		deleteString( &m_caption );
	}


#ifdef _useKBFilter_
	static int CheckAndGetPassword( cPassword *pwd, cKbdEdit *edtA, cKbdEdit *edtB, cPassword *pPwdConfirm = NULL );
	static int CheckAndGetPassword( cPassword *pwd, cKbdEdit *edt );
#else
	static int CheckAndGetPassword( cPassword *pwd, HWND hEdtA, HWND hEdtB = NULL, cPassword *pPwdConfirm = NULL );
#endif // #ifdef _useKBFilter_

	//cPasswordDlg( const char *dlgText, int IDText, int IDPwd, const char *caption, cPassword *pwd )
	//{
	//	bf[0].Add( dlgText );
	//	bf[1].Add( caption );

	//	m_dlgText = ( bf[0].isEmpty() ? NULL : bf[0].m_szBody );
	//	m_caption = ( bf[1].isEmpty() ? NULL : bf[1].m_szBody );
	//	m_pwd = pwd;

	//	m_IDText = IDText;
	//	m_IDPwdA = IDPwd;
	//	m_IDPwdB = 0;
	//	m_errIDS_T = NULL;
	//}

 
	LRESULT RealDlgProc( UINT uMsg, WPARAM wParam, LPARAM lParam );
};

//
//	class cPasswordDlgLoop
//		Calls DoModal in loop.
//		If DoModal returns IDOK calls pwdActor otherwise stops loop.
//		If pwdActor returns enum from actorRes :
//			- errInvalidPassword,	calls showResult with m_idsInvalidPassword and continue loop.
//			- errNo,				stops loop with result IDOK.
//			- errInProcessing,		stops loop with result IDABORT.
//
class cPasswordDlgLoop : public cPasswordDlg
{
protected:
	cPassword m_pwd;
	int m_idsInvalidPassword;
	int m_idsErrorInProcessing;
public:
	
	cPasswordDlgLoop(
		const wchar_t *dlgText, int IDText, int IDPwd, const wchar_t *caption,
		int idsInvalidPassword, int idsErrorInProcessing )
			: cPasswordDlg( dlgText, IDText, IDPwd, caption, &m_pwd )
	{
		m_idsInvalidPassword = idsInvalidPassword;
		m_idsErrorInProcessing = idsErrorInProcessing;
	}

	enum actorRes { errNo = 0, errInvalidPassword, errInProcessing };

	virtual int pwdActor() = 0;

	virtual void showResult( UINT ids )
	{
		DlgMessageBoxT( ids, MB_ICONINFORMATION  );
	}

	INT_PTR DoModalLoop( HINSTANCE hInst, UINT idRes, HWND hWnd, DLGPROC externalDlgProc = NULL )
	{
		while( true )
		{
			INT_PTR iRes = DoModal( hInst, idRes, hWnd, externalDlgProc );
			if ( iRes != IDOK )
				return iRes;
		
			int res = pwdActor();

			if ( res == errNo )
				return IDOK;

			showResult( res == errInvalidPassword ? m_idsInvalidPassword : m_idsErrorInProcessing );

			if ( res != errInvalidPassword )
				return IDABORT;
		}
	}
};
