#include "ShellLink.h"

#include "UnicodeFuncs.h"

#include <shlobj.h>
#include <io.h>
#include <stdio.h>

// GetLinkInfo() fills the filename and path buffer
// with relevant information.
// hWnd         - calling application's window handle.
// 
// lpszLinkName - name of the link file passed into the function.
// 
// lpszPath     - the buffer that receives the file's path name.
// 
// lpszDescription - the buffer that receives the file's
// description.
//BOOL
//GetLinkInfo( HWND    hWnd,
//            LPCTSTR lpszLinkName,
//            LPSTR   lpszPath,
//            LPSTR   lpszDescription)
//{
//	
//	HRESULT hres;
//	IShellLink *pShLink;
//	WIN32_FIND_DATAA wfd;
//	
//	if ( lpszPath )		   *lpszPath = '\0';
//	if ( lpszDescription ) *lpszDescription = '\0';
//	
//	// Call CoCreateInstance to obtain the IShellLink
//	// Interface pointer. This call fails if
//	// CoInitialize is not called, so it is assumed that
//	// CoInitialize has been called.
//	hres = CoCreateInstance( CLSID_ShellLink,
//		NULL,
//		CLSCTX_INPROC_SERVER,
//		IID_IShellLink,
//		(LPVOID *)&pShLink );
//	
//	if (SUCCEEDED(hres))
//	{
//		IPersistFile *ppf;
//		
//		// The IShellLink Interface supports the IPersistFile
//		// interface. Get an interface pointer to it.
//		hres = pShLink->QueryInterface(	IID_IPersistFile, (LPVOID *)&ppf );
//
//		if (SUCCEEDED(hres))
//		{
//			cWCStr wcs( lpszLinkName );
//			
//			// Load the file.
//			hres = ppf->Load( wcs.m_szBody, STGM_READ );
//			if (SUCCEEDED(hres))
//			{
//				// Resolve the link by calling the Resolve() interface function.
//				// This enables us to find the file the link points to even if
//				// it has been moved or renamed.
//				
//				hres = pShLink->Resolve( hWnd, SLR_ANY_MATCH | SLR_NO_UI );//  | SLR_NOSEARCH );
//
//				if (SUCCEEDED(hres))
//				{
//					// Get the path of the file the link points to.
//					if ( lpszPath != NULL )
//						hres = pShLink->GetPath( lpszPath, MAX_PATH, &wfd, SLGP_SHORTPATH );
//					
//					// Only get the description if we successfully got the path
//					// (We can't return immediately because we need to release ppf &
//					//  pShLink.)
//					if( SUCCEEDED( hres ) && ( lpszDescription != NULL ) )
//					{
//						// Get the description of the link.
//						hres = pShLink->GetDescription( lpszDescription, MAX_PATH );
//					}
//				}
//			}
//			ppf->Release();
//		}
//		pShLink->Release();
//	}
//
//	return SUCCEEDED(hres);
//}

BOOL
GetLinkInfoT(
	HWND    hWnd,
	const wchar_t *lpszLinkName,
	wchar_t *lpszPath,
	wchar_t *lpszDescription )
{
	if ( !_isWinNT )
		return false;
	//{
	//	cMBStr link( lpszLinkName );

	//	char bf1[ MAX_PATH ], bf2[ MAX_PATH ];

	//	char *path = ( lpszPath == NULL ? NULL : bf1 );
	//	char *desc = ( lpszDescription == NULL ? NULL : bf2 );

	//	BOOL res = GetLinkInfo( hWnd, link.m_szBody, path, desc );

	//	if ( !res )
	//		return false;

	//	if ( path != NULL )	_mbstowcs( lpszPath, MAX_PATH, path );
	//	if ( desc != NULL )	_mbstowcs( lpszDescription, MAX_PATH, desc );

	//	return true;
	//}
	
	HRESULT hres;
	IShellLinkW *pShLink;
	WIN32_FIND_DATAW wfd;
	
	if ( lpszPath )		   *lpszPath = L'\0';
	if ( lpszDescription ) *lpszDescription = L'\0';
	
	// Call CoCreateInstance to obtain the IShellLink
	// Interface pointer. This call fails if
	// CoInitialize is not called, so it is assumed that
	// CoInitialize has been called.
	hres = CoCreateInstance( CLSID_ShellLink,
		NULL,
		CLSCTX_INPROC_SERVER,
		IID_IShellLinkW,
		(LPVOID *)&pShLink );
	
	if (SUCCEEDED(hres))
	{
		IPersistFile *ppf;
		
		// The IShellLink Interface supports the IPersistFile
		// interface. Get an interface pointer to it.
		hres = pShLink->QueryInterface(	IID_IPersistFile, (LPVOID *)&ppf );

		if (SUCCEEDED(hres))
		{
			hres = ppf->Load( lpszLinkName, STGM_READ );
			if (SUCCEEDED(hres))
			{
				// Resolve the link by calling the Resolve() interface function.
				// This enables us to find the file the link points to even if
				// it has been moved or renamed.
				
				hres = pShLink->Resolve( hWnd, SLR_ANY_MATCH | SLR_NO_UI );//  | SLR_NOSEARCH );

				if (SUCCEEDED(hres))
				{
					// Get the path of the file the link points to.
					if ( lpszPath != NULL )
						hres = pShLink->GetPath( lpszPath, MAX_PATH, &wfd, SLGP_SHORTPATH );
					
					// Only get the description if we successfully got the path
					// (We can't return immediately because we need to release ppf &
					//  pShLink.)
					if( SUCCEEDED( hres ) && ( lpszDescription != NULL ) )
					{
						// Get the description of the link.
						hres = pShLink->GetDescription( lpszDescription, MAX_PATH );
					}
				}
			}

			ppf->Release();
		}

		pShLink->Release();
	}

	return SUCCEEDED(hres);
}

HICON GetShellIcon( const char *path, int iconNum, bool expandRequired )
{
	cWCStr wcsPath( path );

	return GetShellIconT( wcsPath.m_szBody, iconNum, expandRequired );
}

HICON GetShellIconT( const wchar_t *path, int iconNum, bool expandRequired )
{
	// Expands system path f.e. "%windir%\\calc.exe" to real path.

	wchar_t data[ MAX_PATH ];

	if ( expandRequired )
	{
		DWORD len = ExpandEnvironmentStringsT( path, data, maxLen( data ) );

		if ( len == 0 )
			return NULL;

		path = data;
	}

	// Gets shell icon

	UINT uFlags = SHGFI_ICON | SHGFI_USEFILEATTRIBUTES | SHGFI_SMALLICON;

	SHFILEINFOW sfi;
	memset( &sfi, 0, sizeof( sfi ) );

	sfi.iIcon = iconNum;

	if ( SHGetFileInfoT( path, 0, &sfi, sizeof(sfi), uFlags ) )
		return sfi.hIcon;
	else
		return NULL;
}

//
//	GetAppIcon
//		Reads application default icon if it was registerred in:
//			HKEY_LOCAL_MACHINE\SOFTWARE\Classes\Applications
//		Returns icon handler or NULL if error.
//
HICON GetAppIcon( const char *appName, int iconNum ) // iconNum == -1 means default icon
{
	cWCStr wcsAppName( appName );

	return GetAppIconT( wcsAppName.m_szBody, iconNum );
}

HICON GetAppIconT( const wchar_t *appName, int iconNum, wchar_t *appPath ) // iconNum == -1 means default icon
{
	const wchar_t *szApplications = L"SOFTWARE\\Classes\\Applications\\";

	wchar_t keyName[ MAX_PATH ];

	wcscat( wcscpy( keyName, szApplications ), appName );

	HKEY hKey;
	wchar_t data[ MAX_PATH ];
	DWORD dwType = REG_SZ;

	LONG res = RegOpenKeyExT( HKEY_LOCAL_MACHINE, keyName, 0, KEY_QUERY_VALUE, &hKey );

	if ( res == ERROR_SUCCESS )
	{
		HKEY hSubKey;

		DWORD dataSize;

		res = RegOpenKeyExT( hKey, L"DefaultIcon", 0, KEY_QUERY_VALUE, &hSubKey );

		if ( res == ERROR_SUCCESS ) // DefaultIcon field is valid
		{
			dataSize = sizeof( data );

			res = RegQueryValueExT( hSubKey, L"", NULL, &dwType, (BYTE *)data, &dataSize ); // Queries default value.

			RegCloseKey( hSubKey );
		}

		if ( res != ERROR_SUCCESS )
		{
			// Explorer has the value for icon.
				
			dataSize = sizeof( data );
			res = RegQueryValueExT( hKey, L"TaskbarGroupIcon", NULL, &dwType, (BYTE *)data, &dataSize );
		}

		if ( res == ERROR_SUCCESS )
		{
			RegCloseKey( hKey ); 

			// Sample:
			// (Default), REG_EXPAND_SZ, "%ProgramFiles%\Windows NT\Accessories\WORDPAD.EXE,1"

			int num;

			/* It does not work properly :(
			char path[ MAX_PATH ];
			int i = sscanf( data, "%s,%d", path, &num );

			if ( i == 0 )
				return NULL;

			return GetShellIcon( path, num, dwType == REG_EXPAND_SZ );
			*/

			int pos = wcslen( data );
			while( pos-- )
			{
				if ( data[ pos ] == L',' )
				{
					data[ pos ] = L'\0';
					num = _wtoi( data + pos + 1 );
					break;
				}
			}

			if ( pos == -1 )
				return NULL;

			if ( iconNum != -1 )
				num = iconNum;

			if ( appPath != NULL )
				wcscpy( appPath, data );

			return GetShellIconT( data, num, dwType == REG_EXPAND_SZ );
		}

		// Tries to read command with executable path name.
		// Subkey:
		//	shell\open\command
		//	Default value
		//	"%ProgramFiles%\Windows NT\Accessories\WORDPAD.EXE" "%1"

		res = RegOpenKeyExT( hKey, L"shell\\open\\command", 0, KEY_QUERY_VALUE, &hSubKey );

		if ( res != ERROR_SUCCESS )
		{
			res = RegOpenKeyExT( hKey, L"shell\\edit\\command", 0, KEY_QUERY_VALUE, &hSubKey );
		}

		dataSize = sizeof( data );

		if ( res == ERROR_SUCCESS )
		{
			res = RegQueryValueExT( hSubKey, L"", NULL, &dwType, (BYTE *)data, &dataSize ); // Queries default value.

			RegCloseKey( hSubKey );
		}

		RegCloseKey( hKey );
	}

	if ( res != ERROR_SUCCESS )
	{
		// Tries to find the app in system directory.

		wchar_t path[ MAX_PATH ];

		GetSystemDirectoryT( path, maxLen( path ) );
		wcscat( wcscat( path, L"\\" ), appName );

		if ( !FileExistsT( path ) )
		{
			GetWindowsDirectoryT( path, maxLen( path ) );
			wcscat( wcscat( path, L"\\" ), appName );

			if ( !FileExistsT( path ) )
				return NULL;
		}

		FormatT( data, maxLen( data ), L"%s %d", path, iconNum );
	}


	// Sample:
	// (Default), REG_EXPAND_SZ, ""%ProgramFiles%\Windows NT\Accessories\WORDPAD.EXE" "%1""

	wchar_t *path;
	if ( data[0] == L'\"' )
	{
		path = (data + 1);

		for ( wchar_t *p = path; ; p++ )
		{
			if ( *p == L'\"' )
			{
				*p = L'\0';
				break;
			}
			else
			if ( *p == L'\0' ) // The string has to contain '\"' symbol
				return NULL;
		}
	}
	else
	{
		// %ProgramFiles%\Windows\Accessories\WORDPAD.EXE %1

		path = data;
		wchar_t *p;

		for ( p = path; ; p++ )
		{
			if ( *p == L' ' )
			{
				*p = L'\0';
				break;
			}
			else
			if ( *p == L'\0' ) // The command has parameter.
				return NULL;
		}

		// MSPAINT.EXE %1
		bool localName = true;

		for( p = path; *p; p++ )
		{
			if ( *p == L'\\' )
			{
				localName = false;
				break;
			}
		}

		if ( localName )
		{
			wchar_t sysDir[ MAX_PATH ];
			GetSystemDirectoryT( sysDir, maxLen( sysDir ) );
			wcscat( wcscat( sysDir, L"\\" ), path );
			wcscpy( path, sysDir );
		}
	}

	int num = 0;

	if ( iconNum != -1 )
		num = iconNum;

	if ( appPath != NULL )
		wcscpy( appPath, path );

	return GetShellIconT( path, num, dwType == REG_EXPAND_SZ );
}

