//
//	Unicode functions for windows 95
//	For Windows TN this function will be changed by macroses at UnicodeFuncs.h .
//	F.e. 
// #define GetFullPathNameT GetFullPathNameW
//

DWORD GetFullPathNameT( 
    LPCWSTR lpFileName,
    DWORD nBufferLength,
    LPWSTR lpBuffer,
    LPWSTR *lpFilePart
	)
{
	DWORD length;

	if ( _isWinNT )
		length = GetFullPathNameW( lpFileName, nBufferLength, lpBuffer, lpFilePart );
	else
	{
		char lpFileNameA[ MAX_PATH ];
		char lpBufferA[ MAX_PATH ];
		_wcstombs( lpFileNameA, sizeof( lpFileNameA ), lpFileName );

		length = GetFullPathNameA( lpFileNameA, sizeof( lpBufferA ), lpBufferA, NULL );

		size_t num;
		mbstowcs_s(	&num, lpBuffer, nBufferLength, lpBufferA, _TRUNCATE );
	}

	return length;
}

BOOL
WaitNamedPipeT(
	LPCWSTR lpNamedPipeName,
	DWORD nTimeOut
	)
{
	if ( _isWinNT )
		return WaitNamedPipeW( lpNamedPipeName, nTimeOut );
	else
	{
		cMBStr mbs( lpNamedPipeName );
		return WaitNamedPipeA( mbs.m_szBody, nTimeOut );
	}
}

HANDLE
OpenSemaphoreT(
    DWORD dwDesiredAccess,
    BOOL bInheritHandle,
    LPCWSTR lpName
    )
{
	if ( _isWinNT )
		return OpenSemaphoreW( dwDesiredAccess, bInheritHandle, lpName );
	else
		return OpenSemaphoreA( dwDesiredAccess, bInheritHandle, cMBStr(lpName) );
}

HANDLE
CreateSemaphoreT(
    LPSECURITY_ATTRIBUTES lpSemaphoreAttributes,
    LONG lInitialCount,
    LONG lMaximumCount,
    LPCWSTR lpName
)
{
	if ( _isWinNT )
		return CreateSemaphoreW( lpSemaphoreAttributes, lInitialCount, lMaximumCount, lpName );
	else
		return CreateSemaphoreA( lpSemaphoreAttributes, lInitialCount, lMaximumCount, cMBStr(lpName) );
}

int renameT( const wchar_t *oldName, const wchar_t *newName )
{
	if ( _isWinNT )
		return _wrename( oldName, newName );

	cMBStr oldA( oldName );
	cMBStr newA( newName );

#define EINVAL          22

	if ( oldA.m_szBody == NULL ) return EINVAL;
	if ( newA.m_szBody == NULL ) return EINVAL;

	return rename( oldA.m_szBody, newA.m_szBody );
}

BOOL
SetWindowTextT(
    HWND hWnd,
    LPCWSTR lpString)
{
	if ( _isWinNT )
		return SetWindowTextW( hWnd, lpString );

	char stringA[ 1024 ];
	_wcstombs( stringA, sizeof( stringA ), lpString );

	return SetWindowTextA( hWnd, stringA );
}

int ListView_InsertItemT( HWND hwnd, LVITEMW *pitem ) 
{
	if ( _isWinNT )
		return (int)SendMessageW((hwnd), LVM_INSERTITEMW, 0, (LPARAM)(const LVITEMW*)(pitem));

	char buffA[ 1024 ];
	if ( pitem->mask & LVIF_TEXT )
		_wcstombs( buffA, sizeof( buffA ), pitem->pszText );

	LVITEMA lviA;

	lviA.mask		= pitem->mask;
	lviA.iItem		= pitem->iItem;
	lviA.iSubItem	= pitem->iSubItem;
	lviA.state		= pitem->state;
	lviA.stateMask	= pitem->stateMask;
	lviA.pszText	= buffA;
	lviA.cchTextMax	= pitem->cchTextMax;
	lviA.iImage		= pitem->iImage;
	lviA.lParam		= pitem->lParam;

#if (_WIN32_IE >= 0x0300)
	lviA.iIndent	= pitem->iIndent;
#endif

	return (int)SendMessageA((hwnd), LVM_INSERTITEMA, 0, (LPARAM)&lviA );
}

BOOL ListView_GetColumnT( HWND hwnd, int iCol, LV_COLUMNW *pcol)
{
	if ( _isWinNT )
		return (BOOL)SendMessageW((hwnd), LVM_GETCOLUMNW, (WPARAM)iCol, (LPARAM)pcol);

	BOOL ok = (BOOL)SendMessageA((hwnd), LVM_GETCOLUMNA, (WPARAM)iCol, (LPARAM)pcol);

	// the pcol->pszText is filled as mulybyte 

	convertToWchar( (char*)pcol->pszText, pcol->cchTextMax * sizeof( wchar_t ) );

	return ok;
}

BOOL ListView_SetColumnT( HWND hwnd, int iCol, const LV_COLUMNW *pcol)
{
	if ( _isWinNT )
		return (BOOL)SendMessageW((hwnd), LVM_SETCOLUMNW, (WPARAM)iCol, (LPARAM)pcol);

	LV_COLUMNA lvcA;
	memcpy( &lvcA, pcol, sizeof( lvcA ) );

	cMBStr mbs( pcol->pszText );

	if ( ( pcol->pszText != NULL ) && ( pcol->mask & LVCF_TEXT ) )
		lvcA.pszText = mbs.m_szBody; // allocateAndCopyStr( pcol->pszText );
	else
		lvcA.pszText = 0;

	BOOL ok = (BOOL)SendMessageA((hwnd), LVM_SETCOLUMNA, (WPARAM)iCol, (LPARAM)&lvcA );

	// freeStr( lvcA.pszText );

	return ok;
}

int ListView_InsertColumnT( HWND hwnd, int iCol, const LV_COLUMNW *pcol )
{
	if ( _isWinNT )
		return (int)SendMessageW((hwnd), LVM_INSERTCOLUMNW, (WPARAM)(int)(iCol), (LPARAM)(const LV_COLUMNW *)(pcol));

	LV_COLUMNA lvcA;
	memcpy( &lvcA, pcol, sizeof( lvcA ) );

	cMBStr mbs( pcol->pszText );

	if ( ( pcol->pszText != NULL ) && ( pcol->mask & LVCF_TEXT ) )
		lvcA.pszText = mbs.m_szBody; // allocateAndCopyStr( pcol->pszText );
	else
		lvcA.pszText = 0;

	int res = ListView_InsertColumn(hwnd, iCol, &lvcA );

	// freeStr( lvcA.pszText );

	return res;
}

BOOL ListView_GetItemT( HWND hwnd, LVITEMW *pitem )
{
	if ( _isWinNT )
	{
		return (BOOL)SendMessageW((hwnd), LVM_GETITEMW, 0, (LPARAM)(LV_ITEMW *)(pitem));
	}
	else
	{
		char buffA[ 1024 ];
		LVITEMA lviA;
		lviA.mask		= pitem->mask;
		lviA.iItem		= pitem->iItem;
		lviA.iSubItem	= pitem->iSubItem;
		lviA.state		= pitem->state;
		lviA.stateMask	= pitem->stateMask;
		lviA.pszText	= buffA;
		lviA.cchTextMax	= maxLen( buffA ); //pitem->cchTextMax;
		lviA.iImage		= pitem->iImage;
		lviA.lParam		= pitem->lParam;

#if (_WIN32_IE >= 0x0300)
		lviA.iIndent	= pitem->iIndent;
#endif
#if (_WIN32_IE >= 0x560)
		lviA.iGroupId	= pitem->iGroupId;
		lviA.cColumns	= pitem->cColumns;
		lviA.puColumns	= pitem->puColumns;
#endif

		BOOL res = (BOOL)SendMessageA((hwnd), LVM_GETITEMA, 0, (LPARAM)(LV_ITEMA *)(&lviA));

		if ( !res )
			return FALSE;

		if ( pitem->mask & LVIF_TEXT )
		{
			size_t num;
			mbstowcs_s( &num, pitem->pszText, pitem->cchTextMax, buffA, _TRUNCATE );
		}

		pitem->mask			= lviA.mask;
		pitem->iItem		= lviA.iItem;
		pitem->iSubItem		= lviA.iSubItem;
		pitem->state		= lviA.state;
		pitem->stateMask	= lviA.stateMask;
		pitem->iImage		= lviA.iImage;
		pitem->lParam		= lviA.lParam;

#if (_WIN32_IE >= 0x0300)
		pitem->iIndent	= lviA.iIndent;
#endif
#if (_WIN32_IE >= 0x560)
		pitem->iGroupId	= lviA.iGroupId;
		pitem->cColumns	= lviA.cColumns;
		pitem->puColumns	= lviA.puColumns;
#endif
		return TRUE;
	}
}

BOOL ListView_SetItemT( HWND hwnd, LVITEMW *pitem )
{
	if ( _isWinNT )
		return (BOOL)SendMessageW((hwnd), LVM_SETITEMW, 0, (LPARAM)(const LV_ITEMA FAR*)(pitem));

	char buffA[ 1024 ];
	if ( pitem->mask & LVIF_TEXT )
		_wcstombs( buffA, sizeof( buffA ), pitem->pszText );

	LVITEMA lviA;

	lviA.mask		= pitem->mask;
	lviA.iItem		= pitem->iItem;
	lviA.iSubItem	= pitem->iSubItem;
	lviA.state		= pitem->state;
	lviA.stateMask	= pitem->stateMask;
	lviA.pszText	= buffA;
	lviA.cchTextMax	= pitem->cchTextMax;
	lviA.iImage		= pitem->iImage;
	lviA.lParam		= pitem->lParam;

#if (_WIN32_IE >= 0x0300)
	lviA.iIndent	= pitem->iIndent;
#endif

	return (BOOL)SendMessageA((hwnd), LVM_SETITEMA, 0, (LPARAM)(const LV_ITEMA FAR*)(&lviA));
}

int ListView_FindItemT( HWND hwnd, int iStart, const LVFINDINFOW *plvfi)
{
	if ( _isWinNT )
		return (int)SendMessageW((hwnd), LVM_FINDITEMW, (WPARAM)(iStart), (LPARAM)(plvfi) );

	LVFINDINFOA vfiA;
	memcpy( &vfiA, plvfi, sizeof( vfiA ) );

	cMBStr mbs( plvfi->psz );

	vfiA.psz = mbs.m_szBody;

    return (int)SendMessageA((hwnd), LVM_FINDITEMA, (WPARAM)(iStart), (LPARAM)(&vfiA));
}

BOOL
InsertMenuItemT(
    HMENU hm,
    UINT  ui,
    BOOL  bl,
    LPCMENUITEMINFOW pi
    )
{
	assert( sizeof( MENUITEMINFOA ) == sizeof( MENUITEMINFOW ) );

	if ( _isWinNT )
	{
		BOOL ok = InsertMenuItemW( hm, ui, bl, pi );

		int err = GetLastError();

		return ok;
	}

	MENUITEMINFOA *piA = (MENUITEMINFOA *)pi;
	char bf[ 1024 ];

#if (_WIN32_WINNT >= 0x0500)
	assert( MENUITEMINFO_SIZE_VERSION_400 == ( sizeof( *piA ) - sizeof( piA->hbmpItem ) ) );
#endif

	piA->cbSize = MENUITEMINFO_SIZE_VERSION_400;

	bool dataTypeIsString = false;

	if ( piA->fMask & MIIM_FTYPE )
	{
		piA->fMask &= ~MIIM_FTYPE;
		piA->fMask |= MIIM_TYPE;
	}

	if ( piA->fMask & MIIM_STRING )
	{
		dataTypeIsString = true;

		piA->fMask &= ~MIIM_STRING;
		if ( ( piA->fMask & MIIM_TYPE ) == 0 )
			piA->fType = 0;

		piA->fMask |= MIIM_TYPE;
		piA->fType |= MFT_STRING;
	}
	
	if ( ( piA->fMask & MIIM_TYPE ) && ( piA->fType == MFT_STRING ) )
	{
		dataTypeIsString = true;
	}

	if ( dataTypeIsString )
	{	
		_wcstombs( bf, sizeof( bf ), pi->dwTypeData );
		piA->dwTypeData = bf;
//		MessageBox( 0, bf, "InsertMenuItemA", MB_OK );
	}

	BOOL ok = InsertMenuItemA( hm, ui, bl, piA ); 

	if ( !ok )
	{
		ODS_fls( L"Assertion!" );
	}

	return ok;
}

BOOL
ModifyMenuT(
    IN HMENU hMnu,
    IN UINT uPosition,
    IN UINT uFlags,
    IN UINT_PTR uIDNewItem,
    IN LPCWSTR lpNewItem
    )
{
	if ( _isWinNT )
		return ModifyMenuW( hMnu, uPosition, uFlags, uIDNewItem, lpNewItem );

	char bf[ 1024 ];
	_wcstombs( bf, sizeof( bf ), lpNewItem );
	return ModifyMenuA( hMnu, uPosition, uFlags, uIDNewItem, bf );
}

BOOL
SetMenuItemInfoT(
    IN HMENU hMenu,
    IN UINT uItem,
    IN BOOL fByPosition,
    IN LPMENUITEMINFOW lpmii
    )
{
	if ( _isWinNT )
		return SetMenuItemInfoW( hMenu, uItem, fByPosition, lpmii );

	MENUITEMINFOA miiA = *( MENUITEMINFOA * )lpmii;

	int wcSize = lpmii->cch;

	miiA.cbSize = MENUITEMINFO_SIZE_VERSION_400;

	//
	//	Pure windows 95 compatibility
	//
	if ( ( miiA.fMask & MIIM_STRING ) != 0 )
	{
		miiA.fMask &= ~MIIM_STRING;
		if ( ( miiA.fMask & MIIM_TYPE ) == 0 )
			miiA.fType = 0;

		miiA.fMask |= MIIM_TYPE;
		miiA.fType |= MFT_STRING;
	}

	cMBStr mbs( lpmii->dwTypeData );

	if ( ( miiA.fMask & MIIM_TYPE  ) && ( miiA.fType == MFT_STRING ) ) 
	{
		miiA.dwTypeData = mbs.m_szBody; // allocateAndCopyStr( lpmii->dwTypeData );
	}

	BOOL ok = SetMenuItemInfoA( hMenu, uItem, fByPosition, &miiA );

	// freeStr( miiA.dwTypeData );

	if ( !ok )
	{
		ODS_fls( L"Assertion!" );
		return false;
	}


	return ok;
}

int
GetMenuStringT(
    HMENU hMenu,
    UINT uIDItem,
    LPWSTR lpString,
    int cchMax,
    UINT flags )
{
	if ( _isWinNT )
		return GetMenuStringW( hMenu, uIDItem, lpString, cchMax, flags );

	cAutoDeallocate < char > ad( cchMax );

	if ( ad.m_ptr == NULL )
		return 0;

	int res = GetMenuStringA( hMenu, uIDItem, ad.m_ptr, ad.m_len, flags );

	if ( res == 0 )
		return res;

	__mbstowcs( lpString, cchMax, ad.m_ptr, -1 );

	return res;
}

BOOL
GetMenuItemInfoT(
    IN HMENU hMenu,
    IN UINT uItem,
    IN BOOL fByPosition,
    IN OUT LPMENUITEMINFOW lpmii
    )
{
	if ( _isWinNT )
		return GetMenuItemInfoW( hMenu, uItem, fByPosition, lpmii );

	MENUITEMINFOA *miiA = ( MENUITEMINFOA * )lpmii;

	//
	//	Unfortunately pure winfows 95 does not recognise size with hbmpItem item,
	//	and does not works with fMask MIIM_FTYPE, we use MIIM_TYPE insted
	//
	miiA->cbSize = MENUITEMINFO_SIZE_VERSION_400; // sizeof( *miiA ) - sizeof( miiA->hbmpItem );

	if ( ( miiA->fMask & MIIM_FTYPE ) != 0 )
	{
		miiA->fMask &= ~MIIM_FTYPE;
		miiA->fMask |= MIIM_TYPE;
	}

	if ( ( miiA->fMask & MIIM_STRING ) != 0 )
	{
		miiA->fMask &= ~MIIM_STRING;
		if ( ( miiA->fMask & MIIM_TYPE ) == 0 )
			miiA->fType = 0;

		miiA->fMask |= MIIM_TYPE;
		miiA->fType |= MFT_STRING;
	}

	int wcSize = lpmii->cch;

	BOOL ok = GetMenuItemInfoA( hMenu, uItem, fByPosition, miiA );

	if ( !ok )
	{
		//ODS_fls( L"Assertion!" );
		return false;
	}

	if ( ( miiA->fMask & MIIM_TYPE  ) && ( miiA->fType == MFT_STRING ) ) 
	{
		cWCStr wcs( miiA->dwTypeData );
		wcsncpy_s( lpmii->dwTypeData, wcSize, wcs.m_szBody, _TRUNCATE );
//		convertToWchar( miiA->dwTypeData, wcSize * sizeof( wchar_t ) );
	}

	return ok;
}

BOOL GetTextExtentPoint32T( HDC hdc, LPCWSTR lpString, int cbString, LPSIZE lpSize )
{
	if ( _isWinNT )
		return GetTextExtentPoint32W( hdc, lpString, cbString, lpSize );

	cMBStr mbs( lpString );

	char *textA = mbs.m_szBody; // allocateAndCopyStr( lpString );
	BOOL ok = GetTextExtentPoint32A( hdc, textA, cbString, lpSize );
	// freeStr( textA );

	return ok;
}

int
DrawTextT(
    IN HDC hDC,
    IN LPCWSTR lpString,
    IN int nCount,
    IN OUT LPRECT lpRect,
    IN UINT uFormat )
{
	if ( _isWinNT )
		return DrawTextW( hDC, lpString, nCount, lpRect, uFormat );

	cMBStr mbs( lpString );
	char *textA = mbs.m_szBody; // allocateAndCopyStr( lpString );
	int res = DrawTextA( hDC, textA, nCount, lpRect, uFormat );
	// freeStr( textA );

	return res;
}

LRESULT
CallWindowProcT(
	WNDPROC lpPrevWndFunc,
	HWND hWnd,
	UINT Msg,
	WPARAM wParam,
	LPARAM lParam )
{
	if ( _isWinNT )
		return CallWindowProcW( lpPrevWndFunc, hWnd, Msg, wParam, lParam );

	return CallWindowProcA( lpPrevWndFunc, hWnd, Msg, wParam, lParam );
}

UINT
GetWindowsDirectoryT(
    LPWSTR lpBuffer,
    UINT uSize
    )
{
	if ( _isWinNT )
		return GetWindowsDirectoryW( lpBuffer, uSize );

	char buffA[ MAX_PATH ];
	UINT len = GetWindowsDirectoryA( buffA, sizeof( buffA ) );

	if ( len == 0 )
		return 0;

	size_t num = 0;
	mbstowcs_s( &num, lpBuffer, uSize, buffA, _TRUNCATE );
	return (UINT)num;
}

UINT
GetSystemDirectoryT(
    OUT LPWSTR lpBuffer,
    IN UINT uSize
    )
{
	if ( _isWinNT )
		return GetSystemDirectoryW( lpBuffer, uSize );

	char buffA[ MAX_PATH ];
	UINT len = GetSystemDirectoryA( buffA, sizeof( buffA ) );

	size_t num = 0;
	mbstowcs_s( &num, lpBuffer, uSize, buffA, _TRUNCATE );
	return (UINT)num;
//	return (UINT)mbstowcs( lpBuffer, buffA, uSize );
}

HINSTANCE
FindExecutableT(
	LPCWSTR lpFile,
	LPCWSTR lpDirectory,
	LPWSTR lpResult
	)
{
	if ( _isWinNT )
		return FindExecutableW( lpFile, lpDirectory, lpResult );

	char fileA[ MAX_PATH ];
	char directoryA[ MAX_PATH ] = "";
	char resultA[ MAX_PATH ];


	_wcstombs( fileA, sizeof( fileA ), lpFile );
	if ( lpDirectory )
		_wcstombs( directoryA, sizeof( directoryA ), lpDirectory );

	HINSTANCE hInst = FindExecutableA( fileA, directoryA, resultA );

	size_t num = 0;
	mbstowcs_s( &num, lpResult, MAX_PATH, resultA, _TRUNCATE );

	return hInst;
}

HINSTANCE 
ShellExecuteT( HWND hwnd, LPCWSTR lpOperation, LPCWSTR lpFile, LPCWSTR lpParameters, LPCWSTR lpDirectory, INT nShowCmd)
{
	if ( _isWinNT )
	{
		HINSTANCE hRes = ShellExecuteW( hwnd, lpOperation, lpFile, lpParameters, lpDirectory, nShowCmd );

		//
		//	03.04.2012
		//	ShellExecuteW does not recognize the long name convention...
		//
		//if ( hRes > (HINSTANCE)32 )
		//	return hRes;

		//if ( cLongName::isTooLongForShortName( lpFile ) )
		//	hRes = ShellExecuteW( hwnd, lpOperation, cLongName().generate( lpFile ), lpParameters, lpDirectory, nShowCmd );

		return hRes;
	}

	cMBStr mbsT[4];

	char *lpOperationA	= mbsT[0].SetAndGet( lpOperation );
	char *lpFileA		= mbsT[1].SetAndGet( lpFile );
	char *lpParametersA	= mbsT[2].SetAndGet( lpParameters );
	char *lpDirectoryA	= mbsT[3].SetAndGet( lpDirectory );

	return ShellExecuteA( hwnd, lpOperationA, lpFileA, lpParametersA, lpDirectoryA, nShowCmd );
}

BOOL ShellExecuteExT( LPSHELLEXECUTEINFOW lpExecInfo )
{
	if ( _isWinNT )
		return ShellExecuteExW( lpExecInfo );

	SHELLEXECUTEINFOA *shex = (SHELLEXECUTEINFOA *)lpExecInfo;

	memcpy( shex, lpExecInfo, sizeof( *shex ) );
	shex->cbSize = sizeof( *shex );

	cMBStr v( lpExecInfo->lpVerb );
	shex->lpVerb			= v.m_szBody; 

	cMBStr f( lpExecInfo->lpFile );
	shex->lpFile			= f.m_szBody; 

	cMBStr p( lpExecInfo->lpParameters );
	shex->lpParameters	= p.m_szBody; 

	cMBStr d( lpExecInfo->lpDirectory );
	shex->lpDirectory	= d.m_szBody; 

	cMBStr c( lpExecInfo->lpClass );
	shex->lpClass		= c.m_szBody;
  
	return ShellExecuteExA( shex );
}

UINT DragQueryFileT( HDROP hDrop, UINT num, LPWSTR wbuff, UINT wbflen )
{
	if ( _isWinNT )
		return DragQueryFileW( hDrop, num, wbuff, wbflen );

	char abuff[ MAX_PATH ];

	int len = min( wbflen, sizeof( abuff ) );

	UINT ur = DragQueryFileA( hDrop, num, abuff, len );

	size_t size;

	if ( wbuff != NULL )
		mbstowcs_s( &size, wbuff, wbflen, abuff, len );

	return ur;
}

UINT
GetDriveTypeT(
    LPCWSTR lpRootPathName
    )
{
	if ( _isWinNT )
		return GetDriveTypeW( lpRootPathName );

	char buffA[ MAX_PATH ];
	_wcstombs( buffA, sizeof( buffA ), lpRootPathName );

	return GetDriveTypeA( buffA );
}

BOOL
GetDiskFreeSpaceT(
    LPCWSTR lpRootPathName,
    LPDWORD lpSectorsPerCluster,
    LPDWORD lpBytesPerSector,
    LPDWORD lpNumberOfFreeClusters,
    LPDWORD lpTotalNumberOfClusters
    )
{
	if ( _isWinNT )
		return GetDiskFreeSpaceW(
					lpRootPathName,
					lpSectorsPerCluster,
					lpBytesPerSector,
					lpNumberOfFreeClusters,
					lpTotalNumberOfClusters );

	char rootA[ MAX_PATH ];

	_wcstombs( rootA, sizeof( rootA ), lpRootPathName );

	return GetDiskFreeSpaceA(
					rootA,
					lpSectorsPerCluster,
					lpBytesPerSector,
					lpNumberOfFreeClusters,
					lpTotalNumberOfClusters );
}

BOOL
WinHelpT(
    HWND hWndMain,
    LPCWSTR lpszHelp,
    UINT uCommand,
    DWORD dwData
    )
{
	if ( _isWinNT )
		return WinHelpW( hWndMain, lpszHelp, uCommand, dwData );

	char helpA[ MAX_PATH ];

	_wcstombs( helpA, sizeof( helpA ), lpszHelp );

	return WinHelpA( hWndMain, helpA, uCommand, dwData );
}

#ifdef _UseHtmlHelp_

HWND HtmlHelpT( HWND hwndCaller, LPCWSTR pszFile, UINT uCommand, DWORD_PTR dwData )
{
	// Note 1. the htmlHelp.lib has to be mantioned at included list.
	// Note 2. wow64 redirector does not allows to start HtmlHelp at all. 
	//	We have to stop redirection first and continue after. 
	//	See comments for cDisableWow64Redirection in kernelDll.h

	if ( _isWinNT )
		return HtmlHelpW( hwndCaller, pszFile, uCommand, dwData );

	cMBStr file( pszFile );
	cAutoDeallocate <char> data;

	if ( uCommand == HH_DISPLAY_INDEX )
	{
		if ( dwData != NULL )
		{
			data.AllocateAndCopy( getMBStr( (wchar_t*)dwData  ) );
			dwData = (DWORD_PTR)data.m_ptr;
		}
	}
	else
	if ( uCommand == HH_DISPLAY_TEXT_POPUP )
	{
		HH_POPUPW *php = (HH_POPUPW *)dwData;
		if ( php->pszText != NULL )
			php->pszText = ( wchar_t * )data.AllocateAndCopy( getMBStr( php->pszText ) );
	}

	return HtmlHelpA( hwndCaller, file.m_szBody, uCommand, dwData );
}

//struct sHelpParams
//{
//	HWND hWnd;
//	const wchar_t *pszFile;
//	UINT uCommand;
//	DWORD_PTR dwData;
//};
//
//static DWORD WINAPI threadHelp( void *lpParameter )
//{
//	sHelpParams *hp = ( sHelpParams * )lpParameter;
//
//	_HtmlHelpT( hp->hWnd, hp->pszFile, hp->uCommand, hp->dwData );
//
//	hp->hWnd = NULL;
//
//	return 0;
//}
//
//HWND HtmlHelpT( HWND hwndCaller, LPCWSTR pszFile, UINT uCommand, DWORD_PTR dwData )
//{
//	if ( uCommand != HH_DISPLAY_TOPIC )
//		return _HtmlHelpT( hwndCaller, pszFile, uCommand, dwData );
//
//	DWORD id;
//
//	sHelpParams hp;
//	hp.hWnd = hwndCaller;
//	hp.pszFile = pszFile;
//	hp.uCommand = uCommand;
//	hp.dwData = dwData;
//
//	HANDLE hThread = CreateThread( NULL, 0, threadHelp, &hp, 0, &id );
//
//	if ( hThread != NULL )
//	{
//		CloseHandle( hThread );
//
//		for( int i = 10; i--; )
//		{
//			Sleep( 10 );
//
//			if ( hp.hWnd == NULL )
//				break;
//		}
//
//		return HWND(1);
//	}
//
//	return 0;
//}

#endif // #ifdef _UseHtmlHelp_

BOOL TreeView_GetItemT( HWND hwnd, TV_ITEMW *pitem )
{
#ifndef _SupportW95
	return (BOOL)SendMessageW((hwnd), TVM_GETITEMW, 0, (LPARAM)(TV_ITEMW *)(pitem));
#else
	if ( _isWinNT )
//		return TreeView_GetItemW( hWnd, pitem );
		return (BOOL)SendMessageW((hwnd), TVM_GETITEMW, 0, (LPARAM)(TV_ITEMW *)(pitem));

	TV_ITEMA tviA;
	char buffA[ 1024 ];

	tviA.mask = pitem->mask;
	tviA.hItem = pitem->hItem;
	tviA.state = pitem->state;
	tviA.stateMask = pitem->stateMask;
	tviA.pszText = buffA;
	tviA.cchTextMax = sizeof( buffA );
	tviA.iImage = pitem->iImage;
	tviA.iSelectedImage = pitem->iSelectedImage;
	tviA.cChildren = pitem->cChildren;
	tviA.lParam = pitem->lParam;

	BOOL res = 
		(BOOL)SendMessageA((hwnd), TVM_GETITEMA, 0, (LPARAM)(TV_ITEMA *)(&tviA));
	//	TreeView_GetItemA( hWnd, &tviA );

	pitem->mask			= tviA.mask;
	pitem->hItem		= tviA.hItem;
	pitem->state		= tviA.state;
	pitem->stateMask	= tviA.stateMask;
	pitem->iImage		= tviA.iImage;
	pitem->iSelectedImage	= tviA.iSelectedImage;
	pitem->cChildren	= tviA.cChildren;
	pitem->lParam		= tviA.lParam;

	if ( pitem->mask & TVIF_TEXT )
		mbstowcs( pitem->pszText, tviA.pszText, pitem->cchTextMax );

	return res;
#endif
}

BOOL TreeView_SetItemT( HWND hwnd, const TV_ITEMW *pitem )
{
	if ( _isWinNT )
//		return TreeView_SetItemW( hWnd, pitem );
		return (BOOL)SendMessageW((hwnd), TVM_SETITEMW, 0, (LPARAM)(const TV_ITEMW *)(pitem));

	char buffA[ 1024 ];
	if ( pitem->mask & TVIF_TEXT )
		_wcstombs( buffA, sizeof( buffA ), pitem->pszText );

	TV_ITEMA tvi;

	tvi.mask = pitem->mask;
	tvi.hItem = pitem->hItem;
	tvi.state = pitem->state;
	tvi.stateMask = pitem->stateMask;
	tvi.pszText = buffA;
	tvi.cchTextMax = pitem->cchTextMax;
	tvi.iImage = pitem->iImage;
	tvi.iSelectedImage = pitem->iSelectedImage;
	tvi.cChildren = pitem->cChildren;
	tvi.lParam = pitem->lParam;

	return (BOOL)SendMessageA((hwnd), TVM_SETITEMA, 0, (LPARAM)(const TV_ITEMA *)(&tvi));
	//	TreeView_SetItemA( hwnd, &tvi );
}

HTREEITEM TreeView_InsertItemT( HWND hwnd, LPTV_INSERTSTRUCTW lpis )
{
	if ( _isWinNT )
//		return TreeView_GetItemW( hWnd, pitem );
		return (HTREEITEM)SendMessageW((hwnd), TVM_INSERTITEMW, 0, (LPARAM)(LPTV_INSERTSTRUCTW)(lpis));

	char buffA[ 1024 ];
	if ( lpis->item.mask & TVIF_TEXT )
		_wcstombs( buffA, sizeof( buffA ), lpis->item.pszText );

	TVINSERTSTRUCTA tvisA;
	tvisA.hParent = lpis->hParent;
	tvisA.hInsertAfter = lpis->hInsertAfter;
//	tvisA.item = lpis->item;
	tvisA.item.mask = lpis->item.mask;
	tvisA.item.hItem = lpis->item.hItem;
	tvisA.item.state = lpis->item.state;
	tvisA.item.stateMask = lpis->item.stateMask;
	tvisA.item.pszText = buffA; // lpis->item.pszText;
	tvisA.item.cchTextMax = sizeof( buffA ); // lpis->item.cchTextMax;
	tvisA.item.iImage = lpis->item.iImage;
	tvisA.item.iSelectedImage = lpis->item.iSelectedImage;
	tvisA.item.cChildren = lpis->item.cChildren;
	tvisA.item.lParam = lpis->item.lParam;

	return (HTREEITEM)SendMessageA((hwnd), TVM_INSERTITEMA, 0, (LPARAM)(LPTV_INSERTSTRUCTA)(&tvisA));
}

int
WINAPI
MessageBoxIndirectT(
    IN CONST MSGBOXPARAMSW *lpMsgBoxParams )
{
	if ( _isWinNT )
		return MessageBoxIndirectW( ( MSGBOXPARAMSW * )lpMsgBoxParams );


	MSGBOXPARAMSA mbpA;
	memcpy( &mbpA, lpMsgBoxParams, sizeof( mbpA ) );

	cMBStr mbsT[2];
	mbsT[0].Add( lpMsgBoxParams->lpszCaption );
	mbsT[1].Add( lpMsgBoxParams->lpszText );

	mbpA.lpszCaption	= mbsT[0].m_szBody; // allocateAndCopyStr( lpMsgBoxParams->lpszCaption );
	mbpA.lpszText		= mbsT[1].m_szBody; // allocateAndCopyStr( lpMsgBoxParams->lpszText );

	int res = MessageBoxIndirectA( &mbpA );

	//freeStr( (char*)mbpA.lpszCaption );
	//freeStr( (char*)mbpA.lpszText );

	return res;
}

BOOL SetCurrentDirectoryT( LPCWSTR lpPathName )
{
	//ODSFuncS( lpPathName );

	if ( _isWinNT )	return SetCurrentDirectoryW( lpPathName );
	else			return SetCurrentDirectoryA( cMBStr( lpPathName ) );
}

DWORD GetCurrentDirectoryT(
  DWORD nBufferLength,  // size of directory buffer
  LPWSTR lpBuffer       // directory buffer
)
{
	//ODSFunc;

	DWORD dwRes;

	if ( _isWinNT )
		dwRes = GetCurrentDirectoryW( nBufferLength, lpBuffer );
	else
	{
		cMBStr mbs( MAX_PATH );

		dwRes = GetCurrentDirectoryA( mbs.m_len, mbs.m_szBody );

		if ( dwRes != 0 )
			mbs.WriteTo( lpBuffer, nBufferLength );
	}
	
	//ODSFuncRes( lpBuffer );

	return dwRes;
}

BOOL
CreateProcessT(
    IN LPCWSTR lpApplicationName,
    IN LPWSTR lpCommandLine,
    IN LPSECURITY_ATTRIBUTES lpProcessAttributes,
    IN LPSECURITY_ATTRIBUTES lpThreadAttributes,
    IN BOOL bInheritHandles,
    IN DWORD dwCreationFlags,
    IN LPVOID lpEnvironment,
    IN LPCWSTR lpCurrentDirectory,
    IN LPSTARTUPINFOW lpStartupInfo,
    OUT LPPROCESS_INFORMATION lpProcessInformation
    )
{
	assert( sizeof( STARTUPINFOW ) == sizeof( STARTUPINFOA ) );

	if ( _isWinNT )
		return CreateProcessW(
				lpApplicationName,
				lpCommandLine,
				lpProcessAttributes,
				lpThreadAttributes,
				bInheritHandles,
				dwCreationFlags,
				lpEnvironment,
				lpCurrentDirectory,
				lpStartupInfo,
				lpProcessInformation
			);
	else
	{
		char cb[3][MAX_PATH];

		return CreateProcessA(
				getCharBuf( lpApplicationName, cb[0], maxLen( cb[0] ) ),
				getCharBuf( lpCommandLine, cb[1], maxLen( cb[1] ) ),
				lpProcessAttributes,
				lpThreadAttributes,
				bInheritHandles,
				dwCreationFlags,
				lpEnvironment,
				getCharBuf( lpCurrentDirectory, cb[2], maxLen( cb[2] ) ),
				(STARTUPINFOA *)lpStartupInfo,
				lpProcessInformation
			);
	}
}

LONG
RegCreateKeyExT (
	IN HKEY hKey,
	IN LPCWSTR lpSubKey,
	IN DWORD Reserved,
	IN LPWSTR lpClass,
	IN DWORD dwOptions,
	IN REGSAM samDesired,
	IN LPSECURITY_ATTRIBUTES lpSecurityAttributes,
	OUT PHKEY phkResult,
	OUT LPDWORD lpdwDisposition
    )
{
	if ( _isWinNT )
		return RegCreateKeyExW( hKey, lpSubKey, Reserved, lpClass, dwOptions, samDesired, lpSecurityAttributes,  phkResult, lpdwDisposition );
	
	cMBStr mbsT[2];

	mbsT[0].Add( lpSubKey );
	mbsT[1].Add( lpClass );

	char *subKeyA = mbsT[0].m_szBody; // allocateAndCopyStr( lpSubKey );
	char *classA =  mbsT[1].m_szBody; // allocateAndCopyStr( lpClass );

	LONG res = RegCreateKeyExA( hKey, subKeyA, Reserved, classA, dwOptions, samDesired, lpSecurityAttributes,  phkResult, lpdwDisposition );

	//freeStr( subKeyA );
	//freeStr( classA );

	return res;
}

LONG
RegCreateKeyT (
    IN HKEY hKey,
    IN LPCWSTR lpSubKey,
    OUT PHKEY phkResult
    )
{
	if ( _isWinNT )	return RegCreateKeyW( hKey, lpSubKey, phkResult );
	
	cMBStr mbs( lpSubKey );

	char *subKeyA = mbs.m_szBody; // allocateAndCopyStr( lpSubKey );
	LONG res = RegCreateKeyA( hKey, subKeyA, phkResult );
	// freeStr( subKeyA );

	return res;
}

LONG
RegOpenKeyExT (
    IN HKEY hKey,
    IN LPCWSTR lpSubKey,
    IN DWORD ulOptions,
    IN REGSAM samDesired,
    OUT PHKEY phkResult
    )
{
	if ( _isWinNT )
		return RegOpenKeyExW( hKey, lpSubKey, ulOptions, samDesired, phkResult );

	cMBStr mbs( lpSubKey );
	char *subKeyA = mbs.m_szBody; // allocateAndCopyStr( lpSubKey );
	LONG res = RegOpenKeyExA( hKey, subKeyA, ulOptions, samDesired, phkResult );
	//freeStr( subKeyA );

	return res;
}

LONG
RegQueryInfoKeyT (
    IN HKEY hKey,
    OUT LPWSTR lpClass,
    IN OUT LPDWORD lpcbClass,
    IN LPDWORD lpReserved,
    OUT LPDWORD lpcSubKeys,
    OUT LPDWORD lpcbMaxSubKeyLen,
    OUT LPDWORD lpcbMaxClassLen,
    OUT LPDWORD lpcValues,
    OUT LPDWORD lpcbMaxValueNameLen,
    OUT LPDWORD lpcbMaxValueLen,
    OUT LPDWORD lpcbSecurityDescriptor,
    OUT PFILETIME lpftLastWriteTime
    )
{
	LONG res = -1;
#ifndef _SupportW95
	if ( _isWinNT )
		res = RegQueryInfoKeyW( hKey, lpClass, lpcbClass, lpReserved, lpcSubKeys, lpcbMaxSubKeyLen,
								lpcbMaxClassLen, lpcValues, lpcbMaxValueNameLen, lpcbMaxValueLen,
								lpcbSecurityDescriptor, lpftLastWriteTime );
#else
	else
	{
		char *classA = NULL;
		int len = 0;

		cMBStr mbs;

		if ( ( lpcbClass != NULL ) && ( *lpcbClass != 0 ) )
		{
			len = *lpcbClass;
			classA = mbs.SetAndGet( len ); // new char[ len ];
		}

		res = RegQueryInfoKeyA( hKey, classA, lpcbClass, lpReserved, lpcSubKeys, lpcbMaxSubKeyLen,
									lpcbMaxClassLen, lpcValues, lpcbMaxValueNameLen, lpcbMaxValueLen,
									lpcbSecurityDescriptor, lpftLastWriteTime );

		if ( res != ERROR_MORE_DATA )
		{
			if ( classA != NULL )
			{
				mbstowcs( lpClass, classA, len );
			}
		}

		// freeStr( classA );
	}
#endif
	return res;
}

LONG
RegDeleteKeyT(
    IN HKEY hKey,
    IN LPCWSTR lpSubKey
    )
{
	if ( _isWinNT )
		return RegDeleteKeyW( hKey, lpSubKey );
	else
		return RegDeleteKeyA( hKey, cMBStr( lpSubKey ).m_szBody );
}

LONG
RegDeleteValueT(
    IN HKEY hKey,
    IN LPCWSTR lpValueName
    )
{
	if ( _isWinNT )
		return RegDeleteValueW( hKey, lpValueName );
	else
		return RegDeleteValueA( hKey, cMBStr( lpValueName ).m_szBody );
}

LONG
RegEnumKeyExT (
    IN HKEY hKey,
    IN DWORD dwIndex,
    OUT LPWSTR lpName,
    IN OUT LPDWORD lpcbName,
    IN LPDWORD lpReserved,
    IN OUT LPWSTR lpClass,
    IN OUT LPDWORD lpcbClass,
    OUT PFILETIME lpftLastWriteTime
    )
{
#ifndef _SupportW95
	return RegEnumKeyExW( hKey, dwIndex, lpName, lpcbName, lpReserved, lpClass, lpcbClass, lpftLastWriteTime );
#else
	if ( _isWinNT )
		return RegEnumKeyExW( hKey, dwIndex, lpName, lpcbName, lpReserved, lpClass, lpcbClass, lpftLastWriteTime );

	char *classA = NULL;
	char *nameA = NULL;

	DWORD cbName = 0;
	DWORD cbClass = 0;
	
	if ( lpcbClass != NULL )
	{
		cbClass = *lpcbClass;
		classA = new char[ cbClass ];
	}

	if ( lpcbName != NULL )
	{
		cbName = *lpcbName;
		nameA = new char[ cbName ];
	}

	LONG res = RegEnumKeyExA( hKey, dwIndex, nameA, lpcbName, lpReserved, classA, lpcbClass, lpftLastWriteTime );

	if ( nameA != NULL )
		mbstowcs( lpName, nameA, cbName );

	if ( classA != NULL )
		mbstowcs( lpClass, classA, cbClass );

	return res;
#endif
}

LONG
RegEnumValueT (
    IN HKEY hKey,
    IN DWORD dwIndex,
    OUT LPWSTR lpValueName,
    IN OUT LPDWORD lpcbValueName,
    IN LPDWORD lpReserved,
    OUT LPDWORD lpType,
    OUT LPBYTE lpData,
    IN OUT LPDWORD lpcbData
    )
{
#ifndef _SupportW95
	return RegEnumValueW( hKey, dwIndex, lpValueName, lpcbValueName, lpReserved, lpType, lpData, lpcbData );
#else
	if ( _isWinNT )
		return RegEnumValueW( hKey, dwIndex, lpValueName, lpcbValueName, lpReserved, lpType, lpData, lpcbData );

	char *value = NULL;
	DWORD len = *lpcbValueName;
	
	if ( len != 0 )
		value = new char[ len ];

	int bfSize = 0;
	if ( lpcbData != NULL )
		bfSize = *lpcbData;

	LONG res = RegEnumValueA( hKey, dwIndex, value, &len, lpReserved, lpType, lpData, lpcbData );

	if ( value != NULL )
	{
		mbstowcs( lpValueName, value, *lpcbValueName );
		delete value;
	}

	bool multiString = false;

	if ( ( res == ERROR_SUCCESS ) && ( bfSize != 0 ) )
	{
		switch ( *lpType )
		{
			case REG_MULTI_SZ: multiString = true;
			case REG_SZ:
			case REG_EXPAND_SZ:
			case REG_LINK:
				convertToWchar(	(char *)lpData, bfSize, multiString );
				*lpcbData *= 2;
		}
	}

	return res;
#endif
}

LONG
RegSetValueExT (
    IN HKEY hKey,
    IN LPCWSTR lpValueName,
    IN DWORD Reserved,
    IN DWORD dwType,
    IN CONST BYTE* lpData,
    IN DWORD cbData
    )
{
	if ( _isWinNT )
		return RegSetValueExW( hKey, lpValueName, Reserved, dwType, lpData, cbData );

	cMBStr value( lpValueName );

	cAutoDeallocate <char> data;

	if ( ( dwType == REG_SZ ) || ( dwType == REG_EXPAND_SZ ) || ( dwType == REG_MULTI_SZ ) || ( dwType == REG_LINK ) )
	{
		if ( data.Allocate( cbData / sizeof( wchar_t ) ) == NULL )
			return -1;

		assert( data.m_len <= 2048 );

		_wcstombs( data.m_ptr, data.m_len, (wchar_t *)lpData, -1, ( dwType == REG_MULTI_SZ ) );

		lpData = (BYTE*)data.m_ptr;
		cbData = data.m_len;
	}

	LONG res = RegSetValueExA( hKey, value.m_szBody, Reserved, dwType, lpData, cbData );

	//delete value;

	return res;
}

BOOL SetDlgItemTextT( HWND hDlg, int nIDDlgItem, const wchar_t *lpString )
{
	if ( _isWinNT )
		return SetDlgItemTextW( hDlg, nIDDlgItem, lpString );
	else
	{
		cMBStr str( lpString );
		return SetDlgItemTextA( hDlg, nIDDlgItem, str.m_szBody );
	}
}

LRESULT SendSettingChangeMessageT( HWND hWnd, WPARAM wParam, const wchar_t *section )
{
	if ( _isWinNT ) return SendMessageW( hWnd, WM_SETTINGCHANGE, wParam, (LPARAM)section ); 

	cMBStr mbs( section );
	return SendMessageA( hWnd, WM_SETTINGCHANGE, wParam, (LPARAM)mbs.m_szBody );
}

HMODULE
GetModuleHandleT( LPCWSTR lpModuleName )
{
	if ( _isWinNT )
		return GetModuleHandleW( lpModuleName );

	cMBStr mbs( lpModuleName );
	return GetModuleHandleA( mbs.m_szBody );
}

HMODULE
LoadLibraryExT(
    IN LPCWSTR lpLibFileName,
    IN HANDLE hFile,
    IN DWORD dwFlags
    )
{

	if ( _isWinNT )
		return LoadLibraryExW( lpLibFileName, hFile, dwFlags );

	cMBStr mbs(lpLibFileName );
	return LoadLibraryExA( mbs.m_szBody, hFile, dwFlags );
}

BOOL
GetComputerNameT(
    OUT LPWSTR lpBuffer,
    IN OUT LPDWORD nSize
    )
{
#ifndef _SupportW95
	return GetComputerNameW( lpBuffer, nSize );
#else
	if ( _isWinNT )
		return GetComputerNameW( lpBuffer, nSize );

	cAutoDeallocate <char> cad;
	int len = *nSize;
	char *bufferA = cad.Allocate( len );

	BOOL ok = GetComputerNameA( bufferA, nSize );

	if ( ok )
		mbstowcs( lpBuffer, bufferA, len );

	return ok;
#endif
}

DWORD
GetEnvironmentVariableT(
    IN LPCWSTR lpName,
    OUT LPWSTR lpBuffer,
    IN DWORD nSize
    )
{
#ifndef _SupportW95
	return GetEnvironmentVariableW( lpName, lpBuffer, nSize );
#else
	if ( _isWinNT )
		return GetEnvironmentVariableW( lpName, lpBuffer, nSize );

	cMBStr mbsT[2];

	//cAutoDeallocate <char> cad;
	char *bufferA = mbsT[0].SetAndGet( nSize ); // cad.Allocate( nSize );
	char *nameA = mbsT[1].SetAndGet( lpName ); // allocateAndCopyStr( lpName );

	DWORD res = GetEnvironmentVariableA( nameA, bufferA, nSize );

	if ( res != 0  )
		mbstowcs( lpBuffer, bufferA, nSize );

	// freeStr( nameA  );

	return res;
#endif
}

ATOM
RegisterClassExT( CONST WNDCLASSEXW * lpWndClass )
{
	if ( _isWinNT )
		return RegisterClassExW( lpWndClass );

	WNDCLASSEXA wndClass;

	memcpy( &wndClass, lpWndClass, sizeof( wndClass ) );
	cMBStr mbsMN( lpWndClass->lpszMenuName );
	cMBStr mbsCN( lpWndClass->lpszClassName );

	wndClass.lpszMenuName = mbsMN.m_szBody;
	wndClass.lpszClassName = mbsCN.m_szBody;

	return RegisterClassExA( &wndClass );
}

ATOM
RegisterClassT( CONST WNDCLASSW *lpWndClass )
{

#ifdef UNICODE

	return RegisterClassW( lpWndClass );

#else
	//
	//	Problem with the main window name displaing.
	//		Unicode version shows chinise simbols.
	//		Non unicode version with RegisterClassExW shows first simbol only
	//	:) The problem was in DefWindowProcA

	if ( _isWinNT )
		return RegisterClassW( lpWndClass );

	WNDCLASSA wndClass;

	memcpy( &wndClass, lpWndClass, sizeof( wndClass ) );
	cMBStr mbsMN( lpWndClass->lpszMenuName );
	cMBStr mbsCN( lpWndClass->lpszClassName );

	wndClass.lpszMenuName = mbsMN.m_szBody;
	wndClass.lpszClassName = mbsCN.m_szBody;

	return RegisterClassA( &wndClass );

#endif
}


HWND
FindWindowExT(
    __in_opt HWND hWndParent,
    __in_opt HWND hWndChildAfter,
    __in_opt LPCWSTR lpszClass,
    __in_opt LPCWSTR lpszWindow )
{
	if ( _isWinNT )
		return FindWindowExW( hWndParent, hWndChildAfter, lpszClass, lpszWindow );

	cMBStr className( lpszClass );
	cMBStr windowName( lpszWindow );

	return FindWindowExA( hWndParent, hWndChildAfter, className.m_szBody, windowName.m_szBody );
}

HWND
FindWindowT(
    __in_opt LPCWSTR lpClassName,
    __in_opt LPCWSTR lpWindowName )
{
	if ( _isWinNT )
		return FindWindowW( lpClassName, lpWindowName );

	cMBStr className( lpClassName );
	cMBStr windowName( lpWindowName );

	return FindWindowA( className.m_szBody, windowName.m_szBody );
}

HWND
CreateWindowExT(
    IN DWORD dwExStyle,
    IN LPCWSTR lpClassName,
    IN LPCWSTR lpWindowName,
    IN DWORD dwStyle,
    IN int X,
    IN int Y,
    IN int nWidth,
    IN int nHeight,
    IN HWND hWndParent,
    IN HMENU hMenu,
    IN HINSTANCE hInstance,
    IN LPVOID lpParam )
{
	if ( _isWinNT )
		return CreateWindowExW( dwExStyle, lpClassName, lpWindowName, dwStyle, X, Y, nWidth, nHeight, hWndParent, hMenu, hInstance, lpParam );

	cMBStr className( lpClassName );
	cMBStr windowName( lpWindowName );

	return CreateWindowExA( dwExStyle, className.m_szBody, windowName.m_szBody, dwStyle, X, Y, nWidth, nHeight, hWndParent, hMenu, hInstance, lpParam );
}

LRESULT
DefWindowProcT( HWND hWnd, UINT Msg, WPARAM wParam, LPARAM lParam )
{
	if ( _isWinNT )	return DefWindowProcW( hWnd, Msg, wParam, lParam );
	else		return DefWindowProcA( hWnd, Msg, wParam, lParam );

}

BOOL
GetMessageT( LPMSG lpMsg, HWND hWnd, UINT wMsgFilterMin, UINT wMsgFilterMax )
{
	if ( _isWinNT )	return GetMessageW( lpMsg, hWnd, wMsgFilterMin, wMsgFilterMax );
	else		return GetMessageA( lpMsg, hWnd, wMsgFilterMin, wMsgFilterMax );
}

LRESULT
DispatchMessageT( CONST MSG *lpMsg )
{
	if ( _isWinNT ) return DispatchMessageW( lpMsg );
	else		return DispatchMessageA( lpMsg );
}

int
TranslateAcceleratorT(
	HWND hWnd,
	HACCEL hAccTable,
	LPMSG lpMsg )
{
	if ( _isWinNT ) return TranslateAcceleratorW( hWnd, hAccTable, lpMsg );
	else		return TranslateAcceleratorA( hWnd, hAccTable, lpMsg );
}

BOOL TextOutT( HDC hdc, int nXStart, int nYStart, LPCWSTR lpString, int cbString )
{
	if ( _isWinNT )
		return TextOutW( hdc, nXStart, nYStart, lpString, cbString );

	cMBStr ws( lpString );

	return TextOutA( hdc, nXStart, nYStart, ws.m_szBody, cbString );
}

int
GetDateFormatT(
    IN LCID             Locale,
    IN DWORD            dwFlags,
    IN CONST SYSTEMTIME *lpDate,
    IN LPCWSTR          lpFormat,
    OUT LPWSTR          lpDateStr,
    IN int              cchDate)
{
#ifndef _SupportW95
		return GetDateFormatW( Locale, dwFlags, lpDate, lpFormat, lpDateStr, cchDate );
#else
	if ( _isWinNT )
		return GetDateFormatW( Locale, dwFlags, lpDate, lpFormat, lpDateStr, cchDate );

	cMBStr mbsFormat( lpFormat );

	int len = GetDateFormatA( Locale, dwFlags, lpDate, mbsFormat.m_szBody, NULL, 0 );

	if ( len == 0 )
		return 0;

	cMBStr mbsDateStr( len );

	len = GetDateFormatA( Locale, dwFlags, lpDate, mbsFormat.m_szBody, mbsDateStr.m_szBody, len );

	if ( len == 0 )
		return 0;

	return (int)mbstowcs( lpDateStr, mbsDateStr.m_szBody, cchDate );
#endif
}

int
GetTimeFormatT(
    IN LCID             Locale,
    IN DWORD            dwFlags,
    IN CONST SYSTEMTIME *lpTime,
    IN LPCWSTR          lpFormat,
    OUT LPWSTR          lpTimeStr,
    IN int              cchTime)
{
	if ( _isWinNT )
		return GetTimeFormatW( Locale, dwFlags, lpTime, lpFormat, lpTimeStr, cchTime );

	cMBStr mbsFormat( lpFormat );
	cAutoDeallocate < char > auTimeStr;

	if ( ( cchTime != 0 ) && ( auTimeStr.Allocate( cchTime ) == NULL ) )
		return 0;

	int len = GetTimeFormatA( Locale, dwFlags, lpTime, mbsFormat.m_szBody, auTimeStr.m_ptr, cchTime );

	if ( cchTime != 0 )
		_mbstowcs( lpTimeStr, cchTime, auTimeStr.m_ptr, -1 );

	return len;
}

int TabCtrl_InsertItemT( HWND hwnd, int iItem, TCITEMW *pitem )
{
	if ( _isWinNT )
		return (int)SendMessageW((hwnd), TCM_INSERTITEMW, (WPARAM)(int)(iItem), (LPARAM)(const TC_ITEMW *)(pitem));

	TCITEMA tciA;

	memcpy( &tciA, pitem, sizeof( tciA ) );

	cMBStr mbs( pitem->pszText );

	if ( tciA.mask & TCIF_TEXT )
		tciA.pszText = mbs.m_szBody;

	return (int)SendMessageA((hwnd), TCM_INSERTITEMA, (WPARAM)(int)(iItem), (LPARAM)(const TC_ITEMA *)(&tciA));
}

bool Toolbar_InsertButtonT( HWND hToolbar, int iButton,  TBBUTTON *tbb )
{
	if ( _isWinNT )
		return SendMessageW( hToolbar, TB_INSERTBUTTONW, (WPARAM)iButton, (LPARAM)tbb ) != 0;

	TBBUTTON tbbA;
	memcpy( &tbbA, tbb, sizeof( tbbA ) );

	cMBStr mbs( ( wchar_t * )tbb->iString );
	tbbA.iString = (INT_PTR)mbs.m_len;

	return SendMessageA( hToolbar, TB_INSERTBUTTONA, (WPARAM)iButton, (LPARAM)&tbbA ) != 0;
}

DWORD
GetPrivateProfileStringT(
    IN LPCWSTR lpAppName,
    IN LPCWSTR lpKeyName,
    IN LPCWSTR lpDefault,
    OUT LPWSTR lpReturnedString,
    IN DWORD nSize,
    IN LPCWSTR lpFileName
    )
{
	if ( _isWinNT )
		return GetPrivateProfileStringW( lpAppName, lpKeyName,  lpDefault, lpReturnedString, nSize, lpFileName );

	cMBStr app( lpAppName );
	cMBStr key( lpKeyName );
	cMBStr def( lpKeyName );
	cMBStr fil( lpKeyName );

	char *out = new char[ nSize ];

	if ( out == NULL )
		return 0;

	DWORD res = GetPrivateProfileStringA( 
		lpAppName != NULL ? app.m_szBody : NULL,
		lpKeyName != NULL ? key.m_szBody : NULL,
		lpKeyName != NULL ? def.m_szBody : NULL,
		out,
		nSize,
		lpKeyName != NULL ? fil.m_szBody : NULL );

	_mbstowcs( lpReturnedString, nSize, out );

	return res;
}

DWORD
GetLogicalDriveStringsT(
    IN DWORD nBufferLength,
    OUT LPWSTR lpBuffer
    )
{
	if ( _isWinNT )
		return GetLogicalDriveStringsW( nBufferLength, lpBuffer );

	cAutoDeallocate < char > adChar;
	if ( adChar.Allocate( nBufferLength ) == NULL )
		return false;

	DWORD res = GetLogicalDriveStringsA( nBufferLength, adChar.m_ptr );

	if ( res == 0 )
		return res;

	_mbstowcs( lpBuffer, nBufferLength, adChar.m_ptr );

	return res;
}

DWORD
GetCompressedFileSizeT(
    IN LPCWSTR lpFileName,
    OUT LPDWORD lpFileSizeHigh
    )
{
	if ( _isWinNT )
		return GetCompressedFileSizeW( lpFileName, lpFileSizeHigh );

	cMBStr mbs( lpFileName );

	return GetCompressedFileSizeA( mbs.m_szBody, lpFileSizeHigh );
}

//
//	Sends message in unicode for NT and ANSI for 95
//
LRESULT
	SendMessageT(
		IN HWND hWnd,
		IN UINT MsgW,
		IN UINT MsgA,
		IN WPARAM wParam,
		const wchar_t *lParam )
{
	if ( _isWinNT )
		return SendMessageW( hWnd, MsgW, wParam, (LPARAM)lParam );

	cMBStr mbs( lParam );

	return SendMessageA( hWnd, MsgA, wParam, (LPARAM)mbs.m_szBody );
}

DWORD
ExpandEnvironmentStringsT(
    IN LPCWSTR lpSrc,
    OUT LPWSTR lpDst,
    IN DWORD nSize
    )
{
	if ( _isWinNT )
		return ExpandEnvironmentStringsW( lpSrc, lpDst, nSize );

	cMBStr mbsSrc( lpSrc );
	cAutoDeallocate < char > auDst;

	if ( auDst.Allocate( nSize ) == NULL )
		return 0;

	DWORD res = ExpandEnvironmentStringsA( mbsSrc.m_szBody, auDst.m_ptr, nSize );

	_mbstowcs( lpDst, nSize, auDst.m_ptr );

	return res;
}

HRSRC FindResourceT( HMODULE hInst, int idRes, const wchar_t *rcType )
{
	if ( _isWinNT )
		return FindResourceW( hInst, MAKEINTRESOURCEW(idRes), rcType );
	else
		return FindResourceA( hInst, MAKEINTRESOURCEA(idRes), (LPCSTR)rcType );
}

BOOL
FindNextFileT(
    HANDLE hFindFile,
    LPWIN32_FIND_DATAW lpFindFileData
    )
{
	if ( _isWinNT )
		return FindNextFileW( hFindFile, lpFindFileData );

	WIN32_FIND_DATAA findFileDataA;
	BOOL res = FindNextFileA( hFindFile, &findFileDataA );

	fda2fdw( lpFindFileData, &findFileDataA );

	return res;
}

int GetWindowTextLengthT( HWND hWnd )
{
	if ( _isWinNT )
		return GetWindowTextLengthW( hWnd );
	else
		return GetWindowTextLengthA( hWnd );
}

BOOL SHGetSpecialFolderPathT( HWND hwnd, LPWSTR pszPath, int csidl, BOOL fCreate )
{
	if ( cShell32Dll::getStatic().error )
		return false;

	return cShell32Dll::getStatic().m_SHGetSpecialFolderPathW( hwnd, pszPath, csidl, fCreate );
}

LPITEMIDLIST SHBrowseForFolderT(LPBROWSEINFOW lpbi)
{
#ifndef _SupportW95
	return SHBrowseForFolderW( lpbi );
#else

	if ( _isWinNT )
	{
		// Win 95 does not have the function
		// return SHBrowseForFolderW( lpbi );

		static CDll dll( false, "shell32.dll", "SHBrowseForFolderW", false );
		if ( dll.error )
		{
			ODS_fls( L"Assertion!" );
			return 0;
		}

		typedef LPITEMIDLIST WINAPI 
			typeSHBrowseForFolderW( LPBROWSEINFOW lpbi );

		return ((typeSHBrowseForFolderW *)dll.pProc)( lpbi );
	}

	BROWSEINFOA biA;

	char dispNameA[ MAX_PATH ];
	char titleA[ MAX_PATH ];

//	_wcstombs( dispNameA, sizeof( dispNameA ), lpbi->pszDisplayName );
	_wcstombs( titleA, sizeof( titleA ), lpbi->lpszTitle );

    biA.hwndOwner		= lpbi->hwndOwner;
    biA.pidlRoot		= lpbi->pidlRoot;
    biA.pszDisplayName	= dispNameA;
    biA.lpszTitle		= titleA;
    biA.ulFlags			= lpbi->ulFlags;
    biA.lpfn			= lpbi->lpfn;
    biA.lParam			= lpbi->lParam;

	LPITEMIDLIST pidl = SHBrowseForFolderA( &biA );

	if ( lpbi->pszDisplayName != NULL )
		mbstowcs( lpbi->pszDisplayName, dispNameA, sizeof( dispNameA ) );
	
	return pidl;
#endif
}

BOOL SHGetPathFromIDListT( LPCITEMIDLIST pidl, LPWSTR pszPath )
{
	if ( _isWinNT )
	{
		// Win 95 does not have the function
		// return SHGetPathFromIDListW( pidl, pszPath );

		static CDll dll( false, "shell32.dll", "SHGetPathFromIDListW", false );
		if ( dll.error ) { ODS_fls( L"Assertion!" ); return 0; }
		typedef BOOL WINAPI 
			typeSHGetPathFromIDListW( LPCITEMIDLIST pidl, LPWSTR pszPath );

		return ((typeSHGetPathFromIDListW *)dll.pProc)( pidl, pszPath );
	}

	char pszPathA[ MAX_PATH ];

	BOOL res = SHGetPathFromIDListA( pidl, pszPathA );

	mbstowcs( pszPath, pszPathA, sizeof( pszPathA ) );

	return res;
}

BOOL Shell_NotifyIconT( DWORD dwMessage, PNOTIFYICONDATAW lpData )
{
	if ( _isWinNT )
	{
		// Win 95 does not have the function
		//	return Shell_NotifyIconW( dwMessage, lpData );
		
		static CDll dll( false, "shell32.dll", "Shell_NotifyIconW", false );
		if ( dll.error ) { ODS_fls( L"Assertion!" ); return 0; }
		typedef BOOL WINAPI 
			typeShell_NotifyIconW( DWORD dwMessage, PNOTIFYICONDATAW lpData );

		return ((typeShell_NotifyIconW *)dll.pProc)( dwMessage, lpData );
	}

	NOTIFYICONDATAA ndA;
	memset( &ndA, 0, sizeof( ndA ) );
	ndA.cbSize = sizeof( ndA );
//	nid.uFlags = NIF_ICON | NIF_TIP
	ndA.hWnd = lpData->hWnd;
	ndA.uID = lpData->uID;
	ndA.uFlags = lpData->uFlags;
	ndA.uCallbackMessage = lpData->uCallbackMessage;
	ndA.hIcon = lpData->hIcon;
	if ( ndA.uFlags & NIF_TIP ) wcstombs( ndA.szTip, lpData->szTip, maxLen( ndA.szTip ) );
//#if (_WIN32_IE >= 0x0500)
//	ndA.dwState = lpData->dwState;
//	ndA.dwStateMask = lpData->dwStateMask;
//	if ( ndA.uFlags & NIF_INFO ) wcstombs( ndA.szInfo, lpData->szInfo, maxLen( ndA.szInfo ) );
//	ndA.uTimeout = lpData->uTimeout;
//	if ( ndA.uFlags & NIF_INFO ) wcstombs( ndA.szInfoTitle, lpData->szInfoTitle, maxLen( ndA.szInfoTitle ) );
//	ndA.dwInfoFlags = lpData->dwInfoFlags;
//#endif
//#if (_WIN32_IE >= 0x600)
//	ndA.guidItem = lpData->guidItem;
//#endif

	return Shell_NotifyIconA( dwMessage, &ndA );
}

DWORD
GetLongPathNameT(
    IN LPCWSTR lpszShortPath,
    OUT LPWSTR  lpszLongPath,
    IN DWORD    cchBuffer
    )
{
#ifndef _SupportW95
	return GetLongPathNameW( lpszShortPath, lpszLongPath, cchBuffer );
#else
	if ( _isWinNT )
	{
		// Win NT does not have the function
		//	return GetLongPathNameW( lpszShortPath, lpszLongPath, cchBuffer );

		static CDll kernel32dll( false, "KERNEL32.dll", "GetLongPathNameW", false );
		if ( kernel32dll.error )
		{
			copyString( lpszLongPath, lpszShortPath, cchBuffer );
			return (int)wcslen( lpszLongPath );
		}
		typedef BOOL WINAPI 
			typeGetLongPathNameW(
					IN LPCWSTR lpszShortPath,
					OUT LPWSTR  lpszLongPath,
					IN DWORD    cchBuffer
				);

		return ((typeGetLongPathNameW *)kernel32dll.pProc)( lpszShortPath, lpszLongPath, cchBuffer );
	}

	char shortPathA[ MAX_PATH ];
	_wcstombs( shortPathA, sizeof( shortPathA ), lpszShortPath );


	char longPathA[ MAX_PATH ];
	DWORD res;
//	res = GetLongPathNameA( shortPathA, longPathA, sizeof( longPathA ) );

	// Win 95 does not have the function GetLongPathNameA
	static CDll kernel32dll( false, "KERNEL32.dll", "GetLongPathNameA", false );

	if ( kernel32dll.error )
	{
		copyString( lpszLongPath, lpszShortPath, cchBuffer );
		res = (int)wcslen( lpszLongPath );
	}
	else
	{
		typedef BOOL WINAPI 
			typeGetLongPathNameA(
					IN LPCSTR lpszShortPath,
					OUT LPSTR  lpszLongPath,
					IN DWORD    cchBuffer
				);

		res = ((typeGetLongPathNameA *)kernel32dll.pProc)( shortPathA, longPathA, sizeof( longPathA ) );

		mbstowcs( lpszLongPath, longPathA, cchBuffer );
	}

	return res;
#endif
}

DWORD
GetModuleFileNameT(
    IN HMODULE hModule,
    OUT LPWSTR lpFileName,
    IN DWORD nSize
    )
{
#ifndef _SupportW95
		return GetModuleFileNameW( hModule, lpFileName, nSize );
#else
	if ( _isWinNT )
		return GetModuleFileNameW( hModule, lpFileName, nSize );

	char fileNameA[ MAX_PATH ];

	DWORD res = GetModuleFileNameA( hModule, fileNameA, sizeof( fileNameA ) );

	char path[ MAX_PATH ];

	if ( updatePath( path, fileNameA ) )
		mbstowcs( lpFileName, path, nSize );
	else
		mbstowcs( lpFileName, fileNameA, nSize );

	return res;
#endif
}
