//#include "StdAfx.h"
#include "ODS.h"
#include "unicodefuncs.h"
#include "WmiHelper.h"

//#define _WIN32_DCOM
//#include <iostream>
//using namespace std;
#include <comdef.h>

#pragma comment(lib, "wbemuuid.lib")

typedef void (__stdcall * FREE_DLLS_FN)(DWORD, DWORD);

CWmiHelper::CWmiHelper(void)
{
	HRESULT hres;	
	hres = CoInitializeEx(0, COINIT_MULTITHREADED);
	if (FAILED(hres))
		m_error = err_initError;
	else
		m_error = err_no;
}

CWmiHelper::~CWmiHelper(void)
{
	m_Service.Release();
	m_Locator.Release();
	HMODULE hdll = GetModuleHandle(_T("ole32.dll"));
	if (hdll) {
		FREE_DLLS_FN FreeComDlls = (FREE_DLLS_FN)GetProcAddress(hdll, "CoFreeUnusedLibrariesEx");
		if (FreeComDlls) {
			FreeComDlls(0, 0);
		}
	}

	CoUninitialize();
}

bool CWmiHelper::CheckSecurityCenter() 
{
	OSVERSIONINFOEX	osv;

	osv.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);

	GetVersionEx((OSVERSIONINFO*)&osv);

	if (VER_PLATFORM_WIN32_NT != osv.dwPlatformId) {	// Windows 9x, Me
		return false;
	}

	if (5 > osv.dwMajorVersion) {						// Windows NT 3.xx, 4.0
		return false;
	}

	if (5 < osv.dwMajorVersion) {						// Windows Vista?
		return true;
	}

	osv.dwOSVersionInfoSize = sizeof(osv);
	GetVersionEx((OSVERSIONINFO*)&osv);

	if (0 == osv.dwMinorVersion) {
		return false;	// Windows 2000
	}
/*
	if (1 == osv.dwMinorVersion) {
		if (2 > osv.wServicePackMajor)	{				// Windows XP, XPSP1
			return false;
		} else {
			return true;								// Windows XPSP2
		}
	}

	if (2 == osv.dwMinorVersion) {						// Windows Server 2003/R2
		return true;
	}
*/
	return true;

}

HRESULT CWmiHelper::Connect(BSTR Namespace) 
{
	HRESULT	hr;

	hr = m_Locator.CoCreateInstance( CLSID_WbemLocator ); //(CLSID_WbemAdministrativeLocator);
	if (FAILED(hr)) {
		ODS_fls( cWCStr().FormatT( L"WBEM Locator creation error. Code=0x%08X", hr ) );
		return hr;
	}

	hr = m_Locator->ConnectServer(Namespace ? Namespace : L"root\\SecurityCenter",
									NULL,
									NULL,
									NULL,
									WBEM_FLAG_CONNECT_USE_MAX_WAIT,
									NULL,
									NULL,
									&m_Service);

	if (FAILED(hr))
	{
		ODS_fls( cWCStr().FormatT( L"Security Center connection error. Code=0x%08X", hr ) );
		m_Locator.Release();
	}
	else
		ODS_fls( L"Successfully connected to Security Center" );

	return hr;
}

HRESULT CWmiHelper::FindObject( IWbemClassObject **ppObject, const wchar_t *szProductName, int pos ) // AntiVirus, Firewall
{
	ODSFunc;

	if (!m_Service) {
		return E_FAIL;
	}

	if (NULL == ppObject) {
		return E_POINTER;
	}

	*ppObject = NULL;

	HRESULT	hr;
	ULONG	ret;

	CComPtr<IEnumWbemClassObject>	Enum;
	CComPtr<IWbemClassObject>		cObject;

	hr = m_Service->ExecQuery(	L"WQL", 
								(const BSTR)(const wchar_t *)cWCStr().FormatT( L"select * FROM %sProduct", szProductName ), 
								WBEM_FLAG_FORWARD_ONLY,
								NULL,
								&Enum);
	if (FAILED(hr)) {
		ODS_fls( cWCStr().FormatT( L"%s Product query failed. Code = 0x%08X", szProductName, hr ) );
		return hr;
	}
	
	hr = WBEM_S_NO_ERROR;

	for ( int i = 0; WBEM_S_NO_ERROR == hr; i++ )
	{
		hr = Enum->Next(WBEM_INFINITE, 1L, &cObject, &ret);

		if ( S_OK == hr ) 
		{
			if ( pos == i )
			{
				*ppObject = cObject.Detach();
				ODS_fls( L"cObject Product detected OK." );
				return S_OK;
			}

			cObject.Release();
		}
	}

	ODS_fls( cWCStr().FormatT(  L"%s Product not found.", szProductName ) );
	return S_FALSE;
}

//
//	WMI classes properties
//
//	Windows XP
//	Namespace : SecurityCenter
//	Classes availables: AntiVirusProduct
//
//	AntiVirusProduct-Properties
//
//	companyName
//	displayName
//	enableOnAccessUIMd5Hash
//	enableOnAccessUIParameters
//	instanceGuid
//	onAccessScanningEnabled
//	pathToEnableOnAccessUI
//	pathToUpdateUI
//	productUptoDate
//	updateUIMd5Hash
//	updateUIParameters
//	versionNumber
//
//	Windows Vista and Windows 7
//	Namespace : SecurityCenter2
//	Classes availables : AntiVirusProduct
//
//	AntiVirusProduct  Properties
//
//	displayName
//	instanceGuid
//	pathToSignedProductExe
//	pathToSignedReportingExe
//	productState
//

HRESULT CWmiHelper::GetAntiVirusExecutable( cStringArrayT *T )
{
	ODSFunc;

	if (!m_Service) {
		return E_FAIL;
	}

	HRESULT	hr;
	ULONG	ret;

	CComPtr<IEnumWbemClassObject>	Enum;
	CComPtr<IWbemClassObject>		cObject;
	
	const wchar_t *szProductName = L"AntiVirus";

	hr = m_Service->ExecQuery(	L"WQL", 
								(const BSTR)(const wchar_t *)cWCStr().FormatT( L"select * FROM %sProduct", szProductName ), 
								WBEM_FLAG_FORWARD_ONLY,
								NULL,
								&Enum);
	if (FAILED(hr)) {
		ODS_fls( cWCStr().FormatT(  L"%s Product query failed. Code = 0x%08X", szProductName, hr ) );
		return hr;
	}
	
	hr = WBEM_S_NO_ERROR;

	for ( int i = 0; WBEM_S_NO_ERROR == hr; i++ )
	{
		hr = Enum->Next(WBEM_INFINITE, 1L, &cObject, &ret);

		if ( S_OK == hr ) 
		{
			CComVariant	var;

			//
			//	Windows XP does not have the fields. See WMI classes properties.
			//
			//	pathToSignedProductExe: C:\Program Files\ESET\ESET NOD32 Antivirus\ecmd.exe
			//	pathToSignedReportingExe: C:\Program Files\ESET\ESET NOD32 Antivirus\x86\ekrn.exe

			hr = cObject->Get( L"pathToSignedProductExe", 0, &var, NULL, NULL);

			if ( SUCCEEDED(hr) ) 
				T->addString( (const wchar_t *)var.pbstrVal );
			else
				ODS_fls( cWCStr().FormatT( L"pathToSignedProductExe = %X", hr ) );

			var.Clear();

			hr = cObject->Get( L"pathToSignedReportingExe", 0, &var, NULL, NULL);

			if ( SUCCEEDED(hr) ) 
				T->addString( (const wchar_t *)var.pbstrVal );
			else
				ODS_fls( cWCStr().FormatT( L"pathToSignedReportingExe = %X", hr ) );

			var.Clear();

			cObject.Release();
		}
	}

	return S_FALSE;
}

//#define JpfWmiGuid		L"{A0AAD66A-71ED-434d-B63D-05D269FF4E81}"


//HRESULT CWmiHelper::FindJPF(IWbemClassObject **ppFirewall) 
//{
//	if (!m_Service) {
//		return E_FAIL;
//	}
//
//	if (NULL == ppFirewall) {
//		return E_POINTER;
//	}
//
//	*ppFirewall = NULL;
//
//	HRESULT	hr;
//	ULONG	ret;
//
//	CComPtr<IEnumWbemClassObject>	Enum;
//	CComPtr<IWbemClassObject>		Firewall;
//
//	hr = m_Service->ExecQuery(	L"WQL", 
//								L"select * FROM FirewallProduct", 
//								WBEM_FLAG_FORWARD_ONLY,
//								NULL,
//								&Enum);
//	if (FAILED(hr)) {
//		ATLTRACE(_T("FirewallProduct query failed. Code = 0x%08X\n"), hr);
//		return hr;
//	}
//
//	do {
//		hr = Enum->Next(WBEM_INFINITE, 1L, &Firewall, &ret);
//		if (S_OK == hr) {
//			CComVariant	var;
//			hr = Firewall->Get(L"instanceGuid", 0, &var, NULL, NULL);
//			if (VT_BSTR == var.vt) {
//				CComVariant v1(JpfWmiGuid);
//				if (var == v1) {
//					*ppFirewall = Firewall.Detach();
//					ATLTRACE(_T("JPF FirewallProduct detected OK.\n"));
//					return S_OK;
//				}
//			}
//			Firewall.Release();
//		}
//	} while (WBEM_S_NO_ERROR == hr);
//
//	ATLTRACE(_T("JPF FirewallProduct not found.\n"));
//	return S_FALSE;
//}
//
//HRESULT CWmiHelper::InstallJPFProc(IWbemClassObject **ppFirewall) 
//{
//	if (!m_Service) {
//		return E_FAIL;
//	}
//	HRESULT	hr;
//	CComPtr<IWbemClassObject>	pObj;
//	CComPtr<IWbemClassObject>	pInst;
//
//	hr = m_Service->GetObject(	L"FirewallProduct", 
//								WBEM_FLAG_RETURN_WBEM_COMPLETE,
//								NULL,
//								&pObj,
//								NULL);
//	if (FAILED(hr)) {
//		ATLTRACE(_T("GetObject(FirewallProduct) failed. Code = 0x%08X\n"), hr);
//		return hr;
//	}
//	
//	hr = pObj->SpawnInstance(0, &pInst);
//	if (FAILED(hr)) {
//		ATLTRACE(_T("SpawnInstance(FirewallProduct) failed. Code = 0x%08X\n"), hr);
//		return hr;
//	}
//	
//	CComVariant	varCompany	= L"Jetico, Inc.";
//	CComVariant	varName		= L"Jetico Personal Firewall";
//	CComVariant	varGuid		= JpfWmiGuid;
//	CComVariant	varVersion	= L"2.0";
//	CComVariant	varEnabled	= false;
//
//	pInst->Put(L"instanceGuid", 0, &varGuid, VT_BSTR);
//	pInst->Put(L"displayName", 0, &varName, VT_BSTR);
//	pInst->Put(L"companyName", 0, &varCompany, VT_BSTR);
//	pInst->Put(L"versionNumber", 0, &varVersion, VT_BSTR);
//	pInst->Put(L"enabled", 0, &varEnabled, VT_BOOL);
//
//	if (ppFirewall) {
//		*ppFirewall = pInst.Detach();
//	} else {
//		hr = m_Service->PutInstance(pInst, WBEM_FLAG_CREATE_OR_UPDATE, NULL, NULL);
//		ATLTRACE(_T("JPF installed. Code = 0x%08X\n"), hr);
//	}
//
//	return S_OK;
//}
//
//HRESULT CWmiHelper::SetJPFState(bool bState) 
//{
//	HRESULT	hr;
//	CComPtr<IWbemClassObject>		Firewall;
//
//	if (!m_Service) {
//		return E_FAIL;
//	}
//
//	hr = FindJPF(&Firewall);
//
//	if (S_OK != hr) {
//		hr = InstallJPFProc(&Firewall);
//	}
//
//	if (FAILED(hr)) {
//		return hr;
//	}
//
//	if (!Firewall) {
//		return E_FAIL;
//	}
//
//	CComVariant	var = bState;
//	hr = Firewall->Put(L"enabled", 0, &var, VT_BOOL);
//	ATLTRACE(_T("JPF enabled. Code = 0x%08X\n"), hr);
//	hr = m_Service->PutInstance(Firewall, WBEM_FLAG_CREATE_OR_UPDATE, NULL, NULL);
//	ATLTRACE(_T("JPF state saved. Code = 0x%08X\n"), hr);
//
//	return hr;
//}
//
//
//HRESULT	CWmiHelper::InstallJPF() 
//{
//	HRESULT	hr;
//	CComPtr<IWbemClassObject>		Firewall;
//
//	if (!m_Service) {
//		return E_FAIL;
//	}
//
//	hr = FindJPF(&Firewall);
//
//	if (S_OK != hr) {
//		hr = InstallJPFProc();
//	}
//
//	return hr;
//}
//
//HRESULT	CWmiHelper::UninstallJPF() 
//{
//	HRESULT	hr;
//	CComPtr<IWbemClassObject>		Firewall;
//
//	if (!m_Service) {
//		return E_FAIL;
//	}
//
//	hr = FindJPF(&Firewall);
//	if (FAILED(hr)) {
//		return S_OK;
//	}
//
//	CComBSTR	bstr(L"FirewallProduct.instanceGuid=\"" JpfWmiGuid L"\"");
//	hr = m_Service->DeleteInstance(bstr, 0, NULL, NULL);
//	if (WBEM_E_ACCESS_DENIED == hr) {
//	}
//	return hr;
//
//}

