#pragma once
#include <windows.h>

//
//	cAppFlags
//		Saves and retieves enumerated flags from/to registry
//
class cAppFlags
{
	wchar_t m_keyName[ MAX_PATH ];
	wchar_t m_valueName[ MAX_PATH ];
	int m_maxFlag;
	HKEY m_hRoot;
	bool m_bDefaultValue;

public:

	cAppFlags( HKEY hRoot, const wchar_t * keyName, const wchar_t *valueName = L"Flags", bool bDefaultValue = true, int maxFlag = 64 )
	{
		m_hRoot = hRoot; // HKEY_CURRENT_USER
		wcsncpy_s( m_keyName, keyName, MAX_PATH );
		wcsncpy_s( m_valueName, valueName, MAX_PATH );
		m_maxFlag = 64;
		m_bDefaultValue = bDefaultValue;
	}

	//
	// GetFlagNumByIntT
	//	Find flag number in table
	//		int T = {
	//			int0, flagNum0,
	//			....
	//			intN, flagNumN,
	//		}
	//
	static int GetFlagNumByIntT( int *T, int Q, int iValue )
	{
		int afNum = -1;

		for ( int i = Q; i >= 2;  )
		{
			i -= 2;
			if ( T[i] == iValue )
			{
				afNum = T[ i + 1 ];
				break;
			}
		}

		return afNum;
	}

	//
	// Get and Set function of predefined flags
	// Note, GetFlag function returns m_bDefaultValue if flag has never set before.
	//
	bool GetFlag( int flag )
	{
		if ( ( flag < 0 ) || ( flag > m_maxFlag ) )
			return m_bDefaultValue;

		BYTE *bf;
		int bfSize;

		if ( !readRegistry( &bf, &bfSize, m_hRoot, m_keyName, m_valueName ) )
			return m_bDefaultValue;

		int byteNum = flag / 8;
		int pos		= flag & 7;

		BYTE bt = 0xFF;

		if ( byteNum < bfSize )
			bt = bf[ byteNum ];

		delete bf;

		return ( bt & ( 1 << pos ) ) != 0;
	}

	bool SetFlag( int flag, bool set )
	{
		if ( ( flag < 0 ) || ( flag > m_maxFlag ) )
		{
			assert( false );
			return false;
		}

		BYTE *bf = NULL;
		int bfSize = 0;

		if ( !readRegistry( &bf, &bfSize, m_hRoot, m_keyName, m_valueName ) )
		{
			bf = NULL;
			bfSize = 0;
		}

		int byteNum = flag / 8;
		int pos		= flag & 7;

		if ( byteNum >= bfSize )
		{
			BYTE *bfNew = NULL;
			int bfNewSize = 0;

			bfNewSize = byteNum + 1;
			bfNew = new BYTE[ bfNewSize ];

			if ( bfNew == NULL )
				return false;

			// Set FF for all bytes
			int initVal = m_bDefaultValue ? -1 : 0;
			memset( bfNew, initVal, bfNewSize );

			// Copy old content
			if ( bf != NULL )
			{
				memcpy( bfNew, bf, bfSize );
				delete bf;
			}

			bf = bfNew;
			bfSize = bfNewSize;
		}

		if ( bf == NULL )
			return false;

		if ( set )
			bf[ byteNum ] |= 1 << pos;
		else
			bf[ byteNum ] &= ~(1 << pos);

		bool ok = writeRegistry( bf, bfSize, m_hRoot, m_keyName, m_valueName );

		delete bf;

		return ok;
	}

	//
	//	Read write registry binary data.
	//
	bool readRegistry( BYTE **pBf, int *pBfSize, HKEY hRegRoot, const wchar_t *keyName, const wchar_t *valueName )
	{
		*pBf = NULL;
		*pBfSize = 0;

		HKEY hKey;
		LONG res = RegOpenKeyExW( hRegRoot, keyName, 0, KEY_READ, &hKey );

		if ( res != ERROR_SUCCESS )
			return false;

		DWORD dwSize = 0;
		DWORD dwType = REG_BINARY;
		res = RegQueryValueExW( hKey, valueName, 0, &dwType, NULL, &dwSize );

		if ( dwSize == 0 )
		{
			RegCloseKey( hKey );
			return false;
		}

		*pBf = new BYTE[ dwSize ];

		if ( *pBf == NULL )
		{
			RegCloseKey( hKey );
			return false;
		}

		*pBfSize = dwSize;

		res = RegQueryValueExW( hKey, valueName, 0, &dwType, *pBf, &dwSize );

		RegCloseKey( hKey );

		return ( res == ERROR_SUCCESS );
	}

	bool writeRegistry( BYTE *bf, int bfSize, HKEY hRegRoot, const wchar_t *keyName, const wchar_t *valueName )
	{
		HKEY hKey;
		LONG res = RegCreateKeyExW( hRegRoot, keyName, 0, 0, 0, KEY_WRITE, 0, &hKey, 0 );

		if ( res != ERROR_SUCCESS )
			return false;

		if ( bf == NULL )
			res = RegDeleteValueW( hKey, valueName );
		else
			res = RegSetValueExW( hKey, valueName, 0, REG_BINARY, bf, (DWORD)bfSize );

		RegCloseKey( hKey );

		if ( res != ERROR_SUCCESS )
			return false;

		return true;
	}
};