#include <windows.h>
#include <assert.h>
#include <stdlib.h>
#include "cArgs.h"
 
typedef wchar_t * wcharPtr;

bool addStr ( wchar_t *s, int *rSize, int *pSize, const wchar_t *strAdd ) 
{
	*rSize += (int)wcslen( strAdd );
	if ( ( s != NULL ) && ( *rSize < *pSize ) )
		wcscat_s( s, *pSize, strAdd );

	return true;
}

bool copyArgs( wchar_t *s, int *pSize,  wcharPtr *argv, int argc )
{
	int size = 0;

	for ( int i = 0; i < argc; i++ )
	{
		if ( i != 0 )
			addStr( s, &size, pSize, L" " );

		bool hasProbels = ( wcschr( argv[i], L' ' ) != NULL );

		if ( hasProbels ) addStr( s, &size, pSize, L"\"" );
		addStr( s, &size, pSize, argv[i] );
		if ( hasProbels ) addStr( s, &size, pSize, L"\"" );
	}

	*pSize = size;

	return true;
}

const wchar_t *cArgs::GetCmdLine()
{
	if ( cmdRest != NULL )
	{
		delete cmdRest;
		cmdRest = NULL;
	}

	int size = 0;
	copyArgs( NULL, &size, argv, argc );

	cmdRest = new wchar_t[ ++size ];

	if ( cmdRest == NULL )
		return NULL;

	memset( cmdRest, 0, size * sizeof( wchar_t ) );

	if ( size > 0 )
		copyArgs( cmdRest, &size, argv, argc );

	return cmdRest;
}

cArgs::~cArgs()
{
	if ( argv )
		delete argv;

	if ( buffer )
		delete buffer;

	if ( cmdRest )
		delete cmdRest;

	memset( this, 0, sizeof( *this ) );
}


static int PrepareArgs( wchar_t *cmd, int maxArgs, wcharPtr *argv, bool countOnly )
{
	int count = 0;

	BOOL isWord = FALSE;
	BOOL isLongName = FALSE;

	for( int p = 0; cmd[p] ; p++ )
	{
		if ( !countOnly && ( count > maxArgs ) )
			return count;

		if ( isLongName )
		{
			if ( cmd[ p ] == L'\"' )
			{
				isLongName = FALSE;
				if( !countOnly )
					cmd[ p ] = 0; // end line
			}
		}
		else
		if ( isWord )
		{
			if ( cmd[ p ] == L' ' )
			{
				isWord = FALSE;
				if( !countOnly )
					cmd[ p ] = 0; // end line
			}
		}
		else
		{
			if ( cmd[ p ] == L'\"' )
			{
				isLongName = TRUE;
				if( !countOnly )
					argv[ count ] = cmd + p + 1;
				count++;
			}
			else
			if ( cmd[ p ] != L' ' )
			{	
				isWord = TRUE;
				if( !countOnly )
					argv[ count ] = cmd + p;
				count++;
			}
		}
	}

	return count;
}

#define maxLen( buff ) ( sizeof( buff ) / sizeof( *buff ) )

void cArgs::Init( const char *cmdLine )
{
	memset( this, 0, sizeof( *this ) );

	int size = (int)strlen( cmdLine ) + 1;

	char fileName[ MAX_PATH ];
	int fileNameLen = GetModuleFileNameA( NULL, fileName, maxLen( fileName ) );

	int bufferLen = fileNameLen + 3 + size;
	buffer = new wchar_t[ bufferLen ];

	if ( buffer == 0 )
		return;

	bool addModulePathName = false;

	size_t num = 0;

//
//	Use local #pragma instead of _CRT_SECURE_NO_WARNINGS
//
#pragma warning( disable : 4996 )

	if ( addModulePathName )
	{
		buffer[0] = L'\"';
#ifdef _WIN32		
		mbstowcs_s( &num, buffer + 1, bufferLen - 1, fileName, _TRUNCATE );
#else
		mbstowcs( buffer + 1, fileName, _TRUNCATE );
#endif
		buffer[ fileNameLen + 1 ] = L'\"';
		buffer[ fileNameLen + 2 ] = L' ';

#ifdef _WIN32		
		mbstowcs_s( &num, buffer + fileNameLen + 3, size, cmdLine, _TRUNCATE );
#else
		mbstowcs( buffer + fileNameLen + 3, cmdLine, _TRUNCATE );
#endif
	}
	else
	{
#ifdef _WIN32		
		mbstowcs_s( &num, buffer, bufferLen, cmdLine, _TRUNCATE );
#else
		mbstowcs( buffer, cmdLine, _TRUNCATE );
#endif
	}

#pragma warning( default : 4996 )

	argc = PrepareArgs( buffer, 0, NULL, true );

	if ( argc == 0 )
		return;

	argv = new wcharPtr[ argc ];

	if ( argv == 0 )
	{
		argc = 0;
		return;
	}

	if ( argc != PrepareArgs( buffer, argc, argv, false ) )
		argc = 0;
}

void cArgs::Init( const wchar_t *cmdLine )
{
	memset( this, 0, sizeof( *this ) );

	int size = (int)wcslen( cmdLine ) + 1;

	buffer = new wchar_t[ size ];

	if ( buffer == 0 )
		return;

	wcscpy_s( buffer, size, cmdLine );

	argc = PrepareArgs( buffer, 0, NULL, true );

	if ( argc == 0 )
		return;

	argv = new wcharPtr[ argc ];

	if ( argv == 0 )
	{
		argc = 0;
		return;
	}

	if ( argc != PrepareArgs( buffer, argc, argv, false ) )
		argc = 0;
}

bool cArgs::Remove( int i )
{
	if ( i >= argc )
		return false;

	argc--;

	for( ; i < argc; i++ )
	{
		argv[i] = argv[ i + 1 ];
	}

	return true;
}

int cArgs::Find( const wchar_t *string )
{
	int count;

	for( count = argc; count--; )
	{
		if ( _wcsicmp( string, argv[count] ) == 0 )
			break;
	}

	return count;
}

int cArgs::FindPart( const wchar_t *string )
{
	int count;

	for( count = argc; count--; )
	{
		if ( _wcsnicmp( string, argv[count], wcslen( string ) ) == 0 )
			break;
	}

	return count;
}

bool cArgs::Get( const wchar_t *name, const wchar_t **pValue, const wchar_t *szDefValue ) // = NULL )
{
	int i = Find( name );

	if ( i != -1 )
	{
		if ( i + 1 >= argc ) // Check following argument value.
			return false;

		*pValue = argv[ i + 1 ];

		return Remove( i + 1 ) & Remove( i );
	}
	else
	{
		//
		//	Sample:
		//		name == -a, argument -aAnyValue, so *pValue = AnyValue
		//
		i = FindPart( name );

		if ( i != -1 )
		{
			size_t len = wcslen( name );

			assert( wcslen( argv[i] ) > len );

			//
			//	The check allows to parse 'driver=MftWipeFilter'
			//
			if ( argv[i][len] == L'=' )
			{
				len++;
				if ( argv[i][len] == L'\0' )
				{
					*pValue = argv[ i + 1 ];
					return Remove( i + 1 ) & Remove( i );
				}
			}
			
			*pValue = argv[ i ] + len;

			return Remove( i );
		}
	}
	
	*pValue = szDefValue;

	return false;
}

bool cArgs::Get( int pos, const wchar_t *name )
{
	if ( pos < 0 )
		return false;

	if ( pos >= argc )
		return false;

	if ( _wcsicmp( name, argv[ pos ] ) != 0 )
		return false;

	return Remove( pos );
}

bool cArgs::Get( const wchar_t *name )
{
	int i = Find( name );

	if ( i != -1 )
	{
		return Remove( i );
	}

	return false;
}

//static 
bool cArgs::checkOnOffArg( const wchar_t *type, bool &bOn )
{
	bool bError = false;

	if ( ( _wcsicmp( type, L"Off" ) == 0 ) || 
		 ( _wcsicmp( type, L"False" ) == 0 ) || 
		 ( _wcsicmp( type, L"No" ) == 0 ) || 
		 ( _wcsicmp( type, L"0" ) == 0 ) )
	{
		bOn = false;
	}
	else
	if ( ( _wcsicmp( type, L"On" ) == 0  ) ||
		 ( _wcsicmp( type, L"True" ) == 0  ) ||
		 ( _wcsicmp( type, L"Yes" ) == 0  ) ||
		 ( _wcsicmp( type, L"1" ) == 0  ) )
	{
		bOn = true;
	}
	else
		bError = true;

	return !bError;
}

bool cArgs::ReadAndRemoveArgumentWithIntValue( const wchar_t *name, int *pValue, int defValue )
{
	int iValue;
	if ( pValue == NULL )
		pValue = &iValue;

	const wchar_t *szValue = NULL;
	if ( !Get( name, &szValue ) )
	{
		*pValue = defValue;
		szValue = L"0";
	}
	else
		*pValue = _wtoi( szValue );

	if ( ( *pValue == 0 ) && ( wcscmp( szValue, L"0" ) != 0 ) )
		return false;

	return true;
}