#ifndef _Dlg_H_
#define _Dlg_H_

#include <windows.h>
#include <commctrl.h>

#include "UnicodeFuncs.h"
#include "Autodeallocate.h"

#ifndef GWLP_USERDATA
#define GWLP_USERDATA GWL_USERDATA
#define SetWindowLongPtr SetWindowLong
#define GetWindowLongPtr GetWindowLong
#define GWLP_WNDPROC GWL_WNDPROC
#define LONG_PTR LONG
#define DWLP_USER DWL_USER
#endif

#ifndef LPDLGITEMTEMPLATEEX
#define LPDLGITEMTEMPLATEEX

#pragma pack(push, 1)
struct DLGTEMPLATEEX
{
	WORD dlgVer;
	WORD signature;
	DWORD helpID;
	DWORD exStyle;
	DWORD style;
	WORD cDlgItems;
	short x;
	short y;
	short cx;
	short cy;

	// Everything else in this structure is variable length,
	// and therefore must be determined dynamically

	// sz_Or_Ord menu;			// name or ordinal of a menu resource
	// sz_Or_Ord windowClass;	// name or ordinal of a window class
	// WCHAR title[titleLen];	// title string of the dialog box
	// short pointsize;			// only if DS_SETFONT is set
	// short weight;			// only if DS_SETFONT is set
	// short bItalic;			// only if DS_SETFONT is set
	// WCHAR font[fontLen];		// typeface name, if DS_SETFONT is set
};
struct DLGITEMTEMPLATEEX
{
	DWORD helpID;
	DWORD exStyle;
	DWORD style;
	short x;
	short y;
	short cx;
	short cy;
	DWORD id;

	// Everything else in this structure is variable length,
	// and therefore must be determined dynamically

	// sz_Or_Ord windowClass;	// name or ordinal of a window class
	// sz_Or_Ord title;			// title string or ordinal of a resource
	// WORD extraCount;			// bytes following creation data
};
#pragma pack(pop)

#endif // #define LPDLGITEMTEMPLATEEX

//
//	class cTheme
//		Use this to create themed dialogue.
//
//	Usage:
//		1. Create dilogue page class derived from the cTheme public class.
//
//			class cNewDlg : public cDlgBase, public cTheme
//
//		2. Call cTheme::DlgProc function at RealDlgProc before message prcessing.
//			if this function return true, emmeditly return assigned lRes:
//
//			LRESULT cWDPage::RealDlgProc( UINT uMsg, WPARAM wParam, LPARAM lParam )
//			{
//				LRESULT lRes;
//
//				if ( cTheme::DlgProc( m_hDlg, uMsg, wParam, lParam, lRes ) )
//					return lRes;
//
//				switch(uMsg)
//				...
//
//
class cTheme
{
	//
	//	Recommedations from 
	//		http://www.codeproject.com/wtl/ThemedDialog.asp
	//
	//	Themed tab control draws background with gradient, the child dialogues by default use gray background.
	//	The following functions allows to see native tab control background color.
	//
	////	The trick intercepts following messages
	//	case WM_INITDIALOG:
	//		UpdateBackgroundBrush();
	//		break;
	//	case WM_ERASEBKGND:
	//		return m_bThemeActive;
	//	case WM_CTLCOLORBTN:
	//		if ( m_bThemeActive )
	//			SetBkMode((HDC)(wParam), TRANSPARENT);
	//			return(LRESULT)(GetStockObject(WHITE_BRUSH));
	//		break;
	//	case WM_CTLCOLORSTATIC:
	//		return OnCtlColor( wParam, lParam  );
	//	case WM_DESTROY:
	//		if ( m_hBrush != NULL )
	//			DeleteObject( m_hBrush );
	//		return true;
	//
	bool m_bThemeActive;
	HBRUSH m_hBrush;
	void UpdateBackgroundBrush( HWND hDlg );
	LRESULT OnCtlColor( HWND hDlg, WPARAM wParam, LPARAM lParam );

public:

	//
	//	DlgProc
	//	returns true if message is completely processed and return result lRes has to be passed to caller.
	//
	bool DlgProc( HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam, LRESULT &lRes );

	cTheme()
	{
		m_bThemeActive = false;
		m_hBrush = NULL;
	}

	~cTheme()
	{
	}

};

class cDlg : public cDlgFuncsHolder
{
protected:
	char *m_tmpStr;
	bool m_autoDestroy;
	cAutoDeallocate < BYTE > m_adbfRC;

public:
	int m_idd;

	//cDelayVisible m_delayVisible;

	bool m_isNT;

	static POINT &getStaticPos()
	{
		static POINT pos = {-1,-1 };
		return pos;
	}

	short Spin_GetPos(int id)							{ return LOWORD( SendMessage( GetDlgItem( m_hDlg, id ), UDM_GETPOS, 0, 0 ) ); }
	static short Spin_GetPos( HWND hSpin )			{ return LOWORD( SendMessage( hSpin, UDM_GETPOS, 0, 0 ) ); }
	short Spin_SetPos( int id, short pos )			{ return LOWORD( SendMessage( GetDlgItem( m_hDlg, id ), UDM_SETPOS, 0, (LPARAM) MAKELONG( pos, 0 ) ) ); }
	static void Spin_SetRange( HWND hWnd, short min = UD_MINVAL, short max = UD_MAXVAL )
															{ SendMessage( hWnd, UDM_SETRANGE, 0, (LPARAM) MAKELONG( max, min ) );  }
	void Spin_SetRange(int id, short min = UD_MINVAL, short max = UD_MAXVAL )	{	Spin_SetRange( GetDlgItem( m_hDlg, id ), min, max ); }
	static short Spin_SetPos( HWND hSpin, short pos )	{	return LOWORD( SendMessage( hSpin, UDM_SETPOS, 0, (LPARAM) MAKELONG( pos, 0 ) ) ); }
	int Spin_GetRange( int id, short *pmin = NULL, short *pmax = NULL )
															{	return Spin_GetRange( GetDlgItem( m_hDlg, id ), pmin, pmax ); }
	static int Spin_GetRange( HWND hSpin, short *pmin = NULL, short *pmax = NULL )
	{ 
		LRESULT range = SendMessage( hSpin, UDM_GETRANGE, 0, 0 ); 

		if ( pmin != NULL ) *pmin = HIWORD( range );
		if ( pmax != NULL ) *pmax = LOWORD( range );

		return (int)range;
	}

	DWORD Spin_GetPos32(int id)										{ return DWORD( SendMessage( GetItem( id ), UDM_GETPOS32, 0, 0 ) ); }
	static DWORD Spin_GetPos32( HWND hSpin )						{ return DWORD( SendMessage( hSpin, UDM_GETPOS32, 0, 0 ) ); }
	int Spin_SetPos32( int id, int pos )							{ return int( SendMessage( GetItem( id ), UDM_SETPOS32, 0, (LPARAM) pos ) ); }
	static void Spin_SetRange32( HWND hWnd, int min = 0, int max = 0x7FFFffff )
																	{ SendMessage( hWnd, UDM_SETRANGE32, (WPARAM)min, (LPARAM)max );  }
	void Spin_SetRange32(int id, int min, int max = 0x7FFFffff  )	{ Spin_SetRange32( GetItem( id ), min, max ); }
	static int Spin_SetPos32( HWND hSpin, int pos )					{ return int( SendMessage( hSpin, UDM_SETPOS32, 0, (LPARAM)pos ) ); }

	static void Spin_GetRange32( HWND hSpin, int *pmin, int *pmax )	{ SendMessage( hSpin, UDM_GETRANGE32, (WPARAM)pmin, (LPARAM)pmax ); }
	void Spin_GetRange32( int id, int *pmin, int *pmax )			{ SendMessage( GetItem( id ), UDM_GETRANGE32, (WPARAM)pmin, (LPARAM)pmax ); }

	//
	//	UDNDeltapos
	//	Process UDN_DELTAPOS from Spin control notification and update Edit control
	//		Edit control here is simple integer. 
	//		It does same as spin control has 'Auto Buddy' and 'Set Buddy Integer' attributes.
	//
	bool UDNDeltapos( NMUPDOWN *ud, int id_edt, int pMin, int pMax, int pDelta );
	static bool UDNDeltapos( NMUPDOWN *ud, HWND hWnd, int pMin, int pMax, int pDelta );

	//
	//	UpdateEditSize
	//	Process UDN_DELTAPOS from Spin control notification and update Edit control
	//		Edit control here is window with size text with figure and dimension( B,KB,MB,GB,TB ) ( f.e. 10MB ).
	//		This function increases/or decreases figures only.
	//		If text size is not recognised it returns (DWORDLONG)-1 and shows message box if idsErrorSize != 0.
	//		Sample error text is "Size is invalid"
	//
	static DWORDLONG UpdateEditSize( HWND hEdit, int delta, int idsErrorSize = 0, int idsErrorCaption = 0);
	DWORDLONG UpdateEditSize( int idcEdit, int delta, int idsErr = 0, int idsCaption = 0 )
	{ return UpdateEditSize( GetItem( idcEdit ), delta, idsErr, idsCaption ); }

	static void ScreenToClientRect( HWND hDlg, RECT *r )
	{ MapWindowPoints( NULL, hDlg, (LPPOINT)r, 2 );	}

	void ScreenToClientRect( RECT *r )
	{ ScreenToClientRect( m_hDlg, r ); }

	static void GetDlgItemRect( HWND hDlg, HWND hItem, RECT *r )
	{
		GetWindowRect( hItem, r );
		ScreenToClientRect( hDlg, r );
	}

	static bool GetDlgItemRect( HWND hDlg, int iItem, RECT *r )
	{
		HWND hItem = GetDlgItem( hDlg, iItem );

		if ( hItem == NULL )
			return false;

		GetDlgItemRect( hDlg, hItem, r );

		return true;
	}

	static bool SetDlgItemPos( HWND hDlg, int iItem, int X, int Y, int cx = 0, int cy = 0, HWND hWndInsertAfter = NULL, UINT uFlags = SWP_NOSIZE | SWP_NOZORDER | SWP_SHOWWINDOW )
	{
		HWND hItem = GetDlgItem( hDlg, iItem );

		if ( hItem == NULL )
			return false;

		BOOL ok = SetWindowPos( hItem, hWndInsertAfter, X, Y, cx, cy, uFlags );

		return ok != FALSE;
	}

	void GetDlgItemRect( HWND hItem, RECT *r ) { GetDlgItemRect(m_hDlg, hItem, r ); }

	static HICON SetDlgIcon( HWND hDlg, HICON hIcon, int iType = ICON_SMALL )	{ return (HICON)SendMessage( hDlg, WM_SETICON, iType, (LPARAM)hIcon ); }
	HICON SetDlgIcon( HICON hIcon, int iType = ICON_SMALL )	{ return SetDlgIcon( m_hDlg, hIcon, iType ); }

	HICON SetDlgIcon( int idIcon, int iType ) // = ICON_SMALL )
	{
		HICON hIcon;
		if ( iType == ICON_SMALL )
			hIcon = (HICON)LoadImage( m_hInst, MAKEINTRESOURCE( idIcon ), IMAGE_ICON, 16, 16, LR_DEFAULTCOLOR | LR_SHARED );
		else
			hIcon = LoadIcon( m_hInst, MAKEINTRESOURCE( idIcon ) );

		HICON hOld = SetDlgIcon( hIcon, iType );

		//if ( hIcon != NULL )
		//	DestroyIcon( hIcon );

		return hOld;
	}

	void SetDlgIcon( int idIcon )
	{
		SetDlgIcon( idIcon, ICON_SMALL );
		SetDlgIcon( idIcon, ICON_BIG );
	}

	DLGPROC m_externalDlgProc;

	cDlg()
	{
		m_autoDestroy = false;
		m_externalDlgProc = NULL;
		m_tmpStr = NULL;
		m_hDlg = NULL;
		m_isNT = GetVersion() < 0x80000000;

#ifdef _UseHtmlHelp_ 
		m_useCtxAsIDS = true;
#endif
		m_bCallHelpInfoByClickingCtxHelpButton = false;
	}

	virtual ~cDlg()
	{
		if ( m_autoDestroy )
		{
			if ( IsWindow( m_hDlg ) )
				DestroyWindow( m_hDlg );
		}

		if ( m_tmpStr != NULL )
			delete m_tmpStr;

		//if ( m_hAccel != NULL )
		//	DestroyAcceleratorTable( m_hAccel );
	}

	LPDLGTEMPLATEW FlipDlgResource( UINT idRes );
	INT_PTR DoModal( HINSTANCE hInst, UINT idRes, HWND hParent, DLGPROC externalDlgProc = NULL );
	HWND DoModeless( HINSTANCE hInst, UINT idRes, HWND hParent, DLGPROC externalDlgProc = NULL );
	HWND DoModelessPage( HINSTANCE hInst, UINT idRes, HWND hParent );

	bool Clipped( HWND hWnd, POINT point )
	{
		RECT r;
		GetWindowRect( hWnd, &r );
		ClientToScreen( m_hDlg, &point );

		bool res = PtInRect( &r, point ) != FALSE;
		return res;
	}

	bool isMinimized()	{	return IsIconic( m_hDlg ) != 0; }

	//
	//	All the values are assigned just before WM_INITDALOG message,
	//	see DlgProc realisation.
	//
	//	And are served in WM_GETMINMAXINFO, see DlgProc realisation.
	//
	int m_trackingFixType; // default = ttfMinX | ttfMinY
	enum { ttfNone = 0, ttfMinX = 1, ttfMinY = 2, ttfMaxX = 4, ttfMaxY = 8 };
	POINT m_minSize;
	POINT m_maxSize;

	//
	//	Context help functions.
	//
	static bool FoundControlId( DWORD *dwPtr, DWORD id, DWORD *pIDH = NULL );

	virtual wchar_t *GetHelpFileName( wchar_t *path, int wcSize );

	cWCS m_wcsHlpFileName;
	const wchar_t *GetHelpFileName()
	{
		if ( m_wcsHlpFileName.isEmpty() )
		{
			m_wcsHlpFileName.SetLength( MAX_PATH );
			GetHelpFileName( m_wcsHlpFileName.m_szBody, m_wcsHlpFileName.m_len );
		}

		return m_wcsHlpFileName.m_szBody;
	}

	void OnHelpInfo( HELPINFO* pHelpInfo, DWORD *MenuHelpIDs, bool showDefault = false );

	//
	//	If m_bCallHelpInfoByClickingCtxHelpButton is true,
	//	So it allows to skip cxt '?' cursor with following clicking.
	//	It block default action and call directly virtual OnHelpInfo(0)
	//
	bool m_bCallHelpInfoByClickingCtxHelpButton;
	//
	//	CtxHelp_NCLBUTTONDOWN
	//		Detects if 'non client click' happens on context help button.
	//
	bool CtxHelp_NCLBUTTONDOWN( LPARAM lParam );

	//
	//	OnHelpInfo
	//		Calls virtual help info.
	//
	virtual void OnHelpInfo( HELPINFO *pHelpInfo = NULL ) {}

#ifdef _UseHtmlHelp_ 

	bool m_useCtxAsIDS; // Treat that MenuHelpIDs is IDS instead of HID
						// If true OnHTMPHelpInfo uses resource instead of .chm file

	static void OnHTMPHelpInfo( HELPINFO* pHelpInfo, DWORD *MenuHelpIDs, const wchar_t *htmlHlpName, bool useCtxAsIDS = false );
	void OnHTMPHelpInfo( HELPINFO* pHelpInfo, DWORD *MenuHelpIDs );
	static void OnHelpPopup( HWND hItem, POINT pt, const wchar_t *str );
	static void OnHelpPopup( HELPINFO* pHelpInfo, const wchar_t *str );
	static void OnHelpPopup( HELPINFO* pHelpInfo, int ids );
	static void OnHelpPopup( HELPINFO* pHelpInfo, DWORD *MenuHelpIDs );

#endif

	static cDlg *GetDlgClass( HWND hDlg );
	static void SetDlgClass( HWND hDlg, cDlg *dlg );

	//
	//	BelongToScreen
	//		Checks if points table belongs to screen.
	//		If delta is not zero so border around screen with the size is excluded too.
	//
	static bool BelongToScreen( POINT *ptT, int ptQ = 1, int delta = 0 );

	static void CenterPosition( RECT *r, int px = 50, int py = 50 );
	static void CenterPosition( HWND hDlg, int px = 50, int py = 50 );
	void CenterPosition( int px = 50, int py = 50) { CenterPosition( m_hDlg, px, py ); }
	static void MoveWindowToCursorPosition( HWND hDlg );
	void MoveWindowToCursorPosition() { MoveWindowToCursorPosition( m_hDlg ); }
	static void MapToScreen( HWND hDlg );
	void MapToScreen() { MapToScreen( m_hDlg ); }

	void EnableItem( int id, bool enable = true ) {	EnableWindow( GetDlgItem( m_hDlg, id ),	enable ); }
	void CheckAndEnableItem( int id, bool bCheck = true, bool bEnable = true );
	//
	//	Calculates size of text in pixels  
	//
	static bool GetTextSize( HWND hDlg, const wchar_t *text, SIZE *s, DWORD dwDTParam = DT_WORDBREAK );
	static bool GetTextSize( HWND hDlg, const wchar_t *text, SIZE *s, DWORD dwDTParam, HFONT hFont );
	static bool GetTextSize( HDC hdc, const wchar_t *text, SIZE *s, DWORD dwDTParam, HFONT hFont );

	static LRESULT CALLBACK DlgProc( HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam );
	static LRESULT CALLBACK DlgPropPageProc(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam);

	//
	//	Redirect_ClickToCmd
	//		Redirect NM_CLICK notification to WM_COMMAND from this control
	//
	static bool Redirect_ClickToCmd( HWND hDlg, UINT uMsg, int idControl, LPARAM lParam  );

	virtual LRESULT RealDlgProc( UINT uMsg, WPARAM wParam, LPARAM lParam );
	virtual LRESULT DrawColorProc( UINT uMsg, WPARAM wParam, LPARAM lParam ) { return false; };

	//
	//	Checks screen resolution and does not allow to set position beyond the screen borders.
	//
	bool setWindowRect( RECT &r );

};


//
//	CRefCount...
//	OLE interface based classes are used the class...
//
class CRefCount
{
public:
	CRefCount() : m_cRef(0) {};

	UINT AddRef()  { return(++m_cRef); }
	UINT Release() { return(--m_cRef); }
	UINT GetRef()  { return(  m_cRef); }

private:
	UINT m_cRef;
};


class cSizedCtrls
{
  public:
	enum ePosType {
		ptNo  = 0,	// Mo move.
		ptPR  = 1,	// Controls wich Positions depend on Rigth conner of dialogue
		ptPD  = 2,  // Controls wich Positions depend on Down side of dialogue
		ptSR  = 4,	// Controls wich Sizes depend on Rigth conner of dialogue
		ptSD  = 8,	// Controls wich Sizes depend on Down conner of dialogue

		ptPHC  = 0x10,	// Controls wich Positions depend on Horizontal Center of dialogue
		ptPVC  = 0x20,  // Controls wich Positions depend on Vertical Center of dialogue

		ptPSHC = ptPHC | 0x40,	// Controls wich Positions and size depend on Horizontal Center of dialogue
		ptSHC  = 0x80,	// Controls wich size depends on Horizontal Center of dialogue

		ptPDR = ptPR | ptPD,	// Controls wich Positions depend on Rigth-Down conner of dialogue
		ptSDR = ptSR | ptSD, 	// Controls wich Sizes depend on Rigth-Down conner of dialogue

		ptPDSR = ptPD | ptSR,
		ptPRSD = ptPR | ptSD,

		ptPHCSD = ptPHC |ptSD,
		ptPSHCSD = ptPSHC | ptSD,
		ptSHCSD = ptSHC | ptSD,
		ptPHCSDR = ptPHC | ptSDR,
		ptPVCSDR = ptPVC | ptSDR,
		ptPCenter = ptPVC | ptPHC,

		ptSizeMask = ptSR | ptSD
	};

	int m_x, m_y; // Initial dialogue size
	HWND m_hDlg;
	HICON m_hIcon;
	RECT m_rOrig;

	class cCtrl
	{
		bool hidden;
	  public:
		HWND hWnd;
		ePosType pt;
		int dx;
		int dy;
		RECT r;
		RECT rOrig;

		cCtrl *next;

		void Show( bool show );

		cCtrl()
		{
			memset( this, 0, sizeof( *this ) );
		}

	} *head;

private:
	cCtrl *FindCtrl( HWND hWnd );
public:

	cSizedCtrls();
	~cSizedCtrls();

	void Init( HWND hDlg ); //, HICON hIcon = NULL );
	void UpdateItem( cCtrl *ctrl );

	//
	//	AddCtrl
	//		Allows to reassign element type ( see FindCtrl private function ).
	//		It is useful if collective assigment function is used AddAll or AddBelow before.
	//
	void AddCtrl( HWND hWnd, ePosType type );
	void AddCtrl( int idc, ePosType type ) { AddCtrl( GetDlgItem( m_hDlg, idc ), type ); }

	void OnSize( int x, int y );
	void OnSize( LPARAM lParam ) { OnSize( LOWORD( lParam ), HIWORD( lParam ) ); }

	void AddBelow( HWND hDlg, HWND hWnd );
	void AddAll( HWND hDlg, ePosType pType, int *excludeT = NULL, int excludeQ = 0 );
};

//
//	Class cGetDlgCodeCtrl ( see cGetDlgCodeCtrl.h )
//		By deafult controls at dialog do not get VK_RETURN and VK_ESCAPE.
//		Dialog gets IDOK or IDCANCEL instead.
//		This class allows to catch VK_RETURN and VK_ESCAPE with WM_KEYDOWN message for correponded control.
//

class cDlgBase : public cDlg
{
protected:
	virtual LONG_PTR OnNotify( NMHDR *pnmh ) { return false; }
	virtual bool InitDialog() { return true; }
	virtual bool OnDestroy() { return false; }
	virtual void OnSize( int x, int y ) {}
	virtual bool OnCommand( WORD ctrlId, HWND hCtrl = 0, WORD isAccelerator = 0 )
	{
		if ( !isAccelerator )
			EndDialog( m_hDlg, ctrlId );
		return true;
	}

	virtual bool OnCommand( WPARAM wParam, LPARAM lParam )
	{
		return OnCommand(  LOWORD( wParam ), (HWND)lParam, HIWORD( wParam ) ); // ?  0 : 1 );
	}

	//virtual void OnHelpInfo( HELPINFO *pHelpInfo = NULL ) {}

	virtual bool OnScroll( bool bVertScroll, HWND hCtrl, WORD wPos, WORD wRequest ) { return false; }

	//
	//	OnInitPopupMenu
	//		Opens submenu, walks it's item and and updates them by virtual OnUpdatePopupMenuItem.
	//
	bool OnInitPopupMenu( HMENU hMenu, WORD pos, BOOL isWindowMenu );
	virtual bool OnUpdatePopupMenuItem( HMENU hMenu, UINT id ) { return false; }


public:
	
	int m_idcInfo; // Main information control. The text is used as caption when the window is minimised.
	cWCStr m_wcsCaption; // Initially is updated at RealDlgProc catchs WM_INITDIALOG 
	virtual void UpdateCaption();
	//virtual void SetCaption() {	m_wcsCaption.GetWindowTextT( m_hDlg ); }

	cDlgBase()
	{
		m_idcInfo = 0; // It does not update caption if is minimized by default.

		//
		//	It does not call UserMessage by default.
		//	User hast specify m_wmUser_last more then m_wmUser_first;
		//
		m_wmUser_first = WM_USER;
		m_wmUser_last = m_wmUser_first;
	}

	cSizedCtrls m_sizedCtrls;

	UINT m_wmUser_first;
	UINT m_wmUser_last;

	virtual LRESULT UserMessage( UINT uMsg, WPARAM wParam, LPARAM lParam ) { return 0; }

	virtual LRESULT RealDlgProc( UINT uMsg, WPARAM wParam, LPARAM lParam );
};


#define _EnableMenuItem( hmInit, id, enable )\
	EnableMenuItem( hmInit, id, MF_BYCOMMAND | ( enable ? 0 : MF_GRAYED ) )

#define _CheckMenuItem( hmInit, id, check )\
	CheckMenuItem( hmInit, id, MF_BYCOMMAND | ( check ? MF_CHECKED : MF_UNCHECKED ) )

#define _GetMenuCheck( hmInit, id )\
	( GetMenuState( hmInit, id, MF_BYCOMMAND ) == MF_CHECKED )

LPARAM	ListView_GetItemParam	( HWND hLView, int itemNum );
void	ListView_SetItemParam	( HWND hLView, int itemNum, LPARAM param );
int		ListView_GetItemImage	( HWND hLView, int itemNum );
void	ListView_SetItemImage	( HWND hLView, int itemNum, int imageNum );
int		ListView_GetItemIndent	( HWND hLView, int itemNum );

BOOL ToolBar_EnableButton( HWND hToolbar, int id, BOOL enable );
BOOL ToolBar_CheckButton( HWND hToolbar, int id, BOOL check );
int ToolBar_AddButton( HWND hToolbar, int iBmp, int iCmd, int style = 0, int state = TBSTATE_ENABLED );
int ToolBar_AddSep( HWND hToolbar  );
DWORD Toolbar_GetStyle( HWND hToolbar );
void Toolbar_SetStyle( HWND hToolbar, DWORD dwStyle );
BOOL Toolbar_SetUnicodeFormat( HWND hToolbar, bool bSet = true );

HIMAGELIST Toolbar_GetImageList( HWND hToolbar  ); // TB_GETIMAGELIST

bool FillRect_WithGradient( HDC hdc,  RECT *r, COLORREF crStart, COLORREF crFinish, DWORD dwMode = GRADIENT_FILL_RECT_H ); // GRADIENT_FILL_TRIANGLE

class sIdcAndStr
{
friend class cChkDlg;

	wchar_t *str;
public:
	int idc;
	int ids;
	
	sIdcAndStr()
	{
		memset( this, 0, sizeof( *this ) );
	}

	void Set( int pIDC, const wchar_t *szText )
	{
		idc = pIDC;
		ids = 0;
		updateString( &str, szText );
	}

	void Set( int pIDC, int pIDS )
	{
		idc = pIDC;
		ids = pIDS;
		deleteString( &str );
	}

	~sIdcAndStr()
	{
		deleteString( &str );
	}
};


class cRect : public RECT
{
public:
	cRect() { memset( this, 0, sizeof( *this ) ); }

	static LONG Width ( RECT *r ) { return r->right - r->left; }
	static LONG Height( RECT *r ) { return r->bottom - r->top; }

	LONG Width()  { return Width( this );  }
	LONG Height() { return Height( this ); }

	void Set( RECT &r )
	{
		memcpy( this, &r, sizeof( r ) );
	}

	static bool Belong( const RECT *r, int x, int y )
	{
		if ( x < r->left	)	return false;
		if ( x > r->right	)	return false;
		if ( y < r->top		)	return false;
		if ( y > r->bottom	)	return false;
		return true;
	}

	bool Belong( int x, int y )
	{
		return Belong( this, x, y );
	}
};

class cSize : public SIZE
{
public:
	cSize( cRect *r = NULL )
	{
		memset( this, 0, sizeof( *this ) );
		if ( r != NULL )
			GetSize( r );
	}

	bool GetSize( cRect *r )
	{
		if ( r == NULL )
			return false;

		cx = r->right - r->left;
		cy = r->bottom - r->top;

		return true;
	}
};

class cDlgRect : public cRect
{
public:
	cDlgRect() : cRect() {}
};

class cFonts
{
public:
	HFONT m_boldFont;
	HFONT m_titleFont;
	HFONT m_boldFontMin;
	
	cFonts()
	{
		m_boldFont = NULL;
		m_titleFont = NULL;
		m_boldFontMin = NULL;
	}

	~cFonts()
	{
		if ( m_boldFont )
			DeleteObject( m_boldFont );

		if ( m_boldFontMin )
			DeleteObject( m_boldFontMin );

		if ( m_titleFont )
			DeleteObject( m_titleFont );
	}

	void CreateFonts( HDC hDC )
	{
		//int nHeight = -MulDiv(12, GetDeviceCaps( hDC, LOGPIXELSY ), 72);
		//pApp->m_titleFont = CreateFontA( nHeight, 0, 0, 0, 
		//	FW_BOLD, FALSE, FALSE, FALSE, DEFAULT_CHARSET, 
		//	OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS, DEFAULT_QUALITY, VARIABLE_PITCH | FF_SWISS, "ARIAL");

		//nHeight = -MulDiv( 10, GetDeviceCaps( hDC, LOGPIXELSY ), 72);
		//pApp->m_boldFont = CreateFontA( nHeight, 0, 0, 0, 
		//	FW_BOLD, FALSE, FALSE, FALSE, DEFAULT_CHARSET, 
		//	OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS, DEFAULT_QUALITY, VARIABLE_PITCH | FF_SWISS, "MS Shell Dlg" );

		int nHeight = -MulDiv(12, GetDeviceCaps( hDC, LOGPIXELSY ), 72);

		m_titleFont = CreateFontA( nHeight, 0, 0, 0, 
			FW_BOLD, FALSE, FALSE, FALSE, DEFAULT_CHARSET, 
			OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS, DEFAULT_QUALITY, VARIABLE_PITCH | FF_SWISS, "ARIAL");

		nHeight = -MulDiv( 10, GetDeviceCaps( hDC, LOGPIXELSY ), 72);
		m_boldFont = CreateFontA( nHeight, 0, 0, 0, 
			FW_BOLD, FALSE, FALSE, FALSE, DEFAULT_CHARSET, 
			OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS, DEFAULT_QUALITY, VARIABLE_PITCH | FF_SWISS, "MS Shell Dlg" );

		nHeight = -MulDiv( 8, GetDeviceCaps( hDC, LOGPIXELSY ), 72);
		m_boldFontMin = CreateFontA( nHeight, 0, 0, 0, 
			FW_BOLD, FALSE, FALSE, FALSE, DEFAULT_CHARSET, 
			OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS, DEFAULT_QUALITY, VARIABLE_PITCH | FF_SWISS, "MS Shell Dlg" );
	}
};

//
//	Class cColoredDlg
//		1. Changes background from top to m_idcLine control position
//		2. Changes font for m_idcBold and m_idcTitle controls
//
class cColoredDlg : public cDlg, public cFonts
{
	// Header start color for gradient filling
	COLORREF m_crStart;

public:
	wchar_t *m_caption;
	int m_idcHead;
	int m_idcLine;
	RECT m_lineRect;

	bool m_largeBold;

	int m_idcBold;
	int m_idcTitle; // Title text!
	int m_idiIcon;

	//
	//	Pointer to dialog rectangle to inherite position for wizard like chain dlgs.
	//
	cDlgRect *m_dlgRect;
	cSizedCtrls m_sizedCtrls;
	int m_cxAdd; // = 100 , it is added if text size is more then default box size.
	void ResizeByText( HWND hText, DWORD swpFlag = 0 );

	//
	//	The coloring is used to make dialog style Windows 7 like.
	//	But XP does not properly redraw items after background erasing.
	//	So it make be assigned once to disable coloring at all.
	//	F.e. m_bDisbleColoring = !IH.isW7();
	//	Default is false -> coloring enabled.
	//
	static bool m_bDisbleColoring;
	bool m_changeBkColor;

	int m_gradientMax;

	cColoredDlg() : cDlg()
	{
		m_dlgRect = NULL;
		SetRCValues();
		m_crStart = 0;
		m_gradientMax = 0;
		m_changeBkColor = true;
		m_cxAdd = 100;
		memset( &m_lineRect, 0, sizeof( m_lineRect ) );
		m_largeBold = true;
	}

	~cColoredDlg()
	{
		deleteString( &m_caption );
	}

	void SetRCValues( int idcLine = 0, int idcBold = 0, int idcTitle = 0, int idiIcon = 0, const wchar_t *caption = NULL, int idcHead = 0, COLORREF hdrStartColor = RGB( 0xFF, 0xFF, 0xFF ) ) // RGB( 0x40, 0x80, 0x40 ) - dark green
	{
		m_idcHead = idcHead;
		m_idcLine = idcLine;
		m_idcBold = idcBold;
		m_idcTitle = idcTitle;
		m_idiIcon = idiIcon;
		m_caption = newString( caption );
		
		m_crStart = hdrStartColor;

		if ( m_hDlg != NULL )
			DrawColorProc( WM_INITDIALOG, 0, 0 );
	}

	virtual LRESULT DrawColorProc( UINT uMsg, WPARAM wParam, LPARAM lParam );
	virtual void SetItemTextColor( HDC hDC, HWND hWnd )
	{
		// Sample:
		// If ( hWnd == GetItem( IDC_GreenColor ) )
		//		 SetTextColor( hDC, RGB(0,99,0));
	};

};

//
//	Dialog with text and checked items...
//
//	Sample:
//	int MessageDlg_WithChk( HWND hWnd, const wchar_t *text, const wchar_t *chkText, const wchar_t *caption, bool *pBChecked )
//	{
//		cMBTextDlg dlg( text, IDC_Text, *pBChecked, chkText, IDC_chk, caption );
//		int res = dlg.DoModal( theApp.m_hInst, IDD_ChkOkCancel, hWnd );
//
//		*pBChecked = dlg.m_bChk;
//
//		return res;
//	}
//
class cMBTextDlg : public cColoredDlg
{
	//wchar_t *m_caption;
	wchar_t *m_dlgText;
	wchar_t *m_chkText;
	wchar_t *m_dlgBoldText;

	int m_idcSizedText;
	int m_idcChk;

	HICON m_hIcon;
	int m_idcIcon;
	LPTSTR m_iStd;

public:

	POINT m_pos; // Windows position, by default is -1, -1, if so it is moved to cursor position.
	cRect m_parentRect; // Parent rectangle, if assigend place window in center of the rectangle.

	//cSizedCtrls m_sizedCtrls;
	bool m_bChk;
	bool m_disableCancel; // Does not allow close by left-right X button.

	cMBTextDlg( const wchar_t *dlgText, int idcText, const wchar_t *caption = NULL, int timeoutButton = -1, int idcLine = 0 )
		:	m_itemTextT( 0 )
	{
		Init(  dlgText, idcText, 0, NULL, 0, caption, timeoutButton, 0, idcLine );		
	}

	cMBTextDlg( const wchar_t *dlgText, int idcText, bool bChk, const wchar_t *chkText, int idcChk, const wchar_t *caption,
				int timeoutButton = -1, int idcLine = 0, int idiMainIcon = 0, int idiIcon = 0, LPTSTR iStd = IDI_EXCLAMATION )
		:	m_itemTextT( 0 )
	{
		Init( dlgText, idcText, bChk, chkText, idcChk, caption, timeoutButton, idiMainIcon, idcLine, idiIcon, iStd );
	}

	void setBoldText( const wchar_t *dlgText, int idcText, bool bLarge = false )
	{
		updateString( &m_dlgBoldText, dlgText );
		m_idcBold = idcText;
		m_largeBold = bLarge;
	}

	struct cItemTextInfo {
		const wchar_t *m_text;
		int m_idc;
	};

	cADBlock < cItemTextInfo > m_itemTextT;

	void setItemText( const wchar_t *text, int idcText )
	{
		cItemTextInfo elem;
		elem.m_idc = idcText;
		elem.m_text = newString( text );
		m_itemTextT.Add( elem );
	}

	void Init( const wchar_t *dlgText, int idcText, bool bChk, const wchar_t *chkText, int idcChk, const wchar_t *caption = NULL,
		int timeoutButton = -1, int idiMainIcon = 0, int idcLine = 0, int idiIcon = 0, LPTSTR iStd = IDI_EXCLAMATION  )
	{
		SetRCValues( idcLine, 0, 0, idiMainIcon, caption );
		m_dlgText = newString( dlgText );
		m_chkText = newString( chkText );
		m_dlgBoldText = NULL;

		m_idcSizedText = idcText;
		m_idcChk = idcChk;
		m_bChk = bChk;
		m_timeoutButton = timeoutButton;
		m_timeoutButtonId = -1;
		m_hTimeoutButton = NULL;
		m_timeoutCountdown = 10;
		m_timerEvent = 0;

		m_pos.x = -1; // -1 means -> move to cursor position.
		m_pos.y = -1;
		memset( &m_parentRect, 0, sizeof( m_parentRect ) );
		m_parentRect.top = -1;

		m_disableCancel = false;

		m_hIcon = NULL;
		m_idcIcon = idiIcon;
		m_iStd = iStd;
	}
	
	~cMBTextDlg()
	{
		deleteString( &m_dlgText );
		deleteString( &m_chkText );
		deleteString( &m_dlgBoldText );

		for( int i = 0; i < m_itemTextT.m_Q; i++ )
			deleteString( &m_itemTextT.m_T[i].m_text );
	}

	void setTimeoutCount( int timeoutButtonId, int sec, int timeoutButtonNumber = -1 )
	{
		//
		//	One of the value 'number' or 'id' may be used
		//	Prev version uses m_timeoutButton ( number ) at Init()
		//
		m_timeoutButtonId = timeoutButtonId;
		m_timeoutButton = timeoutButtonNumber;
		m_timeoutCountdown = sec;
	}

 private:

	int m_timeoutButton;	// Number of timeout button in tab order, by default is -1
	int m_timeoutButtonId;	// Identifier of timeout button, by default is -1

	HWND m_hTimeoutButton;
	UINT_PTR m_timerEvent;
	cWCStr m_wcsTimeoutButtonText;
	int m_timeoutCountdown; // Defult 10 is assigned at WM_INITDIALOG if m_timeoutButton is not -1

	UINT_PTR timeoutProc( UINT uMsg, UINT_PTR idEvent, DWORD dwTime );
	static VOID CALLBACK static_timeoutProc( HWND hDlg, UINT uMsg, UINT_PTR idEvent, DWORD dwTime );

	LRESULT RealDlgProc( UINT uMsg, WPARAM wParam, LPARAM lParam );
	static BOOL CALLBACK addButton( HWND hWnd, LPARAM lParam );
};

struct sIdcAndChk
{
	int idc;
	bool chk;

	sIdcAndChk()
	{
		memset( this, 0, sizeof( *this ) );
	}
};

//
//	Dialog box with text ( m_idText ) and one check box ( m_idCheck )
//	m_bChk value is updated when dialog is closed by buttons IDOK or IDCANCEL or IDYES or IDNO
//	DoModal function returns the close ID
//
class cChkDlg : public cColoredDlg // cDlg
{
	sIdcAndStr *m_strT;
	int m_strQ;
	sIdcAndStr m_singleStr;

	sIdcAndChk *m_chkT;
	int m_chkQ;
	sIdcAndChk m_singleChk;

	//
	//	Single item text with several foramated strings.
	//
	const wchar_t **m_formatTextT;
	int m_formatTextQ;

public:

	bool m_bChk;
	HICON m_hIcon;
	int m_idIcon;

	//
	//	Automtically find the single check box item.
	//	Inites m_bChk as false.
	//	User can check the m_bChk after the dialog closing.
	//
	cChkDlg()
	{
		Init();
	}

	//
	//	The dialog reads string with idetifier idText as pattern ( f.e. str == " ... %1 ... %2 ... %3 " ),
	//	formats it by FormatT( bf, bflen, str, textT[0], textT[1], textT[2] ),
	//	and set the 'bf' text to item idText back.
	//
	cChkDlg( int idText, const wchar_t **textT, int textQ, int idCheck = 0, bool bCheck = false )
	{
		m_singleStr.idc = idText;
		//m_singleStr.str = NULL;

		if ( idCheck == 0 )
		{
			m_chkT = NULL;
			m_chkQ = 0;
		}
		else
		{
			m_singleChk.idc = idCheck;
			m_singleChk.chk = bCheck;
			m_chkT = &m_singleChk;
			m_chkQ = 1;
		}

		Init( &m_singleStr, 1, m_chkT, m_chkQ, textT, textQ );
	}

	cChkDlg( int idText, const wchar_t *szText, int idCheck = 0, bool bCheck = false )
	{		
		m_singleStr.Set( idText, szText );

		if ( idCheck == 0 )
		{
			m_chkT = NULL;
			m_chkQ = 0;
		}
		else
		{
			m_singleChk.idc = idCheck;
			m_singleChk.chk = bCheck;
			m_chkT = &m_singleChk;
			m_chkQ = 1;
		}

		Init( &m_singleStr, 1, m_chkT, m_chkQ );
	}

	void Init( sIdcAndStr *strT = NULL, int strQ = 0, sIdcAndChk *chkT = NULL, int chkQ = 0, const wchar_t **textT = NULL, int textQ = 0)
	{
		m_hIcon = NULL;
		m_idIcon = 0;

		m_strT = strT;
		m_strQ = strQ;
		m_chkT = chkT;
		m_chkQ = chkQ;
		m_formatTextT = textT;
		m_formatTextQ = textQ;
	}

	cChkDlg( sIdcAndStr *strT, int strQ, sIdcAndChk *chkT, int chkQ )
	{
		Init( strT, strQ, chkT, chkQ );
	}

	cChkDlg( sIdcAndStr *strT, int strQ, int idCheck = 0, bool bCheck = false )
	{
		m_singleChk.idc = idCheck;
		m_singleChk.chk = bCheck;

		if ( idCheck == 0 )
		{
			m_chkT = NULL;
			m_chkQ = 0;
		}
		else
		{
			m_singleChk.idc = idCheck;
			m_singleChk.chk = bCheck;
			m_chkT = &m_singleChk;
			m_chkQ = 1;
		}

		Init( strT, strQ, m_chkT, m_chkQ );
	}

	virtual LRESULT RealDlgProc( UINT uMsg, WPARAM wParam, LPARAM lParam );
};

class cEnumDlgItem
{
public:
	int m_startFrom;
	
	cEnumDlgItem( int startFrom = 0 )
	{
		m_startFrom	= startFrom;
	}

	static BOOL CALLBACK staticEnumChildProc( HWND hWnd, LPARAM lParam )
	{
		cEnumDlgItem *pThis = (cEnumDlgItem *)lParam;

		if ( pThis == NULL )
			return false;

		return pThis->EnumChildProc( hWnd );
	}

	bool EnumChildProc( HWND hWnd )
	{
		int id = GetDlgCtrlID( hWnd );

		if ( m_startFrom != 0 )
		{
			if ( m_startFrom != id )
				return true; // Continue searching.

			m_startFrom = 0; // Start item is found.
		}

		return childActor( hWnd );
	}

	bool Walk( HWND hDlg )
	{
		return EnumChildWindows( hDlg, staticEnumChildProc, (LPARAM)this ) == TRUE;
	}

	//
	//	findActor
	//		return false to stop enumeration.
	//
	virtual bool childActor( HWND hWnd ) = 0;
};

//
//	class cFindDlgControl
//	Enumerates all controls in dialog and set m_hCtrl value if control with corresponded class name is found.
//	Note, if className = "msctls_progress", so "msctls_progress32" class name will be used as well.
//
class cFindDlgControl
{
public:
	HWND m_hCtrl;
	const char *m_className;
	int m_len;

	static BOOL CALLBACK enumFuncFind( HWND hWnd, LPARAM lParam )
	{
		cFindDlgControl *pThis = (cFindDlgControl *)lParam;

		if ( pThis == NULL )
			return false;

		return pThis->find( hWnd );
	}

	bool find( HWND hWnd )
	{
		char szClass[ 16 ];

		if (!::GetClassNameA(hWnd, szClass, sizeof(szClass)))
			return true;

		if ( _strnicmp( szClass, m_className, m_len ) != 0 )
			return true;

		m_hCtrl = hWnd;

		return false;
	}

	cFindDlgControl( HWND hDlg, const char *className )
	{
		m_hCtrl = NULL;
		m_className = className;

		if ( m_className == NULL )
			return ;

		m_len = (int)strlen( m_className );

		EnumChildWindows( hDlg, enumFuncFind, (LPARAM)this );
	}
};


//
//	_LoadPopupMenu Helper function, it is used for TrackPopup menu
//	By strange reasons it requires really submenu of any menu...
//
HMENU _LoadPopupMenu( HINSTANCE hInst, UINT id, UINT uSubMenu);

DWORD LV_getStyle( HWND hList );
void LV_setStyle( HWND hList, DWORD style );

#ifndef OPENFILENAME_SIZE_VERSION_400
#define OPENFILENAME_SIZE_VERSION_400 sizeof( OPENFILENAME )
#endif

//
//	DlgEnableItems
//	Enumerates dialog items and update the all placed 
//	after idAfter bottom but before idBefore top positions.
//
bool DlgEnableItems( HWND hDlg, int idAfter, int idBefore, bool bEnable );

//
//	GeDisplayInfo
//		Use Toolbar_SetUnicodeFormat to forse toolbar to get unicode display info with TTN_GETDISPINFOW instead of TTN_GETDISPINFOA
//
bool GetDispInfo( HINSTANCE hInst, LPNMTTDISPINFOA lpnmtdi, int ids );
bool GetDispInfo( HINSTANCE hInst, LPNMTTDISPINFOW lpnmtdi, int ids );

//
//	DrawFileIcon
//		Draws hIcon above the icon defined by id, it is called at WM_PAINT
//
void DrawFileIcon( HWND hDlg, HDC dc, HICON hIcon, int id );

#endif // _Dlg_H_
