#pragma once

#include "cDlg.h"

//
//	The dialog allows to:
//	1. Change control static color of the dialog item which ids are defined at table ( see parameter T at constructor ) 
//	2. Perform one of actions if the item is clicked on:
//		a. ShellExecute( "Open", item text, ... ) -->Default
//		b. ShellExecute( "Open", m_shellCmdStrT[ item number ], ... )
//		c. Call m_actorT[ item number ]( m_hDlg, item identifier );
//

typedef bool typeHitStrActor( HWND hDlg, int id );

class cDlgHitStr: public cColoredDlg
{
	//
	//	Is true if prev hit string processing is not finished.
	//
	static bool m_bExecuted;

public:

	//
	// Number of item in the following tables
	// Note:
	//	1. All tables have same size. 
	//	2. Alternative tables may be NULL or may contain NULL pointers. If so default command is executed.
	//
	int m_Q;

	//
	// Table if dlg item IDs what have to be processed as Hit String.
	//	When Hit String is processed so we find the string at m_T, and check following tables items.
	//	If no one following tables are assigend or corresponded element in the tables is NULL,
	//	so the dialog use ShellExecute command with real text of the item.
	//		Sample: ShellExecute( "open", windowText, ... )
	//	
	//
	int *m_T;

	//
	//	Alternative table with command lines to be executed ( Note as windowText, see above ).
	//	There are three variants for execution:
	//		1. If command line has '!' symbol as prefix. It means the command has to be started with elevating.
	//		2. Every executable path has to be concluded to '\"' symbols. If so the command is processed by RunProcessWithWaitingT( 'command line', ... )
	//		3. If command does not starts from '\"' symbols, so ShellExecute( "open", 'command line', ... ) function is used
	//
	const wchar_t **m_shellCmdStrT;

	//
	// Actor function table to be called when a item is clicked mouse on.
	//
	typeHitStrActor **m_leftActorT;
	typeHitStrActor **m_rightActorT;
	
	typeHitStrActor **m_actorT; // Current actor is assigned by OnLButtonUp/OnRButtonUp

	// The text has to be formatted and assigned in initdialog
	//	int m_idText;
	//	wchar_t *m_text;
	// Use following code at initdialog insted.
	//if ( m_idText != 0 )
	//{
	//	cWCStr pattern;
	//	pattern.GetWindowTextT( GetDlgItem( m_idText ) );

	//	const wchar_t text;
	//	text.FormatT( pattern, m_text );

	//	SetDlgItemTextT( m_idText, text.m_szBody );
	//}

	virtual LRESULT RealDlgProc( UINT uMsg, WPARAM wParam, LPARAM lParam );
	virtual LRESULT DrawColorProc( UINT uMsg, WPARAM wParam, LPARAM lParam );

	cDlgHitStr( int Q, int *T, const wchar_t **shellCmdT = NULL, typeHitStrActor **leftActorT = NULL, typeHitStrActor **rightActorT = NULL )
	{
		Init();
		m_Q = Q;
		m_T = T;
		m_shellCmdStrT	= shellCmdT;
		m_leftActorT	= leftActorT;
		m_rightActorT	= rightActorT;
	}

	void Init()
	{
		m_actorT = NULL;
		m_shellCmdStrT = NULL;
		m_T = NULL;
		m_Q = 0;
		//m_changeBkColor = true;
		//m_bExecuted = false;
	}

	~cDlgHitStr()
	{
		if ( m_brBkgnd != NULL )
			DeleteObject( m_brBkgnd );
	}

protected:

	COLORREF m_clrText, m_blackText;
	HBRUSH m_brBkgnd;
	HCURSOR m_hcFinger;
	HCURSOR m_hcArrow;

	HBRUSH CtlColor( HDC hDC, HWND hWnd, bool *pFound );
	void OnButtonUp( HWND hWnd, POINT point, bool bRight = false );
	virtual void InitDialog();
	virtual int OnCommand( WPARAM wParam, LPARAM lParam );
};
