/*
 *	Copyright 2015 Jetico Inc. All rights reserved.
 *	cDlgMT.h
 *	The dialog starts modal loop at separate thread. It processes IDCANCEL command only and assign cancelled pointer.
 *	DoModelessMT waits dialog initialization and returns handle of created dialog.
 *	Distructor calls EndDialog so thread is stopped.
 *	Created:	02.02.2015
 *	Author:		Alexander Pichuev.
 */

#pragma once

#include "cSemaphore.h"
#include "cDlg.h"

class cDlgMT : public cDlgBase
{
	cSemaphore m_initDlgSemaphore;
	HANDLE m_hThread;
	UINT m_wmUserStop;
	bool *m_pCancelled;
	bool m_cancelled;

	HINSTANCE m_hInst;
	UINT m_idRes;
	HWND m_hParent;

	static DWORD WINAPI startDialogThread( void *lpThreadParameter )
	{
		cDlgMT * pThis = ( cDlgMT * )lpThreadParameter;
		if ( pThis == NULL )
		{
			ODS_fls( L"Assertion! Invalid thread paraneter!" );
			return (DWORD)-1;
		}

		return (DWORD)pThis->DoModal( pThis->m_hInst, pThis->m_idRes, pThis->m_hParent );
	}

public:

	virtual bool OnCommand( WORD ctrlId, HWND hCtrl = 0, WORD isAccelerator = 0 )
	{
		if ( ctrlId == IDCANCEL )
		{
			*m_pCancelled = true;
			EnableWindow( GetItem( IDCANCEL ), false );
		}

		return true;
	}

	virtual bool InitDialog()
	{
		m_initDlgSemaphore.releaseSemaphore();
		return true;
	}
	
	virtual LRESULT UserMessage( UINT uMsg, WPARAM wParam, LPARAM lParam )
	{
		if ( uMsg == m_wmUserStop )
			EndDialog( m_hDlg, IDOK );

		return 0;
	}

	void setCancelledPtr( bool *pCancelled )
	{
		m_pCancelled = pCancelled;
		if ( m_pCancelled == NULL )
			m_pCancelled = &m_cancelled;
	}

	cDlgMT( bool *pCancelled = NULL )
	{
		m_hThread = NULL;
		m_wmUserStop = m_wmUser_last++;
		m_cancelled = false;
		setCancelledPtr( pCancelled );
	}

	virtual ~cDlgMT()
	{
		if ( m_hThread == NULL ) // Was not inited.
			return;

		StopDialogAndCloseThread();
	}

	bool StopDialogAndCloseThread()
	{
		DWORD dwRes = -1;

		if ( m_hDlg == NULL )
			ODS_fls( L"Assertion!" );
		else
		{
			PostMessage( m_hDlg, m_wmUserStop, 0, 0  );

			dwRes = WaitForSingleObject( m_hThread, 1000 );

			if ( dwRes == WAIT_TIMEOUT )
				ODS_fls( L"Assertion! Wait thread timeout!" );
		}

		CloseHandle( m_hThread );
		m_hThread = NULL;

		return dwRes == WAIT_OBJECT_0;
	}

	HWND DoModelessMT( HINSTANCE hInst, UINT idRes, HWND hParent, const wchar_t *caption = NULL )
	{
		m_hInst = hInst;
		m_idRes = idRes;
		m_hParent = hParent;

		m_initDlgSemaphore.createSemaphore( NULL );
		m_hThread = CreateThread( 0, 0, startDialogThread, this, 0, NULL );
		DWORD dwRes = m_initDlgSemaphore.waitForSemaphore( 1000 );

		if ( dwRes == WAIT_TIMEOUT )
		{
			ODS_fls( L"Assertion! Wait thread timeout!" );
			return NULL;
		}

		if ( caption != NULL )
			SetWindowTextT( m_hDlg, caption );

		return m_hDlg;
	}
};
