//*******************************************************************************************
//
//	Filename : DropTarget.h
//	Desription:
//		Contains class cDropTarget.
//	Modified:
//		05.03.2015 - Simplefied using for cDlg is implemented.
//			1. Use this class as base for any nested from cDlg
//			2. Call  cDropTarget::DlgProc( HWND hWnd, UINT uMsg ) at RealDlgProc of the class.
//			3. Implement virtual Drop( cFilesList &ofl ) function
//			4. Implementation of others virtual function is not nessesary, but in this case
//				m_bSupportDrop varibles has to be assigned to accept drag functions
//
//*******************************************************************************************

#pragma once

#include "cDlg.h"
#include "cFilesList.h"

class cDropTarget : public IDropTarget
{
	CRefCount m_cRef;
	HWND m_hWnd;
	DWORD m_keyState;

	HRESULT m_hrOLE;

 public:

	cDropTarget( HWND hWnd = NULL )
	{
		m_hrOLE = NULL;
		m_hWnd = NULL;
		m_bSupportDrop = false;

		if ( hWnd != NULL )
			Register( hWnd );
	}

	int Register( HWND hWnd )
	{
		m_hWnd = hWnd;
		m_hrOLE = OleInitialize( NULL );

		AddRef();

		int res = CoLockObjectExternal( this, TRUE, FALSE );

		if ( res == S_OK )
			res = RegisterDragDrop( m_hWnd, this );

		// Errors:
		//		S_OK
		//		E_OUTOFMEMORY

		return res;
	}

	void Revoke()
	{
		RevokeDragDrop( m_hWnd );
		CoLockObjectExternal( this, FALSE, TRUE );
		Release();

		if ( m_hrOLE == S_OK )
			OleUninitialize();
	}

	~cDropTarget()
	{
	}

	STDMETHODIMP QueryInterface(REFIID riid, LPVOID *ppvObj )
	{
		*ppvObj = NULL;

		LPUNKNOWN pObj;
 
		if (riid == IID_IUnknown)
		{
			pObj = (IUnknown*)((IDropTarget*)this); 
		}
		else if (riid == IID_IDropTarget)
		{
			pObj = (IUnknown*)((IDropTarget*)this); 
		}
		else
		{
   			return(E_NOINTERFACE);
		}

		pObj->AddRef();
		*ppvObj = pObj;

		return(NOERROR);
	}

	STDMETHODIMP_(ULONG) AddRef()
	{
		return(m_cRef.AddRef());
	}

	STDMETHODIMP_(ULONG) Release()
	{
		if (!m_cRef.Release())
		{
   			delete this;
			return(0);
		}

		return(m_cRef.GetRef());
	}

	// Just call it RealDlgProc
	void DlgProc( HWND hWnd, UINT uMsg )
	{
		if ( uMsg == WM_INITDIALOG )
		{
			Register( hWnd );

			//
			// Add reference here does not allow to delete this object when last reference is released.
			// It is required if this class is used as based for dialog.
			// Otherwise it tries to destroy entire dialogue at Revoke() call.
			//
			AddRef();
		}
		else
		if ( uMsg == WM_DESTROY )
			Revoke();
	}

	bool m_bSupportDrop;

	virtual STDMETHODIMP DragEnter( IDataObject *pDataObject, DWORD grfKeyState, POINTL pt, DWORD *pdwEffect )	{ return m_bSupportDrop ? S_OK : S_FALSE; };
	virtual STDMETHODIMP DragOver( DWORD grfKeyState, POINTL pt, DWORD * pdwEffect )							{ return m_bSupportDrop ? S_OK : S_FALSE; };
	virtual STDMETHODIMP DragLeave( void )																		{ return m_bSupportDrop ? S_OK : S_FALSE; };
	virtual HRESULT Drop( cFilesList &ofl ) { return S_FALSE; }
	virtual STDMETHODIMP Drop( IDataObject *pDataObject, DWORD grfKeyState, _POINTL pt, DWORD * pdwEffect )
	{
		//
		//	Get source file names...
		//

		#define SETDefFormatEtc(fe, cf, med)   \
		{\
		(fe).cfFormat=cf;\
		(fe).dwAspect=DVASPECT_CONTENT;\
		(fe).ptd=NULL;\
		(fe).tymed=med;\
		(fe).lindex=-1;\
		}

		FORMATETC fe =	{ CF_TEXT, NULL, DVASPECT_CONTENT, -1, TYMED_HGLOBAL };
	    
		int res;
		CLIPFORMAT cf;

		for ( cf = 0; cf < CF_MAX; cf++ )
		{
			SETDefFormatEtc( fe, cf, TYMED_HGLOBAL );

			res = pDataObject->QueryGetData(&fe);

			if ( res == S_OK )
				break;
		}

		if ( res != S_OK )
			return res;

		STGMEDIUM mdm;

		memset( &mdm, 0, sizeof( mdm ));

		res = pDataObject->GetData( &fe, &mdm );

		if ( res != S_OK )
		{
			assert( res == S_OK );
			return res;
		}

		if ( (cf == CF_HDROP ) && ( mdm.tymed == TYMED_HGLOBAL ) )
		{
			DROPFILES *pdf = (DROPFILES *)mdm.hGlobal;

			wchar_t *glbPtr = (wchar_t *)((char *)pdf->pFiles + *(DWORD *)(pdf->pFiles));

			cFilesList ofl( glbPtr );
			res = Drop( ofl );

			GlobalFree( mdm.hGlobal );
		}
		
		if ( mdm.pUnkForRelease != NULL )
			mdm.pUnkForRelease->Release();

		return res;
	}
} ;
