/*
 *	Copyright 2015 Jetico, Inc. All rights reserved.
 *	OpenFilesList.h
 *	Description:
 *		Class cFilesList
 *			Walks double zero termintaed multy string buffer. Base class for cOpenFilesList.
 *		Sample:
 *
 *			cFilesList fl( multySZbuffer, -1 );
 *
 *			while( fl.getNext() != NULL )
 *				.. Do somthing with fl.getCur()
 *
 *	Modified:	05.03.2015
 *	Author:		Alexander Pichuev.
 */

#pragma once
#include <windows.h>

class cFilesList
{
protected:

	const wchar_t *m_bf;
	int m_bfSize;
	const wchar_t *m_pos;

public:
	//
	//	dzwcslen
	//		Gets double zero terminated buffer length.
	//
	static inline int dzwcslen( const wchar_t *bf )
	{
		for ( int len = 0;; len++ )
		{
			if ( ( bf[len] == 0 ) && ( bf[++len] == 0 ) )
				return ++len;
		}
	}

	cFilesList( const wchar_t *bf, int bfSize = -1 )
	{
		if ( bfSize == -1 )
			bfSize = dzwcslen( bf );

		m_bf = bf;
		m_bfSize = bfSize;
		m_pos = NULL;
	}

	virtual const wchar_t *getCur()
	{
		return m_pos;
	}
	
	virtual const wchar_t *getNext()
	{
		if ( m_bf == 0 )
			return NULL;

		if ( m_pos == NULL )
		{
			m_pos = m_bf;
		}
		else
		{
			int len = (int)wcslen( m_pos );

			if ( m_pos >= m_bf + m_bfSize )
				return NULL;

			if ( *m_pos == L'\0' )
				return NULL;

			m_pos += len + 1;

			if ( m_pos >= m_bf + m_bfSize )
				return NULL;

			if ( *m_pos == L'\0' )
				return NULL;
		}

		return m_pos;
	}
};

