#pragma once

#include "cWCS.h"

//
// The cLongName checks the file name and generates unicode 32K limited name.
//	The class is used by CreateFileT, FindFirstFileT and accessT functions.
//
// MSDN reports.
//	In the ANSI version of this function, the name is limited to MAX_PATH characters. 
//	To extend this limit to 32,767 wide characters, call the Unicode version of the 
//	function and prepend "\\?\" to the path. For more information, see Naming a File.
//
class cLongName : public cWCS
{
public:

	//
	//	isTooLongForShortName
	//		allows generation of long name from the path
	//
	static bool isTooLongForShortName( const wchar_t *path )
	{
		if ( path == NULL )
		{
			return false;
		}

		if ( wcsncmp( path + 1, L":\\", 2 ) != 0 )
		{
			return false; // Is not valid short name
		}

		if ( wcslen( path ) < MAX_PATH )
		{
			return false; // Is not too long name
		}

		return true;
	}

	void makeLong( const wchar_t *path )
	{
		if ( wcsncmp( path + 1, L":\\", 2 ) == 0 )
			Set( L"\\\\?\\" );

		Add( path );
	}

	const wchar_t *generate( const wchar_t *path )
	{
		if ( path == NULL )
			return NULL;
		
		if ( !isTooLongForShortName( path ) )
			return path;

		makeLong( path );

		return m_szBody;
	}
};
