#ifndef _cMBStr_h_
#define _cMBStr_h_

#include "windefs.h"

#define _getMbs( _WCS ) ( cMBStr( _WCS ).m_szBody )

class cMBStr
{
	static char *allocateAndCopyStr( const wchar_t *wcs, int *pAllocatedLen, int wcsLen = -1, bool doubleZeroTerminated = false, UINT codePage = -1, int *pLen = NULL );
	static char *allocateStr( int strLen, int *pAllocatedLen );
	static void freeStr( char **pStr, int *pAllocatedLen );

public:
	char *m_szBody;
	int m_len; // Current position
	int m_allocatedLen;

	operator const char * ()
	{
		return m_szBody;
	}

	cMBStr( const wchar_t *ws, int len = -1, bool doubleZeroTerminataed = false, UINT codePage = (UINT)-1 );
	cMBStr( const char *mbs, int len = -1 );
	cMBStr( int len = 0 );
	~cMBStr();

	bool Set( const wchar_t *ws, int len = -1, bool doubleZeroTerminated = false, UINT codePage = (UINT)-1 );
	bool Set( const char *mbs, int mbsLen = -1	);
	bool Set( int mbsLen ) { return Set( (const char *)NULL, mbsLen ); }

	template < class T> const char *SetBody( const T *s, int len = -1 )
	{
		Set( s, len );
		return m_szBody;
	}

	int Add( const wchar_t *wcs, int len = -1, bool doubleZeroTerminated = false, UINT codePage = (UINT)-1 );
	int Add( const char *mbs = NULL, int len = -1 );
	int Add( int len ) 	{ return Add( (const char *)NULL, len ); }
	int Add( const unsigned char *mbs, int len ) { return Add( (const char *)mbs, len ); }
	int Add( const char ch ) { return Add( (const char *)&ch, 1 ); }

	char *SetAndGet( const wchar_t *ws, int len = -1, bool doubleZeroTerminated = false, UINT codePage = (UINT)-1 )
	{
		if ( !Set( ws, len, doubleZeroTerminated, codePage ) )
			return NULL;

		return m_szBody;
	}

	char *SetAndGet( const char *mbs, int mbsLen = -1	)
	{
		if ( !Set( mbs, mbsLen ) )
			return NULL;

		return m_szBody;
	}

	char *SetAndGet( int mbsLen ) { return SetAndGet( (const char *)NULL, mbsLen ); }

	int WriteTo( wchar_t *bf, int bfMaxLen, UINT codePage = (UINT)-1 );

	void Free();

	void Swap( cMBStr *mbs )
	{
		char *szBody = m_szBody;
		int len = m_len;
		int allocatedLen = m_allocatedLen;

		m_szBody = mbs->m_szBody;
		m_len = mbs->m_len;
		m_allocatedLen = mbs->m_allocatedLen;

		mbs->m_szBody = szBody;
		mbs->m_len = len;
		mbs->m_allocatedLen = allocatedLen;
	}

	bool isEmpty()
	{
		return GetLength() == 0;
	}

	bool equ( const char *str )
	{
		if ( str == NULL )
			return isEmpty();

		if ( isEmpty() )
			return *str == '\0';

		return strcmp( m_szBody, str ) == 0;
	}

	int SetLength( int len )
	{
		if ( m_len >= len )
		{
			memset( m_szBody + len, '\0', m_len - len );
			m_len = len;
			//m_szBody[ m_len ] = '\0';
		}
		else
		if ( Add( (const char*)NULL, len - m_len ) == 0 )
			return 0;

		return m_len;
	}

	int GetLength()
	{
		if ( m_szBody == 0 )
			return 0;

		return (int)strlen( m_szBody );
	}

	const char *MakeUpper();
	const char *MakeLower();
};

#endif // #ifndef _cMBStr_h_

