#pragma once

#include <windows.h>
#ifdef _noAdvApi_
#include "..\Includes\ODS.h"
#else
#include "..\Includes\AdvAPI.h"
#endif

class cMutex
{
	HANDLE m_Handle;
	int  m_err;

public:

	BOOL m_newCreated;

	cMutex( const wchar_t *szName = NULL, BOOL bInitialOwner = TRUE, bool bSecure = false )
	{ 
		m_err = ERROR_SUCCESS;
		m_newCreated = FALSE;
		m_Handle = NULL;

		if ( szName != NULL )
			create( szName, bInitialOwner, bSecure );
	}

	HANDLE create( const wchar_t *szName, BOOL bInitialOwner = TRUE, bool bSecure = false )
	{
		ODSFuncS( szName );

		if ( bSecure )
		{
#ifdef _noAdvApi_
			ODS( L"Create non secure semaphore" );
			ODS_fls( L"Assertion!" );
			m_Handle = NULL;
			m_err = 0xfffffffe;
#else
			m_Handle = CreateSecureMutex( bInitialOwner, szName, &m_err ); 
#endif
		}
		else
		{
			m_Handle = CreateMutexW( NULL, bInitialOwner, szName ); 
			m_err = GetLastError();
		}

		if ( m_Handle == NULL )
			m_err = 0xffffffff;
		else
		{
			if ( m_err != ERROR_ALREADY_EXISTS )
				m_newCreated = TRUE;
		}

		return m_Handle;
	}
	
	~cMutex()
	{
		if ( m_Handle != NULL )
		{
			BOOL res = ReleaseMutex( m_Handle );
			CloseHandle( m_Handle );
			m_Handle = NULL;
		}
	}
};