#include "cProfileTable.h"
#include "cMBStr.h"

int cProfileTableSection::GetValueNumber( const wchar_t *szValueName, int len )
{
	if ( len == -1 )
		len = (int)wcslen( szValueName );

	int i = m_valueT.m_Q;

	while( i-- )
	{
		if ( wcslen( m_valueT.m_T[i].m_name ) != len )
			continue;

		if ( wcsncmp( szValueName, m_valueT.m_T[i].m_name, len ) == 0 )
			break;
	}

	return i;
}

cProfileTableValue *cProfileTableSection::FindValue( const wchar_t *szValueName, int len, bool bCreate )
{
	cProfileTableValue *tpValue = NULL;

	int i = GetValueNumber( szValueName, len );
	
	if ( i != -1 )
	{
		tpValue = m_valueT.m_T + i;
		return tpValue;
	}

	if ( !bCreate )
		return NULL;
	
	tpValue = m_valueT.Add( NULL, 1 );

	if ( tpValue != NULL )
		tpValue->m_name = newString( szValueName, len );

	return tpValue;
}

bool cProfileTableSection::RemoveValue( const wchar_t *szName, int nameLen )
{
	int i = GetValueNumber( szName, nameLen );

	if ( i == -1 )
		return false;

	m_valueT.m_T[i].Free();

	int res = m_valueT.Remove( i );

	return res >= 0;
}

cProfileTableValue *cProfileTableSection::AddValue( const wchar_t *szName, int nameLen, const wchar_t *szValue, int valueLen )
{
	cProfileTableValue *tpValue = FindValue( szName, nameLen, true );
	
	//
	//	The value has been found with create attribute.
	//
	if ( tpValue == NULL )
	{
		ODS_fls( L"Asertion!" );
		return NULL;
	}

	if ( szValue != NULL )
		tpValue->m_value = newString( szValue, valueLen );

	return tpValue;
}


cProfileTableSection *cProfileTable::FindSection( const wchar_t *szSection, int len, bool bCreate )
{
	cProfileTableSection *ptSection = NULL;

	if ( len == -1 )
		len = (int)wcslen( szSection );

	for ( int i = m_sectionT.m_Q; i--; )
	{
		if ( wcslen( m_sectionT.m_T[i].m_name ) != len )
			continue;

		if ( wcsncmp( m_sectionT.m_T[i].m_name, szSection, len ) == 0 )
		{
			ptSection = m_sectionT.m_T + i;
			return ptSection;
		}
	}
	
	if ( bCreate )
	{
		ptSection = m_sectionT.Add( NULL, 1 );

		if ( ptSection == NULL )
			return NULL;

		ptSection->m_name = newString( szSection, len );
	}

	return ptSection;
}


bool cProfileTable::ReadFromBuffer( BYTE *bf, int bfSize )
{
	cStringBuffer sb;
	if ( !sb.SetBuffer( bf, bfSize ) )
		return false;

	return ReadFromStringBuffer( &sb );
}

bool cProfileTable::ReadFromStringBuffer( cStringBuffer *sb )
{
	// sb->Restart();

	cWCStr bf( 1024 );
	
	cProfileTableSection *curSection = NULL;

	while( sb->ReadStringT( bf.m_szBody, bf.m_len ) )
	{
		ODS_fls( cWCStr().FormatT( L"'%s", bf.m_szBody ) );

		if ( bf[0] == L'[' )
		{
			int len = (int)wcslen( bf.m_szBody );

			if ( len < 2 )
			{
				ODS_fls( cWCStr().FormatT( L"Assertion at string '%s'", bf.m_szBody ) );
				return false;
			}

			if ( bf[len-1] != L']' )
			{
				ODS_fls( cWCStr().FormatT( L"Assertion at string '%s'", bf.m_szBody ) );
				return false;
			}

			curSection = FindSection( bf.m_szBody + 1, len - 2, true );
			continue;
		}

		if ( curSection == NULL )
		{
			ODS_fls( cWCStr().FormatT( L"Assertion! No one section is initialised, '%s'", bf.m_szBody ) );
			return false;
		}
		
		int pos = 0;

		for(;; pos++ )
		{
			if ( bf.m_szBody[pos] == L'\0' )
			{
				ODS_fls( cWCStr().FormatT( L"Assertion! The string is not valid value - '%s', buffer size %s", bf.m_szBody, cWCStr().GetSizeText( sb->m_bfSize ) ) );
				return false;
			}

			if ( bf.m_szBody[pos] == L'=' )
				break;
		}

		if ( curSection->FindValue( bf.m_szBody, pos, false ) != NULL )
		{
			ODS_fls( cWCStr().FormatT( L"Assertion! Same value already exists - '%s'", bf.m_szBody ) );
			return false;
		}

		if ( curSection->AddValue( bf.m_szBody, pos, bf.m_szBody + pos + 1 ) == NULL )
		{
			ODS_fls( cWCStr().FormatT( L"Assertion! Error in adding value - '%s'", bf.m_szBody ) );
			return false;
		}
	}

	return true;
}

static bool writeStrToPtb( cPlainTextBlock *ptb, bool bUnicode, const wchar_t *str )
{
	if ( bUnicode )
		return ptb->WriteString( str, false ) != 0;

	cMBStr mbs( str );

	return ptb->WriteString( mbs.m_szBody, false ) != 0;
}

bool cProfileTable::WriteToPtb( cPlainTextBlock *ptb, bool bUnicode )
{
	for( int i = 0; i < m_sectionT.m_Q; i++ )
	{
		if (  m_sectionT.m_T[i].m_valueT.m_Q == 0 )
			continue;

		cWCStr wcs;

		wcs.FormatT( L"\r\n[%s]\r\n", m_sectionT.m_T[i].m_name );

		if ( !writeStrToPtb( ptb, bUnicode, wcs.m_szBody ) )
			return false;

		for( int j = 0; j < m_sectionT.m_T[i].m_valueT.m_Q; j++ )
		{
			wcs.FormatT( L"%s=%s\r\n", m_sectionT.m_T[i].m_valueT.m_T[j].m_name, m_sectionT.m_T[i].m_valueT.m_T[j].m_value );

			if ( !writeStrToPtb( ptb, bUnicode, wcs.m_szBody ) )
				return false;
		}
	}

	return true;
}
