#include "cRandomGen.h"
#include <assert.h>
#include <time.h>

// #define USE_SHA1_RANDOM

#ifdef USE_SHA1_RANDOM
#include "Z:\CertificateLib\RNG\sha1random.h"
#endif

bool cRandomGenMT::WriteToFile( HANDLE hFile, DWORD size, DWORD *pWritten )
{
	assert( (int)size <= buffSize );

	DWORD written = 0;

	bool res = WriteFile( hFile, bfReady, size, &written, NULL )
				&& ( written == size );

	if ( pWritten )
		*pWritten = written;

	if ( !res )
	{
		return false;
	}

	while( true )
	{
		if ( stop )
			break;

#ifdef _TestWaiting
		int cl = clock();
#endif
		DWORD dwRes = WaitForSingleObject( hWriteSemaphore, 1000 );
#ifdef _TestWaiting
		fillWaiting += clock() - cl;
#endif

		if ( dwRes != WAIT_TIMEOUT )
			break;
	}

	BYTE *ptr = bfReady;
	bfReady = bfFilled;
	bfFilled = ptr;

	LONG prev;
	ReleaseSemaphore( hSemaphore, 1, &prev );

	return true;
}

void cRandomGenMT::GenerationLoop()
{
	initRandom();

	while( true )
	{
		LONG prev;
		ReleaseSemaphore( hWriteSemaphore, 1, &prev );

		while( true )
		{
			if ( stop )
				return;
#ifdef _TestWaiting
			int cl = clock();
#endif
			DWORD dwRes = WaitForSingleObject( hSemaphore, 1000 ); // INFINITE );
#ifdef _TestWaiting
			writeWaiting += clock() - cl;
#endif
			if ( dwRes != WAIT_TIMEOUT )
				break;
		}

		fillBuffer( bfFilled, buffSize );
	}
}

cRandomGenMT::~cRandomGenMT()
{
	if ( ( hThread != NULL ) && ( hThread != (HANDLE)-1 ) )
	{
		WaitForSingleObject( hThread, 10000 );
		CloseHandle( hThread );
	}

	if ( bfReady != NULL ) delete bfReady;
	if ( bfFilled != NULL ) delete bfFilled;

	if ( hSemaphore != NULL )
		CloseHandle( hSemaphore );

	if ( hWriteSemaphore != NULL )
		CloseHandle( hWriteSemaphore );

	freeRandom();
}

void cRandomGenMT::StopIt()
{
	stop = true;
	LONG prev;
	ReleaseSemaphore( hSemaphore, 1, &prev );
	ReleaseSemaphore( hWriteSemaphore, 1, &prev );

//		if ( ( hThread != NULL ) && ( hThread != (HANDLE)-1 ) )
//			WaitForSingleObject( hThread, 100000 );
}

bool cRandomGenMT::Init( int bfSize )
{
	assert( buffSize == 0 ); // Checks init once.

	generate = bfSize != 0;

	if ( !generate )
		return false;

#ifdef _TestWaiting
	totalClocks = clock();
#endif

	buffSize = bfSize;

	bfReady = new BYTE[ buffSize ];
	bfFilled = new BYTE[ buffSize ];

	if ( ( bfReady == NULL ) || ( bfFilled == NULL ) )
	{
		assert( false );
		MessageBoxA( 0, "Error in memory allocating", "cRandomGenMT", MB_ICONSTOP );
	}

	initRandom();

	fillBuffer( bfReady, buffSize );
	fillBuffer( bfFilled, buffSize );

	hSemaphore = CreateSemaphore( NULL, 0, 100, NULL );
	hWriteSemaphore = CreateSemaphore( NULL, 0, 100, NULL );

	unsigned int id;
	hThread = (HANDLE)_beginthreadex( NULL, 0, GenRandomThreadProc, this, 0, &id );

	return hThread != NULL;
}

#ifdef USE_SHA1_RANDOM

void cRandomGenMT::initRandom()
{
	int result; 

	result = SHA1RandomAllocate( &sRandom );
	if ( result != 0 )
		return;

	result = SHA1RandomInitialize( sRandom );
	if ( result != 0 )
		return;

	DWORD *bf = new DWORD[ 1 ];
		
	DWORD t[4];
	t[0] = clock();
	t[1] = time(0);
	t[2] = DWORD( bf );
	t[3] = bf[0];

	delete bf;
	
	BYTE *seedBuffer = (BYTE *)t;
	int seedBufferSize = sizeof( t );

	result = SHA1RandomRandomize( sRandom, seedBuffer, seedBufferSize );
}

void cRandomGenMT::freeRandom()
{
	if ( sRandom != NULL )
	{
		SHA1RandomFinalize( sRandom );
		SHA1RandomFree( &sRandom );
	}
}

void cRandomGenMT::fillBuffer( BYTE *bf, int bfSize )
{
	if ( sRandom != NULL )
		SHA1RandomGetRandomBytes( sRandom, bf, bfSize );
}

#else // USE_SHA1_RANDOM

void cRandomGenMT::initRandom()
{
	//int val = (int)clock() ^ (int)time(0);
	//srand( val );
	ULARGE_INTEGER uli;
	uli.HighPart = (DWORD)clock();
	uli.LowPart = (DWORD)time(0);
	sRandCtx::setSeed( uli.QuadPart );
}

void cRandomGenMT::freeRandom()
{
}

void cRandomGenMT::fillBuffer( BYTE *bf, int bfSize )
{
	// rand() it return in range 0 RAND_MAX  == 0x7FFF:(
	// small trick -> 
	//		uses small eight bits for first byte
	//		uses seven alder bits and one bits randomly from first byte for the second

	//if ( bfSize <= 1 ) 
	//{
	//	if ( bfSize == 1 )
	//		bf[ 0 ] = (BYTE)rand();

	//	return;
	//}

	//for( int i = bfSize - 2; i >= 0 ; i -= 2)
	//{
	//	WORD w = rand();
	//	bf[ i ] = BYTE( w );
	//	bf[ i+1 ] = BYTE( w >> 7 ) ^ ( BYTE ( w >> ( w & 7 ) ) & 1 );
	//}

	DWORDLONG *dwlBuff = (DWORDLONG*)bf;
	DWORD dwBuffSize = bfSize / sizeof( *dwlBuff );

	int tail = dwBuffSize * sizeof( *dwlBuff );

	for( DWORD i = 0; i < dwBuffSize; i++ )	{ dwlBuff[ i ] = getDWL(); }
	for( ; tail < bfSize; tail++ ) { bf[ tail ] = (BYTE)getDW(); }
}

#endif // USE_SHA1_RANDOM