#include "cRegFuncs.h"
#include "UnicodeFuncs.h"

cRegFuncs::cRegFuncs( const wchar_t *szMainSection )
{
	SetMainSection( szMainSection );
}

void cRegFuncs::SetMainSection( const wchar_t *szMainSection )
{
	if ( szMainSection == NULL )
		m_szMainSection = L"";
	else
		m_szMainSection = getStaticWcs().SetBody( szMainSection );
}

bool cRegFuncs::SetRegBytes( HKEY root, const wchar_t *pSection, const wchar_t * subKey, BYTE *value, int size, bool useMainSection )
{
	cWCSF section;

	if ( useMainSection )
		section.FormatT( L"%s\\", m_szMainSection );

	if ( pSection )
		section.Add( pSection );

	HKEY hKey;

	if ( RegOpenKeyExT( root, section.m_szBody, 0, KEY_SET_VALUE,  &hKey ) != ERROR_SUCCESS )
	{
		if ( RegCreateKeyT( root, section.m_szBody, &hKey ) != ERROR_SUCCESS )
			return FALSE;
	}

	bool ok = ( RegSetValueExT( hKey, subKey, NULL, REG_BINARY, value, size ) == ERROR_SUCCESS );
	
	RegCloseKey(hKey);

	return ok;
}

//
//	GetRegBytes.
//	Opens existing key, allocates required buffer in *pValue, and reads the data.
//	Note the allocated buffer has to be deleted after using.
//
bool cRegFuncs::GetRegBytes( HKEY root, const wchar_t *pSection, const wchar_t *subKey, BYTE **pValue, int *pSize, bool useMainSection )
{
	cWCStr section;

	if ( useMainSection )
		section.FormatT( L"%s\\", m_szMainSection );

	if ( pSection )
		section.Add( pSection );

	HKEY hKey;

	if (RegOpenKeyExT( root, section.m_szBody, 0, KEY_READ, &hKey	) != ERROR_SUCCESS)
		return FALSE;

	DWORD cbData = 0;

	if (RegQueryValueExT( hKey, subKey, NULL, NULL, NULL, &cbData ) != ERROR_SUCCESS )
		return FALSE;

	*pValue = new BYTE [ cbData ];

	bool res = ( RegQueryValueExT( hKey, subKey, NULL, NULL, *pValue, &cbData ) == ERROR_SUCCESS );
	
	if (RegCloseKey(hKey) != ERROR_SUCCESS)
		res = FALSE;

	if ( !res )
	{
		if ( *pValue )
			delete *pValue;
		*pValue = NULL;
		*pSize = 0;

		return FALSE;
	}

	*pSize = cbData;

	return TRUE;
}


bool cRegFuncs::SetRegDW( HKEY root, const wchar_t * pSection, const wchar_t * subKey, DWORD value, bool useMainSection, int sveFlags )
{
	cWCStr section;

	if ( useMainSection )
		section.FormatT( L"%s\\", m_szMainSection );

	if ( pSection )
		section.Add( pSection );

	return SetRegDW( root, section.m_szBody, subKey, value, sveFlags );
}

//static
bool cRegFuncs::SetRegDW( HKEY root, const wchar_t *section, const wchar_t *subKey, DWORD value, int sveFlags ) // = sveNative )
{
	HKEY hKey;

	REGSAM samDesired = KEY_ALL_ACCESS | sveFlags; // KEY_SET_VALUE;

	if ( RegOpenKeyExT(
			root, // HKEY_LOCAL_MACHINE,
			section,
			0,
			samDesired,
			&hKey
		) != ERROR_SUCCESS)
	{
		ODS_fls( _FormatT( L"Error %d in opening key %X, %s, %X", GetLastError(), (DWORD)root, section, samDesired ) );

		DWORD disp;
		if ( RegCreateKeyExT( root, section, 0, NULL, 0, KEY_WRITE | sveFlags, NULL, &hKey, &disp ) != ERROR_SUCCESS )
		{
			ODS_fls( _FormatT( L"Error %d in creating key %X, %s", GetLastError(), (DWORD)root, section ) );
			return FALSE;
		}
	}

	bool ok = ( RegSetValueExT(
			hKey,
			subKey,
			NULL,
			REG_DWORD,
			LPBYTE(&value),
			sizeof(DWORD)
		) == ERROR_SUCCESS );
	
	RegCloseKey(hKey);

	if ( !ok )
	{
		ODS_fls( _FormatT( L"Error %d in setting value %X, %s, %s, %X", GetLastError(), (DWORD)root, subKey, section, value ) );
		return FALSE;
	}

	return ok;
}

DWORD cRegFuncs::GetRegDW( HKEY root, const wchar_t * pSection, const wchar_t * subKey, bool useMainSection, DWORD defValue, int sveFlags )
{
	ODSFuncS( _FormatT( L" %s, %s, %s", m_szMainSection, pSection, subKey ) );
	cWCStr section;

	if ( useMainSection )
		section.FormatT( L"%s\\", m_szMainSection );

	if ( pSection )
		section.Add( pSection );

	HKEY hKey;

	if (RegOpenKeyExT(
			root, // HKEY_LOCAL_MACHINE,
			section.m_szBody,
			0,
			KEY_READ | sveFlags,	// security access mask 
			&hKey			// address of handle of open key 
		) != ERROR_SUCCESS)
	{
		ODS_fl;
		return defValue;
	}

	DWORD Data = defValue;
	DWORD cbData = sizeof(DWORD);

	LONG res = RegQueryValueExT(
			hKey,		// handle of key to query 
			subKey,	// address of name of value to query 
			NULL,
			NULL,	// address of buffer for value type 
			LPBYTE(&Data),	// address of data buffer 
			&cbData 	// address of data buffer size 
		);
	
	RegCloseKey(hKey);

	if ( res != ERROR_SUCCESS)
	{
		ODS_fl;
		return defValue;
	}
	
	ODSFuncRes( Data );
	return Data;
}

bool cRegFuncs::GetRegString( HKEY root, const wchar_t * section, const wchar_t *subKey, cWCS &wcs )
{
	wcs.SetLength( 1024 );
	bool ok = GetRegString( root, section, subKey, wcs.m_szBody, wcs.m_len );
	wcs.Truncate();

	return ok == TRUE;
}

bool cRegFuncs::GetRegString( HKEY root, const wchar_t * section, const wchar_t *subKey, wchar_t * bf, int size)
{
	HKEY hKey;

	if (RegOpenKeyExT(
			root, // HKEY_LOCAL_MACHINE,
			section,
			0,
			KEY_READ,	// security access mask 
			&hKey			// address of handle of open key 
		) != ERROR_SUCCESS)
		return FALSE;

	DWORD cbData = size;

	if (RegQueryValueExT(
			hKey,		// handle of key to query 
			subKey,	// address of name of value to query 
			NULL,
			NULL,	// address of buffer for value type 
			(BYTE *)bf,	// address of data buffer 
			&cbData 	// address of data buffer size 
		) != ERROR_SUCCESS)
		return FALSE;
	
	if (RegCloseKey(hKey) != ERROR_SUCCESS)
		return FALSE;

	return TRUE;
}

bool cRegFuncs::SetRegString( HKEY root, const wchar_t * section, const wchar_t *subKey, const wchar_t * bf )
{
	HKEY hKey;

	if (RegOpenKeyExT(
			root, // HKEY_LOCAL_MACHINE,
			section,
			0,
			KEY_WRITE,	// security access mask 
			&hKey			// address of handle of open key 
		) != ERROR_SUCCESS)
		return FALSE;

	bool res;

	if ( ( bf == NULL ) || ( *bf == 0 ) )
		res = ( RegDeleteValueT( hKey, subKey ) == ERROR_SUCCESS );
	else
	{
		DWORD cbData = ( wcslen( bf ) + 1 ) * sizeof( wchar_t );

		res =
			RegSetValueExT(
				hKey,
				subKey,
				NULL,
				REG_SZ,
				(const BYTE *)bf,
				cbData
				) == ERROR_SUCCESS;
	}
	
	RegCloseKey(hKey);

	return res;
}

//
//	Get and Set function of predefined flags
//	Note, GetFlag function returns true if flag has never set befor.
//
static const wchar_t *szAppFlags = L"AppFlags";

bool cRegFuncs::GetFlag( DWORD flag )
{
	assert( flag < 32 );
	DWORD dwFlags = GetRegDW( HKEY_CURRENT_USER, NULL, szAppFlags, true, -1 );

	if ( dwFlags == -1 ) // Setup assigns defaults
		dwFlags = GetRegDW( HKEY_LOCAL_MACHINE, NULL, szAppFlags, true, -1 );

	DWORD mask = 1 << flag;

	return ( dwFlags & mask ) != 0;
}

bool cRegFuncs::SetFlag( DWORD flag, bool set, HKEY hRoot  ) //= HKEY_CURRENT_USER  )
{
	assert( flag < 32 );
	DWORD dwFlags = GetRegDW( hRoot, NULL, szAppFlags, true, -1 );

	DWORD mask = 1 << flag;

	if ( set )
		dwFlags |= mask;
	else
		dwFlags &= ~mask;

	return SetRegDW( hRoot, NULL, szAppFlags, dwFlags, true ) == TRUE;
}
