#include "cSemaphore.h"
#include "ODS.h"
#ifdef _ODS
#include "FormatT.h"
#endif

#ifndef _noAdvApi_
#include "advapi.h"
#endif // #ifndef _noAdvApi_

bool cSemaphore::createSemaphore( long lInitialCount, long lMaximumCount, const wchar_t *lpName, bool bSecure )
{
	ODSFuncS( _FormatT( L"%d,%d,%s%s", lInitialCount, lMaximumCount, lpName, ( bSecure ? L", Secure": L"" ) ) );

	closeSemaphore();

	if ( bSecure )
	{
#ifdef _noAdvApi_
		ODS( L"Create non secure semaphore" );
		ODS_fls( L"Assertion!" );
		return false;
#else
		m_hSemaphore = CreateSecureSemaphore(  lInitialCount, lMaximumCount, lpName, &m_err ); 
#endif
	}
	else
	{
		m_hSemaphore = CreateSemaphoreW( NULL, lInitialCount, lMaximumCount, lpName ); 
		m_err = GetLastError();
	}

	bool ok = m_hSemaphore != NULL;
	m_bExisted = ok && ( m_err == ERROR_ALREADY_EXISTS );

	ODSFuncRes( _FormatT( L"%04X%s", m_hSemaphore, ( m_bExisted ? L" object existed!" : L"") ) );

	return ok;
}

bool cSemaphore::openSemaphore( DWORD dwDesiredAccess, bool bInheritHandle, const wchar_t *lpName )
{
	m_hSemaphore = OpenSemaphoreW( dwDesiredAccess, bInheritHandle, lpName );
	m_err = GetLastError();
	bool ok = ( m_hSemaphore != NULL );

	if ( !ok && ( dwDesiredAccess == SYNCHRONIZE ) && ( m_err != ERROR_FILE_NOT_FOUND ) )
		ODS_f( _FormatT( L"Error %d", m_err ) );

	return ok;
}

DWORD cSemaphore::waitForSemaphore( DWORD dwMilliseconds ) // = INFINITE
{
	ODSFuncS( _FormatT( L"%04X, %d", m_hSemaphore, dwMilliseconds ) );

	DWORD dwRes = -1;

	if ( m_hSemaphore != NULL )
	{
//#ifdef _ODS
//		if ( dwMilliseconds == INFINITE )
//		{
//			DWORD mSecs = 10 * 60 * 1000;
//
//			dwRes = WaitForSingleObject( m_hSemaphore, mSecs );
//
//			if ( dwRes == WAIT_TIMEOUT )
//			{
//				ODS_fls( _FormatT( L"Assertion: Semaphore %04X, timeout %d mSec", m_hSemaphore, mSecs ) );
//			}
//		}
//
//		if ( dwRes == WAIT_TIMEOUT )
//#endif
		dwRes = WaitForSingleObject( m_hSemaphore, dwMilliseconds );
		m_err = GetLastError();
	}

	ODSFuncRes( dwRes );

	return dwRes;
}

bool cSemaphore::releaseSemaphore( long lReleaseCount, long *lpPreviousCount )
{
	m_prev = 0;

	bool ok = ( ReleaseSemaphore( m_hSemaphore, lReleaseCount, &m_prev ) == TRUE );

	if ( lpPreviousCount != NULL )
		*lpPreviousCount = m_prev;

	m_err = GetLastError();

	return ok;
}

//static 
bool cSemaphore::releaseSemaphore( const wchar_t *semaphoreName, LONG lReleaseCount, LPLONG lpPreviousCount )
{
	HANDLE hSemaphore = OpenSemaphoreW( SEMAPHORE_MODIFY_STATE, false, semaphoreName );

	int err = GetLastError();

	if ( hSemaphore == NULL )
	{
		ODSF( L"Error %x in opening semaphore %s", err, semaphoreName );
		return false;
	}

	LONG prev = -1;
	
	if ( lpPreviousCount == NULL )
		lpPreviousCount = &prev;
	
	BOOL bRes = ReleaseSemaphore( hSemaphore, 1, lpPreviousCount );

	err = GetLastError();

	if ( !bRes )
		ODSF( L"ReleaseSemaphore returns false, err %d", err );

	CloseHandle( hSemaphore );

	return bRes != 0;
}

cCriticalSection::cCriticalSection( const wchar_t *lpName, bool bSecure )
	: cSemaphore()
{
	ODSFuncS( lpName );

	if ( lpName == NULL )
	{
		ODS_f( L"Assertion:: lpName == NULL" );
		ODS_fls( L"Assertion!" );
		return;
	}

	while( true )
	{
		waitFor( lpName );

		if ( createSemaphore( 0, 1, lpName, bSecure ) )
		{
			//
			//	Paranoid check, but.
			//	We have to check this if semaphore has been created just after waitFor but before createSemaphore
			//
			if ( m_err == ERROR_ALREADY_EXISTS )
			{
				closeSemaphore();
				ODSF( L"%s: Semaphore '%s' exists", _CRT_WIDE( __FUNCTION__ ), lpName );
				continue;
			}

			break;
		}
		else
		{
			ODSF( L"%s: Error in creating semaphore '%s'", _CRT_WIDE( __FUNCTION__ ), lpName );
			ODS_fls( L"Assertion!" );
			return;
		}
	}
}
