#ifndef __cStreams__
#define __cStreams__

#include <windows.h>
#include "streams.h"

typedef NTSTATUS (__stdcall typeNtQueryInformationFile)( 
    IN HANDLE FileHandle,
    OUT PIO_STATUS_BLOCK IoStatusBlock,
    OUT PVOID FileInformation,
    IN ULONG Length,
    IN FILE_INFORMATION_CLASS FileInformationClass
    );

class cStreams
{
    PFILE_STREAM_INFORMATION  streamInfo, streamInfoPtr;

public:

	void QueryInfo( const wchar_t *fileName )
	{
		ULONG    streamInfoSize;
		NTSTATUS status;
		HANDLE   fileHandle;
		IO_STATUS_BLOCK ioStatus;

		streamInfoSize = 0;
		streamInfo = NULL;
		fileHandle = INVALID_HANDLE_VALUE;
		m_error = true;

		//
		// Load the NTDLL entry point we need
		//
		static typeNtQueryInformationFile *NtQueryInformationFile = (typeNtQueryInformationFile *)-1;

		if ( NtQueryInformationFile == (typeNtQueryInformationFile *)-1 )
			NtQueryInformationFile = (typeNtQueryInformationFile *) GetProcAddress( GetModuleHandleA("ntdll.dll"), "NtQueryInformationFile" );
		
		if (  NtQueryInformationFile == NULL )
			return;

		fileHandle = 
			CreateFileT( fileName, GENERIC_READ,
						FILE_SHARE_READ|FILE_SHARE_WRITE, NULL, 
						OPEN_EXISTING, 
						FILE_FLAG_BACKUP_SEMANTICS, 0 );

		if( fileHandle == INVALID_HANDLE_VALUE )
			return;

		memset( &ioStatus, 0, sizeof( ioStatus ) );
		streamInfoPtr = NULL;

		//
		//	Query stream info
		//
		streamInfoSize = 16384;

		if ( streamInfo != NULL )
			free( streamInfo );

		try
		{
			streamInfo = (FILE_STREAM_INFORMATION *)malloc( streamInfoSize );
			status = STATUS_BUFFER_OVERFLOW;

			while( status == STATUS_BUFFER_OVERFLOW )
			{
				if ( streamInfo == NULL )
				{
					ODS_f( L"Assertion! allocated size is NULL" );
					return;
				}

				status = NtQueryInformationFile( fileHandle, &ioStatus,
												 streamInfo, streamInfoSize,
												 FileStreamInformation );

				if( status == STATUS_BUFFER_OVERFLOW ) {

					free( streamInfo );
					streamInfoSize += 16384;
					streamInfo = (FILE_STREAM_INFORMATION *)malloc( streamInfoSize );

				}
				else
					break;
			}
		}
		catch( ... )
		{
			ODS_f( L"Assertion! Allocation exception." );
			status = ((NTSTATUS)-1);
			ioStatus.Information = 0;
		}

		if( !NT_SUCCESS( status ) )
		{
			ODS_f( L"Status is invalid" );
			m_error = true;
		}
		else
		// 10.10.2013 NtQueryInformationFile report ioStatus.Information == 0 for for directories ( if the do not have streams )
		// For files it is never 0 because at least one ::$DATA stream always present.
		if ( ioStatus.Information == 0 )
			m_error = true;
		else
		{
			streamInfoPtr = streamInfo;
			m_error = false;
		}

		CloseHandle( fileHandle );
	}

	bool m_error;

	cStreams( const wchar_t *fileName )
	{
		streamInfo = NULL;
		QueryInfo( fileName );
	}

	~cStreams()
	{
		if ( streamInfo != NULL )
			free( streamInfo );
	}

	bool getNextName( wchar_t *nameBuff, bool skipStandardData = true, FILE_STREAM_INFORMATION *pStreamInfo = NULL, bool *pStandardData = NULL )
	{
		//ODSFunc;

		if ( m_error )
			return false;

		int len = streamInfoPtr->NameLength/2;

		if ( len >= MAX_PATH )
		{
			ODS_f( L"Assertion" );
			return false;
		}

		const wchar_t *streamName = ( const wchar_t * )streamInfoPtr->Name;

		if ( nameBuff != NULL )
		{
			wcsncpy_s( nameBuff, MAX_PATH, streamName, len );
			nameBuff[ len ] = 0;
		}

		if ( pStreamInfo != NULL )
			*pStreamInfo = *streamInfoPtr;

		if( !streamInfoPtr->NextEntry )
			m_error = true; // Error for next call.
		else
			streamInfoPtr = (PFILE_STREAM_INFORMATION) ((char *) streamInfoPtr + streamInfoPtr->NextEntry );

		bool bStandardData = false;

		ODS_fls( cWCStr( streamName, len ).m_szBody );

		//
		// Skip the standard Data stream
		//
		if ( _wcsnicmp( streamName, L"::$DATA", len ) == 0 )
		{
			bStandardData = true;

			if ( skipStandardData )
				return getNextName( nameBuff );
		}

		if  ( pStandardData )
			*pStandardData = bStandardData;

		return true;
	}
};

#endif // __cStreams__
