#include <windows.h>
#include "..\CertificateLib\SMemory\SBlocks.h"
#include <string.h>
#include <assert.h>
#include "cStringArray.h"
#include "UnicodeFuncs.h"
#include "cStringBuffer.h"


int cStringArrayT::findString( const wchar_t *str ) const
{
	int i = Q;

	while (  i-- )
		if ( _wcsicmp( str, T[i] ) == 0 )
			break;

    return i;
}

//int cStringArrayT::findString( const wchar_t *str, int len ) const
//{
//	int i = Q;
//
//	while (  i-- )
//		if ( wcsncmp( str, T[i], len ) == 0 )
//			break;
//
//    return i;
//}

int cStringArrayT::findString( const wchar_t *str, int len, const wchar_t **pTail ) const
{
	int i = Q;

	if ( len == -1 )
		len = (int)wcslen( str );

	while (  i-- )
	{
		if ( wcsncmp( str, T[i], len ) == 0 )
			break;
	}

	if ( ( i >= 0 ) && ( pTail != NULL ) )
		*pTail = T[i] + len;

    return i;
}

int cStringArrayT::findStringIncluded( const wchar_t *str, int len )
{
	int i = Q;

	if ( len == -1 )
		len = (int)wcslen( str );

	while (  i-- )
	{
		int lenIncluded = (int)wcslen( T[i] );

		if ( lenIncluded == 0 )
			continue;

		if ( lenIncluded > len )
			continue;

		if ( wcsncmp( str, T[i], lenIncluded ) == 0 )
			break;
	}

    return i;
}

int cStringArrayT::addString( const wchar_t *str )
{
	int len = 0;
	if ( str != NULL )
		len = (int)wcslen( str );

	return addString( str, len );
}

static bool isDoubleZeroTerminated( const wchar_t *str )
{
	// it simple checks if the string has the { L'\', L'0' } symbol sequence

	if ( str[0] == 0 )
		return false;

	for ( int i = 1; str[i] != 0; i++ )
	{
		if ( ( str[i] == L'0' ) && ( str[i-1] == L'\\' ) )
			return true;
	}

	return false;
}

//static 
wchar_t *cStringArrayT::allocateStr( int len )
{
	// return new wchar_t[ len ];

	int allocatedSize = 0;

	return cWCStr::allocateStr( len, &allocatedSize );
}

//static 
void cStringArrayT::freeStr( wchar_t *wcs )
{
	// if ( wcs != NULL ) delete [] wcs;

	int allocatedSize = 0;
	if ( wcs != NULL )
		allocatedSize = (int)wcslen( wcs );

	return cWCStr::freeStr( wcs, &allocatedSize );
}

//static 
wchar_t *cStringArrayT::createCopy( const wchar_t *str, int len, bool checkDoubleZero )
{
	// checks if the string is double zero terminated
	// Reason: RCFile some times does not add the last L'\0' for double zero terminated strings but sometimes does.
	// so to avoid the ambiguity we will add the L'\0' for double zero terminated string if it was skipped.

	if ( len == -1 )
		len = (int)wcslen( str );

	bool addZero = checkDoubleZero && isDoubleZeroTerminated( str );

	if ( addZero )
	{
		if ( len < 3 )
			addZero = false;
		else
		if ( wcscmp( str + len - 1, L"\"" ) != 0 )
			addZero = false;
		else
		if ( wcscmp( str + len - 3, L"\\0\"" ) == 0 )
			addZero = false;
	}

	int newStrLen = len + 1 + ( addZero ? 2 : 0 );

	wchar_t *newStr = cStringArrayT::allocateStr( newStrLen );

	if ( newStr == NULL )
		return NULL;

	memcpy( newStr, str, len * sizeof( wchar_t ) );
	newStr[ len ] = L'\0';

	if ( addZero )
		copyString( newStr + len - 1, L"\\0\"", newStrLen - ( len - 1 ) );

	return newStr;
}

const wchar_t *cStringArrayT::getCurrent()
{
	if ( ( m_current >= Q ) || ( m_current < 0 ) )
	{
		//ODS_fls( _FormatT( L"%d/%d - %s", m_current, Q, NULL ) );
		return NULL;
	}

	//ODS_fls( _FormatT( L"%d/%d - %s", m_current, Q, T[ m_current ] ) );
	return T[ m_current];
}

void cStringArrayT::freeTable()
{
	while( Q-- )
	{
		//deleteString( T + Q );
		freeStr( T[Q] );
	}
	
	T = NULL;
	Q = 0;
	
	ptbTable.Free();
}

int cStringArrayT::addStringOnce( const wchar_t *str, int len )
{
	if ( len == -1 )
		len = (int)wcslen( str );

	if ( findString( str, len ) != -1 )
		return 0;

	return addString( str, len );
}

bool cStringArrayT::copyFrom( const cStringArrayT *sa, bool bAppend, bool checkDoubleZero )
{
	if ( !bAppend )
		freeTable();

	if ( getCount( sa ) == 0 )
		return true;

	return addTable( sa->T, sa->Q, checkDoubleZero ) > 0;
}

int cStringArrayT::addTable( const wcharPtr *pT, int pQ, bool checkDoubleZero )
{
	int i = 0;
	for ( ; i < pQ; i++ )
		addString( pT[i], -1, checkDoubleZero );

	return i;
}

int cStringArrayT::addString( const wchar_t *str, int len, bool checkDoubleZero )
{
	wchar_t *newString = createCopy( str, len, checkDoubleZero );

	if ( newString == NULL )
	{
		freeTable();
		return 0;
	}

	if ( addStringPtr( newString ) == 0 )
	{
		//delete newString;
		freeStr( newString );

		return 0;
	}

	return Q;
}

int cStringArrayT::addStringPtr( wchar_t *ptr )
{
	if ( ptbTable.WriteBytes( NULL, sizeof( wchar_t * ) ) == 0  )
	{
		freeTable();
		return 0;
	}

	T = ( wchar_t **)ptbTable.buffer;

	T[Q] = ptr;

	Q++;

	return Q;
}

bool cStringArrayT::updateString( int pos, const wchar_t *str, int len, bool checkDoubleZero )
{
	if ( pos >= Q )
		return false;

	if ( len == -1 )
	{
		if ( str == NULL )
			len = 0;
		else
			len = (int)wcslen( str );
	}

	wchar_t *newString = createCopy( str, len, checkDoubleZero );

	if ( newString == NULL )
		return false;

	//deleteString( T + pos );
	freeStr( T[pos] );

	T[pos] = newString;

	return true;
}

bool cStringArrayT::removeString( int num )
{
	if ( num >= Q )
		return false;

	// deleteString( T + num );
	freeStr( T[num] );

	Q--;

	for( int i = num; i < Q; i++ )
	{
		T[i] = T[i+1];
	}

	return true;
}

//int cStringArrayT::createTable( BYTE *buff, int buffSize, const wchar_t *prefixStr, bool bForceUnicode )
//{
//	freeTable();
//
//	if ( ( buff == NULL ) || ( buffSize == 0 ) )
//		return 0;
//
//	int pos = 0;
//
//	wchar_t path[ 4096 ];
//
//	if ( prefixStr != NULL )
//	{
//		pos = (int)wcslen( prefixStr );
//		if ( pos >= maxLen( path ) )
//			pos = maxLen( path ) - 1;
//
//		wcsncpy_s( path, prefixStr, pos );
//	}
//
//	path[ pos ] = L'\0';
//
//	int lPos = pos;
//
//	cStringBuffer sb( buff, buffSize );
//
//	if ( bForceUnicode )
//		sb.m_isUnicode = true;
//
//	while( sb.ReadStringT( path + pos, maxLen( path ) - pos ) )
//	{
//		int len = (int)wcslen( path );
//
//		if ( ( len != 0 ) && m_treatBackSlashAsStringConcatenation )
//		{
//			while( path[ len - 1 ] == L'\\' )
//				sb.ReadStringT( path + len, maxLen( path ) - len );
//		}
//
//		if ( addString( path ) == 0 )
//			return 0;
//	}
//
//	return Q;
//}

int cStringArrayT::createTable( const BYTE *buff, int buffSize, const wchar_t *prefixStr, bool bForceUnicode )
{
	freeTable();

	if ( ( buff == NULL ) || ( buffSize == 0 ) )
		return 0;

	cStringBuffer sb( buff, buffSize, m_bTabulationAsNewLine );

	if ( bForceUnicode )
		sb.m_isUnicode = true;

	cWCStr wcs;
	int len = 0;

	while( sb.ReadStringT( &wcs ) )
	{
		while ( m_treatBackSlashAsStringConcatenation &&
			   ( ( len = wcs.GetLength() ) != 0 ) &&
			   ( wcs.getAt( len - 1 ) == L'\\' )
			)
		{
			wcs.SetLength( len - 1 );

			cWCStr wcsNew;
			if ( !sb.ReadStringT( &wcsNew ) )
				break;

			wcs.Add( wcsNew.m_szBody );
		}

		//
		//	Fill string with prefix and swap it with wcs
		//
		if ( prefixStr != NULL )
		{
			cWCStr wcsWithPrefix( prefixStr );
			wcsWithPrefix.Add( wcs.m_szBody );
			wcs.Swap( &wcsWithPrefix );
		}

		//
		//	cStringArrayT uses same allocation method as cWCStr does, so the allocated pointers may be bravelly swapped.
		//
		if ( addStringPtr( wcs.m_szBody ) == 0 )
			return 0;
		else
		{
			wcs.m_szBody = NULL;
			wcs.m_len = 0;
			wcs.m_allocatedLen = 0;
		}		
	}

	return Q;
}

//
//	compare
//		returns  1 if this array has at least one different string
//		returns -1 if 'sa' array has at least one different string
//		returns 0 if there are no different string.//		
//
int cStringArrayT::compare( cStringArrayT *sa )
{
	for( int i = sa->Q; i--; )
	{
		if ( findString( sa->T[i] ) == -1 )
			return 1;
	}

	for( int i = Q; i--; )
	{
		if ( sa->findString( T[i] ) == -1 )
			return -1;
	}

	return 0;
}

//static
void cStringArrayT::ODSTable( const cStringArrayT *sa )
{
	ODS_fls( _FormatT( L"Path table Q = %d, current %d", sa->Q, sa->m_current ) );

	for( int i = 0; i< sa->Q; i++ )
		ODS_fls( _FormatT( L"%d) '%s'", i, sa->T[i] ) );
}
