#include "cDll.h"
#include "SignatureCheck.h"
#include <stdlib.h>
#include "ODS.h"

#pragma once

class cTrustDll : public CDll
{
public:

	enum { errInVerification = CDll::errGetProcAddress + 1 };

	//
	//	 WinVerifyTrust requires full path only.
	//
	inline bool VerifyAndLoad( const wchar_t *path, const char *procName = NULL, bool showMessage = false )
	{
		bool ok = CDll::Load( true, path, procName, showMessage );

		if ( !ok )
			return false;

		CSignatureCheck chk;

		char pathA[ MAX_PATH ];
		size_t size;
		wcstombs_s( &size, pathA, path, _TRUNCATE );

	#ifdef UNICODE
		const TCHAR *pathT = path;
	#else
		const char *pathT = pathA;
	#endif

		ok = chk.VerifyEmbeddedSignature( pathT ) == TRUE;

		bool bValidPublicKey = chk.isValidKeyID();

		if ( ok && !bValidPublicKey )
		{
			if ( chk.getRegDelay() )
				ok = false;
		}

		if ( !ok )
		{
			bool bUnload = true;

			if ( showMessage )
			{
				char msg[1024];
				sprintf_s( msg,
					"File signature is invalid!\nPath: '%s'\n"
					"WinTrust API status: 0x%X - %s\n\n"
					"Continue untrusted?",
					pathA, chk.m_status, chk.m_statusStr );

				int id = MessageBoxA( 0, msg, "Error in verification!", MB_ICONSTOP | MB_YESNO | MB_DEFBUTTON2 );

				if ( id == IDYES )
					bUnload = false;
			}

	#ifdef _DEBUG
//			bUnload = false;
	#endif
			if ( bUnload )
			{
				CDll::Unload();
				error = errInVerification;
				ODS_fls( L"Warning! DLL signature is invalid" );
				ok = false;
			}
		}

		return ok;
	}
};
