#include <windows.h>

// Sample:
//
//<!DOCTYPE HTML PUBLIC "-//IETF//DTD HTML 2.0//EN">
//<html><head>
//<title>401 Authorization Required</title>
//</head><body>
//<h1>Authorization Required</h1>
//
class cXMLBlock
{
public:
	
	const BYTE *headT;	// header 
	int headL;			// header lenght
	int entireL;		// entire block lenght

	const BYTE *bodyT;	// body
	int bodyL;			// body size

	const BYTE *nextT;	// next block
	int nextL;			// next block size

	cXMLBlock()
	{
		memset( this, 0, sizeof( *this ) );
	}

	const BYTE *find( 
			const BYTE *bf, int bfSize, 
			const BYTE *findBlock, int findBlockSize,
			bool bBegin // determinates if begin or end of the find block position is to be returned. 
		)
	{
		for ( int i = 0; i <= bfSize - findBlockSize; i++ )
		{
			if ( memcmp( bf + i, findBlock, findBlockSize ) == 0 )
				return bf + i + ( bBegin ? 0 : findBlockSize );
		}

		return NULL;
	}

	bool parse( const BYTE *bf, int bfSize )
	{
		while( bfSize )
		{
			switch ( *bf )
			{
				case '\r':
				case '\n':
				case '\t':
				case ' ':
				{
					bfSize--;
					bf++;
					continue;
				}
				default:
					break;
			}

			break;
		}

		const BYTE *p;

		if ( bfSize < 3 )
			return false;

		headT = bf;

		if ( headT [0] != '<' )
			return false;

		BYTE *tT;	// tail 
		int tL;		//	tail lenght

		if ( headT[1] == '!' )
		{
			// <!DOCTYPE HTML PUBLIC "-//IETF//DTD HTML 2.0//EN">
			headL = 2;
			tL = 1;
			tT = new BYTE[ tL ];
			if ( tT == NULL )
				return false;

			tT[0] = '>';
		}
		else
		{
			//<title>401 Authorization Required</title>
			p = find( headT + 1, bfSize -  1, (const BYTE *)">", 1, false );

			if ( p == NULL )
				return false;

			headL = (int)(p - headT);

			tL = headL + 1;

			if ( headL <= 2 )
				return false;

			tT = new BYTE[ tL ];
			if ( tT == NULL )
				return false;

			tT[0] = '<';
			tT[1] = '/';
			memcpy( tT + 2, headT + 1, headL - 1 );
		}

		bodyT = headT + headL;

		p = find( bodyT, bfSize - headL, tT, tL, true );

		delete tT;

		if ( p == NULL )
			return false;

		bodyL = (int)(p - bodyT);

		nextT = p + tL;
		entireL = (int)(nextT - headT );

		nextL = bfSize - entireL;

		return true;
	}
};