#include "configFile.h"
#include <stdio.h>
#include "UnicodeFuncs.h"

cConfigFile::cConfigFile( const wchar_t *filePath, HWND hReportWnd )
	: cStringFile( filePath )
{
	if ( hReportWnd && m_err )
	{
		wchar_t msg[ 512 ];
		FormatT( msg, maxLen( msg ),
			L"Error in opening file\n"
			L"'%s'",
			filePath );

		MessageBoxT( hReportWnd, msg, L"BCUpdate", MB_ICONSTOP );
	}
}

static bool isEmpty( wchar_t ch )
{
	if ( ch == L' ' ) return true;
	if ( ch == L'\t' ) return true;
	return false;
}

bool cConfigFile::FindString( const wchar_t *group, const wchar_t *str, wchar_t *valueBuff, int buffSize )
{
	Init();
	wchar_t buff[ 1024 ];

	// find group
	while ( true )
	{
		if ( !ReadStringT( buff, maxLen( buff ) ) )
			return false;

		if ( _wcsicmp( buff, group ) == 0 )
			break;
	}

	if ( str == NULL )
		return true;

	int strLen = (int)wcslen( str );

	while ( true )
	{
		if ( !ReadStringT( buff, maxLen( buff ) ) )
			return false;

		int len = (int)wcslen( buff );

		// Next group found
		if ( ( buff[0] == L'[' ) && ( buff[len-1] == L']' ) )
			return false;

		if ( _wcsnicmp( buff, str, strLen ) == 0 )
		{
			int pos = strLen;

			while( isEmpty( buff[pos] ) ) pos++;
			if ( buff[pos++] != L'=' ) continue;
			while( isEmpty( buff[pos] ) ) pos++;

			wcsncpy( valueBuff, buff + pos, buffSize );

			return true;
		}
	}
}

bool cConfigFile::ReadAndParseString( wchar_t *name, int nameSize, wchar_t *value, int valueSize )
{
	if ( nameSize == 0 )
		return false;

	if ( valueSize == 0 )
		return false;

	name[0] = value[0] = L'\0';

	wchar_t buff[ MAX_PATH ];
	memset( buff, 0, sizeof( buff ));
	
	if ( !cStringFile::ReadStringT( buff, maxLen( buff ) ) )
		return false;

	int i;

	for( i = 0; i< maxLen(buff); i++ )
	{
		if ( buff[i] == L'\0' ) // No value is found
		{
			wcsncpy( name, buff, nameSize );
			name[nameSize - 1] = L'\0';
			break;
		}

		if ( buff[i] == L'=' ) // Value separator is found
		{
			wcsncpy( name, buff, nameSize );
			name[ nameSize - 1 ] = L'\0';
			if ( nameSize > i )
				name[ i ] = L'\0';

			for( ; ; )	// Skips probels and fills value
			{
				if ( ++i >= maxLen( buff ) )
					return false;

				if ( buff[i] != L' ' )
				{
					valueSize = min( (int)maxLen( buff ) - i, valueSize );

					wcsncpy( value, buff + i, valueSize );
					value[ valueSize - 1 ] = L'\0';
					break;
				}
			}

			break;
		}
	}

	// Cuts probels at tails
	for( i = (int)wcslen( name  ); --i; ) if ( name [i] == L' ' ) name [i] = L'\0'; else break;
	for( i = (int)wcslen( value ); --i; ) if ( value[i] == L' ' ) value[i] = L'\0'; else break;

	return true;
}