#include "getFileVersion.h"
#include <stdio.h>
#include "FormatT.h"
#include "cWCS.h"
#include "AutoDeallocate.h"
#include "cMBStr.h"
#include <io.h>

static BOOL fileExists( const wchar_t *lpPathName )
{
	if ( lpPathName == NULL )
		return false;

	return _waccess( lpPathName, 0 ) == 0;
}

#define VerQueryValueT VerQueryValueW
//BOOL
//VerQueryValueT(
//        const LPVOID pBlock,
//        LPWSTR lpSubBlock,
//        LPVOID * lplpBuffer,
//        PUINT puLen,
//		bool isStringValue = false
//        )
//{
//	bool isNT = ( GetVersion() < 0x80000000 );
//
//	if ( isNT )
//	{
//		return VerQueryValueW( pBlock, lpSubBlock, lplpBuffer, puLen );
//	}
//	else
//	{
//		cMBStr mbs( lpSubBlock );
//
//		BOOL ok = VerQueryValueA( pBlock, mbs.m_szBody, lplpBuffer, puLen );
//
//		//
//		//	If value is not string value so conversion is not required.
//		//
//		if ( !isStringValue )
//			return ok;
//
//		if ( ok )
//		{
//			static cAutoDeallocate <wchar_t> ad;
//			
//			ad.AllocateAndCopy( getWCStr( (char*)*lplpBuffer ), *puLen );
//
//			if ( ad.m_ptr == NULL )
//				return false;
//			
//			*lplpBuffer = (void *)ad.m_ptr;
//		}
//		
//		return ok;
//	}
//}

DWORD
	GetFileVersionInfoSizeT(
        LPCWSTR lptstrFilename,
        LPDWORD lpdwHandle
        )
{
	if ( !fileExists( lptstrFilename ) )
		return FALSE;

	return GetFileVersionInfoSizeW( (LPWSTR)lptstrFilename, lpdwHandle );

	//bool isNT = ( GetVersion() < 0x80000000 );

	//if ( isNT )
	//	return GetFileVersionInfoSizeW( (LPWSTR)lptstrFilename, lpdwHandle );
	//else
	//{
	//	char lptstrFilenameA[ MAX_PATH ];
	//	_wcstombs( lptstrFilenameA, sizeof( lptstrFilenameA ), lptstrFilename );
	//	return GetFileVersionInfoSizeA( lptstrFilenameA, lpdwHandle );
	//}
}

BOOL
	GetFileVersionInfoT(
        LPCWSTR lptstrFilename,
        DWORD dwHandle,
        DWORD dwLen,
        LPVOID lpData
        )
{
	if ( !fileExists( lptstrFilename ) )
		return FALSE;
	
	return GetFileVersionInfoW( (LPWSTR) lptstrFilename, dwHandle, dwLen, lpData );

	//bool isNT = ( GetVersion() < 0x80000000 );

	//if ( isNT )
	//	return GetFileVersionInfoW( (LPWSTR) lptstrFilename, dwHandle, dwLen, lpData );
	//else
	//{
	//	char lptstrFilenameA[ MAX_PATH ];
	//	_wcstombs( lptstrFilenameA, sizeof( lptstrFilenameA ), lptstrFilename );
	//	return GetFileVersionInfoA( lptstrFilenameA, dwHandle, dwLen, lpData );
	//}
}

//static wchar_t *makeSBName( WORD w0, WORD w1 )
//{
//	static wchar_t name[ 20 ];
//
//	return FormatT( name, maxLen( name ), L"%04x%04x", w0, w1 );
//}
//
static BOOL 
	queryVersionValueT(
			BYTE *versionInfo,
			void **blockValue,
			unsigned int *dwSymbols,
			DWORD *cbSize,
			const wchar_t *szSubBlockValue,
			WORD lang = 0
		)
{
	*dwSymbols = 0;
	*blockValue = 0;

	try
	{	
		struct LANGANDCODEPAGE {
			WORD wLanguage;
			WORD wCodePage;
		} *lpTranslate = NULL;

		UINT cbTranslate = 0;

		// Read the list of languages and code pages.

		VerQueryValueT( versionInfo, 
					L"\\VarFileInfo\\Translation",
					(LPVOID*)&lpTranslate,
					&cbTranslate);

		if ( lpTranslate == NULL )
			return false;

		// Read the file description for each language and code page.

		wchar_t SubBlock[ 256 ];

		for( unsigned int i=0; i < (cbTranslate/sizeof(struct LANGANDCODEPAGE)); i++ )
		{
			if ( ( lang != 0 ) && ( lang != lpTranslate[i].wLanguage ) )
				continue;

			FormatT( SubBlock, maxLen( SubBlock ),
				L"\\StringFileInfo\\%04x%04x\\%s",
				lpTranslate[i].wLanguage,
				lpTranslate[i].wCodePage,
				szSubBlockValue
				);

			// Retrieve file description for language and code page "i". 
			*dwSymbols = 0;
			VerQueryValueT(  versionInfo, 
							SubBlock, 
							blockValue, 
							dwSymbols );
							//, true ); // The string value will be converted for window95

			if ( *dwSymbols != 0 )
				break;
		}
	}
	catch(...)
	{
		*blockValue = 0;
	}

	return (*cbSize >= *dwSymbols) && *blockValue && *dwSymbols;
}

bool getFileVersion ( const char *name, char *blockValue, DWORD *cbSize, const char *szSubBlockValue, WORD  wLang )
{ return GetFileVersionBlock( name, blockValue, cbSize, szSubBlockValue, wLang ) == TRUE; }

bool getFileVersionT( const wchar_t *name, wchar_t *blockValue, DWORD *cbSize, const wchar_t *szSubBlockValue, WORD  wLang )
{ return GetFileVersionBlockT( name, blockValue, cbSize, szSubBlockValue, wLang ) == TRUE; }

BOOL GetFileVersionBlockT(
		const wchar_t *name,
		wchar_t *version,
		DWORD *cbSize,
		const wchar_t *szSubBlockValue, //  = L"FileVersion"		// L"FileDescription"
		WORD lang
		)
{
	if ( !fileExists( name ) )
		return FALSE;

	DWORD zero;
	DWORD size = GetFileVersionInfoSizeT( (wchar_t *)name, &zero );

	if ( size == 0 )
		return false;

	BYTE *versionInfo = new BYTE[size];

	GetFileVersionInfoT((wchar_t *)name,zero,size,versionInfo);

	if ( version != NULL )
		*version = 0;

	void *fileVersion = 0;
	unsigned int dwSymbols = 0;

	BOOL ok = queryVersionValueT( versionInfo, &fileVersion, &dwSymbols, cbSize, szSubBlockValue, lang );

	if ( ok && ( version != NULL ) )
		memcpy( version, fileVersion, dwSymbols * sizeof( wchar_t ) );
	
	delete versionInfo;

	*cbSize = dwSymbols;

	return ok;
}

static BOOL 
	queryVersionValue(
			BYTE *versionInfo,
			void **blockValue,
			unsigned int *dwSymbols,
			DWORD *cbSize,
			const char *szSubBlockValue, //  = "FileVersion"		// "FileDescription"
			WORD  lang // = 0
		)
{
	*dwSymbols = 0;
	*blockValue = 0;

	try
	{
		struct LANGANDCODEPAGE {
			WORD wLanguage;
			WORD wCodePage;
		} *lpTranslate = NULL;

		UINT cbTranslate = 0;

		// Read the list of languages and code pages.

		VerQueryValueA( versionInfo, 
					"\\VarFileInfo\\Translation",
					(LPVOID*)&lpTranslate,
					&cbTranslate);

		if ( lpTranslate == NULL )
			return false;

		// Read the file description for each language and code page.

		char SubBlock[ 256 ];

		for( unsigned int i=0; i < (cbTranslate/sizeof(struct LANGANDCODEPAGE)); i++ )
		{
			sprintf_s( SubBlock, 
				"\\StringFileInfo\\%04x%04x\\%s",
				lpTranslate[i].wLanguage,
				lpTranslate[i].wCodePage,
				szSubBlockValue
				);

			// Retrieve file description for language and code page "i". 
			*dwSymbols = 0;
			VerQueryValueA(  versionInfo, 
							SubBlock, 
							blockValue, 
							dwSymbols); 

			if ( *dwSymbols != 0 )
				break;
		}
	}
	catch(...)
	{
		*blockValue = 0;
	}

	return (*cbSize >= *dwSymbols) && *blockValue && *dwSymbols;
}

BOOL GetFileVersionBlock(
		const char *name,
		char  *blockValue,
		DWORD *cbSize,
		const char *szSubBlockValue,//  = "FileVersion",		// "FileDescription" // "LegalCopyright"
		WORD  wLang		// = Lang_UDF
	)
{
	if ( _access( name, 0 ) != 0 )
		return FALSE;

	DWORD zero;
	DWORD size = GetFileVersionInfoSizeA( (char *)name, &zero );

	if ( size == 0 )
		return false;

	BYTE *versionInfo = new BYTE[size];

	if ( versionInfo == NULL )
	{
		ODS_fls( L"Assertion!" );
		return false;
	}

	GetFileVersionInfoA((char *)name,zero,size,versionInfo);

	*blockValue = 0;
	void *fileBlockValue = 0;
	unsigned int dwSymbols = 0;

	BOOL ok = queryVersionValue( versionInfo, &fileBlockValue, &dwSymbols, cbSize, szSubBlockValue, wLang );

	if ( ok )
		memcpy( blockValue, fileBlockValue, dwSymbols * sizeof( char ) );
	
	delete versionInfo;

	*cbSize = dwSymbols;

	return ok;
}

static int fillVersionTable( int T[4], const char *s )
{
	int q = sscanf_s( s, "%d.%d.%d.%d", &(T[0]), &(T[1]), &(T[2]), &(T[3]) );

	if ( q == 1 )
		q = sscanf_s( s, "%d,%d,%d,%d", &(T[0]), &(T[1]), &(T[2]), &(T[3]) );

	return q;
}

int CompareVersions( const char *s0, const char *s1, int *pFieldsQ )
{
	int v0[4] = { 0 }; int q0 = fillVersionTable( v0, s0 );
	int v1[4] = { 0 }; int q1 = fillVersionTable( v1, s1 );

	//int q0 = sscanf_s( s0, "%d.%d.%d.%d", &(v0[0]), &(v0[1]), &(v0[2]), &(v0[3]) );

	//int v1[] = { 0, 0, 0, 0 };
	//int q1 = sscanf_s( s1, "%d.%d.%d.%d", &(v1[0]), &(v1[1]), &(v1[2]), &(v1[3]) );

	if ( pFieldsQ != NULL )
		*pFieldsQ = min( q0, q1 );

	for ( int i = 0; i < maxLen( v0 ); i++ )
	{
		int res = v0[i] - v1[i];

		if ( res != 0 )
			return res;
	}

	return 0;
}

int CompareVersionsT( const wchar_t *srcVer, const wchar_t *dstVer, int *pFieldsQ )
{
	char srcVerA[ 128 ];
	char dstVerA[ 128 ];

	size_t num;
	wcstombs_s( &num, srcVerA, srcVer, _TRUNCATE );
	wcstombs_s( &num, dstVerA, dstVer, _TRUNCATE );

	return CompareVersions( srcVerA, dstVerA, pFieldsQ );
}

int CheckFileVersion( const wchar_t *path, const wchar_t *szVersion, int *pFieldsQ )
{
	wchar_t curVersion[ 200 ] = L"";
	if ( curVersion[0] == 0 )
	{
		DWORD size = maxLen( curVersion );

		if ( !getFileVersionT( path, curVersion, &size ) )
		{
			return -1;
		}
	}

	return CompareVersionsT( curVersion, szVersion, pFieldsQ );
}

bool getModuleVersion( HINSTANCE hInst, wchar_t *version, DWORD *cbSize, const wchar_t *szSubBlockValue )
{
	cWCS wcs( MAX_PATH );
	if ( wcs.m_szBody == NULL )
		return false;

	if ( GetModuleFileNameW( hInst, wcs.m_szBody, wcs.m_len ) == 0 )
		return false;

	return GetFileVersionBlockT( wcs.m_szBody, version, cbSize, szSubBlockValue ) == TRUE;
}

const wchar_t *getModuleVersion( cWCS &wcs, HINSTANCE hInst, const wchar_t *szSubBlockValue )
{
	cWCS wcsPath( MAX_PATH );
	if ( wcsPath.m_szBody == NULL )
		return NULL;

	if ( GetModuleFileNameW( hInst, wcsPath.m_szBody, wcsPath.m_len ) == 0 )
		return NULL;

	return getFileVersion( wcs, wcsPath.m_szBody, szSubBlockValue );
}

const wchar_t *getFileVersion( cWCS &wcs, const wchar_t *path, const wchar_t *szSubBlockValue )
{
	DWORD cbSize = 0;
	getFileVersionT( path, NULL, &cbSize, szSubBlockValue );

	if ( cbSize == 0 )
		return NULL;

	wcs.SetLength( cbSize );
	cbSize = wcs.m_len;

	if ( !getFileVersionT( path, wcs.m_szBody, &cbSize, szSubBlockValue ) )
		return NULL;

	wcs.Truncate();

	return wcs.m_szBody;
}

int CompareFileVersionsT( const wchar_t *path0, const wchar_t *path1, int *pFieldsQ )
{
	wchar_t version0[ 100 ];
	DWORD size = maxLen( version0 );
	if ( !getFileVersionT( path0, version0, &size ) )
		return -100;

	wchar_t version1[ 100 ];
	size = maxLen( version1 );
	if ( !getFileVersionT( path1, version1, &size ) )
		return 100;
	
	return CompareVersionsT( version0, version1, pFieldsQ );
}

const wchar_t *GetLegalCopyright( HMODULE hInst )
{
	static cWCS wcsVer;

	cWCS wcs( MAX_PATH );

	GetModuleFileNameW( hInst, wcs.m_szBody, wcs.m_len );
	DWORD len = 0;
	GetFileVersionBlockT( wcs.m_szBody, NULL, &len, L"LegalCopyright" );

	wcsVer.SetLength( len );
	GetFileVersionBlockT( wcs.m_szBody, wcsVer.m_szBody, &len, L"LegalCopyright" );

	return wcsVer.m_szBody;
}

