#include "UnicodeFuncs.h"
#include "CDll.h"
#include "grantPrivileges.h"
#include "advApi.h"
#include "cFindFile.h"


//bool grantPrivileges( const wchar_t *path, const wchar_t *privileges = L"D,F,WDAC", const wchar_t *szUserName = NULL );
// "AD" append to subdirectory

//
//	The user SID is prefered as user name because the function does not properly works if name contains probels
//	Note, the '*' prefix has to be added to sid it it is used instead user name.
//

bool grantPrivileges( const wchar_t *path, const wchar_t *privileges, const wchar_t *szUserName )
{
	ODSFuncS( cWCStr().FormatT( L"%1,%2,%3", path, privileges, szUserName ) );

	cWCStr wcsName;

	if ( szUserName == 0 )
	{
		cWCStr wcsSid;
		if ( !GetCurrentThreadSid( &wcsSid ) )
		{
			ODS_f( L"Assertion" );
			return false;
		}
		
		wcsName.Set( L"*" );
		wcsName.Add( wcsSid.m_szBody );

		if ( szUserName == NULL )
			szUserName = wcsName.m_szBody;
	}

	const wchar_t *iaclsCmdPattern = L"icacls \"%s\" /%s %s:(%s) /C /T";

	cWCStr cmd;
	cmd.FormatT( iaclsCmdPattern, path, L"grant", szUserName, privileges );

	int res;
	HANDLE hProcess = RunProcessWithWaitingT( cmd, true, true, NULL, &res );

	ODSF( L"CMD: %s", cmd );
	ODSFuncRes( res );

	return res == 0;
}


//
//	MSDN sample 'Taking Object Ownership in C++'
//
//	The following example tries to change the DACL of a file object by taking ownership of that object. 
//	This will succeed only if the caller has WRITE_DAC access to the object or is the owner of the object. 
//	If the initial attempt to change the DACL fails, an administrator can take ownership of the object. 
//	To give the administrator ownership, the example enables the SE_TAKE_OWNERSHIP_NAME privilege in the 
//	caller's access token, and makes the local system's Administrators group the owner of the object. 
//	If the caller is a member of the Administrators group, the code will then be able to change the object's DACL.
//
bool TakeOwnership( const wchar_t *lpszOwnFile, const wchar_t *stringSid ) // = NULL ) 
{
	ODSFuncS( cWCStr().FormatT( L"%s, %s", lpszOwnFile, stringSid ) );

    bool bRetval = false;

    HANDLE hToken = NULL; 
    PSID pSIDAdmin = NULL;
    PSID pSIDEveryone = NULL;
	PSID pSIDUser = NULL;
    PACL pACL = NULL;
    SID_IDENTIFIER_AUTHORITY SIDAuthWorld = SECURITY_WORLD_SID_AUTHORITY;
    SID_IDENTIFIER_AUTHORITY SIDAuthNT = SECURITY_NT_AUTHORITY;
    const int NUM_ACES = 3;
    EXPLICIT_ACCESS ea[NUM_ACES];
    DWORD dwRes;

    // Specify the DACL to use.
    // Create a SID for the Everyone group.
    if (!CAdvAPIDll::getStatic().AllocateAndInitializeSid(&SIDAuthWorld, 1,
                     SECURITY_WORLD_RID,
                     0,
                     0, 0, 0, 0, 0, 0,
                     &pSIDEveryone)) 
    {
        ODSF( L"AllocateAndInitializeSid (Everyone) error %u", GetLastError());
        goto Cleanup;
    }

    // Create a SID for the BUILTIN\Administrators group.
    if (!CAdvAPIDll::getStatic().AllocateAndInitializeSid(&SIDAuthNT, 2,
                     SECURITY_BUILTIN_DOMAIN_RID,
                     DOMAIN_ALIAS_RID_ADMINS,
                     0, 0, 0, 0, 0, 0,
                     &pSIDAdmin)) 
    {
        ODSF( L"AllocateAndInitializeSid (Admin) error %u",
                GetLastError());
        goto Cleanup;
    }

    ZeroMemory(&ea, NUM_ACES * sizeof(EXPLICIT_ACCESS));

	int usedNum = 0;
	
	DWORD grfInheritance = SUB_CONTAINERS_AND_OBJECTS_INHERIT; // NO_INHERITANCE;

    // Set read access for Everyone.
    ea[usedNum].grfAccessPermissions = GENERIC_READ;
    ea[usedNum].grfAccessMode = SET_ACCESS;
    ea[usedNum].grfInheritance = grfInheritance;
    ea[usedNum].Trustee.TrusteeForm = TRUSTEE_IS_SID;
    ea[usedNum].Trustee.TrusteeType = TRUSTEE_IS_WELL_KNOWN_GROUP;
    ea[usedNum].Trustee.ptstrName = (LPTSTR) pSIDEveryone;
	
	usedNum++;

    // Set full control for Administrators.
    ea[usedNum].grfAccessPermissions = GENERIC_ALL;
    ea[usedNum].grfAccessMode = SET_ACCESS;
    ea[usedNum].grfInheritance = grfInheritance;
    ea[usedNum].Trustee.TrusteeForm = TRUSTEE_IS_SID;
    ea[usedNum].Trustee.TrusteeType = TRUSTEE_IS_GROUP;
    ea[usedNum].Trustee.ptstrName = (LPTSTR) pSIDAdmin;

	usedNum++;

	if ( stringSid != NULL )
	{
	    // Set full control for User identified by SID.
		CAdvAPIDll::getStatic().ConvertStringSidToSid( stringSid, &pSIDUser ); 
		ea[usedNum].grfAccessPermissions = GENERIC_ALL;
		ea[usedNum].grfAccessMode = SET_ACCESS;
		ea[usedNum].grfInheritance = grfInheritance;
		ea[usedNum].Trustee.TrusteeForm = TRUSTEE_IS_SID;
		ea[usedNum].Trustee.TrusteeType = TRUSTEE_IS_GROUP;
		ea[usedNum].Trustee.ptstrName = (LPTSTR) pSIDUser;

		usedNum++;
	}
	else
	{
	    ea[0].grfAccessPermissions = GENERIC_ALL;
	}

    if ( ERROR_SUCCESS != CAdvAPIDll::getStatic().SetEntriesInAcl( usedNum, ea, NULL, &pACL ) )
    {
        ODSF( L"Failed SetEntriesInAcl");
        goto Cleanup;
    }

    // Try to modify the object's DACL.
    dwRes = CAdvAPIDll::getStatic().SetNamedSecurityInfo(
        lpszOwnFile,                 // name of the object
        SE_FILE_OBJECT,              // type of object
        DACL_SECURITY_INFORMATION,   // change only the object's DACL
        NULL, NULL,                  // do not change owner or group
        pACL,                        // DACL specified
        NULL);                       // do not change SACL

    if (ERROR_SUCCESS == dwRes) 
    {
        ODSF( L"Successfully changed DACL%s", L"" );
        bRetval = true;
        // No more processing needed.
        goto Cleanup;
    }

    if (dwRes != ERROR_ACCESS_DENIED)
    {
        ODSF( L"First SetNamedSecurityInfo call failed: %u", dwRes ); 
        goto Cleanup;
    }

    // If the preceding call failed because access was denied, 
    // enable the SE_TAKE_OWNERSHIP_NAME privilege, create a SID for 
    // the Administrators group, take ownership of the object, and 
    // disable the privilege. Then try again to set the object's DACL.

    // Open a handle to the access token for the calling process.
    if (!CAdvAPIDll::getStatic().OpenProcessToken(GetCurrentProcess(), 
                          TOKEN_ADJUST_PRIVILEGES, 
                          &hToken)) 
       {
          ODSF( L"OpenProcessToken failed: %u", GetLastError()); 
          goto Cleanup; 
       } 

    // Enable the SE_TAKE_OWNERSHIP_NAME privilege.
    if (!CAdvAPIDll::getStatic().SetPrivilege( hToken, SE_TAKE_OWNERSHIP_NAME, true ) ) 
    {
        ODSF( L"You must be logged on as Administrator.");
        goto Cleanup; 
    }

    // Set the owner in the object's security descriptor.
    dwRes = CAdvAPIDll::getStatic().SetNamedSecurityInfo(
        lpszOwnFile,                 // name of the object
        SE_FILE_OBJECT,              // type of object
        OWNER_SECURITY_INFORMATION,  // change only the object's owner
        pSIDAdmin,                   // SID of Administrator group
        NULL,
        NULL,
        NULL); 

    if (dwRes != ERROR_SUCCESS) 
    {
        ODSF( L"Could not set owner. Error: %u", dwRes); 
        goto Cleanup;
    }
        
    // Disable the SE_TAKE_OWNERSHIP_NAME privilege.
    if (!CAdvAPIDll::getStatic().SetPrivilege( hToken, SE_TAKE_OWNERSHIP_NAME, false ) ) 
    {
        ODSF( L"Failed SetPrivilege call unexpectedly.");
        goto Cleanup;
    }

    // Try again to modify the object's DACL,
    // now that we are the owner.
    dwRes = CAdvAPIDll::getStatic().SetNamedSecurityInfo(
        lpszOwnFile,                 // name of the object
        SE_FILE_OBJECT,              // type of object
        DACL_SECURITY_INFORMATION,   // change only the object's DACL
        NULL, NULL,                  // do not change owner or group
        pACL,                        // DACL specified
        NULL);                       // do not change SACL

    if (dwRes == ERROR_SUCCESS)
    {
        ODSF( L"Successfully changed DACL%s", L", Try again" );
        bRetval = true; 
    }
    else
    {
        ODSF( L"Second SetNamedSecurityInfo call failed: %u", dwRes ); 
    }

Cleanup:

    if ( NULL != pSIDAdmin )
        CAdvAPIDll::getStatic().FreeSid(pSIDAdmin); 

    if ( NULL != pSIDEveryone )
        CAdvAPIDll::getStatic().FreeSid(pSIDEveryone); 

    if ( NULL != pSIDUser )
        CAdvAPIDll::getStatic().FreeSid(pSIDUser); 

    if ( NULL != pACL )
       LocalFree(pACL);

    if ( NULL != hToken )
       CloseHandle(hToken);

    return bRetval;

}

bool TakeOwnershipRec( const wchar_t *path, const wchar_t *stringSid, WIN32_FIND_DATAW *pFd ) // pFd = NULL
{
	ODSFuncS( path );

	WIN32_FIND_DATAW fd;

	if ( pFd == NULL )
	{
		pFd = &fd;

		if ( !FileExistsT( path, pFd ) )
			return false;
	}

	bool ok = TakeOwnership( path, stringSid );

	if ( !ok )
	{
		ODSF( L"Error: in granting privileges\tPath: %s\tSID: %s", path, stringSid );
		ok = TakeOwnership( path, NULL );
	}

	if ( ( pFd->dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY ) == 0 )
		return ok;

	cFindFile cff( path, L"*" );

	if ( !cff.findFirstFile() )
		return ok;

	do
	{
		ok = TakeOwnershipRec( cff.getCurrentPath(), stringSid, &cff.m_data );
	}
	while( cff.findNextFile() );

	return ok;

//	int len = wcslen( path );
//
//	cWCStr wcsFind( len + maxLen( fd.cFileName ) + 2 );
//
//	wchar_t *findPath = wcsFind.m_szBody;
//
//	if ( findPath[ len - 1 ] != L'\\' )
//	{
//		findPath[ len - 1 ] = L'\\';
//		len++;
//	}
//
//	findPath[ len ] = L'*';
//
//	HANDLE hFind = FindFirstFileT( findPath, &fd );
//
//	if ( hFind == INVALID_HANDLE_VALUE )
//	{
//		ODS_f( L"Assertion" );
//		return false;
//	}
//
//	do
//	{
//		if ( fd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY ) 
//		{
//			if ( ( wcscmp( fd.cFileName, L"." ) == 0 ) ||
//			     ( wcscmp( fd.cFileName, L".." ) == 0 )
//			)
//				continue;
//		}
//		
//		copyString( findPath + len, fd.cFileName, maxLen( fd.cFileName ) );
//
//		ok = TakeOwnershipRec( findPath, stringSid, &fd );
//
//		if ( !ok )
//		{
////			break;
//		}
//	}
//	while( FindNextFileT( hFind, &fd ) );
//
//	FindClose( hFind );
//
//	return ok;
}
