#pragma once

#include <windows.h>
#include "cSysCriticalSection.h"
#include "ODS.h"

class CProcessInfo
{
public:
	CProcessInfo() : hProc(NULL), hToken(NULL) {}
	~CProcessInfo() 
	{
		if (hProc) { 
			CloseHandle(hProc); 
		}
		if (hToken) {
			CloseHandle(hToken);
		}
		hProc = NULL;
		hToken = NULL;
	}

	BOOL Open(DWORD pid) 
	{
		hProc = OpenProcess(PROCESS_QUERY_INFORMATION|PROCESS_VM_READ, FALSE, pid);
		if (NULL == hProc) {
#ifdef  _ODSM
		DWORD err = GetLastError();
		ODSF(L"OpenProcess failed: %d", err);
#endif	
			return FALSE;
		}

		BOOL rc = OpenProcessToken(hProc, TOKEN_QUERY, &hToken);
		if (!rc) {
#ifdef  _ODSM
		DWORD err = GetLastError();
		ODSF(L"OpenProcessToken failed: %d", err);
#endif	
			return FALSE;
		}
		return TRUE;
	}
	
	DWORD GetSessionId() 
	{
		if (!hToken) {
			return 0;
		}

		DWORD	SessionId = 0, sz;
		BOOL rc = GetTokenInformation(hToken, TokenSessionId, &SessionId, sizeof(SessionId), &sz);

		if (!rc) {
#ifdef  _ODSM
		DWORD err = GetLastError();
		ODSF(L"GetSessionId failed: %d", err);
#endif	
			SessionId = 0;
		}

		return SessionId;
	}

	BOOL FindModule( const wchar_t *str );

	HANDLE GetHandle() { return hProc; }

private:
	HANDLE	hProc;	
	HANDLE	hToken;
};

#include "cDll.h"

typedef DWORD WINAPI typeGetModuleBaseNameW(
    HANDLE hProcess,
    HMODULE hModule,
    LPWSTR lpBaseName,
    DWORD nSize
    );

typedef BOOL WINAPI typeEnumProcessModules(
    HANDLE hProcess,
    HMODULE *lphModule,
    DWORD cb,
    LPDWORD lpcbNeeded
    );

typedef BOOL WINAPI typeEnumProcesses(
    DWORD * lpidProcess,
    DWORD   cb,
    DWORD * cbNeeded
    );

typedef DWORD WINAPI typeGetModuleFileNameExW(
    HANDLE hProcess,
    HMODULE hModule,
    LPWSTR lpFilename,
    DWORD nSize
    );


class cPSAPI_Dll : public CDll
{
public:
	static __declspec(noinline) cPSAPI_Dll &getStatic()
	{
		static cSysCriticalSection cs;
		cSysCriticalSection_Block csb( &cs );

		static cPSAPI_Dll dll;
		return dll;
	}

	typeGetModuleBaseNameW		*GetModuleBaseNameW;
	typeEnumProcessModules		*EnumProcessModules;
	typeEnumProcesses			*EnumProcesses;
	typeGetModuleFileNameExW	*GetModuleFileNameExW;
	
	cPSAPI_Dll() : CDll( false, L"psapi.dll", NULL, false )
	{
		if ( !GetProcAddress( "GetModuleBaseNameW", false, (void**)&GetModuleBaseNameW ) ) return;
		if ( !GetProcAddress( "EnumProcessModules", false, (void**)&EnumProcessModules ) ) return;
		if ( !GetProcAddress( "EnumProcesses", false, (void**)&EnumProcesses ) ) return;
		if ( !GetProcAddress( "GetModuleFileNameExW", false, (void**)&GetModuleFileNameExW ) ) return;
	}
};

//
//	Uses psapi for retriving process information.
//

bool GetProcessIdByName( const wchar_t *processName, DWORD *pDWID = NULL );

//
//	Same as GetProcessIdByName
//	but started at separate thread without impersonation.
//	
bool GetProcessIdByNameI( const wchar_t *processName, DWORD *pDWID = NULL );

//
//	TerminateProcessByName
//		Process name sample:  "soffice.exe" is same as Windows Task Manager shows at Processes tab as Image Name
//		Returns:
//			ERROR_SUCCESS if process is terminated.
//			ERROR_FILE_NOT_FOUND if process does not exist.
//			System last error if process was not terminated.
//
//int TerminateProcessByName( const wchar_t *processName, DWORD exitCode );