#include "token_info.h"
#include "cDll.h"
#include "ODS.h"

bool isWindowsVista_Internal()
{
	OSVERSIONINFO osvi;
	osvi.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);
	GetVersionEx ( &osvi );
	bool bIsWindowsVista = ( osvi.dwPlatformId == VER_PLATFORM_WIN32_NT ) && ( osvi.dwMajorVersion >= 6 );
	return bIsWindowsVista;
}

bool isWindowsVista()
{
	static bool bResult = isWindowsVista_Internal();
	return bResult;
}

static xTOKEN_ELEVATION_TYPE getElevationType()
{
	if ( !isWindowsVista() )
		return xTokenElevationTypeError;

	CDll dll( false, "advapi32.dll", NULL, false );

	if ( dll.error )
		return xTokenElevationTypeError;

	HANDLE hToken;
	void *proc;


	if ( !dll.GetProcAddress( "OpenProcessToken", false, &proc ) )
		return xTokenElevationTypeError;

	typedef 
		WINADVAPI
		BOOL
		WINAPI
		typeOpenProcessToken (
			IN HANDLE ProcessHandle,
			IN DWORD DesiredAccess,
			OUT PHANDLE TokenHandle
			);

	if ( !(( typeOpenProcessToken *)proc)( GetCurrentProcess(), TOKEN_READ, &hToken) )
		return xTokenElevationTypeError;


	if ( !dll.GetProcAddress( "GetTokenInformation", false, &proc ) )
		return xTokenElevationTypeError;

	DWORD infoLen;
	xTOKEN_ELEVATION_TYPE elevationType;

	typedef 
		WINADVAPI
		BOOL
		WINAPI
		typeGetTokenInformation (
			IN HANDLE TokenHandle,
			IN TOKEN_INFORMATION_CLASS TokenInformationClass,
			OUT LPVOID TokenInformation,
			IN DWORD TokenInformationLength,
			OUT PDWORD ReturnLength
			);

	BOOL ok = ( ( typeGetTokenInformation * ) proc)( hToken, (TOKEN_INFORMATION_CLASS)TokenElevationType, &elevationType, sizeof(elevationType), &infoLen );

	CloseHandle( hToken );

	if ( !ok )
		return xTokenElevationTypeError;

//#ifdef _DEBUG
//	wchar_t msg[ 1024 ];
//	FormatT( msg, maxLen( msg ), L"TokenElevationType %d %s", elevationType, (elevationType == TokenElevationTypeFull ? L"FULL" : L"LIMITED" ) );
//	MessageBoxT( 0, msg, L"isWindowsVistaLimitedProcess", MB_OK );
//#endif

	//
	// If user account control is disabled elevationType == TokenElevationTypeDefault
	//	so we have to check really limited here.
	//
	return elevationType; // == xTokenElevationTypeLimited; //!= TokenElevationTypeFull;
}

//
//	isWindowsVistaLimitedProcess
//		Returns TRUE if the application is started in Windows Vista invironment as limited process.
//		Otherwise and in others windows environments it returns false.
//
bool isWindowsVistaLimitedProcess( bool bForced )
{
	//ODSFuncS( getWCStr( bForced ) );

	//
	// If user account control is disabled elevationType == TokenElevationTypeDefault
	//
	static xTOKEN_ELEVATION_TYPE elevationType = getElevationType();

	if ( bForced )
		elevationType = getElevationType();

	//ODSFuncRes( bResult );
	
	return elevationType == xTokenElevationTypeLimited;
}

class sOpenThreadToken
{
public:

	HANDLE m_ThreadHandle;
	DWORD m_DesiredAccess;
	BOOL m_OpenAsSelf;
	PHANDLE m_pTokenHandle;

	DWORD m_lastError;

	BOOL m_res;

	const wchar_t *m_odsFunc;
	int m_odsLine;

	sOpenThreadToken( HANDLE ThreadHandle, DWORD DesiredAccess, BOOL OpenAsSelf, PHANDLE TokenHandle, const wchar_t *odsFunc, int odsLine )
	{
		m_ThreadHandle = ThreadHandle;
		m_DesiredAccess = DesiredAccess;
		m_OpenAsSelf = OpenAsSelf;
		m_pTokenHandle = TokenHandle;
		
		m_lastError = 0;
		
		m_res = false;

		m_odsFunc = odsFunc;
		m_odsLine = odsLine;
	}

	void func()
	{
		//ODSFunc;

		m_res = OpenThreadToken( m_ThreadHandle, m_DesiredAccess, m_OpenAsSelf, m_pTokenHandle );
		
		if ( !m_res )
		{
			m_lastError = GetLastError();
			ODS_fls( _FormatT( L"OpenThreadToken error 0x%X, Called from %s:%d, ", m_lastError, m_odsFunc, m_odsLine ) );
		}

		//ODSFuncRes( _FormatT( L"%d, last error 0x%X", m_res, m_lastError ) );
	}

	static DWORD WINAPI funcStatic( sOpenThreadToken *pThis )
	{
		pThis->func();

		return 0;
	}
};

//
//	OpenThreadToken function return error if we get token from impersonated thread.
//	The idea is:
//		We'll take impersonated thread token by new thread ( the new is not impersonated ).
//
static bool OpenThreadTokenMT( HANDLE ThreadHandle, DWORD DesiredAccess, BOOL OpenAsSelf, PHANDLE TokenHandle, const wchar_t *odsFunc, int odsLine )
{
	ODSFuncS( _FormatT( L"%p, %X, %d, %p", ThreadHandle, DesiredAccess, OpenAsSelf, TokenHandle ) );

	sOpenThreadToken sOTT( ThreadHandle, DesiredAccess, OpenAsSelf, TokenHandle, odsFunc, odsLine );

	DWORD dwID = 0;

	HANDLE hThread = CreateThread( NULL, 0, (LPTHREAD_START_ROUTINE)sOpenThreadToken::funcStatic, &sOTT, 0, &dwID );

	if ( hThread == NULL )
	{	ODS_fls( L"hThread == NULL" );	return false; }

	DWORD res = WaitForSingleObject( hThread, INFINITE );

	CloseHandle( hThread );

	if ( res != WAIT_OBJECT_0 )
	{	ODS_fls( _FormatT( L"res = %X", res ) );	return false; }

	if ( sOTT.m_res == TRUE )
	{ ODSFuncRes( true ); return true; }

	ODS_fls( _FormatT( L"Last error - %X", sOTT.m_lastError ) );

	SetLastError( sOTT.m_lastError );

	return false;
}

//
//	Returns handle to the process current thread impersonated token.
//
HANDLE GetProcessImpersonatedToken( const wchar_t *odsFunc, int odsLine )
{
	HANDLE hTokenI = NULL;
	HANDLE hToken = NULL;
	HANDLE hThread = NULL;
	HANDLE hProcess = GetCurrentProcess();
	
	SetLastError( 0 );

	BOOL ok = DuplicateHandle( hProcess, GetCurrentThread(), hProcess, &hThread, 0, true, DUPLICATE_SAME_ACCESS );

	int err = GetLastError();

	//ODSF( L"DuplicateHandle( 0x%X, 0x%X ) == %s, err = 0x%X", hProcess, hThread, getWCStr( ok ), err );

	if ( hProcess != NULL )
		CloseHandle( hProcess );

	ok = OpenThreadTokenMT( hThread,  TOKEN_QUERY | TOKEN_DUPLICATE , true, &hToken, odsFunc, odsLine ); // TOKEN_ALL_ACCESS

	err = GetLastError();

	//ODSF( L"OpenThreadToken( 0x%X ) == %s, err = 0x%X", hThread, getWCStr( ok ), err );

	if ( ok )
	{
		ok = DuplicateTokenEx( hToken, MAXIMUM_ALLOWED, NULL, SecurityImpersonation, TokenImpersonation, &hTokenI );

		err = GetLastError();
		//ODSF( L"DuplicateTokenEx( 0x%X ) == %s, hTokenI = 0x%X, err = 0x%X", hToken, getWCStr( ok ), hTokenI, err );
	}

	if ( hThread != NULL )
		CloseHandle( hThread );

	if ( hToken != NULL )
		CloseHandle( hToken );

	return hTokenI;
}

//
//	GetPipeToken
//		1. Impersonates to pipe token.
//		2. Gets the token.
//		3. Reverts to self back.
//	Returns impersonanted pipe token or NULL.
//
HANDLE GetPipeToken( HANDLE hPipe )
{
	BOOL rc = ImpersonateNamedPipeClient(hPipe);
	if (!rc)
	{
		DWORD err = GetLastError();
		ODS_fls( cWCSF().FormatT( L"Impersonation failed %x", err ) );
		return NULL;
	}

	HANDLE hTokenI = GetProcessImpersonatedToken_ODS();
	if ( !RevertToSelf() )
	{
		CloseHandle( hTokenI );
		hTokenI = NULL;
		DWORD err = GetLastError();
		ODS_fls( cWCSF().FormatT( L"RevertToSelf failed %x", err ) );
	}
	return hTokenI;
}
