#include "unicodeFuncs.h"
#include "fileView.h"
#include "autoDeallocate.h"
#include "translateDll.h"
#include "cStringArray.h"
#include <stdio.h>

cTranslateDll *pTranslateDll = NULL;

static const wchar_t *szLangFileDllName = L"langfile2.dll";
const wchar_t *szLangFileTxtName = L"langfile.txt";

void SetTranslateDll( cTranslateDll *dll = NULL )
{
	pTranslateDll = dll;
}

void cTranslateDll::ReloadHelpModule()
{
	m_wcsLangFilePath.Free();

	//
	// A small problem.
	//	Html help does not load new chm file with same name, even if locations are different.
	//	It caches the name :(
	//	So new help file with same name does not changed without application reload.
	//	The two choise were implemented. But no one works properly, see below.
	//	There is one recommendation only, the names of help files must be different.
	//
	//

	//
	//	Choise 1. The function does not reset the help file name.
	//
	//HMODULE hdll = GetModuleHandleT( L"ole32.dll" );
	//typedef void STDAPICALLTYPE typeCoFreeUnusedLibrariesEx(IN DWORD dwUnloadDelay, IN DWORD dwReserved);
	//typeCoFreeUnusedLibrariesEx *FreeComDlls = NULL;

	//if ( hdll )
	//{
	//	FreeComDlls = (typeCoFreeUnusedLibrariesEx *)::GetProcAddress( hdll, "CoFreeUnusedLibrariesEx" );

	//	if ( FreeComDlls != NULL )
	//	{
	//		FreeComDlls(0, 0);
	//		return;
	//	}
	//}

	//
	//	Choise 2. It was found at RSDN, may be it reload the chached name but
	//	System crushes at first calling of HtmlHelp
	//

	//const wchar_t *helpModuleName = L"HHCtrl.OCX";
	//HMODULE hModule = GetModuleHandleT( helpModuleName );
	//if(hModule)
	//{
	//	FreeLibrary(hModule);
	//	hModule = LoadLibraryT( helpModuleName );
	//}
}


//
// Prepares utility clear name
//
// static
bool cTranslateDll::makeClearName( wchar_t *unOut, const wchar_t *unIn )
{
	if ( ( unIn != NULL ) && ( unIn != unOut ) )
		wcscpy_s( unOut, MAX_PATH, unIn );

	int len = (int)wcslen( unOut );

	if ( ( _wcsicmp( unOut + len - 4, L".exe" ) == 0 ) ||
	  ( _wcsicmp( unOut + len - 4, L".dll" ) == 0 ) )
	{
		 unOut[ len - 4 ] = L'\0';
	}

	return true;
}

// Variants:
//	1. Bestcrypt.exe, BCArchive.exe ....
//	2. BCWipe.dll
//	3. BCShExt.dll 	64 bit version tries to load nearest langfile.dll first
//	4. BCKeyMan.dll
//	5. BCUnIstall.exe

// First case...
bool cTranslateDll::SetDefault( HINSTANCE hDllInst )
{
	cAutoDeallocate < wchar_t > adPath;
	cAutoDeallocate < wchar_t > adDllPath;

	wchar_t *path = adPath.Allocate( MAX_PATH );
	if ( path == NULL )
		return false;

	GetModuleFileNameT( NULL, path, MAX_PATH );

	int len = splitPath( path, path );
	wchar_t *sn = path + len;
	wchar_t *un = sn;

	if ( hDllInst != NULL )
	{
		wchar_t *dllPath = adDllPath.Allocate( MAX_PATH );

		if ( dllPath == NULL )
			return false;
		
		GetModuleFileNameT( hDllInst, dllPath, MAX_PATH );

		len = splitPath( dllPath );
		un = dllPath + len;
	}

	return SetDefault( path, sn, un );
}

bool cTranslateDll::LoadLangFileDll(
		const wchar_t *langFileFolder	// Folder ( or subfolder ) where langFile.dll may be found
	)
{
	//ODSFunc;

	if ( getDefaultCodePage() != CP_UTF8 )
		return false;

	wchar_t path[ MAX_PATH ];

	if ( langFileFolder == NULL )
		GetModuleFileNameT( NULL, path, maxLen( path ) );
	else
		wcscpy_s( path, langFileFolder );

	int len = (int)wcslen( path );

	if ( path[ len - 1 ] != L'\\' )
	{
		wcscpy_s( path + len, MAX_PATH - len, L"\\" );
		len++;
	}

	// Loads from home dir

	wcscpy_s( path + len, MAX_PATH - len, szLangFileDllName );

	return InitAndLoad( path );
}

bool cTranslateDll::InitAndLoad( const wchar_t *path )
{
	Load( true, path, NULL, false );

	if ( error ) // Tries the folder above
	{
		wchar_t pathAbove[ MAX_PATH ];
		splitPath( path, pathAbove );

		int len = splitPath( pathAbove );
		wcscpy_s( pathAbove + len, MAX_PATH - len, szLangFileDllName );
		Load( true, pathAbove, NULL, false );

		if ( error )
			return false;
	}

	if ( !GetProcAddress( "lfTranslateStrW", false, (void**)&m_proc_lfTranslateStrW ) ) return false;
	if ( !GetProcAddress( "lfTryTranslateStrW", false, (void**)&m_proc_lfTryTranslateStrW ) ) return false;
	if ( !GetProcAddress( "lfTranslateDialogW", false, (void**)&m_proc_lfTranslateDialogW ) ) return false;
	if ( !GetProcAddress( "lfTranslateMenuW", false, (void**)&m_proc_lfTranslateMenuW ) ) return false;
	if ( !GetProcAddress( "lfLoadLang", false, (void**)&m_proc_lfLoadLang ) ) return false;
	if ( !GetProcAddress( "lfReset", false, (void**)&m_proc_lfReset ) ) return false;
	if ( !GetProcAddress( "lfSetDialogLayout", false, (void**)&m_proc_lfSetDialogLayout ) ) return false;
	if ( !GetProcAddress( "lfSetTextDirection", false, (void**)&m_proc_lfSetTextDirection ) ) return false;

	error = false;

	return true;
}

bool cTranslateDll::LinkTo( HINSTANCE hInst )
{
	if ( error ) return false;

	typedef void typeLinkTranslateDll( cTranslateDll *pDll );

	typeLinkTranslateDll *procLinkTranslateDll = (typeLinkTranslateDll *)::GetProcAddress( hInst, "LinkTranslateDll" );

	if ( procLinkTranslateDll == NULL ) return false;

	procLinkTranslateDll( this );

	return true;
}

bool cTranslateDll::SetDefault(
		const wchar_t *langFileFolder,	// Folder ( or subfolder ) where langFile.dll may be found
		const wchar_t *localSectionName,// Registry section with 'local' DWORD value
		const wchar_t *utilityName,		// Name of utility to be translated
		bool linkToTheApp				// = true, link the dll to current application 
	)
{
//	if ( !error )
//		return true;

	if ( !LoadLangFileDll( langFileFolder ) )
		return false;

	const wchar_t *langFile = NULL;

	wchar_t path[ MAX_PATH ];

	if ( GetLangFilePath( localSectionName, utilityName, path ) )
		langFile = path;
	else
	{
		GetModuleFileNameT( NULL, path, MAX_PATH );
		int len = splitPath( path );
		wcscpy_s( path + len, maxLen( path ) - len, szLangFileTxtName );

		if ( FileExistsT( path ) )
			langFile = path;
		else
			langFile = NULL;
	}

	if ( LoadLang( langFile ) )
	{
		if ( linkToTheApp )
			SetTranslateDll( this );
	}

	return !error;
}

static bool RegGetString( HKEY root, const wchar_t *section, const wchar_t *value, wchar_t *str, int strMaxLen )
{
	HKEY hKey;

	LONG lRes = RegOpenKeyExT(
				root,
				section, 0,
				KEY_QUERY_VALUE, //  | REG_wow6432Key,
				&hKey );

	if (lRes != ERROR_SUCCESS)
			return false;

	DWORD dwType;
	ULONG size = strMaxLen * sizeof( *str );

	lRes = RegQueryValueExT( hKey, value, 0, &dwType, (BYTE *)str, &size );

	RegCloseKey( hKey );

	if ( ( lRes != ERROR_SUCCESS ) || ( size > strMaxLen * sizeof( *str ) ) )
		return false;

	return true;
}

static bool RegSetString( HKEY root, const wchar_t *section, const wchar_t *value, const wchar_t *str )
{
	HKEY hKey;
	
	DWORD dwDisp;	

	LONG lRes =
		RegCreateKeyExT(
			root,
			section, 0,
			NULL,
			0,
			KEY_SET_VALUE, //  | REG_wow6432Key,
			NULL,
			&hKey,
			&dwDisp );
	
	if ( lRes != ERROR_SUCCESS )
		return false;

	DWORD dwType = REG_SZ;	

	ULONG size = 0;

	if ( str == NULL )
		str = L"";

	size = ( (ULONG)wcslen( str ) + 1 ) * sizeof( wchar_t );

	lRes = RegSetValueExT( hKey, value, 0, dwType, (BYTE*)str, size );

	RegCloseKey( hKey );

	if ( lRes != ERROR_SUCCESS ) 
		return false;

	return true;
}


static const wchar_t *szJeticoSection = L"SOFTWARE\\Jetico";

//
//	getDefaultUtilityName
//		Returns first section name at HKEY_CURRENT_USER\Software\Jetico key
//
static const wchar_t *getDefaultUtilityName()
{
	static cAutoDeallocate < wchar_t > ad;

	if ( ad.m_ptr != NULL )
		return ad.m_ptr;

	// Gets default utility name

	if ( ad.Allocate( 255 ) == NULL )
		return NULL;

	HKEY hKey;

	LONG lRes = RegOpenKeyExT(
				HKEY_CURRENT_USER,
				szJeticoSection, 0,
				KEY_ENUMERATE_SUB_KEYS, //  | REG_wow6432Key,
				&hKey );

	if ( lRes != ERROR_SUCCESS )
		return NULL;

	DWORD cnt = 0;
	DWORD size = ad.m_len;

	lRes = RegEnumKeyExT( hKey, cnt, ad.m_ptr, &size, NULL, NULL, NULL, NULL );

	RegCloseKey( hKey );

	if ( lRes != ERROR_SUCCESS )
		return NULL;

	return ad.m_ptr;
}

static int defaultTranslation( wchar_t *wsOut, int wsMaxLen, const char *szMbs, int keysize, bool *pFound )
{
	size_t size;
	mbstowcs_s( &size, wsOut, wsMaxLen, szMbs, _TRUNCATE );

	if ( pFound != 0 )
		*pFound = false;

	return (int)size;
}

///////////////////
//
//	Code for checking format parameters like "%d %s" or "%1 %2"
//	It have to be checked when lang file is prepared by parser, as well as when the string is trabslated
//	becuse the changed parameters can crash system.
//

//
//template < class T > int countSymbols( const T *ptr, T symbol )
//{
//	int q = 0;
//
//	for( ; *ptr != 0; ptr++ )
//	{
//		if ( *ptr == symbol )
//			q++;
//	}
//
//	return q;
//}
//
//static bool checkParamCounts( const wchar_t *wsOut, const char *szMbs )
//{
//	return countSymbols( wsOut, L'%' ) == countSymbols( szMbs, '%' );
//}

static int isFigure( const wchar_t *wcs )
{
	if ( wcs[0] != L'%' )	return 0;
	if ( wcs[1] < L'1' )	return 0;
	if ( wcs[1] > L'9' )	return 0;
	//if ( ( wcs[2] != L' ' )
	//	|| ( wcs[2] != L'\0' ) )
	//						return 0;

	return wcs[1] - L'0';
}

static int isFigure( const char *wcs )
{
	if ( wcs[0] != '%' )	return 0;
	if ( wcs[1] < '1' )	return 0;
	if ( wcs[1] > '9' )	return 0;
	//if ( ( wcs[2] != ' ' )
	//	|| ( wcs[2] != '\0' ) )
	//						return 0;

	return wcs[1] - '0';
}

static bool T[10];

static const wchar_t *getNextParam( const wchar_t *wcs )
{
	const wchar_t *wc = wcschr( wcs, L'%' );

	if ( wc == NULL )
		return wc;

	if ( wc[1] != L'%' )
		return wc;

	return getNextParam( wc + 2 );
}

static const char *getNextParam( const char *wcs )
{
	const char *wc = strchr( wcs, '%' );

	if ( wc == NULL )
		return wc;

	if ( wc[1] != '%' )
		return wc;

	return getNextParam( wc + 2 );
}

//static bool fillTable( const wchar_t *wcs )
//{
//	for ( const wchar_t *wc = getNextParam( wcs ); wc != NULL; wc = getNextParam( wc + 2 ) )
//	{
//		int argNum = isFigure( wc );
//
//		if ( ( argNum == 0 ) || ( argNum >= maxLen( T ) ) )
//		{
//			ODS_fls( L"Assertion!" );
//			return false;
//		}
//
//		if ( T[ argNum ] )
//		{
//			ODS_fls( L"Assertion!" );
//			return false;
//		}
//
//		T[ argNum ] = true;
//	}
//
//	return true;
//}

template < class WC > static bool fillTable( WC *wcs, bool treatSAsFigure = false )
{
	int count = 0;

	for ( WC *wc = getNextParam( wcs ); wc != NULL; wc = getNextParam( wc + 2 ) )
	{
		int argNum = 0;

		//
		//	See Note 1, at checkParamCounts
		//
		if ( treatSAsFigure  )
		{
			if ( wc[1] != (WC)'s' )
			{
				ODS_fls( L"Assertion!" );
				return false;
			}

			argNum = ++count;
		}
		else
			argNum = isFigure( wc );

		if ( ( argNum == 0 ) || ( argNum >= maxLen( T ) ) )
		{
			ODS_fls( L"Assertion!" );
			return false;
		}

		if ( T[ argNum ] )
		{
			ODS_fls( L"Assertion!" );
			return false;
		}

		T[ argNum ] = true;
	}

	return true;
}

template < class CT > static int checkFigureCount( CT *wcs, bool treatSAsFigure = false )
{
	memset( T, 0, sizeof( T ) );

	if ( !fillTable( wcs, treatSAsFigure ) )
		return false;

	int count = maxLen( T );

	while( !T[--count] ) {}

	for( int i = 1; i < count; i++ )
	{
		if ( !T[i] )
		{
			ODS_fls( L"Assertion!" );
			return false;
		}
	}

	return count;
}

static bool isSameValues( const wchar_t *wc, const char *mb )
{
	for ( int i = 0; i < 2; i++ )
	{
		if ( wc[i] != ( const wchar_t )mb[i] )
			return false;
	}

	return true;
}

static bool checkParamCounts( const wchar_t *wsOut, const char *szMbs )
{
	//
	//	BestCrypt and other utilities uses resource string with '%' symbols for 
	//	double zero terminated strings( the '%' symbols are changed to '\0' ). 
	//	Distinguished fieture of the string is the '%' symbol at the string end.
	//
	//	So, we will not check parameters for the strings.
	//

	int len = (int)strlen( szMbs );

	if ( ( len != 0 ) && ( szMbs[len-1] == '%' ) )
	{
		return true;
	}

	const wchar_t *wc = getNextParam( wsOut );
	const char *mb = getNextParam( szMbs );

	if ( wc == NULL )
	{
		if ( mb == NULL )
			return true;
		
		ODS_fls( L"Assertion!" );
		return false;
	}

	if ( isFigure( wc ) )
	{
		if ( checkFigureCount( wsOut ) == checkFigureCount( szMbs ) )
			return true;

		ODS_fls( L"Assertion!" );

		return false;
	}

	for ( int i = 0; ; wc = getNextParam( wc + 2 ), mb = getNextParam( mb + 2 ), i++ )
	{
		if ( wc == NULL )
		{
			if ( mb == NULL )
				return true;
			
			ODS_fls( L"Assertion!" );
			return false;
		}

		if ( !isSameValues( wc, mb ) )
		{
			//
			//	Note 1. The translators may change parameters
			//	'%s %s' = '%1 %2'
			//	So,
			if ( isFigure( mb ) )
			{
				if ( checkFigureCount( wsOut, true ) == checkFigureCount( szMbs ) )
					return true;
			}

			ODS_fls( L"Assertion!" );
			return false;
		}
	}
}

//
// End of code for parameters checking.
//
//////////////

int cTranslateDll::Translate( wchar_t *wsOut, int wsMaxLen, const char *szMbs, int keysize, bool *pFound, bool bCheckParams )
{
	if ( error || ( hInst == NULL ) || ( m_hDict == undefinedDictionary ) )
		return defaultTranslation( wsOut, wsMaxLen, szMbs, keysize, pFound );

	int res = 0;

	if ( pFound != NULL )
	{
		*pFound = false;

		if ( m_proc_lfTryTranslateStrW != NULL )
		{
			res = m_proc_lfTryTranslateStrW( m_hDict, wsOut, wsMaxLen, szMbs,  keysize );

			if ( res > 0 )
			{
				if ( bCheckParams && !checkParamCounts( wsOut, szMbs ) )
				{
					ODS_fls( L"Assertion!" );
					return defaultTranslation( wsOut, wsMaxLen, szMbs, keysize, pFound );
				}

				*pFound = true;
				return res;
			}
		}
	}

	res = m_proc_lfTranslateStrW( m_hDict, wsOut, wsMaxLen, szMbs,  keysize );

	if ( ( res > 0 ) && bCheckParams && !checkParamCounts( wsOut, szMbs ) )
	{
		ODS_fls( L"Assertion!" );
		return defaultTranslation( wsOut, wsMaxLen, szMbs, keysize, pFound );
	}

	return res;
}

int cTranslateDll::Translate( char *mbsOut, int mbsMaxLen, const char *szMbs, int keySize )
{
	cWCStr wcs( mbsMaxLen );

	int res = Translate( wcs.m_szBody, wcs.m_len, szMbs, keySize );
	
	size_t size;
	wcstombs_s( &size, mbsOut, mbsMaxLen, wcs.m_szBody, _TRUNCATE );

	return (int)res;
}

int cTranslateDll::TranslateDialog( HWND hDlg, bool disableFlip )
{
	//ODSFunc;
	//
	//	The disableFlip is true by default.
	//	Every RTL dialog is flipped at domodall or domodeless function by resource preloading.
	//	if m_flipDialog is assigned.
	//	The m_flipDialog value is assigned at cTranslateDll::LoadLang
	//	by reading of lang file or by getting process layout it depends on _USRDLL define
	//	because dll inherites lang file from process.
	// 
	if ( !disableFlip && ( m_proc_lfSetDialogLayout != NULL ) )
	{
		int layout = m_flipDialog ? LF_RTL : LF_LTR;
		m_proc_lfSetDialogLayout( hDlg, layout );
	}

	if ( error || ( hInst == NULL ) || ( m_hDict == undefinedDictionary ) )	return -1;

	return m_proc_lfTranslateDialogW( m_hDict, hDlg );
}

int cTranslateDll::TranslateMenu( HMENU hMenu )
{
	if ( error || ( hInst == NULL ) || ( m_hDict == undefinedDictionary ) )	return -1;
	return m_proc_lfTranslateMenuW( m_hDict, hMenu );
}

//
//	Checks if the resource file mantioned as resource file
//	// RCFile "..\BCResident\BCResident.rc"
//
bool cTranslateDll::CheckResourceFile( const wchar_t *rcFileName, bool turnTranslationOffIfError )
{
	if ( m_wcsLangFilePath.isEmpty() )
		return false;

	cStringFile sf( m_wcsLangFilePath.m_szBody );

	if ( sf.m_err )
		return false;

	cAutoDeallocate < wchar_t > adbf( MAX_PATH + 20 );

	const wchar_t *szLastStr = L"// Resource file start";
	int lastStrLen = (int)wcslen( szLastStr );

	const wchar_t *szRCFileStr = L"// RCFile \"";
	int rcFileSrtLen = (int)wcslen( szRCFileStr );

	int fileNameLen = (int)wcslen( rcFileName );

	bool found = false;

	while( sf.ReadStringT( adbf.m_ptr, adbf.m_len ) )
	{
		if ( wcsncmp( adbf.m_ptr, szLastStr, lastStrLen ) == 0 )
			break;

		//
		// Sample : '// RCFile "..\BCResident\BCResident.rc"'
		//

		if ( wcsncmp( adbf.m_ptr, szRCFileStr, rcFileSrtLen ) != 0 )
			continue;

		int len = (int)wcslen( adbf.m_ptr );
		while( adbf.m_ptr[len-1] != L'\"' )
		{
			adbf.m_ptr[--len] = L'\0';
		}

		adbf.m_ptr[--len] = L'\0';

		if ( _wcsicmp( adbf.m_ptr + len - fileNameLen, rcFileName ) == 0 )
		{
			found = true;
			break;
		}
	}

	if ( found )
		return true;

	if ( turnTranslationOffIfError )
		LoadLang( NULL );

	return false;
}

//
//	Loads( or reload ) new lang file.
//	If langFile parameter is NULL it simple resets translation.
//	
bool cTranslateDll::LoadLang( const wchar_t *langFile )
{
	if ( hInst == NULL )	return false;

	if ( m_proc_lfReset == NULL )
	{
		ODS_fls( L"Assertion!" );
		return false;
	}

	if ( m_hDict != undefinedDictionary )
	{
		m_proc_lfReset( m_hDict );
		m_hDict = undefinedDictionary;
	}

	error = false;
	m_flipDialog = false;
	
	m_wcsLangFilePath.Set( langFile );

	WIN32_FIND_DATAW findData;

	if ( ( langFile == NULL ) || ( langFile[0] == L'\0' ) )
	{
#ifndef _USRDLL 
		setProcessDefaultLayout( false );
#endif	
		error = true;
		return true;
	}
	else
	if ( !FileExistsT( langFile, &findData ) || 
			(
				// Is Empty, the file has not translation..
				( findData.nFileSizeLow == 0 ) &&
				( findData.nFileSizeHigh == 0 )
			)
		)
	{
		error = true;
		return true;
	}

	bool ok = m_proc_lfLoadLang( langFile, &m_hDict );
	if ( !ok )
		return false;


#ifdef _USRDLL 
	// Dll inherits RTL from process
	DWORD dwLayout;
	if ( getProcessDefaultLayout( &dwLayout ) )
		m_flipDialog = ( dwLayout & LAYOUT_RTL ) != 0;
#else
	wchar_t buff[ 256 ];
	if ( Translate( buff, maxLen( buff ), "RightToLeft" ) && (_wcsicmp( buff, L"TRUE" ) == 0 ) )
		m_flipDialog = true;
	else
	if ( Translate( buff, maxLen( buff ), "LANGUAGE_READ_START" ) )
		m_flipDialog = (_wcsicmp( buff, L"Right" ) == 0 );

	//
	//	23.03.2010
	//	Dlls inherit the assigned layout, see commants above.
	//
	ok = setProcessDefaultLayout( m_flipDialog );
#endif

	return ok;
}

bool cTranslateDll::ReadHelpFileName( wchar_t *path, int pathLen )
{
	bool bFound = false;

	if ( Translate( path, pathLen, "HELP_FILE", 0, &bFound ) && bFound )
		return true;

	return false;
}

bool cTranslateDll::ReadHelpFileDownloadPath( wchar_t *path, int pathLen )
{
	bool bFound = false;

	if ( Translate( path, pathLen, "HELP_FILE_WEB_SITE", 0, &bFound ) && bFound )
		return true;

	return false;
}

const wchar_t *szTranslationFolder = L"TranslationFolder";
const wchar_t *szTranslation = L"Translation";
const wchar_t *szTranslationProduct = NULL;

//#include <Psapi.h>
//static 
bool cTranslateDll::getModuleTranslationFolder( wchar_t *folderName, const wchar_t *modulePath )
{
	wchar_t dir[ MAX_PATH ];
	wchar_t utilName[ MAX_PATH ];

	if ( ( modulePath == NULL ) || ( wcslen( modulePath ) == 0 ) )
	{
		GetModuleFileNameT( NULL, utilName, maxLen( utilName ) );
		GetLongPathNameT( utilName, dir, maxLen( dir ) ); 
		modulePath = dir;
	}

	splitPath( modulePath, dir, utilName );

	makeClearName( utilName );

	// Variants are:
	// C:\Program Files\Jetico\BCWipe
	// C:\Program Files\Jetico\BCArchive
	// C:\Program Files\Jetico\BestCrypt
	// C:\Program Files\Jetico\BestCrypt\BCArchive

	// Translation root folder to be found is
	// C:\Program Files\Jetico\Translation
	splitPath( dir, dir ); // C:\Program Files\Jetico or C:\Program Files\Jetico\BestCrypt

	do
	{
		FormatT( folderName, MAX_PATH, L"%s\\%s", dir, szTranslation );
		if ( FileExistsT( folderName ) )
		{
			wcscat_s( folderName, MAX_PATH, L"\\" );
			wcscat_s( folderName, MAX_PATH, utilName );
			return FileExistsT( folderName ) == TRUE;
		}
	}
	while( splitPath( dir, dir ) > 4 );

	return false;
}


//
//	regSetTranslationFolder
//
// static
bool cTranslateDll::regSetTranslationFolder( HKEY root, const wchar_t *fileFolderName, const wchar_t *utilityName, const wchar_t *vendorName )
{
	wchar_t dir[ MAX_PATH ];

	if ( fileFolderName != NULL )
	{
		wcscpy_s( dir, fileFolderName );

		DWORD dwAttr = GetFileAttributesT( dir );

		if ( ( dwAttr & FILE_ATTRIBUTE_DIRECTORY ) == 0 )
		{
			int pos = splitPath( dir );
			dir[ pos ] = L'\0';
		}
		
		int len = (int)wcslen( dir );

		if ( ( len > 0 ) && ( dir[ len - 1 ] != L'\\' ) )
			wcscpy_s( dir + len, maxLen( dir ) - len, L"\\" );

		fileFolderName = dir;
	}

	wchar_t bf[ MAX_PATH ];

	if ( utilityName == NULL )
	{
		GetModuleFileNameT( NULL, bf, maxLen( bf ) );
		wchar_t *un = bf + splitPath( bf );
		makeClearName( un );
		utilityName = un;
	}

	wchar_t section[ MAX_PATH ];
	FormatT( section, maxLen( section ), L"SOFTWARE\\%s\\%s", vendorName, L"LogView" );
	RegSetString( root, section, szTranslationFolder, fileFolderName );

	FormatT( section, maxLen( section ), L"SOFTWARE\\%s\\%s", vendorName, utilityName );

	return RegSetString( root, section, szTranslationFolder, fileFolderName );
}

//static 
bool cTranslateDll::regGetTranslationFolder( HKEY root, wchar_t *folder, const wchar_t *utilityName, const wchar_t *vendorName )
{
	wchar_t bf[ MAX_PATH ];

	if ( utilityName == NULL )
	{
		GetModuleFileNameT( NULL, bf, maxLen( bf ) );
		wchar_t *un = bf + splitPath( bf );
		makeClearName( un );
		utilityName = un;
	}

	//
	// Reads assigned local from registry by utility name.
	//		Checks HKEY_CURRENT_USER - the utility setting first
	//		Checks HKEY_LOCAL_MACHINE - the setup setting then.
	//
	wchar_t section[ MAX_PATH ];
	FormatT( section, maxLen( section ), L"SOFTWARE\\%s\\%s", vendorName, utilityName );

	//DWORD local = -1;

	//if ( RegGetDW( HKEY_CURRENT_USER, bf, L"TranslationLocal", &local ) ||
	//	 RegGetDW( HKEY_LOCAL_MACHINE, bf, L"TranslationLocal", &local )
	//	)
	//		return local;

	if ( !RegGetString( root, section, szTranslationFolder, folder, MAX_PATH ) )
		return false;

	int len = (int)wcslen( folder );

	if ( len != 0 )
	{
		if ( folder[ len - 1 ] != L'\\' )
			wcscpy_s( folder + len, MAX_PATH - len, L"\\" );
	}

	return true;
}

//static 
bool cTranslateDll::regGetTranslationFolder( wchar_t *folder, const wchar_t *utilityName )
{
	wchar_t bf[ MAX_PATH ];

	if ( utilityName == NULL )
	{
		GetModuleFileNameT( NULL, bf, maxLen( bf ) );
		utilityName = bf + splitPath( bf );
	}

	wchar_t clearSectionName[ MAX_PATH ];
	makeClearName( clearSectionName, utilityName );

	//
	// Reads assigned local from registry by utility name.
	//		Checks HKEY_CURRENT_USER - the utility setting first
	//		Checks HKEY_LOCAL_MACHINE - the setup setting then.
	//
	if ( cTranslateDll::regGetTranslationFolder( HKEY_CURRENT_USER, folder, clearSectionName ) )
	{
		// The string with translation folder exists
		// The app translation was changed
		// so it returs true even if the folder name is empty - the default ( English ) translation was assigned.
		// if ( wcslen( folder ) != 0 )
		return true;
	}
	else
	if ( cTranslateDll::regGetTranslationFolder( HKEY_LOCAL_MACHINE, folder, clearSectionName ) )
	{
		if ( wcslen( folder ) != 0 )
			return true;
	}

	return false;
}

static bool isExecutable( const wchar_t *name )
{
	int len = (int)wcslen( name );

	if ( len < 4 )
		return false;

	return _wcsicmp( name + len - 4, L".exe" ) == 0;
}

//
//	See III. Module translation. at V:\Includes\!TranslationStrategy.txt
//
bool cTranslateDll::GetLangFilePath(
		const wchar_t *localSectionName,// Registry section with 'local' DWORD value
		const wchar_t *utilityName,		// Name of utility to be translated
		wchar_t *path					// Output result( MAX_PATH value )
	)
{
	//ODSFunc;

	if ( localSectionName == NULL )
		return false;

	if ( utilityName == NULL )
		utilityName = localSectionName;

	wchar_t bf[ MAX_PATH ];

	bool ok = false;
	
	if ( !ok ) ok = regGetTranslationFolder( bf, localSectionName );

	const wchar_t *appT[] = {
		L"BestCrypt",
		L"BCWipe"
	};

	//
	//	BestCrypt cannot read other sections.
	//
	if ( !ok && ( wcsncmp( localSectionName, appT[0], 9 ) != 0 ) )
	{
		for( int i = 0; i < maxLen( appT ); i++ )
		{
			ok = regGetTranslationFolder( bf, appT[i] );
			if ( ok )
				break;
		}

		//
		//	Executable file reads BestCrypt or it's own section only.
		//
		if ( !ok && !isExecutable( localSectionName ) )
			ok = regGetTranslationFolder( bf, getDefaultUtilityName() );
	}

	if ( !ok )
	{
		// error = true;
		return false;
	}

	int len = (int)wcslen( bf );

	if ( len == 0 )
	{
		path[0] = L'\0';
		return ok;
	}
	
	if ( bf[len -1] != L'\\' )
	{
		wcscpy_s( bf + len, maxLen( bf ) - len, L"\\" );
		len++;
	}
	
	wcscpy_s( bf + len, maxLen( bf ) - len, utilityName );
	wcscat_s( bf + len, maxLen( bf ) - len, L".txt" );
	
	if ( !FileExistsT( bf ) )
	{
		wcscpy_s( bf + len, maxLen( bf ) - len, szLangFileTxtName );

		if ( !FileExistsT( bf ) )
			return false;
	}

	wcscpy_s( path, MAX_PATH, bf );
	return true;
//
//	return FindLocalFile( bf, local, path );
}

static bool getDWValue( const wchar_t *bf, DWORD *pLocal )
{
	cMBStr mbs( bf );

	DWORD dwValue = -1;

	sscanf_s( mbs.m_szBody, "=\"0x%x\"", &dwValue );

	if ( dwValue == -1 )
		sscanf_s( mbs.m_szBody, "0x%x", &dwValue );

	if ( dwValue == -1 )
		sscanf_s( mbs.m_szBody, "=\"%d\"", &dwValue );

	if ( dwValue == -1 )
		sscanf_s( mbs.m_szBody, "%d", &dwValue );

	if ( dwValue == -1 )
		return false;

	*pLocal = dwValue;

	return true;
}

DWORD cTranslateDll::GetLocaleNumber()
{
	if ( error )
		return 0;

	bool bFound = false;

	cWCStr str( 50 );
	DWORD locale = 0;

	if ( Translate( str.m_szBody, str.m_len, "LANGUAGE_LOCALE_NUMBER", 0, &bFound )
		 && bFound
		 &&	getDWValue( str.m_szBody, &locale )
		)
			return locale;

	return 0;
}


// static 
bool cTranslateDll::GetLocaleNumber( const wchar_t *langFile, DWORD *pLocal )
{
	// Skips not txt files.
	int len = (int)wcslen( langFile );

	if ( _wcsnicmp( langFile + len - 4, L".txt", 4 ) != 0 )
		return false;

	cStringFile csf( langFile );

	if ( csf.m_err )
		return false;

	// "LANGUAGE_LOCALE_NUMBER"="0x407"

	wchar_t bf[ 100 ];

	if ( !csf.FindString( bf, maxLen( bf ), L"\"LANGUAGE_LOCALE_NUMBER\"", true ) )
		return false;

	return getDWValue( bf, pLocal );
}

static bool getClearString( wchar_t *bf, int maxBfLen, const wchar_t *dirtyStr )
{
	cAutoDeallocate < wchar_t > adbf( maxBfLen + 20 );

	// dirtyStr is: '="Deutschland"'
	// bf     will: 'Deutschland'

	int ps = (int)wcscspn( dirtyStr, L"\"" );
	if ( dirtyStr[ ps ] != L'\"' )
		return false;
	
	dirtyStr += ps + 1;

	ps = (int)wcscspn( dirtyStr, L"\"" );
	if ( dirtyStr[ ps ] != L'\"' )
		return false;

	if ( ps > maxBfLen )
		return false;
 
	wcsncpy_s( bf, maxBfLen, dirtyStr, ps );
	bf[ ps ] = L'\0';

	return true;
}

//
//	getNativeName
//		Reads language file and returns it's native name
//	Returns false if native name was not found or the file does not look like language file.
//
//static 
bool cTranslateDll::getNativeName( wchar_t *nativeName, int nameMaxLength, const wchar_t *filePath )
{
	//ODSFunc;

	cStringFile sf( filePath );

	if ( sf.m_err )
		return false;

	cAutoDeallocate < wchar_t > adbf( nameMaxLength + 20 );

	if ( sf.FindString( nativeName, nameMaxLength, L"// NativeName = " ) )
		return true;
	else
	if ( sf.FindString( adbf.m_ptr, adbf.m_len, L"\"LANGUAGE_NAME_NEUTRAL\"" ) &&
		getClearString( nativeName, nameMaxLength, adbf.m_ptr ) )
			return true;
	else
	if ( sf.FindString( adbf.m_ptr, adbf.m_len, L"\"LANGUAGE_NAME_NATIVE\"" ) &&
		getClearString( nativeName, nameMaxLength, adbf.m_ptr ) )
			return true;

	return false;
}

//
//	enumLangFiles
//		Walks subfolder and open langFile.txt or setupLangFile.txt,
//		See II. Installation. and III.5 Enumeration at !TranslationStrategy.txt in ..\Includes.
//
//static 
bool cTranslateDll::enumLangFiles( cStringArrayT *pathT, cStringArrayT *nativeNameT, const wchar_t *folder, const wchar_t *exeName )
{
	//ODSFunc;

	wchar_t path[ MAX_PATH ];
	wcscpy_s( path, folder );

	int len = (int)wcslen( path );
	if ( path[ len - 1 ] != L'\\' )
		path[ len++ ] = L'\\';

	BOOL exists = false;

	if ( exeName != NULL )
	{
		FormatT( path + len, maxLen( path ) - len, L"%s.txt", exeName );
		exists = FileExistsT( path );
	}

	//if ( !exists )
	//{
	//	FormatT( path + len, maxLen( path ) - len, L"%s.txt", L"setup.exe" );
	//	exists = FileExistsT( path );
	//}

	if ( !exists )
	{
		FormatT( path + len, maxLen( path ) - len, L"%s.txt", L"langFile" );
		exists = FileExistsT( path );
	}

	wchar_t nativeName[ 256 ];

	if ( exists && getNativeName( nativeName, maxLen( nativeName ), path ) )
	{
		pathT->addString( path );

		if ( nativeNameT != NULL )
			nativeNameT->addString( nativeName );

		// return true;
	}
	
	//
	// Enumerates subdirectories
	//

	wcscpy_s( path + len, maxLen( path ) - len, L"*" );

	WIN32_FIND_DATAW ffd;
	HANDLE hFind = FindFirstFileT( path, &ffd );

	if ( hFind == INVALID_HANDLE_VALUE )
		return false;

	do
	{
		if ( (( ffd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY ) == 0 ) ||
			 ( wcscmp( ffd.cFileName, L".." ) == 0 ) ||
		     ( wcscmp( ffd.cFileName, L"." ) == 0 ) )
				continue;

		wcscpy_s( path + len, maxLen( path ) - len, ffd.cFileName );

		enumLangFiles( pathT, nativeNameT, path, exeName );

		if ( pathT->Q >= maxLangs )
			break;
	}
	while( FindNextFileT( hFind, &ffd ) );

	FindClose( hFind );

	return pathT->Q > 0;
}


//
//	See 'IV. Tranlated help file.' section at V:\Includes\!TranslationStrategy.txt
//
bool cTranslateDll::GetHelpFileName( wchar_t *path, int pathLen ) // , bool bContext )
{
	if ( !m_wcsLangFilePath.isEmpty() )
	{
		wcsncpy_s( path, pathLen, m_wcsLangFilePath.m_szBody, _TRUNCATE );
		int pos = splitPath( path );

		if ( pos <= 0 )
			return false;

		if ( !ReadHelpFileName( path + pos, pathLen - pos ) )
			return false;

		return FileExistsT( path ) == TRUE;
	}

	return false;
}

bool UpdateTranslatorInfo( HWND hDlg, int idGrp, int idName, int idEMail, int idHTTP )
{
	HWND hGroup = GetDlgItem( hDlg, idGrp );

	if ( hGroup == NULL )
		return false;

	wchar_t bf[ 256 ];

	bool found = false;

	if ( ( pTranslateDll == NULL ) ||
		( pTranslateDll->error ) ||
		( pTranslateDll->Translate( bf, maxLen( bf ), "TRANSLATOR_NAME", 0, &found ) == -1 ) ||
		!found )
	{
		// Truncates dilaog bottom

		ShowWindow( hGroup, SW_HIDE );
		RECT r;
		GetWindowRect( hGroup, &r );
		int bottom = r.top + 3;
		GetWindowRect( hDlg, &r );
		SetWindowPos( hDlg, 0, 0, 0, r.right - r.left, bottom - r.top, SWP_NOMOVE | SWP_NOZORDER ); 

		return false;
	}

	// Sets translator name.
	SetWindowTextT( GetDlgItem( hDlg, idName ), bf ); 

	// Sets translator eMail.
	if ( pTranslateDll->Translate( bf, maxLen( bf ), "TRANSLATOR_EMAIL" ) == -1 ) bf[0] = L'\0';
	SetWindowTextT( GetDlgItem( hDlg, idEMail ), bf ); 

	// Sets translator http.
	if ( pTranslateDll->Translate( bf, maxLen( bf ), "TRANSLATOR_WEB" ) == -1 ) bf[0] = L'\0';
	SetWindowTextT( GetDlgItem( hDlg, idHTTP ), bf ); 

	return true;
}

bool TranslateDialog( HWND hPageDlg )
{
	if ( pTranslateDll == NULL ) 
		return false;

	return pTranslateDll->TranslateDialog( hPageDlg ) != -1;
}

bool TranslatePropSheet( HWND hPropSheet )
{
	if ( !TranslateDialog( hPropSheet ) )
		return false;

	HWND hPage;

#ifndef PropSheet_IndexToHwnd
#define PSM_INDEXTOHWND            (WM_USER + 130)
#define PropSheet_IndexToHwnd(hDlg, i) \
        (HWND)SNDMSG(hDlg, PSM_INDEXTOHWND, (WPARAM)(i), 0)
#endif

	for ( int i = 0; hPage = PropSheet_IndexToHwnd( hPropSheet, i ); i++ )
	{
		char text[ 256 ];
		GetWindowTextA( hPage, text, maxLen( text ) );

		wchar_t textW[ 256 ];
		pTranslateDll->Translate( textW, maxLen( textW ), text );

		PropSheet_SetHeaderTitleT( hPropSheet, i, textW );
	}

	return true;
}

bool TranslatePropPage( HWND hPageDlg )
{
	if ( !TranslateDialog( hPageDlg ) )
		return false;

	HWND hPropSheet = GetParent( hPageDlg );

#ifndef PropSheet_HwndToIndex
#define PSM_HWNDTOINDEX            (WM_USER + 129)
#define PropSheet_HwndToIndex(hDlg, hwnd) \
        (int)SNDMSG(hDlg, PSM_HWNDTOINDEX, (WPARAM)(hwnd), 0)
#endif

	int index = PropSheet_HwndToIndex( hPropSheet, hPageDlg );

	if ( index == -1 )
		return false;

	char text[ 256 ];
	GetWindowTextA( hPageDlg, text, maxLen( text ) );

	wchar_t textW[ 256 ];
	pTranslateDll->Translate( textW, maxLen( textW ), text );

	PropSheet_SetHeaderTitleT( hPropSheet, index, textW );

	return true;
}

typedef int CALLBACK typePropSheetProc( HWND hwndDlg, UINT uMsg, LPARAM lParam );
typePropSheetProc *pspOrig = NULL;

static int CALLBACK translatePropSheetProc(
		HWND hwndDlg,
	    UINT uMsg,
		LPARAM lParam
	)
{
	if ( uMsg == PSCB_INITIALIZED ) 
	{
		if ( pTranslateDll != NULL )
			pTranslateDll->TranslateDialog( hwndDlg, false );

		// PropSheet_Translate( hwndDlg );
	}
	
	if ( pspOrig != NULL )
		return pspOrig( hwndDlg, uMsg, lParam );

	return 0;
}

//
//	New version ( 15.05.06 ) of langfile.dll translates tab control headers, 
//	so the following text does not required.
//
//static int PropSheet_Translate( HWND hwndDlg )
//{
//	if ( pTranslateDll == NULL ) 
//		return 0;
//
//	pTranslateDll->TranslateDialog( hwndDlg );
//
//	return 1;
//
//#define TabCtrl_GetItemA(hwnd, iItem, pitem) (BOOL)SendMessageA((hwnd), TCM_GETITEMA, (WPARAM)(int)(iItem), (LPARAM)(TC_ITEM *)(pitem))
//#define TabCtrl_SetItemW(hwnd, iItem, pitem) (BOOL)SendMessageW((hwnd), TCM_SETITEMW, (WPARAM)(int)(iItem), (LPARAM)(TC_ITEM *)(pitem))
//
//	HWND hTab = PropSheet_GetTabControl( hwndDlg );
//
//	int q = TabCtrl_GetItemCount( hTab );
//	for ( int i = 0; i < q; i++ )
//	{
//		TCITEMA tci;
//		char text[ 256 ];
//
//		memset( &tci, 0, sizeof( tci ) );
//		tci.mask = TCIF_TEXT;
//		tci.pszText = text;
//		tci.cchTextMax = maxLen( text );
//
//		TabCtrl_GetItemA( hTab, i, &tci );
//
//		wchar_t textW[ 256 ];
//		pTranslateDll->Translate( textW, maxLen( textW ), text );
//
//		TCITEMW tciW;
//		memset( &tciW, 0, sizeof( tciW ) );
//		tciW.mask = TCIF_TEXT;
//		tciW.pszText = textW;
//		tciW.cchTextMax = maxLen( textW );
//
//		TabCtrl_SetItemW( hTab, i, &tciW );
//	}
//
//	return q;
//}

int PropSheet_SetHeaderTitleT( HWND hDlg, int index, const wchar_t *lpwszText )
{
	bool isNT = ( GetVersion() < 0x80000000 );

#ifndef PSM_SETHEADERTITLEA
#define PSM_SETHEADERTITLEA     (WM_USER + 125)
#define PSM_SETHEADERTITLEW     (WM_USER + 126)
#endif

	if ( isNT )
		return (int)SendMessageW(hDlg, PSM_SETHEADERTITLEW, (WPARAM)(index), (LPARAM)(lpwszText));

	cMBStr mbs( lpwszText );

	return (int)SendMessageA(hDlg, PSM_SETHEADERTITLEA, (WPARAM)(index), (LPARAM)(mbs.m_szBody));
}

static INT_PTR _PropertySheetT( PROPSHEETHEADERW *lppsph )
{
	if ( !cComCtl32Dll::getStatic().LoadOnce(  ) )
		return -1;

	bool isNT = ( GetVersion() < 0x80000000 );

	if ( isNT )
	{
		typedef INT_PTR WINAPI typePropertySheetW( LPCPROPSHEETHEADERW unnamed1 );

		static typePropertySheetW *procPropertySheetW = (typePropertySheetW *)GetProcAddress( cComCtl32Dll::getStatic().hInst,"PropertySheetW" );
		if ( procPropertySheetW == NULL )
			return -1;

		return procPropertySheetW( lppsph );
	}

	cMBStr mbs( lppsph->pszCaption );

	lppsph->pszCaption = (const wchar_t *)mbs.m_szBody;

	typedef INT_PTR WINAPI typePropertySheetA( LPCPROPSHEETHEADERA unnamed1 );

	static typePropertySheetA *procPropertySheetA = (typePropertySheetA *)GetProcAddress( cComCtl32Dll::getStatic().hInst,"PropertySheetA" );
	if ( procPropertySheetA == NULL )
		return -1;

	return procPropertySheetA( ( const PROPSHEETHEADERA * )lppsph );
}

INT_PTR PropertySheetT( PROPSHEETHEADERW *lppsph )
{
	if ( pTranslateDll == NULL )
		return _PropertySheetT( lppsph );

	if ( lppsph->dwFlags & PSH_USECALLBACK )
		pspOrig = (typePropSheetProc *)lppsph->pfnCallback;
	else
		pspOrig = NULL;

	lppsph->dwFlags |= PSH_USECALLBACK;
	lppsph->pfnCallback = (PFNPROPSHEETCALLBACK)translatePropSheetProc;

	return _PropertySheetT( lppsph );
}
