#pragma once

#ifndef _NoTranslation_

#include <windows.h>
#include "cDll.h"
#include "AutoDeallocate.h"
#include "cSysCriticalSection.h"

//#ifndef _NoRTL_
//#include "FileView.h"
//#endif

#ifndef	LF_LTR

#define	LF_LTR		0
#define	LF_RTL		1
//#define	LF_FLIP		2

#endif

class cStringArrayT;

extern const wchar_t *szLangFileTxtName;
extern const wchar_t *szTranslation;
extern const wchar_t *szTranslationProduct; // = NULL;


class cTranslateDll : public CDll
{
public:

	// Do not use the in global constructor, and avoid global contruction at all.
	static __declspec(noinline) cTranslateDll &getStatic()
	{
		static cSysCriticalSection cs;
		SYS_CRITICAL_SECTION_BLOCK( &cs );

		static cTranslateDll dll;
		return dll;
	}

	bool m_flipDialog;
	DWORD m_hDict;

	//
	//	Loads( or reload ) new lang file.
	//	If langFile parameter is NULL it simple resets translation.
	//	
	bool LoadLang( const wchar_t *langFile );

	//
	//	Checks if the resource file mantioned as resource file
	//	// RCFile "..\BCResident\BCResident.rc"
	//
	bool CheckResourceFile( const wchar_t *rcFileName, bool turnTranslationOffIfError = true );

	int TranslateDialog( HWND hDlg, bool disableFlip = true );

	int TranslateMenu( HMENU hMenu );
	
	//
	//	Translates ACP string to wide char string
	//
	int Translate( wchar_t *wsOut, int wsMaxLen, const char *szMbs, int keysize = 0, bool *pFound = NULL, bool bCehckParams = true );
	int Translate( char *mbsOut, int mbsMaxLen, const char *szMbs, int keySize = 0 );

	//
	//	Usualy LoadString does not load the last zero for double zero terminated strings.
	//	So we patch the result string and the return value.
	//	Returns zero if resource is not found, 
	//	otherwise returns number of bytes copied to the buffer ( instead of usual LoadString does not include zero byte ).
	//	
	static int LoadRCString( char *mbs, int size, HINSTANCE rc_hInst, int rc_ids )
	{
		int res = LoadStringA( rc_hInst, rc_ids, mbs, size );

		if ( res == 0 )
			return res;

		bool doubleZeroTerminated = res > ( int )strlen( mbs );

		if ( res < size )
		{
			mbs[ res++ ] = '\0';
		}

		if ( doubleZeroTerminated && ( res < size ) )
		{
			if ( mbs[ res - 2 ] != '\0' )
				mbs[ res++ ] = '\0';
		}

		return res;
	}

	//
	//	Loads resource string by identifier and translates it to wide char string
	//
	int Translate( wchar_t *wsOut, int wsMaxLen, HINSTANCE rc_hInst, int rc_ids, bool bCehckParams = true  )
	{
		char mbs[ 2048 ];
		int res = LoadRCString( mbs, sizeof( mbs ), rc_hInst, rc_ids );
		if ( res == 0 )
		{
			if ( wsMaxLen != 0 ) *wsOut = L'\0';
			return 0;
		}		

		bool found;

		return Translate( wsOut, wsMaxLen, mbs, res, &found, bCehckParams );
	}

	//
	//	Loads resource string by identifier and translates it to wide char string
	//
	int Translate( char *mbsOut, int mbsMaxLen, HINSTANCE rc_hInst, int rc_ids )
	{
		char mbs[ 2048 ];
		int res = LoadRCString( mbs, sizeof( mbs ), rc_hInst, rc_ids );
		if ( res == 0 )
		{
			if ( mbsMaxLen != 0 ) *mbsOut = '\0';
			return 0;
		}		

		return Translate( mbsOut, mbsMaxLen, mbs, res );
	}


#ifdef _UnicodeFuncs_

	cWCStr m_wcsLangFilePath;

	//
	//	If help module was loaded once it keeps name of help file.
	//	We have to reload the module if help file path name is changed.
	//
	void ReloadHelpModule();

	//
	//	Reads assigned help file name from lang file and adds to langFile home directory.
	//
	bool GetHelpFileName( wchar_t *path, int pathLen );
	
	//
	//	Reads corresponded string with help file name from current langFile.txt
	//
	bool ReadHelpFileName( wchar_t *path, int pathLen );

	//
	//	ReadHelpFileDownloadPath
	//		Reads the HELP_FILE_WEB_SITE section at langfile.txt
	//
	bool ReadHelpFileDownloadPath( wchar_t *path, int pathLen );

	//	See Includes\!TranslationStrategy.txt

	//
	//	Set Local information by module name
	//
	bool SetDefault( HINSTANCE hDllInst = NULL );

	//
	//	LinkTo
	//	 Allows to use this translate dll by client instance
	//	 F.e BestCrypt uses cTranslateDll::LinkTo( kgDll.hInst );
	//	 The client instance has to export following function:
	//
	//		extern "C" __declspec( dllexport )
	//			void LinkTranslateDll( cTranslateDll *pDll )
	//		{
	//			SetTranslateDll( pDll );
	//		}
	//	
	bool LinkTo( HINSTANCE hInst );

	typedef int WINAPI type_lfTranslateStrW(DWORD hDict, wchar_t *wbuf, int bufSize, const char *key, int keySize );
	type_lfTranslateStrW *m_proc_lfTranslateStrW;

	typedef int WINAPI type_lfTryTranslateStrW(DWORD hDict, wchar_t *wbuf, int bufSize, const char *key, int keySize );
	type_lfTryTranslateStrW *m_proc_lfTryTranslateStrW;

	typedef int WINAPI type_lfTranslateDialogW(DWORD hDict, HWND hDlg);
	type_lfTranslateDialogW *m_proc_lfTranslateDialogW;

	typedef int WINAPI type_lfTranslateMenuW(DWORD hDict, HMENU hMenu);
	type_lfTranslateMenuW *m_proc_lfTranslateMenuW;

	typedef void WINAPI type_lfReset(DWORD hDict);
	type_lfReset *m_proc_lfReset;

	typedef bool WINAPI type_lfLoadLang(const wchar_t *p, DWORD *phDict);
	type_lfLoadLang *m_proc_lfLoadLang;

	typedef int WINAPI type_lfSetDialogLayout(HWND hDlg, int layout);
	type_lfSetDialogLayout *m_proc_lfSetDialogLayout;

	typedef int WINAPI type_lfSetTextDirection(HWND hWnd, int layout);
	type_lfSetTextDirection *m_proc_lfSetTextDirection;

	bool LoadLangFileDll(
		const wchar_t *langFileFolder = NULL	// Folder ( or subfolder ) where langFile.dll may be found
	);
	
	bool InitAndLoad( const wchar_t *path );

	enum { undefinedDictionary = -1 };

	cTranslateDll() : CDll()
	{
		m_hDict = undefinedDictionary;

		m_proc_lfTranslateStrW = NULL;
		m_proc_lfTryTranslateStrW = NULL;
		m_proc_lfTranslateDialogW = NULL;
		m_proc_lfTranslateMenuW = NULL;
		m_proc_lfReset = NULL;
		m_proc_lfLoadLang = NULL;
		m_proc_lfSetDialogLayout = NULL;

		if ( error )
			return;
	}

	void Reset()
	{
		if ( m_hDict != undefinedDictionary )
		{
			m_proc_lfReset( m_hDict );
			m_hDict = undefinedDictionary;
		}
	}

	~cTranslateDll()
	{
		Reset();
	}

	bool SetDefault(
			const wchar_t *langFileFolder,	// Folder ( or subfolder ) where langFile.dll may be found
			const wchar_t *localSectionName,// Registry section with 'local' DWORD value
			const wchar_t *utilityName,		// Name of utility to be translated
			bool linkToApp = true
		);

	static bool GetLangFilePath( 
			const wchar_t *localSectionName,// Registry section with 'local' DWORD value
			const wchar_t *utilityName,		// Name of utility to be translated
			wchar_t *path					// Output result( MAX_PATH value )
		);

	DWORD GetLocaleNumber();
	static bool GetLocaleNumber( const wchar_t *langFile, DWORD *pLocal );
	static bool makeClearName( wchar_t *unOut, const wchar_t *unIn = NULL );

	//
	// Gets and sets folder name from/to registry with current translation.
	// If utilityName == NULL it uses name of current executable module without path and extension.
	//
	static bool regGetTranslationFolder( HKEY root, wchar_t *folderName, const wchar_t *utilityName = NULL, const wchar_t *vendorName = L"Jetico" );
	static bool regGetTranslationFolder( wchar_t *folder, const wchar_t *utilityName = NULL );
	static bool regSetTranslationFolder( HKEY root, const wchar_t *folderName, const wchar_t *utilityName = NULL, const wchar_t *vendorName = L"Jetico" );

	// Finds translation root folder for corresponded module.
	static bool getModuleTranslationFolder( wchar_t *folderName, const wchar_t *modulePath = NULL );

	static bool getNativeName( wchar_t *nativeName, int nameMaxLength, const wchar_t *filePath );

	enum { maxLangs = 100 };
	static bool enumLangFiles( cStringArrayT *pathT, cStringArrayT *nativeNameT, const wchar_t *folder, const wchar_t *exeName = NULL );

#endif // _UnicodeFuncs_
};

//extern cTranslateDll translateDll;

//
//	Sample of using.
//
//#include "..\Includes\TranslateDll.h"
//cTranslateDll translateDll;
//
//{
//	// Set translate functions
//	wchar_t path[ MAX_PATH ];
//	wcscpy_s( path, app.SrcPath );
//	int len = wcslen( path );
//
//	wcsncpy( path + len, L"langfile.dll", maxLen( path ) - len );
//	translateDll.Load( true, path );
//
//	wcsncpy( path + len, L"langfile.txt", maxLen( path ) - len );
//	if ( translateDll.LoadLang( path ) )
//		SetTranslateDll( &translateDll );
//}

extern cTranslateDll *pTranslateDll;

//
//	SetTranslateDll
//		1. Sets translation for resource loading functions, see LoadLibraryT and LoadLibraryTA
//		2. cDlg WM_INITDIALOG message at DlgProc and DlgPageProc uses the pTranslateDll pointer
//			for translation of dialog items.
//		if the dll paramater is NULL the travslation is switched off.
//
void SetTranslateDll( cTranslateDll *dll );

bool TranslateDialog( HWND hPageDlg );
bool TranslatePropPage( HWND hPageDlg );
bool TranslatePropSheet( HWND hPropSheet );

//
//	UpdateTranslatorInfo
//		If translateDll is not loaded and traslator info is not assigned, it truncates dlg tail down to idGrp
//		Otherwise about dlg translator info from langfile.txt
//
bool UpdateTranslatorInfo( HWND hDlg, int idGrp, int idName, int idEMail, int idHTTP );

#endif // #ifndef _NoTranslation_
