#include "jpegread.h"

#include <stdio.h>

extern "C"
{
#include "jpeglib.h"
#include "memmgr.h"

int jpeg_memory_src (j_decompress_ptr cinfo, unsigned char *buffer, int size );
}

#include <memory.h>
#include <malloc.h>
#include <setjmp.h>

struct my_error_mgr {
  struct jpeg_error_mgr pub;	/* "public" fields */

  jmp_buf setjmp_buffer;	/* for return to caller */
  bool valid;
} srcerr;

typedef struct my_error_mgr * my_error_ptr;


METHODDEF(void) my_error_exit (j_common_ptr cinfo)
{
	/* cinfo->err really points to a my_error_mgr struct, so coerce pointer */
	my_error_ptr myerr = (my_error_ptr) cinfo->err;

	myerr->pub.num_warnings++;

	char buffer[JMSG_LENGTH_MAX];

	/* Create the message */
	(*cinfo->err->format_message) (cinfo, buffer);

	/* Always display the message. */
	//MessageBox(NULL,buffer,"JPEG Fatal Error",MB_ICONSTOP);


	if ( !myerr->valid )	return;
	/* Return control to the setjmp point */
	longjmp(myerr->setjmp_buffer, 1);
	myerr->valid = false;
}


cJPegRead::cJPegRead( unsigned char *buffer, int size )
{
	memset( this, 0, sizeof( *this ) );

	cinfo = (void *) malloc( sizeof( jpeg_decompress_struct ) );
	if ( cinfo == NULL )
		return;

	jpeg_decompress_struct *srcinfo = (jpeg_decompress_struct *)cinfo;

	jpeg_create_decompress(srcinfo);

	/* Specify data source for decompression */
	jpeg_memory_src( srcinfo, buffer, size );

	srcinfo->err = jpeg_std_error(&srcerr.pub);
	srcerr.pub.error_exit = my_error_exit;

	srcerr.valid = true;

	// Establish the setjmp return context for my_error_exit to use.
	if (setjmp(srcerr.setjmp_buffer))
	{
		srcerr.valid = false;
		err = -1;
		//
		//	Jpeg lib does not return if a error occuares
		//	So the jump is used.
		//
		return;
	}

	/* Read file header */
	(void) jpeg_read_header( srcinfo, TRUE );

	srcerr.valid = false;

	err = srcerr.pub.num_warnings;
	if ( err != 0 )
		return;

	jpeg_start_decompress( srcinfo );

	/* All done. */
	err = srcerr.pub.num_warnings;

	height = srcinfo->output_height;
	width  = srcinfo->output_width;
	bytesPerPixel = srcinfo->output_components * sizeof( JSAMPLE );
}

cJPegRead::~cJPegRead()
{
	if ( cinfo != NULL )
	{
		jpeg_decompress_struct *srcinfo = (jpeg_decompress_struct *)cinfo;
		
		if ( err == 0 )
			jpeg_finish_decompress( srcinfo );

		jpeg_destroy_decompress( srcinfo );

		free( cinfo );
	}
}

class cRowTab
{
public:
	JSAMPROW *tab;
	cRowTab( unsigned char *bf, int rowSizeInBytes, int rowQ )
	{
		tab = new JSAMPROW[ rowQ ];
		for( int i = 0; i < rowQ; i++ )
		{
			tab[i] = bf + i * rowSizeInBytes;
		}
	}

	~cRowTab()
	{
		if ( tab )
			delete tab;
	}
};

int cJPegRead::ScanLines( unsigned char *dst, int lineQ, int byteWidth )
{
	//
	//	The array can be word alligned so the byteWidth is the alligned width.
	//
	if ( byteWidth == 0 )
		byteWidth = width * bytesPerPixel;
	else
	if ( byteWidth < width * bytesPerPixel )
		return 0;

	cRowTab rTab( dst, byteWidth, lineQ );

	jpeg_decompress_struct *srcinfo = (jpeg_decompress_struct *)cinfo;

	int pos = 0;

	while ( srcinfo->output_scanline < srcinfo->output_height )
	{
		int lines =  jpeg_read_scanlines( srcinfo, &(rTab.tab[ pos ]), lineQ - pos );
		pos += lines;
	}

	return pos;
}

int cJPegRead::SwapRGB( unsigned char *buffer, int byteWidth, int height )
{
	int h,w;
	unsigned char *ptr, ch;

	for( h = 0; h < height; h++ )
		for ( w = 0; w + 3 <= byteWidth; w+=3 )
		{
			ptr = buffer + h * byteWidth + w;
			ch = ptr[0];
			ptr[0] = ptr[2];
			ptr[2] = ch;
		}
}