#include "memmgr.h"

typedef struct mem_src_mgr {
  struct jpeg_source_mgr pub;	/* public fields */

  JOCTET * buffer;		/* start of buffer */
  int bfSize;
  int curPos;
} *mem_mgr_ptr;

void mem_init_source( j_decompress_ptr cinfo )
{
  mem_mgr_ptr src = (mem_mgr_ptr) cinfo->src;

  src->curPos = 0;
}

boolean mem_fill_input_buffer(j_decompress_ptr cinfo)
{
  mem_mgr_ptr src = (mem_mgr_ptr) cinfo->src;

  src->pub.next_input_byte = src->buffer;
  src->pub.bytes_in_buffer = src->bfSize;
  src->curPos = src->bfSize;

  return TRUE;
}

void mem_term_source(j_decompress_ptr cinfo){}
void mem_skip_input_data(j_decompress_ptr cinfo, long size ){}


int jpeg_memory_src (j_decompress_ptr cinfo, unsigned char *buffer, int size )
{
	mem_mgr_ptr src;

	/* The source object and input buffer are made permanent so that a series
	* of JPEG images can be read from the same file by calling jpeg_stdio_src
	* only before the first one.  (If we discarded the buffer at the end of
	* one image, we'd likely lose the start of the next one.)
	* This makes it unsafe to use this manager and a different source
	* manager serially with the same JPEG object.  Caveat programmer.
	*/
	if (cinfo->src == NULL) {	/* first time for this JPEG object? */
		cinfo->src = (struct jpeg_source_mgr *)
		(*cinfo->mem->alloc_small) ((j_common_ptr) cinfo, JPOOL_PERMANENT,
				  sizeof( struct mem_src_mgr ));
	}

	src = (mem_mgr_ptr) cinfo->src;
	src->pub.init_source = mem_init_source;
	src->pub.fill_input_buffer = mem_fill_input_buffer;
	src->pub.skip_input_data = mem_skip_input_data;
	src->pub.resync_to_restart = jpeg_resync_to_restart; /* use default method */
	src->pub.term_source = mem_term_source;

	src->buffer = buffer;
	src->bfSize = size;
	src->curPos = 0;

	src->pub.bytes_in_buffer = 0; /* forces fill_input_buffer on first read */
	src->pub.next_input_byte = NULL; /* until buffer loaded */

	return 0;
}

