///////////////////////////////////
//
//	Copyright (c) 1994-2012 Jetico, Inc., Finland
//	All rights reserved.
//
//	File: cCertificateLibEncryptionKey.h
//	Revision: 
//	Created:	22.04.2012
//	Description:
//		It allows to create BestCrypt key handle without driver.
//		It creates the key in certificate lib format
//		The key handle is saved at application based so it may be used whithin the application is alive or the handle not be freed.
//
//////////////////////////////////
#pragma once

#include "..\..\..\CertificateLib\symmetricAlgs\symmetricAlgs.h"
#include "..\..\..\Includes\AutoDeallocate.h"
#include "..\..\..\Includes\UnicodeFuncs.h"
#include "Kblock.h"

#pragma pack(1)

class cCertificateLibEncryptionKey
{
public:
	HSA m_hsa;
	BYTE m_pool[ DB_PoolSize ];
	DWORD m_encModeId;

	DWORD m_keyID; // unique key identifier.
	
	cCertificateLibEncryptionKey()
	{
		m_hsa = NULL;
		memset( m_pool, 0, sizeof( m_pool ) );
		m_keyID = -1;
		m_encModeId = -1;
	}
	
	~cCertificateLibEncryptionKey()
	{
		free();
	}

	bool create( WORD algId, DWORD encModeId, BYTE *keyBuff, int keyBuffSize, BYTE *pool )
	{
		int res = sa_createAlgorithm( &m_hsa, algId, keyBuff, keyBuffSize );

		if ( res != sae_ok )
			return false;

		memcpy( m_pool, pool, sizeof( m_pool ) );
		
		m_encModeId = encModeId;

		return true;
	}

	void free()
	{
		memset( m_pool, 0, sizeof( m_pool ) );
		m_encModeId = 0;

		if ( m_hsa != NULL )
			sa_free( m_hsa );
		
		m_hsa = NULL;
	}
};

#pragma pack()

class cCertificateLibEncryptionKeyManager
{
public:
	static cADBlock < cCertificateLibEncryptionKey > m_keyT;
	static DWORD m_lastId;
	static HANDLE m_hSemaphore;

	static bool AddToTable( cCertificateLibEncryptionKey *alg );
	static cCertificateLibEncryptionKey *FindKeyById( DWORD keyId, bool bDelete = false );
	static void RemoveFromTable( DWORD keyId ) { FindKeyById( keyId, true ); }

	static void CheckSemaphore();
	static void ReleaseSemaphore();
};