/*******************************************************************\
*
*	Copyright (c) 2007-2015 Jetico, Inc., Finland
*	All rights reserved.
*
*	File: kblock.h
*	Revision: 
*	Created:
*	Description: declaration of DATA_BLOCK structure and 
*		procedures to work with Data Block that stores 
*		encryption keys
*
\*******************************************************************/

#pragma once

#include <windows.h>
#include <assert.h>
#include "..\..\..\CertificateLib\SMemory\SBlocks.h"
#include "..\..\..\CertificateLib\hashAlgs\hashAlgs.h"
#include "..\..\Comm_def\Kg_defs.h"
#include "dbHeaderBase.h"
#include "kgParams.h"
#include "ALG.h"
#include "ODS.h"
#ifdef _ODS
#include "UnicodeFuncs.h"
#endif

BOOL CalculateDigest( DWORD hashAlgID, BYTE *digest, int digestSize, BYTE *buff, int buffSize, BYTE *buff_1 = NULL, int buffSize_1 = 0 );
BOOL VerifyDigest( DWORD hashAlgID, BYTE *digest, BYTE *buff, int buffSize, BYTE *buff_1 = NULL, int buffSize_1 = 0 );

//
//	verifyDigest
//		If pHashAlgID is NULL or if *pHashAlgID is invalid ( f.e. -1 )
//		It enumerates all available hash algorithms and and tries to verify.
//
bool verifyDigest( DWORD *pHashAlgID, BYTE *digest, BYTE *bf, DWORD bfSize );

/*******************************************************
*
*	Structures that describe format of the Data block 
*	used in the CreateKeyHandle(). The DataBlock contains
*	information in encrypted form and is stored inside 
*	file-container
*
\*******************************************************/

#define _4K ( 4 * 1024 )
#define DRoundUp( value, blockSize ) ( ( (value) +  (blockSize) - 1 ) & ~((blockSize) - 1) )
#define DRoundDown( value, blockSize ) ( (value) & ~((blockSize) - 1) )
#define DRound( value, blockSize ) DRoundUp( value, blockSize )
#define DRoundTail( value, blockSize ) ( DRound( (value), blockSize ) - (value) )

#define DB_FormatVersion_2			  2
#define DB_FormatVersion			  3
#define DB_MinSize				    _4K

//
//	DB_MaxSize is completely calkulated at dbHeader size.
//
// Rest from all requred fields is busied by key map table 
// The size == 
//			max number of key map table items * DB_MinEncryptedKeyBlockSize 
//			DB_HeaderSizeWithKeyMap +
//			DB_PoolSize
//	= 5 * 4K = 20K
//#define DB_MaxSize	DRound( DB_maxKeyQ * DB_MinEncryptedKeyBlockSize + DB_HeaderSizeWithKeyMap + DB_PoolSize, _4K )

#define DB_KeyMapSize				512
#define DB_HeaderSizeWithKeyMap		1024
#define DB_PoolSize					512 
#define DB_MinEncryptedKeyBlockSize	256 // Minimal step for key block ckecking
#define DB_MaxEncBlockSize			16	// Maximal encryption block size

#define	MAXIMUM_KEY_SIZE_BYTES			64
#define	MAXIMUM_DIGEST_SIZE_BYTES		32
#define	MAXIMUM_DIGEST_SIZE_BYTES_64	64

#define SALT_SIZE					8
#define DEFAULT_SALT_SLOT			0	// Default slot for KEYGEN_ID_V5 salt see DATA_BLOCK::updateSalt, if header is not encrypted it will be skipped automatically,
										// but encrypted header has to skip it for first step.

#define DB_minKeyQ (( DB_MinSize -  DB_HeaderSizeWithKeyMap - DB_PoolSize ) / DB_MinEncryptedKeyBlockSize ) // ( 4k - 3 * 512 ) / 256 == 10 -> 10 keys may be stored at minimal key block
#define DB_maxKeyQ ( DB_KeyMapSize / sizeof( dbKeyId ) ) // 512 / 8 == 64	-> 64 minimal keys may be added to container ( maximal data block size is 5 * 4K ).

//
//	We have to emulate the initial part of previose hidden sector
//	because:
//	1. Guard utility detects the OEMid ( 'LOCOS94 ' ) field to work with conatiner
//	2. Old versions of BestCrypt have to definitely do not work with the new container
//		otherwise they may damage any fields ( desctyprtion, busied name, file system ID, and etc. ).
//		If the volume label field will be changed form 'CRYPTED_DSK' ( 11 symbols ) to key gen name 'BC_KeyGenID'
//
#include "BootRecord.h"

#pragma pack(1)

//
//	Data block header
//		if data block is not encrypted the volumeLabel field is 'BC_KeyGenID',
//		otherwise the field and all others excluding the containerID field
//		are completely filled by random data
//

#define dbSignature "LOCOS94 "
#define dbVolumeLabel "BC_KeyGenID"

struct DATA_BLOCK;

static const DWORD maxHPBlockSize = _4K;

//struct sSpaceUnit : public sSpaceUnitBase
//{
//};

//
//	dbContInfo
//
struct dbContInfo : public sSpaceUnitBase
{
	static bool isForHiddenPart( sSpaceUnitBase *contInfo )
	{
		return contInfo->length == maxHPBlockSize;
	}

	static bool isValid( sSpaceUnitBase *contInfo )
	{
		return contInfo->length >= maxHPBlockSize;
	}
};

struct dbHeader : public dbHeaderBase
{
	bool setKeyGenIdV5( DWORD count );
	static DWORD getIterationCount( const dbHeaderBase &dbh );
	void setVersion( WORD ver );
	int getVersion();

	//
	//	Functions.
	//

	void InitFirst();
	void Init();


	//
	//	Marks as empty if type == kbtEmpty, otherwise marks as busied
	//
	bool MarkBusied( int pos, int q, ekbType type, WORD size = 0 ) { return MarkBusied( keyMap, pos, q, type, size ); }
	static bool MarkBusied( dbKeyId *kmT, int pos, int q, ekbType type, WORD size = 0 );
	static bool MarkBusied( dbKeyId *kmT, sKeyMapInfo *kmi ){ return MarkBusied( kmT, kmi->position, kmi->quantity, (ekbType)kmi->kbType, kmi->kbSize ); 	}
	static bool MarkEmpty( dbKeyId *kmT, int pos, int q )	{ return MarkBusied( kmT, pos, q, kbtEmpty, 0 ); 	}

	//
	//	Updates internal or external key map
	//
	static
	int updateKeyMap( dbKeyId *extKMT, const dbKeyId *keyMap );//, bool internal = true );

	int getLastEmptyChainStartPos();

	//
	//	isEncrypted
	//		The function allows ambiguity so we have to check if cont 
	//		info is really encrypted after encryption!!!
	//
	static bool isEncrypted( const dbHeaderBase &dbh );

	static bool isValidFileSize( DWORDLONG fileSize, const dbHeaderBase &dbh )
	{
		return  ( fileSize >= dbh.contInfo.offset + dbh.contInfo.length ) &&
			    ( fileSize <= dbh.contInfo.offset + dbh.contInfo.length + DB_MaxSize );
	}


	//
	//	If header is encrypted returns -1
	//
	DWORD getRealSize();

	static DWORD getContAlgId( dbHeaderBase *dbh ) { return ( isEncrypted(*dbh) ? -1 : dbh->contAlgId ); }

	//
	//	If header is not encrypted so some it's fields may be changed after digest
	//	is updates, so we have to fill by zero the service fields
	//
	bool whiteServiceInfo( cPlainTextBlock *ptb );

	//
	//	findEmptySpace
	//
	static int findEmptySpace( dbKeyId *kmT, int sizeRequired, int first = 0, int maxQ = DB_maxKeyQ, bool bFromTail = false );

	//
	//	dbGetMaxEmptyKeyBlockSpace
	//		returns position of max space found
	//		if treatAsFree is not -1 the chain will be free later
	//
	static int getMaxEmptyKeys( dbKeyId *keyMap, int keyQ, int *pMaxFree, int meanAsFree = -1 );
	static DWORD checkFreeSpace( dbKeyId *keyMap, DWORD	requiredSize, DWORD *pMaxFreeSize, int treatAsFree = -1 );

	//
	//	freeKeyBlockChain
	//		1. Shreds data block at specified position in key map info (kmi).
	//		2. Mark as empty the sredded place at key map table (map)
	//
	//	Note. It does not update dbHeader at DATA_BLOCK. It has to be updated later.
	//	
	static void freeKeyBlockChain( DATA_BLOCK *dataBlock, DWORD dbSize, dbKeyId *map, sKeyMapInfo *kmi );
	static int  freeKeyBlockChain( DATA_BLOCK *dataBlock, DWORD dbSize, dbKeyId *map, int pos );

	//  
	//	shredKeyTab
	//		Overwrites the ketTab item only at specified position in key map info (kmi).
	//		The header is not modified so it is not required to be updated.
	//
	static bool shredKeyTab( DATA_BLOCK *dataBlock, DWORD dbSize, sKeyMapInfo *kmi );
};


//
//	sKeyInfo
//		the structure is used as block for encoding and decoding key
//		by different key block type PBE, PKE or SSS
//
struct sKeyInfo
{
	BYTE	size;			// real size of the key info in bytes
	BYTE	hashAlgId;		// Digest algorithm used for key block verification ( same as pgp id ( f.e. pgphaSHA1 or bchaSHA256 see pktdefs.h )
	DWORD	keyAlgId;		// Algorithm is used for encryption of header, contInfo, keyMap and container body
	DWORD	encModeId;		// The mode is used for encryption for encryption algorithm above

	sSpaceUnitBase contInfo;		// Information about container

	enum { sizeWithoutKey = 10 + sizeof( sSpaceUnitBase ) };

	BYTE	key [ MAXIMUM_KEY_SIZE_BYTES ];

	//
	//	isValid
	//		simple size check only
	//		the key is completely checked by it's implemeting to container info ( see sSpaceUnit )
	//
	bool isValid();

	// Gets size of key depended on algorithm id only
	static int getAlgKeySizeBytes( int algId, DWORD encModeId );
	int getAlgKeySizeBytes();

	// Gets the structure buffer and key size
	static int getBufferSize( int keySize );
	int getKeySizeBytes();

	//
	//	See bdHeader::whiteServiceInfo description
	//
	bool whiteServiceInfo( cPlainTextBlock *ptb )
	{
		// the sKeyInfo struct does not have service fields.
		return false;
	}
};

//
//	sVerificationBlock
//		empty block is used for verification of tcEncBlock < sVerificationBlock >
//
struct sVerificationBlock
{
	WORD sig;

	//
	//	For any case it saves here algorithm values( same as in sKeyInfo ). 
	//
	BYTE	hashAlgId;
	DWORD	keyAlgId;
	DWORD	encModeId;

	sVerificationBlock()
	{
		memset( this, 0, sizeof( *this ) );
		sig = 0xBC08;
	}

	bool check()
	{
		return sig == 0xBC08;
	}

	bool whiteServiceInfo( cPlainTextBlock *ptb ){ return false; }
};

//
//	struct EBTail encryption block tail
//
struct EBTail32 // 32 bytes digest.
{
	BYTE	m_digest[ MAXIMUM_DIGEST_SIZE_BYTES ];
	BYTE	m_iv[ IVECTOR_LENGTH_DB ];

	bool verifyDigest( DWORD *pHashAlgID, BYTE *bf, DWORD bfSize )
	{
		return ::verifyDigest( pHashAlgID, m_digest, bf, bfSize );
	}

	bool calculateDigest( DWORD	hashAlgID, BYTE *bf, DWORD bfSize )
	{
		return CalculateDigest( hashAlgID, m_digest, sizeof( m_digest ), bf, bfSize ) == TRUE;
	}
};

struct EBTail64 // 64 bytes digest.
{
	BYTE	m_digest[ 64 ];
	BYTE	m_iv[ IVECTOR_LENGTH_DB ];

	bool verifyDigest( DWORD *pHashAlgID, BYTE *bf, DWORD bfSize )
	{
		return ::verifyDigest( pHashAlgID, m_digest, bf, bfSize );
	}

	bool calculateDigest( DWORD	hashAlgID, BYTE *bf, DWORD bfSize )
	{
		return CalculateDigest( hashAlgID, m_digest, sizeof( m_digest ), bf, bfSize ) == TRUE;
	}
};

static inline bool encryptIV( BYTE *iv, BYTE *bfOut, const BYTE *bfIn, int bfSize, KEY_HANDLE keyHandle, DWORD algId, DWORD encModeId )
{
	bfSize -= IVECTOR_LENGTH_DB;
	memcpy( bfOut + bfSize, iv, IVECTOR_LENGTH_DB );
	return cSymAlgorithm::encrypt( iv, bfOut, bfIn, bfSize, keyHandle, algId, encModeId );
}

static inline bool decryptIV( BYTE *iv, BYTE *bfOut, const BYTE *bfIn, int bfSize, KEY_HANDLE keyHandle, DWORD algId, DWORD encModeId )
{
	assert( bfIn + bfSize - IVECTOR_LENGTH_DB == iv );
	bfSize -= IVECTOR_LENGTH_DB;
	return cSymAlgorithm::decrypt( iv, bfOut, bfIn, bfSize, keyHandle, algId, encModeId );
}

template < class TBody, class TTail  > 
	bool verifyBodyTail( TBody &body, TTail &tail, DWORD hashAlgId, DWORD algId, DWORD encModeId, KEY_HANDLE keyHandle )
{
	DWORD decBlockSize = sizeof( TTail );

	cPlainTextBlock ptb;
	if ( ptb.Allocate( decBlockSize ) == NULL )
		return false;

	TTail *decBlock = ( TTail * )ptb.buffer;

	if ( !decryptIV( tail.m_iv, (BYTE *)decBlock, (BYTE *)&tail, decBlockSize, keyHandle, algId, encModeId ) )
		return false;

	cPlainTextBlock ptbWhite;

	if ( body.whiteServiceInfo( &ptbWhite ) )
	{
		if ( decBlock->verifyDigest( &hashAlgId, ptbWhite.buffer, sizeof( TBody ) ) )
			return true;
	}
	else
	{
		if ( decBlock->verifyDigest( &hashAlgId, (BYTE *)&body, sizeof( TBody ) ) )
			return true;
	}

	return false;
}

template < class TBody, class TTail  > 
	bool updateBodyTail( TBody &body, TTail &tail, DWORD hashAlgId, DWORD algId, DWORD encModeId, KEY_HANDLE keyHandle )
{
		DWORD encBlockSize = sizeof( tail );

		// IV is shreded here.
		ShredData( (BYTE *)&tail, encBlockSize );

		cPlainTextBlock ptb;
		if ( body.whiteServiceInfo( &ptb ) )
			tail.calculateDigest( hashAlgId, ptb.buffer, sizeof( body ) );
		else
			tail.calculateDigest( hashAlgId, (BYTE *)&body, sizeof( body ) );

		bool res = encryptIV( tail.m_iv, (BYTE *)&tail, (BYTE *)&tail, encBlockSize, keyHandle, algId, encModeId );

#ifdef _DEBUG
		if ( res )
			res = verifyBodyTail( body, tail, hashAlgId, algId, encModeId, keyHandle );
#endif

		return res;
}

const int tail32tail64diff = ( sizeof( EBTail64 ) - sizeof( EBTail32 ) );

//
//	tcEncBlock
//		Template class Encryption Block
//
template < class TT > class tcEncBlock
{
public:
	TT		m_body;

#ifndef SKIP_RESERVED // See description at tcEncBlock constructor.
	BYTE	m_reserved[ DRoundTail( sizeof( TT ), DB_MaxEncBlockSize ) ];
#endif

	union {
		struct {
			EBTail32	m_tail32;
			BYTE		m_tail32empty[ tail32tail64diff ];
		};
		EBTail64	m_tail64;
	};

	int getBlockSize( int digestSize )
	{
		if ( digestSize <= 32 )
			return sizeof( *this ) - tail32tail64diff;
		else
			return sizeof( *this );
	}

	tcEncBlock()
	{
		//
		// The m_reserved field is rounded the class size up to DB_MaxEncBlockSize,
		// but if the rounded size is 0 compiller generates error.
		// So we have to remove the zero size m_reserved field.
		//
		// The assertion checks if a SKIP_RESERVED was not undefend after skipping.
		//
		assert( ( sizeof( *this ) % DB_MaxEncBlockSize ) == 0 );
	}

	~tcEncBlock()
	{
		memset( this, 0, sizeof( *this ) );
	}

	bool calculateDigest( DWORD	hashAlgID, int digestSize = 0 )
	{
		if ( digestSize == 0 )
			if ( ha_getDigestSize( &digestSize, hashAlgID ) != hae_ok )
				return false;

		if ( digestSize == 64 )
			return m_tail64.calculateDigest( hashAlgID, (BYTE*)&m_body, sizeof( m_body ) );
		else
		if ( digestSize <= 32 )
			return m_tail32.calculateDigest( hashAlgID, (BYTE*)&m_body, sizeof( m_body ) );

		return false;
	}

 	bool verifyDigest( DWORD *pHashAlgID )
	{
		int digestSize = 0;
		if ( ha_getDigestSize( &digestSize, *pHashAlgID ) == hae_ok )
		{
			if ( digestSize == 64 )
				return m_tail64.verifyDigest( pHashAlgID, (BYTE *)&m_body, sizeof( m_body ) );
			else
			if ( digestSize <= 32 )
				return m_tail32.verifyDigest( pHashAlgID, (BYTE *)&m_body, sizeof( m_body ) );

			return false;
		}

		if ( m_tail32.verifyDigest( pHashAlgID, (BYTE *)&m_body, sizeof( m_body ) ) )
			return true;

		return m_tail64.verifyDigest( pHashAlgID, (BYTE *)&m_body, sizeof( m_body ) );
	}

	//
	//	Decrypts and verify tail digest
	//
	bool verifyTail( DWORD algId, DWORD encModeId, KEY_HANDLE keyHandle, DWORD hashAlgId )
	{	
		int digestSize;

		if ( ha_getDigestSize( &digestSize, hashAlgId ) != hae_ok )
			return false;

		if ( digestSize == 64 )
			return verifyBodyTail( m_body, m_tail64, hashAlgId, algId, encModeId, keyHandle );
		else
		if ( digestSize <= 32 )
			return verifyBodyTail( m_body, m_tail32, hashAlgId, algId, encModeId, keyHandle );

		return false;
	}

	bool verifyTail( DWORD algId, DWORD encModeId, KEY_HANDLE keyHandle, DWORD *pHashAlgID = NULL )
	{
		DWORD hashAlgId = -1;

		if ( pHashAlgID != NULL )
			hashAlgId  = *pHashAlgID;
		else
			pHashAlgID = &hashAlgId;

		cEnumHashAlgIDs &hashAlgs = cEnumHashAlgIDs::getStatic();

		// Checks validity.
		bool valid = hashAlgs.checkValid( *pHashAlgID );

		if ( valid )
			return verifyTail( algId, encModeId, keyHandle, *pHashAlgID );

		// Enumeration
		for( int i = 0; i < hashAlgs.Q; i++ )
		{
			if ( !verifyTail( algId, encModeId, keyHandle, hashAlgs.T[i] ) )
				continue;

			*pHashAlgID = hashAlgs.T[i];
			return true;
		}

		return false;
	}

	//
	//	Calculates body digest and encrypt tail only
	//
	bool updateTail( DWORD hashAlgId, DWORD algId, DWORD encModeId, KEY_HANDLE keyHandle )
	{
		if ( hashAlgId == -1 )
			verifyTail( algId, encModeId, keyHandle, &hashAlgId );

		int digestSize = 0;
		if ( ha_getDigestSize( &digestSize, hashAlgId ) != hae_ok )
			return false;

		if ( digestSize == 64 )
			return updateBodyTail( m_body, m_tail64, hashAlgId, algId, encModeId, keyHandle );
		else
		if ( digestSize <= 32 )
			return updateBodyTail( m_body, m_tail32, hashAlgId, algId, encModeId, keyHandle );

		return false;
	}

	bool updateTail( sKeyInfo *keyInfo, BYTE *pool )
	{
		cSymAlgorithm cAlg;
		if ( !cAlg.createKey( keyInfo->keyAlgId, keyInfo->encModeId, keyInfo->key, keyInfo->getKeySizeBytes(), pool ) )
			return false;

		return updateTail( keyInfo->hashAlgId, cAlg.m_algId, cAlg.m_encModeId, cAlg.m_keyHandle );
	}

	//
	//	Allows to decrypt same buffer
	//
	bool decryptFrom( tcEncBlock < TT > *ebIn, DWORD algId, DWORD encModeId, KEY_HANDLE keyHandle )
	{
		DWORD decBlockSize = sizeof(*this );

		cPlainTextBlock ptb;
		if ( ptb.Allocate( decBlockSize ) == NULL )
		{
			assert( false );
			return false;
		}

		tcEncBlock< TT > *decBlock = ( tcEncBlock< TT > * )ptb.buffer;

		if ( !decryptIV( ebIn->m_tail32.m_iv, (BYTE *)decBlock, (BYTE *)ebIn, decBlockSize - tail32tail64diff, keyHandle, algId, encModeId ) )
			return false;

		DWORD hashAlgId = -1;

		if ( !decBlock->verifyDigest( &hashAlgId ) )
		{
			if ( !decryptIV( ebIn->m_tail64.m_iv, (BYTE *)decBlock, (BYTE *)ebIn, decBlockSize, keyHandle, algId, encModeId ) )
				return false;

			if ( !decBlock->verifyDigest( &hashAlgId ) )
				return false;
		}

		int digestSize = 0;
		if ( ha_getDigestSize( &digestSize, hashAlgId ) != hae_ok )
			return false;

		memcpy( this, decBlock, decBlock->getBlockSize( digestSize ) );

		return updateTail( hashAlgId, algId, encModeId, keyHandle );
	}

	bool decryptFrom( tcEncBlock< TT > *ebIn, sKeyInfo *keyInfo, BYTE *pool )
	{
		cSymAlgorithm cAlg;

		if ( !cAlg.createKey( keyInfo->keyAlgId, keyInfo->encModeId, keyInfo->key, keyInfo->getKeySizeBytes(), pool ) )
			return false;

		return decryptFrom( ebIn, cAlg.m_algId, cAlg.m_encModeId, cAlg.m_keyHandle );
	}

	bool decryptBodyTo( TT *ttOut, sKeyInfo *keyInfo, BYTE *pool )
	{
		cSymAlgorithm cAlg;
		if ( !cAlg.createKey( keyInfo->keyAlgId, keyInfo->encModeId, keyInfo->key, keyInfo->getKeySizeBytes(), pool ) )
			return false;

		return decryptBodyTo( ttOut, cAlg.m_algId, cAlg.m_encModeId, cAlg.m_keyHandle );
	}

	bool decryptBodyTo( TT *ttOut, DWORD algId, DWORD encModeId, KEY_HANDLE keyHandle )
	{
		DWORD decBlockSize = sizeof(*this );

		cPlainTextBlock ptb;
		if ( ptb.Allocate( decBlockSize ) == NULL )
		{
			assert( false );
			return false;
		}

		tcEncBlock < TT > *blc = ( tcEncBlock < TT > * )ptb.buffer;;

		if ( !blc->decryptFrom( this, algId, encModeId, keyHandle ) )
			return false;

		if ( ttOut != NULL )
			memcpy( ttOut, blc, sizeof( TT ) );

		return true;
	}

	bool encryptTo( tcEncBlock < TT > *ebOut, DWORD algId, DWORD encModeId, DWORD hashAlgId, KEY_HANDLE keyHandle )
	{
		return encrypt( ebOut, &(this->m_body), algId, encModeId, hashAlgId, keyHandle );
	}

	//bool encryptTo( tcEncBlock < TT > *ebOut, DWORD algId, DWORD encModeId, DWORD hashAlgId, KEY_HANDLE keyHandle )
	//{
	//	if ( hashAlgId == -1 )
	//		if ( !verifyTail( algId, encModeId, keyHandle, &hashAlgId ) )
	//			return false;

	//	int digestSize = 0;
	//	if ( ha_getDigestSize( &digestSize, hashAlgId ) != hae_ok )
	//		return false;

	//	DWORD encBlockSize = sizeof( *this );

	//	cPlainTextBlock ptb;
	//	if ( ptb.Allocate( encBlockSize ) == NULL )
	//		return false;

	//	tcEncBlock < TT > *blc = ( tcEncBlock < TT > * )ptb.buffer;

	//	// Shred entirely tail union.
	//	ShredData( (BYTE*)&blc->m_tail64, sizeof( blc->m_tail64 ) );
	//	memcpy( &blc->m_body, &m_body, sizeof( m_body ) );

	//	if ( !blc->calculateDigest( hashAlgId, digestSize ) )
	//		return false;

	//	if ( digestSize == 64 )
	//		return encryptIV( blc->m_tail64.m_iv, (BYTE *)ebOut, (BYTE *)blc, encBlockSize, keyHandle, algId, encModeId );
	//	else
	//	if ( digestSize <= 32 )
	//		//return encryptIV( blc->m_tail32.m_iv, (BYTE *)ebOut, (BYTE *)blc, encBlockSize - tail32tail64diff, keyHandle, algId, encModeId );
	//		return encryptIV( blc->m_tail32.m_iv, (BYTE *)ebOut, (BYTE *)blc, encBlockSize, keyHandle, algId, encModeId );

	//	return false;
	//}

	bool encryptTo( tcEncBlock < TT > *ebOut, sKeyInfo *keyInfo, BYTE *pool )
	{
		cSymAlgorithm cAlg;

		if ( !cAlg.createKey( keyInfo->keyAlgId, keyInfo->encModeId, keyInfo->key, keyInfo->getKeySizeBytes(), pool ) )
			return false;

		return encrypt( ebOut, &(this->m_body), cAlg.m_algId, cAlg.m_encModeId, keyInfo->hashAlgId, cAlg.m_keyHandle );
	}

	bool encryptFromBody( TT *bodyIn, DWORD algId, DWORD encModeId, DWORD hashAlgId, KEY_HANDLE keyHandle )
	{
		return encrypt( this, bodyIn, algId, encModeId, hashAlgId, keyHandle );
	}

	static bool encrypt( tcEncBlock < TT > *ebOut, TT *bodyIn, DWORD algId, DWORD encModeId, DWORD hashAlgId, KEY_HANDLE keyHandle )
	{
		DWORD encBlockSize = sizeof( *ebOut );

		cPlainTextBlock ptb;
		if ( ptb.Allocate( encBlockSize ) == NULL )
			return false;

		ShredData( ptb.buffer + sizeof( TT ), encBlockSize - sizeof( TT ) );

		tcEncBlock < TT > *blc = ( tcEncBlock < TT > * )ptb.buffer;

		memcpy( &blc->m_body, bodyIn, sizeof( TT ) );

		int digestSize = 0;
		if ( ha_getDigestSize( &digestSize, hashAlgId ) != hae_ok )
			return false;

		if ( !blc->calculateDigest( hashAlgId, digestSize ) )
			return false;

		if ( digestSize == 64 )
			return encryptIV( blc->m_tail64.m_iv, (BYTE *)ebOut, (BYTE *)blc, encBlockSize, keyHandle, algId, encModeId );
		else
		if ( digestSize <= 32 )
		{
			//
			//	Version 2.00.0.4 - 22.09.2014 Bug #1644 Hidden part:: strings of zeroes.
			//	The blc with tail 32 bytes looks like
			//	[ body X bytes ][ digest 32 bytes ][ IV 8 bytes ][empty 32 bytes]
			//	The encryptIV does not encrypt IV but we has to encrypt the empty bytes otherwise they are left zero.
			//
			if ( !encryptIV( blc->m_tail32.m_iv, (BYTE *)ebOut, (BYTE *)blc, encBlockSize - tail32tail64diff, keyHandle, algId, encModeId ) )
				return false;
			return cSymAlgorithm::encrypt(  blc->m_tail32.m_iv, (BYTE *)(ebOut->m_tail32empty), (BYTE *)(blc->m_tail32empty), sizeof(blc->m_tail32empty), keyHandle, algId, encModeId );
		}

		return false;
	}

	bool encryptFromBody(  TT *bodyIn, sKeyInfo *keyInfo, BYTE *pool )
	{
		cSymAlgorithm cAlg;
		if ( !cAlg.createKey( keyInfo->keyAlgId, keyInfo->encModeId, keyInfo->key, keyInfo->getKeySizeBytes(), pool ) )
			return false;
		
		return encrypt( this, bodyIn, cAlg.m_algId, cAlg.m_encModeId, keyInfo->hashAlgId, cAlg.m_keyHandle );
	}
};

class cKeyBlockParams;

struct DATA_BLOCK
{
	DATA_BLOCK() { assert( sizeof( DATA_BLOCK ) == ( DB_MinSize ) ); }

	tcEncBlock < dbHeader > header;

	BYTE		reserved_1K[ 1024 - sizeof( tcEncBlock < dbHeader > ) ];

// -------------- 1K Bytes -----------------------------//

	BYTE		pool[ DB_PoolSize ];

// -------------- 1K + 512 Bytes -----------------------------//

	BYTE		keyTab[ DB_minKeyQ ][ DB_MinEncryptedKeyBlockSize ]; // minimal number of key may be placed in the 4K - 1.5K space left

	static int getKeyQ( DWORD dbSize )
	{
		DWORD keyTabSize = DB_minKeyQ * DB_MinEncryptedKeyBlockSize;
		DWORD keyTabPos = sizeof( DATA_BLOCK ) - keyTabSize;
		DWORD realKeyTabSize = dbSize - keyTabPos;
		if ( dbSize < keyTabPos )
			return 0;

		int keyQ = realKeyTabSize / DB_MinEncryptedKeyBlockSize;

		if ( keyQ > DB_maxKeyQ )
			keyQ = DB_maxKeyQ;

		return keyQ;
	}

	bool updateSalt( cKeyBlockParams *kbParams, bool bReadFromDBToKbParams, int kgId );
};

////////////////////////////////////
//
//	extern functions
//

class cKeyBlockParams;

extern "C" __declspec( dllexport )
	bool GetLocalPKID( BYTE *db, DWORD dbSize, int slot, wchar_t *bf, DWORD bfSize );

extern "C" __declspec( dllexport )
	bool ClearSlots( BYTE *dbBuff, DWORD dbSize, int slotN, int slotQ );

// 
//	DataBlockAllocate
//		Allocates memory for DATA_BLOCK structure 
//		The procedure which calls DataBlockAllocate is responsible for 
//		releasing the memory with the DataBlockFree function.
//
extern "C" __declspec( dllexport )
	BOOL DataBlockAllocate( 
			DATA_BLOCK	**pDataBlock,
			DWORD		*pDataSize
		);

//
//	if dataSize is 0 the block is freed without shreding
//
extern "C" __declspec( dllexport )
	BOOL DataBlockFree( DATA_BLOCK **pDataBlock, DWORD *pDataSize = 0 );

#define _UseSimpleAndExetendedBlockOnly_

//
//	DataBlockWriteKey
//
//		Writes new key to key table.
//		Note, the data block may be extended so data block pointer and it's size may be changed!
//
//		1. Writes encoded from kbParams key to key table
//			1.1 If kbParams is NULL it does not write key block, 
//				it simple free the key block chain started form kbNumber.
//			1.2 If keyMapInfo.position == -1 it writes key to any empty location
//				otherwise it shreds specified key block number and writes the new one.
//
//		2. Updates key map table in dataBlock header
//
//		3. Calculates digest for entire data block and encrypts info.
//
//			3.1 If bEncryptIt is FALSE
//				a. Calculates digest of DATA_BLOCK::dbHeader->m_body to DATA_BLOCK::dbHeader->m_digest field
//
//			3.2 If bEncryptIt is TRUE
//				a. Calculates digest of DATA_BLOCK::dbHeader->m_body to DATA_BLOCK::dbHeader->m_digest field
//				b. Encrypts DATA_BLOCK::dbHeader
//
extern DWORD
	DataBlockWriteKey(
		DATA_BLOCK			**ppDB,		// completely or partialy encrypted data block
		DWORD				*pDBSize,
		bool				bEncryptIt,	// usually the header tail is encrypted only
										// if the flag is true so entire header are completely encrypted.
		sKeyInfo			*keyInfo,	// key for data block encryption
		cKeyBlockParams		*kbParams,	// = NULL key block parameters required to encode keyInfo to the key block has
										// to be written to output data block

		sKeyMapInfo			*keyMapInfo,// if is  NULL or has undefined position -1, writes new key to first empty position
										// if not NULL shreds existing position first, fills the key map info with new data.
		dbKeyId				*keyMapT,	// if not NULL it contains information even about hidden part locations 
										// ( if theirs passwords were accepted see AddNewKey ChangePassword )
		dbHeader			*decHeader,	// if not NULL copy here updated header info ( is decrypted )
		bool				bDataBlockIsDamaged = false// It is true at ChangePassword only.
	);

//
//	DataBlockInit
//		Simple shreds data block upto dbSize and fills signature if bEncryptIt id false.
//
extern BOOL
	DataBlockInit(
			DATA_BLOCK	*db,
			DWORD		dbSize,
			DWORD		algId,
			DWORD		encModeId,
			DWORD		hashAlgId,
			cKeyBlockParams	&kbParams,
			sSpaceUnitBase	*pContInfo
		);

//
//	GenerateKey
//		Generates keys and initial vectors according to PKCS#12 recommendations.
//		Note! If password is used as seed so the buffer has to be in unicode format.
//
extern DWORD
	GenerateKey(
			BYTE *Key,
			int   KeyLength,
			BYTE *salt,
			int	  saltSize,
			BYTE *Seed,
			int   SeedLength,
			BYTE  hashAlgID,
			bool isKey // determintes is kenereted key or initial vector
		);

//
//	DataBlockVerifyKeyBlockAndGetKey
//
//	1. Tries to decode key info, it walks key blocks and implement kbParams to 
//		PBE - decrypt key by key formed from password.
//		PKE - decrypt by public key
//		SSS - implement every share to the key block
//	 Notes
//		a. If kbParams type is undefined ( kbtEmpty ), so all available 
//			encoding shemes will be implemented to DATA_BLOCK
//		b. If pKeyHandle and pKeyInfoCopy both are NULL, so the function
//			checks if the password in kbParams.kbPBE.pwd is compatible with any key block.
//			It returns ERROR_NO if at least the password is compatible even with one pwd 
//			for block with kbtSSS type
//
//	2. If decoded the key info is for hidden part, it reads next bytes as verification block, 
//			See DataBlockWriteKey_Hidden
//
extern 
DWORD DataBlockVerifyKeyBlockAndGetKey(
		DATA_BLOCK	*dataBlock,					// Data block have read directly from container
		DWORD		dbSize,
		cKeyBlockParams *kbParams,				// Key block paramteres used for key encryption.
		sKeyMapInfo *keyMapInfo = NULL,
		KEY_HANDLE	*pKeyHandle = NULL,			//	creates key handle if the parameter is not NULL 
		sKeyInfo	*pKeyInfoCopy = NULL,
		dbHeaderBase	*pDecryptedHeader = NULL,	//	place for decrypted header copy
		sSpaceUnitBase	*pContInfo = NULL,			// place for decrypted cont info for hidden or usual part corresponded to pKeyHandle
		HWND		hProgress = NULL,
		bool		*pCancelled = NULL,
		bool		bFirst4K = true				// The parsing may be devided on two part for encrypted header emprovment. First4K  or other part.
												// If this flag is false so first 4K is skipped.
	);

//
//	DataBlockWriteKey_Hidden
//
//		1. Writes hidden key info encrypted by kbParams
//		2. Writes simple verification block at the key tail encrypted by hidden key info
//
//	How the key read, see DataBlockVerifyKeyBlockAndGetKey
//
//	Note 1. The function does not extend key block for hidden part key. 
//			It will security nonsens if block is extended but key map reports that no one key presents in extended part.
//
//	Note 2. It shreds old key block information if keyMapInfo specified ( is not NULL or position != -1 )
//
extern DWORD 
	DataBlockWriteKey_Hidden(
		DATA_BLOCK			**ppDB,
		DWORD				*pDBSize,
		sKeyInfo			*hiddenKeyInfo,	//	= NULL	if null, so it does not required to save encoded key info
											//	the existing hidden part from started position is simple shreded
		cKeyBlockParams		*kbParams,		//	= NULL

		sKeyMapInfo			*keyMapInfo,	// = NULL,	See Note 2.

		dbKeyId				*keyMap,		// the map is used for detecting busied items.
											// Ifnormation about existing hidden parts positions may be accumulated by and sent to the
											// function create new hidden part. See CreateHiddenPart
		bool				bDataBlockIsDamaged = false// It is true at ChangePassword only.
	);

//
// DataBlockGetContainerId()
//		Container ID is initially randomized for main and hidden contInfo.
//		The info is not changed even if main contInfo is encrypted.
//
extern BOOL DataBlockGetContainerId( DATA_BLOCK *DataBlock, BYTE *ContainerId );

//
//	DataBlockCheck
//		Verifies that DATA_BLOCK structure is not damaged
//		and checks that DATA_BLOCK stores at least 
//		KeysCount keys.
//
extern DWORD
	DataBlockCheck(
		DATA_BLOCK *DataBlock,
		DWORD      DataSize,
		DWORD      AlgId
	);

//
//	DataBlock_CheckFreeSpace
//		Checks if exists space for new key block in the datablock
//		It may treat as free one existing key block for the case if the key block will be removed later
//
extern DWORD
	DataBlock_CheckFreeSpace(
		DATA_BLOCK	*dataBlock,
		DWORD		requiredSize,
		DWORD		*pMaxFreeSize,
		int			treatAsFree = -1
	);

extern void 
	ShredData( BYTE *data, int dataLength );

extern DWORD 
	DataBlockRemoveKeyBlock( DATA_BLOCK **ppDB, DWORD *pDBSize, cKeyBlockParams *kbParams );

class cPassword;
struct sEncIDs;

//
//	FindSamePassword
//		Tryes to decode every key block in the data block with different algorithms
//		Returns true if one key block is compatible, otherwise return false
//
bool FindSamePassword(
		DATA_BLOCK	*dataBlock,
		DWORD		dataBlockSize,
		cPassword	*pwd,
		ekbType		*pKBType = NULL,	// [Out] key block type were same password has been found
		sEncIDs		*encIds = NULL
);

template < class T > void copyEx( T *out, T *in )
{
//	assert( false ); // check this
	if ( out != NULL )
	{
		if ( in == NULL )
			memset( out, 0, sizeof( T ) );
		else
			memcpy( out, in, sizeof( T ) );
	}	
};

#define memCopyEx( out, in, const_size )\
{\
	if ( out != NULL )\
	{\
		int size = const_size;\
		\
		if ( in == NULL )\
		{\
			if ( size > sizeof( *(out) ) )\
			{\
				assert( false );\
				size = sizeof( *(out) );\
			}\
			\
			memset( out, 0, size );\
		}\
		else\
		{\
			if ( size > sizeof( *(out) ) )\
			{\
				assert( false );\
				size = sizeof( *(out) );\
			}\
			\
			if ( sizeof( *(out) ) != sizeof( *(in) ) )\
			{\
				assert( false );\
			}\
			\
			memcpy( out, in, size );\
		}\
	}\
}

//
//	decryptHeaderEx
//		decrypts header by a key handle, the key handle may be for usual or for hidden part
//	
//		1. If dataBlock is not encrypted ( see DATA_BLOCK::header.isEncrypted ) it verifies header
//			and copies dataBlock to output buffer and returns true.
//		2. Tries to decrypt dataBlock::contInfo with the algoritm id and key handle.
//			If the contInfo is properly decrypted, the function decrypts header with keyMap
//			with same key handler, copies decrypted result to out buff and returns true.
//		3. Tries to decrypt hidden part with the algorithm and key handle.
//			If decrypted hidden part was not properly verified returns false.
//		4. Repeats step 1 and 2 but with key derived from decrypted hidden part.
//		5. Otherwise returns false.
//
bool
	decryptHeaderEx(
		DATA_BLOCK	*dataBlock,
		DWORD		algId,				// Encryption Algorithm Identificator
		DWORD		*pHashAlgId,		//  = NULL, [Out] pointer to hash algorithm decrypted from encBlock tail
		KEY_HANDLE	keyHandle,			// and Key Handle it may be key handle for hidden part
		tcEncBlock <dbHeader>
					*headerOut   = NULL,// Decrypted header, caller procedure is responsible for secure erasing of the decrypted data
		DWORD		*pErrorCode	 = NULL,// if the function returns FALSE, look on the Error code
		sSpaceUnitBase	*contInfoOut = NULL,// Decrypted cont info, the procedure copies usual or hidden cont info here, depends on key handle.
		sKeyInfo	*mainKeyInfo = NULL // [OUT] if hidden part properly decrypted the value is fille by decrypted main key info
	);							

//
//	encryptHeaderEx
//		encrypts entire header or it's tail with digest by a key handle for usual part.
//		Note, if the key handle is for hidden part the usual part key is derived from decrypted hidden part.
//	
//		1. If dataBlock is encrypted ( see DATA_BLOCK::header.isEncrypted )
//			1.1 it checks if key handle is valid for the data block by decryptHeaderEx.
//			1.2 If is valid the function copies dataBlock to output buffer and returns true.
//			If is not valid the function returns false
//		2. Checks if key handle is for hidden part.
//			If so derive new key handle form decrypted hidden part
//		3. Encrypts header by the key handle for usual part
//
bool
	encryptHeaderEx(
		DATA_BLOCK	*dataBlock,
		DWORD		algId,			// Encryption Algorithm Identificator
		DWORD		encModeId,		// Encryption Algorithm Identificator
		DWORD		hashAlgId,		// Hash algorithm is used for digest
		DWORD		keyHandle,		// and Key Handle it may be key handle for hidden part
		tcEncBlock <dbHeader>
					*dbOut,			// Entire encrypted block with tail
		DWORD		*pErrorCode,	// if the function returns FALSE, look on the Error code
		bool		bEntire			// = true, Defines if entire or tail only encryption is used
	);

//
//	decryptHeaderEx
//		decrypts header by a key handle, the key handle may be for usual part only
//			( hidden part key consists of both hidden and main parts keys ) see DataBlockVerifyKeyBlockAndGetKey
//	
//		1. If dataBlock is not encrypted ( see DATA_BLOCK::header.isEncrypted ) it verifies header
//			and copies dataBlock to output buffer and returns true.
//		2. Tries to decrypt dataBlock::contInfo with the algoritm id and key handle.
//			If the contInfo is properly decrypted, the function decrypts header with keyMap
//			with same key handler, copies decrypted result to out buff and returns true.
//		3. Otherwise returns false.
//
bool
	decryptHeaderEx(
		DATA_BLOCK	*dataBlock,
		DWORD		contAlgId,	// Encryption Algorithm Identificator
		DWORD		encModeId,
		DWORD		*pHashAlgId,	//  = NULL, [Out] pointer to hash algorithm decrypted from encBlock tail
		DWORD		keyHandle,		// and Key Handle it may be key handle for hidden part
		tcEncBlock <dbHeader>
					*headerOut,		// = NULL, Decrypted header, caller procedure is responsible for secure erasing of the decrypted data
		DWORD		*pErrorCode		// = NULL, if the function returns FALSE, look on the Error code
	);

//
//	EncryptHeader
//		Asks password for encryption/decryption
//		Encrypts decrypted header and decrypts encrypted.
//
extern
	bool EncryptHeader(
		HWND			hWnd,
		const wchar_t	*contPath,
		tcEncBlock < dbHeader >
					*db,
		DWORD		dbSize,
		DWORD		*pErrCode,
		DWORD		*pFlags,
		sCKHParams	*ps
	);

//
//	decryptAndVerifyHeader
//
DWORD
	decryptAndVerifyHeader(
		DATA_BLOCK	*dataBlock,
		sKeyInfo	*keyInfo,
		tcEncBlock < dbHeader > 
					*decryptedHeader,
		KEY_HANDLE	*pDelegateKeyHandle = NULL // key handle to be delegeted
	);

////
////	HP_GetMainPartContInfo
////		1. Finds position of encrypted main part info
////			See DataBlockWriteKey_Hidden.
////		2. Decryptes main part contInfo
////		3. Delegates main part key handle for further mounting
////
//extern bool 
//	HP_GetMainPartContInfo( DATA_BLOCK *db, KEY_HANDLE keyHandleHP, sCKHParams *psHP, KEY_HANDLE *pKeyHandleMain, sCKHParams *psMain );
//

//
//	CreateHandle_MainPart
//		Main part password required.
//
bool CreateHandle_MainPart(

		HWND hWnd,

		const wchar_t *text,
		const wchar_t *caption,

		DATA_BLOCK			**ppDB,
		DWORD				*pDBSize,
		
		sKeyInfo			*keyInfo = NULL,

		KEY_HANDLE			*pKeyHandle = NULL,
		sCKHParams			*ps = NULL,
		dbKeyId				*keyMapT = NULL
	);

extern bool 
	updateKeyParams( sCKHParams *ps, sKeyInfo *keyInfo, dbHeader *decHeader = NULL );


// Password Based Encrypted key block
struct PBE_KeyBlock
{
	// Salt is generated before encoding.
	BYTE		salt[ SALT_SIZE ];

	// Note, we cannot store here iteration count and hash algorithm id because it is not random data
	//	but the block has to be completely randomised or encrypted.

	tcEncBlock < sKeyInfo > keyInfo;
};

// Password Based Encrypted key block for version 5
struct PBE_KeyBlock_V5
{
	tcEncBlock < sKeyInfo > keyInfo;
};

// Public Key Encrypted key block
struct PKE_KeyBlock
{
	BYTE buff[ 1 ]; // Real buffer size defined by secret key size.
					// The stored buffer is pure product of public key encrypt operation.
};

// Shamir Shared Scheme key block
struct SSS_KeyBlock
{
	BYTE buff[ 1 ]; // Real buffer size defined by choosen shared sheme.
					// Where it may be stored for decrypted, may be MAX_SSS_SIZE or up to the data 
					//	block baund ( it is limited by DB_TotalSize )
					// Every ecrypted share contains number of required shares to recover key.
};

struct sKGCmdParams
{
	DWORD flags;	// Old version compatible flag
	DWORD size;	// size of the structure

	enum { stInit = 0, stUpdated = 1 };
	DWORD state;	// signaled state, initially is zero,
			// if the command successfully processed state = stUpdated

	sSpaceUnitBase mainContInfo;

// Parameters for mounting 
	HWND hWnd;
	sSpaceUnitBase contInfo;
	KEY_HANDLE *pKeyHandle;
//	int favDrive;
	BOOL readOnly;
	wchar_t mountPoint[ MAX_PATH ];
	
	sKGCmdParams()
	{
		memset( this, 0, sizeof( *this ) );
		size = sizeof( *this );
	}
};

#pragma pack()

//
//	verifyKeyHandle
//	
//	Every create key handle function may has an ready key handle
//	The funvtion allows to verify the handle before using.
//
//	returns ERROR_NO if success.
//
DWORD verifyKeyHandle( DATA_BLOCK *db, KEY_HANDLE keyHandle, DWORD algId, DWORD encModeId, dbHeader *decHeader = NULL );
