/*******************************************************************
 *
 *	Copyright (c) 2005-2014 Jetico, Inc., Finland
 *	All rights reserved.
 *
 *	File: kgGhost.h
 *	Revision: 
 *	Created:
 *	Description: main header file of the KGGhost.DLL with the 
 *		functions that may be called from external programs
 *
 *******************************************************************/

#pragma once

#include <windows.h>
#include "kblock.h"
#include "cAppFlags.h"

#define KEYGEN_ID    0x00000004
#define KEYGEN_ID_V5 0x00000005

#define DISPLAY_NAME "KG-Ghost"
#define SERVICE_NAME "kgGhost.dll"

#define MAJOR_VERSION 0x00000001
#define MINOR_VERSION 0x00000000

//
//	Key generator has uniquie identifier
//		But can suport different.
//		This query allows to get min and max supported identifiers,
//		if function is not available, the unique idetifier is ised only.
//
extern "C" __declspec( dllexport )
	bool GetSupportedIDs(
		DWORD *MajorID, 
		DWORD *MinorID 
		);

extern "C" __declspec( dllexport )
	BOOL GetNameAndVersion( 
			char  *Name, 
			UINT  nameSize, 
			DWORD *Major, 
			DWORD *Minor 
			);

extern "C" __declspec( dllexport ) 
	BOOL CreateKeyHandle(
		char		*AlgServiceName,// what Encryption Algorithm will use the encryption key
		DWORD		AlgId,			// Encryption Algorithm Identificator
		DWORD		AlgKeyLength,	// encryption key length for the Encryption Algorithm
		char		*Text,			// Any information text (optional parameter)
		char		*Caption,		// Caption of the "get password" dialog window (optional parameter)
		DWORD		CreateFlag,		// whether to create new key or open existing
		BYTE		**vDataBlock,	// block of data that is stored inside file-container in encrypted form
		DWORD		*DataSize,		// size of the block of data
		KEY_HANDLE	*KeyHandle,		// returned Key Handle
		DWORD		*ErrorCode,		// if the function returns FALSE, look at the Error code
		HWND		hWndParent		// parent Window for any dialogs drawn by the DLL (optional parameter, may be NULL))
		);


extern "C" __declspec( dllexport ) 
	BOOL CreateKeyHandleEx(
		char		*AlgServiceName,	// what Encryption Algorithm will use the encryption key
		DWORD		AlgId,			// Encryption Algorithm Identificator
		DWORD		AlgKeyLength,	// encryption key length for the Encryption Algorithm
		char		*Text,			// Text, containing filename of the container
		char		*Caption,		// Caption of the "get password" dialog window
		DWORD		CreateFlag,		// whether to create new key or open existing
		BYTE		**vDataBlock,	// block of data that is stored inside file-container in encrypted form
		DWORD		*DataSize,		// size of the block of data
		KEY_HANDLE	*KeyHandle,		// returned Key Handle
		DWORD		*ErrorCode,		// if the function returns FALSE, look on the Error code
		HWND		hWndParent,		// parent Window for any dialogs drawn by the DLL
		DWORD		*FavoriteDrive, // if 0 <= *Favorite drive < 26, user will be asked to enter drive letter
		DWORD		*Flags          // see the KG_DEFS.H file for the flags (FLAG_...) that may be set/get
		);

typedef DWORD typeGetPasswordPWD(
		const wchar_t *text,
		const wchar_t *caption,
		wchar_t *pwd,
		int pwdSize,
		bool bVerify,
		wchar_t *mountPoint,
		int mountPointSize
	);

extern "C" __declspec( dllexport ) 
	BOOL CreateKeyHandle_V2(
		HWND		hWndParent,		// parent Window for any dialogs drawn by the DLL
		const wchar_t	*text,		// Text, containing filename of the container
		const wchar_t	*caption,	// Caption of the "get password" dialog window
		DWORD		contAlgId,		// Encryption Algorithm Identificator
		DWORD		CreateFlag,		// whether to create new key or open existing
		BYTE		**vDataBlock,	// block of data that is stored inside file-container in encrypted form
		DWORD		*pDataSize,		// size of the block of data
		KEY_HANDLE	*pKeyHandle,	// returned Key Handle
		DWORD		*pErrorCode,	// if the function returns FALSE, look at the Error code
		DWORD		*pFlags,        // see section "Advanced flags" in kg_defs.h
		sParams		*params
		);

extern "C" __declspec( dllexport ) 
	DWORD UpdateAdvancedSettings( DWORD *Flags, HWND  hWndParent );

extern "C" __declspec( dllexport ) 
	DWORD FreeDataBlock( BYTE **vDataBlock );

extern "C" __declspec( dllexport ) 
	DWORD FreeKeyHandle( 
		char  *AlgServiceName,
		DWORD KeyHandle 
		);

DWORD FreeKeyHandle( DWORD algId, DWORD KeyHandle );

//
//	CreateHiddenPartEx
//	It does:
//		1. Gets new key block params for hidden part
//			1.1 Checks if same password already exists.
//		// See ( Description 1.1 ) below...	
//			1.2 If the kbParams is for hidden part, it asks confirmation to overwrite the existing.
//				So space at key map table may be freed.
//			1.3 It mart the field is busied and does not allow to overwrite when key key is written to data block.
//
//		2. Verifies all we did before and gets key handle
//		
//	Rules:
//		1. It does not modify dbHeader so main part key handle is not required.
//		2. It uses keyMapT to detect empty key places, and writes new key directly to data block 
//		3. It does not extend key block, so it has to treat as busied all key in map beyond real data block size. 
//
extern "C" __declspec( dllexport ) 
	BOOL CreateHiddenPartEx(
		HWND		hWnd,
		wchar_t		*contFileName,

		DATA_BLOCK	*pDB,			// block of data that is stored inside file-container in encrypted form
		DWORD		dbSize,			// size of the block of data

		KEY_HANDLE	*pKeyHandleHP,	// hidden part key handle 
		sCKHParams	*psHP,			// parameters for created hidden part
		
		dbKeyId		*keyMapT,		// the information about busied key blocks, new hidden part block will be created at empty key places

		DWORD		*pErrCode,		// if the function returns FALSE, look at the Error code

		DWORD		flags
	);

//
//	Real space unit table may be stored at specified address.
//	The DecryptTable tries to decrypt it.
//
enum { dt_errNo = 0, dt_bufferTooSmall = 1, dt_invalidSignature, dt_errorInFileOpening, dt_invalidPointer, dt_errorInFileWritting };

//struct sSpaceUnit;

//
//	08.10.2014 Additional command did never used by BestCrypt.
//		See coments ..\BCSources\VersionInfo.txt, v. 9.00.0.8b - 07.10.2014, p.3.b
//
//extern "C" __declspec( dllexport ) 
//	BOOL GetAdditionalCommandNames(
//		wchar_t  *CommandNamesBuffer,   // buffer where the function returns strings of commands, like "comm1\0comm2\coom3\0\0"
//		int   *BufferSize,           // size of the buffer for the command names
//		DWORD *ErrorCode,            // if the function returns FALSE, ErrorCode is valid, see kg_defs.h
//		DWORD *Flags                 // some flags BestCrypt tells to the Key Generator, see kg_defs.h
//		);
//
//extern "C" __declspec( dllexport ) 
//	BOOL RunAdditionalCommand(
//		HWND    hWnd,
//		wchar_t    *CommandName,   
//		wchar_t    *ContainerPath,
//		BYTE    *DataBlock,
//		int     DataBlockSize,
//		void   *params           // see the kg_defs.h file
//		);

extern "C" __declspec( dllexport ) 
	BOOL CheckBestCryptVersion(
		HWND    hWnd,
		DWORD   MajorVersion,
		DWORD   MinorVersion 
	);

extern "C" __declspec( dllexport ) 
	BOOL QueryFunction(
		const BYTE	**vDataBlock,	// block of data that is stored inside file-container in encrypted form
		const DWORD	dataSize,		// size of the block of data
		const DWORD	functionID		// identefier of function used as CerateKeyHandle(.., CreateFlag, ... )
		);

#pragma pack(1)

enum {
	UBFlag_Decrypt = 0,
	UBFlag_Encrypt = 1,
	UBFlag_Encrypt_Digest_Only = 2
};

struct sUBParam
{
	DWORD	size;
	DWORD	flags;
	
	sUBParam()
	{
		memset( this, 0, sizeof( *this ) );
		size = sizeof( *this );
	}
};

struct sParamEncrypt : public sUBParam
{
	DWORD hashAlgId;
};

#pragma pack()

//
//	UpdateHeader
//		The function decryptes/encrypts data block to outBuff
//		Note, inBuff may be same as outBuff
//		If data block is encrypted so the header, contInfo and keyMap is encrypted
//
extern "C" __declspec( dllexport ) 
	BOOL UpdateHeader(
		HWND	hWnd,		// parent Window for any dialogs drawn by the DLL
		BYTE	*dataBlock,
		DWORD	dataBlockSize,
		DWORD	algId,			// Encryption Algorithm Identificator
		DWORD	encModeId,		// Encryption mode
		DWORD	keyHandle,		// Key Handle
		BYTE	*outBuff,		// Caller procedure is responsible for secure erasing of the decrypted data
		DWORD	buffSize,
		DWORD	*pErrCode,		// if the function returns FALSE, look on the Error code
		sUBParam	*param		// = NULL
	);

extern "C" __declspec( dllexport )
	int ShowMessage( HWND hWnd, sParams *params );

BOOL
	GetKeyHandle(
		HWND	hWnd,
		const wchar_t	*Text,
		const wchar_t	*Caption,

		DATA_BLOCK	**ppDataBlock,
		DWORD		*pDataSize,

		DWORD		*pErrCode,
		DWORD		*pFlags,

		KEY_HANDLE	*pKeyHandle,
		//wchar_t		*mountPoint,

		sCKHParams	*ps,

		sKeyInfo	*pKeyInfo  = NULL,

		typeGetPasswordPWD *getPwdFunc = NULL
	);

//
//	08.10.2014
//		The function did never used. So are removed with Additional commands.
//
//extern "C" __declspec( dllexport ) 
//	bool GetAlarmCrashHotkey( DWORD *pDWValue );
//
//extern "C" __declspec( dllexport ) 
//	bool SetAlarmCrashHotkey( DWORD dwValue );

const int DEFAULT_ITERATIONS_V4 = 256;
const int DEFAULT_ITERATIONS_V5 = 16 * 1024;

extern bool isDefaultIterations( int iterations );
extern int defIterationsT[2];

enum eFlags { efShowHPInfo = 0, efUseV9Default, efDoNotEncryptHeader };
extern cAppFlags appFlags;

//
//	28.12.2015
//		MP4 merging fuctionality does not alow to mount main part with patched header.
//		But we we have all required information for mounting as well as hidden part.
//		This define allows to propogate error ERROR_INCORRECT_BLOCK_SIGNATURE to BestCrypt with filling all fields.
//		BestCrypt may ask confirmation and allows mounting of damaged header ( not only patched by MP4 merging ).
//
#define _allowBadHeader

#ifdef _allowBadHeader
extern int _lastIterations;
#endif
