#include "ShowKeyBlockInfo.h"
#include "resource.h"
#include "kgGhost.h"
#include "cDlg.h"
#include "ptbRegs.h"
#include "cStringBuffer.h"
#include <stdio.h>

static int getKeyQ( dbHeader *db, ekbType kbType = kbtEmpty )
{
	int count = 0;

	for ( int i = 0; i < DB_maxKeyQ; i++ )
	{
		if ( ( db->keyMap[ i ].kbType != kbtEmpty ) && ( db->keyMap[ i ].kbType != kbtTail ) )
		{
			if ( ( kbType == kbtEmpty ) || 
				 ( db->keyMap[ i ].kbType == kbType ) )
			{
				count++;
			}
		}
	}

	return count;
}

void addStringWithStr( HWND hList, const wchar_t *str, const wchar_t *strTail )
{
	LVITEMW lvi;
	memset( &lvi, 0, sizeof( lvi ) );

	lvi.mask = LVIF_TEXT;
	lvi.iItem = 0xFFFF; // Append to tail.
	lvi.pszText = (wchar_t *)str;

	int i = ListView_InsertItemT( hList, &lvi );

	ListView_SetItemTextT( hList, i, 1, strTail );
}

void addStringWithStr( HWND hList, int strID, const wchar_t *tail )
{
	addStringWithStr( hList, _LoadIDS( strID ), tail );
}

void addStringWithSize( HWND hList, int strID, DWORDLONG size )
{
	wchar_t tmp[ 128 ];
	GetSizeTextT( tmp, maxLen( tmp ), size );
	addStringWithStr( hList, _LoadIDS( strID ), tmp );
}

void addStringWithInt( HWND hList, int strID, int num )
{
	wchar_t tmp[ 128 ];
	_itow( num, tmp, 10 );
	addStringWithStr( hList, _LoadIDS( strID ), tmp );
}

void addStringWithInt( HWND hList, const wchar_t *str, int num )
{
	wchar_t tmp[ 128 ];
	_itow( num, tmp, 10 );
	addStringWithStr( hList, str, tmp );
}

void addStringWithBytes( HWND hList, int strID, BYTE *bf, int bfSize )
{
	wchar_t tmp[ 128 ];
	bytes2string( tmp, maxLen( tmp ), bf, bfSize );
	addStringWithStr( hList, _LoadIDS( strID ), tmp );
}

static UINT ColumnLabelIDS[] = 
{
	IDS_szName, IDS_szValue
};

static const int NUM_COLUMNS = maxLen( ColumnLabelIDS );

static int _gnColumnFmt[ NUM_COLUMNS ]=
{
	LVCFMT_LEFT, LVCFMT_LEFT
};

static int _gnColumnWidth[NUM_COLUMNS]=
{
	200, 200
};

static bool InitListViewColumns( HINSTANCE hInst, HWND hLView )
{
	int i;
	LVCOLUMNW lvc;

	memset( &lvc, 0, sizeof( lvc ) );

	lvc.mask = LVCF_WIDTH;

	//
	//	Check if already initialised...
	//
	if ( ListView_GetColumnT( hLView, 0, &lvc ) )
		return true;

	lvc.mask = LVCF_FMT | LVCF_WIDTH | LVCF_TEXT | LVCF_SUBITEM;

	wchar_t text[ 80 ];

	int res;

	for(i=0; i<NUM_COLUMNS; i++)
	{
		LoadStringT( hInst, ColumnLabelIDS[i], text, maxLen( text ) );

		lvc.iSubItem=i;
		lvc.pszText = text; // _gszColumnLabel[i];
		lvc.cx = _gnColumnWidth[i];
		lvc.fmt=_gnColumnFmt[i];

		res = ListView_InsertColumnT( hLView, i, &lvc );

		if ( res == -1 )
			return false;
	}

	return true;
}

const wchar_t *pkeskbDlgPos = L"kbInfoDlgPos";
const wchar_t *regBCKeyName = L"SOFTWARE\\Jetico\\BestCrypt";

cRegDlgValues dlgVals( HKEY_CURRENT_USER, regBCKeyName, pkeskbDlgPos, sizeof(cRegDlgValues), NUM_COLUMNS );

class cDlgKBInfo : public cColoredDlg
{
public:

	HWND m_hLV;

	const wchar_t *m_contPath;
	dbHeader *m_db;
	int m_dataBlockSize;
	sKGCmdParams *m_sParams;

	cDlgKBInfo( const wchar_t *path, dbHeader *db, int dbSize, sKGCmdParams *sParams )
	{
		m_contPath = path;
		m_db = db;
		m_dataBlockSize = dbSize;
		m_sParams = sParams;
	}

	BOOL InitDialog();
	bool FillListView();

	LRESULT RealDlgProc( UINT uMsg, WPARAM wParam, LPARAM lParam )
	{
		switch (uMsg) 
		{ 
			case WM_INITDIALOG:
				return InitDialog();

	        case WM_COMMAND: 
			{
				switch(wParam) 
				{ 
					case IDOK: 
					case IDCANCEL: 
						EndDialog( m_hDlg, IDOK );
						return TRUE; 
					default:
						return 0; 
				}
			}

			case WM_DESTROY:
				dlgVals.Save( m_hDlg, m_hLV );
				break;

			default:
				return 0; 
		}

		return 0;
	}
};

bool cDlgKBInfo::FillListView()
{
	ListView_DeleteAllItems( m_hLV );

	LVITEMW lvi;
	memset( &lvi, 0, sizeof( lvi ) );

	wchar_t buff[ MAX_PATH ];

	int pos = splitPath( m_contPath, buff );

	addStringWithStr  ( m_hLV, IDS_szFileName, m_contPath + pos );
	addStringWithStr  ( m_hLV, IDS_szLocation, buff );

	addStringWithSize ( m_hLV, IDS_DataBlockSize, m_dataBlockSize );
	addStringWithBytes( m_hLV, IDS_ContainerID, m_db->containerID, sizeof( m_db->containerID ) );
	addStringWithStr  ( m_hLV, IDS_KeyGenerator, getWCStr( DISPLAY_NAME ) );

	cSymAlgorithm symAlg;
	
	if ( symAlg.Init( m_db->contAlgId, m_db->encModeId ) )
	{
		addStringWithStr( m_hLV, IDS_EncryptionAlgorithm, symAlg.getAlgName( buff, maxLen( buff ) ) );
	}
	else
		addStringWithInt( m_hLV, cWCStr().FormatT( L"%s ID", _LoadIDS( IDS_EncryptionAlgorithm ) ), m_db->contAlgId );

	addStringWithSize ( m_hLV, IDS_VDOffset, m_db->contInfo.offset );
	addStringWithSize ( m_hLV, IDS_VDSize,   m_db->contInfo.length );

	int keyQ;
	keyQ = getKeyQ( m_db, kbtPBE ); if ( keyQ ) addStringWithInt  ( m_hLV, IDS_PBE_KeyNumber, keyQ );
	keyQ = getKeyQ( m_db, kbtPKE ); if ( keyQ ) addStringWithInt  ( m_hLV, IDS_PKE_KeyNumber, keyQ );
	keyQ = getKeyQ( m_db, kbtSSS ); if ( keyQ ) addStringWithInt  ( m_hLV, IDS_SSS_KeyNumber, keyQ );

	return true;
}

BOOL cDlgKBInfo::InitDialog()
{
	SendMessage( m_hDlg, WM_SETICON,0, (LPARAM)LoadIcon( m_hInst, MAKEINTRESOURCE( IDI_BCDoc ) ) );

	m_hLV = GetDlgItem( m_hDlg, IDC_List );
	InitListViewColumns( m_hInst, m_hLV );

	ListView_SetExtendedListViewStyleEx( m_hLV, LVS_EX_GRIDLINES, LVS_EX_GRIDLINES );

	m_sizedCtrls.Init( m_hDlg );
	m_sizedCtrls.AddCtrl( IDC_List, m_sizedCtrls.ptSDR );

	dlgVals.Restore( m_hDlg, m_hLV );

	return FillListView();
}

bool ShowKeyBlockInfo( HWND hWnd, const wchar_t *contPath, dbHeader *db, int dbSize )
{
	if ( dbHeader::isEncrypted( *db ) )
	{
		cWCStr text;
		text.FormatT( 
			_LoadIDS( IDS_S_DB_isEncrypted_DBSize_S ),
			cWCStr().FormatT( _LoadIDS( IDS_Cont_S ), contPath ),
			cWCStr().GetSizeText( dbSize ) );

		MessageBoxT( hWnd, text.m_szBody, _LoadIDS( IDS_cmdShowKeyBlockInfo ), MB_OK );
	}
	else
	{
		sKGCmdParams sParams;
		sParams.hWnd = hWnd;

		cDlgKBInfo dlg( contPath, db, dbSize, &sParams );
		
		dlg.DoModal( IH.m_hInst, IDD_KBInfo, hWnd );
	}

	return TRUE;
}
