#include "cDlgEncHeaderParse.h"
#include "resource.h"
#include "AutoDeallocate.h"
#include "..\..\..\CertificateLib\hashAlgs\hashAlgs.h"
#include "BC_ioctl.h"
#include "ODS.h"
#include "KgGhost.h"

BOOL cProgressDlg_EncHeaderParse::InitDialog()
{
	m_hProgress = GetItem( IDC_PROGRESS );
	m_pattern.GetWindowTextT( GetItem( IDC_HashAlgTxt ) );

	DWORD threadId;
	HANDLE hThread = CreateThread( NULL, 0, threadFunc, this, 0, &threadId );

	if ( hThread == NULL )
	{
		MessageBoxT( m_hDlg, L"Error in thread creating", _LoadIDS( IDS_KeyGen ), MB_ICONSTOP );
		EndDialog( m_hDlg, -1 );
		return false;
	}

	CloseHandle( hThread );

	return TRUE;
}

LRESULT cProgressDlg_EncHeaderParse::RealDlgProc( UINT uMsg, WPARAM wParam, LPARAM lParam )
{
	switch (uMsg) 
	{ 
		case WM_INITDIALOG:
			return InitDialog();

	    case WM_COMMAND: 
		{
			switch(wParam) 
			{ 
				case IDOK: 
					EndDialog( m_hDlg, m_res );
					// ??? Mistic feature, but app is crashes in Vista without the wating.
					// Cancel of decoding when container with encrypted header is been mounting.
					Sleep( 100 );
					return TRUE; 
				case IDCANCEL:
				{
					int res = MessageBoxT(
								m_hDlg,
								_LoadIDS( IDS_Q_StopKeyEncodingProcess ),
								_LoadIDS( IDS_KeyGen ),
								MB_ICONQUESTION | MB_YESNO | MB_DEFBUTTON2 );

					if ( res != IDYES )
							return 0;

					m_cancelled = true;
					EnableWindow( GetDlgItem( m_hDlg, IDCANCEL ), false );
					return TRUE;
				}
				default:
					return 0; 
			}
		}
		default:
			return 0; 
	}

	return 0;
}

struct sAlgInfo{
	int m_encAlgId;
	int m_ecnModeId;
	int m_keySize;
};

class cEnumSortedT
{
public:
	sAlgInfo *T;
	int Q;
	
	cEnumSortedT( cEnumAlgIDs &algs, cEnumEncModes &encModes )
	{
		Q = algs.Q * encModes.Q;
		T = new sAlgInfo[ Q ];

		if ( T == NULL )
		{ Q = 0; return; }

		cSymAlgorithm cAlg;

		for( int i = 0, k = 0; i < algs.Q; i++ )
		{
			for( int j = 0; j < encModes.Q; j++, k++ )
			{
				if ( k >= Q )
					return;

				T[k].m_encAlgId = algs.T[i];
				T[k].m_ecnModeId = encModes.T[j].id;
				T[k].m_keySize = 0;

				if ( !cAlg.Init( T[k].m_encAlgId, T[k].m_ecnModeId ) || !cAlg.supportMode( T[k].m_ecnModeId ) )
					continue;
				
				T[k].m_keySize = ( cAlg.m_keyLengthInBites + 7 ) / 8;
			}
		}

		// Sort by key size.
		for( int i = Q-1; i >= 1; i-- )
		{
			for( int j = 0; j < i; j++ )
			{
				if ( T[j].m_keySize < T[j+1].m_keySize )
				{
					// Swap T[j], T[j+1]
					sAlgInfo ai = T[j+1];
					T[j + 1] = T[j];
					T[j] = ai;
				}
			}
		}

		for( ; Q > 0 ; Q-- )
		{
			if ( T[Q - 1].m_keySize != 0 )
				break;
		}
	}
	
	~cEnumSortedT()
	{
		if ( T != NULL )
			delete T;

		T = NULL;
		Q = 0;
	}
};

int cProgressDlg_EncHeaderParse::Run( cEnumHashAlgIDs &hashAlgs, cEnumSortedT &sortedAlgs, bool bFirst4K )
{
	assert( m_pKBParams->kbType == kbtEmpty );

	bool bCheckSK = bFirst4K;

	int countV5 = sortedAlgs.Q * hashAlgs.Q * maxLen( defIterationsT );
	int countV4 = sortedAlgs.Q * hashAlgs.Q;

	SendMessage( m_hProgress, PBM_SETRANGE, 0, MAKELPARAM( 0, countV5 + countV4 + bCheckSK ) );
	SendMessage( m_hProgress, PBM_SETSTEP, 1, 0 );

	//
	//	30.06.2011
	//	We know enough for secret key opening password and keyIDSBuffer.
	//	Note if secret key was not found at BCDB it will not asked and opened.
	//
	if ( bCheckSK ) // && ( m_pKBParams->kbPKE.ska == NULL ) )
	{
		SetDlgItemTextT( IDC_HashAlgTxt, L"Check secret key" );

		DWORD maxKeyQ = DATA_BLOCK::getKeyQ( m_dbSize );

		for ( DWORD i=0; i < maxKeyQ; i++ )
		{
			BYTE *kbEncrypted = &(m_dataBlock->keyTab[i][0]);

			if ( m_pKBParams->kbPKE.openSecretKey( m_pKBParams->kbSSS.hWnd, &m_pKBParams->kbPBE.m_pwd, kbEncrypted, true ) )
				break;
		}

		SendMessage( m_hProgress, PBM_STEPIT, 0, 0 );
	}

	if ( RunLoops( hashAlgs, sortedAlgs, cParamsPBE::eDecodeType_V5, bFirst4K ) )
		return true;

	if ( RunLoops( hashAlgs, sortedAlgs, cParamsPBE::eDecodeType_V4, bFirst4K ) )
		return true;

	return false;
}

static bool isNewHashAlgorithm( int hashAlgId )
{
	int newHashT[] = { bchaWhirlpool512, pgphaSHA512, bchaSHA3_512, bchaSkein512 };

	for( int i = maxLen( newHashT ); i--; )
	{
		if ( newHashT[i] == hashAlgId )
			return true;
	}

	return false;
}

bool cProgressDlg_EncHeaderParse::RunLoops( cEnumHashAlgIDs &hashAlgs, cEnumSortedT &sortedAlgs, int decodeType, bool bFirst4K )
{
	m_pKBParams->kbPBE.m_decodeType = cParamsPBE::eDecodeType( decodeType );

	cADBlock < dbHeader > adbDecHeader;

	bool bIsVersion4 = decodeType == cParamsPBE::eDecodeType_V4;

	cWCS wcsTail( L"" );

	if ( bIsVersion4 )
		wcsTail.Set( L"v.4 " );

	if ( !bFirst4K )
		wcsTail.Add( L"extended" ); 

	if ( adbDecHeader.m_T == NULL )
	{
		assert( false );
		*m_pErrCode = ERROR_INTERNAL_PROBLEM;
		return 0;
	}

	for( int j = 0; j < hashAlgs.Q; j++ )
	{
		m_pKBParams->kbPBE.m_hashAlgId = (BYTE)hashAlgs.T[j];
		SetDlgItemTextT( IDC_HashAlgTxt,
			_FormatT( m_pattern.m_szBody,
				_FormatT( L"(%d/%d) - %s %s", j+1, hashAlgs.Q, _getWcs( ha_getHashAlgName( (BYTE)hashAlgs.T[j] ) ), wcsTail.m_szBody )
		) );
		
		m_dataBlock->updateSalt( m_pKBParams, true, bIsVersion4 ? KEYGEN_ID : KEYGEN_ID_V5 ); // 

		m_pKeyMapInfo->setUndefined();

		for( int k = 0; k < maxLen( defIterationsT ); k++ )
		{
			if ( bIsVersion4 )
			{
				m_pKBParams->kbPBE.m_iterations = DEFAULT_ITERATIONS_V4;
				k = maxLen( defIterationsT );
			}
			else
				m_pKBParams->kbPBE.m_iterations = defIterationsT[k];

			for( int i = 0; i < sortedAlgs.Q; i++ )
			{
				m_pKBParams->kbPBE.m_algId		= sortedAlgs.T[i].m_encAlgId;
				m_pKBParams->kbPBE.m_encModeId	= sortedAlgs.T[i].m_ecnModeId;

				SendMessage( m_hProgress, PBM_STEPIT, 0, 0 );

				if ( sortedAlgs.T[i].m_keySize == 0 ) // Mode does not supported
					continue;

				if ( bIsVersion4 && isNewHashAlgorithm( m_pKBParams->kbPBE.m_hashAlgId ) )
					continue;

				*m_pErrCode =
					DataBlockVerifyKeyBlockAndGetKey( m_dataBlock, bFirst4K ? _4K : m_dbSize, m_pKBParams, m_pKeyMapInfo, m_pKeyHandle, m_pKeyInfoCopy, adbDecHeader.m_T,
						NULL,NULL, &m_cancelled, bFirst4K );

				if ( m_cancelled )
				{	*m_pErrCode = ERROR_CANCELLED_BY_USER; return false; }

#ifdef _allowBadHeader
				if ( *m_pErrCode == ERROR_INCORRECT_BLOCK_SIGNATURE )
				{
					_lastIterations = m_pKBParams->kbPBE.m_iterations;
					return true;
				}
#endif

 				if ( *m_pErrCode == ERROR_NO )	break;
			}

			if ( *m_pErrCode == ERROR_NO ) break;
		}

		if ( *m_pErrCode == ERROR_NO ) break;
	}

	if ( *m_pErrCode != ERROR_NO )
		return false;

	copyEx( m_pDecryptedHeader, adbDecHeader.m_T );

	return true;
}

// static 
DWORD WINAPI cProgressDlg_EncHeaderParse::threadFunc( void *ptr )
{
	cProgressDlg_EncHeaderParse *pThis = ( cProgressDlg_EncHeaderParse * )ptr;
	if ( pThis == NULL )
		return -1;

	SetWindowTextT( pThis->GetItem( IDC_HashAlgTxt ), L"Prepare algorithm table" );

	cEnumAlgIDs algs; // ( bcsaRIJN );
	cEnumHashAlgIDs hashAlgs( bchaWhirlpool512 );
	hashAlgs.setFirst( hashAlgs.T + 1, hashAlgs.Q - 1, pgphaSHA512, true );
	hashAlgs.setFirst( hashAlgs.T + 2, hashAlgs.Q - 2, bchaSHA256, true );
	hashAlgs.setFirst( hashAlgs.T + 3, hashAlgs.Q - 3, bchaSHA3_512, true );

	cEnumEncModes encModes( EX_ENCRYPT_XTS_MODE_FLAG );

	cEnumSortedT sortedAlgs( algs, encModes );

	pThis->Run( hashAlgs, sortedAlgs, true ); // Parses first 4K

	if ( *pThis->m_pErrCode == ERROR_INCORRECT_PASSWORD )
	{
		int res = pThis->DlgMessageBoxT( IDS_KeyIsNotFoundInMainKeyBlock, MB_CANCELTRYCONTINUE | MB_ICONEXCLAMATION );
		if ( res == IDCONTINUE )
			pThis->Run( hashAlgs, sortedAlgs, false ); // Parses left 16K
		else
		if ( res == IDCANCEL )
		{
			pThis->m_cancelled = true;
			*pThis->m_pErrCode = ERROR_CANCELLED_BY_USER;
		}
		else // id == IDTRYAGAIN
		{
			*pThis->m_pErrCode = ERROR_INCORRECT_PASSWORD;
		}
	}

	PostMessage( pThis->m_hDlg, WM_COMMAND, IDOK, 0 );

	return 0;
}
