#include "cDlgKBParams.h"
#include "resource.h" 
#include <windowsx.h>
#include "cDlgKeyFileMan.h"

bool cPageKBParams::Init( cDlgKBParams *dlg )
{
	m_dlg = dlg;
	if ( dlg == NULL )
	{
		assert( false );
		return false;
	}

	m_hDlg = m_dlg->m_hDlg;
	m_useSecureTyping = m_dlg->m_useSecureTyping;

	return InitV();
}

void cPageKBParams::UpdateRange( int id, int min, int pos, bool init )
{
	if ( m_dlg == NULL )
		return;

	if ( !init && ( m_dlg->Spin_GetPos( id ) > pos ) )
		m_dlg->Spin_SetPos( id, pos );

	m_dlg->Spin_SetRange( id, min, pos );
	if ( init )
		m_dlg->Spin_SetPos( id, min );
}

cPageKBParams *cDlgKBParams::CurPage()
{
	switch( m_kbParams->kbType )
	{
		case kbtPBE: return &PBE;
		case kbtPKE: return &PKE;
		case kbtSSS: return &SSS;
		default:
			assert( false );
			return &PBE;
	}
}

void cDlgKBParams::ShowItems()
{
	m_bSupportDrop = ( m_kbParams->kbType == ( kbtPBE ) );

	PBE.ShowItems( m_kbParams->kbType == kbtPBE );
	PKE.ShowItems( m_kbParams->kbType == kbtPKE );
	SSS.ShowItems( m_kbParams->kbType == kbtSSS );
}

int cDlgKBParams::OnNotify( NMHDR *phdr )
{
	if ( phdr->code == NM_CLICK ) 
	{
		if ( phdr->idFrom == IDC_btnAdvanced )
		{
			PostMessage( m_hDlg, WM_COMMAND, IDC_btnAdvanced, 0 );
			return 0;
		}
	}

	int res = SSS.Notify( phdr );

	if ( res != 0 )
		return res;

	return 0;
}

#define WM_ShowItems ( WM_USER + 2 )

bool cDlgKBParams::OnInitDialog()
{
	//
	//	It disables standard context menu.
	//
	m_bCallHelpInfoByClickingCtxHelpButton = true;

	SetDlgIcon( IDI_BCDoc, ICON_SMALL );

	if ( m_caption  != NULL ) SetWindowTextT( m_hDlg, m_caption );
	if ( m_fileName != NULL ) SetWindowTextT( GetItem( IDC_TextName), cWCS( m_fileName ).ChangeWords( L"&", L"&&" ) );

	kbd.GenerateEncodingTable( m_useSecureTyping );

	HWND hCmbType = GetItem( IDC_cmbType );

	// We have to attach both kbdEdits anyway.
	PBE.Init( this );

	HWND hFirstEdit = GetItem( IDC_pwdPBEA );
	SetFocus( hFirstEdit );

	bool bTravellerMode = ( ( m_keyGenFlags & FLAG_TRAVELER_MODE ) != 0 );
	EnableWindow( GetItem( IDC_btnAdvanced ), !bTravellerMode );

	if ( hCmbType == NULL ) // It is not advanced dialog
	{
		return true;
	}

	PKE.Init( this );
	SSS.Init( this );


	// Init key block type
	ComboBox_AddStringT( hCmbType, _LoadIDS( IDS_szPBE ) );
	ComboBox_AddStringT( hCmbType, _LoadIDS( IDS_szPKE ) );
	ComboBox_AddStringT( hCmbType, _LoadIDS( IDS_szSSS ) );

	ComboBox_SetCurSel( hCmbType, m_kbParams->kbType - 1 );

	// IDC_TextName is edit box instead of simple mode, so we have to set file name without changing & to && as we did above for static text.
	if ( m_fileName != NULL ) SetWindowTextT( GetItem( IDC_TextName), m_fileName );

	// Set the default push button to "Cancel" 
	SendMessage( m_hDlg, DM_SETDEFID, (WPARAM) IDCANCEL, (LPARAM) 0); 

	// It allows to set required focus after the dialog initialisation
	PostMessage( m_hDlg, WM_ShowItems, 0, 0 );
//	ShowItems();
	return true;
}

HWND FindSpinBuddy( HWND hDlg, HWND hEdit )
{
	HWND hItem = GetNextDlgGroupItem( hDlg, hEdit, false );

	if ( (HWND)SendMessage( hItem, UDM_GETBUDDY, 0, 0) == hEdit )
		return hItem;

	return NULL;
}

void UpdateSpinData( HWND hDlg, HWND hEdit )
{
	HWND hSpin = FindSpinBuddy( hDlg, hEdit );
	if ( hSpin == NULL )
		return;

	short min = 0,max = 0;
	cDlg::Spin_GetRange( hSpin, &min, &max );

	if ( !min && !max )
		return;

	char text[ 100 ];

	GetWindowTextA( hEdit, text, sizeof( text ) );
	int pos = atoi( text );

	if ( pos > max )
		pos = max;
	else
	if ( pos < min )
		pos = min;
	else
		return;

	cDlg::Spin_SetPos( hSpin, pos );
	SetFocus( hEdit );
}

BOOL cDlgKBParams::OnCommand( WPARAM wParam, LPARAM lParam )
{
	cPageKBParams *curPage = CurPage();

	WORD cmd = LOWORD( wParam );

	switch( cmd )
	{
		case IDC_btnAdvanced:
		{
			cPagePBE *pbePage = dynamic_cast < cPagePBE * > ( curPage );

			if ( pbePage == NULL )
			{	assert( false ); return false; 	}

			cKbdEdit::CheckAndGetPassword( &m_kbParams->kbPBE.m_pwd, &pbePage->m_edtA, true );
			cKbdEdit::CheckAndGetPassword( &m_kbParams->kbPBE.m_pwdConf, &pbePage->m_edtB, true );

			EndDialog( m_hDlg, ERROR_AdvancedDlgRequired );
			return true;
		}
		case IDCANCEL:
		{
			EndDialog( m_hDlg, (int)ERROR_CANCELLED_BY_USER); 
			return TRUE;
		}
		case IDC_btnKeyfiles:
		{
			cDlgKeyFileMan dlg( m_kbParams->kbPBE.m_keyFileList );
			if ( dlg.DoModal( m_hInst, IDD_KeyFilesManager, m_hDlg ) != IDOK )
				return false;

			m_kbParams->kbPBE.m_keyFileList.updateWindowText( curPage->m_dlg->GetItem( IDC_btnKeyfiles ) );
			return true;
		}
		default:
			break;
	}

	if ( HIWORD (wParam) == EN_KILLFOCUS )
		UpdateSpinData( m_hDlg, (HWND)lParam );
	else
	if ( HIWORD (wParam) == EN_SETFOCUS )
		HWND hEdit = (HWND)lParam;
	else
	if ( HIWORD (wParam) == CBN_SELCHANGE ) 
	{
		if ( LOWORD( wParam ) == IDC_cmbType )
		{
			m_kbParams->kbType = (ekbType)( ComboBox_GetCurSel( GetItem( IDC_cmbType )) + 1 );
			ShowItems();
		}
	}

	return curPage->OnCommand( wParam );
}

//virtual
HRESULT cDlgKBParams::Drop( cFilesList &ofl )
{
	//cPageKBParams *curPage = CurPage();
	//cPagePBE *pbePage = dynamic_cast < cPagePBE * > ( curPage );
	//if ( pbePage == NULL )
	//{	assert( false ); return S_FALSE; 	}

	bool bV8Compatible = ( Button_GetCheck( GetItem( IDC_chkV8Compatible ) ) == BST_CHECKED );
	if ( bV8Compatible )
	{
		MessageBoxT(
			NULL,
			_LoadIDS( IDS_szKeyFilesDoesNoEffectOnOldContainers ),
			_LoadIDS( IDS_WARNING ),
			MB_ICONEXCLAMATION  );
		return S_OK;
	}

	while( ofl.getNext() )
		m_kbParams->kbPBE.m_keyFileList.m_sa.addString( ofl.getCur());

	m_kbParams->kbPBE.m_keyFileList.updateWindowText( GetItem( IDC_btnKeyfiles ) );

	return S_OK;
}

LRESULT cDlgKBParams::RealDlgProc( UINT uMsg, WPARAM wParam, LPARAM lParam )
{
	cDropTarget::DlgProc( m_hDlg, uMsg );

	switch (uMsg) 
	{ 
		case WM_ShowItems:
			ShowItems();
			break;

        case WM_INITDIALOG:
			return OnInitDialog();

		case WM_userUpdateRanges:
			SSS.UpdateRanges();
			break;

		case WM_NOTIFY:
		{
			int res = OnNotify( (LPNMHDR)lParam );
			SetWindowLong( m_hDlg, DWL_MSGRESULT, res );
			return res;
		}
		case WM_DESTROY:

			PBE.Destroy();
			PKE.Destroy();
			SSS.Destroy();
			kbd.FreeTables( m_useSecureTyping );
			break;

        case WM_COMMAND: 
			return OnCommand( wParam, lParam );

		case WM_HELP:
			OnHelpInfo( (HELPINFO*)lParam );
			break;

		default:
			return 0;
    }

	return 0;
}

