#include "cDlgKBParams.h"
#include "resource.h" 
#include <windowsx.h>
#include "..\..\..\CertificateLib\hashAlgs\hashAlgs.h"
//#include "HELP\HelpIds.h"
#include "alg.h"
#include "..\..\..\Includes\cStringBuffer.h"
#include "KgGhost.h"

////////////////////////////
//
//	PBE Functions 
//

static int PBE_ItemTab[] = {
	IDC_stPBE1,
	IDC_stPBE2,
	IDC_stPBESalt,
	IDC_stPBE4,
	IDC_stPBE6,
	IDC_cmbPBE_HashAlg,
	IDC_edtPBESalt,
	IDC_chkV8Compatible,
	IDC_chkEncryptHeader,
	IDC_stEstimationTime,
	IDC_edtPBEIterations,
	IDC_pwdPBEA,
	IDC_linePBEA,
	IDC_pwdPBEB,
	IDC_linePBEB,
	IDC_grpType2,	
	IDC_btnPBEUpdateSalt,
	IDC_btnKeyfiles
};


void cPagePBE::ShowItems( bool show )
{
	static bool bAllowKeyFiles = true; // isDebug();

	for ( int i = 0; i < sizeof( PBE_ItemTab )/sizeof( int ); i++ )
	{
		ShowWindow( GetDlgItem( m_hDlg, PBE_ItemTab[i] ), show ? SW_SHOW : SW_HIDE );
	}

	if ( !bAllowKeyFiles )
		ShowWindow( GetDlgItem( m_hDlg, IDC_btnKeyfiles ), SW_HIDE );

	if ( !m_useSecureTyping )
	{
		ShowWindow( GetDlgItem( m_hDlg, IDC_linePBEA ), SW_HIDE );
		ShowWindow( GetDlgItem( m_hDlg, IDC_linePBEB ), SW_HIDE );
	}

	if ( show )
	{
		SetFocus( m_edtA.m_hEdit );
		bool enable = cKbdEdit::CheckAndGetPassword( &(m_dlg->m_kbParams->kbPBE.m_pwd ), &m_edtA, &m_edtB, true );
		EnableWindow( GetDlgItem( m_hDlg, IDOK ), enable );
	}
}

static cEnumEncModes modes;

static void updateModesCmb( HWND hCmb, DWORD algId, cDlgKBParams *dlg )
{
	ComboBox_ResetContent( hCmb );

	int def = 0;

	for ( int i = 0 ; i < modes.Q; i++ )
	{
		cSymAlgorithm cAlg;

		if ( !cAlg.Init( algId, modes.T[i].id ) )
			continue;

		if ( modes.T[i].id == dlg->m_kbParams->kbPBE.m_encModeId )
			def = i;

		ComboBox_InsertStringT( hCmb, i, _LoadIDS( modes.T[i].ids ) );
		//ComboBox_AddStringT( hCmb, dlg->LoadStringOnceT( modes.T[i].ids ) );
	}

	int err = ComboBox_SetCurSel( hCmb, def );
	
	dlg->m_kbParams->kbPBE.m_encModeId = modes.T[def].id;
}

const UINT maxIterations = 1000000000; // 0x7FFFffff;
const UINT minIterations = 256;

//
//	getIterations
//		return true is user assigned values was valid.
//
bool cPagePBE::getIterations( bool bSilent )
{
	HWND hWnd = m_dlg->GetItem( IDC_edtPBEIterations );
	if ( hWnd == NULL )
		return true;

	BOOL bTranslated = false;
	UINT ui = GetDlgItemInt( m_hDlg, IDC_edtPBEIterations, &bTranslated, false );

	bool bUseDefault = true;

	if ( !bTranslated || ( ui > maxIterations ) )
	{
		if ( GetWindowTextLength( hWnd ) == 0 )
			ui = minIterations;
		else
			ui = maxIterations;

		SetDlgItemInt( m_hDlg, IDC_edtPBEIterations, ui, false );
	}
	else
	if ( ui < minIterations )
	{
		ui = minIterations;
		SetDlgItemInt( m_hDlg, IDC_edtPBEIterations, ui, false );
	}
	else
		bUseDefault = false;

	m_dlg->m_kbParams->kbPBE.m_iterations = ui;

	if ( !bSilent && bUseDefault )
		m_dlg->DlgMessageBoxT( _FormatT( _LoadIDS( IDS_IterationCountChangedTo_D ), ui ), MB_ICONINFORMATION );

	return !bUseDefault;
}

bool cPagePBE::readSaltAndIterations()
{
	getIterations();

	HWND hEB = m_dlg->GetItem( IDC_edtPBESalt );
	if ( hEB == NULL )
		return true;

	BYTE *salt = m_dlg->m_kbParams->kbPBE.m_salt.buffer;
	int size = m_dlg->m_kbParams->kbPBE.getSaltSize();

	cWCStr wcs;
	wcs.GetWindowTextT( hEB );

	if ( string2bytes( salt, size, wcs.m_szBody ) == size )
		return true;

	m_dlg->DlgMessageBoxT( IDS_InvalidSaltSymbol, MB_ICONSTOP );
	SetFocus( hEB );
	return false;
}

void cPagePBE::writeSalt()
{
	HWND hEB = m_dlg->GetItem( IDC_edtPBESalt );
	if ( hEB == NULL )
		return;

	const BYTE *salt = m_dlg->m_kbParams->kbPBE.m_salt.buffer;
	int size = m_dlg->m_kbParams->kbPBE.getSaltSize();

	cWCSF wcs;

	for( int i = 0; i < size; i++ )
	{
		wcs.AddF( L"%02X ", salt[i] );
	}

	SetWindowTextT( hEB, wcs.m_szBody );
	SetWindowTextT( m_dlg->GetItem( IDC_stPBESalt ), _FormatT( _LoadIDS( IDS_Salt_D_Bytes ), size ) );
}

static int iPos_Sha256 = -1;
static bool isOldHash( int id )
{
	static int IDsTab[] = {
		pgphaMD5,
		pgphaSHA1,
		pgphaRIPEMD160,
		//bchaSHA256,
		//pgphaSHA512,
		//bchaWhirlpool512,
		//bchaSHA3_512
	};

	for( int i = maxLen( IDsTab ); i--; )
	{
		if ( IDsTab[i] == id )
			return true;
	}

	return false;
}

static const wchar_t *szBestCryptRegKey = L"Software\\Jetico\\BestCrypt";
static const wchar_t *szLastHashName = L"LastHashName";
static const wchar_t *szLastIterations = L"LastIterations";

static bool getDefaultValues( const wchar_t **pAlgName, int *pIterations )
{
	static cWCStr wcs;
	if ( wcs.isEmpty() )
	{
		if ( !wcs.RegGetString( HKEY_CURRENT_USER, szBestCryptRegKey, szLastHashName ) )
			wcs.RegGetString( HKEY_LOCAL_MACHINE, szBestCryptRegKey, szLastHashName );
	}
	
	if ( wcs.isEmpty() )
		*pAlgName = NULL;
	else
		*pAlgName = wcs.m_szBody;

	static int iterations = -1;
	if ( iterations == -1 )
	{
		DWORD dwType;
		DWORD cbData = sizeof( iterations );
		if ( RegGetValueT( HKEY_CURRENT_USER, szBestCryptRegKey, szLastIterations, 0, &dwType, &iterations, &cbData ) != ERROR_SUCCESS )
		{
			cbData = sizeof( iterations );
			RegGetValueT( HKEY_CURRENT_USER, szBestCryptRegKey, szLastIterations, 0, &dwType, &iterations, &cbData );
		}
	}

	if ( iterations != -1 )
		*pIterations = iterations;

	return true;
}

static bool setDefaultValues( const wchar_t *algName, int iterations )
{
	bool ok = cWCStr::RegSetString( HKEY_CURRENT_USER, szBestCryptRegKey, szLastHashName, algName );
	ok &= ( RegSetValueT( HKEY_CURRENT_USER, szBestCryptRegKey, szLastIterations, REG_DWORD, &iterations, sizeof( iterations ) ) == ERROR_SUCCESS );
	if ( ok )
		appFlags.SetFlag( efUseV9Default, false );

	return ok;
}

void cPagePBE::initEditControls()
{
	ODSFunc;

	m_edtA.Attach( GetDlgItem( m_hDlg, IDC_pwdPBEA ), GetDlgItem(m_hDlg, IDC_linePBEA) );
	m_edtB.Attach( GetDlgItem( m_hDlg, IDC_pwdPBEB ), GetDlgItem(m_hDlg, IDC_linePBEB) );

	if ( !m_useSecureTyping )
	{
		ShowWindow( GetDlgItem(m_hDlg, IDC_linePBEA), SW_HIDE );
		ShowWindow( GetDlgItem(m_hDlg, IDC_linePBEB), SW_HIDE );
	}

	m_edtA.addSymbols( m_dlg->m_kbParams->kbPBE.m_pwd.wcBuffer );
	m_edtB.addSymbols( m_dlg->m_kbParams->kbPBE.m_pwdConf.wcBuffer );
}

void cPagePBE::EnableWindowIteration()
{
	bool bEncryptHeader = Button_GetCheck( m_dlg->GetItem( IDC_chkEncryptHeader ) ) == BST_CHECKED;
	bool bV8Compatible = isV8Compatible();

	bool bEnable = m_dlg->m_kbParams->m_allowChangeAlgorithms && !bEncryptHeader && !bV8Compatible;;
	EnableWindow( m_dlg->GetItem( IDC_edtPBEIterations ), bEnable );
}

//
//	initKeyStrechingOptions
//		Key streching options are: hash algorithm, iterations and salt.
//		Gets default values and fills controls according the defalt.
//
void cPagePBE::initKeyStrechingOptions()
{
	ODSFunc;
	int defIterations = -1;
	const wchar_t *defHashAlgName = NULL;
	int defHashId = -1;
	getDefaultValues( &defHashAlgName, &defIterations );

	int def = -1, def0 = -1;

	HWND hCmb = m_dlg->GetItem( IDC_cmbPBE_HashAlg );

	cEnumHashAlgIDs hashAlgs;

	if ( hCmb == NULL )
	{
		// Bug #1817 Create container without opening Advanced Settings

		ODS_fls( _FormatT( L"Default hash alg name %s", defHashAlgName ) );

		defHashId = hashAlgs.getAlgId( getMBStr( defHashAlgName ) );
	}
	else
	{
		iPos_Sha256 = -1;

		for ( int i = 0; i<hashAlgs.Q; i++ )
		{
			int hashAlgId = hashAlgs.T[i];

			if ( isOldHash( hashAlgId ) )
				continue;

			cWCStr wcsHashAlgName( getWCStr( ha_getHashAlgName( (BYTE)hashAlgs.T[i] ) ) );

			int pos = ComboBox_AddStringT( hCmb, wcsHashAlgName.m_szBody );
			if ( pos < 0 )
			{ assert( false ); continue; }

			if ( ( defHashAlgName != NULL ) && wcsHashAlgName.equ( defHashAlgName ) )
				def0 = pos;
			else
			if ( ( m_dlg->m_kbParams->kbPBE.m_hashAlgId == hashAlgId ) && ( def == -1 ) )
				def = pos;

			if ( ( hashAlgId == pgphaSHA256 ) ||
				 ( hashAlgId == bchaSHA256 ) )
					iPos_Sha256 = pos;
		}

		//
		// Version 2.00.2 - 06.01.2015
		//	If allow to change algorithm so last saved is preferable
		//
		if ( ( m_dlg->m_kbParams->m_allowChangeAlgorithms ) && ( def0 != -1 ) )
			def = def0;

		if ( def == -1 ) { assert( false ); def = def0; }
		if ( def == -1 ) { assert( false ); def = 0; }

		ComboBox_SetCurSel( hCmb, def );
		EnableWindow( hCmb, m_dlg->m_kbParams->m_allowChangeAlgorithms );
	}

	//
	// Version 2.00.1 - 15.12.2014
	//	Checks m_allowChangeAlgorithms before changing m_dlg->m_kbParams->kbPBE.m_iterations
	//
	if ( m_dlg->m_kbParams->m_allowChangeAlgorithms )
	{
		if ( defIterations != -1 )
			m_dlg->m_kbParams->kbPBE.m_iterations = defIterations;

		if ( defHashId != -1 )
			m_dlg->m_kbParams->kbPBE.m_hashAlgId = defHashId;

		m_dlg->m_kbParams->kbPBE.setSaltSize();
	}

	HWND hWnd = m_dlg->GetItem( IDC_edtPBEIterations );

	if ( hWnd != NULL )
	{
		//EnableWindow( hWnd, m_dlg->m_kbParams->m_allowChangeAlgorithms );
		EnableWindowIteration();
		EnableWindow( m_dlg->GetItem( IDC_chkV8Compatible ), m_dlg->m_kbParams->m_allowChangeAlgorithms );
		setIterations();
	}

	if ( m_dlg->m_kbParams->m_allowChangeAlgorithms && 
		!isDefaultIterations( m_dlg->m_kbParams->kbPBE.m_iterations ) ) // Not defualt iterations
	{
		if ( hWnd == NULL ) // Not advansed view
			PostMessage( m_dlg->m_hDlg, WM_COMMAND, MAKELONG( IDC_btnAdvanced, 0 ), 0 ); 
		else
		{
			// Assign here for checkUpdateIterations show warning, because setIterations() already change this control, see aboth
			m_dlg->m_kbParams->kbPBE.m_iterations = DEFAULT_ITERATIONS_V5;
			ShowWindow( m_hDlg, SW_SHOW );
			PostMessage( m_hDlg, WM_COMMAND, MAKELONG( IDC_edtPBEIterations, EN_CHANGE ), 0 );
			//checkUpdateIterations();
		}
	}

	writeSalt();

	hWnd = m_dlg->GetItem( IDC_btnKeyfiles );
	m_dlg->m_kbParams->kbPBE.m_keyFileList.updateWindowText( hWnd );
}

bool cPagePBE::InitV()
{
	ODSFunc;

	//
	//	By default this flag is true, and is reset after first user parameters are saved ( see setDefaultValues ).
	//	Version 2.01.5.3 - 13.12.2016
	//		At hidden part creating and changing password the m_dlg->m_kbParams->m_allowChangeAlgorithms is false,
	//		But is was not checked here. As result new key was generated with default parameters and was not opened if 
	//		current parameters are not default.
	//
	if ( appFlags.GetFlag( efUseV9Default ) && m_dlg->m_kbParams->m_allowChangeAlgorithms )
		m_dlg->m_kbParams->setBCV9Default();

	//
	//	This dialog does not have 'Help' button, it uses the '?' icon from dialog system menu instead.
	//	So context menu is disabled  ->//	m_MenuHelpIDs = (DWORD*)MenuHelpIDsT;
	//	The m_helpPath is assigned and cDlg::m_bCallHelpInfoByClickingCtxHelpButton = true.
	//
	m_helpPath = L"/HTML_help/html/08_1_new_key_generator/01_new_key_generator.htm";

	initEditControls();
	initKeyStrechingOptions();

	OnCheckEncryptHeader( true );
	OnCheckV8Compatibility( true );

	Start_TimeEstimationThread();

	return true;
}

bool cPagePBE::isV8Compatible()
{
	return Button_GetCheck( m_dlg->GetItem( IDC_chkV8Compatible ) ) == BST_CHECKED;
}

void cPagePBE::setIterations()
{
	SetDlgItemInt( m_hDlg, IDC_edtPBEIterations, m_dlg->m_kbParams->kbPBE.m_iterations, false );
}

void cPagePBE::Destroy()
{
	m_edtA.Detach();
	m_edtB.Detach();
}


void cPagePBE::Wait_TimeEstimationThread()
{
	ODSFunc;

	if ( m_hTread != NULL )
	{
		DWORD dwRes = WaitForSingleObject( m_hTread, INFINITE );
		CloseHandle( m_hTread );
		ODSFuncRes( dwRes );
		m_hTread = NULL;
	}
}

void cPagePBE::Start_TimeEstimationThread()
{
	ODSFunc;

	Wait_TimeEstimationThread();

	m_hTread = CreateThread( NULL, 0, estimateTimeStatic, this, 0, NULL );
}

DWORD cPagePBE::estimateTime()
{
	ODSFunc;

	cParamsPBE &pbe = m_dlg->m_kbParams->kbPBE;

	cSymAlgorithm sPwdAlg;
	if ( !sPwdAlg.Init( pbe.m_algId, pbe.m_encModeId ) )
	{
		ODS_fls( L"Assertyion!" );
		return ERROR_INVALID_ALGORITHM;
	}

	cPlainTextBlock ptb;
	int pwdKeyBuffSize = ( sPwdAlg.m_keyLengthInBites + 7 ) / 8;
	if ( ptb.Allocate( pwdKeyBuffSize ) == NULL )
	{
		ODS_fls( L"Assertyion!" );
		return ERROR_INTERNAL_PROBLEM;
	}

	cPassword pwd( L"Test" );

	clock_t start = clock();
	
	pbe.generateKeyFromPassword( ptb.buffer, pwdKeyBuffSize, &pwd, pbe.m_hashAlgId, pbe.m_salt.buffer, pbe.getSaltSize(), 256 );

	m_timeFor256Iterations = (DWORDLONG)clock() - start;

	if ( m_timeFor256Iterations < 10 )
	{
		start = clock();
		pbe.generateKeyFromPassword( ptb.buffer, pwdKeyBuffSize, &pwd, pbe.m_hashAlgId, pbe.m_salt.buffer, pbe.getSaltSize(), 16 * 1024 );
		m_timeFor16KIterations = (DWORDLONG)clock() - start;
	}

	PostMessage( m_hDlg, WM_COMMAND, MAKELONG( IDC_edtPBEIterations, EN_CHANGE ), 0 );

	ODSFuncRes( _FormatT( L"256 - %I64d, 16K -  %I64d" , m_timeFor256Iterations, m_timeFor16KIterations ) );
	return 0;
}

time_t cPagePBE::setEstimationTime()
{
	if ( m_dlg == NULL )
	{
		//assert( false );
		return 0;
	}

	HWND hWnd = m_dlg->GetItem( IDC_edtPBEIterations );
	if ( hWnd == NULL )
		return 0;

	BOOL bTranslated = false;
	UINT ui = GetDlgItemInt( m_hDlg, IDC_edtPBEIterations, &bTranslated, false );
	
	if ( !bTranslated )
		return 0;

	//
	// m_timeFor256Iterations is calculated by separate thread, see Start_TimeEstimationThread
	//
	time_t res = 0;
	cWCStr wcs;
	if ( m_timeFor16KIterations != 0 )
	{
		res = (time_t)( m_timeFor16KIterations * ui / ( 16 * 1024 * CLOCKS_PER_SEC ) );
		wcs.GetTimeText( res );
	}
	else
	if ( m_timeFor256Iterations != 0 )
	{
		res = (time_t)( m_timeFor256Iterations * ui / ( 256 * CLOCKS_PER_SEC ) );
		wcs.GetTimeText( res );
	}
	else
		wcs.LoadIDS( IDS_isNotReady );

	SetWindowTextT( m_dlg->GetItem( IDC_stEstimationTime ), wcs.m_szBody );

	return res;
}

void cPagePBE::OnCheckEncryptHeader( bool bInit )
{
	ODSFunc;

	HWND hWnd = m_dlg->GetItem( IDC_chkEncryptHeader );
	if ( hWnd == NULL )
		return;

	if ( bInit )
	{
		bool bChecked = false;

		if ( !m_dlg->m_kbParams->m_allowChangeAlgorithms )
			EnableWindow( hWnd, false );
		else
			bChecked = !appFlags.GetFlag( efDoNotEncryptHeader );

		Button_SetCheck( hWnd, bChecked ? BST_CHECKED : BST_UNCHECKED );
	}

	bool bEncryptHeader = Button_GetCheck( hWnd ) == BST_CHECKED;
	bool bV8Compatible = isV8Compatible();

	if ( bEncryptHeader )
	{
		m_dlg->m_kbParams->kbPBE.m_iterations = bV8Compatible ? DEFAULT_ITERATIONS_V4 : DEFAULT_ITERATIONS_V5;
		setIterations();
	}

	//bool bEnable = m_dlg->m_kbParams->m_allowChangeAlgorithms && !bEncryptHeader;
	//EnableWindow( m_dlg->GetItem( IDC_edtPBEIterations ), bEnable );
	EnableWindowIteration();
}

void cPagePBE::OnCheckV8Compatibility( bool bInit )
{
	ODSFunc;

	HWND hWnd = m_dlg->GetItem( IDC_chkV8Compatible );
	if ( hWnd == NULL )
		return;

	if ( bInit )
	{
		if ( !m_dlg->m_kbParams->m_allowChangeAlgorithms )
			EnableWindow( hWnd, false );

		if ( !m_dlg->m_kbParams->kbPBE.isKeyGenV5() )
			Button_SetCheck( hWnd, BST_CHECKED );
	}

	bool bV8Compatible = ( Button_GetCheck( hWnd ) == BST_CHECKED );

	int saltSize = 8;
	if ( bV8Compatible )
	{
		bool bHasKeyFiles = m_dlg->m_kbParams->kbPBE.m_keyFileList.hasFiles();

		if ( bHasKeyFiles )
		{
			int res = m_dlg->DlgMessageBoxT( IDS_szHasKeyFilesQ, MB_YESNO | MB_ICONQUESTION );
			if ( res == IDNO )
			{
				Button_SetCheck( hWnd, BST_UNCHECKED );
				return;
			}

			m_dlg->m_kbParams->kbPBE.m_keyFileList.Clear();
			m_dlg->m_kbParams->kbPBE.m_keyFileList.updateWindowText( m_dlg->GetItem( IDC_btnKeyfiles  ) );
		}

		m_dlg->m_kbParams->kbPBE.m_iterations = DEFAULT_ITERATIONS_V4;
		m_dlg->m_kbParams->setHashAlgId( bchaSHA256 );

		if ( iPos_Sha256 == -1 ) // position of pgphaSHA256
			assert( false );
		else
			ComboBox_SetCurSel( m_dlg->GetItem( IDC_cmbPBE_HashAlg ), iPos_Sha256 );
		
		setIterations();
	}
	else
		ha_getDigestSize( &saltSize, m_dlg->m_kbParams->kbPBE.m_hashAlgId );

	EnableWindow( m_dlg->GetItem( IDC_btnKeyfiles  ), !bV8Compatible );

	m_dlg->m_kbParams->kbPBE.setSaltSize( saltSize );
	writeSalt();
	

	bool bEnable = m_dlg->m_kbParams->m_allowChangeAlgorithms && !bV8Compatible;

	// V8 allows to change salt,  it is saved at every PBE slot, instead of V9 uses common salt at 
	EnableWindow( m_dlg->GetItem( IDC_btnPBEUpdateSalt ), m_dlg->m_kbParams->m_allowChangeAlgorithms || bV8Compatible );
	EnableWindow( m_dlg->GetItem( IDC_cmbPBE_HashAlg ), bEnable );
	EnableWindowIteration();
	//EnableWindow( m_dlg->GetItem( IDC_edtPBEIterations ), bEnable );
}

static void setFlagEncryptHeader( HWND hWnd )
{
	if ( hWnd == NULL )
		return;

	bool bEncryptHeader = Button_GetCheck( hWnd ) == BST_CHECKED;
	appFlags.SetFlag( efDoNotEncryptHeader, !bEncryptHeader );
}

bool cPagePBE::checkUpdateIterations(  bool bSilent )
{
	int oldIterations = m_dlg->m_kbParams->kbPBE.m_iterations;
	bool bValidValue = getIterations( bSilent );
	setEstimationTime();

	if ( !isDefaultIterations( oldIterations ) )
		return bValidValue;

	int newIterations = m_dlg->m_kbParams->kbPBE.m_iterations;

	if ( isDefaultIterations( newIterations ) )
		return bValidValue;

	static bool bShowMessage = true;
	if ( !bShowMessage )
		return bValidValue;

	cWCStr text;
	text.FormatT( _LoadIDS( IDS_DefaultIterationForEncryptedHeader_S ), DEFAULT_ITERATIONS_V5 );

	cMBTextDlg dlg( text.m_szBody, IDC_Text, bShowMessage, _LoadIDS( IDS_showThisMessageNextTime ), IDC_chk, _LoadIDS( IDS_Attention ) );
	int res = dlg.DoModal( m_dlg->m_hInst, IDD_ChkYesNo, m_hDlg );

	bShowMessage = dlg.m_bChk;

	if ( res == IDYES )
		return bValidValue;

	m_dlg->m_kbParams->kbPBE.m_iterations = DEFAULT_ITERATIONS_V5;
	setIterations();
	return false;
}

bool cPagePBE::OnCommand( WPARAM wParam )
{
	// Set the default push button to "OK" when the user enters text 
	if( HIWORD (wParam) == EN_CHANGE ) 
	{
		//if ( LOWORD( wParam ) == IDC_edtPBEIterations )
		//	checkUpdateIterations();
		setEstimationTime();

		SendMessage( m_hDlg, DM_SETDEFID, (WPARAM) IDOK, (LPARAM) 0);
		EnableWindow( GetDlgItem( m_hDlg, IDOK ), true );
	}
	else
	if ( HIWORD (wParam) == CBN_SELCHANGE ) 
	{
		if ( LOWORD( wParam ) == IDC_cmbPBE_HashAlg )
		{
			HWND hCmb = GetDlgItem( m_hDlg, LOWORD( wParam ) );
			char name[ 128 ];
			GetWindowTextA( hCmb, name, maxLen( name ) );

			cEnumHashAlgIDs &hashAlgIds = cEnumHashAlgIDs::getStatic();
			int id = hashAlgIds.getAlgId( name );

			if ( id == -1 )
			{
				assert( false );
				return false;
			}

			m_dlg->m_kbParams->setHashAlgId( id );
			writeSalt();

			Start_TimeEstimationThread();

			return true;
		}
	}

	switch(LOWORD( wParam )) 
	{ 
		case IDC_chkV8Compatible:
		{
			OnCheckV8Compatibility();
			return 0;
		}

		case IDC_chkEncryptHeader:
		{
			OnCheckEncryptHeader();
			return 0;
		}

		case IDC_btnPBEUpdateSalt:
		{
			m_dlg->m_kbParams->kbPBE.updateSalt();
			writeSalt();
			return 0;
		}
		case IDOK: 
			if ( !checkUpdateIterations( false ) )
				return false;

			return OnOK();

		default:
			return 0; 
	} 
}

bool cPagePBE::OnOK()
{
	if ( !readSaltAndIterations() )
		return false;

	time_t estTime = setEstimationTime();
	if ( estTime > 5 * 60 )
	{
		int res = m_dlg->DlgMessageBoxT( _FormatT( _LoadIDS( IDS_PasswordGenerationTimeIs_S ), cWCStr().GetTimeText( estTime ) ), MB_YESNO | MB_ICONQUESTION );
		if ( res == IDNO )
		{
			SetFocus( m_dlg->GetItem( IDC_edtPBEIterations ) );
			return false;
		}
	}

	if ( !cKbdEdit::CheckAndGetPassword( &(m_dlg->m_kbParams->kbPBE.m_pwd ), &m_edtA, &m_edtB ) )
		return false;

	setDefaultValues( getWCStr( ha_getHashAlgName( (BYTE)m_dlg->m_kbParams->kbPBE.m_hashAlgId ) ), m_dlg->m_kbParams->kbPBE.m_iterations );

	if ( m_dlg->m_kbParams->m_allowChangeAlgorithms )
		setFlagEncryptHeader( m_dlg->GetItem( IDC_chkEncryptHeader ) );

	EndDialog( m_hDlg, ERROR_NO );

	return true;
}

//
//	End of PBE functions
//
//////////////////////////
