#include "cDlg.h"
#include "cKeyBlockParams.h"
#include "cDlgKbParams.h"
#include "pwdlg.h"
#include "resource.h"
#include "..\..\..\CertificateLib\PGPPkts\PKTDEFS.H"
#include "..\..\..\CertificateLib\x509pkts\pkcs12_funcs.h"
#include "..\..\..\CertificateLib\symmetricAlgs\symmetricAlgs.h"
#include <assert.h>
#include "ods.h"
#include "KgGhost.h"

bool cKeyBlockParams::KeyBlockParamsInit(
		DATA_BLOCK *dataBlockArray,
		DWORD dataBlockArraySizeInBytes,
		int		contAlgId,		// = -1
		DWORD	encMode,		// = -1
		int		contHashAlgId,	// = -1
		int		iterations,		// = DEFAULT_ITERATIONS_V5
		DWORD	*pErrorCode,	// = NULL
		DWORD	*pFlags,		// = NULL
		HWND	hWndParent		// = NULL
	)
{
	ODSFunc;

	if ( iterations == -1 )
		iterations = DEFAULT_ITERATIONS_V5;

	dataBlock = dataBlockArray;
	dataBlockSize = dataBlockArraySizeInBytes;

	// kbSSS may ask password for additional shares, so it needs in the parameters.
	if ( pErrorCode )	kbSSS.pErrorCode = pErrorCode;
	if ( pFlags )		kbSSS.pFlags = pFlags;
	if ( hWndParent )	kbSSS.hWnd = hWndParent;

	kbPBE.m_algId = contAlgId;
	kbPBE.m_hashAlgId = contHashAlgId;
	kbPBE.m_encModeId = encMode;
	kbPBE.m_iterations = iterations;

	//
	//
	// See http://nsk.jetico.com/redmine/projects/best-crypt-for-windo/wiki/Wiki
	//
	if ( dataBlock != NULL )
	{
		dbHeader &dbh = dataBlockArray->header.m_body;

		bool bIsEncrypted = dbHeader::isEncrypted( dbh );

		if ( !bIsEncrypted )
		{
			kbPBE.m_iterations = dbHeader::getIterationCount( dbh );

			if ( contAlgId == -1 )
			{
				kbPBE.m_algId = dbh.contAlgId;
				kbPBE.m_hashAlgId = dbh.hashAlgId;
				kbPBE.m_encModeId = dbh.encModeId;
			}

			// Read salt from DataBlock to this kbParams
			dataBlock->updateSalt( this, true, dbh.wKeyGenId );
		}
		else
			dataBlock->updateSalt( this, true, KEYGEN_ID_V5 );
	}
	
	if ( kbPBE.m_algId == -1 )
		return false;

	return algInfo.Init( kbPBE.m_algId, encMode );
}

cKeyBlockParams::cKeyBlockParams()
{
	ODSFuncS( _FormatT( L"PBE address %X", &kbPBE ) );

	kbType = kbtEmpty;
	kbPBE.m_pKBParams = this;
	m_allowChangeAlgorithms = false;
}

cKeyBlockParams::~cKeyBlockParams()
{
	ODSFunc;
}

bool cKeyBlockParams::decodeKey( sKeyInfo *keyBuff, int keyBuffSize, BYTE *kbEncoded, int kbEncodedSize, bool checkPasswordOnly )
{
	ODSFunc;
	bool res = false;

	ekbType saved = kbType;

//	bool checkPasswordOnly = false;

	// Checks for all types
	if ( checkPasswordOnly )
		kbType = kbtEmpty;

	kbSSS.encodedSize = 0;

	switch ( kbType )
	{
		case kbtPBE:
			res = kbPBE.decodeKey( (BYTE *)keyBuff, keyBuffSize, kbEncoded, kbEncodedSize, algInfo.m_algBlockSize );
#ifdef _DEBUG
//#define _Michael_Burley_Problem
#endif
#ifdef _Michael_Burley_Problem
			//
			// Michael Burley encounter a problem.
			// The 0x00A7 symbol has been read as 0xFFA7 early
			// So old containers cannot be opened again.
			//
			// If decoding result is not true, we'll check the 0x00A7 symbol, change it 0xFFA7 
			//	and try to decode again. 
			//	The changed password will be restored.
			//
			if ( !res )
			{
				int len = (int)wcslen(kbPBE.m_pwd.wcBuffer);
				int pos = -1;
				bool found = false;
				bool foundB5 = false;
				int pos_0402 = -1;
				for ( pos = len; pos--; )
				{
					if ( kbPBE.m_pwd.wcBuffer[pos] == 0x00A7 )
					{
						found = true;
						kbPBE.m_pwd.wcBuffer[pos] = 0xFFA7;
					}
					else
					if ( kbPBE.m_pwd.wcBuffer[pos] == 0x00B5 )
					{
						foundB5 = true;
						kbPBE.m_pwd.wcBuffer[pos] = 0xFFB5;
					}
					else
					if ( kbPBE.m_pwd.wcBuffer[pos] == 0x0402 )
					{
						pos_0402 = pos;
					}
				}

				if ( found || foundB5 )
				{
					wchar_t path[ MAX_PATH ];

					wchar_t f1 = -1, f3 = -1;

					if ( ODS_GetFileName( path, maxLen( path ) ) )
					{
						ODS_CreateNewFile( false );

						// Read Last Values 
						HANDLE hFile = CreateFileT( path, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, 0, NULL );
						if ( hFile != INVALID_HANDLE_VALUE )
						{
							wchar_t bf[ 100 ];
							int len = maxLen( bf );
							DWORD dwRead;

							SetFilePointer( hFile, -( ( len - 1 ) * sizeof( bf[0] ) ) , NULL, FILE_END );
							bool ok = ReadFile( hFile, (BYTE*)bf, (len - 1) * sizeof( bf[0] ), &dwRead, NULL );

							CloseHandle( hFile );

							if ( ok )
							{
								
								len = dwRead / sizeof( bf[0] );

								// 33:901 pwd[1] - 0000, pwd[3] - 00C9

								if ( len > 35 )
								{
									bf[ --len ] = L'\0';
									bf[ --len ] = L'\0';
								
									wchar_t *ptr = wcsrchr( bf, L':' );
									if ( ptr != NULL )
									{
										cMBStr mbs( ptr );
										DWORD t, dw1, dw3;
										sscanf( mbs.m_szBody, ":%d pwd[1] - %x, pwd[3] - %x", &t, &dw1, &dw3 );

										int res = sscanf( mbs.m_szBody, ":%d pwd[1] - %x, pwd[3] - %x", &t, &dw1, &dw3 );
										if ( res == 3 )
										{
											f1 = dw1;
											f3 = dw3;

											ODS( FormatOnceT( L"Found processed: pwd[1] - %04X, pwd[3] - %04X", f1, f3 ) );
										}
									}
								}
							}
						}
					}
					else
					{
						ODS_Enable();
						ODS( L"Found unicode symbols" );
					}

					// It tries to add FF to char buffer
					// @~|I.just love and adore Donna-she is so cute and.Cuddly\}][{


					if ( !foundB5 )
						res = kbPBE.decodeKey( (BYTE *)keyBuff, keyBuffSize, kbEncoded, kbEncodedSize, algInfo.m_algBlockSize );
					else
					{
						wchar_t wc1 = kbPBE.m_pwd.wcBuffer[1];
						wchar_t wc3 = kbPBE.m_pwd.wcBuffer[3];

						if ( ( f1 != -1 ) && ( f3 != -1 ) )
						{
							kbPBE.m_pwd.wcBuffer[1] = f1;
							kbPBE.m_pwd.wcBuffer[3] = f3;
						}

						for( int i = 0x10000; i--; )
						{
							ODS( FormatOnceT( L"pwd[1] - %04X, pwd[3] - %04X", (DWORD)kbPBE.m_pwd.wcBuffer[1] , (DWORD)kbPBE.m_pwd.wcBuffer[3] ) );

							for( int j = 0x10000; j--; )
							{
								res = kbPBE.decodeKey( (BYTE *)keyBuff, keyBuffSize, kbEncoded, kbEncodedSize, algInfo.m_algBlockSize );
								if ( res )
								{
									ODS( FormatOnceT( L"Found !!! pwd[1] - %04X, pwd[3] - %04X", (DWORD)kbPBE.m_pwd.wcBuffer[1] , (DWORD)kbPBE.m_pwd.wcBuffer[3] ) );
									break;
								}

								kbPBE.m_pwd.wcBuffer[3] += 0x1;
								
							}

							if ( res )
								break;
							
							kbPBE.m_pwd.wcBuffer[1] += 0x1;
						}
						
						kbPBE.m_pwd.wcBuffer[1] = wc1;
						kbPBE.m_pwd.wcBuffer[3] = wc3;
					}

					for ( pos = len; pos--; )
					{
						if ( kbPBE.m_pwd.wcBuffer[pos] == 0xFFA7 )
							kbPBE.m_pwd.wcBuffer[pos] = 0x00A7;
						else
						if ( kbPBE.m_pwd.wcBuffer[pos] == 0xFFB5 )
							kbPBE.m_pwd.wcBuffer[pos] = 0x00B5;
					}
				}
			}
#endif
			break;

		case kbtPKE:

			//	See 'Secret key notes' at DataBlockVerifyKeyBlockAndGetKey.

			if ( kbPKE.ska == NULL )
				kbPKE.openSecretKey( kbSSS.hWnd, &kbPBE.m_pwd, kbEncoded, true );

			if ( kbPKE.ska == NULL )
				res = false;
			else
				res = kbPKE.decodeKey( (BYTE *)keyBuff, keyBuffSize, kbEncoded, kbEncodedSize );

			break;
		case kbtSSS:
ODS_fl;
			kbSSS.SetEncIds( &kbPBE );

			kbSSS.version = -1;

ODS_fl;
			if ( kbSSS.itemQ != 0 )
				res = kbSSS.decodeKey( (BYTE *)keyBuff, keyBuffSize, kbEncoded, kbEncodedSize );
			else
			{
ODS_fl;
				kbSSS.encodedSize = kbEncodedSize;
				res = kbSSS.decodeKeyWithFirstPwd( &kbPBE.m_pwd, (BYTE *)keyBuff, keyBuffSize, kbEncoded, &(kbSSS.encodedSize), checkPasswordOnly );
			}
ODS_fl;
			break;
		case kbtEmpty:
		{
			//
			//	Note, if kbParams type is undefined ( kbtEmpty ), so all available 
			//	encoding shemes are implemented to DATA_BLOCK,
			//	see kbParams->decodeKey
			//

			res = kbPBE.decodeKey( (BYTE *)keyBuff, keyBuffSize, kbEncoded, kbEncodedSize, algInfo.m_algBlockSize );

			if ( res ) { saved = kbtPBE; break; }

			if ( kbPKE.ska != NULL )
			{
				//
				//	For future improvemen, the check function for encrypted ( or hidden part ) key block enumeration 
				//	may be moved to one loop above ( excluding PBE algorithm enumeration, including the key block enumertion only )
				//	In the case we have to change loops order key block enumeration must be first and algorithm verification second,
				//	but it may be slowly, because we have to load all algorithms for every block again.
				//
				res = kbPKE.decodeKey( (BYTE *)keyBuff, keyBuffSize, kbEncoded, kbEncodedSize );
				if ( res ) { saved = kbtPKE; break; }
			}

			//kbSSS.algId = kbPBE.m_algID; //keyAlgId;
			//kbSSS.encModeId = kbPBE.m_encModeId;
			//kbSSS.hashAlgId = kbPBE.m_hashAlgId;
			kbSSS.SetEncIds( &kbPBE );
			kbSSS.version = -1;

			kbSSS.encodedSize = kbEncodedSize;
			res = kbSSS.decodeKeyWithFirstPwd( &kbPBE.m_pwd, (BYTE *)keyBuff, keyBuffSize, kbEncoded, &kbSSS.encodedSize, checkPasswordOnly );
			if ( res ) { saved = kbtSSS; break; } 

/*
			if ( kbPKE.ska == NULL )
			{
				//
				//	Tries to get secret key again and demand
				//
				kbPKE.openSecretKey( kbSSS.hWnd, &kbPBE.m_pwd, kbEncoded );

				if ( kbPKE.ska != NULL )
				{	res = kbPKE.decodeKey( (BYTE *)keyBuff, keyBuffSize, kbEncoded, kbEncodedSize );
					if ( res ) { saved = kbtPKE; break; }
				}
			}
*/
			break;
		}
		default:
			res = false;
	}

	if ( res && ! checkPasswordOnly )
	{
		res = keyBuff->isValid();
	}

	kbType = saved;

	return res;
}

bool cKeyBlockParams::encodeKey( BYTE *kbEncoded, int kbEncodedSize, sKeyInfo *keyBuff, int keyBuffSize )
{
	bool res = false;

	assert( keyBuff->size <= keyBuffSize );

	switch ( kbType )
	{
		case kbtPBE: res = kbPBE.encodeKey( kbEncoded, kbEncodedSize, (BYTE *)keyBuff, keyBuffSize ); break;
		case kbtPKE: res = kbPKE.encodeKey( kbEncoded, kbEncodedSize, (BYTE *)keyBuff, keyBuffSize ); break;
		case kbtSSS: res = kbSSS.encodeKey( kbEncoded, kbEncodedSize, (BYTE *)keyBuff, keyBuffSize ); break;
		default:
		{
			assert( false );
			return false;
		}
	}

	return res;
}

int cKeyBlockParams::getEncodedKeySize( int keyLength )
{
	switch ( kbType )
	{
		case kbtPBE: return kbPBE.getEncodedKeySize( keyLength );
		case kbtPKE: return kbPKE.getEncodedKeySize( keyLength );
		case kbtSSS: return kbSSS.getEncodedKeySize( keyLength );
		default:
		{
			assert( false );
			return false;
		}
	}
}

//
//	Tries to get a password from key block parameters.
//	The password is required as client network password for corporate version.
//
bool cKeyBlockParams::getPassword( char *pwdBuff )
{
	assert( false );
	return false;
}