#include "Kblock.h"
#include "KgGhost.h"

//
//	KG-Ghost v5 allows to change iteration
//	1. It changes key gen identifier, so old key generator cannot open this container.
//		Otherwise it will report invalid password.
//	2. Uses the version field as iteration count. But keyGenId DWORD files uses as WORD wKeyGenId, WORD wVesrion.
//	3. Key generator ID - 5 does not allow to encrypt header, otherwise enumeration becames imposible.
//
//	See http://nsk.jetico.com/redmine/projects/best-crypt-for-windo/wiki/Wiki
//
bool dbHeader::setKeyGenIdV5( DWORD count )
{
	if ( isEncrypted(*this) )
	{	assert( false ); return false; }

	if ( wKeyGenId != KEYGEN_ID_V5 )
	{
		wKeyGenId = KEYGEN_ID_V5;
		wVersion = (WORD)version;
		assert( wVersion == version );
	}

	iterations = count;

	return true;
}

// static
DWORD dbHeader::getIterationCount( const dbHeaderBase &dbh )
{
	if ( !isEncrypted( dbh ) && ( dbh.wKeyGenId == KEYGEN_ID_V5 ) )
		return dbh.iterations;
	else
		return DEFAULT_ITERATIONS_V4;
}

void dbHeader::setVersion( WORD ver )
{
	if ( wKeyGenId == KEYGEN_ID_V5 )
		wVersion = ver;
	else
	{
		version = ver;
		wVersion = 0;
	}
}

int dbHeader::getVersion()
{
	if ( wKeyGenId == KEYGEN_ID_V5 )
		return wVersion;
	else
		return version;
}

void dbHeader::InitFirst()
{
	memset( busiedFlag, 0, sizeof( busiedFlag ) );
	memcpy( volumeLabel, dbVolumeLabel, sizeof( volumeLabel ) );
	memcpy( signature, dbSignature, sizeof( signature ) );
	memset( busiedName, 0, sizeof( busiedName ) );
}

void dbHeader::Init()
{
	memset( this, 0, sizeof( *this ) );
	InitFirst();
	version = DB_FormatVersion;
}

// static 
DWORD dbHeader::checkFreeSpace(
		dbKeyId *keyMap,
		DWORD requiredSize,
		DWORD *pMaxFreeSize,
		int treatAsFree // = -1
	)
{
	int maxFound = 0;

	int pos = dbHeader::getMaxEmptyKeys( keyMap, DB_maxKeyQ, &maxFound, treatAsFree );

	if ( pMaxFreeSize != NULL )
		*pMaxFreeSize = (DWORD)maxFound * DB_MinEncryptedKeyBlockSize;

	if ( (DWORD)maxFound * DB_MinEncryptedKeyBlockSize < requiredSize )
		return	ERROR_NOT_ENOUGH_SPACE_FOR_KEY;

	return ERROR_NO;
}

//
//	isEncrypted
//		The function allows ambiguity so we have to check if cont 
//		info is really encrypted after encryption!!!
//
// static 
bool dbHeader::isEncrypted( const dbHeaderBase &dbh )
{
	return !Alg_GetAlgInfoByID( dbh.contAlgId, dbh.encModeId ) && 
			( strncmp( dbh.volumeLabel, "BC_KeyGenID", sizeof( dbh.volumeLabel ) )!= 0 ) ;
}

// static 
//	Overwrites the ketTab item only. The header is not modified so it is not required to be updated.
bool dbHeader::shredKeyTab( DATA_BLOCK *dataBlock, DWORD dbSize, sKeyMapInfo *kmi )
{
	if ( kmi->position < 0 )
		return false;

	int maxKeyQ = DATA_BLOCK::getKeyQ( dbSize );
	int quantity = kmi->quantity;
	if ( kmi->position >= maxKeyQ )
		return false;

	int tailQ = maxKeyQ - kmi->position;

	if ( quantity > tailQ )
		quantity = tailQ;

	ShredData( dataBlock->keyTab[ kmi->position ], quantity * DB_MinEncryptedKeyBlockSize );

	return true;
}

// static 
void dbHeader::freeKeyBlockChain( DATA_BLOCK *dataBlock, DWORD dbSize, dbKeyId *map, sKeyMapInfo *kmi )
{
	if ( ( kmi == NULL ) || ( kmi->position == -1 ) )
		return;

	int quantity = freeKeyBlockChain( dataBlock, dbSize, map, kmi->position );

	assert( quantity == kmi->quantity );
}

// static 
int dbHeader::freeKeyBlockChain( DATA_BLOCK *dataBlock, DWORD dbSize, dbKeyId *map, int cur )
{
	int quantity = 0;

	int maxKeyQ = DATA_BLOCK::getKeyQ( dbSize );
	do
	{
		map[cur].kbType = kbtEmpty;
		map[cur].kbSize = 0;
		map[cur].keyParam = 0;

		if ( cur < maxKeyQ )
			ShredData( dataBlock->keyTab[cur], DB_MinEncryptedKeyBlockSize );

		quantity++;

		if ( map[++cur].kbType != (WORD)kbtTail )
			break;
	}
	while( cur < DB_maxKeyQ );

	return quantity;
}

bool dbHeader::whiteServiceInfo( cPlainTextBlock *ptb )
{
	if ( ptb->WriteBytes( (BYTE *)this, ( sizeof( *this ) ), true ) == 0 )
		return false;
	
	dbHeader *hdr = ( dbHeader * )ptb->buffer;

	hdr->InitFirst();

	return true;
}

int dbHeader::getLastEmptyChainStartPos()
{
	int i = DB_maxKeyQ;

	for ( ; i ; i-- )
	{
		if ( keyMap[ i-1 ].kbType != kbtEmpty )
			break;
	}

	return i;
}

//
//	Updates internal or external key map
//	
//static
int dbHeader::updateKeyMap( dbKeyId *kmT, const dbKeyId *keyMap ) //bool internal ) // internal = true
{
	ODSFunc;

	if ( kmT == NULL )
		return 0;

	int i = DB_maxKeyQ;

	for ( ; i--; )
	{
		//if ( internal )
		//{
		//	if ( kmT[i].kbType != kbtEmpty )
		//		keyMap[ i ].kbType = kmT[i].kbType;
		//}
		//else
		{
			if ( keyMap[i].kbType != kbtEmpty )
				kmT[i].kbType = keyMap[i].kbType;
		}
	}

	return i;
}

//static
bool dbHeader::MarkBusied( dbKeyId *kmT, int pos, int q, ekbType type, WORD size )
{
	if ( size != 0 )
		q = dbKeyId::GetRequiredQ( size );

	ekbType tailType = ( type == kbtEmpty ) ? kbtEmpty : kbtTail;

	if ( pos >= DB_maxKeyQ )
		return false;

	if ( q == 0 )
		return false;

	if ( (DWORD)q > DB_maxKeyQ - pos )
	{
		assert( false );
		q = DB_maxKeyQ - pos;
	}

	int i = q;

	while( true )
	{
		if ( --i )
		{
			kmT[ pos + i ].kbType = tailType;
			kmT[ pos + i ].kbSize = 0;
			kmT[ pos + i ].keyParam = 0;
		}
		else
		{
			kmT[ pos + i ].kbType = type;
			kmT[ pos + i ].kbSize = size;
			kmT[ pos + i ].keyParam = 0;
			break;
		}
	}

	//
	//	Real data block size may be small nad does not occupies items up to DB_maxKeyQ
	//
	////
	////	Check next value, it has to be kbtEmpty or a busied, but not kbtTail
	////
	//if ( pos + q < DB_maxKeyQ )
	//{
	//	assert( kmT[ pos + q ].kbType != kbtTail );
	//	return false;
	//}

	return true;
}

//
//	findEmptySpace
//		returns position number if found,
//		otherwise returns -1
//
int dbHeader::findEmptySpace( dbKeyId *kmT, int sizeRequired, int first, int maxQ, bool bFromTail ) 
																	//int first = 0, int last = DB_maxKeyQ, bool bFromTail = false )
{
	int foundEmptyQ = 0;
	int foundEmptyN = -1;

	if ( bFromTail )
	{
		for( int p = maxQ - 1; p >= first; p-- )
		{
			if ( kmT[ p ].kbType == kbtEmpty )
			{
				foundEmptyQ++;
				if ( foundEmptyQ == sizeRequired )
					return p;
			}
		}
	}
	else
	{
		for( int p = first; p < maxQ; p++ )
		{
			if ( kmT[ p ].kbType == kbtEmpty )
			{
				if ( foundEmptyN == -1 )
				{
					foundEmptyN = p;
					foundEmptyQ = 0;
				}

				foundEmptyQ++;

				if ( foundEmptyQ == sizeRequired )
					return foundEmptyN;
			}
			else
				foundEmptyN = -1;
		}
	}

	return -1;
}

//
//	treatAsEmpty
//		checks if current block consists of chain from start block
//
static bool treatAsEmpty( dbKeyId *map, int startNum, int curNum )
{
	if ( startNum == -1 )
		return false;

	// Checks low bound
	if ( curNum < startNum )
		return false;

	//
	//	e - empty |   eeebtbttteebtbt
	//	b - busy  |        | |
	//	t - tail  |   start| |cur
	//

	// Gets finish number

	int nextNum = startNum;

	for( ; ++nextNum < DB_maxKeyQ;  )
	{
		if (  map[nextNum].kbType != (WORD)kbtTail )
			break;
	}

	// Returns high bound

	return ( curNum < nextNum ) ;
}

////
////	dbGetEmptyKeyBlock
////		Walks fields and return number of first empty field chain required size.
////		Returns -1 if required place is not found.
////		If meanAsFree is not -1 the key block is going to be overwriten next step,
////			see DataBlockWriteKey with kbNumber = -1
////
//static int dbGetEmptyKeyBlock( dbKeyId *map, int keyBlockSize, int meanAsFree = -1 )
//{
//	WORD foundSize = 0;
//
//	for ( int i = 0; i < DB_maxKeyQ; i++ )
//	{
//		if ( ( map[i].kbType == kbtEmpty ) || treatAsEmpty( map, meanAsFree, i ) )
//		{	
//			foundSize += DB_MinEncryptedKeyBlockSize;
//			if ( foundSize >= keyBlockSize )
//				return i + ( 1 - foundSize / DB_MinEncryptedKeyBlockSize );
//		}
//		else
//		{
//			foundSize = 0;
//		}
//	}
//
//	return -1;
//}
//
//
//int dbHeader::getMaxEmptyKeyBlockSpace( dbKeyId *keyMap, DWORD *pMaxFreeSize, int meanAsFree = -1 )
//{
//	int maxFree = 0;
//	int pos = dbHeader::getMaxEmptyKeys( keyMap, DB_maxKeyQ, pMaxFree, meanAsFree );
//
//	if ( pos == -1 )
//		return -1;
//
//	if ( pMaxFreeSize != 0 )
//		*pMaxFreeSize = maxFree * DB_MinEncryptedKeyBlockSize;
//
//	return pos;
//}

int dbHeader::getMaxEmptyKeys( dbKeyId *keyMap, int keyQ, int *pMaxFree, int meanAsFree )
{
//	assert( false ); // check this 
	int foundSize = 0;
	int founPos = -1;
	int maxFoundSize = 0;
	int maxFoundPos = -1;
	
	for ( int i = 0; i < keyQ; i++ )
	{
		if ( ( keyMap[i].kbType == kbtEmpty )
			|| treatAsEmpty( keyMap, meanAsFree, i ) )
		{
			if ( foundSize == 0 )
				founPos = i;

			foundSize += 1;
		}
		else
		{
			if ( maxFoundSize < foundSize )
			{
				maxFoundSize = foundSize;
				maxFoundPos = founPos;
			}

			foundSize = 0;
		}
	}

	if ( maxFoundSize < foundSize )
	{
		maxFoundSize = foundSize;
		maxFoundPos = founPos;
	}

	if ( pMaxFree != NULL )
		*pMaxFree = maxFoundSize;

	return maxFoundPos;
}

DWORD dbHeader::getRealSize() // DATA_BLOCK *db )
{
	if ( isEncrypted(*this) ) return DWORD(-1);

	int lastEmpty =	getLastEmptyChainStartPos(); // db->header.m_body.getLastEmptyChainStartPos();

	DATA_BLOCK db;
	DWORD sizeRequired = DWORD((BYTE*)&(db.keyTab[ lastEmpty ][ 0 ]) - (BYTE*)(&db));

	return sizeRequired;
}
