//
//	BestCrypt driver access...
//

#include "diskDriver.h"
#include <windef.h> 
#include "..\..\..\Includes\ODS.h"
#ifdef _ODS
#include "..\..\..\Includes\UnicodeFuncs.h"
#endif

#define FILE_DEVICE_DISK   0x00000007
#define IOCTL_DISK_BASE    FILE_DEVICE_DISK
#define METHOD_BUFFERED    0
#define METHOD_IN_DIRECT   1
#define FILE_ANY_ACCESS    0
#define STATUS_SUCCESS     0

#define maxLen( buff ) ( sizeof( buff ) / sizeof( *buff ) )

#define CTL_CODE( DeviceType, Function, Method, Access ) (                 \
    ((DeviceType) << 16) | ((Access) << 14) | ((Function) << 2) | (Method) \
)


//typedef struct _BUFFER_HEADER { char   Signature[8];	// input: "LOCOS97",00 ; output: "LS06CXX",00
//                                UCHAR   Status;
//                              } BUFFER_HEADER, *PBUFFER_HEADER;
//
/*******************************************************************************
***************** IOCTLs for BestCrypt Disk Driver Service *********************
********************************************************************************/


// Get Logical Disk Mask command & structure definition

#define IOCTL_PRIVATE_GET_BCDISK_MASK \
              CTL_CODE(IOCTL_DISK_BASE, 0x801, METHOD_IN_DIRECT, FILE_ANY_ACCESS)


//typedef struct  _GETMASK_BUFFER {
//	char   Signature[8];	// input: "LOCOS97",00 ; output: "LS06CXX",00
//	UCHAR  Status;
//	DWORD  Mask;
//
//	//
//	//	Mask of the drives available for transparent ecryption.
//	//	To set transparent ecryption we will use same command as for usual mount
//	//	 but for bcrmpart.vxd (or .sys) "BestCrypt Removable Partitions"
//	//
//	DWORD  hostsMask;
//} GETMASK_BUFFER, *PGETMASK_BUFFER;


//static const char *getServiseName()
//{
//	OSVERSIONINFO vi;
//	vi.dwOSVersionInfoSize = sizeof( vi );
//
//	GetVersionEx( &vi );
//
//	BOOL isNT = (vi.dwPlatformId  == VER_PLATFORM_WIN32_NT);
//
//	if ( isNT )
//		return ("\\\\.\\BCDISK");
//	else
//		return ("\\\\.\\bcrypt.vxd");
//}

#define DISK_SERVICE_NAME_95    L"\\\\.\\bcrypt.vxd"
#define DISK_SERVICE_NAME_NT4   L"\\\\.\\BCDISK"
#define DISK_SERVICE_NAME_W2000 L"\\\\.\\BCBUS"

static const wchar_t *getDiskServiseName()
{
	OSVERSIONINFO vi;
	vi.dwOSVersionInfoSize = sizeof( vi );
	bool isNT = false, isW2000 = false;

	GetVersionEx( &vi );

	isNT = (vi.dwPlatformId  == VER_PLATFORM_WIN32_NT);

	if ( vi.dwMajorVersion >= 5 ) // && vi.dwMinorVersion >= 0 )
		isW2000 = true;

	if (isW2000)
		return DISK_SERVICE_NAME_W2000;
	else if ( isNT )
		return DISK_SERVICE_NAME_NT4;
	else
		return DISK_SERVICE_NAME_95;
}

static BOOL SendCommandToDiskDriver( LPVOID Buffer, DWORD BufferSize, DWORD CtlCode )
{
	HANDLE         hDevice;
	DWORD          lBytesReturned = 0;
	BUFFER_HEADER	*bufHead = (BUFFER_HEADER *)Buffer;

	ODSFuncS( _FormatT( L"0x%x" , CtlCode ) );

	hDevice = CreateFileW( getDiskServiseName(), GENERIC_READ | GENERIC_WRITE,
						  0, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL );

	if ( hDevice == INVALID_HANDLE_VALUE )
	{
		DWORD dwErr = GetLastError();
		ODS_fls( _FormatT( L"Error '%s' in openig file '%s'", _GetSystemError( dwErr ), getDiskServiseName() ) );
		return FALSE;
	}

	memcpy( bufHead->Signature, "LOCOS97 ", maxLen( bufHead->Signature ) );

	if (!DeviceIoControl( hDevice, CtlCode,
						  Buffer, BufferSize,
						  Buffer, BufferSize,
						  &lBytesReturned, NULL))
	{
		DWORD dwErr = GetLastError();
		ODS_fls( _FormatT( L"Error '%s' in DeviceIoControl", _GetSystemError( dwErr ) ) );
		CloseHandle(hDevice);
		return FALSE;
	}

	CloseHandle(hDevice);

	if ( strncmp(bufHead->Signature,"LS06CXX ", 8 ) != 0 )
	{
		ODS_fls( L"Assertion! Signature is invalid!" );
		return FALSE;
	}

	if ( bufHead->Status != STATUS_SUCCESS )
	{
		ODS_fls( _FormatT( L"Assertion! Status %d is invalid!", bufHead->Status ) );
		return FALSE;
	}

	return TRUE;
}

/*****************************************************************/

BOOL CmdGetVDiskMask( DWORD &Mask )
{
	GETMASK_BUFFER mBuf;

	Mask = 0;

	if (!SendCommandToDiskDriver(
			(LPVOID)(&mBuf), 
			sizeof(GETMASK_BUFFER),
			IOCTL_PRIVATE_GET_BCDISK_MASK) )
				return false;

	if ( mBuf.Status != STATUS_SUCCESS )
		return false;

	Mask = mBuf.Mask;
	ODS_fls( _FormatT( L"VDiskMask = 0x%X", Mask ) );

	return true;
}

static bool prepareContNameForDriver( wchar_t *contNameForDriver,  const wchar_t *contNameOrig, int len )
{
	const wchar_t *wcPrefix = L"\\??\\";

	if ( (size_t)len < wcslen( contNameOrig ) + wcslen( wcPrefix ) )
		return false;

	wcsncat( wcsncpy( contNameForDriver, wcPrefix, len ), contNameOrig, len );

	return true;
}

BOOL GetVDiskInfo( DISKINFO_BUFFER_V8_0 *dBuf, wchar_t *name, bool byMountPoint  )
{
	// Version 2.01.7.4 - 24.10.2018 - Was 'if ( name != NULL )' 
	if ( name == NULL )
		return FALSE;

	memset( dBuf, 0, sizeof( *dBuf ));
	dBuf->structSize = sizeof( *dBuf );
	dBuf->command = CMD_GetInformation;
	
	if ( byMountPoint )
	{
		wcsncpy( dBuf->mountPoint, name, maxLen( dBuf->mountPoint ) );
		dBuf->validFieldsMask = MASK_MountPoint;
	}
	else
	{
		if ( !prepareContNameForDriver( dBuf->fileName,  name, maxLen( dBuf->fileName ) ) )
		{
			return FALSE;
		}

		dBuf->validFieldsMask = MASK_ContainerName;
	}

	return SendCommandToDiskDriver( (LPVOID)(dBuf), dBuf->structSize, IOCTL_PRIVATE_V8_COMMAND );
}

//
//	_GetLogicalDrives()
//	We have to ask BestCrypt about used mask before.
//	Reasons:
//	In Windows NT environment, if any BC's drive was not full dismount 
//	( only symbolik link removed ) the drive reported by system as free and
//	the container can be mounted again only on previous drive letter 
//	( see FLAG_MOUNT_ON_SINGLE_LETTER ).
//	But! If we mount other container the letter is to be band for mounting.
//	We have to ask BestCrypt driver for the really used virtual drive mask.
//
DWORD _GetLogicalDrives()
{
	OSVERSIONINFO ver;
	DWORD mask;
	BOOL  isWin2000 = FALSE;

	mask = GetLogicalDrives();

	ver.dwOSVersionInfoSize = sizeof(ver);

	GetVersionEx( &ver );

	if ( ( ver.dwPlatformId == VER_PLATFORM_WIN32_NT ) &&
		 ( ver.dwMajorVersion == 5 ) && 
		 ( ver.dwMinorVersion >= 0 )
	   )
		isWin2000 = TRUE;
	else 
		isWin2000 = FALSE;
	
	if (!isWin2000)
	{
		DWORD bcMask = 0;

		if ( CmdGetVDiskMask( bcMask ) )
			mask |= bcMask;
	}

	return mask;
}


