
/*******************************************************************\
*
*	Copyright (c) 2007-2015 Jetico, Inc., Finland
*	All rights reserved.
*
*	File: kgParams.h
*	Revision: 
*	Created:
*
\*******************************************************************/

#pragma once

#include "dbHeaderBase.h"
#include "..\..\..\Includes\cSymAlgInfo.h"
#include "..\..\..\Includes\ODS.h"
#include "..\..\..\Includes\UnicodeFuncs.h"

#pragma pack(1)
struct sParams { DWORD size; };

struct sCKHParams : public sParams
{
	sCKHParams()
	{
		size = sizeof ( *this );
		memset( &contInfo, 0, sizeof( contInfo ) );
		memset( &decryptedHeader, 0, sizeof( decryptedHeader ) );
		memset( &keyMapInfo, 0, sizeof( keyMapInfo ) );
		contAlgId = 0;
		encModeId = 0;
		hashAlgId = 0;
		bEncrypted = false;
	}

	sSpaceUnitBase	contInfo;	// [IN ] information about the new container is being created
								// [OUT] information about current ( usual or hidden ) part
								// If hidden part key block was activated.

	DWORD contAlgId;			// [OUT] returns back real algorithm id for encrypted block.
	DWORD encModeId;			// Define encryption mode CBC or LRW
	DWORD hashAlgId;			// sets initial for new key block and returns back real hash algorithm id for verification and password generation.
//	DWORD iteration;			// sets iteration count and return it back, included from v.1.06.6

	bool bEncrypted;
	dbHeaderBase decryptedHeader;

	//
	//	Key map information where key was found or created.
	//	It is useful for several hidden part creation.
	//
	sKeyMapInfo keyMapInfo;

	virtual ~sCKHParams() { memset( this, 0, sizeof ( *this )); }

	static int getIterationCount( sCKHParams *ps );

	virtual wchar_t *getMountPoint() { ODS_fls( L"Assertion!" ); return NULL; }
	virtual bool setMountPoint( const wchar_t *mp ) { assert(false); return false; }
};

struct sCKHParamsAndMountPoint : public sCKHParams
{
	sCKHParamsAndMountPoint()
	{
		memset( mountPoint, 0, sizeof(mountPoint) );
		size = sizeof ( *this );
	}

	wchar_t mountPoint[ MAX_PATH ]; // [IN]  mount point saved befor
									// [OUT] new mount point to be saved

	~sCKHParamsAndMountPoint() { memset( this, 0, sizeof ( *this )); }

	virtual wchar_t *getMountPoint()
	{
		ODS_fls( _FormatT( L"mountPoint = %x - %s", mountPoint, mountPoint ) );
		return mountPoint;
	}

	virtual bool setMountPoint( const wchar_t *mp )
	{
		copyString( mountPoint, mp, maxLen( mountPoint ) );
		return false;
	}
};



//
//	Best solution to pass algorithm info is separate function.
//
//
//	Includes algorithms info
//
struct sCKHParams_MP_AI : public sCKHParamsAndMountPoint
{
	sCKHParams_MP_AI()
	{
		memset( this, 0, sizeof ( *this ));
		size = sizeof ( *this );
		symAlgT = NULL;
		symAlgQ = 0;
	}

	bool bUseMountPoint;

	cSymAlgInfo *symAlgT;
	int symAlgQ;

	~sCKHParams_MP_AI() { memset( this, 0, sizeof ( *this )); }

	virtual wchar_t *getMountPoint()
	{
		ODS_fls( _FormatT( L"bUseMountPoint = %s, mountPoint %x, %s", getWCStr(bUseMountPoint), mountPoint, mountPoint ) );

		if ( !bUseMountPoint )
			return NULL;

		return mountPoint;
	}
};

struct sCKHParamsAndMountPointEx :public sCKHParams_MP_AI
//struct sCKHParamsAndMountPointEx :public sCKHParamsAndMountPoint
{
	sCKHParamsAndMountPointEx()
	{
		size = sizeof ( *this );
		contPath = NULL;
		flags = 0;
	}

	const wchar_t *contPath;
	enum { eFlagMountForCurrentUser = 1, eFlagSimpleDialog = 2, eFlagAdvancedMode = 4, eFlagMountAsRemovable = 0x8 };
	DWORD flags; 

	bool getFlagMountForCurrentUser()
	{
		return ( flags & eFlagMountForCurrentUser ) != 0;
	}

	void setFlag( DWORD mask, bool bSet )
	{
		if ( bSet )
			flags |= mask;
		else
			flags &= ~mask;
	}
};

#pragma pack()
