/******************************************************************\
*
*	Copyright (c) 2005-14 Jetico, Inc., Finland
*	All rights reserved.
*
*	File: GeberateSeed.cpp
*	Revision: 
*	Created:
*	Description: generation of seed value for random key generation
*
\******************************************************************/

#include <windows.h>
#include <windowsx.h>
#include <commctrl.h>
#include <string.h>
#include <stdio.h>
#include "GenerateSeed.h"
//#include "resource.h" 
#include "kg_errors.h"
#include <time.h>
#include <assert.h>
#include "..\..\CertificateLib\x509pkts\pkcs12_funcs.h"
#include "..\..\Includes\sysRand.h"
//
//	For GetPasswordFromCommandLine only
//
#include "pwdlg.h"
#include "..\..\CertificateLib\SMemory\cPassword.h"


#include "UnicodeFuncs.h"
//#include "Translatedll.h"
#include "cDlg.h"

#define MSG_SEED_READY		(WM_USER + 0x529)
#define MSG_STOP_LOOP		(WM_USER + 0x530)

typedef BOOL CALLBACK DlgProcType( HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam );

BOOL CALLBACK getRandomDlgProc( HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam );
BOOL UpdateSeed( BYTE *Seed, DWORD dwData );

/* Global data */

#define BUFFER_SIZE 1024

BYTE    *Seed;
DWORD   SeedCounter, SeedLength, SeedMinLength;

WNDPROC wpOrigWndProc; // Progerss original procedure

static int st_idDlg = 0;
static int st_idProgress = 0;
static int st_idsStop = 0;
static int st_idIcon = 0;

extern void setDialogIds( int idDlg, int idProgress, int idsStop, int idIcon )
{
	st_idDlg = idDlg;
	st_idProgress = idProgress;
	st_idsStop = idsStop;
	st_idIcon = idIcon;
}

DWORD
	GetSeedValue( 
		HWND  hWndParent,
		BYTE  *pSeed, 
		DWORD pSeedLength,
		DWORD pSeedMinLen
	)
{
	if ( st_idDlg == 0 )
	{
		// The setDialogIds( IDD_RANDOMDLG, IDC_PROGRESS, IDS_szStop ) was not called.
		assert( false );
		ODS_fls( L"Assertion!" );
		return ERROR_INTERNAL_PROBLEM;
	}

	if ((!pSeed) || (pSeedLength <= 0)) 
	{
		assert( false );
		ODS_fls( L"Assertion!" );
		return ERROR_INTERNAL_PROBLEM;
	}

	Seed = pSeed;
	SeedCounter = 0;
	SeedLength = pSeedLength;
	SeedMinLength = min( pSeedMinLen, pSeedLength );

	//
	//	07.09.2014
	//		Strange problem. 
	//		Progress control does not catch WM_KEY messages, even it's sublass procedure returns DLGC_WANTALLKEYS for WM_GETDLGCODE
	//	So modeless dialog is used.
	//

	cDlg dlg;
	HWND hWnd = dlg.DoModeless( IH.m_hInst, st_idDlg, hWndParent, (DLGPROC)getRandomDlgProc );
	ShowWindow(hWnd, SW_SHOW);

	MSG msg;
	while ( GetMessage(&msg, NULL, 0, 0) ) 
	{
		if ( (msg.message == WM_KEYDOWN) ||
			 (msg.message == WM_SYSKEYDOWN )
			 )
		{
			SendMessage(hWnd, msg.message, msg.wParam, msg.lParam);
		}

		if ( msg.message == MSG_STOP_LOOP )
			break;

		TranslateMessage(&msg);
		DispatchMessage(&msg);
	}

	DWORD dwErr = ERROR_NO;

	//
	//	v.2.01.4.7 - 12.04.2016
	//
	// Stop button was used.
	if ( SeedCounter < SeedLength )
	{
		DWORDLONG dwl = sRandCtx::getTimeSeed();
		if ( 0 != P12_producePseudoRandomBytes( pSeed + SeedCounter, SeedLength - SeedCounter, (BYTE*)&dwl, sizeof(dwl), 1, pSeed, SeedCounter, P12_prgm4IV, pgphaSHA256 ) ) // -1 = pgphaSHA1
			dwErr = ERROR_INTERNAL_PROBLEM;
	}

	//result = (int)dlg.DoModal( IH.m_hInst, IDD_RANDOMDLG, hWndParent, (DLGPROC)getRandomDlgProc );

	//if (result == -1) /* system error */
	//{
	//	assert( false );
	//	return ERROR_INTERNAL_PROBLEM;
	//}

	//if ( result == 0 )
	//{
	//	assert( false );
	//}

	return dwErr;
}

static void enableOkButton( bool bEnable, HWND hOk = NULL )
{
	static HWND hSave = NULL;
	if ( hOk != NULL )
		hSave = hOk;

	static int iEnabled = -1;
	if ( iEnabled == (int)bEnable )
		return;

	EnableWindow( hSave, bEnable );
	iEnabled = (int)bEnable;
}


// Subclass procedure
LRESULT APIENTRY progressSubclassProc( HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam )
{
	HWND parentHwnd;

	switch (uMsg)
	{
		case WM_GETDLGCODE:
			return DLGC_WANTALLKEYS;

		case WM_CHAR:
		case WM_KEYUP:
		case WM_KEYDOWN:
		case WM_SYSKEYDOWN:
			if (!UpdateSeed( Seed, (DWORD)wParam ))
			{	parentHwnd = GetParent( hwnd );
				SendMessage( parentHwnd, MSG_SEED_READY, 0, 0);
			}
			else
			{	SendMessage( hwnd, PBM_SETPOS, (WPARAM)(SeedCounter), 0);
			}

			break;
		case WM_MOUSEMOVE:
			if (!UpdateSeed( Seed, (DWORD)lParam ))
			{	SendMessage( GetParent( hwnd ), MSG_SEED_READY, 0, 0);
			}
			else
			{	SendMessage( hwnd, PBM_SETPOS, (WPARAM)(SeedCounter), 0);
			}

			break;
		default:
			break;
	}

	if ( SeedCounter >= SeedMinLength )
		enableOkButton( true );

	return CallWindowProc(wpOrigWndProc, hwnd, uMsg, wParam, lParam);
}

//
// v. 1.06.5 - 30.07.2014
//	Dialog closes automatically at buffer is filled enough
//
#define _stopAutomatically

BOOL CALLBACK getRandomDlgProc( HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam )
{
	HWND   prHwnd;
	
	switch (uMsg) 
	{ 
		case WM_CHAR:
		case WM_KEYUP:
		case WM_KEYDOWN:
		case WM_SYSKEYDOWN:
			if (!UpdateSeed( Seed, (DWORD)wParam ))
			{	
				SendMessage( hwnd, MSG_SEED_READY, 0, 0);
			}
			else
			{	SendMessage( GetDlgItem( hwnd, st_idProgress )  , PBM_SETPOS, (WPARAM)(SeedCounter), 0);
			}

			break;

		case WM_MOUSEMOVE:
			if (!UpdateSeed( Seed, (DWORD)lParam ))
			{
				SendMessage( hwnd, MSG_SEED_READY, 0, 0);
			}
			else
			{	SendMessage( GetDlgItem( hwnd, st_idProgress  ), PBM_SETPOS, (WPARAM)(SeedCounter), 0);
			}
			break;

        case WM_INITDIALOG:

			//if ( pTranslateDll != NULL )
			//	pTranslateDll->TranslateDialog( hwnd );

			// Retrieve the handle of the progress indicator control
			prHwnd = GetDlgItem(hwnd, st_idProgress); 
			SendMessage( prHwnd, PBM_SETRANGE, 0, MAKELPARAM(0, SeedLength));

			if ( st_idIcon != 0 )
				cDlg::SetDlgIcon( hwnd, LoadIcon( IH.m_hInst, MAKEINTRESOURCE( st_idIcon ) ), ICON_SMALL );
 
#ifdef _stopAutomatically
			SetDlgItemTextT( hwnd, IDOK, _LoadIDS( st_idsStop ) );
			enableOkButton( false, GetDlgItem( hwnd, IDOK ) );
#else
			EnableWindow( GetDlgItem(hwnd, IDOK), FALSE); 
#endif

			// Subclass the  progress indicator control
			wpOrigWndProc = (WNDPROC) SetWindowLongPtrT(prHwnd, GWLP_WNDPROC, (LONG_PTR) progressSubclassProc);
			//SetWindowLongPtrT(GetDlgItem(hwnd, IDC_Text), GWLP_WNDPROC, (LONG_PTR) progressSubclassProc);
			return TRUE;

		case WM_DESTROY:
			// Retrieve the handle of the  progress indicator control
			prHwnd = GetDlgItem(hwnd, st_idProgress);

			// Remove the subclass from the progress indicator control.
			SetWindowLongPtrT(prHwnd, GWLP_WNDPROC, (LONG_PTR) wpOrigWndProc); 

			break;
 
		case MSG_SEED_READY:

#ifdef _stopAutomatically
			EndDialog( hwnd, IDOK ); 
			PostMessage( NULL, MSG_STOP_LOOP, 0, 0 );
			return TRUE; 
#else
			EnableWindow( GetDlgItem(hwnd, IDOK), TRUE); 

			SetFocus( GetDlgItem(hwnd, IDOK) ); 
			break;
#endif

		case WM_COMMAND:
			switch (wParam)
			{
				case IDOK: 
					EndDialog(hwnd, (int)ERROR_CANCELLED_BY_USER); 
					PostMessage( NULL, MSG_STOP_LOOP, 0, 0 );
					return TRUE; 
			}
			break;

		default:
			return 0;
	}
	return 0;
}

BOOL UpdateSeed( BYTE *Seed, DWORD dwData )
{
#define _CheckPrevData
#ifdef _CheckPrevData

	// Skips series of same values

	static int failCount = 0;
	static DWORD prevData = -1;

	if ( SeedCounter == 0 )
		failCount = 0;

	dwData ^= (DWORD)clock();

	if ( dwData == prevData )
	{
		failCount ++;
		return TRUE;
	}

	prevData = dwData;

#endif _CheckPrevData

	if (SeedCounter >= SeedLength)
	{

//#ifdef _DEBUG
//#ifdef _CheckPrevData
//
//		static bool shown = false;
//		if ( ( failCount != 0 ) && !shown )
//		{
//			shown = true;
//			MessageBoxT( 0, FormatOnceT( L"Fail count %d data %X", failCount, dwData ), L"UpdateSeed DBG", MB_OK );
//			failCount = 0;
//		}
//#endif // _CheckPrevData
//#endif // _DEBUG

		return FALSE;
	}

	BYTE value = 0;
	for( int i = sizeof( dwData ); i--; )
		value ^= BYTE( dwData >> ( i * 8 ) );

	Seed[ SeedCounter ++ ] = value;

	return TRUE;
}