/*******************************************************************
 *	Copyright (c) 1994-2009 Jetico, Inc., Finland
 *	All rights reserved.
 *
 *	File: kbd.cpp
 *	Revision: 
 *	Created:
 *	Description: interface with the BestCrypt keyboard driver
 *
 *******************************************************************/

#include <windows.h>
#include <winuser.h>
#include <memory.h>
#include <ctype.h>
#include <assert.h>
#include "kbd.h"
//#include "bc_ioctl.h"
//#include "kg_misc.h"
#include "mhk.h"
#include "..\..\Includes\cdll.h" 
#include "..\..\Includes\ODS.h" 

static void ShredData( BYTE *data, int dataLength )
{
	if ( ( data == NULL ) || ( dataLength < 0 ) )
	{
		assert( false );
		return;
	}

	static DWORD dwOld = 0; // It happed that ShredData was called within one time-clock interval 

	DWORD dwInit;
	do {
		ULARGE_INTEGER uli;
		uli.QuadPart = (DWORDLONG)time( NULL ) * (DWORDLONG)clock();
		dwInit = uli.LowPart ^ uli.HighPart;
	}
	while( dwInit == dwOld );
	
	dwOld = dwInit;

	srand( dwInit );

	for( int i = 0; i < dataLength; i++ )
		data[i] = (BYTE)rand();
}

//const char *kbdParams::szSemaphoreName = "Local\\BCKeyGenSemaphore";
const char *kbdParams::szSemaphoreName = "BCKeyGenSemaphore";

HHOOK HookNext = NULL;

LRESULT CALLBACK HookKeyboardProc(
  int    nCode,   // hook code
  WPARAM wParam,  // virtual-key code
  LPARAM lParam   // keystroke-message information
);

BYTE *kbdTmp;

UINT CharacterOrder;

void freeTables(UINT   **VkOriginal,
				UINT   **VkMessed,
				USHORT **ScodOriginal,
				USHORT **ScodMessed,
				BYTE   **ShiftTable,
				UINT   ShiftTableSize )
{
	if (*VkOriginal)	
	{	ShredData( (BYTE *)(*VkOriginal), ENCODING_TABLE_SIZE*2*sizeof(UINT) );
		delete *VkOriginal;
		*VkOriginal = NULL;
	}
	if (*VkMessed)
	{	ShredData( (BYTE *)(*VkMessed), ENCODING_TABLE_SIZE*2*sizeof(UINT) );
		delete *VkMessed;
		*VkMessed = NULL;
	}
	if (*ScodOriginal)
	{	ShredData( (BYTE *)(*ScodOriginal), ENCODING_TABLE_SIZE*2*sizeof(USHORT) );
		delete *ScodOriginal;
		*ScodOriginal = NULL;
	}
	if (*ScodMessed)
	{	ShredData( (BYTE *)(*ScodMessed), ENCODING_TABLE_SIZE*2*sizeof(USHORT) );
		delete *ScodMessed;
		*ScodMessed = NULL;
	}
	if (*ShiftTable)
	{	ShredData( (BYTE *)(*ShiftTable), ShiftTableSize*sizeof(BYTE) );
		delete *ShiftTable;
		*ShiftTable = NULL;
	}
}


static BOOL virtualKeyIsValid( ULONG Vk )
{
	BYTE  keyState[256];
	HKL   hkl;
  	UINT  scanCode;
	WORD  chars;
	int   charsGot;

	memset( keyState,  0x00, 256 ); // nothing is pressed or toggled

	hkl = GetKeyboardLayout( 0 );
	
	/* call ToAsciiEx just to clear keyboard from dead key, if it was pressed */
	scanCode = MapVirtualKeyEx( Vk, 0, hkl ); 
	charsGot = ToAsciiEx( Vk, scanCode, keyState, &chars, 0, hkl );

	/* Get scan code for the virtual key */
	scanCode = MapVirtualKeyEx( Vk, 0, hkl ); 

	/* Get character for the given virtual key */
	charsGot = ToAsciiEx(  Vk, scanCode, keyState, &chars, 0, hkl );
		
	if (charsGot != 1)
	{
		return FALSE;
	}

	return TRUE;
}


#define ADD_VIRTUAL_KEY( Vk, i ) \
	if (virtualKeyIsValid( Vk )) \
	{ tableNumL[i] = tableSmal[i] = Vk;	i++; } \
	else \
	{ (*SizeNumL)--; (*SizeSmal)--; }


void addToTableAltGr(  UINT VirtualKey, UINT *TableAltGr, UINT *IndexAltGr )
{
	BYTE  keyState[256];
	HKL   hkl;
  	UINT  scanCode;
	WORD  chars;
	int   charsGot;

	memset( keyState,  0x00, 256 ); // nothing is pressed or toggled
	keyState[ VK_MENU ]  = 0x80;    // ALT is pressed
	keyState[ VK_RMENU ] = 0x80;    // RIGHT ALT (AltGr) is pressed

	keyState[ VK_CONTROL ] = 0x81;
	keyState[ VK_LSHIFT ]  = 1;
	keyState[ VK_RSHIFT ]  = 1;
	keyState[ VK_LCONTROL ]  = 0x80;
	keyState[ VK_RCONTROL ]  = 1;

	hkl = GetKeyboardLayout( 0 );
	
	/* Get scan code for the virtual key */
	scanCode = MapVirtualKeyEx( VirtualKey, 0, hkl ); 

	/* Get character for the given virtual key */
	charsGot = ToAsciiEx(  VirtualKey, scanCode, keyState, &chars, 0, hkl );
		
	if (charsGot != 1)
	{
		return;
	}

	TableAltGr[ *IndexAltGr ] = VirtualKey;
	(*IndexAltGr)++;
}



//#define ENCODING_TABLE_SIZE 64 - max numbers
#define NUMBER_OF_LETTERS       26
#define NUMBER_OF_USUAL_DIGITS  10
#define NUMBER_OF_SPACE_LETTERS 1
//#define NUMBER_OF_OTHER_VK      11
#define NUMBER_OF_OTHER_VK      8
#define NUMBER_OF_NUMPAD_DIGITS 10


BOOL createOriginalTable_VK( UINT **Table, UINT *SizeNumL, UINT *SizeSmal, UINT *SizeAltGr )
{
	UINT i,j, indexAltGr;
	UINT start, *tableNumL, *tableSmal, *tableAltGr;

	// set the SizeNumL and SizeSmal to maximum numbers
	*SizeNumL = NUMBER_OF_USUAL_DIGITS + NUMBER_OF_LETTERS + NUMBER_OF_SPACE_LETTERS + NUMBER_OF_OTHER_VK + NUMBER_OF_NUMPAD_DIGITS;
	*SizeSmal = NUMBER_OF_USUAL_DIGITS + NUMBER_OF_LETTERS + NUMBER_OF_SPACE_LETTERS + NUMBER_OF_OTHER_VK;

	if ( ((*SizeNumL) > ENCODING_TABLE_SIZE) || ((*SizeSmal) > ENCODING_TABLE_SIZE) )
		return FALSE;
	
	*Table = new UINT[ ENCODING_TABLE_SIZE * 3 ];
	if (!(*Table)) return FALSE;

	memset( *Table, 0, ENCODING_TABLE_SIZE * 3 * sizeof(UINT) );

	tableNumL  = *Table;
	tableSmal  = &((*Table)[ENCODING_TABLE_SIZE]);
	tableAltGr = &((*Table)[ENCODING_TABLE_SIZE * 2]);

	i = 0;
	indexAltGr = 0;

	// Fill in two halves of the table - one of them for the case when NumLock is pressed

	start = 0x30; /* VK_0 ... VK_9 */
	for( j=0; j<10; i++, j++ )
	{
		tableNumL[i] = tableSmal[i] = start + j;
		addToTableAltGr(  start + j, tableAltGr, &indexAltGr );
	}


	start = 0x41; /* VK_A ... VK_Z */
	for( j=0; j<26; i++, j++ )
	{
		tableNumL[i] = tableSmal[i] = start + j;
		addToTableAltGr(  start + j, tableAltGr, &indexAltGr );
	}

	/* Plus additional VK_SPACE virtual key */

	ADD_VIRTUAL_KEY( VK_SPACE, i )
	addToTableAltGr( VK_SPACE, tableAltGr, &indexAltGr );

	/* Plus OTHER_VK */

	// main keyboard
	ADD_VIRTUAL_KEY( 0xBD, i )
	addToTableAltGr( 0xBD, tableAltGr, &indexAltGr );
	ADD_VIRTUAL_KEY( 0xBB, i )
	addToTableAltGr( 0xBB, tableAltGr, &indexAltGr );
	ADD_VIRTUAL_KEY( 0xBF, i )
	addToTableAltGr( 0xBF, tableAltGr, &indexAltGr );
	ADD_VIRTUAL_KEY( 0xBE, i )
	addToTableAltGr( 0xBE, tableAltGr, &indexAltGr );
////	ADD_VIRTUAL_KEY( 0xDC, i )                         - does not work in German keyboard
////	addToTableAltGr( 0xDC, tableAltGr, &indexAltGr );  - does not work in German keyboard
	ADD_VIRTUAL_KEY( 0xBC, i )
	addToTableAltGr( 0xBC, tableAltGr, &indexAltGr );
	ADD_VIRTUAL_KEY( 0xC0, i )
	addToTableAltGr( 0xC0, tableAltGr, &indexAltGr );
/////	ADD_VIRTUAL_KEY( 0xDB, i ) - for some reason it also does not work Norwegian keyboard correctly
//////	ADD_VIRTUAL_KEY( 0xDD, i ) - excluded, because on Norwegian keyboard it is a "transforming" key, i.e. after prssing it Windows waits for next key and transforms it...
	ADD_VIRTUAL_KEY( 0xBA, i )
	addToTableAltGr( 0xBA, tableAltGr, &indexAltGr );
	ADD_VIRTUAL_KEY( 0xDE, i )
	addToTableAltGr( 0xDE, tableAltGr, &indexAltGr );

	/* Plus additional VK_NUMPAD* virtual keys */

	tableNumL[i++] = VK_NUMPAD0;
	tableNumL[i++] = VK_NUMPAD1;
	tableNumL[i++] = VK_NUMPAD2;
	tableNumL[i++] = VK_NUMPAD3;
	tableNumL[i++] = VK_NUMPAD4;
	tableNumL[i++] = VK_NUMPAD5;
	tableNumL[i++] = VK_NUMPAD6;
	tableNumL[i++] = VK_NUMPAD7;
	tableNumL[i++] = VK_NUMPAD8;
	tableNumL[i++] = VK_NUMPAD9;

	*SizeAltGr = indexAltGr;

	return TRUE;
}

int getRandomToMess( BYTE *RandomArray, int Bound )
{	
	int result; /* result always < 256 because maximum number of virtual keys is 256 */

	result = (int)RandomArray[ Bound ];

	if (Bound > 0) result = result % Bound;
	else           result = 0;

	return result;
}


BOOL messTable( UINT **Table, UINT Size )
{
	int  i, index;
	UINT tmp;
	BYTE randomArray[ ENCODING_TABLE_SIZE ];

	/* Initialize random numbers array. ShredData uses SHA-1 PRNG */
    ShredData( randomArray, ENCODING_TABLE_SIZE );

	if (!Size)
	{
		return TRUE;
	}
	
	/*	Use the Moses-Oakford shuffling algorithm 
		(Tables of Random Permutations, Stanford University Press, 1963)
	 */
	for( i=Size-1; i>0; i-- )
	{
		index = getRandomToMess(randomArray, i-1);
		tmp = (*Table)[i];
		(*Table)[i] = (*Table)[index];
		(*Table)[index] = tmp;
	}
	return TRUE;
}



BOOL createMessedTable_VK(UINT *TableOriginal, 
						  UINT **TableMessed, 
						  UINT SizeNumL,
						  UINT SizeSmal,
						  UINT SizeAltGr )
{
	UINT *tableMessSmal, *tableMessAltGr;
	
	*TableMessed = new UINT[ ENCODING_TABLE_SIZE * 3 ];
	if (!(*TableMessed)) return FALSE;

	memcpy( *TableMessed, TableOriginal, ENCODING_TABLE_SIZE * 3 * sizeof(UINT));

	if (!messTable( TableMessed, SizeNumL )) return FALSE;

	tableMessSmal = &((*TableMessed)[ENCODING_TABLE_SIZE]);

	if (!messTable( &tableMessSmal, SizeSmal )) return FALSE;

	tableMessAltGr = &((*TableMessed)[ENCODING_TABLE_SIZE * 2]);

	if (!messTable( &tableMessAltGr, SizeAltGr )) return FALSE;

	return TRUE;
}


BOOL createTable_scode( UINT *VkTable, USHORT **ScodTable, UINT SizeNumL, UINT SizeSmal, UINT SizeAltGr )
{
	UINT i;
	USHORT *table;
	HKL    hkl= GetKeyboardLayout( 0 );

	table = *ScodTable = new USHORT[ ENCODING_TABLE_SIZE * 3 ];
	if (!table) return FALSE;

	memset( table, 0, ENCODING_TABLE_SIZE * 3 * sizeof(USHORT));

	for( i=0; i<SizeNumL; i++)
	{
		/* Get scan code for the virtual key */
		table[i] = MapVirtualKeyEx( VkTable[i], 0, hkl ); 
	}

	table = &((*ScodTable)[ENCODING_TABLE_SIZE]);

	for( i=0; i<SizeSmal; i++)
	{
		/* Get scan code for the virtual key */
		table[i] = MapVirtualKeyEx( VkTable[i+ENCODING_TABLE_SIZE], 0, hkl ); 
	}
	
	if (SizeAltGr == 0)
		return TRUE;

	table = &((*ScodTable)[ENCODING_TABLE_SIZE * 2]);

	for( i=0; i<SizeAltGr; i++)
	{
		/* Get scan code for the virtual key */
		table[i] = MapVirtualKeyEx( VkTable[i+(ENCODING_TABLE_SIZE*2)], 0, hkl ); 
	}
	
	return TRUE;
}


BOOL createShiftTable( BYTE **Table, UINT *Size )
{
	BYTE *table;

	*Size = SHIFT_TABLE_SIZE;

	table = *Table = new BYTE[ *Size ];
	if (!table) return FALSE;

	ShredData( (BYTE *)(*Table), SHIFT_TABLE_SIZE ); // Fill in the table by random data

	return TRUE;
}


BOOL KBD_GenerateEncodingTable(
		UINT   **VkOriginal,
		UINT   **VkMessed,
		USHORT **ScodOriginal,
		USHORT **ScodMessed,
		UINT   *TableSizeNL,
		UINT   *TableSizeSm,
		UINT   *TableSizeAltGr,
   		BYTE   **ShiftTable,
		UINT   *ShiftTableSize,
		BOOL   UseSecureTyping
	)
{
	ODSFuncS( getWCStr( UseSecureTyping ) );

	UINT  tSizeNL = 0, tSizeSm = 0, tsSize = 0, tSizeAltGr = 0;
	BOOL result;

	result = FALSE;

	HookNext = SetWindowsHookEx( WH_KEYBOARD, HookKeyboardProc, NULL, GetCurrentThreadId() );

	if (!UseSecureTyping) return TRUE;

	CharacterOrder = 0;

	*TableSizeNL = tSizeNL = *TableSizeSm = tSizeSm = *TableSizeAltGr = tSizeAltGr = 0;
	*ShiftTableSize = tsSize = 0;

	if (!createOriginalTable_VK( VkOriginal, &tSizeNL, &tSizeSm, &tSizeAltGr ))
		goto error_exit;
	
	if (!createMessedTable_VK( *VkOriginal, VkMessed, tSizeNL, tSizeSm, tSizeAltGr )) 
		goto error_exit;
	
	if (!createTable_scode( *VkOriginal, ScodOriginal, tSizeNL, tSizeSm, tSizeAltGr )) 
		goto error_exit;

	if (!createTable_scode( *VkMessed, ScodMessed, tSizeNL, tSizeSm, tSizeAltGr )) 
		goto error_exit;

	if (!createShiftTable( ShiftTable, &tsSize )) 
		goto error_exit;

	result = TRUE;
	goto success_exit;

error_exit:
	freeTables( VkOriginal, VkMessed, ScodOriginal, ScodMessed, 
		        ShiftTable, *ShiftTableSize );

success_exit:
	*TableSizeNL    = tSizeNL;
	*TableSizeSm    = tSizeSm;
	*ShiftTableSize = tsSize;
	*TableSizeAltGr = tSizeAltGr;
	return result;
}


#define MHK_HOOK_DEVICE_NT    ("\\\\.\\MHK")
#define MHK_HOOK_DEVICE_95    ("\\\\.\\bcrypt.vxd")

void KBD_getServiceName( char *ServiceName, int size )
{
	OSVERSIONINFO vi;
	vi.dwOSVersionInfoSize = sizeof( vi );

	GetVersionEx( &vi );

	BOOL isNT = (vi.dwPlatformId  == VER_PLATFORM_WIN32_NT);

	if ( isNT )  strcpy_s( ServiceName, size, MHK_HOOK_DEVICE_NT );
	else         strcpy_s( ServiceName, size, MHK_HOOK_DEVICE_95 );
}

bool KBD_CheckDriver()
{
	char	kbdService[32];

	KBD_getServiceName( kbdService, maxLen( kbdService ) );

	HANDLE hDevice =
		CreateFileA( kbdService,
				 GENERIC_READ | GENERIC_WRITE, 0, NULL,
				 OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL );

	if ( hDevice == INVALID_HANDLE_VALUE )
		return FALSE;

	CloseHandle( hDevice );

	return true;
}

int KBD_SetEncodingTable( 	USHORT *ScodOriginal,
							USHORT *ScodMessed,
							UINT   TableSizeNL,
							UINT   TableSizeSm,
							UINT   TableSizeAltGr,
   							BYTE   *ShiftTable,
                            UINT   ShiftTableSize,
							BOOL   UseSecureTyping )
{
	char            kbdService[32];
	ENCODING_BUFFER eBuf = {0};
	BOOL            res = TRUE;
	DWORD           lBytesReturned;
	HANDLE          hDevice;
	UINT            i;
	SHORT           state;
	HKL             hkl = GetKeyboardLayout( 0 );

	if (!UseSecureTyping) return TRUE;

	memset( &eBuf, 0xFF, sizeof( eBuf ) );

	ODSFunc;
/*
	ODSF( L"ScodOriginal 0x%04x, 0x%04x, 0x%04x, 0x%04x", (DWORD)ScodOriginal[0], (DWORD)ScodOriginal[1], (DWORD)ScodOriginal[2], (DWORD)ScodOriginal[3] );
	ODSF( L"ScodMessed   0x%04x, 0x%04x, 0x%04x, 0x%04x", (DWORD)ScodMessed[0], (DWORD)ScodMessed[1], (DWORD)ScodMessed[2], (DWORD)ScodMessed[3] );
	ODSF( L"ShiftTable   0x%04x, 0x%04x, 0x%04x, 0x%04x", (DWORD)ShiftTable[0], (DWORD)ShiftTable[1], (DWORD)ShiftTable[2], (DWORD)ShiftTable[3] );
*/
	//CharacterOrder = 0;
	  
	if ( (TableSizeNL > ENCODING_TABLE_SIZE) || (TableSizeSm > ENCODING_TABLE_SIZE) || 
		(!ScodOriginal) || (!ScodMessed))
	{
		ODS_fls( L"Assertion!" );
		return FALSE;
	}

	eBuf.FlagSetEncoding = 1;
	eBuf.TableSizeNL     = (DWORD)TableSizeNL;
	eBuf.TableSizeSm     = (DWORD)TableSizeSm;
	eBuf.TableSizeAltGr  = (DWORD)TableSizeAltGr;
	eBuf.ShiftTableSize  = (DWORD)ShiftTableSize;
	eBuf.CharacterOrder  = CharacterOrder;

	memset( eBuf.OriginalTable, 0, sizeof(USHORT)*ENCODING_TABLE_SIZE*3 );
	memset( eBuf.MessedTable,   0, sizeof(USHORT)*ENCODING_TABLE_SIZE*3 );
	memset( eBuf.ShiftTable,    0, sizeof(BYTE) * SHIFT_TABLE_SIZE );
	memset( eBuf.Info,          0, MHK_INFO_SIZE * sizeof(USHORT));

	eBuf.Info[INFO_NUMLOCK_CODE_BYTE] = MapVirtualKeyEx( VK_NUMLOCK, 0, hkl );
	eBuf.Info[INFO_SHIFT_CODE_BYTE]   = MapVirtualKeyEx( VK_SHIFT, 0, hkl );
	eBuf.Info[INFO_LSHIFT_CODE_BYTE]  = MapVirtualKeyEx( VK_LSHIFT, 0, hkl );
	eBuf.Info[INFO_RSHIFT_CODE_BYTE]  = MapVirtualKeyEx( VK_RSHIFT, 0, hkl );
	eBuf.Info[INFO_CTRL_CODE_BYTE]    = MapVirtualKeyEx( VK_CONTROL, 0, hkl );
	eBuf.Info[INFO_LCTRL_CODE_BYTE]   = MapVirtualKeyEx( VK_LCONTROL, 0, hkl );
	eBuf.Info[INFO_RCTRL_CODE_BYTE]   = MapVirtualKeyEx( VK_RCONTROL, 0, hkl );
	eBuf.Info[INFO_MENU_CODE_BYTE]    = MapVirtualKeyEx( VK_MENU, 0, hkl );
	eBuf.Info[INFO_LMENU_CODE_BYTE]   = MapVirtualKeyEx( VK_LMENU, 0, hkl );
	eBuf.Info[INFO_RMENU_CODE_BYTE]   = MapVirtualKeyEx( VK_RMENU, 0, hkl );

	state = GetKeyState( VK_NUMLOCK );
	if ((state & 1) > 0 )
	{
		  SET_BIT( eBuf.Info[INFO_STATUS_BYTE], STATUS_NUMLOCK_BIT );
	}

	state = GetKeyState( VK_SHIFT );
	if (((state >> 8) & 0x80) > 0 )
	{
		  SET_BIT( eBuf.Info[INFO_STATUS_BYTE], STATUS_SHIFT_BIT );
	}

	state = GetKeyState( VK_CONTROL );
	if (((state >> 8) & 0x80) > 0 )
	{
		  SET_BIT( eBuf.Info[INFO_STATUS_BYTE], STATUS_CTRL_BIT );
	}

	state = GetKeyState( VK_MENU );
	if (((state >> 8) & 0x80) > 0 )
	{
		  SET_BIT( eBuf.Info[INFO_STATUS_BYTE], STATUS_MENU_BIT );
	}

	for( i=0; i<ENCODING_TABLE_SIZE*3; i++)
	{
		eBuf.OriginalTable[i] = ScodOriginal[i];
		eBuf.MessedTable[i]   = ScodMessed[i];
	}

	for( i=0; i<ShiftTableSize; i++)
	{
		eBuf.ShiftTable[i] = ShiftTable[i];
	}

	KBD_getServiceName( kbdService, maxLen( kbdService ) );

	memcpy( eBuf.Signature, "LOCOS97 ", maxLen( eBuf.Signature ) );

	hDevice = CreateFileA( kbdService,
						 GENERIC_READ | GENERIC_WRITE, 0, NULL,
						 OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL );

	if (hDevice == INVALID_HANDLE_VALUE)
	{
		ODS_fls( L"Assertion!" );
		return 0;
	}

	res = DeviceIoControl( hDevice, IOCTL_PRIVATE_SET_RESET_ENCODING,
						 &eBuf, sizeof(ENCODING_BUFFER),
						 &eBuf, sizeof(ENCODING_BUFFER),
						 &lBytesReturned, NULL);
	assert( res );

	CloseHandle(hDevice);

	if (strncmp(eBuf.Signature, "LS06CXX ", 8) != 0)
	{
		ODS_fls( L"Assertion!" );
		return 0;
	}

	if ((eBuf.Status) == STATUS_ERROR_ENCODING_ALREADY_SET )
	{
		ODS_fls( L"Assertion!" );
		return 2;
	}

	if ((eBuf.Status) != STATUS_KBD_SUCCESS)
	{
		ODS_fls( L"Assertion!" );
		return 0;
	}

	return res;
}


BOOL KBD_ResetEncodingTable( BOOL   UseSecureTyping )
{
	char            kbdService[32];
	ENCODING_BUFFER eBuf;
	BOOL            res = TRUE;
	DWORD           lBytesReturned;
	HANDLE          hDevice;

	if (!UseSecureTyping) return TRUE;

	ODSFunc;

	memset( &eBuf, 0, sizeof(ENCODING_BUFFER) );

	eBuf.FlagSetEncoding = 0;
	 
	KBD_getServiceName( kbdService, maxLen( kbdService ) );

	memcpy( eBuf.Signature, "LOCOS97 ", maxLen( eBuf.Signature) );

	hDevice = CreateFileA( kbdService,
						 GENERIC_READ | GENERIC_WRITE, 0, NULL,
						 OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL );
	if (hDevice == INVALID_HANDLE_VALUE) return FALSE;

	res = DeviceIoControl( hDevice, IOCTL_PRIVATE_SET_RESET_ENCODING,
						 &eBuf, sizeof(ENCODING_BUFFER),
						 &eBuf, sizeof(ENCODING_BUFFER),
						 &lBytesReturned, NULL);
	CloseHandle(hDevice);

	if (strncmp(eBuf.Signature, "LS06CXX ", 8) != 0)
	   return FALSE;

	if ((eBuf.Status) != STATUS_KBD_SUCCESS) return FALSE;

	return res;
}

BOOL KBD_FreeTables(UINT   **VkOriginal,
					UINT   **VkMessed,
					USHORT **ScodOriginal,
					USHORT **ScodMessed,
					BYTE   **ShiftTable,
                    UINT   ShiftTableSize,
					BOOL   UseSecureTyping )
{

	if (HookNext)
	{
		UnhookWindowsHookEx( HookNext );
	}

	if ( !UseSecureTyping )
		return TRUE;

	freeTables( VkOriginal, VkMessed, ScodOriginal, ScodMessed,
		        ShiftTable, ShiftTableSize );

	return TRUE;
}




/***************************************************************

 Translate single character depending on whether the following 
 keys are pressed:
	VK_LSHIFT
	VK_RSHIFT
	VK_LCONTROL
	VK_RCONTROL
	VK_CAPITAL
	VK_NUMLOCK

 ***************************************************************/


/*
BOOL createTable_charEx( UINT *VkTable, char **CharTable, UINT Size )
{
	UINT i;
	char  *table;
  	UINT  scanCode;
	BYTE  keyState[256];
	SHORT state;
	WORD  chars;
	int   charsGot;

	table = *CharTable = new char[ Size ];
	if (!table) return FALSE;

	memset( keyState,  0x00, 256 ); // nothing is pressed or toggled
	
	// set correct state of Shift, Control and Caps Lock
	state = GetKeyState( VK_SHIFT );
	keyState[VK_SHIFT] = (((BYTE)state) & 1) | (((BYTE)(state >> 8)) & 0x80);

	state = GetKeyState( VK_CONTROL );
	keyState[VK_CONTROL] = (((BYTE)state) & 1) | (((BYTE)(state >> 8)) & 0x80);

	state = GetKeyState( VK_LSHIFT );
	keyState[VK_LSHIFT] = (((BYTE)state) & 1) | (((BYTE)(state >> 8)) & 0x80);

	state = GetKeyState( VK_RSHIFT );
	keyState[VK_RSHIFT] = (((BYTE)state) & 1) | (((BYTE)(state >> 8)) & 0x80);

	state = GetKeyState( VK_LCONTROL );
	keyState[VK_LCONTROL] = (((BYTE)state) & 1) | (((BYTE)(state >> 8)) & 0x80);

	state = GetKeyState( VK_RCONTROL );
	keyState[VK_RCONTROL] = (((BYTE)state) & 1) | (((BYTE)(state >> 8)) & 0x80);

	state = GetKeyState( VK_CAPITAL );
	keyState[VK_CAPITAL] = (((BYTE)state) & 1) | (((BYTE)(state >> 8)) & 0x80);

	state = GetKeyState( VK_NUMLOCK );
	keyState[VK_NUMLOCK] = (((BYTE)state) & 1) | (((BYTE)(state >> 8)) & 0x80);


	HKL hkl = GetKeyboardLayout( 0 );
	
	// call ToAsciiEx just to clear keyboard from dead key, if it was pressed
	scanCode = MapVirtualKeyEx( VkTable[0], 0, hkl );
	charsGot = ToAsciiEx( VkTable[0], scanCode, keyState, &chars, 0, hkl );


	for( i=0; i<Size; i++)
	{
		// Get scan code for the virtual key
		scanCode = MapVirtualKeyEx( VkTable[i], 0, hkl ); 

		// Get character for the given virtual key
		charsGot = ToAsciiEx(  VkTable[i], scanCode, keyState, &chars, 0, hkl );
		if (charsGot != 1) return FALSE;
		table[i] = (char)chars;
	}
	
	return TRUE;
}
*/

/*************************************************************

  ShiftTable is used as 
   a) direct  transformation: N = (X + R) % L 
   b) reverse transformation: X = (N + (L - (R % L))) % L

  where X - initial index, 
        R - random from ShiftTable, 
		L - TableSize - number of messed characters,
		N - new index

  In KBD_TranslateVirtualKey() we use reverse transformation

**************************************************************/

void definePressedControlKeys( BYTE *KsTable )
{
	SHORT state;

	memset( KsTable,  0x00, 256 ); // nothing is pressed or toggled
	
	// set correct state of Shift, Control and Caps Lock
	state = GetKeyState( VK_SHIFT );
	KsTable[VK_SHIFT] = (((BYTE)state) & 1) | (((BYTE)(state >> 8)) & 0x80);

	state = GetKeyState( VK_CONTROL );
	KsTable[VK_CONTROL] = (((BYTE)state) & 1) | (((BYTE)(state >> 8)) & 0x80);

	state = GetKeyState( VK_LSHIFT );
	KsTable[VK_LSHIFT] = (((BYTE)state) & 1) | (((BYTE)(state >> 8)) & 0x80);

	state = GetKeyState( VK_RSHIFT );
	KsTable[VK_RSHIFT] = (((BYTE)state) & 1) | (((BYTE)(state >> 8)) & 0x80);

	state = GetKeyState( VK_LCONTROL );
	KsTable[VK_LCONTROL] = (((BYTE)state) & 1) | (((BYTE)(state >> 8)) & 0x80);

	state = GetKeyState( VK_RCONTROL );
	KsTable[VK_RCONTROL] = (((BYTE)state) & 1) | (((BYTE)(state >> 8)) & 0x80);

	state = GetKeyState( VK_CAPITAL );
	KsTable[VK_CAPITAL] = (((BYTE)state) & 1) | (((BYTE)(state >> 8)) & 0x80);

	state = GetKeyState( VK_NUMLOCK );
	KsTable[VK_NUMLOCK] = (((BYTE)state) & 1) | (((BYTE)(state >> 8)) & 0x80);

	state = GetKeyState( VK_MENU );
	KsTable[ VK_MENU ] = (((BYTE)state) & 1) | (((BYTE)(state >> 8)) & 0x80);

	state = GetKeyState( VK_LMENU );
	KsTable[ VK_LMENU ] = (((BYTE)state) & 1) | (((BYTE)(state >> 8)) & 0x80);

	state = GetKeyState( VK_RMENU );
	KsTable[ VK_RMENU ] = (((BYTE)state) & 1) | (((BYTE)(state >> 8)) & 0x80);
}


BOOL  KBD_WasDeadChar = FALSE;


LRESULT CALLBACK HookKeyboardProc(
  int    nCode,       // hook code
  WPARAM wParam,  // virtual-key code
  LPARAM lParam   // keystroke-message information
)
{
	UINT      virtualKey, scanCode;
	HWND      hWnd;
	kbdParams *kbd = NULL;
	BOOL      bPressed;
	HKL       hkl;

	bPressed = ((lParam & 0x80000000) == 0);

	if ((KBD_WasDeadChar) && bPressed)
	{
		hWnd = GetFocus();

		kbd = kbdParams::GetParams( GetParent( hWnd ) );

		//
		//	Note 3. It is not required to process dead char if secure typing is off.
		//	Othewise Dutch ( Belgium ) key board does not send WM_CHAR message at all when ALT + ( Double ? )
		//	is pressed ( required result is '~~' ).
		//

		if ( ( kbd == NULL ) || ( !kbd->bUseSecureTyping ) )
		{
			KBD_WasDeadChar = false;
			return CallNextHookEx( HookNext, nCode, wParam, lParam );
		}

		virtualKey = (UINT)wParam;
		scanCode = (UINT)((lParam >> 16) & 0xFF);

		hkl = GetKeyboardLayout( 0 ); // check that printable character was entered
		if (!MapVirtualKeyEx( virtualKey, 2, hkl ))
			return CallNextHookEx( HookNext, nCode, wParam, lParam );

		kbd->ReactOnDeadchar( &virtualKey, kbd->bUseSecureTyping, &scanCode );

		KBD_WasDeadChar = FALSE;
		// kbd->keyState.wasDeadcharBefore = KBD_WasDeadChar;
	}

	return CallNextHookEx( HookNext, nCode, wParam, lParam );
}

static int _ToUnicodeEx( UINT wVirtKey, UINT wScanCode, CONST BYTE *lpKeyState, LPWSTR pwszBuff, int cchBuff, UINT wFlags, HKL dwhkl, char *ReturnChar1, char *ReturnChar2 )
{
	int charsGot;

	static CDll dll( false, "user32.dll", "ToUnicodeEx" );

	if ( dll.error )
	{
		if ( ReturnChar1 == NULL )
		{
			assert( false );
			return 0;
		}

		WORD chars = 0;

		charsGot = ToAsciiEx( wVirtKey, wScanCode, lpKeyState, &chars, wFlags, dwhkl );

		//KeyState->ReturnCharsNumber = (DWORD)charsGot;
		if (charsGot == 2)
		{
			*ReturnChar1 = (char)(chars >> 8);
			*ReturnChar2 = (char)chars;
		}
		else
		{
			*ReturnChar2 = (char)chars;
		}

		return charsGot;
	}

	typedef int WINAPI typeToUnicodeEx(
		UINT wVirtKey,
		UINT wScanCode,
		CONST BYTE *lpKeyState,
		LPWSTR pwszBuff,
		int cchBuff,
		UINT wFlags,
		HKL dwhkl);

	typeToUnicodeEx *procToUnicodeEx = (typeToUnicodeEx *)dll.pProc;

	charsGot = procToUnicodeEx( wVirtKey, wScanCode, lpKeyState, pwszBuff, cchBuff, wFlags, dwhkl );

	if ( ( charsGot > 0 ) && ( charsGot < cchBuff ) )
		pwszBuff[ charsGot ] = L'\0';

	if ( ReturnChar1 == NULL )
		return charsGot;

	//ReturnCharsNumber = (DWORD)charsGot;

	// cMBStr mbs( pwszBuff, charsGot, false, CP_ACP );

	char bf[2] = {0};
	BOOL defCharFlag = false;

	int res = WideCharToMultiByte( CP_ACP, 0, pwszBuff, charsGot, bf, maxLen( bf ), "?", &defCharFlag );
	
	if ( defCharFlag )
	{
		//ODS_fls( _FormatT( L"'%s' %d, %c %c", pwszBuff, charsGot, (wchar_t)bf[0], (wchar_t)bf[1] ) );

		const char *szKBTranslationError = "Key board symbol was not properly translated";
		MessageBoxA( NULL, szKBTranslationError, __FUNCTION__, MB_ICONSTOP | MB_SYSTEMMODAL );

		assert( false );
		return 0;
	}

	if ( charsGot == 2 )
	{
		*ReturnChar1 = bf[0];
		*ReturnChar2 = bf[1];
	}
	else
	{
		*ReturnChar2 = bf[0];
	}

	return charsGot;
}


BOOL KBD_ReactOnDeadchar(KEY_STATE *KeyState,
						UINT  *VirtualKey,
						UINT   **VkOriginal,
						UINT   **VkMessed,
						USHORT **ScodOriginal,
						USHORT **ScodMessed,
						UINT   TableSizeNL,
						UINT   TableSizeSm,
						UINT   TableSizeAltGr,
						BYTE   **ShiftTable,
						UINT   ShiftTableSize,
						BOOL   UseSecureTyping,
						UINT   *ScanCode
						)
{
	UINT   i, index, tabSize;
	UINT   *vkOriginal, *vkMessed;
	USHORT *scodOriginal, *scodMessed;
	BOOL   bNumLock;
	HKL    hkl;
	BYTE   kS[256];
	//int    charsGot;
	//WORD   chars;

	if (!KeyState->wasDeadcharBefore) // Was WM_DEADCHAR before?
		return TRUE;                  // If no, it is an error situation. We should not be here.
		                              
	// KeyState->wasDeadcharBefore will be reset in subsequent WM_CHAR message

	if ( !UseSecureTyping ) 
	{
		hkl = GetKeyboardLayout( 0 );

		UINT scanCode = MapVirtualKeyEx( *VirtualKey, 0, hkl );

		//if ((scanCode & 0x1FF) != ((*ScanCode) & 0x1FF)) 
		//	continue; // the numbers was pressed on num pad

		//index = (i + (tabSize - ( ((*ShiftTable)[CharacterOrder%ShiftTableSize]) % tabSize))) % tabSize;
		//*VirtualKey = vkOriginal[index];
		//*ScanCode   = scodOriginal[index];

		definePressedControlKeys( kS );

		KeyState->ReturnCharsNumber = (DWORD)_ToUnicodeEx( *VirtualKey, *ScanCode, kS, KeyState->wcs, maxLen( KeyState->wcs ), 0, hkl, &(KeyState->ReturnChar1), &(KeyState->ReturnChar2) );

		//
		//	Valid locale properly translates the chars.
		//
		//KeyState->ReturnChar2 = (char)(KeyState->wcs[0]);
		//KeyState->ReturnChar1 = (char)(KeyState->wcs[1]);

		return TRUE;
	}

	if ((!VirtualKey) || (!(*VkOriginal)) || (!(*VkMessed)) || (!(*ScodOriginal)) || (!(*ScodMessed)) )
		goto ts_exit;

	// check if NumLock is pressed
	tabSize = TableSizeSm;
	vkOriginal = &((*VkOriginal)[ENCODING_TABLE_SIZE]);
	vkMessed   = &((*VkMessed)[ENCODING_TABLE_SIZE]);
	scodOriginal = &((*ScodOriginal)[ENCODING_TABLE_SIZE]);
	scodMessed   = &((*ScodMessed)[ENCODING_TABLE_SIZE]);
	bNumLock   = FALSE;

	if ((GetKeyState( VK_NUMLOCK ) & 1) &&
	   (!((GetKeyState( VK_SHIFT ) >> 8) & 0x80)) &&
	   (!((GetKeyState( VK_CONTROL ) >> 8) & 0x80)) )
	{
		tabSize = TableSizeNL;
		vkOriginal = *VkOriginal;
		vkMessed   = *VkMessed;
		scodOriginal = *ScodOriginal;
		scodMessed   = *ScodMessed;
		bNumLock   = TRUE;
	}

	if ( ((GetKeyState( VK_MENU ) >> 8) & 0x80) && ((GetKeyState( VK_RMENU ) >> 8) & 0x80) && (TableSizeAltGr != 0))
	{
		tabSize = TableSizeAltGr;
		vkOriginal = &((*VkOriginal)[ENCODING_TABLE_SIZE*2]);
		vkMessed   = &((*VkMessed)[ENCODING_TABLE_SIZE*2]);
		scodOriginal = &((*ScodOriginal)[ENCODING_TABLE_SIZE*2]);
		scodMessed   = &((*ScodMessed)[ENCODING_TABLE_SIZE*2]);
		bNumLock   = FALSE;
	}

	/*
	if (!createTable_charEx( vkOriginal, &charOriginal, tabSize )) 
		goto ts_exit;

	if (!createTable_charEx( vkMessed, &charMessed, tabSize )) 
		goto ts_exit;
	*/

	for(i=0; i<tabSize; i++)
	{ 
		if ((*VirtualKey) == vkMessed[i])
		{
			hkl = GetKeyboardLayout( 0 );

			UINT scanCode = MapVirtualKeyEx( *VirtualKey, 0, hkl );

			if ((scanCode & 0x1FF) != ((*ScanCode) & 0x1FF)) 
				continue; // the numbers was pressed on num pad

			index = (i + (tabSize - ( ((*ShiftTable)[CharacterOrder%ShiftTableSize]) % tabSize))) % tabSize;
			*VirtualKey = vkOriginal[index];
			*ScanCode   = scodOriginal[index];

			definePressedControlKeys( kS );

			KeyState->ReturnCharsNumber = (DWORD)_ToUnicodeEx( *VirtualKey, *ScanCode, kS, KeyState->wcs, maxLen( KeyState->wcs ), 0, hkl, &(KeyState->ReturnChar1), &(KeyState->ReturnChar2) );

			//
			// Valid locale properly translates the chars.
			//
			//KeyState->ReturnChar2 = (char)(KeyState->wcs[0]);
			//KeyState->ReturnChar1 = (char)(KeyState->wcs[1]);

			//charsGot = ToAsciiEx( *VirtualKey, *ScanCode, kS, &chars, 0, hkl );

			//KeyState->ReturnCharsNumber = (DWORD)charsGot;
			//if (charsGot == 2)
			//{
			//	KeyState->ReturnChar1 = (char)(chars >> 8);
			//	KeyState->ReturnChar2 = (char)chars;
			//}
			//else
			//{
			//	KeyState->ReturnChar2 = (char)chars;
			//}

			// Commented to fix the problem with Windows Commander.
			// See also keyboard hook in mhk.sys (Windows NT) and bcrypt.vxd (Windows 95/98)
			//CharacterOrder++;
			return TRUE;
		}
	}

ts_exit:
	return false;
}

//
// KBD_WM_CHAR
//
//	If an error occuares or if sequre  typing is on but virtual key was not found at vkMessed table the function returs false.
//	It allows to process WM_CHAR message by standard procedurte.
//
BOOL KBD_WM_CHAR(	char   OriginalLetter,
					KEY_STATE *KeyState,
					UINT   **VkOriginal,
					UINT   **VkMessed,
					USHORT **ScodOriginal,
					USHORT **ScodMessed,
					UINT   TableSizeNL,
					UINT   TableSizeSm,
					UINT   TableSizeAltGr,
					BYTE   **ShiftTable,
					UINT   ShiftTableSize,
					BOOL   UseSecureTyping,
					DWORD  *ReturnCharsNumber,
					char   *ReturnChar1,
					char   *ReturnChar2,
					wchar_t *wcs,
					WPARAM *pWP,
					LPARAM *pLP
					)
{
	UINT   i, index, tabSize;
	UINT   *vkOriginal, *vkMessed;
	USHORT *scodOriginal, *scodMessed;
	BOOL   bNumLock;
	HKL    hkl;
	BYTE   kS[256];
	//int    charsGot;
	//WORD   chars;

	ODSFunc;
	//ODSFuncS( _FormatT( L"%X, %c", (int)OriginalLetter, (int)OriginalLetter ) );

	*ReturnCharsNumber = 1;
	if ( ReturnChar2 != NULL ) *ReturnChar2 = OriginalLetter;
	if ( wcs != NULL )
	{
		// service symbols
		if ( pWP != NULL )
		{
#ifdef _UNICODE
			wcs[0] = *pWP;
#else
			_mbstowcs( wcs, 1, (char*)pWP, 1, false, CP_ACP );
#endif
		}
		else
			wcs[0] = OriginalLetter;

		wcs[1] = L'\0';
	}

	if (KeyState->wasDeadcharBefore) // WM_DEADCHAR was before, just return what we have defined in previous second WM_KEYDOWN
	{
		*ReturnCharsNumber = KeyState->ReturnCharsNumber;

		if ( ReturnChar1 != NULL ) *ReturnChar1 = KeyState->ReturnChar1;
		if ( ReturnChar2 != NULL ) *ReturnChar2 = KeyState->ReturnChar2;

		if ( wcs != NULL )
			wcscpy_s( wcs, maxLen( KeyState->wcs ), KeyState->wcs );

		memset( KeyState, 0, sizeof(KEY_STATE) );//  - sizeof( KeyState->wcs ) );

		//ODS_fls( _FormatT( L"wasDeadcharBefore - %s", KeyState->wcs ) );

		return ( *ReturnCharsNumber != 0 );
	}

	if ( !UseSecureTyping )
	{
		hkl = GetKeyboardLayout( 0 );

		UINT scanCode = MapVirtualKeyEx( KeyState->saved_vk, 0, hkl );

		//if ((scanCode & 0x1FF) != ((KeyState->saved_sc) & 0x1FF)) 
		//	continue; // the numbers was pressed on num pad

		//index = (i + (tabSize - ( ((*ShiftTable)[CharacterOrder%ShiftTableSize]) % tabSize))) % tabSize;
		//KeyState->saved_vk = vkOriginal[index];
		//KeyState->saved_sc = scodOriginal[index];

		definePressedControlKeys( kS );

		int wcsSize = 10;

		if ( wcs == NULL )
		{
			wcs = KeyState->wcs;
			wcsSize = maxLen( KeyState->wcs );
		}
		
		*ReturnCharsNumber = (DWORD)_ToUnicodeEx( KeyState->saved_vk, KeyState->saved_sc, kS, wcs, wcsSize, 0, hkl, ReturnChar1, ReturnChar2 );

		memset( KeyState, 0, sizeof(KEY_STATE) );

		if ( *ReturnCharsNumber <= 0 )
		{
			return false;
		}

		//ODS_fls( _FormatT( L"Not UseSecureTyping - %s", KeyState->wcs ) );
		return TRUE;
	}

	// If we here, hence, there were no WM_DEADCHAR before.
	// Let's define only one character that must be returned in this WM_CHAR

	// check if NumLock is pressed
	tabSize    = TableSizeSm;
	vkOriginal = &((*VkOriginal)[ENCODING_TABLE_SIZE]);
	vkMessed   = &((*VkMessed)[ENCODING_TABLE_SIZE]);
	scodOriginal = &((*ScodOriginal)[ENCODING_TABLE_SIZE]);
	scodMessed   = &((*ScodMessed)[ENCODING_TABLE_SIZE]);
	bNumLock   = FALSE;

	if ((GetKeyState( VK_NUMLOCK ) & 1) &&
	   (!((GetKeyState( VK_SHIFT ) >> 8) & 0x80)) &&
	   (!((GetKeyState( VK_CONTROL ) >> 8) & 0x80)) )
	{
		tabSize = TableSizeNL;
		vkOriginal = *VkOriginal;
		vkMessed   = *VkMessed;
		scodOriginal = *ScodOriginal;
		scodMessed   = *ScodMessed;
		bNumLock   = TRUE;
	}

	if ( ((GetKeyState( VK_MENU ) >> 8) & 0x80) && ((GetKeyState( VK_RMENU ) >> 8) & 0x80) && (TableSizeAltGr != 0))
	{
		tabSize = TableSizeAltGr;
		vkOriginal = &((*VkOriginal)[ENCODING_TABLE_SIZE*2]);
		vkMessed   = &((*VkMessed)[ENCODING_TABLE_SIZE*2]);
		scodOriginal = &((*ScodOriginal)[ENCODING_TABLE_SIZE*2]);
		scodMessed   = &((*ScodMessed)[ENCODING_TABLE_SIZE*2]);
		bNumLock   = FALSE;
	}

	if ((!(*VkOriginal)) || (!(*VkMessed)) || (!(*ScodOriginal)) || (!(*ScodMessed)) )
		goto ts_exit;
	
	for(i=0; i<tabSize; i++)
	{ 
		if (KeyState->saved_vk == vkMessed[i])
		{
			hkl = GetKeyboardLayout( 0 );

			UINT scanCode = MapVirtualKeyEx( KeyState->saved_vk, 0, hkl );

			if ((scanCode & 0x1FF) != ((KeyState->saved_sc) & 0x1FF)) 
				continue; // the numbers was pressed on num pad

			index = (i + (tabSize - ( ((*ShiftTable)[CharacterOrder%ShiftTableSize]) % tabSize))) % tabSize;
			KeyState->saved_vk = vkOriginal[index];
			KeyState->saved_sc = scodOriginal[index];

			//
			//	Just restore parameters for original edit box procedure call if service secuence is occuared
			//
			if ( pWP != NULL ) *pWP = KeyState->saved_vk;
			if ( pLP != NULL ) *pLP = ( KeyState->saved_sc << 16 ) | ( *pLP & 0xFFFF );

			definePressedControlKeys( kS );

			int wcsSize = 10;

			if ( wcs == NULL )
			{
				wcs = KeyState->wcs;
				wcsSize = maxLen( KeyState->wcs );
			}
			
			*ReturnCharsNumber = (DWORD)_ToUnicodeEx( KeyState->saved_vk, KeyState->saved_sc, kS, wcs, wcsSize, 0, hkl, ReturnChar1, ReturnChar2 );

			memset( KeyState, 0, sizeof(KEY_STATE) );

			if ( *ReturnCharsNumber <= 0 )
			{
				return false;
			}

			// Commented to fix the problem with Windows Commander.
			// See also keyboard hook in mhk.sys (Windows NT) and bcrypt.vxd (Windows 95/98)
			//CharacterOrder++;
			//ODSFuncRes( KeyState->wcs );
			return TRUE;
		}
	}

ts_exit:
	
	ODSFuncRes( L"Error!" );
	return false; // TRUE;
}



void KBD_IncrementCharacterOrder(int	VirtualKey, 
								 UINT	*VkMessed,
								 UINT	TableSizeNL,
								 UINT	TableSizeSm,
								 UINT	TableSizeAltGr,
								 BOOL	UseSecureTyping )
{
	UINT i;
	UINT *vkMessed, tabSize;

	if (!UseSecureTyping) return;

	ODSFuncS( getWCStr( (DWORD)CharacterOrder ) );

	// check if NumLock is pressed
	tabSize = TableSizeSm;
	vkMessed   = &(VkMessed[ENCODING_TABLE_SIZE]);

	if ((GetKeyState( VK_NUMLOCK ) & 1) &&
	   (!((GetKeyState( VK_SHIFT ) >> 8) & 0x80)) &&
	   (!((GetKeyState( VK_CONTROL ) >> 8) & 0x80)) )
	{
		tabSize  = TableSizeNL;
		vkMessed = VkMessed;
	}

	if ( ((GetKeyState( VK_MENU ) >> 8) & 0x80) && ((GetKeyState( VK_RMENU ) >> 8) & 0x80) && (TableSizeAltGr != 0))
	{
		tabSize = TableSizeAltGr;
		vkMessed   = &(VkMessed[ENCODING_TABLE_SIZE*2]);
	}

	for(i=0; i<tabSize; i++)
	{ 
		if (VirtualKey == (int)(vkMessed[i]))
		{
			CharacterOrder++;
			break;
		}
	}

	ODSFuncRes( (DWORD)CharacterOrder );
}



void KBD_IncrementCharacterOrderCNTRL(	int		VirtualKey, 
										UINT	*VkMessed,
										UINT	TableSizeNL,
										UINT	TableSizeSm,
										BOOL	UseSecureTyping )  
{
    // Commented to fix the problem with Windows Commander.
	// See also keyboard hook in mhk.sys (Windows NT) and bcrypt.vxd (Windows 95/98)

	return;
}

BOOL KBD_sendCommandToMhkDriver( LPVOID Buffer, DWORD BufferSize, DWORD CtlCode )
{
	HANDLE         hDevice;
	DWORD          lBytesReturned = 0;
	PBUFFER_HEADER bufHead = (PBUFFER_HEADER)Buffer;

	char	kbdService[32];

	KBD_getServiceName( kbdService, maxLen( kbdService ) );

	hDevice = CreateFileA( kbdService, GENERIC_READ | GENERIC_WRITE,
					  0, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL );

	if ( hDevice == INVALID_HANDLE_VALUE )
	{
			return FALSE;
	}

	memcpy( bufHead->Signature, "LOCOS97 ", maxLen( bufHead->Signature ) );

	if (!DeviceIoControl( hDevice, CtlCode,
											  Buffer, BufferSize,
											  Buffer, BufferSize,
											  &lBytesReturned, NULL))
	{
			CloseHandle(hDevice);
			return FALSE;
	}

	CloseHandle(hDevice);

	if (strncmp(bufHead->Signature,"LS06CXX ",8) != 0)
	{
			return FALSE;
	}

	if ( bufHead->Status != BCPSTATUS_SUCCESS )
	{
			return FALSE;
	}

	return TRUE;
}

BOOL KBD_IsAnyKeyPressed()
{
	MHK_KEY_PRESSED_BUFFER kBuf={0};

	//memset(&kBuf, 0, sizeof(MHK_KEY_PRESSED_BUFFER));

	if (!KBD_sendCommandToMhkDriver((LPVOID)&kBuf, sizeof(MHK_KEY_PRESSED_BUFFER), IOCTL_PRIVATE_MHK_CONFIRM_KEY_PRESSED))
	{
		return TRUE; // Our MHK driver does not support IOCTL_PRIVATE_MHK_CONFIRM_KEY_PRESSED request, let's return TRUE as if key pressing was intercepted.
	}

	return kBuf.KeyPressed;
}
