/*******************************************************************
 *	Copyright (c) 1994-2013 Jetico, Inc., Finland
 *	All rights reserved.
 *
 *	File: kbd.h
 *	Revision: 
 *	Created:
 *	Description: interface with the BestCrypt keyboard driver
 *		09.12.2004 - Semaphores added for Set/Reset 
 *		encoding table procedure.( See description at
 *		kbdParams::hSemaphore
 *
 *******************************************************************/


#ifndef __KBD
#define __KBD

#include <assert.h>
#include <stdio.h>
#include <time.h>

#include "..\..\Includes\UnicodeFuncs.h"

typedef struct _KEY_STATE
{
	BOOL  wasDeadcharBefore;
	UINT  saved_vk;
	UINT  saved_sc;
	DWORD ReturnCharsNumber;
	char  ReturnChar1;
	char  ReturnChar2;

	wchar_t wcs[10];

} KEY_STATE;

extern BOOL KBD_WasDeadChar;

extern bool KBD_CheckDriver();

extern BOOL KBD_GenerateEncodingTable(
										UINT   **VkOriginal,
										UINT   **VkMessed,
										USHORT **ScodOriginal,
										USHORT **ScodMessed,
										UINT   *TableSizeNL,
										UINT   *TableSizeSm,
										UINT   *TableSizeAltGr,
										BYTE   **ShiftTable,
										UINT   *ShiftTableSize,
										BOOL   UseSecureTyping);


//
//	KBD_SetEncodingTable
//		returns:
//			1 if success
//			0 if error
//			2 if Prev table was not reset yet.
//
extern int KBD_SetEncodingTable( 	USHORT *ScodOriginal,
									USHORT *ScodMessed,
								    UINT   TableSizeNL,
								    UINT   TableSizeSm,
									UINT   TableSizeAltGr,
   									BYTE   *ShiftTable,
									UINT   ShiftTableSize,
									BOOL   UseSecureTyping);

extern BOOL KBD_ResetEncodingTable( BOOL   UseSecureTyping );


extern BOOL KBD_FreeTables(	UINT   **VkOriginal,
							UINT   **VkMessed,
							USHORT **ScodOriginal,
							USHORT **ScodMessed,
							BYTE   **ShiftTable,
							UINT   ShiftTableSize,
							BOOL   UseSecureTyping );


extern BOOL KBD_ReactOnDeadchar(KEY_STATE *KeyState,
								UINT  *VirtualKey,
								UINT   **VkOriginal,
								UINT   **VkMessed,
								USHORT **ScodOriginal,
								USHORT **ScodMessed,
								UINT   TableSizeNL,
								UINT   TableSizeSm,
								UINT   TableSizeAltGr,
								BYTE   **ShiftTable,
								UINT   ShiftTableSize,
								BOOL   UseSecureTyping,
								UINT   *ScanCode );

extern BOOL KBD_WM_CHAR( char OriginalChar,
						KEY_STATE *KeyState,
						UINT   **VkOriginal,
						UINT   **VkMessed,
						USHORT **ScodOriginal,
						USHORT **ScodMessed,
						UINT   TableSizeNL,
						UINT   TableSizeSm,
						UINT   TableSizeAltGr,
						BYTE   **ShiftTable,
						UINT   ShiftTableSize,
						BOOL   UseSecureTyping,
						DWORD  *ReturnCharsNumber,
						char   *ReturnChar1,
						char   *ReturnChar2,
						wchar_t *wcs = NULL,
						WPARAM *pWP = NULL,
						LPARAM *pLP = NULL
					);


extern void KBD_IncrementCharacterOrder( int	VirtualKey, 
										 UINT	*VkMessed,
										 UINT   TableSizeNL,
										 UINT   TableSizeSm,
										 UINT	TableSizeAltGr,
										 BOOL	UseSecureTyping );

extern void KBD_IncrementCharacterOrderCNTRL(	int		VirtualKey, 
												UINT	*VkMessed,
												UINT	TableSizeNL,
												UINT	TableSizeSm,
												BOOL	UseSecureTyping );


extern void KBD_getServiceName( char *ServiceName, int size );
extern BOOL KBD_sendCommandToMhkDriver( LPVOID Buffer, DWORD BufferSize, DWORD CtlCode );

extern BOOL KBD_IsAnyKeyPressed();

// #define _DEBUG_PRINT

#ifndef _DEBUG_PRINT
#define DbgPrint( a, b ) {};
#endif


#ifndef DWLP_USER
#define DWLP_USER DWL_USER
#define GetWindowLongPtr GetWindowLong
#endif

//
//	kbdData has been changed for GWLP_USERDATA because kbd.h saves lbd at parent dialogue
//	but the DWLP_USER storage place is used by cDlg as dlg class body.
//	Note, kgsha was not updated with the new settings ( kgsha256 and kgGhost only ).
//
#define kbdData GWLP_USERDATA // DWLP_USER

struct kbdParams
{
	UINT      *VkOriginal,   *VkMessed;
	USHORT    *ScodOriginal, *ScodMessed;
	BYTE      *ShiftTable;
	UINT      TableSizeNL, TableSizeSm, ShiftTableSize, TableSizeAltGr;
	KEY_STATE keyState;

	//
	//	The semaphore object has been added by one reason:
	//	On any machines the WM_KILLFOCUS message sends later then WM_SETFOCUS.
	//	As result in UseSecureTyping mode the encoding tables in kbd driver 
	//	may be changed not correctly. 
	//	One edit box may try to set tables when other edit box does not reset 
	//	it's own tables.
	//
	HANDLE hSemaphore;

	//
	//	Bug #1527 - Error in setting encoding table
	//	13.05.2014
	//		If input local is changed Windows 8 calls WM_INPUTLANGCHANGE later then WM_KILLFOCUS it is correct, 
	//		All other systems call WM_KILLFOCUS first and the way worked properly.
	//		But for Windows 8 we get timeout at WM_SETFOCUS after WM_INPUTLANGCHANGE.
	//	ODS log for Windows 8:
	//		0820:0B1C 00:090   <Stop>  convertTextToFilenamePath
	//		0820:0B1C 00:121   CreateSemaphore = 0x168, 0, NOT exist
	//		0820:0B1C 00:121   kbdParams::GenerateEncodingTable:278 ReleaseSemaphore 0
	//		0820:0B1C 00:137   WM_SETFOCUS:
	//		0820:0B1C 00:137   kbdParams::SetEncodingTable:309 WaitForSingleObject start
	//		0820:0B1C 00:152   kbdParams::SetEncodingTable:311 WaitForSingleObject stop = 0x0
	//		0820:0B1C 00:152   SetEncodingTable = 0x1
	//		0820:0B1C 01:172   WM_KILLFOCUS:
	//		0820:0B1C 01:172   kbdParams::ResetEncodingTable:295 ReleaseSemaphore 0, res = 1
	//		0820:0B1C 05:247   WM_INPUTLANGCHANGE:
	//		0820:0B1C 05:247   kbdParams::ResetEncodingTable:295 ReleaseSemaphore 1, res = 1
	//		0820:0B1C 05:262   CreateSemaphore = 0x168, 0, NOT exist
	//		0820:0B1C 05:262   kbdParams::GenerateEncodingTable:278 ReleaseSemaphore 0
	//		0820:0B1C 05:278   kbdParams::SetEncodingTable:309 WaitForSingleObject start
	//		0820:0B1C 05:295   kbdParams::SetEncodingTable:311 WaitForSingleObject stop = 0x0
	//		0820:0B1C 05:295   SetEncodingTable = 0x1
	//		0820:0B1C 05:309   WM_SETFOCUS:
	//		0820:0B1C 05:309   kbdParams::SetEncodingTable:309 WaitForSingleObject start
	//		0820:0B1C 10:329   kbdParams::SetEncodingTable:311 WaitForSingleObject stop = 0x102
	//		0820:0B1C 10:329   kbdParams::SetEncodingTable:316 Timeput in wating semaphore
	//		0820:0B1C 10:341   kbdParams::ResetEncodingTable:295 ReleaseSemaphore 0, res = 1
	//		0820:0B1C 10:357   <Start> MessageBoxT( Error in setting encoding table )
	//
	//	The updated version does not change semaphore state at WM_INPUTLANGCHANGE using this flag.
	//	So the semaphore works for WM_SETFOCUS and WM_KILLFOCUS mesages only to sincronizing of different proceses
	//	as it was designed.
	//
	bool noSemaphore;
	bool exist;

	bool bUseSecureTyping;

	bool wasError;
	
	static const char *kbdParams::szSemaphoreName;

	kbdParams()
	{
		Init();
	}

	static kbdParams *Allocate()
	{
		HGLOBAL hgl = GlobalAlloc( GMEM_FIXED, sizeof( kbdParams ) );

		if ( hgl == NULL )
		{
			assert( false );
			return false;
		}

		kbdParams *kbd = (kbdParams *)hgl;

		kbd->Init();		

		return kbd;
	}


	void Init()
	{
		memset( this, 0, sizeof( *this ) );
	}

	
	static void Free( kbdParams *hgl )
	{
		GlobalFree( hgl );
	}
	
	static kbdParams *GetParams( HWND hWnd )
	{
//		assert( false );

		kbdParams *kbd = (kbdParams *)GetWindowLongPtrT( hWnd, kbdData );
		if ( kbd == NULL )	{	assert( false ); return FALSE; }
		return kbd;
	}

	static BOOL SetParams( HWND hWnd, kbdParams *kbd )
	{
		//if ( ( GetWindowLongPtr( hWnd, kbdData ) != NULL )
		//	 && ( kbd != NULL ) )
		//{
		//	assert( false );
		//}

		SetWindowLongPtrT( hWnd, kbdData, (LPARAM)kbd );

		return true;
	}

	BOOL GenerateEncodingTable ( BOOL UseSecureTyping )
	{
		memset( &keyState, 0, sizeof( KEY_STATE ) );

		bUseSecureTyping = ( UseSecureTyping == TRUE );

		if ( VkMessed != 0 )
		{
			ODSF( L"GenerateEncodingTable VkMessed (0x%X) is not NULL", (int)VkMessed );
			assert( false );
		}

		BOOL res = KBD_GenerateEncodingTable(
				&VkOriginal, &VkMessed,
				&ScodOriginal, &ScodMessed,	
				&TableSizeNL, &TableSizeSm, &TableSizeAltGr,
				&ShiftTable, &ShiftTableSize, 
				UseSecureTyping );

		if ( !res )
		{
			ODS_fls( _FormatT( L"Assertion! Error 0x%X", res ) );
			return false;
		}

		if ( noSemaphore )
			return res;

		if ( hSemaphore != NULL )
			CloseHandle( hSemaphore );

		hSemaphore = CreateSemaphoreA( NULL, 0, 100, szSemaphoreName );

		if ( hSemaphore == NULL )
		{
			assert( hSemaphore != NULL );
			ODS_fls( L"Assertion! Error in creating semaphore." );
			return false;
		}

		res = GetLastError();
		exist = ( res == ERROR_ALREADY_EXISTS );
		ODSF( L"CreateSemaphore = 0x%X, %d, %sexist", hSemaphore, res, exist ? L"" : L"NOT " );

		LONG prev = 0;
		if ( !exist )
		{
			ReleaseSemaphore( hSemaphore, 1, &prev );
			ODS_fls( _FormatT( L"ReleaseSemaphore %d", prev ) );
		}		

		return true;
	}

	BOOL ResetEncodingTable( BOOL UseSecureTyping, int &setQ )
	{
		memset( &keyState, 0, sizeof( KEY_STATE ) );
		
		keyState.wasDeadcharBefore = KBD_WasDeadChar;

		BOOL res = KBD_ResetEncodingTable( UseSecureTyping );
		ODS_fls( _FormatT( L"res 0x%X, %d/%d, Set count %d", res, keyState.wasDeadcharBefore, KBD_WasDeadChar, setQ ) );

		if ( res )
			setQ --;

		if ( noSemaphore )
			return res;

		LONG prev = 0;
		ReleaseSemaphore( hSemaphore, 1, &prev );
		ODS_fls( _FormatT( L"ReleaseSemaphore %04X %d, res = %d", hSemaphore, (int)prev, res ) );
		return res;
	}

	
	BOOL SetEncodingTable( BOOL UseSecureTyping, int &setQ )
	{
		ODSFuncS( _FormatT( L"%d, %d", UseSecureTyping, (int)noSemaphore ) );
		DWORD dwRes;

		memset( &keyState, 0, sizeof( KEY_STATE ) );
		
		keyState.wasDeadcharBefore = KBD_WasDeadChar;

		if ( !noSemaphore )
		{
			// Tries to get semaphore state
			ODS_fls( _FormatT( L"WaitForSingleObject start %04X", hSemaphore ) );
			dwRes = WaitForSingleObject( hSemaphore, 5000 );

			if ( dwRes == WAIT_FAILED  )
				ODS_fls( _FormatT( L"WaitForSingleObject Wait Failed stop = %s", cWCStr().GetSystemError() ) );
			else
				ODS_fls( _FormatT( L"WaitForSingleObject stop = 0x%X", dwRes ) );

			if ( dwRes != WAIT_OBJECT_0 )
			{
				if ( dwRes == WAIT_TIMEOUT )
					ODS_fls( L"Timeput in wating semaphore" );
				else
					ODS_fls( _FormatT( L"Wait semaphore rrror - 0x%X", dwRes ) );

				return false;
			}
		}

		int res;
		res = KBD_SetEncodingTable( ScodOriginal, ScodMessed, TableSizeNL, TableSizeSm, TableSizeAltGr, ShiftTable, ShiftTableSize, UseSecureTyping );
		ODS_fls( _FormatT( L"res 0x%X, %d/%d, Set count %d", res, keyState.wasDeadcharBefore, KBD_WasDeadChar, setQ ) );

		if ( res )
			setQ++;

		assert( res == 1 );

		//
		//	The res == 2 means that STATUS_ERROR_ENCODING_ALREADY_SET table is currently set
		//
		//	Note, the code does not work for usual situation - semaphore ( see description at hSemaphore )
		//	properly waits permition for new table setting. But if a password dialog has been acsidentally
		//	terminated the following code allows to reset old invalid table and assing the new one.
		//
		if ( res == 2 )
		{
			KBD_ResetEncodingTable( UseSecureTyping );
			res = KBD_SetEncodingTable( ScodOriginal, ScodMessed, TableSizeNL, TableSizeSm, TableSizeAltGr, ShiftTable, ShiftTableSize, UseSecureTyping );
			ODSF( L"SetEncodingTable = 0x%X - second try", res );
		}

		return res == 1;
	}


	BOOL FreeTables( BOOL UseSecureTyping )
	{
		if ( !noSemaphore && ( hSemaphore != NULL ) )
		{
			ODS_fls( _FormatT( L"Close semaphore %04X!", hSemaphore ) );
			CloseHandle( hSemaphore );
			hSemaphore = NULL;
		}

		BOOL res = KBD_FreeTables( &VkOriginal, &VkMessed, &ScodOriginal, &ScodMessed, &ShiftTable, ShiftTableSize, UseSecureTyping );
		return res;
	}


	BOOL Wm_KeyDown(
				UINT		VirtualKey,
				BOOL		UseSecureTyping,
				UINT		ScanCode )
	{
		keyState.saved_vk = VirtualKey;
		keyState.saved_sc = ScanCode;
		return TRUE;
	}


	BOOL Wm_DeadChar()
	{
		KBD_WasDeadChar = TRUE;

		keyState.wasDeadcharBefore = TRUE;
		return TRUE;
	}


	BOOL ReactOnDeadchar(
				UINT		*VirtualKey,
				BOOL		UseSecureTyping,
				UINT		*ScanCode )
	{
		return KBD_ReactOnDeadchar( &keyState, VirtualKey,
						&VkOriginal, &VkMessed,	&ScodOriginal, &ScodMessed, TableSizeNL, TableSizeSm, TableSizeAltGr,
						&ShiftTable, ShiftTableSize, UseSecureTyping, ScanCode );
	}


	BOOL Wm_Char(char   OriginalChar,
				 BOOL   UseSecureTyping,
				 DWORD  *ReturnCharsNumber,
				 char   *ReturnChar1,
				 char   *ReturnChar2,
				 wchar_t *wcs = NULL,
				 WPARAM *pWP = NULL,
				 LPARAM *pLP = NULL
				 )
	{
		return KBD_WM_CHAR( OriginalChar, &keyState, &VkOriginal, &VkMessed,
							&ScodOriginal, &ScodMessed, TableSizeNL, TableSizeSm, TableSizeAltGr,
							&ShiftTable, ShiftTableSize, 
							UseSecureTyping, ReturnCharsNumber, ReturnChar1, ReturnChar2, wcs, pWP, pLP );
	}


	void IncrementCharacterOrder(
				int			VirtualKey, 
				BOOL		UseSecureTyping
		)
	{
		if ( UseSecureTyping && ( VkMessed == NULL ) )
		{
			assert( false );
			return;
		}

		KBD_IncrementCharacterOrder( VirtualKey, VkMessed, TableSizeNL, TableSizeSm, TableSizeAltGr, UseSecureTyping );
	}

	void IncrementCharacterOrderCNTRL(
				int			VirtualKey,
				BOOL		UseSecureTyping
		)
	{
		if ( UseSecureTyping && ( VkMessed == NULL ) )
		{
			assert( false );
			return;
		}

		KBD_IncrementCharacterOrderCNTRL( VirtualKey, VkMessed, TableSizeNL, TableSizeSm, UseSecureTyping );
	}

#ifdef _DEBUG_PRINT
	void DbgPrint( const char *msg, int res = 0 )
	{
		char str[ 1024 ];
		sprintf( str, "\n%08d : %08X - %s, res %d", GetTickCount(), DWORD(VkMessed ? VkMessed[0]: 0 ), msg, res ); 

		HANDLE hFile = CreateFile( "C:\\dbg.txt", GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ, NULL, OPEN_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL );
		if ( hFile == INVALID_HANDLE_VALUE )
		{
			MessageBox( 0, "Error in opening file", "DbgPrint", MB_ICONSTOP );
			HANDLE hFile = CreateFile( "C:\\dbg.txt", GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ, NULL, OPEN_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL );
			if ( hFile == INVALID_HANDLE_VALUE )
				return;
		}

		DWORD pos = SetFilePointer( hFile, 0, 0, FILE_END );

		DWORD dwWritten = 0;
		WriteFile( hFile, (BYTE *)str, strlen( str ), &dwWritten, NULL );

		CloseHandle( hFile );
	}
#endif
};

//
//	V. 1.03.3 uses ToUnicodeEx in both cases secure typing or not.
//
#ifdef UNICODE
#define KBD_CallWindowProc CallWindowProcW
#define KBD_SetWindowProc( __hEdit, __wndProc ) (LONG_PTR)SetWindowLongPtrW( __hEdit, GWLP_WNDPROC, (LONG_PTR)__wndProc )
#else
#define KBD_CallWindowProc CallWindowProcA
#define KBD_SetWindowProc( __hEdit, __wndProc ) (LONG_PTR)SetWindowLongPtrA( __hEdit, GWLP_WNDPROC, (LONG_PTR)__wndProc )
#endif // UNICODE

extern bool gOldModeCompatible;

#endif /* __KBD */
