#include "commonFuncs.h"
#include <process.h>

const unsigned int rgERROR = (unsigned int)(-1);

const wchar_t *szSharedDllSection =
	L"SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\SharedDLLs";

int GetRegCount( const wchar_t *fileName )
{
	HKEY hKey;

	if (RegOpenKeyExT(
			HKEY_LOCAL_MACHINE,
			szSharedDllSection,
			0,
			KEY_READ,	// security access mask 
			&hKey			// address of handle of open key 
		) != ERROR_SUCCESS)
		return 0; //rgERROR;

	wchar_t buff[MAX_PATH];
	DWORD buffSize = maxLen(buff);

//	FILETIME ftLastWriteTime;

	DWORD regCount = 0;
	DWORD regCountSize = sizeof(DWORD);
	DWORD dwType = REG_DWORD;

	DWORD Data;
	DWORD cbData = sizeof(Data);

	LONG res;

	for(DWORD dwIndex = 0; ; dwIndex++)
	{
		buffSize = maxLen(buff);
  		res = RegEnumValueT(
				hKey,	// handle of key to enumerate 
				dwIndex,	// index of subkey to enumerate 
				buff,	// address of buffer for subkey name 
				&buffSize,	// address for size of subkey buffer 
				NULL,	// reserved 
				&dwType,
				(LPBYTE)&Data,	// address of buffer for value data 
				&cbData 	// address for size of data buffer 
 			);
		if (res == ERROR_NO_MORE_ITEMS) break;

		if (res != ERROR_SUCCESS) continue; //return rgERROR;

		if (_wcsicmp(fileName,buff) == 0)
		{	regCount = Data;
			break;
		}
	}
	
	if (RegCloseKey(hKey) != ERROR_SUCCESS)
		return rgERROR;

	return regCount;
}

int SetRegCount( const wchar_t *fileName, const DWORD count )
{
	ODSFuncS( _FormatT( L"%s, %d", fileName, count ) );

	HKEY hKey;
	DWORD dwDisposition;

	if (RegCreateKeyExT(
			HKEY_LOCAL_MACHINE,
			szSharedDllSection,
			0,
			NULL,
			REG_OPTION_NON_VOLATILE,
			KEY_WRITE,
			NULL,
			&hKey,
			&dwDisposition
		) != ERROR_SUCCESS)
		return rgERROR;

	if ((int)count <= 0)
	{	if (RegDeleteValueT(
				hKey,
				fileName
			) != ERROR_SUCCESS)
			return rgERROR;
	}
	else
	{	if (RegSetValueExT(
				hKey,
				fileName,
				0,
				REG_DWORD,
				(const BYTE *)&count,
				sizeof(DWORD)
			) != ERROR_SUCCESS)
			return rgERROR;
	}

	if ( RegCloseKey(hKey) != ERROR_SUCCESS )
		return rgERROR;

	return count;
}

int hKey2int(HKEY hKey)
{
	if (hKey == HKEY_CLASSES_ROOT)	return 0;
	if (hKey == HKEY_CURRENT_USER)	return 1;
	if (hKey == HKEY_LOCAL_MACHINE)	return 2;
	if (hKey == HKEY_USERS)			return 3;
	return -1;
}

const wchar_t *hKey2str(HKEY hKey)
{
	if (hKey == HKEY_CLASSES_ROOT)	return L"HKEY_CLASSES_ROOT";
	if (hKey == HKEY_CURRENT_USER)	return L"HKEY_CURRENT_USER";
	if (hKey == HKEY_LOCAL_MACHINE)	return L"HKEY_LOCAL_MACHINE";
	if (hKey == HKEY_USERS)			return L"HKEY_USERS";
	return L"Unknown";
}

HKEY int2hKey(int i)
{
	if (i == 0)	return HKEY_CLASSES_ROOT;
	if (i == 1)	return HKEY_CURRENT_USER;
	if (i == 2)	return HKEY_LOCAL_MACHINE;
	if (i == 3)	return HKEY_USERS;

	MessageBoxT( 0, L"Invalid hKey identificator", L"UNInstall error",MB_ICONSTOP | MB_TOPMOST );
	return HKEY_USERS;
}

BOOL GetUninstallString( const wchar_t *productName, wchar_t *value, int cbSize )
{
	//ODSFuncS( _FormatT( L"%s, %x, %d", productName, value, cbSize ) );

	wchar_t *subKey = L"UninstallString";

	HKEY hKey;

	if (ERROR_SUCCESS !=
			RegOpenKeyExT(
				HKEY_LOCAL_MACHINE,
				_FormatT( L"%s%s", L"SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\Uninstall\\", productName ), //section,
				0,
				KEY_QUERY_VALUE,
				&hKey))
	{
		//ODSFuncRes( false );
		return FALSE;
	}

	DWORD size;
	BOOL ok;
	DWORD type;

	ok = ( RegQueryValueExT(
				hKey,
				subKey,
				0,
				&type,
				NULL,
				&size)
			== ERROR_SUCCESS );

	//ODSF( L"OK = %d, type = %X, size %d", ok, type, size );

	if (ok && ((type != REG_SZ) || ( size > (DWORD)cbSize ) ) )
	{
		ODSF( L"Invalid type" );
		ok = FALSE;
	}

	if (ok)
	{
		ok = ( RegQueryValueExT(
					hKey,
					subKey,
					0,
					&type,
					(BYTE *)value,
					&size)
				== ERROR_SUCCESS );

		//ODSF( L"OK = %d, type = %X, size %d", ok, type, size );
	}

	RegCloseKey(hKey);

	if ( ok )
	{
		// ODSF( L"value %s", value );
	}

	// ODSFuncRes( ok );

	return (ok);
}