//
// Copyright (c) 2006, Jetico, Inc., Finland. All rights reserved.
//
// Implementation of XTS encryption mode for Little Endian Platform (Intel)
//

#include "xts_mode.h"

#if defined(__cplusplus)
extern "C"
{
#endif

#ifndef RECOVERY16

//
// For 32/64 bit compiler
//

int XTS_EncryptDecrypt( ProcEncryptECBmode16ByteBlockAlgorithm *EncryptProc,
                        ProcEncryptECBmode16ByteBlockAlgorithm *EncryptOrDecryptProc,
                        ULONGLONG                              SectorNumber, // or IV - Initial Vector
                        unsigned char                          *InBuffer,
                        unsigned char                          *OutBuffer,
				        unsigned long                          BufferLength,
                        unsigned char                          *Context_1,
                        unsigned char                          *Context_2 )
{
	ULONGLONG     alpha_J_array[ SECTOR_SIZE / sizeof(ULONGLONG) ];
	ULONGLONG     ulSectorCounter, ulBlockCounter, ulBlockMax;
	ULONGLONG     i, ul0, ul1, ulSpecial;
	ULONGLONG     ulPP[ ALGORITHM_BLOCK_SIZE / sizeof(ULONGLONG) ];
	ULONGLONG     *ulInBuffer, *ulOutBuffer;

	ulInBuffer  = (ULONGLONG *)InBuffer;
	ulOutBuffer = (ULONGLONG *)OutBuffer;

	ulBlockCounter = BufferLength >> ALGORITHM_BLOCK_SIZE_POWER;

	for (ulSectorCounter = 0; ulSectorCounter < ((BufferLength >> SECTOR_SIZE_POWER) + 1); ulSectorCounter++)
	{
		alpha_J_array[0] = SectorNumber + ulSectorCounter;
		alpha_J_array[1] = 0;

		EncryptProc( (unsigned char *)alpha_J_array, (unsigned char *)alpha_J_array, Context_2 );
		
		for (i = 2; i < (SECTOR_SIZE / ALGORITHM_BLOCK_SIZE) * 2; i = i + 2 )
		{
			ul0 = alpha_J_array[ i - 2 ];
			ul1 = alpha_J_array[ i - 1 ];

			ulSpecial = 0;

			if (ul1 & 0x8000000000000000)
			{
				ulSpecial = 135;
			}

			ul1 <<= 1;

			if (ul0 & 0x8000000000000000)
			{
				ul1 |= 1;
			}

			ul0 <<= 1;

			ul0 ^= ulSpecial;

			alpha_J_array[i]   = ul0;
			alpha_J_array[i+1] = ul1;
		}

		if (ulBlockCounter < (SECTOR_SIZE / ALGORITHM_BLOCK_SIZE))
		{
			ulBlockMax = ulBlockCounter;
		}
		else
		{
			ulBlockMax = SECTOR_SIZE / ALGORITHM_BLOCK_SIZE;
		}

		for ( i = 0; i < (ulBlockMax << 1); i += 2 )
		{
			ulPP[0] = alpha_J_array[i]   ^ (*ulInBuffer++);
			ulPP[1] = alpha_J_array[i+1] ^ (*ulInBuffer++);

			EncryptOrDecryptProc( (unsigned char *)ulPP, (unsigned char *)ulPP, Context_1 );

			*ulOutBuffer++ = alpha_J_array[i]   ^ ulPP[0];
			*ulOutBuffer++ = alpha_J_array[i+1] ^ ulPP[1];

			ulBlockCounter--;
		}
	}

	return 0;
}


#else
//
// For 16 bit compiler
//

#define SECTOR_SIZE          512
#define SECTOR_SIZE_POWER    9

#define ALGORITHM_BLOCK_SIZE       16
#define ALGORITHM_BLOCK_SIZE_POWER 4

#define SHIFT_LEFT_ELEM_ALPHA_J(n) \
			alpha_J[n] <<= 1; \
			if (alpha_J[n-1] & 0x80000000) \
				alpha_J[n] |= 1;

int XTS_EncryptDecrypt( ProcEncryptECBmode16ByteBlockAlgorithm *EncryptProc,
                        ProcEncryptECBmode16ByteBlockAlgorithm *EncryptOrDecryptProc,
                        unsigned long                          *SectorNumber, // 2 32-bit numbers or IV - Initial Vector
                        unsigned char                          *InBuffer,
                        unsigned char                          *OutBuffer,
                        unsigned long                          BufferLength,
                        unsigned char                          *Context_1,
                        unsigned char                          *Context_2 )
{
	unsigned long     ivLocal[4], alpha_J[4], usPP[4], *in_dw, *out_dw, ulSpecial;
	unsigned short    usSectorCounter, usBlockCounter, usBlockMax, i, usLength;
		
	in_dw  = (unsigned long *)InBuffer;
	out_dw = (unsigned long *)OutBuffer;

	usLength = (unsigned short)BufferLength;
	
	// we get IV as sector number, will just transform it to 32-bit dwords
	ivLocal[0] = SectorNumber[0];
	ivLocal[1] = SectorNumber[1];
	ivLocal[2] = ivLocal[3] = 0;
					
	// Run XTS encryption

	usBlockCounter = usLength >> ALGORITHM_BLOCK_SIZE_POWER;
	
	for (usSectorCounter = 0; usSectorCounter < ((usLength >> SECTOR_SIZE_POWER) + 1); usSectorCounter++)
	{
		EncryptProc( (unsigned char *)ivLocal, (unsigned char *)alpha_J, Context_2 );

		if (usBlockCounter < (SECTOR_SIZE / ALGORITHM_BLOCK_SIZE))
		{
			usBlockMax = usBlockCounter;
		}
		else
		{
			usBlockMax = SECTOR_SIZE / ALGORITHM_BLOCK_SIZE;
		}
		
		for ( i = 0; i < (usBlockMax << 1); i += 2 )
		{
			usPP[0] = alpha_J[0] ^ (*in_dw++);
			usPP[1] = alpha_J[1] ^ (*in_dw++);
			usPP[2] = alpha_J[2] ^ (*in_dw++);
			usPP[3] = alpha_J[3] ^ (*in_dw++);

			EncryptOrDecryptProc( (unsigned char *)usPP, (unsigned char *)usPP, Context_1 );
			
			*out_dw++ = alpha_J[0] ^ usPP[0];
			*out_dw++ = alpha_J[1] ^ usPP[1];
			*out_dw++ = alpha_J[2] ^ usPP[2];
			*out_dw++ = alpha_J[3] ^ usPP[3];
			
			// Calculate next alpha_J
			ulSpecial = 0;
			
			if (alpha_J[3] & 0x80000000)
			{
				ulSpecial = 135;
			}

			SHIFT_LEFT_ELEM_ALPHA_J(3)
			SHIFT_LEFT_ELEM_ALPHA_J(2)
			SHIFT_LEFT_ELEM_ALPHA_J(1)
			alpha_J[0] <<= 1;
			
			alpha_J[0] ^= ulSpecial;	

			usBlockCounter--;
		}
		
		// Increment ivLocal
		if (ivLocal[0] == 0xFFFFFFFF)
		{
			ivLocal[0] = 0;
			ivLocal[1] += 1;
		}
		else
		{
			ivLocal[0] += 1;
		}
	}

	return 0;
}


#endif // RECOVERY16

#if defined(__cplusplus)
}
#endif

