#include <windows.h>

#ifndef __WIN95__
#include <windef.h>
#else
#include <basedef.h>
#endif

#include <stdio.h>
#include <string.h>
#include "..\bc_serp.h"

// Serpent test vectors from the AES submission. Only test vectors for 256-bits keys are included

struct
{
	unsigned char key[32];
	unsigned char plain[16];
	unsigned char cipher[16];
	int length;
} testVector =
{
	{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
	{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
	{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 }, 16
};

void stringToBinary(BYTE dest[], BYTE src[], int length)
{
	int i;
	for (i = 0; i < length; i+=2, src-=2, dest++)
	{
		if (src[0] > '9') dest[0] = src[0] - 'a' + 10; else dest[0] = src[0] - '0';
		dest[0] = dest[0] << 4;
		if (src[1] > '9') dest[0] += src[1] - 'a' + 10; else dest[0] += src[1] - '0';
	}
}

int main() {
	int i,j;
	FILE *fd;
	BYTE buffer[80];
	DWORD outputData[4];
	DWORD extendedKey[KEY_LENGTH_EXTENDED / 32];

	// Simple Encryption/Decryption Test

	fd = fopen("vectors.txt", "r");
	if (fd == 0)
	{
		printf("\n(ERROR) Could not open file with test vectors!\n");
		fclose(fd);
		return 0;
	}

	i = 0;
	while (!feof(fd))
	{
		// Read the next test vector from file
		fgets(buffer, 80, fd);
		stringToBinary(testVector.key, buffer+4+62, 64);

		fgets(buffer, 80, fd);
		stringToBinary(testVector.plain, buffer+3+30, 32);

		fgets(buffer, 80, fd);
		stringToBinary(testVector.cipher, buffer+3+30, 32);

		// Skip the newline
		fgets(buffer, 80, fd);

		// Begin testing
		KeyExtend(testVector.key, extendedKey);

		ECB_Encrypt((DWORD *)testVector.plain, outputData, extendedKey);

		if (memcmp(outputData, testVector.cipher, testVector.length) != 0)
		{
			printf("\n(ERROR) Test Vector %d: The Serpent implementation failed the test (encryption)!\n",i);
			fclose(fd);
			return 0;
		}

		ECB_Decrypt((DWORD *)testVector.cipher, outputData, extendedKey);

		if (memcmp((DWORD *)testVector.plain, outputData, testVector.length) != 0)
		{
			printf("\n(ERROR) Test Vector %d: The Serpent implementation failed the test (decryption)!\n",i);
			fclose(fd);
			return 0;
		}

		printf("Test Vector %d: The Serpent implementation passes the test successfully.\n",i);

		i++;
	}
	fclose(fd);

	// Monte Carlo Encryption Test

	fd = fopen("vectors1.txt", "r");
	if (fd == 0)
	{
		printf("\n(ERROR) Could not open file with test vectors!\n");
		fclose(fd);
		return 0;
	}

	while (!feof(fd))
	{
		// Read the next test vector from file
		fgets(buffer, 80, fd);
		stringToBinary(testVector.key, buffer+4+62, 64);

		fgets(buffer, 80, fd);
		stringToBinary(testVector.plain, buffer+3+30, 32);

		fgets(buffer, 80, fd);
		stringToBinary(testVector.cipher, buffer+3+30, 32);

		// Skip the newline
		fgets(buffer, 80, fd);

		// Begin testing
		KeyExtend(testVector.key, extendedKey);

		ECB_Encrypt((DWORD *)testVector.plain, outputData, extendedKey);
		for (j = 0; j < 9999; j++)
		{
			ECB_Encrypt(outputData, outputData, extendedKey);
		}

		if (memcmp(outputData, testVector.cipher, testVector.length) != 0)
		{
			printf("\n(ERROR) Test Vector %d: The Serpent implementation failed the test (encryption)!\n",i);
			fclose(fd);
			return 0;
		}

		printf("Test Vector %d: The Serpent implementation passes the test successfully.\n",i);

		i++;
	}
	fclose(fd);

	// Monte Carlo Decryption Test

	fd = fopen("vectors2.txt", "r");
	if (fd == 0)
	{
		printf("\n(ERROR) Could not open file with test vectors!\n");
		fclose(fd);
		return 0;
	}

	while (!feof(fd))
	{
		// Read the next test vector from file
		fgets(buffer, 80, fd);
		stringToBinary(testVector.key, buffer+4+62, 64);

		fgets(buffer, 80, fd);
		stringToBinary(testVector.plain, buffer+3+30, 32);

		fgets(buffer, 80, fd);
		stringToBinary(testVector.cipher, buffer+3+30, 32);

		// Skip the newline
		fgets(buffer, 80, fd);

		// Begin testing
		KeyExtend(testVector.key, extendedKey);

		ECB_Decrypt((DWORD *)testVector.cipher, outputData, extendedKey);
		for (j = 0; j < 9999; j++)
		{
			ECB_Decrypt(outputData, outputData, extendedKey);
		}

		if (memcmp(outputData, testVector.plain, testVector.length) != 0)
		{
			printf("\n(ERROR) Test Vector %d: The Serpent implementation failed the test (decryption)!\n",i);
			fclose(fd);
			return 0;
		}

		printf("Test Vector %d: The Serpent implementation passes the test successfully.\n",i);

		i++;
	}
	fclose(fd);

	printf("\nThis implementation passes all test vectors!\n");

	return 0;
}
