#include <windows.h>

#ifndef __WIN95__
#include <windef.h>
#else
#include <basedef.h>
#endif

#include <stdio.h>
#include <string.h>
#include "..\bc_tfish.h"

/* TWOFISH test vector, from the ECB_IVAL.TXT file 
   available for downloading from the http://www.counterpane.com
 */
struct 
    {    UCHAR key[ KEY_LENGTH / 8 ];
         UCHAR plain[16];
         UCHAR cipher[16];
		 DWORD dataLength;
	} testVectors[] =
{
	{ { 0x01, 0x23, 0x45, 0x67, 0x89, 0xAB, 0xCD, 0xEF,
		0xFE, 0xDC, 0xBA, 0x98, 0x76, 0x54, 0x32, 0x10, 
		0x00, 0x11, 0x22, 0x33, 0x44, 0x55, 0x66, 0x77, 
		0x88, 0x99, 0xAA, 0xBB, 0xCC, 0xDD, 0xEE, 0xFF },
	  
	  { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },

	  { 0x37, 0x52, 0x7B, 0xE0, 0x05, 0x23, 0x34, 0xB8,
	    0x9F, 0x0C, 0xFC, 0xCA, 0xE8, 0x7C, 0xFA, 0x20 }, 16
	},
	{ { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
	  { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
	  { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
	    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 }, 0 
	}
};


int main()
{
  int   i;
  UCHAR outputData[16];
  BOOL  result;
  KeyInstance extendedKey;

  result = TRUE; // Assume everything will be Ok

  for (i=0; testVectors[i].dataLength != 0; i++)
  {
    if (!KeyExtend( testVectors[i].key, (DWORD *)(&extendedKey) ))
	{
		printf("\n(ERROR) Test Vector N %d: error to extend encryption key!\n",i);
		return 0;
    }		
	  
	memcpy( outputData, testVectors[i].plain, 16 );

	if (!blockEncrypt( (DWORD *)outputData, &extendedKey ))
	{
		printf("\n(ERROR) Test Vector N %d: internal error (encryption)!\n",i);
		return 0;
    }		

    if( memcmp( outputData, testVectors[i].cipher, testVectors[i].dataLength ) != 0)
	{
		printf("\n(ERROR) Test Vector N %d: the TWOFISH implementation failed the test (encryption)!\n",i);
		return 0;
	}

	if (!blockDecrypt( (DWORD *)outputData, &extendedKey ))
	{
		printf("\n(ERROR) Test Vector N %d: internal error (decryption)!\n",i);
		return 0;
    }		

    if( memcmp( outputData, testVectors[i].plain, testVectors[i].dataLength ) != 0)
	{
		printf("\n(ERROR) Test Vector N %d: the TWOFISH implementation failed the test (decryption)!\n",i);
		return 0;
	}

    printf("Test Vector N %2d: the TWOFISH implementation passes the test successfully.\n",i+1);
  }

  return 0;
}

